(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.0'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     27324,        851]*)
(*NotebookOutlinePosition[     58654,       1563]*)
(*  CellTagsIndexPosition[     56633,       1497]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["Introduction", "Section",
  CellTags->"b:3"],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< ProPac`\)], "Input",
  CellLabel->"In[1]:=",
  CellTags->"b:3"],

Cell[BoxData[
    \("\n    This is ProPac - Development Version 3.xx\n \n ProPac has been \
successfully loaded. It contains several packages\n for multibody dynamics, \
linear control and nonlinear control.\n The package names are:\n    GeoTools\n\
    Dynamics\n    ControlL\n    ControlN\n    DifferentialForms\n    \
ObserverTools\n    NDTools\n    VSCTools\n    MatlabInterface\n    MEXTools\n \
Appropriate packages will be loaded automatically as required.\n To obtain a \
list of functions in a package and links to function\n declarations, enter:\n \
   ?ProPac`PackageName`*\n Help is also available in the Help Browser under \
Add-ons."\)], "Print",
  CellLabel->"From In[1]:=",
  CellTags->"b:3"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(GeometricTools\)], "Input",
  CellLabel->"In[2]:=",
  CellTags->"b:3"],

Cell[BoxData[
    \(GeometricTools\)], "Output",
  CellLabel->"Out[2]=",
  CellTags->"b:3"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Dynamics", "Input",
  CellLabel->"In[3]:=",
  CellTags->"b:3"],

Cell[BoxData[
    \("  *** Dynamics successfully loaded ***"\)], "Print",
  CellLabel->"From In[3]:=",
  CellTags->"b:3"],

Cell[BoxData[
    \(Dynamics\)], "Output",
  CellLabel->"Out[3]=",
  CellTags->"b:3"]
}, Open  ]],

Cell[CellGroupData[{

Cell["?Dynamics", "Input",
  CellLabel->"In[4]:=",
  CellTags->"b:3"],

Cell[BoxData[
    \("Functions included in the Dynamics package are:\n \n Kinematics:\n \
RotationMatrixEuler, ConfigurationMatrixEuler,\n \
AToATilda,ATildaToA,SimpleJointKinematics,\n \
CompoundJointKinematics,CompoundJointMap,Joints,\n \
JointRotation,JointTranslation,CompoundJointConfiguration,\n \
SimpleJointConfiguration,RotationMatrixToEuler,\n \
EndEffector,RelativeConfiguration,\n NodeVelocity,KinematicReplacements,\n \n \
Generalized Forces & Potential Functions:\n \
GeneralizedForce,DamperForce,SpringForce,GravPotential,\n \
SpringPotential,DamperPotential,LeafPotential,\n \
RayleighDissipationForce,BacklashPotential,\n \
BacklashForce,JointFrictionPotential,\n FlxDissPot,FlxStrnPot,\n \n Inertial \
Parameters: RgdBdyInrShift,ChainInertia,TreeInertia,\n \n Model Assembly:\n \
CMatrix,CreateModelSpecial,CreateModel,\n \
PoincareCoefficients,PoincareFunctionCombined,PoincareFunction,\n \
AlgebraicConstraints,DifferentialConstraints,\n MakeODEs,StateTransformation, \
MakeLagrangeEquations,\n CreateModelMEX,FindMostFrequentTerms,\n \n \
Miscellaneous:\n SeriesExpansion,Truncate,MatlabForm\n "\)], "Print",
  CellTags->{"Info3268728757-3969975", "b:3"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Differential Geometry ", "Section"],

Cell[CellGroupData[{

Cell["Grad", "Subsection",
  CellTags->{"Grad", "i:1", "b:4.0"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?Grad\)\)], "Input",
  CellLabel->"In[5]:=",
  CellTags->{"Grad", "i:1", "b:4.0"}],

Cell[BoxData[
    \("Grad[f,varlist] computes the gradient of the scalar function f with  \
respect to the variables varlist."\)], "Print",
  CellTags->{"Info3268728757-8371087", "b:4.0"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Grad[Sin[x]\ y + y^2, {x, y}]\)], "Input",
  CellLabel->"In[6]:=",
  CellTags->{"Grad", "i:1", "b:4.0"}],

Cell[BoxData[
    \({y\ Cos[x], 2\ y + Sin[x]}\)], "Output",
  CellLabel->"Out[6]=",
  CellTags->{"Grad", "i:1", "b:4.0"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Jacob", "Subsection",
  CellTags->{"Jacob", "i:2", "b:4.1"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?Jacob\)\)], "Input",
  CellLabel->"In[7]:=",
  CellTags->{"Jacob", "i:2", "b:4.1"}],

Cell[BoxData[
    \("Jacob[f,x] computes the Jacobian of the list of functions f with\n  \
respect to the list of variables x.\n \n Example: Jacob[{x y, Sin[x^2 +x y]}, \
{x,y}]"\)], "Print",
  CellTags->{"Info3268728757-7417227", "b:4.1"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Jacob[{x\ y, Sin[x^2 + x\ y]}, {x, y}]\)], "Input",
  CellLabel->"In[8]:=",
  CellTags->{"Jacob", "i:2", "b:4.1"}],

Cell[BoxData[
    \({{y, x}, {\((2\ x + y)\)\ Cos[x\^2 + x\ y], 
        x\ Cos[x\^2 + x\ y]}}\)], "Output",
  CellLabel->"Out[8]=",
  CellTags->{"Jacob", "i:2", "b:4.1"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["LieBracket & Ad", "Subsection",
  CellTags->{"LieBracket", "i:3", "b:4.2"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?LieBracket\)\)], "Input",
  CellLabel->"In[9]:=",
  CellTags->{"LieBracket", "i:3", "b:4.2"}],

Cell[BoxData[
    \("LieBracket[f,g,var] computes the Lie Bracket of the vector\n functions \
f,g with respect to the vector of variables var.\n \n Example: \
LieBracket[{x,Sin[x^2 +x y]},{y,Cos[x y]},{x,y}]"\)], "Print",
  CellTags->{"Info3268728757-4720349", "b:4.2"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(LieBracket[{x, Sin[x^2\  + x\ y]}, {y, Cos[x\ y]}, {x, y}]\)], "Input",
  CellLabel->"In[10]:=",
  CellTags->{"LieBracket", "i:3", "b:4.2"}],

Cell[BoxData[
    \({\(-y\) + Sin[x\^2 + x\ y], \(-y\)\ \((2\ x + y)\)\ Cos[x\^2 + x\ y] - 
        x\ Cos[x\ y]\ Cos[x\^2 + x\ y] - x\ y\ Sin[x\ y] - 
        x\ Sin[x\ y]\ Sin[x\^2 + x\ y]}\)], "Output",
  CellLabel->"Out[10]=",
  CellTags->{"LieBracket", "i:3", "b:4.2"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?Ad\)\)], "Input",
  CellLabel->"In[11]:=",
  CellTags->{"LieBracket", "i:3", "b:4.2"}],

Cell[BoxData[
    \("Ad[f,g,var,n] computes the nth Adjoint (iterated Lie Bracket) of\n the \
vector fields f,g with respect to the vector of variables var.\n         \
Ad[f,g,var,0]=g\n         Ad[f,g,var,n]=LieBracket[f,Ad[f,g,var,n-1],var]\n   \
      Ad[f,g,var]=Ad[f,g,var,1] "\)], "Print",
  CellTags->{"Info3268728757-8448341", "b:4.2"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Ad[{x, Sin[x^2\  + x\ y]}, {y, Cos[x\ y]}, {x, y}, 2]\)], "Input",
  CellLabel->"In[12]:=",
  CellTags->{"LieBracket", "i:3", "b:4.2"}],

Cell[BoxData[
    \({y + x\ \((2\ x + y)\)\ Cos[x\^2 + x\ y] - 
        Sin[x\^2 + x\ y] + \((\(-1\) + x\ Cos[x\^2 + x\ y])\)\ Sin[
            x\^2 + x\ y], \(-\((2\ x + y)\)\)\ Cos[
            x\^2 + x\ y]\ \((\(-y\) + Sin[x\^2 + x\ y])\) + 
        Sin[x\^2 + x\ y]\ \((\(-x\^2\)\ y\ Cos[x\ y] - 
              y\ Cos[x\^2 + x\ y] - \((2\ x + y)\)\ Cos[x\^2 + x\ y] - 
              x\ Sin[x\ y] + x\ y\ \((2\ x + y)\)\ Sin[x\^2 + x\ y])\) + 
        x\ \((\(-x\)\ y\^2\ Cos[x\ y] - 2\ y\ Cos[x\^2 + x\ y] - 
              Cos[x\ y]\ Cos[x\^2 + x\ y] - y\ Sin[x\ y] + 
              x\ y\ Cos[x\^2 + x\ y]\ Sin[x\ y] - 
              x\ \((2\ x + y)\)\ Cos[x\^2 + x\ y]\ Sin[
                  x\ y] - \((\(-2\)\ x - y)\)\ y\ \((2\ x + y)\)\ Sin[
                  x\^2 + x\ y] - 
              x\ \((\(-2\)\ x - y)\)\ Cos[x\ y]\ Sin[x\^2 + x\ y] - 
              x\ y\ Cos[x\ y]\ Sin[x\^2 + x\ y] - 
              Sin[x\ y]\ Sin[x\^2 + x\ y])\) - 
        x\ Cos[x\^2 + x\ y]\ \((\(-y\)\ \((2\ x + y)\)\ Cos[x\^2 + x\ y] - 
              x\ Cos[x\ y]\ Cos[x\^2 + x\ y] - x\ y\ Sin[x\ y] - 
              x\ Sin[x\ y]\ Sin[x\^2 + x\ y])\)}\)], "Output",
  CellLabel->"Out[12]=",
  CellTags->{"LieBracket", "i:3", "b:4.2"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["LieDerivative", "Subsection",
  CellTags->{"LieDerivative", "i:4", "b:4.3"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?LieDerivative\)\)], "Input",
  CellLabel->"In[13]:=",
  CellTags->{"LieDerivative", "i:4", "b:4.3"}],

Cell[BoxData[
    \("LieDerivative[f,h,x] computes the Lie derivative of the real valued\n \
function h of the vector x along the direction defined by the vector\n field \
f:\n   LDf[h](x) = < dh(x),f(x) >\n \n LieDerivative[f,h,x,k] computes the \
k-th order Lie derivative of the\n real valued function h of the vector x \
along the direction defined by\n  the vector field f:\n   LD^k f[h](x) = \
LD^(k-1) f[LDf[h]](x)\n   LD^0 f[h](x) = h(x)  \n LieDerivative[f,w,x], where \
w is a one-form computes the Lie derivative\n   LDf[w](x)=(dw/dx).f+w.(df/dx)\
\n "\)], "Print",
  CellTags->{"Info3268728758-4755545", "b:4.3"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(LieDerivative[{Cos[x^2\ y], x + y^2}, {x^2, x\ y + y}, {x, 
        y}]\)], "Input",
  CellLabel->"In[14]:=",
  CellTags->{"LieDerivative", "i:4", "b:4.3"}],

Cell[BoxData[
    \({2\ x\ Cos[x\^2\ y], \((1 + x)\)\ \((x + y\^2)\) + 
        y\ Cos[x\^2\ y]}\)], "Output",
  CellLabel->"Out[14]=",
  CellTags->{"LieDerivative", "i:4", "b:4.3"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(LieDerivative[{Cos[x^2\ y], x + y^2}, {x^2, x\ y + y}, {x, y}, 
      2]\)], "Input",
  CellLabel->"In[15]:=",
  CellTags->{"LieDerivative", "i:4", "b:4.3"}],

Cell[BoxData[
    \({\(-2\)\ x\^3\ \((x + y\^2)\)\ Sin[x\^2\ y] + 
        Cos[x\^2\ y]\ \((2\ Cos[x\^2\ y] - 
              4\ x\^2\ y\ Sin[x\^2\ y])\), \((x + 
              y\^2)\)\ \((2\ \((1 + x)\)\ y + Cos[x\^2\ y] - 
              x\^2\ y\ Sin[x\^2\ y])\) + 
        Cos[x\^2\ y]\ \((1 + 2\ x + y\^2 - 
              2\ x\ y\^2\ Sin[x\^2\ y])\)}\)], "Output",
  CellLabel->"Out[15]=",
  CellTags->{"LieDerivative", "i:4", "b:4.3"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Involutive", "Subsection",
  CellTags->{"Involutive", "b:4.4"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?Involutive\)\)], "Input",
  CellLabel->"In[16]:=",
  CellTags->{"Involutive", "b:4.4"}],

Cell[BoxData[
    \("Involutive[X,x] tests a set X of vector fields to determine if it is\n \
involutive. It reports the result as a value True or False. The argument\n  X \
must be a list of vector fields, and x must be a list of the\n \
variables."\)], "Print",
  CellTags->{"Info3268728758-2063324", "b:4.4"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Involutive[{{Sin[x], y\ Cos[x]}, {Cos[y], Sin[x]}}, {x, y}]\)], "Input",\

  CellLabel->"In[17]:=",
  CellTags->{"Involutive", "b:4.4"}],

Cell[BoxData[
    \(True\)], "Output",
  CellLabel->"Out[17]=",
  CellTags->{"Involutive", "b:4.4"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Rank", "Subsection",
  CellTags->{"Rank", "i:5", "b:4.5"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?Rank\)\)], "Input",
  CellLabel->"In[18]:=",
  CellTags->{"Rank", "i:5", "b:4.5"}],

Cell[BoxData[
    \("Rank[X] computes the (generic) rank of a list of vector fields X.\n \
Rank[X,x,x0] returns the rank at a specified point x=x0. Note that\n \
Rank[X]/.x->x0 is the same as Rank[X,x,x0] only at regular points of X."\)], \
"Print",
  CellTags->{"Info3268728758-9909747", "b:4.5"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Rank[{{Sin[x], y\ Cos[x]}, {Cos[y], Sin[x]}}]\)], "Input",
  CellLabel->"In[19]:=",
  CellTags->{"Rank", "i:5", "b:4.5"}],

Cell[BoxData[
    \(2\)], "Output",
  CellLabel->"Out[19]=",
  CellTags->{"Rank", "i:5", "b:4.5"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Span", "Subsection",
  CellTags->{"Span", "i:6", "b:4.6"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?Span\)\)], "Input",
  CellLabel->"In[20]:=",
  CellTags->{"Span", "i:6", "b:4.6"}],

Cell[BoxData[
    \("Span[X] gives a basis set for the span of a list of vector fields,\n \
X={v1(x)...Vr(x)}. The result is the 'generic span', i.e., the result\n is \
valid at regular points of the distribution. \n Span[X,x,x0] gives a basis \
set for the span of a list of vector fields X\n evaluated at a specified \
x=x0. x0 may be a regular point or a singular point. "\)], "Print",
  CellTags->{"Info3268728758-4807394", "b:4.6"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Span[{{Sin[x], y\ Cos[x]}, {Cos[y], Sin[x]}}]\)], "Input",
  CellLabel->"In[21]:=",
  CellTags->{"Span", "i:6", "b:4.6"}],

Cell[BoxData[
    \({{1, 0}, {0, 1}}\)], "Output",
  CellLabel->"Out[21]=",
  CellTags->{"Span", "i:6", "b:4.6"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["SpecialNullSpace", "Subsection",
  CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?SpecialNullSpace\)\)], "Input",
  CellLabel->"In[22]:=",
  CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],

Cell[BoxData[
    \("SpecialNullSpace[A] gives a basis for the Null Space of A. It\n \
returns an analytic set of basis vector fields if A(x) is\n analytic."\)], \
"Print",
  CellTags->{"Info3268728758-2216006", "b:4.7"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(A = {{x\^3\ Sin[x], x^2\ Cos[x]}, {x\ Sin[x], Cos[x]}}\)], "Input",
  CellLabel->"In[23]:=",
  CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],

Cell[BoxData[
    \({{x\^3\ Sin[x], x\^2\ Cos[x]}, {x\ Sin[x], Cos[x]}}\)], "Output",
  CellLabel->"Out[23]=",
  CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(NullSpace[A]\)], "Input",
  CellLabel->"In[24]:=",
  CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],

Cell[BoxData[
    \({{\(-\(Cot[x]\/x\)\), 1}}\)], "Output",
  CellLabel->"Out[24]=",
  CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(SpecialNullSpace[A]\)], "Input",
  CellLabel->"In[25]:=",
  CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],

Cell[BoxData[
    \({{\(-Cot[x]\), x}}\)], "Output",
  CellLabel->"Out[25]=",
  CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Invariant Distributions", "Subsection",
  CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?LargestInvariantDistribution\)\)], "Input",
  CellLabel->"In[26]:=",
  CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],

Cell[BoxData[
    \("LargestInvariantDistribution[tau,h,var] returns the largest\n \
distribution invariant with respect to the set of vector fields\n tau and \
contained in the distribution annihilator of exact codistribution\n {dh}. The \
fields tau and the vector-valued function h\n are given in the local \
coordinates var. "\)], "Print",
  CellTags->{"Info3268728758-5746287", "b:4.8"}]
}, Open  ]],

Cell[BoxData[{
    \(\(f = {x1\ x3 + x2\ Exp[x2], x3, x4 - x2\ x3, 
          x3^2 + x2\ x4 - x2^2\ x3};\)\), "\n", 
    \(\(g = {x1, 1, 0, x3};\)\), "\n", 
    \(\(h = {x3};\)\), "\n", 
    \(\(var = {x1, x2, x3, x4};\)\)}], "Input",
  CellLabel->"In[27]:=",
  CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(LargestInvariantDistribution[{f, g}, h, var]\)], "Input",
  CellLabel->"In[31]:=",
  CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],

Cell[BoxData[
    \({{0, 1, 0, x3}, {1, 0, 0, 0}}\)], "Output",
  CellLabel->"Out[31]=",
  CellTags->{"Invariant Distributions", "i:8", "b:4.8"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?SmallestInvariantDistribution\)\)], "Input",
  CellLabel->"In[32]:=",
  CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],

Cell[BoxData[
    \("SmallestInvariantDistribution[tau,del,var] returns the smallest\n \
distribution invariant with respect to the set of vector fields\n tau and \
containing the distribution del. The fields tau and the \n distribution del \
are given in the local coordinates var. "\)], "Print",
  CellTags->{"Info3268728758-4311524", "b:4.8"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(del = Span[{g}]\)], "Input",
  CellLabel->"In[33]:=",
  CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],

Cell[BoxData[
    \({{x1, 1, 0, x3}}\)], "Output",
  CellLabel->"Out[33]=",
  CellTags->{"Invariant Distributions", "i:8", "b:4.8"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(SmallestInvariantDistribution[{f, g}, del, var]\)], "Input",
  CellLabel->"In[34]:=",
  CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],

Cell[BoxData[
    \({{1, 0, 0, 0}, {0, 1, 0, x3}}\)], "Output",
  CellLabel->"Out[34]=",
  CellTags->{"Invariant Distributions", "i:8", "b:4.8"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Triangular Decomposition", "Subsection",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?TriangularDecomposition\)\)], "Input",
  CellLabel->"In[35]:=",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],

Cell[BoxData[
    \("TriangularDecomposition[f,Del,x,x0,m] returns {Psi,PsiInverse,fnew},\n \
given an involutive distribution Del that is invariant with respect to the\n \
vector field f - both described in terms of local coordinates x, around a \n \
point x0. Psi and PsiInverse define a transformation to new coordinates z, \
x=Psi(z)\n alligned with/and transverse to the integral manifolds of Del. It \
places f(x) in a\n triangular form given by fnew(z). The transfrom x=Psi(z) \
and its inverse z=PsiInverse(x)\n are obtained as a power series of order m. \
m can be set to Infinity to\n obtain an exact result if it is computable. \
"\)], "Print",
  CellTags->{"Info3268728758-5559249", "b:4.9"}],

Cell[BoxData[
    \("TriangularDecomposition[f,Del,x,x0,m] returns {Psi,PsiInverse,fnew},\n\
given an involutive distribution Del that is invariant with respect to the\n\
vector field f - both described in terms of local coordinates x, around a \n\
point x0. Psi and PsiInverse define transformation to new coordinates z, \
x=Psi(z)\nalligned with/and transverse to the integral manifolds of Del. It \
places f(x) in a\ntriangular form given by fnew(z). \nThe transfrom x=Psi(z) \
and its inverse z=PsiInverse(x)\nare obtained as a power series of order m. m \
can be set to Infinity to obtain an exact result if it is computable. "\)], \
"Print",
  GeneratedCell->False,
  CellAutoOverwrite->False,
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(f = {x1\ x3 + x2\ Exp[x2], x3, x4 - x2\ x3, 
          x3^2 + x2\ x4 - x2^2\ x3};\)\), "\n", 
    \(\(g = {x1, 1, 0, x3};\)\), "\n", 
    \(\(var = {x1, x2, x3, x4};\)\), "\n", 
    \(del = Span[{g}]\)}], "Input",
  CellLabel->"In[36]:=",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],

Cell[BoxData[
    \({{x1, 1, 0, x3}}\)], "Output",
  CellLabel->"Out[39]=",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Del0 = SmallestInvariantDistribution[{f, g}, del, var]\)], "Input",
  CellLabel->"In[40]:=",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],

Cell[BoxData[
    \({{1, 0, 0, 0}, {0, 1, 0, x3}}\)], "Output",
  CellLabel->"Out[40]=",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Del = Join[Del0, {{0, 0, 1, 0}, {0, 0, 0, 1}}]\)], "Input",
  CellLabel->"In[41]:=",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],

Cell[BoxData[
    RowBox[{\(General::"spell1"\), \(\(:\)\(\ \)\), "\<\"Possible spelling \
error: new symbol name \\\"\\!\\(Del\\)\\\" is similar to existing symbol \
\\\"\\!\\(del\\)\\\". \\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", \
ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"General::spell1\\\"]\\)\"\>"}]], "Message",
  CellLabel->"From In[41]:=",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],

Cell[BoxData[
    \({{1, 0, 0, 0}, {0, 1, 0, x3}, {0, 0, 1, 0}, {0, 0, 0, 1}}\)], "Output",
  CellLabel->"Out[41]=",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(TriangularDecomposition[f + g*u, Del, var, {0, 0, 0, 0}, 
      Infinity]\)], "Input",
  CellLabel->"In[42]:=",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],

Cell[BoxData[
    \({{z1, z2, z3, z2\ z3 + z4}, {x1, x2, 
        x3, \(-x2\)\ x3 + x4}, {u\ z1 + \[ExponentialE]\^z2\ z2 + z1\ z3, 
        u + z3, z4, 0}}\)], "Output",
  CellLabel->"Out[42]=",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Working with Complex Arrays", "Section",
  CellTags->{"Complex Arrays", "b:5"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?SpecialInverse\)\)], "Input",
  CellLabel->"In[43]:=",
  CellTags->{"Complex Arrays", "i:9", "b:5"}],

Cell[BoxData[
    \("SpecialInverse[A] gives the inverse of a square (generically) \
invertible\n matrix A. It can be considerably more efficient than the \
standard\n Mathematica function Inverse if the elements of A are complex \
expressions."\)], "Print",
  CellTags->{"Info3268728760-5111011", "b:5", "Complex Arrays"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?SimplifyArray\)\)], "Input",
  CellLabel->"In[44]:=",
  CellTags->{"Complex Arrays", "i:10", "b:5"}],

Cell[BoxData[
    \("SimplifyArray[A] returns a simplified version of the list A.\n It is \
intended to replace the standard function Simplify when\n complex lists of \
expressions are to be simplified. SimplifyArray\n applies Simplify to \
individual expressions in A or groups of\n expressions with combined leaf \
count less than 200."\)], "Print",
  CellTags->{"Info3268728760-3056648", "b:5", "Complex Arrays"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?WriteArray\)\)], "Input",
  CellLabel->"In[45]:=",
  CellTags->{"Complex Arrays", "i:11", "b:5"}],

Cell[BoxData[
    \("WriteArray[filename,A] writes the array A to the file filename.\n \
WriteArray writes individual expressions or groups of expressions\n with leaf \
count less than 200. This partially mitigates the memory\n explosion that \
occurs when writing large arrays using the standard\n Mathematica write \
functions. WriteArray can roughly double the time\n required to write an \
array, but it may reduce the the maximum memory\n used by an order of \
magnitude or more. For example, to write\n and retrieve the two arrays A and \
B use:\n \n       WriteArray[Temp,{A,B}]\n       {A,B}=Get[Temp]\n   "\)], \
"Print",
  CellTags->{"Info3268728760-5984139", "b:5", "Complex Arrays"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Series Expansion", "Section",
  CellTags->{"Series Expansion", "i:16", "b:6"}],

Cell[CellGroupData[{

Cell["?SeriesExpansion", "Input",
  CellLabel->"In[46]:=",
  CellTags->{"Series Expansion", "i:16", "b:6"}],

Cell[BoxData[
    \("SeriesExpansion[A,x,n], Scalar,matrix or vector A(x), vector x and \
integer n.\n  Produces a power series expansion about x=0 up to terms of \
order n.\n "\)], "Print",
  CellTags->{"Info3268728760-6259689", "b:6", "Series Expansion"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["?Truncate", "Input",
  CellLabel->"In[47]:=",
  CellTags->{"Series Expansion", "i:16", "b:6"}],

Cell[BoxData[
    \("Truncate[A,x,n], Scalar,matrix or vector A(x), vector x and integer n.\
\n  Drops x terms of order higher n. Produces results equivalent to \
SeriesExpansion\n but may be in a different (unexpanded) form. Generally, \
significantly\n faster than SeriesExpansion.\n  "\)], "Print",
  CellTags->{"Info3268728760-6093397", "b:6", "Series Expansion"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Matlab Output", "Section",
  CellTags->{"i:17", "Matlab Output", "b:7"}],

Cell[CellGroupData[{

Cell["?MatlabForm", "Input",
  CellLabel->"In[48]:=",
  CellTags->{"i:17", "Matlab Output", "b:7"}],

Cell[BoxData[
    \("  *** MatlabInterface successfully loaded ***"\)], "Print",
  CellLabel->"From In[48]:=",
  CellTags->{"i:17", "Matlab Output", "b:7"}],

Cell[BoxData[
    \("MatlabForm[Matrix,MatrixName] converts the matrix or vector expression \
given  by the first argument Matrix into a Matlab readable and executable \
string with a variable name given by the second argument MatrixName. For \
example,  x_matlabform = MatlabForm[{{1,2},{3,4}},x]  returns a string of the \
form   xmatlabform = x=[1,2;3,4];  This function is useful for transfering \
matrices from Mathematica to Matlab via an ASCII file. The process is as \
follows:     (1) Using the above function, convert the pertinent matrices to  \
      Matlab readable strings.   (2) Write the strings to an ASCII file.   \
(3) At the Matlab prompt, type the name of the file which will execute the \
matrix assignments."\)], "Print",
  CellTags->{"Info3268728760-5557815", "b:7", "Matlab Output"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["?MatlabFormPolynomials", "Input",
  CellLabel->"In[49]:=",
  CellTags->{"i:17", "Matlab Output", "b:7"}],

Cell[BoxData[
    \("MatlabFormPolynomials[Polynomial,order,name,Filename] creates a\n \
Malab readable file 'Filename.m' that defines the matrix coefficients\n of \
matrix-valued multivariate polynomials in the list 'Polynomial'.\n \
Coefficients up to degree 'order' are generated. 'name' is a list of\n base \
names (strings) for the coefficents.\n  Example:\n A = {x^2 + x y + 2 y^2, 2 \
+ x}, {3, x + 3 x^2 + y + x y}}\n Polynomial={A};\n order=3;\n \
names={ToString[A]};\n Filename=ToString[TestFile.m];\n \
MatlabFormPolynomials[Polynomial,order,names,Filename]\n \n Produces an ASCII \
file named TestFile.m containing:\n  \n 'Variables=[x,y]'\n A00=[0,2;3,0];\n \
A01=[0,0;0,1];\n A10=[0,1;0,1];\n A02=[2,0;0,0];\n A11=[1,0;0,1];\n \
A20=[1,0;0,3];\n A03=[0,0;0,0];\n A12=[0,0;0,0];\n A21=[0,0;0,0];\n \
A30=[0,0;0,0];\n \n At the Matlab prompt, type the name of the file and press \
enter\n to execute the matrix assignments.\n \n Matrices of the form Aij are \
coefficents of terms of the form x^i*y^j\n \n The reconstructed polynomial is\
\n A00+A01y+A10x+A02y^2+A11xy+A20x^2+A03y^3+A12xy^2+A21x^2y^2+A30x^3 "\)], \
"Print",
  CellTags->{"Info3268728762-6123324", "b:7", "Matlab Output"}]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"5.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 695}},
ShowPageBreaks->False,
WindowSize->{647, 611},
WindowMargins->{{0, Automatic}, {Automatic, 5}},
Visible->True,
PrintingStartingPageNumber->12,
TaggingRules:>{"NextIndexingNumber" -> 18, "IndexEntries" -> {{"i:1", \
"Grad"}, {"i:10", "SimplifyArray"}, {"i:11", "WriteArray"}, {"i:12", \
"SpecialNullSpace"}, {"i:13", "TriangularDecomposition"}, {"i:17", \
"MatlabFormPolynomials"}, {"i:2", "Jacob"}, {"i:3", "Ad"}, {"i:3", \
"LieBracket"}, {"i:4", "LieDerivative"}, {"i:5", "Rank"}, {"i:6", "Span"}, \
{"i:8", "SmallestInvariantDistribution"}, {"i:9", "SpecialInverse"}, {"i:16", \
"SeriesExpansion", "Truncate"}, {"i:17", "MatlabForm", "", ""}, {"i:8", \
"LargestInvariantDistribution", "", ""}}, "PaginationCache" -> {{2003, 8, 1, \
21, 15, 41.5607024`9.371257756107292}, {12, 12, 12, 12, 12, 12, 12, 12, 13, \
13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, \
15, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, \
16, 16, 16, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 18, 18, \
18, 18, 18, 18, 18, 18, 18, 19, 19, 19, 19, 19, 19, 19, 19, 20, 20, 20, 20, \
20, 20, 20, 20, 20, 20, 20, 20, 21, 21, 21, 21, 21, 21, 21, 21, 21, 22, \
22}}},
ShowCellTags->True,
StyleDefinitions -> "HelpBrowser.nb"
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "b:3"->{
    Cell[1776, 53, 50, 1, 70, "Section",
      CellTags->"b:3"],
    Cell[1851, 58, 86, 3, 43, "Input",
      CellTags->"b:3"],
    Cell[1940, 63, 708, 11, 417, "Print",
      CellTags->"b:3"],
    Cell[2685, 79, 90, 3, 43, "Input",
      CellTags->"b:3"],
    Cell[2778, 84, 91, 3, 41, "Output",
      CellTags->"b:3"],
    Cell[2906, 92, 68, 2, 45, "Input",
      CellTags->"b:3"],
    Cell[2977, 96, 121, 3, 37, "Print",
      CellTags->"b:3"],
    Cell[3101, 101, 85, 3, 41, "Output",
      CellTags->"b:3"],
    Cell[3223, 109, 69, 2, 45, "Input",
      CellTags->"b:3"],
    Cell[3295, 113, 1176, 19, 588, "Print",
      CellTags->{"Info3268728757-3969975", "b:3"}]},
  "Info3268728757-3969975"->{
    Cell[3295, 113, 1176, 19, 588, "Print",
      CellTags->{"Info3268728757-3969975", "b:3"}]},
  "Grad"->{
    Cell[4586, 142, 64, 1, 53, "Subsection",
      CellTags->{"Grad", "i:1", "b:4.0"}],
    Cell[4675, 147, 104, 3, 43, "Input",
      CellTags->{"Grad", "i:1", "b:4.0"}],
    Cell[5007, 160, 124, 3, 43, "Input",
      CellTags->{"Grad", "i:1", "b:4.0"}],
    Cell[5134, 165, 122, 3, 41, "Output",
      CellTags->{"Grad", "i:1", "b:4.0"}]},
  "i:1"->{
    Cell[4586, 142, 64, 1, 53, "Subsection",
      CellTags->{"Grad", "i:1", "b:4.0"}],
    Cell[4675, 147, 104, 3, 43, "Input",
      CellTags->{"Grad", "i:1", "b:4.0"}],
    Cell[5007, 160, 124, 3, 43, "Input",
      CellTags->{"Grad", "i:1", "b:4.0"}],
    Cell[5134, 165, 122, 3, 41, "Output",
      CellTags->{"Grad", "i:1", "b:4.0"}]},
  "b:4.0"->{
    Cell[4586, 142, 64, 1, 53, "Subsection",
      CellTags->{"Grad", "i:1", "b:4.0"}],
    Cell[4675, 147, 104, 3, 43, "Input",
      CellTags->{"Grad", "i:1", "b:4.0"}],
    Cell[4782, 152, 188, 3, 56, "Print",
      CellTags->{"Info3268728757-8371087", "b:4.0"}],
    Cell[5007, 160, 124, 3, 43, "Input",
      CellTags->{"Grad", "i:1", "b:4.0"}],
    Cell[5134, 165, 122, 3, 41, "Output",
      CellTags->{"Grad", "i:1", "b:4.0"}]},
  "Info3268728757-8371087"->{
    Cell[4782, 152, 188, 3, 56, "Print",
      CellTags->{"Info3268728757-8371087", "b:4.0"}]},
  "Jacob"->{
    Cell[5305, 174, 66, 1, 53, "Subsection",
      CellTags->{"Jacob", "i:2", "b:4.1"}],
    Cell[5396, 179, 106, 3, 43, "Input",
      CellTags->{"Jacob", "i:2", "b:4.1"}],
    Cell[5782, 193, 134, 3, 43, "Input",
      CellTags->{"Jacob", "i:2", "b:4.1"}],
    Cell[5919, 198, 171, 4, 41, "Output",
      CellTags->{"Jacob", "i:2", "b:4.1"}]},
  "i:2"->{
    Cell[5305, 174, 66, 1, 53, "Subsection",
      CellTags->{"Jacob", "i:2", "b:4.1"}],
    Cell[5396, 179, 106, 3, 43, "Input",
      CellTags->{"Jacob", "i:2", "b:4.1"}],
    Cell[5782, 193, 134, 3, 43, "Input",
      CellTags->{"Jacob", "i:2", "b:4.1"}],
    Cell[5919, 198, 171, 4, 41, "Output",
      CellTags->{"Jacob", "i:2", "b:4.1"}]},
  "b:4.1"->{
    Cell[5305, 174, 66, 1, 53, "Subsection",
      CellTags->{"Jacob", "i:2", "b:4.1"}],
    Cell[5396, 179, 106, 3, 43, "Input",
      CellTags->{"Jacob", "i:2", "b:4.1"}],
    Cell[5505, 184, 240, 4, 94, "Print",
      CellTags->{"Info3268728757-7417227", "b:4.1"}],
    Cell[5782, 193, 134, 3, 43, "Input",
      CellTags->{"Jacob", "i:2", "b:4.1"}],
    Cell[5919, 198, 171, 4, 41, "Output",
      CellTags->{"Jacob", "i:2", "b:4.1"}]},
  "Info3268728757-7417227"->{
    Cell[5505, 184, 240, 4, 94, "Print",
      CellTags->{"Info3268728757-7417227", "b:4.1"}]},
  "LieBracket"->{
    Cell[6139, 208, 81, 1, 53, "Subsection",
      CellTags->{"LieBracket", "i:3", "b:4.2"}],
    Cell[6245, 213, 116, 3, 43, "Input",
      CellTags->{"LieBracket", "i:3", "b:4.2"}],
    Cell[6671, 227, 160, 3, 43, "Input",
      CellTags->{"LieBracket", "i:3", "b:4.2"}],
    Cell[6834, 232, 274, 5, 60, "Output",
      CellTags->{"LieBracket", "i:3", "b:4.2"}],
    Cell[7145, 242, 109, 3, 43, "Input",
      CellTags->{"LieBracket", "i:3", "b:4.2"}],
    Cell[7638, 257, 155, 3, 43, "Input",
      CellTags->{"LieBracket", "i:3", "b:4.2"}],
    Cell[7796, 262, 1226, 21, 212, "Output",
      CellTags->{"LieBracket", "i:3", "b:4.2"}]},
  "i:3"->{
    Cell[6139, 208, 81, 1, 53, "Subsection",
      CellTags->{"LieBracket", "i:3", "b:4.2"}],
    Cell[6245, 213, 116, 3, 43, "Input",
      CellTags->{"LieBracket", "i:3", "b:4.2"}],
    Cell[6671, 227, 160, 3, 43, "Input",
      CellTags->{"LieBracket", "i:3", "b:4.2"}],
    Cell[6834, 232, 274, 5, 60, "Output",
      CellTags->{"LieBracket", "i:3", "b:4.2"}],
    Cell[7145, 242, 109, 3, 43, "Input",
      CellTags->{"LieBracket", "i:3", "b:4.2"}],
    Cell[7638, 257, 155, 3, 43, "Input",
      CellTags->{"LieBracket", "i:3", "b:4.2"}],
    Cell[7796, 262, 1226, 21, 212, "Output",
      CellTags->{"LieBracket", "i:3", "b:4.2"}]},
  "b:4.2"->{
    Cell[6139, 208, 81, 1, 53, "Subsection",
      CellTags->{"LieBracket", "i:3", "b:4.2"}],
    Cell[6245, 213, 116, 3, 43, "Input",
      CellTags->{"LieBracket", "i:3", "b:4.2"}],
    Cell[6364, 218, 270, 4, 94, "Print",
      CellTags->{"Info3268728757-4720349", "b:4.2"}],
    Cell[6671, 227, 160, 3, 43, "Input",
      CellTags->{"LieBracket", "i:3", "b:4.2"}],
    Cell[6834, 232, 274, 5, 60, "Output",
      CellTags->{"LieBracket", "i:3", "b:4.2"}],
    Cell[7145, 242, 109, 3, 43, "Input",
      CellTags->{"LieBracket", "i:3", "b:4.2"}],
    Cell[7257, 247, 344, 5, 113, "Print",
      CellTags->{"Info3268728757-8448341", "b:4.2"}],
    Cell[7638, 257, 155, 3, 43, "Input",
      CellTags->{"LieBracket", "i:3", "b:4.2"}],
    Cell[7796, 262, 1226, 21, 212, "Output",
      CellTags->{"LieBracket", "i:3", "b:4.2"}]},
  "Info3268728757-4720349"->{
    Cell[6364, 218, 270, 4, 94, "Print",
      CellTags->{"Info3268728757-4720349", "b:4.2"}]},
  "Info3268728757-8448341"->{
    Cell[7257, 247, 344, 5, 113, "Print",
      CellTags->{"Info3268728757-8448341", "b:4.2"}]},
  "LieDerivative"->{
    Cell[9071, 289, 82, 1, 53, "Subsection",
      CellTags->{"LieDerivative", "i:4", "b:4.3"}],
    Cell[9178, 294, 123, 3, 43, "Input",
      CellTags->{"LieDerivative", "i:4", "b:4.3"}],
    Cell[9966, 313, 176, 4, 43, "Input",
      CellTags->{"LieDerivative", "i:4", "b:4.3"}],
    Cell[10145, 319, 182, 4, 41, "Output",
      CellTags->{"LieDerivative", "i:4", "b:4.3"}],
    Cell[10364, 328, 177, 4, 43, "Input",
      CellTags->{"LieDerivative", "i:4", "b:4.3"}],
    Cell[10544, 334, 438, 9, 79, "Output",
      CellTags->{"LieDerivative", "i:4", "b:4.3"}]},
  "i:4"->{
    Cell[9071, 289, 82, 1, 53, "Subsection",
      CellTags->{"LieDerivative", "i:4", "b:4.3"}],
    Cell[9178, 294, 123, 3, 43, "Input",
      CellTags->{"LieDerivative", "i:4", "b:4.3"}],
    Cell[9966, 313, 176, 4, 43, "Input",
      CellTags->{"LieDerivative", "i:4", "b:4.3"}],
    Cell[10145, 319, 182, 4, 41, "Output",
      CellTags->{"LieDerivative", "i:4", "b:4.3"}],
    Cell[10364, 328, 177, 4, 43, "Input",
      CellTags->{"LieDerivative", "i:4", "b:4.3"}],
    Cell[10544, 334, 438, 9, 79, "Output",
      CellTags->{"LieDerivative", "i:4", "b:4.3"}]},
  "b:4.3"->{
    Cell[9071, 289, 82, 1, 53, "Subsection",
      CellTags->{"LieDerivative", "i:4", "b:4.3"}],
    Cell[9178, 294, 123, 3, 43, "Input",
      CellTags->{"LieDerivative", "i:4", "b:4.3"}],
    Cell[9304, 299, 625, 9, 265, "Print",
      CellTags->{"Info3268728758-4755545", "b:4.3"}],
    Cell[9966, 313, 176, 4, 43, "Input",
      CellTags->{"LieDerivative", "i:4", "b:4.3"}],
    Cell[10145, 319, 182, 4, 41, "Output",
      CellTags->{"LieDerivative", "i:4", "b:4.3"}],
    Cell[10364, 328, 177, 4, 43, "Input",
      CellTags->{"LieDerivative", "i:4", "b:4.3"}],
    Cell[10544, 334, 438, 9, 79, "Output",
      CellTags->{"LieDerivative", "i:4", "b:4.3"}]},
  "Info3268728758-4755545"->{
    Cell[9304, 299, 625, 9, 265, "Print",
      CellTags->{"Info3268728758-4755545", "b:4.3"}]},
  "Involutive"->{
    Cell[11031, 349, 69, 1, 53, "Subsection",
      CellTags->{"Involutive", "b:4.4"}],
    Cell[11125, 354, 110, 3, 43, "Input",
      CellTags->{"Involutive", "b:4.4"}],
    Cell[11587, 369, 156, 4, 43, "Input",
      CellTags->{"Involutive", "b:4.4"}],
    Cell[11746, 375, 100, 3, 41, "Output",
      CellTags->{"Involutive", "b:4.4"}]},
  "b:4.4"->{
    Cell[11031, 349, 69, 1, 53, "Subsection",
      CellTags->{"Involutive", "b:4.4"}],
    Cell[11125, 354, 110, 3, 43, "Input",
      CellTags->{"Involutive", "b:4.4"}],
    Cell[11238, 359, 312, 5, 94, "Print",
      CellTags->{"Info3268728758-2063324", "b:4.4"}],
    Cell[11587, 369, 156, 4, 43, "Input",
      CellTags->{"Involutive", "b:4.4"}],
    Cell[11746, 375, 100, 3, 41, "Output",
      CellTags->{"Involutive", "b:4.4"}]},
  "Info3268728758-2063324"->{
    Cell[11238, 359, 312, 5, 94, "Print",
      CellTags->{"Info3268728758-2063324", "b:4.4"}]},
  "Rank"->{
    Cell[11895, 384, 64, 1, 53, "Subsection",
      CellTags->{"Rank", "i:5", "b:4.5"}],
    Cell[11984, 389, 105, 3, 43, "Input",
      CellTags->{"Rank", "i:5", "b:4.5"}],
    Cell[12427, 404, 141, 3, 43, "Input",
      CellTags->{"Rank", "i:5", "b:4.5"}],
    Cell[12571, 409, 98, 3, 41, "Output",
      CellTags->{"Rank", "i:5", "b:4.5"}]},
  "i:5"->{
    Cell[11895, 384, 64, 1, 53, "Subsection",
      CellTags->{"Rank", "i:5", "b:4.5"}],
    Cell[11984, 389, 105, 3, 43, "Input",
      CellTags->{"Rank", "i:5", "b:4.5"}],
    Cell[12427, 404, 141, 3, 43, "Input",
      CellTags->{"Rank", "i:5", "b:4.5"}],
    Cell[12571, 409, 98, 3, 41, "Output",
      CellTags->{"Rank", "i:5", "b:4.5"}]},
  "b:4.5"->{
    Cell[11895, 384, 64, 1, 53, "Subsection",
      CellTags->{"Rank", "i:5", "b:4.5"}],
    Cell[11984, 389, 105, 3, 43, "Input",
      CellTags->{"Rank", "i:5", "b:4.5"}],
    Cell[12092, 394, 298, 5, 75, "Print",
      CellTags->{"Info3268728758-9909747", "b:4.5"}],
    Cell[12427, 404, 141, 3, 43, "Input",
      CellTags->{"Rank", "i:5", "b:4.5"}],
    Cell[12571, 409, 98, 3, 41, "Output",
      CellTags->{"Rank", "i:5", "b:4.5"}]},
  "Info3268728758-9909747"->{
    Cell[12092, 394, 298, 5, 75, "Print",
      CellTags->{"Info3268728758-9909747", "b:4.5"}]},
  "Span"->{
    Cell[12718, 418, 64, 1, 53, "Subsection",
      CellTags->{"Span", "i:6", "b:4.6"}],
    Cell[12807, 423, 105, 3, 43, "Input",
      CellTags->{"Span", "i:6", "b:4.6"}],
    Cell[13389, 439, 141, 3, 43, "Input",
      CellTags->{"Span", "i:6", "b:4.6"}],
    Cell[13533, 444, 113, 3, 41, "Output",
      CellTags->{"Span", "i:6", "b:4.6"}]},
  "i:6"->{
    Cell[12718, 418, 64, 1, 53, "Subsection",
      CellTags->{"Span", "i:6", "b:4.6"}],
    Cell[12807, 423, 105, 3, 43, "Input",
      CellTags->{"Span", "i:6", "b:4.6"}],
    Cell[13389, 439, 141, 3, 43, "Input",
      CellTags->{"Span", "i:6", "b:4.6"}],
    Cell[13533, 444, 113, 3, 41, "Output",
      CellTags->{"Span", "i:6", "b:4.6"}]},
  "b:4.6"->{
    Cell[12718, 418, 64, 1, 53, "Subsection",
      CellTags->{"Span", "i:6", "b:4.6"}],
    Cell[12807, 423, 105, 3, 43, "Input",
      CellTags->{"Span", "i:6", "b:4.6"}],
    Cell[12915, 428, 437, 6, 151, "Print",
      CellTags->{"Info3268728758-4807394", "b:4.6"}],
    Cell[13389, 439, 141, 3, 43, "Input",
      CellTags->{"Span", "i:6", "b:4.6"}],
    Cell[13533, 444, 113, 3, 41, "Output",
      CellTags->{"Span", "i:6", "b:4.6"}]},
  "Info3268728758-4807394"->{
    Cell[12915, 428, 437, 6, 151, "Print",
      CellTags->{"Info3268728758-4807394", "b:4.6"}]},
  "SpecialNullSpace"->{
    Cell[13695, 453, 89, 1, 53, "Subsection",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
    Cell[13809, 458, 130, 3, 43, "Input",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
    Cell[14200, 472, 163, 3, 44, "Input",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
    Cell[14366, 477, 161, 3, 41, "Output",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
    Cell[14564, 485, 121, 3, 43, "Input",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
    Cell[14688, 490, 135, 3, 54, "Output",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
    Cell[14860, 498, 128, 3, 43, "Input",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
    Cell[14991, 503, 128, 3, 41, "Output",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}]},
  "i:12"->{
    Cell[13695, 453, 89, 1, 53, "Subsection",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
    Cell[13809, 458, 130, 3, 43, "Input",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
    Cell[14200, 472, 163, 3, 44, "Input",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
    Cell[14366, 477, 161, 3, 41, "Output",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
    Cell[14564, 485, 121, 3, 43, "Input",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
    Cell[14688, 490, 135, 3, 54, "Output",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
    Cell[14860, 498, 128, 3, 43, "Input",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
    Cell[14991, 503, 128, 3, 41, "Output",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}]},
  "b:4.7"->{
    Cell[13695, 453, 89, 1, 53, "Subsection",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
    Cell[13809, 458, 130, 3, 43, "Input",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
    Cell[13942, 463, 221, 4, 75, "Print",
      CellTags->{"Info3268728758-2216006", "b:4.7"}],
    Cell[14200, 472, 163, 3, 44, "Input",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
    Cell[14366, 477, 161, 3, 41, "Output",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
    Cell[14564, 485, 121, 3, 43, "Input",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
    Cell[14688, 490, 135, 3, 54, "Output",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
    Cell[14860, 498, 128, 3, 43, "Input",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
    Cell[14991, 503, 128, 3, 41, "Output",
      CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}]},
  "Info3268728758-2216006"->{
    Cell[13942, 463, 221, 4, 75, "Print",
      CellTags->{"Info3268728758-2216006", "b:4.7"}]},
  "Invariant Distributions"->{
    Cell[15168, 512, 102, 1, 53, "Subsection",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[15295, 517, 148, 3, 43, "Input",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[15853, 531, 316, 7, 103, "Input",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[16194, 542, 159, 3, 43, "Input",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[16356, 547, 145, 3, 41, "Output",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[16538, 555, 149, 3, 43, "Input",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[17073, 570, 130, 3, 43, "Input",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[17206, 575, 132, 3, 41, "Output",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[17375, 583, 162, 3, 43, "Input",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[17540, 588, 145, 3, 41, "Output",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}]},
  "i:8"->{
    Cell[15168, 512, 102, 1, 53, "Subsection",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[15295, 517, 148, 3, 43, "Input",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[15853, 531, 316, 7, 103, "Input",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[16194, 542, 159, 3, 43, "Input",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[16356, 547, 145, 3, 41, "Output",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[16538, 555, 149, 3, 43, "Input",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[17073, 570, 130, 3, 43, "Input",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[17206, 575, 132, 3, 41, "Output",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[17375, 583, 162, 3, 43, "Input",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[17540, 588, 145, 3, 41, "Output",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}]},
  "b:4.8"->{
    Cell[15168, 512, 102, 1, 53, "Subsection",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[15295, 517, 148, 3, 43, "Input",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[15446, 522, 392, 6, 113, "Print",
      CellTags->{"Info3268728758-5746287", "b:4.8"}],
    Cell[15853, 531, 316, 7, 103, "Input",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[16194, 542, 159, 3, 43, "Input",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[16356, 547, 145, 3, 41, "Output",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[16538, 555, 149, 3, 43, "Input",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[16690, 560, 346, 5, 94, "Print",
      CellTags->{"Info3268728758-4311524", "b:4.8"}],
    Cell[17073, 570, 130, 3, 43, "Input",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[17206, 575, 132, 3, 41, "Output",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[17375, 583, 162, 3, 43, "Input",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
    Cell[17540, 588, 145, 3, 41, "Output",
      CellTags->{"Invariant Distributions", "i:8", "b:4.8"}]},
  "Info3268728758-5746287"->{
    Cell[15446, 522, 392, 6, 113, "Print",
      CellTags->{"Info3268728758-5746287", "b:4.8"}]},
  "Info3268728758-4311524"->{
    Cell[16690, 560, 346, 5, 94, "Print",
      CellTags->{"Info3268728758-4311524", "b:4.8"}]},
  "Triangular Decomposition"->{
    Cell[17734, 597, 105, 1, 53, "Subsection",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[17864, 602, 145, 3, 43, "Input",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[18719, 619, 758, 12, 227, "Print",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[19514, 636, 320, 7, 103, "Input",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[19837, 645, 134, 3, 41, "Output",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[20008, 653, 171, 3, 43, "Input",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[20182, 658, 147, 3, 41, "Output",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[20366, 666, 163, 3, 43, "Input",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[20532, 671, 443, 7, 49, "Message",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[20978, 680, 175, 3, 41, "Output",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[21190, 688, 190, 4, 43, "Input",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[21383, 694, 254, 5, 60, "Output",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}]},
  "i:13"->{
    Cell[17734, 597, 105, 1, 53, "Subsection",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[17864, 602, 145, 3, 43, "Input",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[18719, 619, 758, 12, 227, "Print",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[19514, 636, 320, 7, 103, "Input",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[19837, 645, 134, 3, 41, "Output",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[20008, 653, 171, 3, 43, "Input",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[20182, 658, 147, 3, 41, "Output",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[20366, 666, 163, 3, 43, "Input",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[20532, 671, 443, 7, 49, "Message",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[20978, 680, 175, 3, 41, "Output",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[21190, 688, 190, 4, 43, "Input",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[21383, 694, 254, 5, 60, "Output",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}]},
  "b:4.9"->{
    Cell[17734, 597, 105, 1, 53, "Subsection",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[17864, 602, 145, 3, 43, "Input",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[18012, 607, 704, 10, 265, "Print",
      CellTags->{"Info3268728758-5559249", "b:4.9"}],
    Cell[18719, 619, 758, 12, 227, "Print",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[19514, 636, 320, 7, 103, "Input",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[19837, 645, 134, 3, 41, "Output",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[20008, 653, 171, 3, 43, "Input",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[20182, 658, 147, 3, 41, "Output",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[20366, 666, 163, 3, 43, "Input",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[20532, 671, 443, 7, 49, "Message",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[20978, 680, 175, 3, 41, "Output",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[21190, 688, 190, 4, 43, "Input",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
    Cell[21383, 694, 254, 5, 60, "Output",
      CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}]},
  "Info3268728758-5559249"->{
    Cell[18012, 607, 704, 10, 265, "Print",
      CellTags->{"Info3268728758-5559249", "b:4.9"}]},
  "Complex Arrays"->{
    Cell[21698, 706, 85, 1, 70, "Section",
      CellTags->{"Complex Arrays", "b:5"}],
    Cell[21808, 711, 123, 3, 43, "Input",
      CellTags->{"Complex Arrays", "i:9", "b:5"}],
    Cell[21934, 716, 323, 5, 75, "Print",
      CellTags->{"Info3268728760-5111011", "b:5", "Complex Arrays"}],
    Cell[22294, 726, 123, 3, 43, "Input",
      CellTags->{"Complex Arrays", "i:10", "b:5"}],
    Cell[22420, 731, 416, 6, 113, "Print",
      CellTags->{"Info3268728760-3056648", "b:5", "Complex Arrays"}],
    Cell[22873, 742, 120, 3, 43, "Input",
      CellTags->{"Complex Arrays", "i:11", "b:5"}],
    Cell[22996, 747, 695, 10, 246, "Print",
      CellTags->{"Info3268728760-5984139", "b:5", "Complex Arrays"}]},
  "b:5"->{
    Cell[21698, 706, 85, 1, 70, "Section",
      CellTags->{"Complex Arrays", "b:5"}],
    Cell[21808, 711, 123, 3, 43, "Input",
      CellTags->{"Complex Arrays", "i:9", "b:5"}],
    Cell[21934, 716, 323, 5, 75, "Print",
      CellTags->{"Info3268728760-5111011", "b:5", "Complex Arrays"}],
    Cell[22294, 726, 123, 3, 43, "Input",
      CellTags->{"Complex Arrays", "i:10", "b:5"}],
    Cell[22420, 731, 416, 6, 113, "Print",
      CellTags->{"Info3268728760-3056648", "b:5", "Complex Arrays"}],
    Cell[22873, 742, 120, 3, 43, "Input",
      CellTags->{"Complex Arrays", "i:11", "b:5"}],
    Cell[22996, 747, 695, 10, 246, "Print",
      CellTags->{"Info3268728760-5984139", "b:5", "Complex Arrays"}]},
  "i:9"->{
    Cell[21808, 711, 123, 3, 43, "Input",
      CellTags->{"Complex Arrays", "i:9", "b:5"}]},
  "Info3268728760-5111011"->{
    Cell[21934, 716, 323, 5, 75, "Print",
      CellTags->{"Info3268728760-5111011", "b:5", "Complex Arrays"}]},
  "i:10"->{
    Cell[22294, 726, 123, 3, 43, "Input",
      CellTags->{"Complex Arrays", "i:10", "b:5"}]},
  "Info3268728760-3056648"->{
    Cell[22420, 731, 416, 6, 113, "Print",
      CellTags->{"Info3268728760-3056648", "b:5", "Complex Arrays"}]},
  "i:11"->{
    Cell[22873, 742, 120, 3, 43, "Input",
      CellTags->{"Complex Arrays", "i:11", "b:5"}]},
  "Info3268728760-5984139"->{
    Cell[22996, 747, 695, 10, 246, "Print",
      CellTags->{"Info3268728760-5984139", "b:5", "Complex Arrays"}]},
  "Series Expansion"->{
    Cell[23740, 763, 84, 1, 70, "Section",
      CellTags->{"Series Expansion", "i:16", "b:6"}],
    Cell[23849, 768, 107, 2, 45, "Input",
      CellTags->{"Series Expansion", "i:16", "b:6"}],
    Cell[23959, 772, 257, 4, 94, "Print",
      CellTags->{"Info3268728760-6259689", "b:6", "Series Expansion"}],
    Cell[24253, 781, 100, 2, 45, "Input",
      CellTags->{"Series Expansion", "i:16", "b:6"}],
    Cell[24356, 785, 368, 5, 132, "Print",
      CellTags->{"Info3268728760-6093397", "b:6", "Series Expansion"}]},
  "i:16"->{
    Cell[23740, 763, 84, 1, 70, "Section",
      CellTags->{"Series Expansion", "i:16", "b:6"}],
    Cell[23849, 768, 107, 2, 45, "Input",
      CellTags->{"Series Expansion", "i:16", "b:6"}],
    Cell[24253, 781, 100, 2, 45, "Input",
      CellTags->{"Series Expansion", "i:16", "b:6"}]},
  "b:6"->{
    Cell[23740, 763, 84, 1, 70, "Section",
      CellTags->{"Series Expansion", "i:16", "b:6"}],
    Cell[23849, 768, 107, 2, 45, "Input",
      CellTags->{"Series Expansion", "i:16", "b:6"}],
    Cell[23959, 772, 257, 4, 94, "Print",
      CellTags->{"Info3268728760-6259689", "b:6", "Series Expansion"}],
    Cell[24253, 781, 100, 2, 45, "Input",
      CellTags->{"Series Expansion", "i:16", "b:6"}],
    Cell[24356, 785, 368, 5, 132, "Print",
      CellTags->{"Info3268728760-6093397", "b:6", "Series Expansion"}]},
  "Info3268728760-6259689"->{
    Cell[23959, 772, 257, 4, 94, "Print",
      CellTags->{"Info3268728760-6259689", "b:6", "Series Expansion"}]},
  "Info3268728760-6093397"->{
    Cell[24356, 785, 368, 5, 132, "Print",
      CellTags->{"Info3268728760-6093397", "b:6", "Series Expansion"}]},
  "i:17"->{
    Cell[24773, 796, 78, 1, 70, "Section",
      CellTags->{"i:17", "Matlab Output", "b:7"}],
    Cell[24876, 801, 99, 2, 45, "Input",
      CellTags->{"i:17", "Matlab Output", "b:7"}],
    Cell[24978, 805, 156, 3, 37, "Print",
      CellTags->{"i:17", "Matlab Output", "b:7"}],
    Cell[25982, 826, 110, 2, 45, "Input",
      CellTags->{"i:17", "Matlab Output", "b:7"}]},
  "Matlab Output"->{
    Cell[24773, 796, 78, 1, 70, "Section",
      CellTags->{"i:17", "Matlab Output", "b:7"}],
    Cell[24876, 801, 99, 2, 45, "Input",
      CellTags->{"i:17", "Matlab Output", "b:7"}],
    Cell[24978, 805, 156, 3, 37, "Print",
      CellTags->{"i:17", "Matlab Output", "b:7"}],
    Cell[25137, 810, 808, 11, 208, "Print",
      CellTags->{"Info3268728760-5557815", "b:7", "Matlab Output"}],
    Cell[25982, 826, 110, 2, 45, "Input",
      CellTags->{"i:17", "Matlab Output", "b:7"}],
    Cell[26095, 830, 1201, 17, 664, "Print",
      CellTags->{"Info3268728762-6123324", "b:7", "Matlab Output"}]},
  "b:7"->{
    Cell[24773, 796, 78, 1, 70, "Section",
      CellTags->{"i:17", "Matlab Output", "b:7"}],
    Cell[24876, 801, 99, 2, 45, "Input",
      CellTags->{"i:17", "Matlab Output", "b:7"}],
    Cell[24978, 805, 156, 3, 37, "Print",
      CellTags->{"i:17", "Matlab Output", "b:7"}],
    Cell[25137, 810, 808, 11, 208, "Print",
      CellTags->{"Info3268728760-5557815", "b:7", "Matlab Output"}],
    Cell[25982, 826, 110, 2, 45, "Input",
      CellTags->{"i:17", "Matlab Output", "b:7"}],
    Cell[26095, 830, 1201, 17, 664, "Print",
      CellTags->{"Info3268728762-6123324", "b:7", "Matlab Output"}]},
  "Info3268728760-5557815"->{
    Cell[25137, 810, 808, 11, 208, "Print",
      CellTags->{"Info3268728760-5557815", "b:7", "Matlab Output"}]},
  "Info3268728762-6123324"->{
    Cell[26095, 830, 1201, 17, 664, "Print",
      CellTags->{"Info3268728762-6123324", "b:7", "Matlab Output"}]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"b:3", 29101, 886},
  {"Info3268728757-3969975", 29805, 907},
  {"Grad", 29914, 910},
  {"i:1", 30267, 919},
  {"b:4.0", 30622, 928},
  {"Info3268728757-8371087", 31089, 939},
  {"Jacob", 31198, 942},
  {"i:2", 31555, 951},
  {"b:4.1", 31914, 960},
  {"Info3268728757-7417227", 32385, 971},
  {"LieBracket", 32499, 974},
  {"i:3", 33150, 989},
  {"b:4.2", 33803, 1004},
  {"Info3268728757-4720349", 34664, 1023},
  {"Info3268728757-8448341", 34790, 1026},
  {"LieDerivative", 34908, 1029},
  {"i:4", 35487, 1042},
  {"b:4.3", 36068, 1055},
  {"Info3268728758-4755545", 36762, 1070},
  {"Involutive", 36877, 1073},
  {"b:4.4", 37232, 1082},
  {"Info3268728758-2063324", 37700, 1093},
  {"Rank", 37809, 1096},
  {"i:5", 38165, 1105},
  {"b:4.5", 38523, 1114},
  {"Info3268728758-9909747", 38994, 1125},
  {"Span", 39103, 1128},
  {"i:6", 39460, 1137},
  {"b:4.6", 39819, 1146},
  {"Info3268728758-4807394", 40292, 1157},
  {"SpecialNullSpace", 40414, 1160},
  {"i:12", 41218, 1177},
  {"b:4.7", 42023, 1194},
  {"Info3268728758-2216006", 42941, 1213},
  {"Invariant Distributions", 43069, 1216},
  {"i:8", 44130, 1237},
  {"b:4.8", 45193, 1258},
  {"Info3268728758-5746287", 46466, 1283},
  {"Info3268728758-4311524", 46594, 1286},
  {"Triangular Decomposition", 46723, 1289},
  {"i:13", 48022, 1314},
  {"b:4.9", 49322, 1339},
  {"Info3268728758-5559249", 50737, 1366},
  {"Complex Arrays", 50858, 1369},
  {"b:5", 51577, 1384},
  {"i:9", 52296, 1399},
  {"Info3268728760-5111011", 52420, 1402},
  {"i:10", 52545, 1405},
  {"Info3268728760-3056648", 52670, 1408},
  {"i:11", 52796, 1411},
  {"Info3268728760-5984139", 52921, 1414},
  {"Series Expansion", 53060, 1417},
  {"i:16", 53591, 1428},
  {"b:6", 53892, 1435},
  {"Info3268728760-6259689", 54441, 1446},
  {"Info3268728760-6093397", 54586, 1449},
  {"i:17", 54714, 1452},
  {"Matlab Output", 55108, 1461},
  {"b:7", 55719, 1474},
  {"Info3268728760-5557815", 56349, 1487},
  {"Info3268728762-6123324", 56493, 1490}
  }
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 50, 1, 70, "Section",
  CellTags->"b:3"],

Cell[CellGroupData[{
Cell[1851, 58, 86, 3, 43, "Input",
  CellTags->"b:3"],
Cell[1940, 63, 708, 11, 417, "Print",
  CellTags->"b:3"]
}, Open  ]],

Cell[CellGroupData[{
Cell[2685, 79, 90, 3, 43, "Input",
  CellTags->"b:3"],
Cell[2778, 84, 91, 3, 41, "Output",
  CellTags->"b:3"]
}, Open  ]],

Cell[CellGroupData[{
Cell[2906, 92, 68, 2, 45, "Input",
  CellTags->"b:3"],
Cell[2977, 96, 121, 3, 37, "Print",
  CellTags->"b:3"],
Cell[3101, 101, 85, 3, 41, "Output",
  CellTags->"b:3"]
}, Open  ]],

Cell[CellGroupData[{
Cell[3223, 109, 69, 2, 45, "Input",
  CellTags->"b:3"],
Cell[3295, 113, 1176, 19, 588, "Print",
  CellTags->{"Info3268728757-3969975", "b:3"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[4520, 138, 41, 0, 56, "Section"],

Cell[CellGroupData[{
Cell[4586, 142, 64, 1, 53, "Subsection",
  CellTags->{"Grad", "i:1", "b:4.0"}],

Cell[CellGroupData[{
Cell[4675, 147, 104, 3, 43, "Input",
  CellTags->{"Grad", "i:1", "b:4.0"}],
Cell[4782, 152, 188, 3, 56, "Print",
  CellTags->{"Info3268728757-8371087", "b:4.0"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[5007, 160, 124, 3, 43, "Input",
  CellTags->{"Grad", "i:1", "b:4.0"}],
Cell[5134, 165, 122, 3, 41, "Output",
  CellTags->{"Grad", "i:1", "b:4.0"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[5305, 174, 66, 1, 53, "Subsection",
  CellTags->{"Jacob", "i:2", "b:4.1"}],

Cell[CellGroupData[{
Cell[5396, 179, 106, 3, 43, "Input",
  CellTags->{"Jacob", "i:2", "b:4.1"}],
Cell[5505, 184, 240, 4, 94, "Print",
  CellTags->{"Info3268728757-7417227", "b:4.1"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[5782, 193, 134, 3, 43, "Input",
  CellTags->{"Jacob", "i:2", "b:4.1"}],
Cell[5919, 198, 171, 4, 41, "Output",
  CellTags->{"Jacob", "i:2", "b:4.1"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[6139, 208, 81, 1, 53, "Subsection",
  CellTags->{"LieBracket", "i:3", "b:4.2"}],

Cell[CellGroupData[{
Cell[6245, 213, 116, 3, 43, "Input",
  CellTags->{"LieBracket", "i:3", "b:4.2"}],
Cell[6364, 218, 270, 4, 94, "Print",
  CellTags->{"Info3268728757-4720349", "b:4.2"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[6671, 227, 160, 3, 43, "Input",
  CellTags->{"LieBracket", "i:3", "b:4.2"}],
Cell[6834, 232, 274, 5, 60, "Output",
  CellTags->{"LieBracket", "i:3", "b:4.2"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[7145, 242, 109, 3, 43, "Input",
  CellTags->{"LieBracket", "i:3", "b:4.2"}],
Cell[7257, 247, 344, 5, 113, "Print",
  CellTags->{"Info3268728757-8448341", "b:4.2"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[7638, 257, 155, 3, 43, "Input",
  CellTags->{"LieBracket", "i:3", "b:4.2"}],
Cell[7796, 262, 1226, 21, 212, "Output",
  CellTags->{"LieBracket", "i:3", "b:4.2"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[9071, 289, 82, 1, 53, "Subsection",
  CellTags->{"LieDerivative", "i:4", "b:4.3"}],

Cell[CellGroupData[{
Cell[9178, 294, 123, 3, 43, "Input",
  CellTags->{"LieDerivative", "i:4", "b:4.3"}],
Cell[9304, 299, 625, 9, 265, "Print",
  CellTags->{"Info3268728758-4755545", "b:4.3"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[9966, 313, 176, 4, 43, "Input",
  CellTags->{"LieDerivative", "i:4", "b:4.3"}],
Cell[10145, 319, 182, 4, 41, "Output",
  CellTags->{"LieDerivative", "i:4", "b:4.3"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[10364, 328, 177, 4, 43, "Input",
  CellTags->{"LieDerivative", "i:4", "b:4.3"}],
Cell[10544, 334, 438, 9, 79, "Output",
  CellTags->{"LieDerivative", "i:4", "b:4.3"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[11031, 349, 69, 1, 53, "Subsection",
  CellTags->{"Involutive", "b:4.4"}],

Cell[CellGroupData[{
Cell[11125, 354, 110, 3, 43, "Input",
  CellTags->{"Involutive", "b:4.4"}],
Cell[11238, 359, 312, 5, 94, "Print",
  CellTags->{"Info3268728758-2063324", "b:4.4"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[11587, 369, 156, 4, 43, "Input",
  CellTags->{"Involutive", "b:4.4"}],
Cell[11746, 375, 100, 3, 41, "Output",
  CellTags->{"Involutive", "b:4.4"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[11895, 384, 64, 1, 53, "Subsection",
  CellTags->{"Rank", "i:5", "b:4.5"}],

Cell[CellGroupData[{
Cell[11984, 389, 105, 3, 43, "Input",
  CellTags->{"Rank", "i:5", "b:4.5"}],
Cell[12092, 394, 298, 5, 75, "Print",
  CellTags->{"Info3268728758-9909747", "b:4.5"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[12427, 404, 141, 3, 43, "Input",
  CellTags->{"Rank", "i:5", "b:4.5"}],
Cell[12571, 409, 98, 3, 41, "Output",
  CellTags->{"Rank", "i:5", "b:4.5"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[12718, 418, 64, 1, 53, "Subsection",
  CellTags->{"Span", "i:6", "b:4.6"}],

Cell[CellGroupData[{
Cell[12807, 423, 105, 3, 43, "Input",
  CellTags->{"Span", "i:6", "b:4.6"}],
Cell[12915, 428, 437, 6, 151, "Print",
  CellTags->{"Info3268728758-4807394", "b:4.6"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[13389, 439, 141, 3, 43, "Input",
  CellTags->{"Span", "i:6", "b:4.6"}],
Cell[13533, 444, 113, 3, 41, "Output",
  CellTags->{"Span", "i:6", "b:4.6"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[13695, 453, 89, 1, 53, "Subsection",
  CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],

Cell[CellGroupData[{
Cell[13809, 458, 130, 3, 43, "Input",
  CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
Cell[13942, 463, 221, 4, 75, "Print",
  CellTags->{"Info3268728758-2216006", "b:4.7"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[14200, 472, 163, 3, 44, "Input",
  CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
Cell[14366, 477, 161, 3, 41, "Output",
  CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[14564, 485, 121, 3, 43, "Input",
  CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
Cell[14688, 490, 135, 3, 54, "Output",
  CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[14860, 498, 128, 3, 43, "Input",
  CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}],
Cell[14991, 503, 128, 3, 41, "Output",
  CellTags->{"SpecialNullSpace", "i:12", "b:4.7"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[15168, 512, 102, 1, 53, "Subsection",
  CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],

Cell[CellGroupData[{
Cell[15295, 517, 148, 3, 43, "Input",
  CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
Cell[15446, 522, 392, 6, 113, "Print",
  CellTags->{"Info3268728758-5746287", "b:4.8"}]
}, Open  ]],
Cell[15853, 531, 316, 7, 103, "Input",
  CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],

Cell[CellGroupData[{
Cell[16194, 542, 159, 3, 43, "Input",
  CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
Cell[16356, 547, 145, 3, 41, "Output",
  CellTags->{"Invariant Distributions", "i:8", "b:4.8"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[16538, 555, 149, 3, 43, "Input",
  CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
Cell[16690, 560, 346, 5, 94, "Print",
  CellTags->{"Info3268728758-4311524", "b:4.8"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[17073, 570, 130, 3, 43, "Input",
  CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
Cell[17206, 575, 132, 3, 41, "Output",
  CellTags->{"Invariant Distributions", "i:8", "b:4.8"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[17375, 583, 162, 3, 43, "Input",
  CellTags->{"Invariant Distributions", "i:8", "b:4.8"}],
Cell[17540, 588, 145, 3, 41, "Output",
  CellTags->{"Invariant Distributions", "i:8", "b:4.8"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[17734, 597, 105, 1, 53, "Subsection",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],

Cell[CellGroupData[{
Cell[17864, 602, 145, 3, 43, "Input",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
Cell[18012, 607, 704, 10, 265, "Print",
  CellTags->{"Info3268728758-5559249", "b:4.9"}],
Cell[18719, 619, 758, 12, 227, "Print",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[19514, 636, 320, 7, 103, "Input",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
Cell[19837, 645, 134, 3, 41, "Output",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[20008, 653, 171, 3, 43, "Input",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
Cell[20182, 658, 147, 3, 41, "Output",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[20366, 666, 163, 3, 43, "Input",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
Cell[20532, 671, 443, 7, 49, "Message",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
Cell[20978, 680, 175, 3, 41, "Output",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[21190, 688, 190, 4, 43, "Input",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}],
Cell[21383, 694, 254, 5, 60, "Output",
  CellTags->{"Triangular Decomposition", "i:13", "b:4.9"}]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[21698, 706, 85, 1, 70, "Section",
  CellTags->{"Complex Arrays", "b:5"}],

Cell[CellGroupData[{
Cell[21808, 711, 123, 3, 43, "Input",
  CellTags->{"Complex Arrays", "i:9", "b:5"}],
Cell[21934, 716, 323, 5, 75, "Print",
  CellTags->{"Info3268728760-5111011", "b:5", "Complex Arrays"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[22294, 726, 123, 3, 43, "Input",
  CellTags->{"Complex Arrays", "i:10", "b:5"}],
Cell[22420, 731, 416, 6, 113, "Print",
  CellTags->{"Info3268728760-3056648", "b:5", "Complex Arrays"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[22873, 742, 120, 3, 43, "Input",
  CellTags->{"Complex Arrays", "i:11", "b:5"}],
Cell[22996, 747, 695, 10, 246, "Print",
  CellTags->{"Info3268728760-5984139", "b:5", "Complex Arrays"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[23740, 763, 84, 1, 70, "Section",
  CellTags->{"Series Expansion", "i:16", "b:6"}],

Cell[CellGroupData[{
Cell[23849, 768, 107, 2, 45, "Input",
  CellTags->{"Series Expansion", "i:16", "b:6"}],
Cell[23959, 772, 257, 4, 94, "Print",
  CellTags->{"Info3268728760-6259689", "b:6", "Series Expansion"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[24253, 781, 100, 2, 45, "Input",
  CellTags->{"Series Expansion", "i:16", "b:6"}],
Cell[24356, 785, 368, 5, 132, "Print",
  CellTags->{"Info3268728760-6093397", "b:6", "Series Expansion"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[24773, 796, 78, 1, 70, "Section",
  CellTags->{"i:17", "Matlab Output", "b:7"}],

Cell[CellGroupData[{
Cell[24876, 801, 99, 2, 45, "Input",
  CellTags->{"i:17", "Matlab Output", "b:7"}],
Cell[24978, 805, 156, 3, 37, "Print",
  CellTags->{"i:17", "Matlab Output", "b:7"}],
Cell[25137, 810, 808, 11, 208, "Print",
  CellTags->{"Info3268728760-5557815", "b:7", "Matlab Output"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[25982, 826, 110, 2, 45, "Input",
  CellTags->{"i:17", "Matlab Output", "b:7"}],
Cell[26095, 830, 1201, 17, 664, "Print",
  CellTags->{"Info3268728762-6123324", "b:7", "Matlab Output"}]
}, Open  ]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

