(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.0'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     15183,        594]*)
(*NotebookOutlinePosition[     39180,       1143]*)
(*  CellTagsIndexPosition[     38432,       1113]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["Introduction", "Section",
  CellTags->"b:8"],

Cell[BoxData[{
    \(\(Off[General::spell];\)\), "\n", 
    \(\(Off[General::spell1];\)\)}], "Input",
  CellLabel->"In[1]:=",
  CellTags->"b:8"],

Cell[CellGroupData[{

Cell["<<ProPac`", "Input",
  CellLabel->"In[3]:=",
  CellTags->"b:8"],

Cell[BoxData[
    \("    This is ProPac - Development Version 3.xx\n \n ProPac has been \
successfully loaded. It contains several packages\n for multibody dynamics, \
linear control and nonlinear control.\n The package names are:\n    GeoTools\n\
    Dynamics\n    ControlL\n    ControlN\n    DifferentialForms\n    \
ObserverTools\n    NDTools\n    VSCTools\n    MatlabInterface\n    MEXTools\n \
Appropriate packages will be loaded automatically as required.\n To obtain a \
list of functions in a package and links to function\n declarations, enter:\n \
   ?ProPac`PackageName`*\n Help is also available in the Help Browser under \
Add-ons."\)], "Print",
  CellLabel->"From In[3]:=",
  GeneratedCell->False,
  CellAutoOverwrite->False,
  CellTags->"b:8"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Exterior Derivative", "Section",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],

Cell["\<\
The exterior derivative gives the total differential of a scalar expression \
(0-form) with all symbols treated as independent variables. The exterior \
derivative generalizes Curl.\
\>", "Text",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(d[x + y + z]\)], "Input",
  CellLabel->"In[4]:=",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],

Cell[BoxData[
    \(d[x] + d[y] + d[z]\)], "Output",
  CellLabel->"Out[4]=",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(d[\ h\ g]\)], "Input",
  CellLabel->"In[5]:=",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],

Cell[BoxData[
    \(h\ d[g] + g\ d[h]\)], "Output",
  CellLabel->"Out[5]=",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(d[h[x, y]\ g]\)], "Input",
  CellLabel->"In[6]:=",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],

Cell[BoxData[
    RowBox[{\(d[g]\ h[x, y]\), "+", 
      RowBox[{"g", " ", \(d[y]\), " ", 
        RowBox[{
          SuperscriptBox["h", 
            TagBox[\((0, 1)\),
              Derivative],
            MultilineFunction->None], "[", \(x, y\), "]"}]}], "+", 
      RowBox[{"g", " ", \(d[x]\), " ", 
        RowBox[{
          SuperscriptBox["h", 
            TagBox[\((1, 0)\),
              Derivative],
            MultilineFunction->None], "[", \(x, y\), "]"}]}]}]], "Output",
  CellLabel->"Out[6]=",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}]
}, Open  ]],

Cell["The following example is from Burke.", "Text",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(d[3\ x^2\ y\ d[y] + d[x]]\)], "Input",
  CellLabel->"In[7]:=",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],

Cell[BoxData[
    \(6\ x\ y\ d[x]\[Wedge]d[y]\)], "Output",
  CellLabel->"Out[7]=",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(d[%]\)], "Input",
  CellLabel->"In[8]:=",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],

Cell[BoxData[
    \(0\)], "Output",
  CellLabel->"Out[8]=",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Forms", "Section",
  CellTags->{"i:2", "b:9.1", "Forms"}],

Cell["Here are some examples of forms.", "Text",
  CellTags->{"i:2", "b:9.1", "Forms"}],

Cell[CellGroupData[{

Cell[BoxData[{
    \(w1 = y^2\ d[x] + d[y]\), "\[IndentingNewLine]", 
    \(w2 = 
      a\ d[x]\[Wedge]d[y] + 
        b\ x\ z\ d[x]\[Wedge]d[z]\), "\[IndentingNewLine]", 
    \(w3 = Exp[x + z]\ d[x]\[Wedge]d[y]\[Wedge]d[z]\)}], "Input",
  CellLabel->"In[9]:=",
  CellTags->{"i:2", "b:9.1", "Forms"}],

Cell[BoxData[
    \(y\^2\ d[x] + d[y]\)], "Output",
  CellLabel->"Out[9]=",
  CellTags->{"i:2", "b:9.1", "Forms"}],

Cell[BoxData[
    \(a\ d[x]\[Wedge]d[y] + b\ x\ z\ d[x]\[Wedge]d[z]\)], "Output",
  CellLabel->"Out[10]=",
  CellTags->{"i:2", "b:9.1", "Forms"}],

Cell[BoxData[
    \(\[ExponentialE]\^\(x + z\)\ d[x]\[Wedge]d[y]\[Wedge]d[z]\)], "Output",
  CellLabel->"Out[11]=",
  CellTags->{"i:2", "b:9.1", "Forms"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(FormDegree[w1]\), "\[IndentingNewLine]", 
    \(FormDegree[w2]\), "\[IndentingNewLine]", 
    \(FormDegree[w3]\)}], "Input",
  CellLabel->"In[12]:=",
  CellTags->{"i:2", "b:9.1", "Forms"}],

Cell[BoxData[
    \(1\)], "Output",
  CellLabel->"Out[12]=",
  CellTags->{"i:2", "b:9.1", "Forms"}],

Cell[BoxData[
    \(2\)], "Output",
  CellLabel->"Out[13]=",
  CellTags->{"i:2", "b:9.1", "Forms"}],

Cell[BoxData[
    \(3\)], "Output",
  CellLabel->"Out[14]=",
  CellTags->{"i:2", "b:9.1", "Forms"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(FormQ[w1]\)], "Input",
  CellLabel->"In[15]:=",
  CellTags->{"i:2", "b:9.1", "Forms"}],

Cell[BoxData[
    \(True\)], "Output",
  CellLabel->"Out[15]=",
  CellTags->{"i:2", "b:9.1", "Forms"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(FormQ[w1 + w2]\)], "Input",
  CellLabel->"In[16]:=",
  CellTags->{"i:2", "b:9.1", "Forms"}],

Cell[BoxData[
    \(False\)], "Output",
  CellLabel->"Out[16]=",
  CellTags->{"i:2", "b:9.1", "Forms"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Wedge Product", "Section",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}],

Cell["The wedge product generalizes the cross product.", "Text",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?Wedge\)\)], "Input",
  CellLabel->"In[17]:=",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}],

Cell[BoxData[
    \(" Wedge[v,w,..] returns the wedge product of the forms v, w,... "\)], \
"Print",
  CellTags->{"Info3268715208-6008824", "i:3", "b:9.2", "Wedge Product"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Wedge[w1, w2]\)], "Input",
  CellLabel->"In[18]:=",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}],

Cell[BoxData[
    \(\(-b\)\ x\ z\ d[x]\[Wedge]d[y]\[Wedge]d[z]\)], "Output",
  CellLabel->"Out[18]=",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}]
}, Open  ]],

Cell["or", "Text",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(w1\[Wedge]w2\)], "Input",
  CellLabel->"In[19]:=",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}],

Cell[BoxData[
    \(\(-b\)\ x\ z\ d[x]\[Wedge]d[y]\[Wedge]d[z]\)], "Output",
  CellLabel->"Out[19]=",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Wedge[w1, w1]\)], "Input",
  CellLabel->"In[20]:=",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}],

Cell[BoxData[
    \(0\)], "Output",
  CellLabel->"Out[20]=",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["FormQ[-Exp[x1+x2] d[x1]\[Wedge]d[x2]]", "Input",
  CellLabel->"In[21]:=",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}],

Cell[BoxData[
    \(True\)], "Output",
  CellLabel->"Out[21]=",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Form Basis", "Section",
  CellTags->{"i:4", "b:9.3", "Form Basis"}],

Cell[CellGroupData[{

Cell["\<\
FormBasis[{a x1^2 Wedge[d[x1],d[x2],d[x3],d[x4]],b \
Wedge[d[x2],d[x1],d[x3],d[x4]]},{x2,x1,x3,x4}]\
\>", "Input",
  CellLabel->"In[22]:=",
  CellTags->{"i:4", "b:9.3", "Form Basis"}],

Cell[BoxData[
    \({d[x1]\[Wedge]d[x2]\[Wedge]d[x3]\[Wedge]d[x4]}\)], "Output",
  CellLabel->"Out[22]=",
  CellTags->{"i:4", "b:9.3", "Form Basis"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
FormBasis[{Wedge[d[x3],d[\[Lambda]]]+Wedge[d[x1],d[x2]]},{x1,x2,x3,\[Lambda]}]\
\
\>", "Input",
  CellLabel->"In[23]:=",
  CellTags->{"i:4", "b:9.3", "Form Basis"}],

Cell[BoxData[
    \({d[x1]\[Wedge]d[x2] + d[x3]\[Wedge]d[\[Lambda]]}\)], "Output",
  CellLabel->"Out[23]=",
  CellTags->{"i:4", "b:9.3", "Form Basis"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Lie Derivative", "Section",
  CellTags->{"i:5", "b:9.4", "Lie Derivative"}],

Cell[CellGroupData[{

Cell["\<\
f={x1^2,2+x3,-x1 x2};w=x3 d[x1]+x2 x1 d[x2];x={x1,x2,x3};
FF=LieDerivative[f,w,x]\
\>", "Input",
  CellLabel->"In[24]:=",
  CellTags->{"i:5", "b:9.4", "Lie Derivative"}],

Cell[BoxData[
    \(\(-x1\)\ \((x2 - 2\ x3)\)\ d[x1] + x1\ \((2 + x1\ x2 + x3)\)\ d[x2] + 
      x1\ x2\ d[x3]\)], "Output",
  CellLabel->"Out[25]=",
  CellTags->{"i:5", "b:9.4", "Lie Derivative"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Codistribution Rank", "Section",
  CellTags->{"i:6", "b:9.5", "Codistribution Rank"}],

Cell["\<\
A codistribution is a list of one forms. RankCodistribution computes the \
generic rank.\
\>", "Text",
  CellTags->{"i:6", "b:9.5", "Codistribution Rank"}],

Cell[CellGroupData[{

Cell["RankCodistribution[{d[x1],d[x2],d[x1]+d[x3]},{x1,x2,x3}]", "Input",
  CellLabel->"In[26]:=",
  CellTags->{"i:6", "b:9.5", "Codistribution Rank"}],

Cell[BoxData[
    \(3\)], "Output",
  CellLabel->"Out[26]=",
  CellTags->{"i:6", "b:9.5", "Codistribution Rank"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Contraction", "Section",
  CellTags->{"i:7", "b:9.6", "Contraction"}],

Cell[CellGroupData[{

Cell["Definition", "Subsection",
  CellTags->{"i:7", "b:9.6", "Contraction"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?Contraction\)\)], "Input",
  CellLabel->"In[27]:=",
  CellTags->{"i:7", "b:9.6", "Contraction"}],

Cell[BoxData[
    \(" Contraction[w,v,x] returns the contraction of the form w with the\n \
vector v. x is the vector of coordinates. "\)], "Print",
  CellTags->{"Info3268715209-5953169", "i:7", "b:9.6", "Contraction"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Example from Olver", "Subsection",
  CellTags->{"i:7", "b:9.6", "Contraction"}],

Cell["Clear[x]", "Input",
  CellLabel->"In[28]:=",
  CellTags->{"i:7", "b:9.6", "Contraction"}],

Cell[CellGroupData[{

Cell["\<\
\[Omega] = \[Alpha] d[y]\[Wedge]d[z] + \[Beta] d[z]\[Wedge]d[x] + \[Gamma] \
d[x]\[Wedge]d[y];
vec1 = {\[Xi]1, \[Eta]1, \[Zeta]1}; 
vec2 = {\[Xi]2, \[Eta]2, \[Zeta]2};

\[Omega]a = Contraction[\[Omega], vec1, {x, y, z}]\
\>", "Input",
  CellLabel->"In[29]:=",
  CellTags->{"i:7", "b:9.6", "Contraction"}],

Cell[BoxData[
    \(\[Gamma]\ \((\(-\[Eta]1\)\ d[x] + \[Xi]1\ d[
                y])\) + \[Alpha]\ \((\(-\[Zeta]1\)\ d[y] + \[Eta]1\ d[
                z])\) - \[Beta]\ \((\(-\[Zeta]1\)\ d[x] + \[Xi]1\ d[
                z])\)\)], "Output",
  CellLabel->"Out[32]=",
  CellTags->{"i:7", "b:9.6", "Contraction"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Contraction[\[Omega]a, vec2, {x, y, z}]\
\>", "Input",
  CellLabel->"In[33]:=",
  CellTags->{"i:7", "b:9.6", "Contraction"}],

Cell[BoxData[
    \(\[Alpha]\ \((\[Zeta]2\ \[Eta]1 - \[Zeta]1\ \[Eta]2)\) - \[Beta]\ \((\
\[Zeta]2\ \[Xi]1 - \[Zeta]1\ \[Xi]2)\) + \[Gamma]\ \((\[Eta]2\ \[Xi]1 - \
\[Eta]1\ \[Xi]2)\)\)], "Output",
  CellLabel->"Out[33]=",
  CellTags->{"i:7", "b:9.6", "Contraction"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Example from Arnold", "Subsection",
  CellTags->{"i:7", "b:9.6", "Contraction"}],

Cell["\<\
form1 = d[x2]\[Wedge]d[x3]; 
form2 = x1 d[x3]\[Wedge]d[x2];
vec1 = {1, 1, 1}; vec2 = {1, 2, 3};
coords = {x1, x2, x3};\
\>", "Input",
  CellLabel->"In[34]:=",
  CellTags->{"i:7", "b:9.6", "Contraction"}],

Cell[CellGroupData[{

Cell["form1a = Contraction[form1, vec1, coords]", "Input",
  CellLabel->"In[38]:=",
  CellTags->{"i:7", "b:9.6", "Contraction"}],

Cell[BoxData[
    \(\(-d[x2]\) + d[x3]\)], "Output",
  CellLabel->"Out[38]=",
  CellTags->{"i:7", "b:9.6", "Contraction"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["formb = Contraction[form1a, vec2, coords]", "Input",
  CellLabel->"In[39]:=",
  CellTags->{"i:7", "b:9.6", "Contraction"}],

Cell[BoxData[
    \(1\)], "Output",
  CellLabel->"Out[39]=",
  CellTags->{"i:7", "b:9.6", "Contraction"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["form2a = Contraction[form2, vec1, coords]", "Input",
  CellLabel->"In[40]:=",
  CellTags->{"i:7", "b:9.6", "Contraction"}],

Cell[BoxData[
    \(\(-x1\)\ \((\(-d[x2]\) + d[x3])\)\)], "Output",
  CellLabel->"Out[40]=",
  CellTags->{"i:7", "b:9.6", "Contraction"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["form2b = Contraction[form2a, vec2, coords]", "Input",
  CellLabel->"In[41]:=",
  CellTags->{"i:7", "b:9.6", "Contraction"}],

Cell[BoxData[
    \(\(-x1\)\)], "Output",
  CellLabel->"Out[41]=",
  CellTags->{"i:7", "b:9.6", "Contraction"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
W={form1,form2};
FormBasis[W,coords]\
\>", "Input",
  CellLabel->"In[42]:=",
  CellTags->{"i:7", "b:9.6", "Contraction"}],

Cell[BoxData[
    \({d[x2]\[Wedge]d[x3]}\)], "Output",
  CellLabel->"Out[43]=",
  CellTags->{"i:7", "b:9.6", "Contraction"}]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Bibliography", "Section",
  CellTags->"b:10"],

Cell[TextData[{
  "V. I. Arnold, ",
  StyleBox["Geometrical Methods in the Theory of Ordinary Differential \
Equations",
    FontSlant->"Italic"],
  ". New York: Springer-Verlag, 1983.",
  "\n",
  "P. J. Olver, ",
  StyleBox["Applications of Lie Groups to Differential Equations",
    FontSlant->"Italic"],
  ". New York: Springer-Verlag, 1986."
}], "Text",
  CellTags->"b:10"]
}, Open  ]]
},
FrontEndVersion->"5.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 695}},
ShowPageBreaks->False,
WindowSize->{716, 629},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
WindowFrame->"Normal",
PrintingStartingPageNumber->23,
TaggingRules:>{"NextIndexingNumber" -> 8, "IndexEntries" -> {{"i:1", "d[ ]"}, \
{"i:2", "FormDegree"}, {"i:2", "FormQ"}, {"i:3", "Wedge"}, {"i:4", \
"FormBasis"}, {"i:5", "LieDerivative"}, {"i:6", "RankCodistribution"}, \
{"i:7", "Contraction"}}, "PaginationCache" -> {{2003, 8, 1, 21, 15, \
42.1415376`9.377285252079231}, {23, 23, 23, 23, 23, 23, 23, 23, 23, 23, 23, \
23, 23, 23, 23, 23, 23, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, 24, \
24, 25, 25, 25, 25, 25, 25, 25, 25, 25, 25, 25, 25, 25, 26, 26, 26, 26, 26, \
26, 26, 26, 26, 26, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 27, 28, 28, \
28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28, 28}}},
ShowCellTags->True,
StyleDefinitions -> "Textbook.nb"
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "b:8"->{
    Cell[1776, 53, 50, 1, 99, "Section",
      CellTags->"b:8"],
    Cell[1829, 56, 144, 4, 64, "Input",
      CellTags->"b:8"],
    Cell[1998, 64, 69, 2, 46, "Input",
      CellTags->"b:8"],
    Cell[2070, 68, 758, 13, 403, "Print",
      CellTags->"b:8"]},
  "i:1"->{
    Cell[2877, 87, 91, 1, 67, "Section",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[2971, 90, 258, 5, 56, "Text",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[3254, 99, 122, 3, 44, "Input",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[3379, 104, 129, 3, 38, "Output",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[3545, 112, 119, 3, 44, "Input",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[3667, 117, 128, 3, 38, "Output",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[3832, 125, 123, 3, 44, "Input",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[3958, 130, 562, 15, 38, "Output",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[4535, 148, 105, 1, 39, "Text",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[4665, 153, 135, 3, 44, "Input",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[4803, 158, 136, 3, 38, "Output",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[4976, 166, 114, 3, 44, "Input",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[5093, 171, 112, 3, 38, "Output",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}]},
  "b:9.0"->{
    Cell[2877, 87, 91, 1, 67, "Section",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[2971, 90, 258, 5, 56, "Text",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[3254, 99, 122, 3, 44, "Input",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[3379, 104, 129, 3, 38, "Output",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[3545, 112, 119, 3, 44, "Input",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[3667, 117, 128, 3, 38, "Output",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[3832, 125, 123, 3, 44, "Input",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[3958, 130, 562, 15, 38, "Output",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[4535, 148, 105, 1, 39, "Text",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[4665, 153, 135, 3, 44, "Input",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[4803, 158, 136, 3, 38, "Output",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[4976, 166, 114, 3, 44, "Input",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[5093, 171, 112, 3, 38, "Output",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}]},
  "Exterior Derivative"->{
    Cell[2877, 87, 91, 1, 67, "Section",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[2971, 90, 258, 5, 56, "Text",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[3254, 99, 122, 3, 44, "Input",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[3379, 104, 129, 3, 38, "Output",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[3545, 112, 119, 3, 44, "Input",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[3667, 117, 128, 3, 38, "Output",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[3832, 125, 123, 3, 44, "Input",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[3958, 130, 562, 15, 38, "Output",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[4535, 148, 105, 1, 39, "Text",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[4665, 153, 135, 3, 44, "Input",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[4803, 158, 136, 3, 38, "Output",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[4976, 166, 114, 3, 44, "Input",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
    Cell[5093, 171, 112, 3, 38, "Output",
      CellTags->{"i:1", "b:9.0", "Exterior Derivative"}]},
  "i:2"->{
    Cell[5254, 180, 63, 1, 99, "Section",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[5320, 183, 87, 1, 31, "Text",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[5432, 188, 300, 7, 62, "Input",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[5735, 197, 114, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[5852, 202, 145, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6000, 207, 154, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6191, 215, 209, 5, 62, "Input",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6403, 222, 99, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6505, 227, 99, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6607, 232, 99, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6743, 240, 106, 3, 34, "Input",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6852, 245, 102, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6991, 253, 111, 3, 34, "Input",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[7105, 258, 103, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}]},
  "b:9.1"->{
    Cell[5254, 180, 63, 1, 99, "Section",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[5320, 183, 87, 1, 31, "Text",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[5432, 188, 300, 7, 62, "Input",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[5735, 197, 114, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[5852, 202, 145, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6000, 207, 154, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6191, 215, 209, 5, 62, "Input",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6403, 222, 99, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6505, 227, 99, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6607, 232, 99, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6743, 240, 106, 3, 34, "Input",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6852, 245, 102, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6991, 253, 111, 3, 34, "Input",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[7105, 258, 103, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}]},
  "Forms"->{
    Cell[5254, 180, 63, 1, 99, "Section",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[5320, 183, 87, 1, 31, "Text",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[5432, 188, 300, 7, 62, "Input",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[5735, 197, 114, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[5852, 202, 145, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6000, 207, 154, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6191, 215, 209, 5, 62, "Input",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6403, 222, 99, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6505, 227, 99, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6607, 232, 99, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6743, 240, 106, 3, 34, "Input",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6852, 245, 102, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[6991, 253, 111, 3, 34, "Input",
      CellTags->{"i:2", "b:9.1", "Forms"}],
    Cell[7105, 258, 103, 3, 29, "Output",
      CellTags->{"i:2", "b:9.1", "Forms"}]},
  "i:3"->{
    Cell[7257, 267, 79, 1, 119, "Section",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[7339, 270, 111, 1, 31, "Text",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[7475, 275, 115, 3, 34, "Input",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[7593, 280, 173, 3, 29, "Print",
      CellTags->{"Info3268715208-6008824", "i:3", "b:9.2", "Wedge Product"}],
    Cell[7803, 288, 118, 3, 34, "Input",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[7924, 293, 148, 3, 29, "Output",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[8087, 299, 65, 1, 31, "Text",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[8177, 304, 117, 3, 34, "Input",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[8297, 309, 148, 3, 29, "Output",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[8482, 317, 118, 3, 34, "Input",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[8603, 322, 107, 3, 29, "Output",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[8747, 330, 126, 2, 35, "Input",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[8876, 334, 110, 3, 29, "Output",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}]},
  "b:9.2"->{
    Cell[7257, 267, 79, 1, 119, "Section",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[7339, 270, 111, 1, 31, "Text",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[7475, 275, 115, 3, 34, "Input",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[7593, 280, 173, 3, 29, "Print",
      CellTags->{"Info3268715208-6008824", "i:3", "b:9.2", "Wedge Product"}],
    Cell[7803, 288, 118, 3, 34, "Input",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[7924, 293, 148, 3, 29, "Output",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[8087, 299, 65, 1, 31, "Text",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[8177, 304, 117, 3, 34, "Input",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[8297, 309, 148, 3, 29, "Output",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[8482, 317, 118, 3, 34, "Input",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[8603, 322, 107, 3, 29, "Output",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[8747, 330, 126, 2, 35, "Input",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[8876, 334, 110, 3, 29, "Output",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}]},
  "Wedge Product"->{
    Cell[7257, 267, 79, 1, 119, "Section",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[7339, 270, 111, 1, 31, "Text",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[7475, 275, 115, 3, 34, "Input",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[7593, 280, 173, 3, 29, "Print",
      CellTags->{"Info3268715208-6008824", "i:3", "b:9.2", "Wedge Product"}],
    Cell[7803, 288, 118, 3, 34, "Input",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[7924, 293, 148, 3, 29, "Output",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[8087, 299, 65, 1, 31, "Text",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[8177, 304, 117, 3, 34, "Input",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[8297, 309, 148, 3, 29, "Output",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[8482, 317, 118, 3, 34, "Input",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[8603, 322, 107, 3, 29, "Output",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[8747, 330, 126, 2, 35, "Input",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}],
    Cell[8876, 334, 110, 3, 29, "Output",
      CellTags->{"i:3", "b:9.2", "Wedge Product"}]},
  "Info3268715208-6008824"->{
    Cell[7593, 280, 173, 3, 29, "Print",
      CellTags->{
      "Info3268715208-6008824", "i:3", "b:9.2", "Wedge Product"}]},
  "i:4"->{
    Cell[9035, 343, 73, 1, 119, "Section",
      CellTags->{"i:4", "b:9.3", "Form Basis"}],
    Cell[9133, 348, 193, 5, 49, "Input",
      CellTags->{"i:4", "b:9.3", "Form Basis"}],
    Cell[9329, 355, 149, 3, 29, "Output",
      CellTags->{"i:4", "b:9.3", "Form Basis"}],
    Cell[9515, 363, 174, 5, 35, "Input",
      CellTags->{"i:4", "b:9.3", "Form Basis"}],
    Cell[9692, 370, 151, 3, 29, "Output",
      CellTags->{"i:4", "b:9.3", "Form Basis"}]},
  "b:9.3"->{
    Cell[9035, 343, 73, 1, 119, "Section",
      CellTags->{"i:4", "b:9.3", "Form Basis"}],
    Cell[9133, 348, 193, 5, 49, "Input",
      CellTags->{"i:4", "b:9.3", "Form Basis"}],
    Cell[9329, 355, 149, 3, 29, "Output",
      CellTags->{"i:4", "b:9.3", "Form Basis"}],
    Cell[9515, 363, 174, 5, 35, "Input",
      CellTags->{"i:4", "b:9.3", "Form Basis"}],
    Cell[9692, 370, 151, 3, 29, "Output",
      CellTags->{"i:4", "b:9.3", "Form Basis"}]},
  "Form Basis"->{
    Cell[9035, 343, 73, 1, 119, "Section",
      CellTags->{"i:4", "b:9.3", "Form Basis"}],
    Cell[9133, 348, 193, 5, 49, "Input",
      CellTags->{"i:4", "b:9.3", "Form Basis"}],
    Cell[9329, 355, 149, 3, 29, "Output",
      CellTags->{"i:4", "b:9.3", "Form Basis"}],
    Cell[9515, 363, 174, 5, 35, "Input",
      CellTags->{"i:4", "b:9.3", "Form Basis"}],
    Cell[9692, 370, 151, 3, 29, "Output",
      CellTags->{"i:4", "b:9.3", "Form Basis"}]},
  "i:5"->{
    Cell[9892, 379, 81, 1, 119, "Section",
      CellTags->{"i:5", "b:9.4", "Lie Derivative"}],
    Cell[9998, 384, 179, 5, 49, "Input",
      CellTags->{"i:5", "b:9.4", "Lie Derivative"}],
    Cell[10180, 391, 197, 4, 29, "Output",
      CellTags->{"i:5", "b:9.4", "Lie Derivative"}]},
  "b:9.4"->{
    Cell[9892, 379, 81, 1, 119, "Section",
      CellTags->{"i:5", "b:9.4", "Lie Derivative"}],
    Cell[9998, 384, 179, 5, 49, "Input",
      CellTags->{"i:5", "b:9.4", "Lie Derivative"}],
    Cell[10180, 391, 197, 4, 29, "Output",
      CellTags->{"i:5", "b:9.4", "Lie Derivative"}]},
  "Lie Derivative"->{
    Cell[9892, 379, 81, 1, 119, "Section",
      CellTags->{"i:5", "b:9.4", "Lie Derivative"}],
    Cell[9998, 384, 179, 5, 49, "Input",
      CellTags->{"i:5", "b:9.4", "Lie Derivative"}],
    Cell[10180, 391, 197, 4, 29, "Output",
      CellTags->{"i:5", "b:9.4", "Lie Derivative"}]},
  "i:6"->{
    Cell[10426, 401, 91, 1, 119, "Section",
      CellTags->{"i:6", "b:9.5", "Codistribution Rank"}],
    Cell[10520, 404, 165, 4, 31, "Text",
      CellTags->{"i:6", "b:9.5", "Codistribution Rank"}],
    Cell[10710, 412, 151, 2, 35, "Input",
      CellTags->{"i:6", "b:9.5", "Codistribution Rank"}],
    Cell[10864, 416, 113, 3, 29, "Output",
      CellTags->{"i:6", "b:9.5", "Codistribution Rank"}]},
  "b:9.5"->{
    Cell[10426, 401, 91, 1, 119, "Section",
      CellTags->{"i:6", "b:9.5", "Codistribution Rank"}],
    Cell[10520, 404, 165, 4, 31, "Text",
      CellTags->{"i:6", "b:9.5", "Codistribution Rank"}],
    Cell[10710, 412, 151, 2, 35, "Input",
      CellTags->{"i:6", "b:9.5", "Codistribution Rank"}],
    Cell[10864, 416, 113, 3, 29, "Output",
      CellTags->{"i:6", "b:9.5", "Codistribution Rank"}]},
  "Codistribution Rank"->{
    Cell[10426, 401, 91, 1, 119, "Section",
      CellTags->{"i:6", "b:9.5", "Codistribution Rank"}],
    Cell[10520, 404, 165, 4, 31, "Text",
      CellTags->{"i:6", "b:9.5", "Codistribution Rank"}],
    Cell[10710, 412, 151, 2, 35, "Input",
      CellTags->{"i:6", "b:9.5", "Codistribution Rank"}],
    Cell[10864, 416, 113, 3, 29, "Output",
      CellTags->{"i:6", "b:9.5", "Codistribution Rank"}]},
  "i:7"->{
    Cell[11026, 425, 75, 1, 119, "Section",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[11126, 430, 77, 1, 49, "Subsection",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[11228, 435, 119, 3, 34, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[11350, 440, 219, 3, 43, "Print",
      CellTags->{"Info3268715209-5953169", "i:7", "b:9.6", "Contraction"}],
    Cell[11618, 449, 85, 1, 49, "Subsection",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[11706, 452, 95, 2, 35, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[11826, 458, 314, 9, 91, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[12143, 469, 310, 6, 29, "Output",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[12490, 480, 134, 4, 35, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[12627, 486, 266, 5, 29, "Output",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[12942, 497, 86, 1, 49, "Subsection",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[13031, 500, 213, 7, 77, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[13269, 511, 128, 2, 35, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[13400, 515, 122, 3, 29, "Output",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[13559, 523, 128, 2, 35, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[13690, 527, 105, 3, 29, "Output",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[13832, 535, 128, 2, 35, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[13963, 539, 137, 3, 29, "Output",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[14137, 547, 129, 2, 35, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[14269, 551, 111, 3, 29, "Output",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[14417, 559, 131, 5, 49, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[14551, 566, 124, 3, 29, "Output",
      CellTags->{"i:7", "b:9.6", "Contraction"}]},
  "b:9.6"->{
    Cell[11026, 425, 75, 1, 119, "Section",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[11126, 430, 77, 1, 49, "Subsection",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[11228, 435, 119, 3, 34, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[11350, 440, 219, 3, 43, "Print",
      CellTags->{"Info3268715209-5953169", "i:7", "b:9.6", "Contraction"}],
    Cell[11618, 449, 85, 1, 49, "Subsection",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[11706, 452, 95, 2, 35, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[11826, 458, 314, 9, 91, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[12143, 469, 310, 6, 29, "Output",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[12490, 480, 134, 4, 35, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[12627, 486, 266, 5, 29, "Output",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[12942, 497, 86, 1, 49, "Subsection",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[13031, 500, 213, 7, 77, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[13269, 511, 128, 2, 35, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[13400, 515, 122, 3, 29, "Output",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[13559, 523, 128, 2, 35, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[13690, 527, 105, 3, 29, "Output",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[13832, 535, 128, 2, 35, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[13963, 539, 137, 3, 29, "Output",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[14137, 547, 129, 2, 35, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[14269, 551, 111, 3, 29, "Output",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[14417, 559, 131, 5, 49, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[14551, 566, 124, 3, 29, "Output",
      CellTags->{"i:7", "b:9.6", "Contraction"}]},
  "Contraction"->{
    Cell[11026, 425, 75, 1, 119, "Section",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[11126, 430, 77, 1, 49, "Subsection",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[11228, 435, 119, 3, 34, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[11350, 440, 219, 3, 43, "Print",
      CellTags->{"Info3268715209-5953169", "i:7", "b:9.6", "Contraction"}],
    Cell[11618, 449, 85, 1, 49, "Subsection",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[11706, 452, 95, 2, 35, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[11826, 458, 314, 9, 91, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[12143, 469, 310, 6, 29, "Output",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[12490, 480, 134, 4, 35, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[12627, 486, 266, 5, 29, "Output",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[12942, 497, 86, 1, 49, "Subsection",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[13031, 500, 213, 7, 77, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[13269, 511, 128, 2, 35, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[13400, 515, 122, 3, 29, "Output",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[13559, 523, 128, 2, 35, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[13690, 527, 105, 3, 29, "Output",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[13832, 535, 128, 2, 35, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[13963, 539, 137, 3, 29, "Output",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[14137, 547, 129, 2, 35, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[14269, 551, 111, 3, 29, "Output",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[14417, 559, 131, 5, 49, "Input",
      CellTags->{"i:7", "b:9.6", "Contraction"}],
    Cell[14551, 566, 124, 3, 29, "Output",
      CellTags->{"i:7", "b:9.6", "Contraction"}]},
  "Info3268715209-5953169"->{
    Cell[11350, 440, 219, 3, 43, "Print",
      CellTags->{"Info3268715209-5953169", "i:7", "b:9.6", "Contraction"}]},
  "b:10"->{
    Cell[14736, 576, 51, 1, 119, "Section",
      CellTags->"b:10"],
    Cell[14790, 579, 377, 12, 52, "Text",
      CellTags->"b:10"]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"b:8", 16558, 623},
  {"i:1", 16828, 632},
  {"b:9.0", 18131, 659},
  {"Exterior Derivative", 19448, 686},
  {"i:2", 20749, 713},
  {"b:9.1", 21957, 742},
  {"Forms", 23165, 771},
  {"i:3", 24371, 800},
  {"b:9.2", 25623, 827},
  {"Wedge Product", 26883, 854},
  {"Info3268715208-6008824", 28152, 881},
  {"i:4", 28290, 885},
  {"b:9.3", 28758, 896},
  {"Form Basis", 29231, 907},
  {"i:5", 29697, 918},
  {"b:9.4", 29997, 925},
  {"Lie Derivative", 30306, 932},
  {"i:6", 30604, 939},
  {"b:9.5", 31020, 948},
  {"Codistribution Rank", 31450, 957},
  {"i:7", 31864, 966},
  {"b:9.6", 33948, 1011},
  {"Contraction", 36038, 1056},
  {"Info3268715209-5953169", 38139, 1101},
  {"b:10", 38270, 1104}
  }
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 50, 1, 99, "Section",
  CellTags->"b:8"],
Cell[1829, 56, 144, 4, 64, "Input",
  CellTags->"b:8"],

Cell[CellGroupData[{
Cell[1998, 64, 69, 2, 46, "Input",
  CellTags->"b:8"],
Cell[2070, 68, 758, 13, 403, "Print",
  CellTags->"b:8"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{
Cell[2877, 87, 91, 1, 67, "Section",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
Cell[2971, 90, 258, 5, 56, "Text",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],

Cell[CellGroupData[{
Cell[3254, 99, 122, 3, 44, "Input",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
Cell[3379, 104, 129, 3, 38, "Output",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[3545, 112, 119, 3, 44, "Input",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
Cell[3667, 117, 128, 3, 38, "Output",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[3832, 125, 123, 3, 44, "Input",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
Cell[3958, 130, 562, 15, 38, "Output",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}]
}, Open  ]],
Cell[4535, 148, 105, 1, 39, "Text",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],

Cell[CellGroupData[{
Cell[4665, 153, 135, 3, 44, "Input",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
Cell[4803, 158, 136, 3, 38, "Output",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[4976, 166, 114, 3, 44, "Input",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}],
Cell[5093, 171, 112, 3, 38, "Output",
  CellTags->{"i:1", "b:9.0", "Exterior Derivative"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[5254, 180, 63, 1, 99, "Section",
  CellTags->{"i:2", "b:9.1", "Forms"}],
Cell[5320, 183, 87, 1, 31, "Text",
  CellTags->{"i:2", "b:9.1", "Forms"}],

Cell[CellGroupData[{
Cell[5432, 188, 300, 7, 62, "Input",
  CellTags->{"i:2", "b:9.1", "Forms"}],
Cell[5735, 197, 114, 3, 29, "Output",
  CellTags->{"i:2", "b:9.1", "Forms"}],
Cell[5852, 202, 145, 3, 29, "Output",
  CellTags->{"i:2", "b:9.1", "Forms"}],
Cell[6000, 207, 154, 3, 29, "Output",
  CellTags->{"i:2", "b:9.1", "Forms"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[6191, 215, 209, 5, 62, "Input",
  CellTags->{"i:2", "b:9.1", "Forms"}],
Cell[6403, 222, 99, 3, 29, "Output",
  CellTags->{"i:2", "b:9.1", "Forms"}],
Cell[6505, 227, 99, 3, 29, "Output",
  CellTags->{"i:2", "b:9.1", "Forms"}],
Cell[6607, 232, 99, 3, 29, "Output",
  CellTags->{"i:2", "b:9.1", "Forms"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[6743, 240, 106, 3, 34, "Input",
  CellTags->{"i:2", "b:9.1", "Forms"}],
Cell[6852, 245, 102, 3, 29, "Output",
  CellTags->{"i:2", "b:9.1", "Forms"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[6991, 253, 111, 3, 34, "Input",
  CellTags->{"i:2", "b:9.1", "Forms"}],
Cell[7105, 258, 103, 3, 29, "Output",
  CellTags->{"i:2", "b:9.1", "Forms"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[7257, 267, 79, 1, 119, "Section",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}],
Cell[7339, 270, 111, 1, 31, "Text",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}],

Cell[CellGroupData[{
Cell[7475, 275, 115, 3, 34, "Input",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}],
Cell[7593, 280, 173, 3, 29, "Print",
  CellTags->{"Info3268715208-6008824", "i:3", "b:9.2", "Wedge Product"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[7803, 288, 118, 3, 34, "Input",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}],
Cell[7924, 293, 148, 3, 29, "Output",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}]
}, Open  ]],
Cell[8087, 299, 65, 1, 31, "Text",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}],

Cell[CellGroupData[{
Cell[8177, 304, 117, 3, 34, "Input",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}],
Cell[8297, 309, 148, 3, 29, "Output",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[8482, 317, 118, 3, 34, "Input",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}],
Cell[8603, 322, 107, 3, 29, "Output",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[8747, 330, 126, 2, 35, "Input",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}],
Cell[8876, 334, 110, 3, 29, "Output",
  CellTags->{"i:3", "b:9.2", "Wedge Product"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[9035, 343, 73, 1, 119, "Section",
  CellTags->{"i:4", "b:9.3", "Form Basis"}],

Cell[CellGroupData[{
Cell[9133, 348, 193, 5, 49, "Input",
  CellTags->{"i:4", "b:9.3", "Form Basis"}],
Cell[9329, 355, 149, 3, 29, "Output",
  CellTags->{"i:4", "b:9.3", "Form Basis"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[9515, 363, 174, 5, 35, "Input",
  CellTags->{"i:4", "b:9.3", "Form Basis"}],
Cell[9692, 370, 151, 3, 29, "Output",
  CellTags->{"i:4", "b:9.3", "Form Basis"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[9892, 379, 81, 1, 119, "Section",
  CellTags->{"i:5", "b:9.4", "Lie Derivative"}],

Cell[CellGroupData[{
Cell[9998, 384, 179, 5, 49, "Input",
  CellTags->{"i:5", "b:9.4", "Lie Derivative"}],
Cell[10180, 391, 197, 4, 29, "Output",
  CellTags->{"i:5", "b:9.4", "Lie Derivative"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[10426, 401, 91, 1, 119, "Section",
  CellTags->{"i:6", "b:9.5", "Codistribution Rank"}],
Cell[10520, 404, 165, 4, 31, "Text",
  CellTags->{"i:6", "b:9.5", "Codistribution Rank"}],

Cell[CellGroupData[{
Cell[10710, 412, 151, 2, 35, "Input",
  CellTags->{"i:6", "b:9.5", "Codistribution Rank"}],
Cell[10864, 416, 113, 3, 29, "Output",
  CellTags->{"i:6", "b:9.5", "Codistribution Rank"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[11026, 425, 75, 1, 119, "Section",
  CellTags->{"i:7", "b:9.6", "Contraction"}],

Cell[CellGroupData[{
Cell[11126, 430, 77, 1, 49, "Subsection",
  CellTags->{"i:7", "b:9.6", "Contraction"}],

Cell[CellGroupData[{
Cell[11228, 435, 119, 3, 34, "Input",
  CellTags->{"i:7", "b:9.6", "Contraction"}],
Cell[11350, 440, 219, 3, 43, "Print",
  CellTags->{"Info3268715209-5953169", "i:7", "b:9.6", "Contraction"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[11618, 449, 85, 1, 49, "Subsection",
  CellTags->{"i:7", "b:9.6", "Contraction"}],
Cell[11706, 452, 95, 2, 35, "Input",
  CellTags->{"i:7", "b:9.6", "Contraction"}],

Cell[CellGroupData[{
Cell[11826, 458, 314, 9, 91, "Input",
  CellTags->{"i:7", "b:9.6", "Contraction"}],
Cell[12143, 469, 310, 6, 29, "Output",
  CellTags->{"i:7", "b:9.6", "Contraction"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[12490, 480, 134, 4, 35, "Input",
  CellTags->{"i:7", "b:9.6", "Contraction"}],
Cell[12627, 486, 266, 5, 29, "Output",
  CellTags->{"i:7", "b:9.6", "Contraction"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[12942, 497, 86, 1, 49, "Subsection",
  CellTags->{"i:7", "b:9.6", "Contraction"}],
Cell[13031, 500, 213, 7, 77, "Input",
  CellTags->{"i:7", "b:9.6", "Contraction"}],

Cell[CellGroupData[{
Cell[13269, 511, 128, 2, 35, "Input",
  CellTags->{"i:7", "b:9.6", "Contraction"}],
Cell[13400, 515, 122, 3, 29, "Output",
  CellTags->{"i:7", "b:9.6", "Contraction"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[13559, 523, 128, 2, 35, "Input",
  CellTags->{"i:7", "b:9.6", "Contraction"}],
Cell[13690, 527, 105, 3, 29, "Output",
  CellTags->{"i:7", "b:9.6", "Contraction"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[13832, 535, 128, 2, 35, "Input",
  CellTags->{"i:7", "b:9.6", "Contraction"}],
Cell[13963, 539, 137, 3, 29, "Output",
  CellTags->{"i:7", "b:9.6", "Contraction"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[14137, 547, 129, 2, 35, "Input",
  CellTags->{"i:7", "b:9.6", "Contraction"}],
Cell[14269, 551, 111, 3, 29, "Output",
  CellTags->{"i:7", "b:9.6", "Contraction"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[14417, 559, 131, 5, 49, "Input",
  CellTags->{"i:7", "b:9.6", "Contraction"}],
Cell[14551, 566, 124, 3, 29, "Output",
  CellTags->{"i:7", "b:9.6", "Contraction"}]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[14736, 576, 51, 1, 119, "Section",
  CellTags->"b:10"],
Cell[14790, 579, 377, 12, 52, "Text",
  CellTags->"b:10"]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

