(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.0'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    551508,      10976]*)
(*NotebookOutlinePosition[    655993,      13406]*)
(*  CellTagsIndexPosition[    654519,      13356]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["Introduction", "Section",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Introduction", "b:11"}],

Cell[BoxData[{
    \(\(Off[General::spell];\)\), "\n", 
    \(\(Off[General::spell1];\)\)}], "Input",
  CellLabel->"In[1]:=",
  CellTags->"b:11"],

Cell[CellGroupData[{

Cell[BoxData[
    \(SetDirectory["\<C:\Projects\ProPac\ProPac_3\Notebooks\>"]\)], "Input",
  CellLabel->"In[3]:=",
  CellTags->"b:11"],

Cell[BoxData[
    \("C:\\Projects\\ProPac\\ProPac_3\\Notebooks"\)], "Output",
  CellLabel->"Out[3]=",
  CellTags->"b:11"]
}, Open  ]],

Cell["\<\
TSi ProPac: Dynamics is a Mathematica package containing functions that \
support the assembly of mathematical models and simulation models for \
mechanical systems.  In its current form, Dynamics accommodates tree \
structures composed of combinations of rigid and flexible bodies and simple \
and compound kinematic joints, and which may also have supplemental algebraic \
and/or differential constraints. Thus, systems with closed loops and \
nonholonomic joints can be treated. 

Dynamics creates fully nonlinear, explicit models. The model building process \
requires that the user provide defining data for individual joints and \
bodies, as well as the system interconnection structure. The joints are \
defined in terms of their primitive action parameters from which all the \
required kinematic relations are derived. Thus, a user can contrive unusual \
joint configurations and is not restricted to a predefined set of standard \
joints. The equations are formulated in Poincare's form of Lagrange's \
equations that admits the standard Lagrange equations as a special case. \
However, Poincare's form allows the exploitation of quasi-velocities which \
can significantly simplify the equations of motion.

Dynamics assembles the equations of motion and it contains functions that \
facilitate simulation. Simulation may done within Mathematica  or externally \
using some other simulation software. Dynamics includes functions that assist \
in performing simulations using the NDSolve function in Mathematica and other \
functions to construct a C-code subroutine that compiles as an S-Function for \
use with MATLAB/SIMULINK.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Introduction", "b:11"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Getting Started", "Section",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Getting Started", "b:12"}],

Cell[TextData[{
  "This notebook illustrates the tools provided in Dynamics including a few \
new functions contained in Version 3.0. Dynamics is contained in the \
directory ProPac. ProPac can be placed anywhwere you like, but a search path \
must be defined for it. It is most convenient to place ProPac in the \
AddOns\\Applications directiory, i.e, C:\\Program Files\\Wolfram Research\\",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "\\3.0\\AddOns\\Applications. Once ProPac isloaded all of the functions in \
Dynamics as well as other packages included in TSi ProPac will be available \
for use and loaded as when needed."
}], "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Getting Started", "b:12"}],

Cell[TextData[{
  "In ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " 3 ProPac can be loaded with the command"
}], "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Getting Started", "b:12"}],

Cell[CellGroupData[{

Cell["<<ProPac`", "Input",
  CellLabel->"In[4]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Getting Started", "b:12"}],

Cell[BoxData[
    \("\n    This is ProPac - Development Version 3.xx\n \n ProPac has been \
successfully loaded. It contains several packages\n for multibody dynamics, \
linear control and nonlinear control.\n The package names are:\n    GeoTools\n\
    Dynamics\n    ControlL\n    ControlN\n    DifferentialForms\n    \
ObserverTools\n    NDTools\n    VSCTools\n    MatlabInterface\n    MEXTools\n \
Appropriate packages will be loaded automatically as required.\n To obtain a \
list of functions in a package and links to function\n declarations, enter:\n \
   ?ProPac`PackageName`*\n Help is also available in the Help Browser under \
Add-ons."\)], "Print",
  CellLabel->"From In[4]:=",
  CellTags->"b:12"]
}, Open  ]],

Cell[TextData[{
  "With ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " 2.2 use the command ProPac`Master`."
}], "Text",
  CellTags->{"Getting Started", "b:12"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["The Basics of TSi Dynamics", "Section",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell["What's in the Package?", "Subsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"What's in the Package", "b:13.0"}],

Cell["\<\
To find out what functions are available in Dynamics simply use the commands:\
\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"What's in the Package", "b:13.0"}],

Cell[CellGroupData[{

Cell["Dynamics", "Input",
  CellLabel->"In[5]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"What's in the Package", "b:13.0"}],

Cell[BoxData[
    \("  *** Dynamics successfully loaded ***"\)], "Print",
  CellLabel->"From In[5]:=",
  CellTags->"b:13.0"],

Cell[BoxData[
    \(Dynamics\)], "Output",
  CellLabel->"Out[5]=",
  CellTags->"b:13.0"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?Dynamics\)\)], "Input",
  CellLabel->"In[6]:=",
  CellTags->{"What's in the Package", "b:13.0"}],

Cell[BoxData[
    \("Functions included in the Dynamics package are:\n \n Kinematics:\n \
RotationMatrixEuler, ConfigurationMatrixEuler,\n \
AToATilda,ATildaToA,SimpleJointKinematics,\n \
CompoundJointKinematics,CompoundJointMap,Joints,\n \
JointRotation,JointTranslation,CompoundJointConfiguration,\n \
SimpleJointConfiguration,RotationMatrixToEuler,\n \
EndEffector,RelativeConfiguration,\n NodeVelocity,KinematicReplacements,\n \n \
Generalized Forces & Potential Functions:\n \
GeneralizedForce,DamperForce,SpringForce,GravPotential,\n \
SpringPotential,DamperPotential,LeafPotential,\n \
RayleighDissipationForce,BacklashPotential,\n \
BacklashForce,JointFrictionPotential,\n FlxDissPot,FlxStrnPot,\n \n Inertial \
Parameters: RgdBdyInrShift,ChainInertia,TreeInertia,\n \n Model Assembly:\n \
CMatrix,CreateModelSpecial,CreateModel,\n \
PoincareCoefficients,PoincareFunctionCombined,PoincareFunction,\n \
AlgebraicConstraints,DifferentialConstraints,\n MakeODEs,StateTransformation, \
MakeLagrangeEquations,\n CreateModelMEX,FindMostFrequentTerms,\n \n \
Miscellaneous:\n SeriesExpansion,Truncate,MatlabForm\n "\)], "Print",
  CellTags->{"Info3248853457-1624643", "b:13.0"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?GeneralizedForce\)\)], "Input",
  CellLabel->"In[7]:=",
  CellTags->{"What's in the Package", "b:13.0"}],

Cell[BoxData[
    \("Q=GeneralizedForce[ChnLst,TerminalNode,BodyLst,X,H,q,p,Force,VelNames],\
\n where\n \n   ChnLst and TerminalNode define the location of the point at \
which the\n   external force is applied. ChnLst defines a chain and \
TerminalNode is\n   the number of the node of interest in the last body of \
the chain.\n \n   BodyLst is the list of system body data.\n \n   X,H,q,p are \
the usual lists of joint data.\n \n   F is a list of 6 expressions which \
defines the external torques (first\n   three) and forces (last three) in \
terms of body velocities (velocities\n   of a body fixed frame at the \
terminal node).\n \n   VelNames is a list of (6) names of the velocities used \
in the expressions\n   F. There must be six names - the first 3 corresponding \
to the angular\n   velocity and the last 3 to the linear velocity.\n \n   Q \
is the generalized force vector.\n \n GeneralizedForce sets up the \
appropriate data structures to use\n NodeVelocity\n \n An alternative calling \
syntax is\n \n \
Q=GeneralizedForce[TerminalNode,TreeLst,BodyLst,X,H,q,p,Force,VelNames]\n \n \
in which case the appropriate chain is identified from the system data"\)], \
"Print",
  CellTags->{"Info3248853457-6771253", "b:13.0"}]
}, Open  ]],

Cell["\<\
Information about any of these functions can be obtained in the usual way \
with ?Function. Most users will find that only a few of these functions are \
necessary for creating models and simulations. The most important are:
   Joints: generates all kinematic functions for list of joints
   CreateModel: creates the dynamic model parameters for tree structures from \
system data   
  EndEffector & NodeVelocity: creates the position, orientation and velocity \
of a body frame located at specified nodes as functions of configuration \
coordinates
   SpringPotential & LeafPotential: the former creates the potential energy \
function of a spring (defined by a local energy function) connected between \
specified nodes, the latter performs a similar function for energy storage \
elements that act between a system node and the space frame
   GeneralizedForce, Spring & DamperForce: the former is a general function \
for forming the generalized force vector from a local force acting at a \
specified node, and the latter two are specific for a spring (defined by a \
local potential function) and a damper (defined by a local disspation \
function) connected between specified system nodes
   AlgebraicConstraints & DifferentialConstraints: these account for \
algebraic and differential constraints added to a tree structure
   MakeODEs & CreateModelMEX: the former assembles the ordinary differential \
equations from the model parmeters generated by other functions into a form \
for numerical computation within Mathematica, the latter builds a C source \
function for computation in MATLAB/SIMULINK.
After describing the basic data structures used by TSi Dynamics we will \
illustrate the use of these and some other functions.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"What's in the Package", "b:13.0"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Data Structures", "Subsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}],

Cell[CellGroupData[{

Cell["Tree Structures", "Subsubsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}],

Cell["\<\
Dynamics builds models for mechanical systems that have an underlying tree \
topology. Chain structures are a special case. Systems with closed loops are \
accommodated by adding constraints to the underlying tree.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}],

Cell[GraphicsData["Metafile", "\<\
CF5dJ6E]HGAYHf4PEfU^I6mgLb15CDHPAVmbKF5d0@00068l0@0006`0000200000P000>81003H0@00
000000000004>`00W3T00215CDH00040?68006P100090000000000000000000000@0000300100@00
l0000000000000000000003R1020Z@<0AP0002`0000P0000ADe6:`500@0L00004000004@`=/10000
H000060000160000>3T002`i0015CDH[<4020100000400000020?am0100<0000000001i00@0<0000
000001e0000D000020000000003G0@008D0700`0000000009D00010000040000000002a0000T0000
600006HimSh000200000P2EDl3iKSj8oFhnR?`Q000F0>000M3P0004@`=/200000`0006@h00010000
K000000000010000A00004<00000000000000588000O200084E=AP000@1T>000>P4000L000000000
000000000000100000<00401003`000000000000000000000>84082Y0`16000001H00?@E0017A4U3
0@00P003003eNd^900000=`E000100T000?^2P001`3F0000000500002`8000001@0000`27`QB20D0
00090_ooo`0500000@8000001@0000T2_moo00D000010P00000400000P4200D000010P0000070000
o0800;oOo`2I:@@0000]0@002@000?X20001000000000280100002d10@3F000090=Y00P0H`L:05P7
401=1aT0@PLW03P7>00]1dd08`MU01T7P@0?1j001PO30?d6j@3d1Q81j`Hn0N@6K07L1Yd1e@KA0Ll6
1P;91Sh2``Ig0[l6/P:k1^l2]`H/0k@6J`>b1ZX3/0KY0jl6:@B_1VT4[`JY1;06j0Bb1RH5]0IT1KL6
X0Fk1]/5_`HE1/<6C0K91X86c`Jf1]D6i`KL1QD7i0I11n/6JPOd1Y07o@Jb1`H7dPL?1nh76@L622<7
6`P]1b`8>0Lj2487@`Q=1dT8F0M;26<7B@Q^1d<8N@Lj28<7;0R>1a/8V0L62:<7kPN]1m87]PNb1l07
T0O91fX7dPM11mX75@OR1nL6jPNf1_47PPKg1d`6o@LE1P88f`D72:052`QT1@l89PDB2>P450RY11H8
J@@F22T45`SY0aH8ZP<F26/350P/0a88k`8?2;822`Qg0PL8?P8220H2o@OA0OL7W@7a1f`1jPLn0N87
4P7J1nT0dPO30<T7X0301h40]PMU0:d7C@2S1cP0V0LW08h76@231a00N@L:06h7201S1`T0003j0P00
00000000000R00@0000]0@801`000?`20@00000000040000;@4300@0003`0@0010000?010@050000
2@;oool01@000042000000@000020@801@000042000000L0003l0P00oooo0000100002d100090000
nP8000/0000000008P040000;@4107h0000T0cd0>00F1S80l@D^0<`5;@2W1Bd0P`D`0645=@101C/0
8@E400@5CP3Z15X0d`AX0<04M`2`18L0X`BH09X4ZP2E1;d0U0CA09H4i02M1?P0Y`@;0KD47P761341
fPA30O84D`4=1F<1:PEa0DX5OP5[1HT1SPFC0K<5VP7G1J01o@FS0B86Y@571Z@1JPJR0Hd6W@6^1YH1
c@J>0NT6Q0431gP16PMZ0Bh7F`4n1d/1B`Lj0E@79`5I1a@1F`L10EP7kP1B1mX0A`O603T7/`0X1j40
4`N?0?/6OP3P1Vl0``IP0:<6E0221TP0G`Ho03/6>00F1P@0000]0@80100002d10`040000l04100D0
00090P00P00500000@800000100000810P0500000@8000001`000?`20000080000040000;@4100T0
003j0P000@000000000R00@0000]0@@0<P0002@35`3o04`70@501`H1=0L?0BX76P4S1bH180Lc0B07
?`4S1dX1:PMB0C@7F0501eT1C0MH0ET7DP5U1dX1KPLo0GD7<`5i1bH1N@LJ0GD73`5^1`H1I@L10ET7
o`1<1`@0000]0@80100002d10`040000l04400D000090_ooo`0500000@800000100000810P050000
0@800000100002d100090000nP8000/0000000008P040000;@4409X0000T0d/0d04b0o414P<B0_<2
<`;E0U@2^09f0Yd2UP:40[H2K0;E0UL2l`940Pl3=08Y0bD2@P<J0UT34@9^0`/2P0<80Y0320:N0`X2
Z@<@0[4360:f0b<2^@<`0[P3@0:e0e<2[`=W0ZH3OP:K0iL2S@>b0W`3cP9Y0n`2E0<:0cd3:P<T0d/3
20=/0n`2S@?>0Zl3[`;@0hl2l@=^0Q44C@8`12/2CP@:0VX4j@651<P1W`BY0KH4RP7:16`1g@A@0Nh4
=@7l11`11`D50A05l@0F1Mh06@G>01X5`@0G1KH04PF^00X5Z03o1:H0l@BV0>44Z@3?1:l0^PBh0:@4
``2;1=80L@CR05D4m@0g10X160@R0OP3>`7H0eH1]P=c0ID3T@5d0k01DP?@0C83100002d10P040000
;@4300@0003`0@@01@0000T2002000D000010P00000400000P4200D000010P0000040000;@4100T0
003j0P000@000000000R00@0000]0@@0<P0002@35`2?01X5T@0=1IH00@FN0?P4Z@3a1;D0k@C20>d4
c`3a1=T0n0CR0045i`0=1NT06PGW02H5hP0b1MT0?0G?04<5`P161KD0APFY04<5WP0l1IH0<PFA02H5
S`0J1@@0000]0@80100002d10`040000l04400D000090_ooo`0500000@800000100000810P050000
0@800000100002d100090000nP8000/0000000008P040000;@440780000T0cL0:P<Y0C43204i0nL0
@P?804d3ZP1H0hh0I@=e0783GP200dX0SP<j09`3;@2Z0b<0^0<N0<D3703B0ah0gP<U0>T3;P3c0c`0
o0==00@4H@0:17T03PBC0144[`0C1<d04PC]01043@4=12l120A@0@44LP7i0i81l0>b0ND3d07J0n/1
c@<50/0370:b0c02Y0=00YH3C@:80eH2NP=L0Vd3GP9P0e/2E0=E0TT3B`8n0ch2=P<]0Rh3608X0`42
8`?W0B03b`4O0jd17`>=0B43K04U0d/1:P<Y0@@0000]0@80100002d10`040000l04400D000090P00
P00500000@800000100000810P0500000@800000100002d10@090000nP800040000000008P040000
;@440380000T0aL0<P=Q0S@3E@8i0dT2@@<o0T`3>09H0cD2I@<e0W83>09l0cl2Q@=90XX3E@:<0f42
RP=^0XD3NP9l0h<2LP>:0VD3SP9H0hh2C0>:0T43P`8i0gX2=0=^0S83H@840000;@4200@0000]0@<0
10000?01100500002@;oool01@000042000000@000020@801@000042000000@0000]0@002@000?X2
000;000000000280100002d1102B000090=706l2R@=c0W`3N`9_0hL2H`>F0UP3Z09>0kh2A@?G0Sh3
l`8g0a83<P<c0bl3EP<]0g/3;0>R0b`3bP<^0o<3<P<L13L3AP@m0g04A0>I14d3`PAF0nX4H@<@1Fd3
=0Ej0eL5Q`=g1ID3U@FT0k05/`?81L83g@GB0nl5hP?m1O8320H110l640@B1Ql44@H]10d6>`@51TP4
n@ED1>X5G`CG1FT4`@Ea1:P5N@B<1Gl4KPF414d5R0@Y1HX410F;1=h4RPBf18P4S@B516<4P0@i17X4
40Ac1>H3JPBn0f04UP=E1703BPA;0cd4:@<`10P38@CZ0Q<4c`841;L2m0>R0^@3T@;E0h<2a@=h0[D3
L@:V0fh2U`=_0XT3100002d10P040000;@4300@0003`0@@01@0000T2002000D000010P0000040000
0P4200D000010P0000040000;@4100T0003j0P000@000000000R00@0000]0@@0<P0002@35`0j0XH3
?09i0d42K@=:0V@3E@9M0f42F@=^0UT3NP9M0hD2I0>=0Vd3T`9i0i@2QP>C0Y83S@:N0hD2Z0=j0Zl3
KP:b0f42/P=E0Zl3BP:X0d42WP<l0Y83>P:60`@0000]0@80100002d10`040000l04400D000090_oo
o`0500000@800000100000810P0500000@800000100002d100090000nP8000/0000000008P040000
;@4406X0000T0c<0=@@K14845PAA11@4H@@E17<46PB512<4V@@^1:d4?@C014l4e0AS1>L4N@Ci1984
2PF/11T5aP@W1N44<`Gm13d56@E51C@5BPE>1Dd5IPE>1Gd5C0FA1DL5X`E01K85=`Fn1B`5a`DO1L`5
40G>1@05c0G^1<L5f`Bo1LP4/`Fd1:@5X0BC1Hd4OPEj16P5J0A@1EL4=PE711/5>@Co12d4h`@S1<P4
70B^11H4U0@C17/44`AU11D4D0@I13l480@`12T490@e11/4100002d10P040000;@4300@0003`0@@0
1@0000T2002000D000010P00000400000P4200D000010P0000040000;@4100T0003j0P000@000000
000R00@0000]0@@0<P0002@35`0C12<45@@G11X42`@R1044;@Cj0cT4m`=61?L3D`Cj0ed40@AV10/4
J`@G16d48`A[1304IP@l15d4A@AC14`4APA@13T4D0@]14`48PA511X4?0@E13044`@S10@0000]0@80
100002d10`040000l04400D000090_ooo`0500000@800000100000810P0500000@800000100002d1
00090000nP8000/0000000008P040000;@4407H0000T0cT0m@G50P`6ZP8T1Y42?0Ii0U@6HP9[1Td2
PPHk0YP6:`:]1Qh2`@HD0]8630;R1PP2l0H70_/62@841`h22`LF0Pl7808@1bh23PLo0PX7D@831fH2
nPIm0^h6UP;P1Zl2d0K:0[h6iP:Z1P83U@HM0gl6>@=X1U<3D0I]0cP6Q@<P1Y`32@J`0o85`P?L1M83
a`GP0k<5jP>Q1O43TPGf0h@5m`=h1OD3K`G`0fT5j0=U1Md3I0G?0fH5_`=Z1J`3L@FG0gX5P@>61FP3
U0E>0j@5<`>f1AP3bPGl0]l5h0;e1LD2100002d10P040000;@4300@0003`0@@01@0000T2002000D0
00010P00000400000P4200D000010P0000040000;@4100T0003j0P000@000000000R00@0000]0@@0
<P0002@35`1>1N03D0GC0eD5a`=M1Kh3J0Fg0g@5/`>11K<3SPFg0iP5_P>Q1LL3YPGC0jP5h0>V1N`3
X@Gh0iP50PB>1@T4P@D<17@530AX1@T4G@D215D5n0=@1N`3CPGP0`@0000]0@80100002d10`040000
l04400D000090_ooo`0500000@80000040000?/2^_lQ0000002l0P000000000R@G9YHF`006h40000
;@440100003k0Q001`000000_08000000@828U=iLgAUK@1^100002d11@040000204000D0000:0P00
000400000P4100@0000^0AP01@0000T2000000@0000]0@@01`0002450@0a02H6e@040000;@4500D0
00090_ooo`0500000@800000100002d110040000;@4500@000080@001@0000X2000000@000020@40
100002h1600500002@800000100002d1100700008@D103809`B90@@0000]0@D01@0000T2oooo00D0
00010P0000040000;@4400@0000]0@D0100000P1000500002P800000100000810@040000;P4H00D0
00090P0000040000;@4400L0000Q1@40<`350h`3100002d11@0500002@;oool01@000042000000@0
000]0@@0100002d11@040000204000D0000:0P00000400000P4100@0000^0AP01@0000T2000000@0
000]0@@01`0002450@0d0281S0<40000;@4500D000090_ooo`0500000@800000100002d110040000
;@4500@000080@001@0000X2000000@000020@40100002h1600500002@800000100002d110070000
8@D103D0i09B1P@0000]0@D01@0000T2oooo00D000010P0000040000;@4400@0000]0@D0100000P1
000500002P800000100000810@040000;P4H00D000090P0000040000;@4400L0000Q1@40=P0R1IX4
100002d11@0500002@;oool01@000042000000@0000]0@@0100002d11@040000204000D0000:0P00
000400000P4100@0000^0AP01@0000T2000000@0000]0@@01`0002450@0`07P7@`@40000;@4500D0
00090_ooo`0500000@80000040000?/2^_lO0000002@0@400000000R@G9YHF`000040000;@4600@0
000]0@D0100000P1000500002P800000100000810@040000;P4H00D000090P0000040000;@4600D0
00090P00P00700008@D103401`OT00D000090P0000040000;@4500D000090_ooo`0500000@800000
100002d11P040000;@4500@000080@001@0000X2000000@000020@40100002h1600500002@800000
100002d11P0500002@8008001`0002450@0b06`5L@0500002@800000100002d11@0500002@;oool0
1@000042000000@0000]0@H0100002d11@040000204000D0000:0P00000400000P4100@0000^0AP0
1@0000T2000000@0000]0@H01@0000T2002000L0000Q1@40<`3c0b821@0000T2000000@0000]0@D0
1@0000T2oooo00D000010P0000040000;@4600@0000]0@D0100000P1000500002P80000010000081
0@040000;P4H00D000090P0000040000;@4600D000090P00P00700008@D103@0e`<L1@D000090P00
00040000;@4500D000090_ooo`0500000@800000100002d11P040000;@4500@000080@001@0000X2
000000@000020@40100002h1600500002@800000100002d11P0500002@8008001`0002450@0e0:82
00<500002@800000100002d11@0500002@;oool01@000042000000@0000]0@H0100002d11@040000
204000D0000:0P00000400000P4100@0000^0AP01@0000T2000000@0000]0@H01@0000T2002000L0
000Q1@40=P0416P41@0000T2000000@0000]0@D00`0000004@0000`0000800002`00010000140000
@`0000T0000@0000A00004<0000:000040000000000000002@000100001B20007`P001P0000<0000
oooo01T0000<0000000001P0000<0000_moo01T0000<000000000180000<00000P0001T0000<0000
000002L0000H00000@000000002ogol0000002D0000<00000@0002H0000L00000P00000000010000
00000000000U000030000080001F0000`0400000000g0000A00004<0001Y0000201S1`X0F0L@04d7
6@121bL0>0Lh02d7C@0S1fD06@N100l7X0061l<0o@KY0?@64P7[1Sh1i0I/0M`6W@7E1]41c`H60/T6
?P;31WL2_`Jb0[/6k`:g1R`3]0I[0k86ZP>`1^T3[`HY1:l6J@B_1ZT4/0KX1;869PFd1V@5]`JP1K/6
f`Fo1QD6``I<1/T6PPK?1[H6e@KW1]`65@OT1T47j`IZ1o@6T0Om1[871POB1`l7kPLI1`H88`LK22d7
;0Ph1cX8@PM324d7B@QH1d/8H`M926h7@`Qi1cX8P`L/28h76`RH1`H8X`O^1jd7dPNf1k87`0N@1lT7
JPOB1d47fPLE1n87i`KZ1kH6l@N21_L7C0Km1aD60PSK1@L8X0D;26@53`PV1A88j0@D2:T45PQY11H8
:@@G2>T35PRZ0aH8J`<D22`34PS_0Pl8/P8;27L21`Pn0P881P;m1m41m`NM0O47K07Z1ch1hPLB0MX7
j@3B1l<0b@NP0<07P@2f1fD0[@M=0:<7>02H1bL0SPLI08<7401i1`X0KPL806<79P0001`000030000
0000000000000000000002D0000<00000`0002D0000<00001@00P2P0000<00000@0002P0000<0000
0P0001P0000<0000oooo01T0000<000000000180000<00000P0001T0000<0000000002L0000H0000
0P000000003oool0000002D0000<00000P0002H0000L00000@000000000;000000000000000U0000
30000040001F000040400040000V00003@0003d0000m0000>00F1S80l@D^0<`5;@2W1Bd0P`D`0645
=@101C/08@E400@5CP3Z15X0d`AX0<04M`2`18L0X`BH09X4ZP2E1;d0U0CA09H4i02M1?P0Y`@;0KD4
7P761341fPA30O84D`4=1F<1:PEa0DX5OP5[1HT1SPFC0K<5VP7G1J01o@FS0B86Y@571Z@1JPJR0Hd6
W@6^1YH1c@J>0NT6Q0431gP16PMZ0Bh7F`4n1d/1B`Lj0E@79`5I1a@1F`L10EP7kP1B1mX0A`O603T7
/`0X1j404`N?0?/6OP3P1Vl0``IP0:<6E0221TP0G`Ho03/6>00F1RD0000<00000`0002D0000<0000
1@00P2P0000<00000@0001P0000<0000002001T0000<000000000180000<00000P0001T0000<0000
000002L0000H00000@00000000000800000002D0000<00000@0002H0000L00001000000000010000
00000000000U0000300000@0001F0000N00000P0000k00002`0003h0000G0000o`1<1`41@0L60C@7
3`4Z1aX18`LV0B07<`4P1cl18`M:0BX7DP4d1eP1@0MI0D`7F05I1e81I@M:0Fh7?`5e1c<1N@LV0GT7
6P5e1`l1KPL60FD70@5I1ol0C0LU0000300000<0000U0000300000D0080X0000300000@0000H0000
30000?ooo`0I000030000000000B000030000080000I000030000000000U000030000080000V0000
700000@0000000002`000000000000009@0000`000040000EP0004P1000500004@0001h0000Z0000
B`000=01<P?a0A834P;c0S<2e@9D0[P2MP:M0YH2Q0:f0V`2e@9G0_<2A08?0c@2:@<U0T836P9I0a42
KP<;0X0320:@0`P2WP<:0ZT340:a0aP2]P<S0[T3<0:h0d02]@=C0Zl3I`:V0gh2V`>G0Xd3/P9l0lh2
J@?/0U@32P<m0bX390=;0`P3K0?/0Xd3cP:_0jl2d0>?0_43KP8A14d2<0@[0Th42P9Z1>T1Q@C80Il4
Z@6f18X1bPA/0Md4D07^13D1o0@L0@L51@4@1O405PGN01T5cP0J1L405`Ff0185[P0:1JP0o`BV0?44
YP3Q1:T0c`B_0;X4^02T1<<0R`CB0744hP1E1?D0=`@:0AP48P7h0c/1f0=F0KH3L`6E0i41M0>`0E83
d04b0bD0000<00000`0002D0000<00001@00P2P0000<0000100001P0000<0000002001T0000<0000
00000180000<00000P0001T0000<0000000002D0000<00000@0002H0000L00001000000000010000
00000000000U0000300000@0001F0000N00000D0000Y00001`0002`0000G0000S`0J1I403@FF0045
WP3h1:T0l@Be0>d4`P3]1<l0l@CI0?P4hP011NL03@GY01X5i`0V1N80<PGI03`5c`131L80APFe04H5
Z@131Ih0?0FF0385T@0V1Hl06PDU0000300000<0000U0000300000D0080X0000300000@0000H0000
30000?ooo`0I000030000000000B000030000080000I000030000000000U000030000080000V0000
700000@0000000002`000000000000009@0000`000040000EP000?P0000J00000@000240000D0000
=`0002X3:@4a0`P1>@?W0483b01=0jX0F0>>06D3M@1b0eh0P0=:08h3>P2L0bd0ZP<S0;P37P350a`0
dP<N0=h39@3Y0bh0l`<l0?`3C@0416402PAi00h4T`0A1:l04`C=0184k@0@10d13@@_0@P4D0411781
n@>B0O03/P7U0m01fP?[0Ld31@;00a`2/P<`0Z@3@0:F0dd2R0=F0WX3G09]0eh2H0=K0U@3E@990d/2
?P<n0SH3;@8^0aP2:0<10R<3i`4P0l/17`>]0Al3S@4Q0f`19@=;0BX3:@4U0000300000<0000U0000
300000D0080X0000300000@0000H000030000000P00I000030000000000B000030000080000I0000
30000000000U000030000040000V0000700000@0000000000@000000000000009@0000`000040000
EP0007P0000J00004P0001d0000E00005`000383H@8d0eD2>@=90T43?`9<0cP2F0<e0VD3=@9b0cP2
O0<o0XD3B@::0eD2S0=Q0XX3KP:50gX2O0>30W83RP9U0hh2F0>>0T`3RP910h<2>@=j0S@3KP8b0f42
9@0000`0000300009@0000`000050020:00000`000040000600000`0003oool06@0000`000000000
4P0000`0000200006@0000`0000000009@0000`0000200009P0001`000040000000000/000000000
000002D0000<0000100005H0000h0@00500001X0000b00009P0004L0001_0XT3L`9l0g/2K`>70V<3
UP9H0jP2CP>n0TD3e`8n0o<2=`<B0c83<`<_0eH3;@=k0b`3XP</0lX3;P?c0c8370@g0dH4?@=`14@3
V@A=0l84EP?Z164340E]0c@5NP=G1HL3M`FE0iD5Y0>`1K<3b0G20md5dP?_1N83o@Gb0`P60@@?1Q04
4PHO1146;@@=1S/41@I81?T5E0CZ1El4e`EY1<45L@BX1GT4S0Eo16h5Q0A=1HP4:@F:10@5R`CN18X4
]PB818d4Q@AS1804>@Aj1104L`CV0fX4_P=P19H3E@A`0dX4B`<m12T3<0@80b44jP8C1<l210Bg0_@3
XP;T0i42e@>30/D3N0:e0g42YP=^0YL3K`:90bD0000<00000`0002D0000<00001@00P2P0000<0000
100001P0000<0000002001T0000<000000000180000<00000P0001T0000<0000000002D0000<0000
0@0002H0000L0000100000000001000000000000000U0000300000@0001F0000N0000180000L0000
5@0001l0000G0000>P:60c`2N@=10Vd3BP9T0eD2G@=Q0UT3KP9I0gX2G@>50V@3S@9]0i<2N@>D0XH3
T`:B0hd2WP>50ZP3NP:_0fh2/P=Q0[83E@:_0dX2Z0=10Yh3?0:B0cX2QP<U0000300000<0000U0000
300000D0080X0000300000@0000H000030000?ooo`0I000030000000000B000030000080000I0000
30000000000U000030000080000V0000700000@0000000002`000000000000009@0000`000040000
EP000>P0000Q00008P0002/0000`0000<`0003D46`A211H4D@@D16445@Ac11X4Q@@S19T4;PB]13d4
`0A?1=@4H`CW17T4n@BB10X5[0@I1LH49`GQ13<5o@@m1AT5A@Dd1DX5CPE=1FH5CPEm1D`5T@E71J<5
@0Fb1CL5_PD/1LL57`G<1A05cPD01L`5kPC71M/4_`G81;<5]0BT1J04T`F=17h5NPAX1FP4D0EG13H5
A`@K1CT4o`@]1><48`C811`4[P@F19@44`Ak11<4I@@E15046@@o1204<0@Y12@4=@@K12D0000<0000
0`0002D0000<00001@00P2P0000<0000100001P0000<0000002001T0000<000000000180000<0000
0P0001T0000<0000000002D0000<00000@0002H0000L0000100000000001000000000000000U0000
300000@0001F0000N0000240000Q0000900002@0000G00004`@S11D45`@J10/48P@112d4nP<i1?L3
APCg0e<4nP=M1044IP@;16/45`A]12<4J`@`16H4?0AM14D4D`A<14H4D0@i1504;@A<1284A@@J13`4
5@@`11<48`@U0000300000<0000U0000300000D0080X0000300000@0000H000030000?ooo`0I0000
30000000000B000030000080000I000030000000000U000030000080000V0000700000@000000000
2`000000000000009@0000`000040000EP000001000/00004@0003X0000Q0000>@000?D5a@8<1ZX2
90JA0S`6N@9D1V82J`I=0X86>`:H1R/2[@HN0/4650;B1P`2hPH80_061`;k1PT210L>0P/75P8?1b02
40L^0Ph7?`8:1e420`MV0_X6O@;^1YH2h0J_0]06bP:n1^H2ZPH20iD67@=o1ST3J0IC0e06K@<h1XD3
80JL0`T6/0?b1L83g0GB0lL5h0>c1NX3X@Ga0i85mP>41OL3N0Ge0fl5l0=Y1NP3I@GM0f@5c`=V1Kl3
JPF/0g45U`=j1H43QPEX0i@5CP>T1C<3]PDH0lX5o0;O1N02m@G50RD0000<00000`0002D0000<0000
1@00P2P0000<0000100001P0000<0000002001T0000<000000000180000<00000P0001T0000<0000
000002D0000<00000@0002H0000L0000100000000001000000000000000U0000300000@0001F0000
N00002/0000O0000;P000240000G0000CPGP0e05d`=E1LL3G@Fn0fP5]`=d1K<3P@Fc0hh5]`>H1Kh3
X@G70jH5d`>X1N03YPG/0j45n0>H1@84SPD9184530Ad1@`4J0D915d50PAE1OP3D0G/0dh5h0<U0000
300000<0000U0000300000D0080X0000300000@0000H000030000?ooo`0I000030000000001B0000
L04000@0002joooo8@00000000000000_08000000000000R@@1b06T0H@1/00000000000000000000
00000000000000000000000000000000000000000000000000000000nc9X9O/bN04C0>Sl6@1h0A<0
H`@0052fo7LF]Oag;KGlM`0000200000/9;k<SC^4P000000iCShM`004`1X^3Hc0000013^4P281Q<0
a>hB09D[n7L001<01000003^4P10]oag20/C006h=S<G0000d>dB0000002HkQ80003hMaPfn7Oooooo
Z>hB089cj7L001<0000006PUnc9l1C1DJ2Gk</1GmgM0BeHa005<01P0001@]_agM>hB0000000Xl180
EQSZMk1ej7Oooooo0@0004mmm7MX9O/b80000;2Bnc;TkQ80J2Gk</@7mGO0EoMgH0DQkP@0001TMP08
000002D0000<000010000580001`0@001@0001000007000000000000002l0P00000000420R9C07T0
L`1d06D0K@0000000000000000000000000000000000000000000000000000000000000000000000
000B0000000Xl180EQSZMk1ej7Oooooo0@0004mmm7MX9O/b80000;2Bnc;TkQ80J2Gk</@7mGO0EoMg
H0DQkP@0001TMP08000008P64`34kQ80UB_hM`004`0400000>hB042go7L82a<00KPf<aL0003@kA80
000009S^4P000?Qg63KhMooooonXkQ80PW?XM`004`000000J2Gk<W`5<5AX9O/b`5OgMd1;ES400D`0
6000052fo7MdkQ80000002S`4P1F6>Yg/7GXMoooool10000CggdMfPUnc8U0000/9;k<_3^4P251@00
U`ceM`40000I0;01Q@D::fAf00P000009@0000`0000500004P0000`0000100005P0000`0000H0000
600000`0000000009@0000`000040000E00005@000070000<@0000L0000b00000@0002QVOcmkj80o
e@0002H600010000C00000000000000000000?ooooooooooD0000340000O00009@0000`000050000
600000`0003oool06@0000`0000000009@0000`0000400009@0000`0000500004P0000`000010000
5P0000`0000H0000600000`0000000009@0000`000040000E00005@0000=0000800000d0000Q0000
0@0002QVOcmkj80oR@4002L400010000C00000000000000000000?ooooooooooD0000380000O0000
9@0000`000050000600000`0003oool06@0000`0000000009@0000`0000400009@0000`000050000
4P0000`0000100005P0000`0000H0000600000`0000000009@0000`000040000E00005@0000M0000
7@0001d0000N00000@0002QVOcmkj80oS0<00<D300010000C00000000000000000000?oooooooooo
D00003<0000O00009@0000`000050000600000`0003oool06@0000`0000000009@0000`000040000
9@0000`0000500004P0000`0000100005P0000`0000H0000600000`0000000009@0000`000040000
E00005@0000M00001`0001d0000800000@0002QVOcmkj80oS0<0028100010000C000000000000000
00000?ooooooooooD00003@0000O00009@0000`000050000600000`0003oool06@0000`000000000
9@0000`0000400009@0000`0000500004P0000`0000100005P0000`0000H0000600000`000000000
9@0000`000040000E00005@0000d00005P0003@0000G00000@0002QVOcmkj80oDPH00>@200010000
C00000000000000000000?ooooooooooD00003D0000O00009@0000`000050000600000`0003oool0
6@0000`0000000009@0000`0000400009@0000`0000500004P0000`0000100005P0000`0000H0000
600000`0000000009@0000`000040000E00005@0000V0000:00002H0000Y00000@0002QVOcmkj80o
VP@0028500010000C00000000000000000000?ooooooooooD00003H0000O00009@0000`000050000
600000`0003oool06@0000`0000000009@0000`0000400009@0000`0000500004P0000`000010000
5P0000`0000H0000600000`0000000009@0000`000040000E00005@0000S0000?00002<0000m0000
0@0002QVOcmkj80o@`@007P700010000C00000000000000000000?ooooooooooD0000300000O0000
9@0000`000050000600000`0003oool06@0000`000000000DP00070100060000^_oooal000000000
000009010001000000008T40LP1Y0640K00000000000000000000000000000000000000000000000
00000000000000000000000000000180000002S`4P1F6>Yg/7GXMoooool10000CggdMfPUnc8U0000
/9;k<_3^4P251@00U`ceM`40000I0;01Q@D::fAf00P00000R0HC0<C^4P2E:oQg000C00@00000kQ80
@;OlM`P;4`01^3Hc5`000=3]4P000000V>hB0000n7LH=_QgooooojS^4P22LnQg000C0000001X9O/b
O0D`E6PUnc;0EoMg@4]F<@01C00H0000D;KlMgC^4P000?EgH0DQk]3=6@3@cAT0/9;k<[2Bnc9e3OEg
b<dI02D0002`T_/bl>hB08T5002G3?Eg0@0001T0/0691@XDI7H02000000U0000300000H0000U0000
300000D0000B000030000040000F0000300001P0000H000030000000000U0000300000H0000H0000
30000000P01D0000E00000L0000h00001`0003T000010000:6Io?g_XP3oT00001`L00040001<0000
0000000000000000oooooooooom@0000<@0001l0000H000030000000000U0000300000D0000H0000
30000?ooo`0I000030000000000U0000300000H0000U0000300000D0000B000030000040000F0000
300001P0000H000030000000000U0000300000H0000H000030000000P01D0000E00000@0000[0000
100002`000010000:6Io?g_XP3ma0000K0D00040001<00000000000000000000oooooooooom@0000
<P0001l0000H000030000000000U0000300000D0000H000030000?ooo`0I000030000000000U0000
300000H0000U0000300000D0000B000030000040000F0000300001P0000H000030000000000U0000
300000H0000H000030000000P01D0000E0000140000O00004@00020000010000:6Io?g_XP3lR0P00
l`<00040001<00000000000000000000oooooooooom@0000<`0001l0000H000030000000000U0000
300000D0000H000030000?ooo`0I000030000000000U0000300000H0000U0000300000D0000B0000
30000040000F0000300001P0000H000030000000000U0000300000H0000H000030000000P01D0000
E00002X0000N0000:P0001l000010000:6Io?g_XP3lL1@00e`<00040001<00000000000000000000
oooooooooom@0000=00001l0000H000030000000000U0000300000D0000H000030000?ooo`0I0000
30000000000U0000300000H0000U0000300000D0000B000030000040000F0000300001P0000H0000
30000000000U0000300000H0000H000030000000P01D0000E00001T0000D00006@0001D000010000
:6Io?g_XP3l00`00XP800040001<00000000000000000000oooooooooom@0000=@0001l0000H0000
30000000000U0000300000D0000H000030000?ooo`0I000030000000000U0000300000H0000U0000
300000D0000B000030000040000F0000300001P0000H000030000000000U0000300000H0000H0000
30000000P01D0000E00002@0000O00009000020000010000:6Io?g_XP3mX100010@00040001<0000
0000000000000000oooooooooom@0000=P0001l0000H000030000000000U0000300000D0000U0000
300000L0080U000030000000080`0000300000l0080U0000300000d0081;00004000000000050000
:00000`000020000:00000`000010000:00000`000030000:00000`000040000:00000`000050000
:00000`0000600003P0001@000000000400001@0000J@010;0000200003ooooo0P00000000000000
002:@P00R4800000j0?X0b4000080000HP0000`0000100008@0000P0001R000030000040000Q0000
2000024000080000L`0000`000000000700000P0000U000030000000080U0000300000L0080U0000
300000h0080`0000300000l0080I000030000?ooo`0H000030000000000B000030000080000C0000
30000040000D0000300000d0000E000030000040000F000030000000000=00004000000000000000
>P0000`0000:00006`00010000000000000002<0000P000095[Q@00000000000]1_Q@00004000010
900002@0000TF^5000000000002d6n500000@00004040000L`0000`0000000003@00010000020000
0P000580001`0@000@0001000007000000000000002l0P00000000L20R9C07T0L`1d06D0K@000000
0000000000000000000000000000000000000000000000000000000000000000003eMe=_IWAgHG9U
G4=/HG=cIG=LBFic400000l0001dia80o@4001SW4P2JMN5gZ>TB0?d1000`ia80B5iV01h000000000
B5iV08PEFP1Pia80UdWQM`000018GVH03`0007CW4P0?00000000033:4`2H@FL0X5eV071NEC4005Da
J4<C03E4=SA4<C55@S=5<30`@dJ<ia80T<?RMb831`0=0000oP400:SY4P010000L5iE<G1NEC5X@a<0
001V0=CW4P3ie>9g8P<700d0003n0@00Z>TB0040001L3SQ/8P<700d0003n0@00Z>TB004M=f`=0000
oP4006Af00P000009@0000`000010000AP0002P0000L0000AdA9@`8000000000000004@000130000
000002<0000P000095[Q@00000000000]1_Q@00004000010900002@0000TF^5000000000002d6n50
0000@000040400008`0002000036Knm3000000000000ZN]30000@000040T000090000<I_kd<00000
0000002Yjd<000100000@0@0000S0000800002AJhD00000000000;@KhD0000100000@2@0000T0000
95[Q@00000000000]1_Q@00004000010100002<0000P000095[Q@00000000000]1_Q@00004000010
900002@0000TF^5000000000002d6n500000@000040400008`0002000007>6Hn00000000001_9FPn
0000@000040T0000900000LhISh00000000006lUJ3h000100000@0@0000H000030000?ooo`0I0000
30000000000H000030000;oOo`0I000030000000000B000030000080000I000030000000000W0000
6000008000000000_moo0000000U000030000080000V0000700000<0000000000@00000000000000
9@0000`000030000EP000<0100030000Q@400>01003H0@00J@0000P0H`L:05P7401=1aT0@PLW03P7
>00]1dd08`MU01T7P@0?1j001PO30?d6j@3d1Q81j`Hn0N@6K07L1Yd1e@KA0Ll61P;91Sh2``Ig0[l6
/P:k1^l2]`H/0k@6J`>b1ZX3/0KY0jl6:@B_1VT4[`JY1;06j0Bb1RH5]0IT1KL6X0Fk1]/5_`HE1/<6
C0K91X86c`Jf1]D6i`KL1QD7i0I11n/6JPOd1Y07o@Jb1`H7dPL?1nh76@L622<76`P]1b`8>0Lj2487
@`Q=1dT8F0M;26<7B@Q^1d<8N@Lj28<7;0R>1a/8V0L62:<7kPN]1m87]PNb1l07T0O91fX7dPM11mX7
5@OR1nL6jPNf1_47PPKg1d`6o@LE1P88f`D72:052`QT1@l89PDB2>P450RY11H8J@@F22T45`SY0aH8
ZP<F26/350P/0a88k`8?2;822`Qg0PL8?P8220H2o@OA0OL7W@7a1f`1jPLn0N874P7J1nT0dPO30<T7
X0301h40]PMU0:d7C@2S1cP0V0LW08h76@231a00N@L:06h7201S1bH0000L00001000000000000000
00000000000U0000300000@0000U0000300000D0080X000030000080000X0000300000<0000H0000
30000?ooo`0I000030000000000B000030000080000I000030000000000W0000600000<000000000
oooo0000000U0000300000<0000V000070000080000000002`000000000000009@0000`000020000
EP000101000900002@4006<0002`0@00?@0003P05PHb0?45;P3<1Bd0Y`D]08<5<01Q1CD0@0Dk0245
A0041Dh0jPAJ0=<4J03017L0/0B70:<4V02J1:X0U@Bm09@4d@2F1>@0W@Ch0:L42`6e11h1aP@a0MX4
@`7b15<13@ES0BX5L@5:1Gh1J`F90Hh5T`6c1IX1e`FP0Od5X`4R1ZD1A`JT0FX6XP6=1Yd1[PJF0Ld6
SP7Y1X@10`Mh0AX7JP4^1e/1?PM;0D/7>P5D1bL1F@LD0E/70@5H1nh0DPOJ04L7aP0i1k<0:0NQ01<7
S`3k1Wh0h0I_0<<6H02S1U@0PPI805l6?`0k1SP05PHU0000300000@0000U0000300000D0080X0000
30000080000H000030000000P00I000030000000000B000030000080000I000030000000000W0000
600000800000000000200000000U000030000080000V0000700000D0000000000@00000000000000
9@0000`000050000EP0007P0000k0000W`400500002d0@005`000?l0C0L10D071P4d1`l1:PLJ0B<7
9P4P1c<180Lo0B<7BP4Z1e81=0MH0D07F@5<1eP1F@MB0FD7BP5^1cl1M@Lc0GT79P5i1aX1M@L?0Fh7
1P5U1`41F@Oo04`79@0000`0000400009@0000`000050020:00000`000050000600000`0003oool0
6@0000`0000000004P0000`0000200006@0000`0000000009@0000`0000300009P0001`000050000
000000/000000000000002D0000<00001@0005H000180@009@0007D0003K0000;@4004/0003@0C83
l@4B0a82l`8c0]D2E0:h0WH2W@:F0X@2]P9/0]D2E`;c0T@23`<d0RT39@920aX2F@<A0Vh32`:00`P2
T0<80Yh32P:Y0a02/@<H0[H38`:i0c02^0=00[D3D`:_0fL2YP=n0Y/3U`:=0k82O0?>0VT3k09D0`X3
?@<Z0b@3B`<80f`3k0:=0lh2[`>_0]03S`;a0fh24@A=0S04:`9>10X2JPCY0HD4b06O1:T1]PB:0LX4
K07M1501kP@e0O`470471@D140Ga01H5gP0I1Lh06PG101L5]P0B1Jh02PFX0?l4YP3a1:H0h@BY0<l4
[`2j1;P0Y0C308/4dP1a1>80E@Ce03L42P4H1281n0<k0MP3EP6f0g<1U@>A0G@3/05B0m01<P<U0000
300000@0000U0000300000D0080X0000300000D0000H000030000000P00I000030000000000B0000
30000080000I000030000000000U000030000080000V0000700000D0000000000@00000000000000
9@0000`000050000EP0007P0000R00007`4003L0000e0@005`0008l06PFA00d5UP011Ih0n0BY0?44
]@3]1<80k@C?0?44f@3h1>800@GW00d5j@0J1NL09PGR0385f@0l1Ll0@`G204H5]@161JT0@`FN03`5
UP0b1I409PF?01X59@0000`0000400009@0000`000050020:00000`000050000600000`0003oool0
6@0000`0000000004P0000`0000200006@0000`0000000009@0000`0000300009P0001`000050000
000000/000000000000002D0000<00001@0005H0003h0000/`0000H0003_0000SP0003L0000Z0bT1
<@<80CT3i`120lP0C@>Z05P3SP1U0gD0LP=N0803BP2>0cX0W0<]0:X38`2h0ah0a@<L0=837P3N0bD0
j@<^0?<3?03l0dd010AQ00X4N@0>19<04@B_01<4c@0B1>d040@=0@d4;`4815010@Ab0OT3TP7`0k81
i@?@0MX3j`7=0`D2`0<L0[83<0:T0d02UP==0XP3EP9j0e`2K@=N0V03F`9D0eD2B@=;0Sh3?P8f0bd2
;P<H0RP30@8S0nL180?;0Al3[@4O0hd18@=/0BD3B`4Z0bT19@0000`0000400009@0000`000050020
:00000`000050000600000`0000008006@0000`0000000004P0000`0000200006@0000`000000000
9@0000`0000200009P0001`0000500000000004000000000000002D0000<00001@0005H0001h0000
^@000880003?0000U`0001L0000b0f42=0=E0ST3B@910cl2C0<h0UP3=@9U0cD2LP<h0W`3?`:50dT2
RP=E0X`3H@::0fh2Q@=j0W`3P`9b0hX2I@>>0UP3SP9<0hX2@@>30ST3NP8d0fh2<P=Q0RD0000<0000
100002D0000<00001@00P2P0000<00001@0001P0000<0000oooo01T0000<000000000180000<0000
0P0001T0000<0000000002D0000<00000`0002H0000L00001@000000000;000000000000000U0000
300000D0001F0000>04008/0002g0000HP4000`100170000K`:90g<2O0=k0Vl3Q`9S0iH2F0>X0Th3
_P950mL2?P?c0SL34P<b0c<3;`=F0bd3N`</0j83;0?:0bh3l`<b0a`4=`=613d3L0A40iT4C@?215H3
jPAQ0a05K@<d1GX3E`F70gL5U@>E1J@3/0Fc0lP5`P?M1M83k`GR0od5lP<81P443`H@11867`@A1Rd4
3@Hk10D6B0Ci1E@4jPEO1=L5J@C11G44Z0Ei18`5O`A^1H@4C@F812T5RP@41H/4gPB:1;H4R0B=18D4
H`B013T4NP@@17<4iP=Z1;h3H0BF0eD4L0=:14/3?@@Y0c0420<Q1>X24`C?0P@4]`;d0j82i0>A0]D3
P`;50gP2]@=a0ZH3KP:G0fl2R@<U0000300000@0000U0000300000D0080X0000300000D0000H0000
30000000P00I000030000000000B000030000080000I000030000000000U000030000080000V0000
700000D0000000000@000000000000009@0000`000050000EP0007P000220000a00009L0003I0000
5`0003X2QP<l0WT3@@9]0dX2I0=E0Ud3H@9I0fh2F@=j0Ud3Q@9T0hd2K@>C0WT3U0:60i<2TP>=0Yh3
Q@:X0gX2[`=^0[83H@:b0eD2[`=:0ZP3@@:N0c`2TP<j0XH39@0000`0000400009@0000`000050020
:00000`000050000600000`0003oool06@0000`0000000004P0000`0000200006@0000`000000000
9@0000`0000300009P0001`000050000000000/000000000000002D0000<00001@0005H0003X0000
jP000>`0000f0@00EP4003<0000e11/4@P@F154450AQ11D4L`@J18D48`BI12h4[@@m1<04C`CD16<4
i`Ai1?T4TP@:1J`46@G612L5h@@c1Od4?@DI1DD5=0E:1Dh5C@EV1Dh5O@E<1I45A`FS1D05/PDg1Kh5
;0G71Al5c0D@1Lh500G<1Nh4a`GK1;l5b0Bc1K@4Y0FP19<5S@An1GX4J0EX1505E`@f1DL46`Di1?l4
;@CS12<4b0@L1:h45PBD11<4N`@C16D45@A@11T4?`@P1304:@@T13D46`@U0000300000@0000U0000
300000D0080X0000300000D0000H000030000000P00I000030000000000B000030000080000I0000
30000000000U000030000080000V0000700000D0000000000@000000000000009@0000`000050000
EP0007P0003/0000j0000041003m00005`0001<48`@E11L46P@;12840@@]1?X3>@Cg0dH4m`=C1?X3
G@@116H42`A[11L4K@@S16/4<0AV13`4G@A515<4C0A61504>@A@12d4C0@R14D46P@l11D4<0@C12<4
9@0000`0000400009@0000`000050020:00000`000050000600000`0003oool06@0000`000000000
4P0000`0000200006@0000`0000000009@0000`0000300009P0001`000050000000000/000000000
000002D0000<00001@0005H000000@00=P4007D0002K0@00j`0003T0003e1LD230JZ0R@6T@8l1WT2
E0IR0V/6C@:21S/2V0H[0Zd67P;11Q@2dPH<0^8620;`1PL2n`H90P@73P8;1aH23`LP0Q07;P8>1cl2
2PMA0P<7IP;j1Wd2kPJF0^06[`;@1/X2_PKV0ZX60P>E1Qd3O`Hi0fP6D`=@1Vd3>0J50b06W0<91[03
lPG20m`5dP?71N03/`GZ0j45l@>B1OH3Q0Gg0gP5m@=_1O03J@GX0fD5g@=T1Ll3IPFo0fX5[0=a1IL3
NPF10hH5J0>D1Dh3Y0Dc0kH560?:1O`2g`GP0_D5a@8U0000300000@0000U0000300000D0080X0000
300000D0000H000030000000P00I000030000000000B000030000080000I000030000000000U0000
30000080000V0000700000D0000000000@000000000000009@0000`000050000EP0007P0000c0@00
f00004P1003]00005`0004h5h0=@1M<3E@G70ed5_P=X1KL3M0Fc0h45/`>>1KL3V0Fn0j45a`>V1M<3
Z0GP0jH5k0>Q1OP3V0D218h52@B11@`4M0D<16P52@AM1@84E@Gh0e05k0=>1N039@0000`000040000
9@0000`000050020:00000`000050000600000`0003oool06@0000`000000000DP00070100050000
^_ooob400000000000000;`2000000001`008T40LP1Y0640K0000000000000000000000000000000
000000000000000000000000000000000000000000000?/bJ2Gk<WP14`3Xo1T0N04C06<4001@]_ag
5[GlMbfeo7L00000P0000;2Bnc8dkQ8000000>Dhn7L001<0J;Pf<`00000@kQ80R0HC0<C^4P2E:oQg
000C00@00000kQ80@;OlM`P;4`01^3Hc5`000=3]4P000000V>hB0000n7LH=_QgooooojS^4P22LnQg
000C0000001X9O/bO0D`E6PUnc;0EoMg@4]F<@01C00H0000D;KlMgC^4P000000:?0B05HHjWN`MNQg
ooooo`40001?OOAgJ2Gk<R00002`T_/bi>hB06PUnc;41oEg`5OgMf058Nh40000I7H02000000U0000
300000D0001B0000L04000H0000@00001`00000000000000_080000000070P8RD`1i07<0M01U06d0
0000000000000000000000000000000000000000000000000000000000000000000000004P000000
:?0B05HHjWN`MNQgooooo`40001?OOAgJ2Gk<R00002`T_/bi>hB06PUnc;41oEg`5OgMf058Nh40000
I7H0200000281Q<0a>hB09D[n7L001<01000003^4P10]oag20/C006h=S<G0000d>dB0000002HkQ80
003hMaPfn7OoooooZ>hB089cj7L001<0000006PUnc9l1C1DJ2Gk</1GmgM0BeHa005<01P0001@]_ag
M>hB0000000Xl180EQSZMk1ej7Oooooo0@0004mmm7MX9O/b9@000;2Bnc;`kQ80Q@D009L<mGL10000
6@2`0HD52R]TMP08000002D0000<00001P000180000<00000@0001H0000<0000600001P0000<0000
000002D0000<00001@0005@0001D0000<P0005P1000h0000JP400040000XIWloNnR0?mD0000V1P00
0@0004`00000000000000000003ooooooooooe00000a00007`0002D0000<00001P0001P0000<0000
oooo01T0000<0000000002D0000<00001@0002D0000<00001P000180000<00000@0001H0000<0000
600001P0000<0000000002D0000<00001@0005@0001D0000FP000>@0001Q0000mP000040000XIWlo
NnR0?hT1000W10000@0004`00000000000000000003ooooooooooe00000b00007`0002D0000<0000
1P0001P0000<0000oooo01T0000<0000000002D0000<00001@0002D0000<00001P000180000<0000
0@0001H0000<0000600001P0000<0000000002D0000<00001@0005@0001D0000cP000<h0003E0000
h0000040000XIWloNnR0?h`300350`000@0004`00000000000000000003ooooooooooe00000c0000
7`0002D0000<00001P0001P0000<0000oooo01T0000<0000000002D0000<00001@0002D0000<0000
1P000180000<00000@0001H0000<0000600001P0000<0000000002D0000<00001@0005@0001D0000
cP0003D0003E0000A`000040000XIWloNnR0?h`3000R0@000@0004`00000000000000000003ooooo
oooooe00000d00007`0002D0000<00001P0001P0000<0000oooo01T0000<0000000002D0000<0000
1@0002D0000<00001P000180000<00000@0001H0000<0000600001P0000<0000000002D0000<0000
1@0005@0001D0000KP4009/0001e0@00[@000040000XIWloNnR0?e86003T0P000@0004`000000000
00000000003ooooooooooe00000e00007`0002D0000<00001P0001P0000<0000oooo01T0000<0000
000002D0000<00001@0002D0000<00001P000180000<00000@0001H0000<0000600001P0000<0000
000002D0000<00001@0005@0001D00002`4001d1000B0@00;`400040000XIWloNnR0?iX4000R1@00
0@0004`00000000000000000003ooooooooooe00000f00007`0002D0000<00001P0001P0000<0000
oooo01T0000<0000000002D0000<00001@0002D0000<00001P000180000<00000@0001H0000<0000
600001P0000<0000000002D0000<00001@0005@0001D0000m`000:@1003n0000]P400040000XIWlo
NnR0?d<4001h1`000@0004`00000000000000000003ooooooooooe00000`00007`0002D0000<0000
1P0001P0000<0000oooo01T0000<000000000580001`0@001`000;[ooolO000000000000002@0@00
0@0000L002910780J@1Q06`000000000000000000000000000000000000000000000000000000000
0000000000000000000B0000000Xl180EQSZMk1ej7Oooooo0@0004mmm7MX9O/b9@000;2Bnc;`kQ80
Q@D009L<mGL100006@2`0HD52R]TMP08000008P64`34kQ80UB_hM`004`0400000>hB042go7L82a<0
0KPf<aL0003@kA80000009S^4P000?Qg63KhMooooonXkQ80PW?XM`004`000000J2Gk<W`5<5AX9O/b
`5OgMd1;ES400D`06000052fo7MdkQ80003eMf058Nk@cAT0d<dI0;2Bnc:`T_/bM@geMlS=6@0U0000
/9;k<_3^4P291@00U`ceM`40000I0;01R@D:56Af00P000009@0000`0000700009@0000`000060000
4P0000`0000100005P0000`0000H0000600000`0000000009@0000`000070000600000`000000800
E00005@0000e0000RP4003`0002L0@000@0002QVOcmkj80oi00000L700010000C000000000000000
00000?ooooooooooD0000340000O0000600000`0000000009@0000`000060000600000`0003oool0
6@0000`0000000009@0000`0000700009@0000`0000600004P0000`0000100005P0000`0000H0000
600000`0000000009@0000`000070000600000`000000800E00005@0000K0000;@4002<0000o0@00
0@0002QVOcmkj80oL@0006`500010000C00000000000000000000?ooooooooooD0000380000O0000
600000`0000000009@0000`000060000600000`0003oool06@0000`0000000009@0000`000070000
9@0000`0000600004P0000`0000100005P0000`0000H0000600000`0000000009@0000`000070000
600000`000000800E00005@0001m0000e`0008D0003Y00000@0002QVOcmkj80o8P800?<300010000
C00000000000000000000?ooooooooooD00003<0000O0000600000`0000000009@0000`000060000
600000`0003oool06@0000`0000000009@0000`0000700009@0000`0000600004P0000`000010000
5P0000`0000H0000600000`0000000009@0000`000070000600000`000000800E00005@0000X0@00
d@0002l1003S00000@0002QVOcmkj80o70D00=L300010000C00000000000000000000?oooooooooo
D00003@0000O0000600000`0000000009@0000`000060000600000`0003oool06@0000`000000000
9@0000`0000700009@0000`0000600004P0000`0000100005P0000`0000H0000600000`000000000
9@0000`000070000600000`000000800E00005@0002_0000R`000;L0002M00000@0002QVOcmkj80o
00<00:8200010000C00000000000000000000?ooooooooooD00003D0000O0000600000`000000000
9@0000`000060000600000`0003oool06@0000`0000000009@0000`0000700009@0000`000060000
4P0000`0000100005P0000`0000H0000600000`0000000009@0000`000070000600000`000000800
E00005@000000@00f`0000P1003]00000@0002QVOcmkj80oJ0@000@400010000C000000000000000
00000?ooooooooooD00003H0000O0000600000`0000000009@0000`0000600009@0000`000070020
9@0000`000000020DP00070100080000400000L00000000000000;`2000000001`828U<0N@1c07@0
I@1]0000000000000000000000000000000000000000000000000000000000000000000000000?Eg
gP8Qh`00000000000@0003cD6@1Hia80DNCfMmh28N<00000000004CD6@2djA80?=@I04:6mWON0R7S
>6//<`L00014e1T0]>TB0;CY4P0000000000423W4P01C?Ag000009SV4P000000H>LB09D[n7LH=_Qg
ooooo`004P22LnQg000C0000003Z3?EggP8QhiB3702DPa`0/9;k<[2Bnc9e3OEgS8<L02D0002`T_/b
U>LB010000000?Eg0@0001T0/04@0901G`WeMmh28N<000209@000100T07oo`0040000;cW4P3j7OAg
gP8Qha00T04U0000I7H02000000U0000300000P0001;00004000000000050000:00000`000030000
:00000`000020000:00000`000040000:00000`000050000:00000`000060000:00000`000070000
AP0001@000080000AdA9@`<0000R000030000?oooolR000030000?oooolU0000300000d0080X0000
30000040000X0000300000P0000R000030000?oooom60000800001@00015CDH[9T00010000040000
000004`0001T00000P000080003R0@00eP40008000020000h@400=D1000Y0:X0000000000000080o
000000000000080o0000000000000000000000000000000000000000000R000030000?oooom60000
700001000015CDH[0T0000`0000000003P0001@000000000400001@0
\>"], "Graphics",
  ImageSize->{305, 298.563},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  CellTags->{"b:13.1", "Data Structures"}],

Cell["\<\
A tree consisting of n bodies also contains n joints. Every system contains a \
base reference frame which is designated body '0'. Otherwise, bodies and \
joints can be numbered arbitrarily. Joint data and body data are organized \
into lists by the analyst, i.e.:
\tJointList={JointData_1,...,BodyData_n}
\tBodyList={BodyData_1,...,BodyData_n}
The structure ofthe individual data objects will be described below. Joints \
and bodies are implicitly numbered by there position in the data lists.
Each body contains a unique 'inboard' node, corresponding to (the outboard \
side of) a joint through which the body connects to an inner branch of the \
tree, or to the root (body 0). Each body may also contain outboard nodes. The \
outboard nodes are distinguished body locations that may be associated with a \
joint location (the inboard side of the joint), a sensor location, a point of \
application of an external force or any other feature of interest. Since one \
joint connects the tree to the root (the root node may be considered an \
outboard node of body 0), there must be at least n-1 outboard nodes among the \
n bodies corresponding to the remaining n-1 joints. These are the n 'outboard \
joint nodes'. The outboard joint nodes must be numbered 1 through n and must \
correspond to the associated joint number. The specific association of \
numbers to joints is not essential but by convention the root node is \
normally assigned the number 1. The remaining outboard notes can be numbered \
arbitrarily. The inboard nodes need not be assigned numbers.
In summary there are two important book-keeping principles:
\t\[Bullet] joints and bodies are numbered according to their position in the \
data lists,
\t\[Bullet] output joint nodes must be numbered consistently with their \
repective joints.
A tree is composed of a set of defining chains. For instance consider a tree \
composed of the following sequences of bodies:
                                       0,1,2,4      0,1,2,3,5      0,1,2,3,6
All defining chains of any tree will start with body 0, so we need not list \
it. However, the body sequences alone do not adequately define a tree. For \
instance bodies 5 and 6 both connect to body 3, but they will do so through \
different joints. This information can be provided by defining each chain as \
an ordered list of pairs - each pair consisting of a body and its inboard \
joint: {inboard joint, body}. For example:
            {{1,1},{2,2},{5,4}}      {{1,1},{2,2},{3,3},{4,5}}      \
{{1,1},{2,2},{3,3},{6,6}}
indicates that body 5 connects to body 3 at joint 4, and body 6 connects to \
body 3 at joint 6. Recall that each joint is uniquely associated with an \
outboard node of a particular body.
A tree is defined by the data structure:
    Tree = {list of chains}
    chain = ordered list of pairs {inboard joint, body}
              =  {{first inboard joint, first body},...,{last inboard joint, \
last body}}\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Reference Frames", "Subsubsection",
  CellTags->{"Data Structures", "b:13.1"}],

Cell["\<\
It is assumed that there is a single inertially fixed reference frame whose \
origin is the inboard side of the root  joint.  Each body has a primary \
reference frame, fixed in the body with origin at the inboard node. Body data \
is defined in this frame. As appropriate, there may be other body fixed \
frames as well with origins at outboard node locations. Normally, the axes of \
these frames are parallel to the primary frame when the body is undeformed. \
For the system as a whole, there is a 'reference configuration' corresponding \
to the nominal joint configurations (associated with zero joint motion \
parameters) and undeformed bodies. In the reference configuration all \
reference frames (body and space) are alligned. The analyst sets up the \
reference configuration when choosing body frame orientations for joint and \
body data definitions. It is recommended that the analyst begin by defining a \
physically meaningful reference configuration from which data definitions \
will logically follow.\
\>", "Text",
  CellTags->{"Data Structures", "b:13.1"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Joints", "Subsubsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}],

Cell["\<\
Joints constrain the relative motion between two bodies, or more precisely, \
between two reference frames  - one fixed in each body on either side of the \
joint. In TSi Dynamics, joints are of two types: simple and compound. Simple \
joints directly define relative motion between the two reference frames, \
whereas compound joints require intermediate reference frames to define the \
overall motion.
Simple joints are characterized by the number of degrees of freedom, r, an \
r-vector of joint quasi-velocities, p, and a 6xr joint map matrix, H. Across \
the joint the relative velocity vector (\[CapitalDelta]w, \[CapitalDelta]V) \
is Hp. Moreover, H is a constant (independent of the joint configuration) and \
the columns represent the joint action axes in the outboard frame (by \
convention). A compound joint is equivalent to a sequence of simple joints. \
Thus, it is necessary to define a set of numbers that represent the degrees \
of freedom associated with each intermediate frame and a corresponding set of \
(constant) joint map matrices. When defining a joint in TSi Dynamics, it is \
necessary to also assign names for both the joint quasi-velocities and joint \
configuration variables.
A k-frame compound joint with n degrees of freedom is defined by the data \
structure:
                                                            {r, H,q,p} 
where 
   r = k-vector whose elements define the number of degrees of freedom for \
each simple joint, with n = r1 +.. + rk. 
   H = [H1 .. Hk], a matrix composed of the k joint map matrices of the \
simple joints.
   q = n-vector of joint coordinate names.
   p = n-vector of joint quasi-velocity names.  
Let us consider some elementary examples.
A two degree of freedom revolute joint as a simple joint (spherical joint):\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}],

Cell["\<\
r1={2};H1={{1,0},{0,0},{0,1},{0,0},{0,0},{0,0}};
q1={a1x,a1z};p1={w1x,w1z};\
\>", "Input",
  CellLabel->"In[8]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}],

Cell["\<\
A two degree of freedom revolute joint as a compound joint (universal joint):\
\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}],

Cell["\<\
r2={1,1};H2={{1,0},{0,0},{0,1},{0,0},{0,0},{0,0}};
q2={a2x,a2z};p2={w2x,w2z};\
\>", "Input",
  CellLabel->"In[10]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}],

Cell["\<\
The function Joints computes all of the necessary quantities for modeling a \
list of joints. These include the kinematic matrix V(q), the Euclidean \
configuration matrix X(q), and the overall joint map matrix H(q). These are \
reported as lists.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}],

Cell[CellGroupData[{

Cell["\<\
JointLst={{r1,H1,q1,p1},{r2,H2,q2,p2}};

{V,X,H}=Joints[JointLst];\
\>", "Input",
  CellLabel->"In[12]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}],

Cell[BoxData[
    \("Computing joint 1 kinematics"\)], "Print",
  CellLabel->"From In[12]:=",
  CellTags->"b:13.1"],

Cell[BoxData[
    \("Computing joint 2 kinematics"\)], "Print",
  CellLabel->"From In[12]:=",
  CellTags->"b:13.1"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Print[V]", "Input",
  CellLabel->"In[14]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}],

Cell[BoxData[
    \({{{1, 0}, {0, Cos[a1x]}}, {{1, 0}, {0, 1}}}\)], "Print",
  CellLabel->"From In[14]:=",
  CellTags->"b:13.1"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Print[X]", "Input",
  CellLabel->"In[15]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}],

Cell[BoxData[
    \({{{Cos[a1z], \(-Cos[a1x]\)\ Sin[a1z], Sin[a1x]\ Sin[a1z], 
          0}, {Sin[a1z], Cos[a1x]\ Cos[a1z], \(-Cos[a1z]\)\ Sin[a1x], 0}, {0, 
          Sin[a1x], Cos[a1x], 0}, {0, 0, 0, 
          1}}, {{Cos[a2z], \(-Cos[a2x]\)\ Sin[a2z], Sin[a2x]\ Sin[a2z], 
          0}, {Sin[a2z], Cos[a2x]\ Cos[a2z], \(-Cos[a2z]\)\ Sin[a2x], 0}, {0, 
          Sin[a2x], Cos[a2x], 0}, {0, 0, 0, 1}}}\)], "Print",
  CellLabel->"From In[15]:=",
  CellTags->"b:13.1"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Print[H]", "Input",
  CellLabel->"In[16]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}],

Cell[BoxData[
    \({{{1, 0}, {0, 0}, {0, 1}, {0, 0}, {0, 0}, {0, 0}}, {{1, 0}, {0, 
          Sin[a2x]}, {0, Cos[a2x]}, {0, 0}, {0, 0}, {0, 0}}}\)], "Print",
  CellLabel->"From In[16]:=",
  CellTags->"b:13.1"]
}, Open  ]],

Cell["\<\
Notice that joints 1 and 2 produce the same Euclidean configuration matrix. \
Thus, the configuration coordinates have the same physical meaning. Other \
observations are: (i) the overall joint map matrix for a compound joint \
depends on the joint configuration parameters, and (ii) the kinematic matrix \
is simpler for a compound joint representation than for a simple joint \
representation. A free or unconstrainedbody would be modeled using a six \
degree of freedom joint, such as\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}],

Cell[CellGroupData[{

Cell["\<\
r3={6}; H3=IdentityMatrix[6];
q3={ax,ay,az,x,y,z}; p3={wx,wy,wz,ux,uy,uz};

JointLst={{r3,H3,q3,p3}};

{V,X,H}=Joints[JointLst];\
\>", "Input",
  CellLabel->"In[17]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}],

Cell[BoxData[
    \("Computing joint 1 kinematics"\)], "Print",
  CellLabel->"From In[17]:=",
  CellTags->"b:13.1"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Print[V]", "Input",
  CellLabel->"In[21]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}],

Cell[BoxData[
    \({{{1, Sin[ax]\ Tan[ay], Cos[ax]\ Tan[ay], 0, 0, 0}, {0, 
          Cos[ax], \(-Sin[ax]\), 0, 0, 0}, {0, Sec[ay]\ Sin[ax], 
          Cos[ax]\ Sec[ay], 0, 0, 0}, {0, 0, 0, Cos[ay]\ Cos[az], 
          Cos[az]\ Sin[ax]\ Sin[ay] - Cos[ax]\ Sin[az], 
          Cos[ax]\ Cos[az]\ Sin[ay] + Sin[ax]\ Sin[az]}, {0, 0, 0, 
          Cos[ay]\ Sin[az], 
          Cos[ax]\ Cos[az] + 
            Sin[ax]\ Sin[ay]\ Sin[az], \(-Cos[az]\)\ Sin[ax] + 
            Cos[ax]\ Sin[ay]\ Sin[az]}, {0, 0, 0, \(-Sin[ay]\), 
          Cos[ay]\ Sin[ax], Cos[ax]\ Cos[ay]}}}\)], "Print",
  CellLabel->"From In[21]:=",
  CellTags->"b:13.1"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Print[X]", "Input",
  CellLabel->"In[22]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}],

Cell[BoxData[
    \({{{Cos[ay]\ Cos[az], Cos[az]\ Sin[ax]\ Sin[ay] - Cos[ax]\ Sin[az], 
          Cos[ax]\ Cos[az]\ Sin[ay] + Sin[ax]\ Sin[az], 
          x}, {Cos[ay]\ Sin[az], 
          Cos[ax]\ Cos[az] + 
            Sin[ax]\ Sin[ay]\ Sin[az], \(-Cos[az]\)\ Sin[ax] + 
            Cos[ax]\ Sin[ay]\ Sin[az], y}, {\(-Sin[ay]\), Cos[ay]\ Sin[ax], 
          Cos[ax]\ Cos[ay], z}, {0, 0, 0, 1}}}\)], "Print",
  CellLabel->"From In[22]:=",
  CellTags->"b:13.1"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Print[H]", "Input",
  CellLabel->"In[23]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}],

Cell[BoxData[
    \({{{1, 0, 0, 0, 0, 0}, {0, 1, 0, 0, 0, 0}, {0, 0, 1, 0, 0, 0}, {0, 0, 0, 
          1, 0, 0}, {0, 0, 0, 0, 1, 0}, {0, 0, 0, 0, 0, 1}}}\)], "Print",
  CellLabel->"From In[23]:=",
  CellTags->"b:13.1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Rigid Bodies", "Subsubsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}],

Cell["\<\
A rigid body is defined by its mass, inertia matrix and the location of \
distinguished points or nodes where joints or sensors may be located. We \
assume the following:
1) There is a distinguished point that corresponds to the inboard joint of \
the body. The body frame has its origin located there. 
2) The center of mass and all other points of interest (nodes) including \
outboard joint locations are defined in the body frame.
3) The inertia matrix is defined in the body frame and it is the inertia \
matrix about the center of mass.
The data for a rigid body is organized in a list as follows. A rigid body \
with k outboard nodes is defined by the data structure:
                                    {com,{out1,..,outk},m,Inertia}
where 
   com is the center of mass location, 
   outi = {node number, location} for the ith outboard node, 
   m is the mass, and 
   Inertia is the 3x3 inertia tensor (about the center of mass). It may also \
be specified as a 6x6 spatial inertia tensor. In this      case it is assumed \
specified about the inboard joint. This type of specification is useful, for \
example, when dealing with submerged bodies where a fluid loaded inertia \
needs to be specified.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Flexible Bodies", "Subsubsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}],

Cell["\<\
Dynamics accommodates flexible bodies which satisfy the following \
assumptions:
- Local deformations are small, so that linear stiffness (quadratic strain \
energy) and dissipation (quadratic dissipation function) relationships apply.
 -  Body deformations can be characterized by a finite set of deformation \
coordinates, denoted by x. 
-  Body frame center of mass location and joint locations and orientations \
can be defined as (affine) linear functions of the deformation coordinates.
Any flexible body model in which a modal representation of flexure is valid \
satisfies these assumptions. Even with these assumptions, large global \
deformations are possible in which case the body inertia matrix as \
represented in the primary body frame may be a function of the deformation \
coordinates.
The primary body reference frame is a frame fixed in an infinitesimal body \
element with origin at the inboard joint location. We use the following \
notation for specifying the location of points in the primary frame under \
deformation. Let x be the n-vector of deformation coordinates. From the \
linearity assumption the position y of any point P in the body under \
deformation can be characterized by a matrix C and the relation
                                                   y = C.{1, x}
where C is a 3x(n+1) matrix. This is the type of information required to \
locate the center of gravity in the deformed body.
More information is typically required for general nodes at which sensors or \
joints are located. Here it is necessary to determine both the position of \
the node and the angular orientation of a local reference frame fixed in an \
infinitesimal element at the node. Under the small deformation assumption the \
orientation can be specified by any set of local parameters. We use Euler \
angles (3-2-1 convention). The orientation and position of the frame is \
defined by a 6-vector Y, the first three elements being the angles and the \
last three the position vector all relative to and specified in the primary \
body frame. As above, Y can be characterized by a matrix C
                                                  Y = C.{1,x}
where C is a 6x(n+1) matrix.
A flexible body with k outboard nodes and n deformation coordinates is \
defined by the data structure:
                       {Ccom,{out1,..,outk},m,{M(x),B,K},x,v}
where Ccom is a 3x(n+1) matrix that defines the center of mass location, 
   outi = {node number, Couti}, where Couti is a 6x(n+1) matrix that defines \
the orientation and location of the ith outboard node
   m is the mass 
   M(x) is the inertia matrix, B is the dissipation matrix, K is the \
stiffness matrix
   x is an n-vector of deformation coordinate names
   v is an n-vector of deformation velocity names\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Data Structures", "b:13.1"}]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["A First Example: The Double Pendulum", "Section",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell["Building the Equations of Motion", "Subsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Building Equations", "b:14.0"}],

Cell[GraphicsData["Metafile", "\<\
CF5dJ6E]HGAYHf4PEfU^I6mgLb15CDHPAVmbKF5d0@000:Ld0@0006`000000000ooooolX000350000
00000000002H6000A1P00215CDH00040=:L00?l200060000000000000000000000@0000300100@00
l0000000000000000000003R1020Z@<0AP0002`0000P0000ADe6:`500@0L00004000004@`=/10000
H000060000160000d0000<@00015CDH[<4020100000400000020?am0100<0000000001i00@0<0000
000001e0000D0000200000400036000024000TP0000l00000A30f`00001>0@0000000:TCD3l20000
0P000080000200000P000000044008100A30f`000000003o24010b`0000P00000A30f`8000000000
[8]j@TJDeSn/RgY2L=oR@P01000E@040400000@0000000008@0000P0001R000030000040000T0000
90000000P3d0000000000000P3d0000000000080001O0000>0000040000h0000000003P000000000
01810100000000000000000000000000000002D0000<00000@0002L0000H00000P00000000000000
000002D0000<00000P0001<0000<00000P0003/0000800006`000100003c0`006`0003H0000@0000
l`<008/0001H0000=000000000000000ooooooooool60000l`>@0>l3T`3[0i<0iP>C0><3T03S0h/0
=P000100003S0`006`0005P0000d000000000000003oooooooooo`H0003S0aL0iP<C0>/34`3_0a<0
l`<G0?<36`0m0000200001/0000@0000l`<00=/0000f000040000?<3001;0@00F00003@000000000
00000?oooooooooo1P000?<3D07_0e<1j`=C0NH3D`7S0e01h`=;0CH0000@0000h`<00=/0001H0000
=000000000000000ooooooooool60000h`?G0>H3d`3[0m<0k`?C0?<3e`3c0m/0?@0000P0000K0000
40000?<3002K0@00=P000100003c0`002`8005P0000d000000000000003oooooooooo`H0003c0a02
k`<C0^/34`;V0a<2h`<@0^<32`8f000040000><3002K0@00F00003@00000000000000?oooooooooo
1P000><3U`7V0i<1j`>C0Nl3T`7c0iL1l`>K0Cd0000800006`000100003c0`00F`8003H0000@0000
l`<00</2001H0000=000000000000000ooooooooool60000l`?@0^l3d`;[0m<2iP?C0^<3d0;S0l/2
=P000100003S0`00F`8005P0000d000000000000003oooooooooo`H0003S0eL2iP=C0^/3D`;_0e<2
l`=G0_<3F`8m0000200001/0000@0000l`<001/3000f000040000?<3002;0`00F00003@000000000
00000?oooooooooo1P000?<3T0?_0i<3j`>C0nH3T`?S0i03h`>;0cH0000@0000h`<001/3001H0000
=000000000000000ooooooooool60000h`<G0nH34`?[0a<3k`<C0o<35`?c0a/3?@0000P0000K0000
40000?<3003K0`00=P000100003c0`00B`@005P0000d000000000000003oooooooooo`H0003c0e04
k`=C1>/3D`CV0e<4h`=@1><3B`@f000040000><3003K0`00F00003@00000000000000?oooooooooo
1P000><3e`?V0m<3j`?C0nl3d`?c0mL3l`?K0cd0000800006`000100003c0`00V`@003H0000@0000
l`<000/5001H0000=000000000000000ooooooooool60000l`<@1Nl34`G[0a<5iP<C1N<340GS0`/5
=P000100003S0`00V`@005P0000d000000000000003oooooooooo`H0003S0iL4iP>C1>/3T`C_0i<4
l`>G1?<3V`@m0000200001/0000@0000l`<005/5000f000040000?<3003;1@00F00003@000000000
00000?oooooooooo1P000?<3d0G_0m<5j`?C1NH3d`GS0m05h`?;1CH0000@0000h`<005/5001H0000
=000000000000000ooooooooool60000h`=G1NH3D`G[0e<5k`=C1O<3E`Gc0e/5?@0000P0000K0000
40000?<3000K1P00=P000100003c0`00R`H005P0000d000000000000003oooooooooo`H0003c0i06
k`>C1^/3T`KV0i<6h`>@1^<3R`Hf000040000><3000K1P00F00003@00000000000000?oooooooooo
1P000><35`KV0a<6j`<C1^l34`Kc0aL6l`<K1Sd0000800006`000100003c0`00f`H003H0000@0000
l`<001L7001H0000=000000000000000ooooooooool60000l`<L1nl37`O[0al7iP<O1n<370OS0aL7
=P000100003S0`00f`H005P0000d000000000000003oooooooooo`H0003S0mL6iP?C1^/3d`K_0m<6
l`?G1_<3f`Hm0000200003`000080000?`0001P0000l0000oooood40001d00004`0000`000010000
9@0000`0000000209@0000`000070020900002@00000081100000000000008110000000000020000
:00000`000010000AP0008P0001l0000ADe6:`Q0008l0000<000004@`=/00000CP000000002RJV50
0P000080000200000P00004@`=/000000000o`Q00@</00008000004@`=/2000000000:b;NT9gW=Q1
Q8g>@WZff4800@005D01010000040000000002@0000T00000020?@00000000000020?@0000000000
0P0005l0000h00000@0003P000000000>000000000000040>0000000000000000000000000000000
9@0000`0000100009@0000`000050020E`0002@0000k0000600006/0001`00000P000>/3/P5e1/H6
9@0000`0000700209@0000`000000020900002@00000081100000000000008110000000000020000
:00000`000010000AP0005@000180000ADe6:`Q00P4P00005000004@`=/100005@0000000?oooooo
2T0002@0000H00000P0000400016U=HoKSCH@0SGld9JSj91900002@00000080m000000000000080m
0000000000020000GP0007@000010000000002@0000`0000E000020000000000:00000P000080000
0@010000000P0000000000000000000000000?ooo`000000V@000<`0001V0000<`0009T0003<0000
IP0003<0000U000030000040000U0000300000P0081<0000I000008000070000N`0001X0000K0000
K@0009l700150@008@3`0000000000000020?`00000000000020?`00000000000000000000000000
00000000000000009@0000`0000700209@0000`000000020900002@0000008110000000000000811
0000000000020000:00000`000010000AP000:@0002H0000ADe6:`Q0008l0000<000004@`=/00000
CP000000002Y4e0o0P000080000200000P00004@`=/000000000o`Q00@=80000?000004@`=/50000
000004JDeSmgW=Q1Fc7g@WNLf45K<OM2KSCH@4JDeSm^==Q0AYCF?gNLf4400@41P@0001E00@0@0000
10000000000T000090000000P3d0000000000000P3d0000000000080001O0000>0000040000h0000
000003P000000000000100d0000000000000000000000000000002D0000<00000@0002D0000<0000
1@00P5H0000`0000000000D0001m00007@0000D0000K0;81^PNb0KX7K@0K06d06`2b0BD0000<0000
1`00P2D0000<00000000P2@0000T00000020@@00000000000020@@00000000000P0002P0000<0000
0@0004H000380000_00004E=AR/8@043Z00009`000014<3K40000000003:45a2NIcH@LT@G4:VlKI1
@;IY@[FVVd6YRgY2]JJK@HV`QD:bYY]1AH><@Yka]T56Pha2MYcH@DJ3S49fW=Q1AX><@WNLf456Pha2
MicH@DJ3S499AoY1R[25@Qc92T:/RgY27<T:@T>fJD8Lb@Y2ba1L@TY7nT7:45a2N9cH@@030`<30`<3
0`<30`<30h<D@060400000@0003Vi^Ko:00000`000020000900002@00000080m000000000000080m
00000000000200009`0001P00002000000000>KViP0000009@0000`0000200004`0000`000010000
>`0000P0001E0000G000000000000000ooooooooool@0000L@>b0G43KP6W0cP1j`<h0Bh4>05U16h1
I@Bb0FD4/P5U1;81I@Bb0FD4m@4^12`2j`</0ZL3;09a0oD1L@>b0C`000080000?P0001P0000g0000
4`0004L0000S00004`0000`0000100009@0000`000000020900002@0000008110000000000000811
0000000000020000AP0005`0001@0000ADe6:`Q0008l0000<000004@`=/00000CP000000002o3Qa0
0P000080000200000P00004@`=/000000000oaE00@0@000010000000000T000090000000P3d00000
00000000P3d0000000000080001O0000>0000040000h0000000003P000000000000102L000000000
0000000000000000000002D0000<00000@0002D0000<00001@00P5D0001L0000=000014000190000
9@000100001a0k81L@=^0JL3>07[0cP1;P@h0FD4KP5U1;81I@Bb0FD4/P5U1;81I@Ce0Bh4;0;[0b`2
Y`</0W43m@5a0k819@0000`0000700209@0000`000000020900002@0000008110000000000000811
0000000000020000:00000`000010000AP000<P0002l0000ADe6:`Q00@>X0000W000004@`=/@0000
000001=@_d9j]]Q24U2o@/E;d4;>8/I22GW9@X:=cT89NLU2=oSF@PQibD;cb]e2`d_@@_C:gD9i]]Q2
m<[M@WVff4;db]e2N[KH@_C:gD9j]]Q2m<[M@RhQhD8hn=I2j_?W@XB=cT;ZlnM2cb;6@^[cid8CD;m2
;R7Q@Q=@_d9j]]Q200<30`<30`<30`<30`<3PaA00H0@0000100000000?lX000030000080000T0000
90000000P3d0000000000000P3d0000000000080000W0000600000800000000000000000000U0000
30000080000C000030000040000k0000200005D0001L000000000000003ooooooooooa00003k1LH6
n`F31S86C0Ie1T`6^0I<1^l6P`K_1/H6k`K61^l6aPK_1/H6k`H:1kP6@0Me1T07<PI01o/52POk1LH6
?00000P0000n0000600005l0001T0000K`0007@0000C000030000040000U000030000000080T0000
90000000P440000000000000P44000000000008000160000G00005000015CDH[24000S`0000`0000
0A30f`00001>000000000:TCD3l200000P000080000200000A30f`000000003o5D01010000040000
000002@0000T00000020?@00000000000020?@00000000000P0005l0000h00000@0003P000000000
>0000000000000403@0000000000000000000000000000009@0000`0000100009@0000`000050020
E@0005`0001N0000H`000740001f000040000?/5aPKk1H<6<PI<1WD6C0Jh1T`6k`J31^l6aPK_1/H6
k`K61^l6aPK_1PX7^0I01gD6@0Lb1T07n`D:1o/5aPHU0000300000L0080U000030000000080T0000
90000000P440000000000000P440000000000080000X00003000004000160000R00007`00015CDH[
24000S`0000`00000A30f`00001>000000000:9ZHD0200000P000080000200000A30f`000000003o
24010b`0000P00000A30f`8000000000oj7C@_G:gD9V]d93oe0Y@`01000E@040400000@000000000
900002@00000080m000000000000080m0000000000020000G`0003P000010000>0000000000h0000
000000000@0h0000000000000000000000000000000U000030000040000U0000300000D0081G0000
900006L0001/0000aP000:d000020000WPK_1R`<UPXU0000300000L0080U000030000000080T0000
90000000P440000000000000P440000000000080000X00003000004000160000b0000;`00015CDH[
24010jP0002L00000A30fa0000000000Khhh@om@:D=_SSQ3Y1/U@lgg>d=6/R539be0@dJb8D>1HTA3
A[8Q@mo;Ad>S6bE3h<]7@oi@:D?PbdM3oU0Y@n3;Ad?oD2U3h<]7@om@:D?PbdM3FHH]@h9RA4>gkc13
:2e0@kO_<4?=mc]3]nl`@fn>>4=IQRe3Khhh@om@:D<00`<30`<30`<30`<30`>35401P10000040000
0000ob@0000T00000020?@00000000000020?@00000000000P0002D0000<00000P0001<0000<0000
0@0003/000080000E@0005`00000000000000?oooooooooo400008T;UPZ92e8:`0/L2P<<70Y731`:
O@aB2Wd<UPYm39H:O@bF2Wd<UPYm3=T:A``?2`<<3`_02`l;R@_I2XT;UPXl0000200003h0000H0000
^0000:4000380000/@0001<0000<00000@0002D0000<00000000P2@0000T00000020@@0000000000
0020@@00000000000P0004H0001L0000D00004E=AR/8@002?000030000014<3K000004h000000000
ZA=@?`80000200000P00008000014<3K000000000?lE@040400000@000000000900002@00000080m
000000000000080m0000000000020000G`0003P000010000>0000000000h0000000000000@0=0000
000000000000000000000000000U000030000040000U0000300000D0081E0000G0000;L0002P0000
bP000;<0000@0000R@^F2XT;DP[02a`:0``L2TL<70Ym358:O@bF2Wd<UPYm39H:O@bF2Wd<f@Y730l;
0``?2l0;3`^92mT:R@^F2RD0000<00001`00P2D0000<00000000P2@0000T00000020@@0000000000
0020@@00000000000P0002P0000<00000@0004H0002D0000R00004E=AR/8@002B00003`000014<3K
000004h100000000ZA=@?`80000200000P000080000200000000@@00P4014<3K000000000?l8@043
;000020000014<3K0P000000002I>kY2Xa:`@UaP3D=jIBi3004001E00@0@000010000000000T0000
90000000P3d0000000000000P3d0000000000080001O0000>0000040000h0000000003P000000000
01810100000000000000000000000000000002D0000<00000@0002D0000<00000P0001<0000<0000
0P0003/0000800006`000100003I1@00O@D003H0000@000040H00=l5001H0000=000000000000000
ooooooooool600004PKS1A46j0D=1^X52@K/1@@6j`D21^L5=P000100003;1@00Q@D005P0000d0000
00000000003oooooooooo`H000391H45b`El1Lh5NPGB1GP5e`Ei1MT5O@Dm0000200001/0000@0000
=`H002D6000f0000400006h600271P00F00003@00000000000000?oooooooooo1P000706RPI^1Xl6
J`JA1VL6U0IR1Y86H0J>1SH0000@0000:@H002d6001H0000=000000000000000ooooooooool60000
9`HY1RP690H/1R86<0HP1SD68@Hg1RD6?@0000P0000K0000400009D6003<1P00=P000100003;1P00
;PL005P0000d000000000000003oooooooooo`H0003=1S87c0Hg1lP6>@O41S/7_`Hj1kd6=PLf0000
400008L6003D1P00F00003@00000000000000?oooooooooo1P0008D6d0J61//6RPK91Xh6a`JB1/T6
U@K<1Sd0000800006`000100003b1P00M0L003H0000@0000:@L00=H7001H0000=000000000000000
ooooooooool60000:`OJ1bX7gPLV1n478POS1ad7h@LK1mh7=P000100003T1P00O0L005P0000d0000
00000000003oooooooooo`H0003R1WP7i0Ic1nL6L@O[1Vl7l0I`1o86M0Lm0000200001/0000@0000
D0L001`8000f0000400008L7001m2000F00003@00000000000000?oooooooooo1P0008T7P@R71hH8
Q0N82807RPQk1hT8N@N523H0000@0000@PL002<8001H0000=000000000000000ooooooooool60000
@0LP24476`Q51aP8B@LF24h760Q@1a`8?@0000P0000K000040000:h700332000=P000100003T1`00
9@T005P0000d000000000000003oooooooooo`H0003V1bT9i@L^2N47<0WM1c89f0La2MH7;@Tf0000
40000:07003;2000F00003@00000000000000?oooooooooo1P0009h7a`RO1l88X`O02:L7_PR[1kl8
[PO323d0000800006`000100000;2000J`T003H0000@0000@PP00<d9001H0000=000000000000000
ooooooooool60000A0S@2D<8e@To2=L9>`SJ2CH8f0Td2=@9=P000100003m1`00L`T005P0000d0000
00000000003oooooooooo`H0003k1fl9o@MZ2@08J0T426H92@QW2@/8J`Tm0000200001/0000@0000
J@P0018:000f000040000:08001d2P00F00003@00000000000000?oooooooooo1P000:88N0ZP27d:
W@Qo2YT8P@ZD280:TPQl2SH0000@0000F`P001X:001H0000=000000000000000ooooooooool60000
F@PF2UX84@YN20l:HPP=2VL83`YY218:?@0000P0000K000040000<L8002j2P00=P000100003M2000
h`X005P0000d000000000000003oooooooooo`H0003P2>L:gPS/2]X8kP[G2?0:dPS_2]08j`Xf0000
40000;T800322P00F00003@00000000000000?oooooooooo1P000;H8_PZh2;T:_0Rg2/08]@[42;H:
a`Rj2Sd000080000?00000P0000o0000600005/0001F0000T0000;40000C000030000040000U0000
30000000080U0000300000L0080T000090000000P440000000000000P440000000000080000X0000
3000004000160000V00008`00015CDH[24000S`0000`00000A30f`00001>000000000:TCD3l20000
0P000080000200000A30f`000000003o24010c`0000`00000A30f`@000000000>E:V@UXYZd:5@9U2
NKRa@SIkRd;Ba[I2]8]j@YHk^T800`<35D01010000040000000002@0000T00000020?@0000000000
0020?@00000000000P0005l0000h00000@0003P000000000>0000000000000403@00000000000000
00000000000000009@0000`0000100009@0000`000050020E@0002`0000m0000E00005D0001O0000
100003<5FPG:18h5G0Bg1N/3dPDU0000300000L0080U000030000000080T000090000000P4400000
00000000P440000000000080000X00003000004000160000N00006`00015CDH[24010eP0001<0000
0A30f`L000000000Xa:`@ZgYYD8eZ:I2cdVc@YF0YT9I9:i2injS@WANZD8A^Im2cfNV@Q:iWd;@IjI2
Xa:`@ZgYYD800@<30`6101A00H0@0000100000000?lT000090000000P3d0000000000000P3d00000
00000080000U000030000080000C000030000040000k0000200001/0000@0000P@D00305000f0000
400003H5002K1@00F00002P00000000000000?oooooooooo0`0003@5LPDP1D/5oP@d1ET0000T0000
00000000003oooooooooo`80003n13@5P@D`1Cd000080000?00000P0000n0000600004l0001C0000
F@0005X0000C000030000040000U000030000000080T000090000000P440000000000000P4400000
0000008000160000V00008`00015CDH[24000S`0000`00000A30f`00001>000000000:TCD3l20000
0P000080000200000A30f`000000003o24010c`0000`00000A30f`@000000000^CPF@eHY5D=dLa53
SXdI@i;J2d=Mlaa3Xl45@`TX7d<00`<35D01010000040000000002@0000T00000020?@0000000000
0020?@00000000000P0005l0000h00000@0003P000000000>0000000000000403@00000000000000
00000000000000009@0000`0000100009@0000`000050020E@0002`000240000T`0009P0002Q0000
100006@9D`TH2IT9_PS@2Ed8l`TU0000300000L0080U000030000000080T000090000000P4400000
00000000P440000000000080000X00003000004000160000N00006`00015CDH[24010eP0001<0000
0A30f`L000000000Sa<J@kP_4D<<W1M3lO/H@a[25T?nSAI31M0D@oZZ54>/Fa93?N@C@jaK4T<ni1=3
Sa<J@kP_4D<00@<30`6101A00H0@0000100000000?lT000090000000P3d0000000000000P3d00000
00000080000U000030000080000C000030000040000k0000200001/0000@0000XPT001<9000f0000
400007X9002@2@00F00002P00000000000000?oooooooooo0`0006d9J@U=2D/99PTo2ET0000T0000
00000000003oooooooooo`80000V2Cl9XPTC2Cd000080000?00000P0000n000060000980002A0000
V`0009T0000C000030000040000U000030000000080T000090000000P440000000000000P4400000
0000008000160000X0T009@90015CDH[8D0700`0000000009D00010000040000000002a0000T0000
600003D6WC`000200000P6SITcc1Ui52iIfd@PQ00`DT2@0060T0004@`=/200000`0000P900010000
K0000380000:0000]P400>0100000000000000100012?00084E=AP000@082@00:@0000<000000000
000000000000100000<00401003`000000000000000000000>84082Y0`0I000030000?ooo`0B0000
30000080000U000030000000080U0000300000d0080U0000300000L0080H000030000000000F0000
30000000000Q0000200001h0000H000000000000000<0P00kP400140000<0000200000`0000@0000
00000000000;0000400000`2003^0@00600000`0000000004P0000`0000100006@0000`0003oool0
5P0000`0000H00002P00010000000000000000T0000@000080800002001B0000L04000400020o_oo
0000000000000000T040004000000P0@E01Y06d0I@1c0200CP1U07L0801B06l0K@1Q06h000000000
0000000000000000000000000000000000000000<C<@ja80P1800<c/4P2E:oQg^3WhMooooooLk180
9k;lMaP74`000000d07k=`00001605@0E`110580A@1L04<0K01Q07<0L`1U07<0G01304`0D`1904@0
G01k0300<00`0380@`150300<`0]0300<00`0300;@0`0300000`02d0@`0`0300<00]0300<00`0300
<00`0300<00`0300=00f07d0G01=06T0L`1S05<0M01Q07@0M@1c0000O@000:Eg003^>000001@k180
_NNUM`800h3XJ3@cK0fVMl9Hm7L90000/9;k<US/4P0ROOdb`0KeM`40001TMP08000002D0000<0000
0@0005@0001D0000<P0000X0003Z0000X`400040000?co1103Ka@C@0001P0@000@0004`000000000
00000000003ooooooooooe00001Q0000`0000580001`0@000P00023oool0000000000000002@0@00
0@000002011D06T0K@1U07<0801>06D0M`0P0580K`1]0640KP000000000000000000000000000000
0000000000000000001c0000O@000:Eg003^>000001@k180_NNUM`800h3XJ3@cK0fVMl9Hm7L90000
O>/B0000003U>?Qg000C040]60000000F>/B08P64`0<k180UB_hMb0in7Oooooo7>`B042go7OX1Q<0
B2dH04P]602XOOdbB2dH0300<0000300;@130300<00`02d0<00`0300<00`0300000C0300<00d03H0
O@1L04d0J@0006<00P000640M01e07<0001m0000YGL00>hh000C00800000;AP040000>QX=3<10000
0@0008c[4P1D0000E>`B09D[n7LH=_QgooooofAf00P000009@0000`000020000:00000`000010000
E00005@0001@0@00k`000;H1003N0@000@0000o?l440=_51I0400<0100010000C000000000000000
00000?ooooooooooD00007P0001S0000DP000701000100008?ooo`00000000000000090100000000
008045@0J@1]06D0L`0P04h0I@1g0200DP1_06d0H@1^000000000000000000000000000000000000
00000000000002LcH6<W<`00YGL00>hh000C00800000;AP040000>QX=3<100000@0008c[4P1lja80
00000>Dhn7L001<0@2dH0000001Hja80R0HC00c/4P2E:oQg83WhMooooolLk180@;OlMnP64`18;AP0
B2dH0?AmoC98;AP0h>/B0000n7LH=_Qgoooooo3[4P22LnQg000C0000000001<0BPD:Q61S9c?0EoMg
C@1Y0000H`020000H@1d0;c[4P000000[>dB05HHjWL001<00P00000]600@0000H6<W<`4000010000
S>/B05@0001Dk180UB_hMaPfn7OoooooI7H02000000U000030000040000X000030000080001D0000
E0000>80003_0000C@400>01000100003lo`@@0flD7[0000`0400040001<00000000000000000000
oooooooooom@0000<@000700001B0000L0400080000@00001`00000000000000_080000000010P8R
D`1i07<0M01U06d00000000000000000000000000000000000000000000000000000000000000000
00000000<C>`8c4c[>dB05HHjWL001<00P00000]600@0000H6<W<`4000010000S>/B07c[4P000000
iCShM`004`10;AP0000005S[4P281Q<03>`B09D[n7LP>OQgoooooac/4P10]oagj0HC04P]6018;AP0
@7km<TP]603Pja80003hMaPfn7Ooooool>/B089cj7L001<0000000004`251@Zm/2<a<l1GmgM=06T0
001S0080001Q07@0_>/B0000002/kA80EQSZM`004`02000002dH0100002`8c4c0@000040002<ja80
E00005C/4P2E:oQg63KhMooooomTMP08000002D0000<00000P0002P0000<00000@0002D0000<0000
1`00P2D0000<00000000P300000<00003`00P2D0000<00003@00P4/000100000<00000D0000P0000
0@000040000@000000000000000<0P00kP4000000000000030800>h1000X000030000080000R0000
30000?ooool>000050000000000@0000500001Y00d0/000080000?ooool20000000000000000@0=4
083g@`00003X0nP38@0000P0001R000030000040000Q00002000024000080000L`0000`000000000
700000P0000U000030000000080U0000300000L0080U0000300000h0080`0000300000l0080I0000
30000?ooo`0H000030000000000B000030000080000C000030000040000D0000300000d0000E0000
30000040000F000030000000000=00004000000000000000>P0000`0000:00006`00010000000000
000002<0000P00000:0<?@0000000000P0l=?@00TT800;I2900002@00000X0`m0000000000203`dm
002B@P00]T840000L`0000`0000000003@00010000190000F`000580001`0@000@00010000070000
00000000002l0P00000000L20R9C07T0L`1d06D0K@00000000000000000000000000000000000000
00000000000000000000000000000000000B09D[n7LH=_QgooooofAf00P00000Wo0<@nZS;T<Hd0e3
2BL^@`000000000000000000001CKfIdMf5bIEa=JF=bKg=_IWALEfU^I6mgLea3MG9bIFidEVEbLfU_
KUa9KW=dHFa/IG9LEG=ULTAQM65LDbda;CD]<CP00000000000000000000000000000000000000000
0000000000000000000000000000000000000000003<iQ8000000000001CKfIdMf5bIEa3K65cLfEc
G4U^La00000E0000M>LB0?d1000Hia80VWGQMjSY4P3m0@00<>LB06Af00P000009@0000`000010000
AP0002P0000L0000AdA9@`8000000000000000`2003^0@00000001T0000<0000oooo0180000<0000
0P0002D0000<00000000P580001`0@000`0001000007000000000000002l0P00000000L20R9C07T0
L`1d06D0K@0000000000000000000000000000000000000000000000000000000000000000000000
000005=_IWAgHG9UG4eYHg9_LfmVM5aGJFiTKgMcG4=eLW9UKWAFIG9cJFm^G4U^LgAQK6aULUaELfEb
A65dHEaC;C4]=Bda>000000000000000000000000000000000000000000000000000000000000000
000000000000000000000<cV4P000000000005=_IWAgHG9UG4=/HG=cIG=LBFic400001D0000U0000
/9;k<]3V4P0@0000003eM`40000I0;01402@0El9mGN@1B4m0000P2D0000@0901ool00100003hiQ80
nQgdMi058Cd@09019@0006Af00P000009@0000`0000300009@0000`000070020600000`000000000
5P0000`0000000008@0000P0000N0000600000000000000030800>h1000S00008000002P33d00000
0000080?3Cd00992002f@R@0000T00000:0<?@0000000000P0l=?@00TT800;I2100002<0000P0000
0:0<?@0000000000P0l=?@00TT800;I2900002@00000X0`m0000000000203`dm002B@P00]T840000
8`0002000030jhm100000000003e6HQ1002B@P00]T8T000090000<3[Sd40000000000?DIR4400992
002f@P@0000H000030000000000B000030000040000I000030000?ooo`0F0000300001P0000S0000
80000<3[Sd40000000000?DIR4400992002f@R@0000T0000`>^?@@0000000000mAV8@@00TT800;I2
100002<0000P0000NG@7?@0000000000mAT8?@00TT800;I2900002@0001iM0Lm00000000003e6@Pm
002B@P00]T840000DP00070100040000P?koo`000000000000000901000100001`8045@0J@1]06D0
L`0P04h0I@1g0200DP1_06d0H@1^000000000000000000000000000000000000000000000000034c
4>/B080B003<k180UB_hMkPin7Oooooog>`B02Nbo7LH1a<000000=01ncL00000AP1D05L0@@1B04D0
G01306`0H@1c07<0I@1c05`0@`1<05<0B@1405`0N`0`0300<00b04<0A@0`03<0;@0`0300<00`02d0
<00`0000<00]04<0<00`0300;@0`0300<00`0300<00`0300<00`03@0=P1m05`0C@1Y07<0H`1C07@0
H@1d07D0L`0007d0002UM`00kSP00000D>`B0;gWYGL200>0j6Pd<f`=YWO2F?Ag2@000;2Bnc9Hk180
8Wgm</06mGL10000I7H02000000U0000300000@0001D0000E00004/0001K0000D@0006T000010000
3lo`@@0flD4d0000H0400040001<00000000000000000000oooooooooom@0000H@000<00001B0000
L04000D0000Poooo0000000000000000T040004000070P0@E01Y06d0I@1c0200CP1U07L0801B06l0
K@1Q06h0000000000000000000000000000000000000000000000000L`0007d0002UM`00kSP00000
D>`B0;gWYGL200>0j6Pd<f`=YWO2F?Ag2@0007c[4P000000iCShM`004`10;AP0000005S[4P281Q<0
3>`B09D[n7LP>OQgoooooac/4P10]oagj0HC04P]6018;AP0Z7gm<TP]600`0300000`02d0@`0`0300
<00]0300<00`0300<00`00004`0`0300=00f07d0G01=06T0001S0080001Q07@0M@1c0000O@000:Eg
003^>0004`02000002dH0100003XJ3@c0@000040002<ja80E00005C/4P2E:oQg63KhMooooomTMP08
000002D0000<00001@0002P0000<0000100005@0001D0000E@0006<0001H0000J`000040000?co11
03Ka@F@100300@000@0004`00000000000000000003ooooooooooe00001h0000H`000580001`0@00
1000023oool0000000000000002@0@00000000L2011D06T0K@1U07<0801>06D0M`0P0580K`1]0640
KP0000000000000000000000000000000000000000000000000W<f1S9c<00:Eg003^>0004`020000
02dH0100003XJ3@c0@000040002<ja80O>/B0000003U>?Qg000C040]60000000F>/B08P64`0<k180
UB_hMb0in7Oooooo7>`B042go7OX1Q<0B2dH04P]603dOOdbB2dH0>3[4P000?Qg63KhMoooooo`ja80
PW?XM`004`000000000C04X52XAPHbLc`5OgMdd0J@0006<00P000640M02lja8000000:c]4P1F6>Yg
000C00800000;AP04000061S9c<100000@0008c[4P1D0000E>`B09D[n7LH=_QgooooofAf00P00000
9@0000`000040000:00000`000050000E00005@0001A0000H`0005@0001[00000@0000o?l440=_51
j`000<0100010000C00000000000000000000?ooooooooooD0000340001`0000DP00070100050000
400000L00000000000000;`2000000001`828U<0N@1c07@0I@1]0000000000000000000000000000
00000000000000000000000000000000000000000000034c/2<a<jc]4P1F6>Yg000C00800000;AP0
4000061S9c<100000@0008c[4P1lja8000000>Dhn7L001<0@2dH0000001Hja80R0HC00c/4P2E:oQg
83WhMooooolLk180@;OlMnP64`18;AP0B2dH041noC98;AP0h>/B0000n7LH=_Qgoooooo3[4P22LnQg
000C0000000001<0Q@D:_K0S<C?0EoMgC@1Y0000H`020000H@1d0;c[4P000000[>dB05HHjWL001<0
0P00000]600@0000/2<a<`4000010000S>/B05@0001Dk180UB_hMaPfn7OoooooI7H02000000U0000
300000D0000X0000300000@0000U0000300000L0080U000030000000080U0000300000<0001;0000
@0000300000500008000004000010000400004T0001K0000F`0006`000190000F`0005/0001/0000
:00000`0000500008P0000`0003oooooAP0001@000080000AdA9@`<0000R000030000?oooolR0000
30000?oooolU0000300000d0080X000030000040000X0000300000<0000R000030000?oooom60000
Y0T009P90015CDH[9T00010000040000000002E0000@000010000000000/@000900001P0000TMZHl
0000P000081PfI<lV>X?@kE45d<8@0@590T001P900014<3K0P0000<000082@000@0006`0000_0000
2@000;h100350@00000000000000@000f3P00215CDH0004020T002T0000300000000000000000000
00@0000300100@00l0000000000000000000003R1020Z@<06@0000`0003oool04P0000`000020000
9@0000`0000000209@0000`0000=00209@0000`000070020600000`0000000005P0000`000000000
8@0000P0000N0000600000000000000030800=81000A0000300000P0000<00004000000000000000
2`000100000<0P00dP4001P0000<000000000180000<00000@0001T0000<0000oooo01H0000<0000
600000X0000@000000000000000900004000040200000P00DP00070100010000P?koo`0000000000
0000090100010000008045@0J@1]06D0L`0P04h0I@1g0200DP1_06d0H@1^00000000000000000000
0000000000000000000000000000034c4>/B080B003<k180UB_hMkPin7Oooooog>`B02Nbo7LH1a<0
000000006C<00000AP1D05L0@@1B04D0G01306`0H@1c07<0I@1c05`0@`1<05<0B@1405`0N`0`0300
<00b04<0A@0`03<0;@0`0300<00`02d0<00`0000<00]04<0<00`0300;@0`0300<00`0300<00`0300
<00`03@0=P1m05`0C@1Y07<0H`1C07@0H@1d07D0L`0007d0002UM`00kcP00000D>`B0;gWYGL200>0
`6Ld<f`=YWO2F?Ag2@000;2Bnc9Hk180nSla<l06mGL10000I7H02000000U000030000040001D0000
E00002l000090000g@0008/100010000>FkS@@2:hd4d0000H0400040001<00000000000000000000
oooooooooom@0000H@000<00001B0000L0400080000Poooo0000000000000000T040004000000P0@
E01Y06d0I@1c0200CP1U07L0801B06l0K@1Q06h00000000000000000000000000000000000000000
00000000L`0007d0002UM`00kcP00000D>`B0;gWYGL200>0`6Ld<f`=YWO2F?Ag2@0007c[4P000000
iCShM`004`28DcPc000005S[4P281Q<03>`B09D[n7LP>OQgoooooac/4P10]oagj0HC091C>3>@DcPc
P40a<i1C>3<`0300000`02d0@`0`0300<00]0300<00`0300<00`00004`0`0300=00f07d0G01=06T0
001S0080001Q07@0M@1c0000O@000:Eg003_>0004`02000005<h<a000030Ic@c0@000040002<ja80
E00005C/4P2E:oQg63KhMooooomTMP08000002D0000<00000P0002P0000<00000@0005@0001D0000
G@400>40002n0@00a0400040000iK^=108[S@HL100300@000@0004`00000000000000000003ooooo
oooooe00001h0000H`000580001`0@000@00023oool0000000000000002@0@0000000002011D06T0
K@1U07<0801>06D0M`0P0580K`1]0640KP0000000000000000000000000000000000000000000000
000N<j2j7S<00:Eg003_>0004`02000005<h<a000030Ic@c0@000040002<ja80O>/B0000003U>?Qg
000C08QC>3<00000F>/B08P64`0<k180UB_hMb0in7Oooooo7>`B042go7OX1Q<0T5<h<i1C>3?<@34c
T5<h<n3[4P000?Qg63KhMoooooo`ja80PW?XM`004`000000000C00P62VZP^Qhc`5OgMdd0J@0006<0
0P000640M02lja8000000:c]4P1F6>Yg000C00800000DcPc40000:2j7S<100000@0008c[4P1D0000
E>`B09D[n7LH=_QgooooofAf00P000009@0000`000010000:00000`000020000E00005@0003S0000
h@0004P100350@000@0003U^hd40R^=1n@000<0100010000C00000000000000000000?oooooooooo
D0000380001`0000DP00070100020000400000L00000000000000;`2000000000@828U<0N@1c07@0
I@1]00000000000000000000000000000000000000000000000000000000000000000000000001hc
P;/N<jc]4P1F6>Yg000C00800000DcPc40000:2j7S<100000@0008c[4P1lja8000000>Dhn7L001<0
R5<h<`00001Hja80R0HC00c/4P2E:oQg83WhMooooolLk180@;OlMnP64`2@DcPcT5<h<aQ1<C>@DcPc
h>/B0000n7LH=_Qgoooooo3[4P22LnQg000C0000000001<0T@D:HX2k7S?0EoMgC@1Y0000H`020000
H@1d0;c[4P000000[>dB05HHjWL001<00P00001C>3<@0000P;/N<`4000010000S>/B05@0001Dk180
UB_hMaPfn7OoooooI7H02000000U000030000080000X000030000040000U0000300000L0080U0000
30000000080`0000300000l0080U0000300000d0081;0000@0000300000500008000004000010000
400000000000000030800=8100000000000000`2003B0@00:00000`0000200008P0000`0003ooooo
3P0001@000000000400001@0000J@0A0;0000200003ooooo0P000000000000000403A020jD<00000
j0?X0b4000080000HP0000`0000100008@0000P0000Q0000200007<0000<0000000001`000080000
9@0000`0000000209@0000`0000700209@0000`0000>0020<00000`0000?00206@0000`0003oool0
600000`0000000004P0000`0000200004`0000`000010000500000`0000=00005@0000`000010000
5P0000`0000000003@00010000000000000003X0000<00002P0001/0000@000000000000000S0000
8000001073d00000000005n85Cd00113000H@b@0000T0000040L?@0000000000GhPE?@0044<001Q3
100007<0000<0000000000d0000@0000T00009P0001B0000L04000<0000@00001`00000000000000
_080000000070P8RD`1i07<0M01U06d0000000000000000000000000000000000000000000000000
0000000000000000000000008Cd0000000000:3VnC;`j180V>Ki<T:6mWN@1B4ml?lR00<0002Pi_Tb
000C00@0000000005@000<9Hm7L100008@2:0@002P0h>3LcF_3dM`d0000I0;018@2:0B400022LnQg
000C0000002PYC4cn04:J:2U<C?0EoMg000Q?A00T04U0000/CWdMgSV4P3H=bXci>LB05HHjWN`MNQg
ooooo`400017A4U3200000SW4P0Lia80Fi_fM`000004jOTb000Q0?cXnC8PGOMg00000=0113H00000
00000000000Lia800@0009058CdLia80PD7dMm0113ITMP08000002D0000<00000`0004H0000X0000
700004M4BD<2000000000000000<0P00dP400000000I000030000?ooo`0B000030000080000U0000
30000000081B0000L0400040000@00001`00000000000000_080000000070P8RD`1i07<0M01U06d0
000000000000000000000000000000000000000000000000000000000000000000000000000E0000
`USdM`40000Q08X1000:03Ph=c=Jl?Ag3@0001T0/04Q08X18@00089cj7L001<000000:2U<C?h0@YX
X:Da<l1GmgL0024m402@0BD0002a>OAgN>HB0=Pg:S?Tia80EQSZMk1ej7Oooooo0@0004M4BD<80000
000B01cW4P1KVoIg000000CYnC800240o>Si<R1MmgL00000d044=P00000000009@000;2Bnc;@iQ80
40000000mGL100006@2`0A00T05O2OEgT0DQ?@00080U0000402@0Ooo000@0000n>HB0?XMm7N@1B4m
402@0BD0001TMP08000002D0000<00000@0002D0000<00001`00P1P0000<0000000001H0000<0000
00000240000800007P0001P000000000000000`2003B0@008`0002000000@1`m00000000001OR1Dm
000@@`0064<T00009000001073d00000000005n85Cd00113000H@`@0000S00008000001073d00000
000005n85Cd00113000H@b@0000T0000040L?@0000000000GhPE?@0044<001Q3100002<0000P0000
P>VO@@00000000007QV8@@0044<001Q3900002@00020jIm100000000000N6HQ1000@@`0064<40000
600000`0000000004P0000`0000100006@0000`0003oool05P0000`0000H00008`0002000020jIm1
00000000000N6HQ1000@@`0064<T00009000083YWd400000000001hIR4400113000H@`@0000S0000
80000><T3Sd00000000001hI23d00113000H@b@0000T0000hb@>?@00000000007QT8?@0044<001Q3
10000580001`0@001@00083nool0000000000000002@0@000@0000L2011D06T0K@1U07<0801>06D0
M`0P0580K`1]0640KP0000000000000000000000000000000000000000000000000a<a3[4P204P00
c>`B09D[n7Nh>OQgooooomc/4P0W/_ag60LC0000000001Tc000004H0E01G0440DP1505`0@`1/0640
L`1c06D0L`1L04<0C01C04T0A01L07/0<00`0300<P1304D0<00c02d0<00`0300<00]0300<0000300
;@130300<00`02d0<00`0300<00`0300<00`0300<00d03H0O@1L04d0J@1c06<0D`1d0640M01e07<0
001m0000YGL00>lh0000053/4P2mijEg0P03P<1W=3=/3JIg`USdM`T0002`T_/bF>`B0?Xo<C?01_Eg
0@0006Af00P000009@0000`000050000E00005@0002B0000V00009P0002V00000@0003U^hd40R^=1
=000060100010000C00000000000000000000?ooooooooooD000064000300000DP00070100040000
8?ooo`000000000000000901000100001`8045@0J@1]06D0L`0P04h0I@1g0200DP1_06d0H@1^0000
0000000000000000000000000000000000000000000007<0001m0000YGL00>lh0000053/4P2mijEg
0P03P<1W=3=/3JIg`USdM`T0001lja8000000>Dhn7L001<0R5<h<`00001Hja80R0HC00c/4P2E:oQg
83WhMooooolLk180@;OlMnP64`2@DcPcT5<h<h10<C>@DcPc<00`0000<00]04<0<00`0300;@0`0300
<00`0300<00001<0<00`03@0=P1m05`0C@1Y0000H`020000H@1d07D0L`0007d0002UM`00kcP001<0
0P00001C>3<@0000`6Ld<`4000010000S>/B05@0001Dk180UB_hMaPfn7OoooooI7H02000000U0000
300000@0000X0000300000D0001D0000E00009h0002P0000X@000:P000010000>FkS@@2:hd670@00
`0400040001<00000000000000000000oooooooooom@0000N00006<0001B0000L04000D0000Poooo
0000000000000000T040000000070P0@E01Y06d0I@1c0200CP1U07L0801B06l0K@1Q06h000000000
00000000000000000000000000000000000000007S>P^Qhc002UM`00kcP001<00P00001C>3<@0000
`6Ld<`4000010000S>/B07c[4P000000iCShM`004`28DcPc000005S[4P281Q<03>`B09D[n7LP>OQg
oooooac/4P10]oagj0HC091C>3>@DcPcc40a<i1C>3?Pja80003hMaPfn7Ooooool>/B089cj7L001<0
000000004`081PYZX;XN<l1GmgM=06T0001S0080001Q07@0_>/B0000002/kA80EQSZM`004`020000
05<h<a00002P^Qhc0@000040002<ja80E00005C/4P2E:oQg63KhMooooomTMP08000002D0000<0000
1@0002P0000<0000100005@0001D0000V@000:00002L0000Z0000040000iK^=108[S@OT000300@00
0@0004`00000000000000000003ooooooooooe00000b0000L0000580001`0@001000010000070000
00000000002l0P00000000L20R9C07T0L`1d06D0K@00000000000000000000000000000000000000
00000000000000000000000000000000000N<h2k7S>/kA80EQSZM`004`02000005<h<a00002P^Qhc
0@000040002<ja80O>/B0000003U>?Qg000C08QC>3<00000F>/B08P64`0<k180UB_hMb0in7Oooooo
7>`B042go7OX1Q<0T5<h<i1C>3<H@C4cT5<h<n3[4P000?Qg63KhMoooooo`ja80PW?XM`004`000000
000C09452V:0^ahc`5OgMdd0J@0006<00P000640M02lja8000000:c]4P1F6>Yg000C00800000DcPc
4000082k7S<100000@0008c[4P1D0000E>`B09D[n7LH=_QgooooofAf00P000009@0000`000040000
:00000`0000500009@0000`0000700209@0000`0000000209@0000`000010000B`000400000`0000
1@000200000100000@000100002@0000V0000:@0002Y0000T00009P0002T0000Z@0002P0000<0000
10000280000<0000ooooodH0000D0000200004M4BD<300008P0000`0003ooooo8P0000`0003ooooo
9@0000`0000=0020:00000`000030000:00000`0000100008P0000`0003oooooAP000<P7002l1`00
ADe6:bI0000@000010000000000U@000400000@000000000;40002@0000H0000[3b1?00008000020
J=VC?6oOhT;UWKA224051DP7000l1`000A30f`8000030000;0L00040001/0000=@0000X0002P0@00
oP40000000000000lCL00010000PADe6000102`7000U00000`000000000000000000000400000`00
@0400?0000000000000000000000hP@0P:T301T0000<0000oooo0180000<00000P0002D0000<0000
0000P2D0000<00003@00P2D0000<00001`00P1P0000<0000000001H0000<00000000024000080000
7P0001P00000000000000<X1000<0P004@0000`0000800003000010000000000000000/0000@0000
bP4000`2000H000030000000000B000030000040000I000030000?ooo`0F0000300001P0000:0000
40000000000000002@00010000300@0000800580001`0@000@00083nool0000000000000002@0@00
0@000002011D06T0K@1U07<0801>06D0M`0P0580K`1]0640KP000000000000000000000000000000
0000000000000000000a<a3[4P204P00c>`B09D[n7Nh>OQgooooomc/4P0W/_ag60LC0000000@2E/c
000004H0E01G0440DP1505`0@`1/0640L`1c06D0L`1L04<0C01C04T0A01L07/0<00`0300<P1304D0
<00c02d0<00`0300<00]0300<0000300;@130300<00`02d0<00`0300<00`0300<00`0300<00d03H0
O@1L04d0J@1c06<0D`1d0640M01e07<0001m0000YGL00?0h0000053/4P2mijEg0P03P61O=3=/3JIg
`USdM`T0002`T_/bF>`B00:b;c?01_Eg0@0006Af00P000009@0000`000010000E00005@0000e0000
2P0004l1002l0@000@0009>Dod40g?m1=000060100010000C00000000000000000000?oooooooooo
D00006d0000D0@00DP000701000200008?ooo`00000000000000090100000000008045@0J@1]06D0
L`0P04h0I@1g0200DP1_06d0H@1^00000000000000000000000000000000000000000000000007<0
001m0000YGL00?0h0000053/4P2mijEg0P03P61O=3=/3JIg`USdM`T0001lja8000000>Dhn7L001<0
l8Ph<`00001Hja80R0HC00c/4P2E:oQg83WhMooooolLk180@;OlMnP64`3hR3Pcn8Ph<hRb;c?hR3Pc
<00`0000<00]04<0<00`0300;@0`0300<00`0300<00001<0<00`03@0=P1m05`0C@1Y0000H`020000
H@1d07D0L`0007d0002UM`00l3P001<00P000028>3<@0000H5ld<`4000010000S>/B05@0001Dk180
UB_hMaPfn7OoooooI7H02000000U000030000080000X000030000040001D0000E00002h1003m0000
X0400?h100010000TiCo@@3Lod4W0@00`0400040001<00000000000000000000oooooooooom@0000
<@000700001B0000L0400040000@00001`00000000000000_080000000010P8RD`1i07<0M01U06d0
0000000000000000000000000000000000000000000000000000000000000000000000009c=0BRLc
002UM`00l3P001<00P000028>3<@0000H5ld<`4000010000S>/B07c[4P000000iCShM`004`3`R3Pc
000005S[4P281Q<03>`B09D[n7LP>OQgoooooac/4P10]oagj0HC0?R8>3?hR3Pce;8_<oR8>3?Pja80
003hMaPfn7Ooooool>/B089cj7L001<0000000004`2C1@Y9@4XW<l1GmgM=06T0001S0080001Q07@0
_>/B0000002/kA80EQSZM`004`02000008Ph<a000010BRLc0@000040002<ja80E00005C/4P2E:oQg
63KhMooooomTMP08000002D0000<00000@0002P0000<00000P0002D0000<00001`00P2D0000<0000
0000P300000<00003`00P2D0000<00003@00P4/000100000<00000D0000P00000@000040000@0000
00000000003:0@003080000000000000bP4000`2000X000030000040000R000030000?ooool>0000
50000000000@0000500001Y01D0/000080000?ooool20000000000000000P>E30403A000003X0nP3
8@0000P0001R000030000040000Q00002000024000080000L`0000`000000000700000P0000U0000
30000000080U0000300000L0080U0000300000h0080`0000300000l0080I000030000?ooo`0H0000
30000000000B000030000080000C000030000040000D0000300000d0000E000030000040000F0000
30000000000=00004000000000000000>P0000`0000:00006`00010000000000000002<0000P0000
X146?@00000000000=04?@00i4800;I2900002@0002P4@Hm000000000000d0@m003T@P00]T840000
L`0000`0000000003@000100001b0000F`000580001`0@000@0001000007000000000000002l0P00
000000L20R9C07T0L`1d06D0K@000000000000000000000000000000000000000000000000000000
0000000000000000000Q?@000000000028H/<o3X4P00QR`c@XKfMi058Cg`;o0b0`0000R6;3<001<0
1000000000000000`USdM`40000Q08X1000:03Ph=c=Jl?Ag3@0001T0/04Q08X18@00089cj7L001<0
000008285`221@ZYP8PG0<1GmgL0024m402@0BD0002a>OAgN>HB0000003Tia80EQSZMk1ej7Oooooo
0@0004M4BD<800002>LB01cW4P1KVoIg000006b8;3<00240I8P/<b1MmgL00000d044=P0000000000
000001cW4P010000T0DQ?AcW4P21@OAgd044=VAf00P000009@0000`000010000AP0002P0000L0000
AdA9@`800000000000000<X1000<0P00000001T0000<0000oooo0180000<00000P0002D0000<0000
0000P580001`0@000`0001000007000000000000002l0P00000000L20R9C07T0L`1d06D0K@000000
0000000000000000000000000000000000000000000000000000000000000000000000000032F?Ag
0@000240RP4000X0>3Pg<e[`m7L=00006@2`0B40RP4Q0000PW?XM`004`000000P8PG08852ZV0R1L0
`5OgM`008Cd@09019@000;4im7MhiQ8000000>CW4P1F6>Yg/7GXMoooool10000AdA9@`P000000180
7>LB05^KmWL00000K8P/<`008@1TR2`c85ggM`00003@0@@f00000000000U0000/9;k<]3V4P0@0000
003eM`40000I0;01402@0El9mGN@1B4m0000P2D0000@0901ool00100003hiQ80nQgdMi058Cd@0901
9@0006Af00P000009@0000`0000300009@0000`000070020600000`0000000005P0000`000000000
8@0000P0000N00006000000000000000bP4000`2000S000080000:0A1Sd000000000003@13d00>A2
002f@R@0000T0000X146?@00000000000=04?@00i4800;I2100002<0000P0000X146?@0000000000
0=04?@00i4800;I2900002@0002P4@Hm000000000000d0@m003T@P00]T8400008`0002000028ffm1
00000000003Pk8M1003T@P00]T8T0000900008SKKd40000000000>3/Qd400>A2002f@P@0000H0000
30000000000B000030000040000I000030000?ooo`0F0000300001P0000S0000800008SKKd400000
00000>3/Qd400>A2002f@R@0000T0000R=]_@@0000000000h>b7@@00i4800;I2100002<0000P0000
_0l9?@0000000000h>`7?@00i4800;I2900002@0002l3`Tm00000000003Pk0Lm003T@P00]T840000
DP00070100040000P?koo`000000000000000901000100001`8045@0J@1]06D0L`0P04h0I@1g0200
DP1_06d0H@1^000000000000000000000000000000000000000000000000034c4>/B080B003<k180
UB_hMkPin7Oooooog>`B02Nbo7LH1a<000000109Fc<00000AP1D05L0@@1B04D0G01306`0H@1c07<0
I@1c05`0@`1<05<0B@1405`0N`0`0300<00b04<0A@0`03<0;@0`0300<00`02d0<00`0000<00]04<0
<00`0300;@0`0300<00`0300<00`0300<00`03@0=P1m05`0C@1Y07<0H`1C07@0H@1d07D0L`0007d0
002UM`00l3P00000D>`B0;gWYGL200>0H5ld<f`=YWO2F?Ag2@000;2Bnc9Hk1800[8_<l06mGL10000
I7H02000000U0000300000@0001D0000E00007@0001K0000O@0006T000010000TiCo@@3Lod4d0000
H0400040001<00000000000000000000oooooooooom@0000K@3ooa@1001B0000L04000D0000Poooo
0000000000000000T040000000070P0@E01Y06d0I@1c0200CP1U07L0801B06l0K@1Q06h000000000
0000000000000000000000000000000000000000L`0007d0002UM`00l3P00000D>`B0;gWYGL200>0
H5ld<f`=YWO2F?Ag2@0007c[4P000000iCShM`004`3`R3Pc000005S[4P281Q<03>`B09D[n7LP>OQg
oooooac/4P10]oagj0HC0?R8>3?hR3PcR;8_<oR8>3<`0300000`02d0@`0`0300<00]0300<00`0300
<00`00004`0`0300=00f07d0G01=06T0001S0080001Q07@0M@1c0000O@000:Eg003`>0004`020000
08Ph<a00001PGc@c0@000040002<ja80E00005C/4P2E:oQg63KhMooooomTMP08000002D0000<0000
1@0002P0000<0000100005@0001D0000O00006<0001o0000J`000040002CU?m10=co@BL100300@00
0@0004`00000000000000000003ooooooooooe00000a0000L0000580001`0@001000010000070000
00000000002l0P00000000L20R9C07T0L`1d06D0K@00000000000000000000000000000000000000
00000000000000000000000000000000000W<d1:9c<00:Eg003`>0004`02000008Ph<a00001PGc@c
0@000040002<ja80O>/B0000003U>?Qg000C0?28>3<00000F>/B08P64`0<k180UB_hMb0in7Oooooo
7>`B042go7OX1Q<0n8Ph<oR8>3?D/Rlcn8Ph<n3[4P000?Qg63KhMoooooo`ja80PW?XM`004`000000
000C09<52TU0BRLc`5OgMdd0J@0006<00P000640M02lja8000000:c]4P1F6>Yg000C00800000R3Pc
4000041:9c<100000@0008c[4P1D0000E>`B09D[n7LH=_QgooooofAf00P000009@0000`000040000
:00000`0000500009@0000`0000700209@0000`0000000209@0000`000030000B`000400000`0000
1@000200000100000@000100001b0000F`000840001/0000LP0005/000210000K00002P0000<0000
10000280000<0000ooooodH0000D0000200004M4BD<300008P0000`0003ooooo8P0000`0003ooooo
9@0000`0000=0020:00000`000010000:00000`0000300008P0000`0003oooooAP000<P7002l1`00
ADe6:bI0000@000010000000000U@000400000@000000000;40002@0000H0000Vjb:?00008000020
J=VC?:HG<4?dNC=324061DP7000l1`000A30f`8000030000;0L00040001/0000=@0000X0002_0@00
oP400000000000001c`00010000PADe6000102`7000U00000`000000000000000000000400000`00
@0400?0000000000000000000000hP@0P:T301T0000<0000oooo0180000<00000P0002D0000<0000
0000P2D0000<00003@00P2D0000<00001`00P1P0000<0000000001H0000<00000000024000080000
7P0001P00000000000000>`1000<0P004@0000`0000800003000010000000000000000/0000@0000
k04000`2000H000030000000000B000030000040000I000030000?ooo`0F0000300001P0000:0000
40000000000000002@000100003P0@0000800580001`0@000@00083nool0000000000000002@0@00
0@000002011D06T0K@1U07<0801>06D0M`0P0580K`1]0640KP000000000000000000000000000000
0000000000000000000a<a3[4P204P00c>`B09D[n7Nh>OQgooooomc/4P0W/_ag60LC0000002`35/c
000004H0E01G0440DP1505`0@`1/0640L`1c06D0L`1L04<0C01C04T0A01L07/0<00`0300<P1304D0
<00c02d0<00`0300<00]0300<0000300;@130300<00`02d0<00`0300<00`0300<00`0300<00d03H0
O@1L04d0J@1c06<0D`1d0640M01e07<0001m0000YGL00?4h0000053/4P2mijEg0P03P1Qn<c=/3JIg
`USdM`T0002`T_/bF>`B089n5P301_Eg0@0006Af00P000009@0000`000010000E00005@0000e0000
2P0004l1002l0@000@00000P0480g?m1=000060100010000C00000000000000000000?oooooooooo
D00006d0000D0@00DP000701000200008?ooo`00000000000000090100000000008045@0J@1]06D0
L`0P04h0I@1g0200DP1_06d0H@1^00000000000000000000000000000000000000000000000007<0
001m0000YGL00?4h0000053/4P2mijEg0P03P1Qn<c=/3JIg`USdM`T0001lja8000000>Dhn7L001<0
D98h<`00001Hja80R0HC00c/4P2E:oQg83WhMooooolLk180@;OlMnP64`1HTSPcF98h<`Qo5P1HTSPc
<00`0000<00]04<0<00`0300;@0`0300<00`0300<00001<0<00`03@0=P1m05`0C@1Y0000H`020000
H@1d07D0L`0007d0002UM`00lCP001<00P00002B>3<@000067hc<`4000010000S>/B05@0001Dk180
UB_hMaPfn7OoooooI7H02000000U000030000080000X000030000040001D0000E00003d1003m0000
[`400?h1000100000200@P3Lod4e0@00`0400040001<00000000000000000000oooooooooom@0000
<P000700001B0000L0400040000@00001`00000000000000_080000000010P8RD`1i07<0M01U06d0
0000000000000000000000000000000000000000000000000000000000000000000000007S=@^Ahc
002UM`00lCP001<00P00002B>3<@000067hc<`4000010000S>/B07c[4P000000iCShM`004`1@TSPc
000005S[4P281Q<03>`B09D[n7LP>OQgoooooac/4P10]oagj0HC05RB>3=HTSPcE7lF05RB>3?Pja80
003hMaPfn7Ooooool>/B089cj7L001<0000000004`271@XZD;TN<l1GmgM=06T0001S0080001Q07@0
_>/B0000<C>/kA80EQSZM`004`020000098h<a00001@^Ahc0@000040002<ja80E00005C/4P2E:oQg
63KhMooooomTMP08000002D0000<00000@0002P0000<00000P0002D0000<00001`00P2D0000<0000
0000P300000<00003`00P2D0000<00003@00P4/000100000<00000D0000P00000@000040000@0000
00000000003/0@003080000000000000k04000`2000X000030000040000R000030000?ooool>0000
50000000000@0000500001Y01T0/000080000?ooool20000000000000000P?I30403A000003X0nP3
8@0000P0001R000030000040000Q00002000024000080000L`0000`000000000700000P0000U0000
30000000080U0000300000L0080U0000300000h0080`0000300000l0080I000030000?ooo`0H0000
30000000000B000030000080000C000030000040000D0000300000d0000E000030000040000F0000
30000000000=00004000000000000000>P0000`0000:00006`00010000000000000002<0000P0000
f>;i?000000000000=04?@00<D<003A3900002@0003Hh_Tl000000000000d0@m000a@`00=4<40000
L`0000`0000000003@000100002a0000]0000580001`0@000`0001000007000000000000002l0P00
000000L20R9C07T0L`1d06D0K@000000000000000000000000000000000000000000000000000000
0000000000000000000Q?@0000000000a9@[<o3X4P2lU2/c@XKfMi058CdH/S4c0`000<BD:c<001<0
1000000000000000`USdM`40000Q08X1000:03Ph=c=Jl?Ag3@0001T0/04Q08X18@00089cj7L001<0
000008285`221@Z[P8PG0<1GmgL0024m402@0BD0002a>OAgN>HB0000003Tia80EQSZMk1ej7Oooooo
0@0004M4BD<800002>LB01cW4P1KVoIg000002RG:c<0024089L[<b1MmgL00000d044=P0000000000
000001cW4P010000T0DQ?AcW4P21@OAgd044=VAf00P000009@0000`000030000AP0002P0000L0000
AdA9@`800000000000000>`1000<0P00000001T0000<0000oooo0180000<00000P0002D0000<0000
0000P580001`0@000@0001000007000000000000002l0P00000000L20R9C07T0L`1d06D0K@000000
0000000000000000000000000000000000000000000000000000000000000000000000000032F?Ag
0@000240RP4000X0>3Pg<e[`m7L=00006@2`0B40RP4Q0000PW?XM`004`000000P8PG08852Z^0R1L0
`5OgM`008Cd@09019@000;4im7MhiQ8000000>CW4P1F6>Yg/7GXMoooool10000AdA9@`P000000180
7>LB05^KmWL00000:9L[<`008@0PUb/c85ggM`00003@0@@f00000000000U0000/9;k<]3V4P0@0000
003eM`40000I0;01402@0El9mGN@1B4m0000P2D0000@0901ool00100003hiQ80nQgdMi058Cd@0901
9@0006Af00P000009@0000`0000100009@0000`000070020600000`0000000005P0000`000000000
8@0000P0000N00006000000000000000k04000`2000S000080000=SRnC`000000000003@13d00353
000d@b@0000T0000f>;i?000000000000=04?@00<D<003A3100002<0000P0000f>;i?00000000000
0=04?@00<D<003A3900002@0003Hh_Tl000000000000d0@m000a@`00=4<400008`000200003l7g11
00000000003Pk8M1000a@`00=4<T000090000?`OL440000000000>3/Qd400353000d@`@0000H0000
30000000000B000030000040000I000030000?ooo`0F0000300001P0000S000080000?`OL4400000
00000>3/Qd400353000d@b@0000T0000o1m`@@0000000000h>b7@@00<D<003A3100002<0000P0000
3a40?@0000000000h>`7?@00<D<003A3900002@0000?4@0m00000000003Pk0Lm000a@`00=4<40000
DP00070100040000P?koo`000000000000000901000100001`8045@0J@1]06D0L`0P04h0I@1g0200
DP1_06d0H@1^000000000000000000000000000000000000000000000000034c4>/B080B003<k180
UB_hMkPin7Oooooog>`B02Nbo7LH1a<000000;0<Fc<00000AP1D05L0@@1B04D0G01306`0H@1c07<0
I@1c05`0@`1<05<0B@1405`0N`0`0300<00b04<0A@0`03<0;@0`0300<00`02d0<00`0000<00]04<0
<00`0300;@0`0300<00`0300<00`0300<00`03@0=P1m05`0C@1Y07<0H`1C07@0H@1d07D0L`0007d0
002UM`00lCP00000D>`B0;gWYGL200>067hc<f`=YWO2F?Ag2@000;2Bnc9Hk180PWhF0<06mGL10000
I7H02000000U0000300000@0001D0000E0000;<0002d0000^`000<80000100000200@P3Lod4d0000
H0400040001<00000000000000000000oooooooooom@0000K@0001@1001B0000L04000D0000Poooo
0000000000000000T040000000070P0@E01Y06d0I@1c0200CP1U07L0801B06l0K@1Q06h000000000
0000000000000000000000000000000000000000L`0007d0002UM`00lCP00000D>`B0;gWYGL200>0
67hc<f`=YWO2F?Ag2@0007c[4P000000iCShM`004`1@TSPc000005S[4P281Q<03>`B09D[n7LP>OQg
oooooac/4P10]oagj0HC05RB>3=HTSPc27lF05RB>3<`0300000`02d0@`0`0300<00]0300<00`0300
<00`00004`0`0300=00f07d0G01=06T0001S0080001Q07@0M@1c0000O@000:Eg003a>0004`020000
098h<a00000HOS<c0@000040002<ja80E00005C/4P2E:oQg63KhMooooomTMP08000002D0000<0000
1@0002P0000<0000100005@0001D0000^`000;`0002n0000a0000040000080120=co@CD100300@00
0@0004`00000000000000000003ooooooooooe00000b0000L0000580001`0@001000010000070000
00000000002l0P00000000L20R9C07T0L`1d06D0K@00000000000000000000000000000000000000
00000000000000000000000000000000000N<e2i7S<00:Eg003a>0004`020000098h<a00000HOS<c
0@000040002<ja80O>/B0000003U>?Qg000C052B>3<00000F>/B08P64`0<k180UB_hMb0in7Oooooo
7>`B042go7OX1Q<0F98h<eRB>3=DOaH0F98h<n3[4P000?Qg63KhMoooooo`ja80PW?XM`004`000000
000C08L52RY@^Ahc`5OgMdd0J@0006<00P000640M02lja80000a<jc]4P1F6>Yg000C00800000TSPc
4000052i7S<100000@0008c[4P1D0000E>`B09D[n7LH=_QgooooofAf00P000009@0000`000040000
:00000`0000500009@0000`0000700209@0000`0000000209@0000`000010000B`000400000`0000
1@000200000100000@000100002a0000]0000<0000350000/@000;@000300000a@0002P0000<0000
10000280000<0000ooooodH0000D0000200004M4BD<300008P0000`0003ooooo8P0000`0003ooooo
9@0000`0000=0020:00000`000030000:00000`0000100008P0000`0003oooooAP000<P7002l1`00
ADe6:bI0000@000010000000000U@000400000@000000000;40002@0000H0000]Qm`?00008000020
I]VC?:@m/D9`[4I224071DP7000l1`000A30f`8000030000;0L00040001/0000>@000000001g0@00
2`80000000000000n3<00010000PADe6000102`7000U00000`000000000000000000000400000`00
@0400?0000000000000000000000hP@0P:T301T0000<0000oooo0180000<00000P0002D0000<0000
0000P2D0000<00003@00P2D0000<00001`00P1P0000<0000000001H0000<00000000024000080000
7P0001P00000000000000:X1000<0P004@0000`0000800003000010000000000000000/0000@0000
ZP4000`2000H000030000000000B000030000040000I000030000?ooo`0F0000300001P0000:0000
40000000000000002@000100002P0@0000800580001`0@000@00083nool0000000000000002@0@00
00000P02011=05@0801507P0M01b0640000000000000000000000000000000000000000000000000
0000000000000000000a<a3[4P204P00c>`B09D[n7Nh>OQgooooomc/4P0W/_ag60LC0000000@2ATc
000004H0E01G0440DP1505`0@`1/0640L`1c06D0L`1L04<0C01C04T0A01L07/0<00`0300<P1304D0
<00c02d0<00`0300<00]0300<0000300;@130300<00`02d0<00`0300<00`0300<00`0300<00d03H0
O@1L04d0J@1c06<0D`1d0640M01e07<0001m0000YGL00?8h0000053/4P2mijEg0P03P2RN<S=/3JIg
`USdM`T0002`T_/bF>`B07Z6<S?01_Eg0@0006Af00P000009@0000`000010000E00005@0000i0000
00000><0000;0P000@0003/10480g?m1>000060100010000C00000000000000000000?oooooooooo
D00006c`002W0000DP000701000200008?ooo`00000000000000090100000000008045@0J@1]06D0
L`0P04h0I@1g0200DP1_06d0H@1^00000000000000000000000000000000000000000000000007<0
001m0000YGL00?8h0000053/4P2mijEg0P03P2RN<S=/3JIg`USdM`T0001lja8000000>Dhn7L001<0
d9`h<`00001Hja80R0HC00c/4P2E:oQg83WhMooooolLk180@;OlMnP64`3HW3Pcf9`h<oR6<S?HW3Pc
<00`0000<00]04<0<00`0300;@0`0300<00`0300<00001<0<00`03@0=P1m05`0C@1Y0000H`020000
H@1d07D0L`0007d0002UM`00lSP001<00P00002L>3<@0000:9hb<`4000010000S>/B05@0001Dk180
UB_hMaPfn7OoooooI7H02000000U000030000080000X000030000040001D0000E00000D1003m0000
M`400?h100010000>`40@P3Lod7o0000`0400040001<00000000000000000000oooooooooom@0000
<@000700001B0000L0400040000@00001`00000000000000_080000000010P8RD`1i07<0M01U06d0
000000000000000000000000000000000000000000000000000000000000000000000000:3?HDBPc
002UM`00lSP001<00P00002L>3<@0000:9hb<`4000010000S>/B07c[4P000000iCShM`004`3@W3Pc
000005S[4P281Q<03>`B09D[n7LP>OQgoooooac/4P10]oagj0HC0=RL>3?HW3PcA8Lb<mRL>3?Pja80
003hMaPfn7Ooooool>/B089cj7L001<0000000004`2I0`[?f54X<l1GmgM=06T0001S0080001Q07@0
_>/B0000002/kA80EQSZM`004`02000009`h<a00003HDBPc0@000040002<ja80E00005C/4P2E:oQg
63KhMooooomTMP08000002D0000<00000@0002P0000<00000P0002D0000<00001`00P2D0000<0000
0000P300000<00003`00P2D0000<00003@00P4/000100000<00000D0000P00000@000040000@0000
00000000002Z0@003080000000000000ZP4000`2000X000030000040000R000030000?ooool>0000
50000000000@0000500001Y01d0/000080000?ooool20000000000000000P=E30403A000003X0nP3
8@0000P0001R000030000040000Q00002000024000080000L`0000`000000000700000P0000U0000
30000000080U0000300000L0080U0000300000h0080`0000300000l0080I000030000?ooo`0H0000
30000000000B000030000080000C000030000040000D0000300000d0000E000030000040000F0000
30000000000=00004000000000000000>P0000`0000:00006`00010000000000000002<0000P0000
@[86?@00000000000=04?@00/T8004Q2900002@00012/PHm000000000000d0@m002b@P00B4840000
L`0000`0000000003@000100001I0000<P000580001`0@000@0001000007000000000000002l0P00
000000L20R9C07T0L`1d06D0K@000000000000000000000000000000000000000000000000000000
0000000000000000000Q?@000000000038LX<o3X4P04QbPc@XKfMi058Cg@ORLc0`0000b7:3<001<0
1000000000000000`USdM`40000Q08X1000:03Ph=c=Jl?Ag3@0001T0/04Q08X18@00089cj7L001<0
000008285`221@Z]P8PG0<1GmgL0024m402@0BD0002a>OAgN>HB0000003Tia80EQSZMk1ej7Oooooo
0@0004M4BD<800002>LB01cW4P1KVoIg00000729:3<00240J8TX<b1MmgL00000d044=P0000000000
000001cW4P010000T0DQ?AcW4P21@OAgd044=VAf00P000009@0000`000010000AP0002P0000L0000
AdA9@`800000000000000:X1000<0P00000001T0000<0000oooo0180000<00000P0002D0000<0000
0000P580001`0@000`0001000007000000000000002l0P00000000L20R9C07T0L`1d06D0K@000000
0000000000000000000000000000000000000000000000000000000000000000000000000032F?Ag
0@000240RP4000X0>3Pg<e[`m7L=00006@2`0B40RP4Q0000PW?XM`004`000000P8PG08852Zf0R1L0
`5OgM`008Cd@09019@000;4im7MhiQ8000000>CW4P1F6>Yg/7GXMoooool10000AdA9@`P000000180
7>LB05^KmWL00000L8TX<`008@1XRBPc85ggM`00003@0@@f00000000000U0000/9;k<]3V4P0@0000
003eM`40000I0;01402@0El9mGN@1B4m0000P2D0000@0901ool00100003hiQ80nQgdMi058Cd@0901
9@0006Af00P000009@0000`0000300009@0000`000070020600000`0000000005P0000`000000000
8@0000P0000N00006000000000000000ZP4000`2000S0000800004:b1Sd000000000003@13d00;92
0018@R@0000T0000@[86?@00000000000=04?@00/T8004Q2100002<0000P0000@[86?@0000000000
0=04?@00/T8004Q2900002@00012/PHm000000000000d0@m002b@P00B48400008`000200002R9611
00000000003Pk8M1002b@P00B48T000090000:8TH440000000000>3/Qd400;920018@P@0000H0000
30000000000B000030000040000I000030000?ooo`0F0000300001P0000S000080000:8TH4400000
00000>3/Qd400;920018@R@0000T0000XRAP@@0000000000h>b7@@00/T8004Q2100002<0000P0000
:>l9?@0000000000h>`7?@00/T8004Q2900002@0000Xk`Tm00000000003Pk0Lm002b@P00B4840000
DP00070100040000P?koo`000000000000000901000000021`8044d0E00P04D0N01d0780H@000000
000000000000000000000000000000000000000000000000000000000000034c4>/B080B003<k180
UB_hMkPin7Oooooog>`B02Nbo7LH1a<0000001096C<00000AP1D05L0@@1B04D0G01306`0H@1c07<0
I@1c05`0@`1<05<0B@1405`0N`0`0300<00b04<0A@0`03<0;@0`0300<00`02d0<00`0000<00]04<0
<00`0300;@0`0300<00`0300<00`0300<00`03@0=P1m05`0C@1Y07<0H`1C07@0H@1d07D0L`0007d0
002UM`00lSP00000D>`B0;gWYGL200>0:9hb<f`=YWO2F?Ag2@000;2Bnc9Hk180NXHb<l06mGL10000
I7H02000000U0000300000@0001D0000E00005/0000b0000H000048000010000>`40@P3Lod4h0000
H0400040001<00000000000000000000oooooooooom@0000K?000:L0001B0000L04000D0000Poooo
0000000000000000T040000000070P0@E01Y06d0I@1c0200CP1U07L0801B06l0K@1Q06h000000000
0000000000000000000000000000000000000000L`0007d0002UM`00lSP00000D>`B0;gWYGL200>0
:9hb<f`=YWO2F?Ag2@0007c[4P000000iCShM`004`3@W3Pc000005S[4P281Q<03>`B09D[n7LP>OQg
oooooac/4P10]oagj0HC0=RL>3?HW3Pcn8Hb<mRL>3<`0300000`02d0@`0`0300<00]0300<00`0300
<00`00004`0`0300=00f07d0G01=06T0001S0080001Q07@0M@1c0000O@000:Eg003b>0004`020000
09`h<a00000XWS8c0@000040002<ja80E00005C/4P2E:oQg63KhMooooomTMP08000002D0000<0000
1@0002P0000<0000100005@0001D0000HP0003X0001U0000@P000040000k0@120=co@Ol000300@00
0@0004`00000000000000000003ooooooooooe00000a0000L0000580001`0@001000010000070000
00000000002l0P00000000L20R9C07T0L`1d06D0K@00000000000000000000000000000000000000
00000000000000000000000000000000000X<mQA:3<00:Eg003b>0004`02000009`h<a00000XWS8c
0@000040002<ja80O>/B0000003U>?Qg000C0=2L>3<00000F>/B08P64`0<k180UB_hMb0in7Oooooo
7>`B042go7OX1Q<0f9`h<mRL>3=4Qc8cf9`h<n3[4P000?Qg63KhMoooooo`ja80PW?XM`004`000000
000C09T32/oHDBPc`5OgMdd0J@0006<00P000640M02lja8000000:c]4P1F6>Yg000C00800000W3Pc
40000=QA:3<100000@0008c[4P1D0000E>`B09D[n7LH=_QgooooofAf00P000009@0000`000040000
:00000`0000500009@0000`0000700209@0000`0000000209@0000`000030000B`000400000`0000
1@000200000100000@000100001I0000<P0006L000130000F@000380001W0000@`0002P0000<0000
10000280000<0000ooooodH0000D0000200004M4BD<300008P0000`0003ooooo8P0000`0003ooooo
9@0000`0000=0020:00000`000010000:00000`0000300008P0000`0003oooooAP000<P7002l1`00
ADe6:bI0000@000010000000000U@000400000@000000000;40002@0000H0000[3b1?00008000020
I=VC?=ae54=T2>e224081DP7000l1`000A30f`8000030000;0L00040001/0000>@00000000250@00
2`80000000000000lCL00010000PADe6000102`7000U00000`000000000000000000000400000`00
@0400?0000000000000000000000hP@0P:T301T0000<0000oooo0180000<00000P0002D0000<0000
0000P2D0000<00003@00P2D0000<00001`00P1P0000<0000000001H0000<00000000024000080000
7P0001P00000000000000<X1000<0P004@0000`0000800003000010000000000000000/0000@0000
bP4000`2000H000030000000000B000030000040000I000030000?ooo`0F0000300001P0000:0000
40000000000000002@00010000300@0000800580001`0@000@00083nool0000000000000002@0@00
00000P02011=05@0801507P0M01b0640000000000000000000000000000000000000000000000000
0000000000000000000a<a3[4P204P00c>`B09D[n7Nh>OQgooooomc/4P0W/_ag60LC000000101`@f
000004H0E01G0440DP1505`0@`1/0640L`1c06D0L`1L04<0C01C04T0A01L07/0<00`0300<P1304D0
<00c02d0<00`0300<00]0300<0000300;@130300<00`02d0<00`0300<00`0300<00`0300<00d03H0
O@1L04d0J@1c06<0D`1d0640M01e07<0001m0000YGL00?<h0000053/4P2mijEg0P03P>27<S=/3JIg
`USdM`T0002`T_/bF>`B08YK9S?01_Eg0@0006Af00P000009@0000`000010000E00005@0000i0000
00000><0000;0P000@0009>Dod40g?m1>000060100010000C00000000000000000000?oooooooooo
D00006c`002W0000DP000701000200008?ooo`00000000000000090100000000008045@0J@1]06D0
L`0P04h0I@1g0200DP1_06d0H@1^00000000000000000000000000000000000000000000000007<0
001m0000YGL00?<h0000053/4P2mijEg0P03P>27<S=/3JIg`USdM`T0001lja8000000>Dhn7L001<0
0;h^<`00001Hja80R0HC00c/4P2E:oQg83WhMooooolLk180@;OlMnP64`08_Rhc2;h^<`QL9S<8_Rhc
<00`0000<00]04<0<00`0300;@0`0300<00`0300<00001<0<00`03@0=P1m05`0C@1Y0000H`020000
H@1d07D0L`0007d0002UM`00lcP001<00P00002n;S<@0000h8Lb<`4000010000S>/B05@0001Dk180
UB_hMaPfn7OoooooI7H02000000U000030000080000X000030000040001D0000E00001<1003m0000
Q@400?h100010000TiCo@@3Lod4=0@00`0400040001<00000000000000000000oooooooooom@0000
<P000700001B0000L0400040000@00001`00000000000000_080000000010P8RD`1i07<0M01U06d0
000000000000000000000000000000000000000000000000000000000000000000000000=C=@D3Dc
002UM`00lcP001<00P00002n;S<@0000h8Lb<`4000010000S>/B07c[4P000000iCShM`004`00_Rhc
000005S[4P281Q<03>`B09D[n7LP>OQgoooooac/4P10]oagj0HC00Rn;S<8_RhcE5`V<`Rn;S?Pja80
003hMaPfn7Ooooool>/B089cj7L001<0000000004`2Q1@XGD50e<l1GmgM=06T0001S0080001Q07@0
_>/B0000002/kA80EQSZM`004`0200000;h^<a00001@D3Dc0@000040002<ja80E00005C/4P2E:oQg
63KhMooooomTMP08000002D0000<00000@0002P0000<00000P0002D0000<00001`00P2D0000<0000
0000P300000<00003`00P2D0000<00003@00P4/000100000<00000D0000P00000@000040000@0000
00000000003:0@003080000000000000bP4000`2000X000030000040000R000030000?ooool>0000
50000000000@0000500001Y0240/000080000?ooool20000000000000000P>E30403A000003X0nP3
8@0000P0001R000030000040000Q00002000024000080000L`0000`000000000700000P0000U0000
30000000080U0000300000L0080U0000300000h0080`0000300000l0080I000030000?ooo`0H0000
30000000000B000030000080000C000030000040000D0000300000d0000E000030000040000F0000
30000000000=00004000000000000000>P0000`0000:00006`00010000000000000002<0000P0000
X146?@00000000000=04?@005D<00>i2900002@0002P4@Hm000000000000d0@m000E@`00kT840000
L`0000`0000000003@000100002E0000M`000580001`0@000`0001000007000000000000002l0P00
000000L20R9C07T0L`1d06D0K@000000000000000000000000000000000000000000000000000000
0000000000000000000Q?@0000000000[?0G0?3X4P2Tl1L0@XKfMi058CehTc8c0`000:c`5`0001<0
1000000000000000`USdM`40000Q08X1000:03Ph=c=Jl?Ag3@0001T0/04Q08X18@00089cj7L001<0
000008285`221@Z_P8PG0<1GmgL0024m402@0BD0002a>OAgN>HB0000003Tia80EQSZMk1ej7Oooooo
0@0004M4BD<800002>LB01cW4P1KVoIg0000013c5`0002402?<G021MmgL00000d044=P0000000000
000001cW4P010000T0DQ?AcW4P21@OAgd044=VAf00P000009@0000`000030000AP0002P0000L0000
AdA9@`800000000000000<X1000<0P00000001T0000<0000oooo0180000<00000P0002D0000<0000
0000P580001`0@000@0001000007000000000000002l0P00000000L20R9C07T0L`1d06D0K@000000
0000000000000000000000000000000000000000000000000000000000000000000000000032F?Ag
0@000240RP4000X0>3Pg<e[`m7L=00006@2`0B40RP4Q0000PW?XM`004`000000P8PG08852Zn0R1L0
`5OgM`008Cd@09019@000;4im7MhiQ8000000>CW4P1F6>Yg/7GXMoooool10000AdA9@`P000000180
7>LB05^KmWL000004?<G00008@08laL085ggM`00003@0@@f00000000000U0000/9;k<]3V4P0@0000
003eM`40000I0;01402@0El9mGN@1B4m0000P2D0000@0901ool00100003hiQ80nQgdMi058Cd@0901
9@0006Af00P000009@0000`0000100009@0000`000070020600000`0000000005P0000`000000000
8@0000P0000N00006000000000000000bP4000`2000S000080000:0A1Sd000000000003@13d001E3
003^@R@0000T0000X146?@00000000000=04?@005D<00>i2100002<0000P0000X146?@0000000000
0=04?@005D<00>i2900002@0002P4@Hm000000000000d0@m000E@`00kT8400008`0002000028ffm1
00000000003Pk8M1000E@`00kT8T0000900008SKKd40000000000>3/Qd4001E3003^@P@0000H0000
30000000000B000030000040000I000030000?ooo`0F0000300001P0000S0000800008SKKd400000
00000>3/Qd4001E3003^@R@0000T0000R=]_@@0000000000h>b7@@005D<00>i2100002<0000P0000
_0l9?@0000000000h>`7?@005D<00>i2900002@0002l3`Tm00000000003Pk0Lm000E@`00kT840000
DP00070100040000P?koo`000000000000000901000000021`8044d0E00P04D0N01d0780H@000000
000000000000000000000000000000000000000000000000000000000000034c4>/B080B003<k180
UB_hMkPin7Oooooog>`B02Nbo7LH1a<00000040713H00000AP1D05L0@@1B04D0G01306`0H@1c07<0
I@1c05`0@`1<05<0B@1405`0N`0`0300<00b04<0A@0`03<0;@0`0300<00`02d0<00`0000<00]04<0
<00`0300;@0`0300<00`0300<00`0300<00`03@0=P1m05`0C@1Y07<0H`1C07@0H@1d07D0L`0007d0
002UM`00lcP00000D>`B0;gWYGL200>0h8Lb<f`=YWO2F?Ag2@000;2Bnc9Hk180RU/V<l06mGL10000
I7H02000000U0000300000@0001D0000E00009L0001g0000W00008L000010000TiCo@@3Lod4h0000
H0400040001<00000000000000000000oooooooooom@0000K?000:L0001B0000L04000D0000Poooo
0000000000000000T040000000070P0@E01Y06d0I@1c0200CP1U07L0801B06l0K@1Q06h000000000
0000000000000000000000000000000000000000L`0007d0002UM`00lcP00000D>`B0;gWYGL200>0
h8Lb<f`=YWO2F?Ag2@0007c[4P000000iCShM`004`00_Rhc000005S[4P281Q<03>`B09D[n7LP>OQg
oooooac/4P10]oagj0HC00Rn;S<8_Rhc25`V<`Rn;S<`0300000`02d0@`0`0300<00]0300<00`0300
<00`00004`0`0300=00f07d0G01=06T0001S0080001Q07@0M@1c0000O@000:Eg003c>0004`020000
0;h^<a00003PQc8c0@000040002<ja80E00005C/4P2E:oQg63KhMooooomTMP08000002D0000<0000
1@0002P0000<0000100005@0001D0000WP0007l0002Q0000Q`000040002CU?m10=co@@d100300@00
0@0004`00000000000000000003ooooooooooe00000b0?ooL0000580001`0@001000010000070000
00000000002l0P00000000L20R9C07T0L`1d06D0K@00000000000000000000000000000000000000
00000000000000000000000000000000000e<e1@=C<00:Eg003c>0004`0200000;h^<a00003PQc8c
0@000040002<ja80O>/B0000003U>?Qg000C002n;S<00000F>/B08P64`0<k180UB_hMb0in7Oooooo
7>`B042go7OX1Q<02;h^<`Rn;S=DG2Hc2;h^<n3[4P000?Qg63KhMoooooo`ja80PW?XM`004`000000
000C0:452QM@D3Dc`5OgMdd0J@0006<00P000640M02lja8000000:c]4P1F6>Yg000C00800000_Rhc
4000051@=C<100000@0008c[4P1D0000E>`B09D[n7LH=_QgooooofAf00P000009@0000`000040000
:00000`0000500009@0000`0000700209@0000`0000000209@0000`000010000B`000400000`0000
1@000200000100000@000100002E0000M`000:@000280000U@0007L0002T0000R00002P0000<0000
10000280000<0000ooooodH0000D0000200004M4BD<300008P0000`0003ooooo8P0000`0003ooooo
9@0000`0000=0020:00000`000030000:00000`0000100008P0000`0003oooooAP000200000D0000
ADe6:bI0000@000010000000001<0000I000004000010000b@000<D0000100000@000<T000350000
:@2Z0000000000000020?`00000000000020?`000000000000000000000000000000000000000000
8P0000`0003oooooAP0001`0000@0000ADe6:`90000<0000000000h0000D000000000100000D0000

\>"], "Graphics",
  ImageSize->{202, 198},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  CellTags->{"b:14.0", "Building Equations"}],

Cell["\<\
Building a model in TSi Dynamics generally proceeds in the following way:
  - Define joint data.
  - Define body data.
  - Define interconnection structure.
  - Define potential energy function and generalized forces.
  - Assemble the model.
The following is a simple example.
Define Joint Data\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Building Equations", "b:14.0"}],

Cell["\<\
r1={1};H1={{1},{0},{0},{0},{0},{0}};
q1={a1x};p1={w1x};

r2={1};H2={{1},{0},{0},{0},{0},{0}};
q2={a2x};p2={w2x};

JointLst={{r1,H1,q1,p1},{r2,H2,q2,p2}};\
\>", "Input",
  CellLabel->"In[24]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Building Equations", "i:1", "b:14.0"}],

Cell["Define Body Data", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Building Equations", "b:14.0"}],

Cell["\<\
com1={0,0,-l1}; mass1=m1; out1={2,{0,0,-l1}};
Inertia1=DiagonalMatrix[{0,0,0}];

com2={0,0,-l2}; mass2=m2; out2={3,{0,0,-l2}};
Inertia2=DiagonalMatrix[{0,0,0}];

BodyLst={{com1,{out1},mass1,Inertia1},{com2,{out2},mass2,Inertia2}};\
\>", "Input",
  CellLabel->"In[29]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Building Equations", "b:14.0"}],

Cell["Define Interconnection Structure", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Building Equations", "b:14.0"}],

Cell["TreeLst={{{1,1},{2,2}}};", "Input",
  CellLabel->"In[34]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Building Equations", "b:14.0"}],

Cell["\<\
Define Potential Energy
In this case only gravity contributes to the potential energy. The only \
generalized forces are external torques acting at the two joints.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Building Equations", "b:14.0"}],

Cell["g=gc; PE=0; Q={T1,T2};", "Input",
  CellLabel->"In[35]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Building Equations", "b:14.0"}],

Cell[CellGroupData[{

Cell["\<\
{V,X,H,M,Cp,Fp,p,q}=CreateModel[JointLst,BodyLst,TreeLst,g,PE,Q];\
\>", "Input",
  CellLabel->"In[36]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Building Equations", "b:14.0"}],

Cell[BoxData[
    \("Computing Joint Kinematics"\)], "Print",
  CellLabel->"From In[36]:=",
  CellTags->"b:14.0"],

Cell[BoxData[
    \("Computing joint 1 kinematics"\)], "Print",
  CellLabel->"From In[36]:=",
  CellTags->"b:14.0"],

Cell[BoxData[
    \("Computing joint 2 kinematics"\)], "Print",
  CellLabel->"From In[36]:=",
  CellTags->"b:14.0"],

Cell[BoxData[
    \("Computing Potential Functions"\)], "Print",
  CellLabel->"From In[36]:=",
  CellTags->"b:14.0"],

Cell[BoxData[
    \("Computing Inertia Matrix"\)], "Print",
  CellLabel->"From In[36]:=",
  CellTags->"b:14.0"],

Cell[BoxData[
    \("Computing Poincare Function"\)], "Print",
  CellLabel->"From In[36]:=",
  CellTags->"b:14.0"]
}, Open  ]],

Cell["We can look at some results.", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Building Equations", "b:14.0"}],

Cell[CellGroupData[{

Cell["Print[MatrixForm[V]]", "Input",
  CellLabel->"In[37]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Building Equations", "b:14.0"}],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {
              RowBox[{"(", "\[NoBreak]", GridBox[{
                    {"1"}
                    }], "\[NoBreak]", ")"}]},
            {
              RowBox[{"(", "\[NoBreak]", GridBox[{
                    {"1"}
                    }], "\[NoBreak]", ")"}]}
            }], "\[NoBreak]", ")"}],
      MatrixForm[ {{{1}}, {{1}}}]]], "Print",
  CellLabel->"From In[37]:=",
  CellTags->"b:14.0"]
}, Open  ]],

Cell["\<\
Recall that V returns as a list of kinematic matrices- one for each joint. \
Hence in this case we get two 1x1 matrices. This can be assembled into a \
single block diagonal matrix but it is more efficient to retain the list \
form.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Building Equations", "b:14.0"}],

Cell[CellGroupData[{

Cell["Print[MatrixForm[M]]", "Input",
  CellLabel->"In[38]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Building Equations", "b:14.0"}],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {\(l1\^2\ m1 + l2\^2\ m2 + l1\ l2\ m2\ Cos[a2x] + 
                l1\ Cos[a2x]\ \((l2\ m2 + l1\ m2\ Cos[a2x])\) + 
                l1\^2\ m2\ Sin[a2x]\^2\), \(l2\^2\ m2 + 
                l1\ l2\ m2\ Cos[a2x]\)},
            {\(l2\^2\ m2 + l1\ l2\ m2\ Cos[a2x]\), \(l2\^2\ m2\)}
            }], "\[NoBreak]", ")"}],
      Function[ BoxForm`e$, 
        MatrixForm[ BoxForm`e$]]]], "Print",
  CellLabel->"From In[38]:=",
  CellTags->"b:14.0"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Print[MatrixForm[Fp]]", "Input",
  CellLabel->"In[39]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Building Equations", "b:14.0"}],

Cell[BoxData[
    InterpretationBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {\(\(-T1\) + 
                gc\ \((l1\ \((m1 + m2)\)\ Sin[a1x] + 
                      l2\ m2\ Sin[a1x + a2x])\)\)},
            {\(\(-T2\) + gc\ l2\ m2\ Sin[a1x + a2x]\)}
            }], "\[NoBreak]", ")"}],
      MatrixForm[ {
        Plus[ 
          Times[ -1, T1], 
          Times[ gc, 
            Plus[ 
              Times[ l1, 
                Plus[ m1, m2], 
                Sin[ a1x]], 
              Times[ l2, m2, 
                Sin[ 
                  Plus[ a1x, a2x]]]]]], 
        Plus[ 
          Times[ -1, T2], 
          Times[ gc, l2, m2, 
            Sin[ 
              Plus[ a1x, a2x]]]]}]]], "Print",
  CellLabel->"From In[39]:=",
  CellTags->"b:14.0"]
}, Open  ]],

Cell["\<\
When doing time consuming computations you may want to save results \
periodically and then reload them for later computations. Since, as an \
analysis proceeds, different data sets may be saved under the same file name, \
it is a good idea to delete any older file before the new one is saved.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Building Equations", "b:14.0"}],

Cell["\<\
DeleteFile[\"C:\\Projects\\ProPac\\ProPac_3\\Notebooks\\dbl_pend.dat\"];
Save[\"C:\\Projects\\ProPac\\ProPac_3\\Notebooks\\dbl_pend.dat\",p,q,V,X,H,Cp,\
Fp,M];
\
\>", "Input",
  CellLabel->"In[40]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Building Equations", "b:14.0"}],

Cell["Of course, the data is reloaded by the standard:", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Building Equations", "b:14.0"}],

Cell["<<C:\\Projects\\ProPac\\ProPac_3\\Notebooks\\dbl_pend.dat;", "Input",
  CellLabel->"In[42]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Building Equations", "b:14.0"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Simulation: Numerical Solution", "Subsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Simulation", "i:2", "b:14.1"}],

Cell["\<\
We can easily construct a simulation within Mathematica. First assemble the \
system parameter matrices computed above into a system of ordinary \
differential equations using the TSi Dynamics function MakeODEs.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Simulation", "i:2", "b:14.1"}],

Cell[CellGroupData[{

Cell["Equations=MakeODEs[p,q,V,M,Cp,Fp,t]", "Input",
  CellLabel->"In[43]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Simulation", "i:2", "b:14.1"}],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          RowBox[{
            SuperscriptBox["a1x", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], "==", \(w1x[t]\)}], 
        ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["a2x", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], "==", \(w2x[t]\)}], 
        ",", 
        RowBox[{
          
          RowBox[{\(-T1\), 
            "+", \(gc\ \((l1\ \((m1 + m2)\)\ Sin[a1x[t]] + 
                  l2\ m2\ Sin[a1x[t] + a2x[t]])\)\), 
            "+", \(w2x[
                t]\ \((\((\(-l1\)\ l2\ m2\ Sin[a2x[t]] + 
                        l1\^2\ m2\ Cos[a2x[t]]\ Sin[a2x[t]] - 
                        l1\ \((l2\ m2 + l1\ m2\ Cos[a2x[t]])\)\ Sin[
                            a2x[t]])\)\ w1x[t] - 
                  l1\ l2\ m2\ Sin[a2x[t]]\ w2x[t])\)\), "+", 
            
            RowBox[{\((l1\^2\ m1 + l2\^2\ m2 + l1\ l2\ m2\ Cos[a2x[t]] + 
                  l1\ Cos[a2x[t]]\ \((l2\ m2 + l1\ m2\ Cos[a2x[t]])\) + 
                  l1\^2\ m2\ Sin[a2x[t]]\^2)\), " ", 
              RowBox[{
                SuperscriptBox["w1x", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}], "+", 
            RowBox[{\((l2\^2\ m2 + l1\ l2\ m2\ Cos[a2x[t]])\), " ", 
              RowBox[{
                SuperscriptBox["w2x", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}]}], "==", "0"}], 
        ",", 
        RowBox[{
          
          RowBox[{\(-T2\), "+", \(gc\ l2\ m2\ Sin[a1x[t] + a2x[t]]\), 
            "-", \(1\/2\ l1\ l2\ m2\ Sin[a2x[t]]\ w1x[t]\ w2x[t]\), 
            "+", \(1\/2\ w1x[
                t]\ \((\(-\((\(-l1\)\ l2\ m2\ Sin[a2x[t]] + 
                          l1\^2\ m2\ Cos[a2x[t]]\ Sin[a2x[t]] - 
                          l1\ \((l2\ m2 + l1\ m2\ Cos[a2x[t]])\)\ Sin[
                              a2x[t]])\)\)\ w1x[t] + 
                  l1\ l2\ m2\ Sin[a2x[t]]\ w2x[t])\)\), "+", 
            RowBox[{\((l2\^2\ m2 + l1\ l2\ m2\ Cos[a2x[t]])\), " ", 
              RowBox[{
                SuperscriptBox["w1x", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}], "+", 
            RowBox[{\(l2\^2\), " ", "m2", " ", 
              RowBox[{
                SuperscriptBox["w2x", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}]}], "==", 
          "0"}]}], "}"}]], "Output",
  CellLabel->"Out[43]=",
  CellTags->"b:14.1"]
}, Open  ]],

Cell["\<\
Now, replace parameter symbols by numbers and set up initial conditions.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Simulation", "i:2", "b:14.1"}],

Cell[CellGroupData[{

Cell["DataReplacements={m1->1,l1->2,m2->1,l2->1,gc->1,T1->0,T2->0}", "Input",
  CellLabel->"In[44]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Simulation", "i:2", "b:14.1"}],

Cell[BoxData[
    \({m1 \[Rule] 1, l1 \[Rule] 2, m2 \[Rule] 1, l2 \[Rule] 1, gc \[Rule] 1, 
      T1 \[Rule] 0, T2 \[Rule] 0}\)], "Output",
  CellLabel->"Out[44]=",
  CellTags->"b:14.1"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Equations1=Equations/.DataReplacements", "Input",
  CellLabel->"In[45]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Simulation", "i:2", "b:14.1"}],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          RowBox[{
            SuperscriptBox["a1x", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], "==", \(w1x[t]\)}], 
        ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["a2x", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], "==", \(w2x[t]\)}], 
        ",", 
        RowBox[{
          
          RowBox[{\(4\ Sin[a1x[t]]\), "+", \(Sin[a1x[t] + a2x[t]]\), 
            "+", \(w2x[
                t]\ \((\((\(-2\)\ Sin[a2x[t]] + 4\ Cos[a2x[t]]\ Sin[a2x[t]] - 
                        2\ \((1 + 2\ Cos[a2x[t]])\)\ Sin[a2x[t]])\)\ w1x[t] - 
                  2\ Sin[a2x[t]]\ w2x[t])\)\), "+", 
            
            RowBox[{\((5 + 2\ Cos[a2x[t]] + 
                  2\ Cos[a2x[t]]\ \((1 + 2\ Cos[a2x[t]])\) + 
                  4\ Sin[a2x[t]]\^2)\), " ", 
              RowBox[{
                SuperscriptBox["w1x", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}], "+", 
            RowBox[{\((1 + 2\ Cos[a2x[t]])\), " ", 
              RowBox[{
                SuperscriptBox["w2x", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}]}], "==", "0"}], 
        ",", 
        RowBox[{
          
          RowBox[{\(Sin[a1x[t] + a2x[t]]\), 
            "-", \(Sin[a2x[t]]\ w1x[t]\ w2x[t]\), 
            "+", \(1\/2\ w1x[
                t]\ \((\(-\((\(-2\)\ Sin[a2x[t]] + 
                          4\ Cos[a2x[t]]\ Sin[a2x[t]] - 
                          2\ \((1 + 2\ Cos[a2x[t]])\)\ Sin[a2x[t]])\)\)\ w1x[
                      t] + 2\ Sin[a2x[t]]\ w2x[t])\)\), "+", 
            RowBox[{\((1 + 2\ Cos[a2x[t]])\), " ", 
              RowBox[{
                SuperscriptBox["w1x", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}], "+", 
            RowBox[{
              SuperscriptBox["w2x", "\[Prime]",
                MultilineFunction->None], "[", "t", "]"}]}], "==", "0"}]}], 
      "}"}]], "Output",
  CellLabel->"Out[45]=",
  CellTags->"b:14.1"]
}, Open  ]],

Cell[CellGroupData[{

Cell["vars=Join[p,q]", "Input",
  CellLabel->"In[46]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Simulation", "i:2", "b:14.1"}],

Cell[BoxData[
    \({w1x, w2x, a1x, a2x}\)], "Output",
  CellLabel->"Out[46]=",
  CellTags->"b:14.1"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
InitialConditions=Table[ToExpression[StringJoin[ToString[vars[[i]]],\"[0]==0\"\
]],{i,Length[vars]}]\
\>", "Input",
  CellLabel->"In[47]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Simulation", "i:2", "b:14.1"}],

Cell[BoxData[
    \({w1x[0] == 0, w2x[0] == 0, a1x[0] == 0, a2x[0] == 0}\)], "Output",
  CellLabel->"Out[47]=",
  CellTags->"b:14.1"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
InitialConditions={w1x[0] == 0, w2x[0] == 0, a1x[0] == .1, a2x[0] == .2}\
\>", "Input",
  CellLabel->"In[48]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Simulation", "i:2", "b:14.1"}],

Cell[BoxData[
    \({w1x[0] == 0, w2x[0] == 0, a1x[0] == 0.1`, a2x[0] == 0.2`}\)], "Output",\

  CellLabel->"Out[48]=",
  CellTags->"b:14.1"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Equations2=Join[Equations1,InitialConditions]", "Input",
  CellLabel->"In[49]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Simulation", "i:2", "b:14.1"}],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          RowBox[{
            SuperscriptBox["a1x", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], "==", \(w1x[t]\)}], 
        ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["a2x", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], "==", \(w2x[t]\)}], 
        ",", 
        RowBox[{
          
          RowBox[{\(4\ Sin[a1x[t]]\), "+", \(Sin[a1x[t] + a2x[t]]\), 
            "+", \(w2x[
                t]\ \((\((\(-2\)\ Sin[a2x[t]] + 4\ Cos[a2x[t]]\ Sin[a2x[t]] - 
                        2\ \((1 + 2\ Cos[a2x[t]])\)\ Sin[a2x[t]])\)\ w1x[t] - 
                  2\ Sin[a2x[t]]\ w2x[t])\)\), "+", 
            
            RowBox[{\((5 + 2\ Cos[a2x[t]] + 
                  2\ Cos[a2x[t]]\ \((1 + 2\ Cos[a2x[t]])\) + 
                  4\ Sin[a2x[t]]\^2)\), " ", 
              RowBox[{
                SuperscriptBox["w1x", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}], "+", 
            RowBox[{\((1 + 2\ Cos[a2x[t]])\), " ", 
              RowBox[{
                SuperscriptBox["w2x", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}]}], "==", "0"}], 
        ",", 
        RowBox[{
          
          RowBox[{\(Sin[a1x[t] + a2x[t]]\), 
            "-", \(Sin[a2x[t]]\ w1x[t]\ w2x[t]\), 
            "+", \(1\/2\ w1x[
                t]\ \((\(-\((\(-2\)\ Sin[a2x[t]] + 
                          4\ Cos[a2x[t]]\ Sin[a2x[t]] - 
                          2\ \((1 + 2\ Cos[a2x[t]])\)\ Sin[a2x[t]])\)\)\ w1x[
                      t] + 2\ Sin[a2x[t]]\ w2x[t])\)\), "+", 
            RowBox[{\((1 + 2\ Cos[a2x[t]])\), " ", 
              RowBox[{
                SuperscriptBox["w1x", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}], "+", 
            RowBox[{
              SuperscriptBox["w2x", "\[Prime]",
                MultilineFunction->None], "[", "t", "]"}]}], "==", "0"}], 
        ",", \(w1x[0] == 0\), ",", \(w2x[0] == 0\), ",", \(a1x[0] == 0.1`\), 
        ",", \(a2x[0] == 0.2`\)}], "}"}]], "Output",
  CellLabel->"Out[49]=",
  CellTags->"b:14.1"]
}, Open  ]],

Cell["\<\
Finally, integrate the equations using the Mathematica function NDSolve and \
plot the results.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Simulation", "i:2", "b:14.1"}],

Cell[CellGroupData[{

Cell["sols=NDSolve[Equations2,vars,{t,0,40}]", "Input",
  CellLabel->"In[50]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Simulation", "i:2", "b:14.1"}],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{"{", 
        RowBox[{
          RowBox[{"w1x", "\[Rule]", 
            TagBox[\(InterpolatingFunction[{{0.`, 40.`}}, "<>"]\),
              False,
              Editable->False]}], ",", 
          RowBox[{"w2x", "\[Rule]", 
            TagBox[\(InterpolatingFunction[{{0.`, 40.`}}, "<>"]\),
              False,
              Editable->False]}], ",", 
          RowBox[{"a1x", "\[Rule]", 
            TagBox[\(InterpolatingFunction[{{0.`, 40.`}}, "<>"]\),
              False,
              Editable->False]}], ",", 
          RowBox[{"a2x", "\[Rule]", 
            TagBox[\(InterpolatingFunction[{{0.`, 40.`}}, "<>"]\),
              False,
              Editable->False]}]}], "}"}], "}"}]], "Output",
  CellLabel->"Out[50]=",
  CellTags->"b:14.1"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Plot[Evaluate[a1x[t]/.sols],{t,0,40}]", "Input",
  CellLabel->"In[51]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Simulation", "i:2", "b:14.1"}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0238095 0.30933 1.67182 [
[.2619 .29683 -6 -9 ]
[.2619 .29683 6 0 ]
[.5 .29683 -6 -9 ]
[.5 .29683 6 0 ]
[.7381 .29683 -6 -9 ]
[.7381 .29683 6 0 ]
[.97619 .29683 -6 -9 ]
[.97619 .29683 6 0 ]
[.01131 .05856 -30 -4.5 ]
[.01131 .05856 0 4.5 ]
[.01131 .14215 -24 -4.5 ]
[.01131 .14215 0 4.5 ]
[.01131 .22574 -30 -4.5 ]
[.01131 .22574 0 4.5 ]
[.01131 .39292 -24 -4.5 ]
[.01131 .39292 0 4.5 ]
[.01131 .47651 -18 -4.5 ]
[.01131 .47651 0 4.5 ]
[.01131 .5601 -24 -4.5 ]
[.01131 .5601 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.2619 .30933 m
.2619 .31558 L
s
[(10)] .2619 .29683 0 1 Mshowa
.5 .30933 m
.5 .31558 L
s
[(20)] .5 .29683 0 1 Mshowa
.7381 .30933 m
.7381 .31558 L
s
[(30)] .7381 .29683 0 1 Mshowa
.97619 .30933 m
.97619 .31558 L
s
[(40)] .97619 .29683 0 1 Mshowa
.125 Mabswid
.07143 .30933 m
.07143 .31308 L
s
.11905 .30933 m
.11905 .31308 L
s
.16667 .30933 m
.16667 .31308 L
s
.21429 .30933 m
.21429 .31308 L
s
.30952 .30933 m
.30952 .31308 L
s
.35714 .30933 m
.35714 .31308 L
s
.40476 .30933 m
.40476 .31308 L
s
.45238 .30933 m
.45238 .31308 L
s
.54762 .30933 m
.54762 .31308 L
s
.59524 .30933 m
.59524 .31308 L
s
.64286 .30933 m
.64286 .31308 L
s
.69048 .30933 m
.69048 .31308 L
s
.78571 .30933 m
.78571 .31308 L
s
.83333 .30933 m
.83333 .31308 L
s
.88095 .30933 m
.88095 .31308 L
s
.92857 .30933 m
.92857 .31308 L
s
.25 Mabswid
0 .30933 m
1 .30933 L
s
.02381 .05856 m
.03006 .05856 L
s
[(-0.15)] .01131 .05856 1 0 Mshowa
.02381 .14215 m
.03006 .14215 L
s
[(-0.1)] .01131 .14215 1 0 Mshowa
.02381 .22574 m
.03006 .22574 L
s
[(-0.05)] .01131 .22574 1 0 Mshowa
.02381 .39292 m
.03006 .39292 L
s
[(0.05)] .01131 .39292 1 0 Mshowa
.02381 .47651 m
.03006 .47651 L
s
[(0.1)] .01131 .47651 1 0 Mshowa
.02381 .5601 m
.03006 .5601 L
s
[(0.15)] .01131 .5601 1 0 Mshowa
.125 Mabswid
.02381 .07528 m
.02756 .07528 L
s
.02381 .09199 m
.02756 .09199 L
s
.02381 .10871 m
.02756 .10871 L
s
.02381 .12543 m
.02756 .12543 L
s
.02381 .15887 m
.02756 .15887 L
s
.02381 .17558 m
.02756 .17558 L
s
.02381 .1923 m
.02756 .1923 L
s
.02381 .20902 m
.02756 .20902 L
s
.02381 .24246 m
.02756 .24246 L
s
.02381 .25918 m
.02756 .25918 L
s
.02381 .27589 m
.02756 .27589 L
s
.02381 .29261 m
.02756 .29261 L
s
.02381 .32605 m
.02756 .32605 L
s
.02381 .34277 m
.02756 .34277 L
s
.02381 .35948 m
.02756 .35948 L
s
.02381 .3762 m
.02756 .3762 L
s
.02381 .40964 m
.02756 .40964 L
s
.02381 .42636 m
.02756 .42636 L
s
.02381 .44308 m
.02756 .44308 L
s
.02381 .45979 m
.02756 .45979 L
s
.02381 .49323 m
.02756 .49323 L
s
.02381 .50995 m
.02756 .50995 L
s
.02381 .52667 m
.02756 .52667 L
s
.02381 .54339 m
.02756 .54339 L
s
.02381 .04184 m
.02756 .04184 L
s
.02381 .02512 m
.02756 .02512 L
s
.02381 .0084 m
.02756 .0084 L
s
.02381 .57682 m
.02756 .57682 L
s
.02381 .59354 m
.02756 .59354 L
s
.02381 .61026 m
.02756 .61026 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .47651 m
.02499 .4766 L
.02605 .47683 L
.02729 .47728 L
.02846 .47787 L
.03053 .47932 L
.03279 .48142 L
.04262 .49437 L
.0452 .49777 L
.04762 .50046 L
.04894 .50161 L
.05015 .50244 L
.05083 .50279 L
.05149 .50304 L
.05217 .50322 L
.05293 .50329 L
.05418 .50312 L
.05532 .50262 L
.05663 .50162 L
.05786 .50023 L
.05901 .49853 L
.06006 .4966 L
.06244 .49089 L
.06488 .48303 L
.0675 .47225 L
.07221 .44664 L
.07782 .40636 L
.08293 .36169 L
.09418 .24837 L
.10458 .14841 L
.10935 .11143 L
.11383 .08406 L
.11608 .07329 L
.11851 .06401 L
.11982 .06 L
.1212 .05656 L
.12238 .05424 L
.12366 .05235 L
.12434 .05163 L
.12507 .05105 L
.12575 .0507 L
.12639 .05054 L
.12759 .05065 L
.12886 .05134 L
.12952 .05192 L
.13023 .05272 L
.13151 .05457 L
.13303 .05745 L
.13442 .06066 L
.13955 .07656 L
Mistroke
.14443 .09567 L
.15404 .13601 L
.15931 .15566 L
.16427 .1709 L
.16891 .18186 L
.17153 .18661 L
.17393 .19016 L
.17818 .19481 L
.18267 .19828 L
.18501 .19993 L
.18755 .20195 L
.18995 .20433 L
.19217 .20717 L
.19476 .21151 L
.1975 .21761 L
.20235 .23293 L
.20685 .25297 L
.21107 .27704 L
.22069 .34901 L
.23921 .51375 L
.24428 .55071 L
.24707 .56744 L
.24969 .58051 L
.25214 .59016 L
.25344 .59421 L
.2548 .59769 L
.25595 .59998 L
.25721 .60182 L
.25836 .60287 L
.25943 .60332 L
.26068 .6032 L
.26138 .60283 L
.26203 .60229 L
.26323 .60083 L
.2645 .59861 L
.26669 .5932 L
.26912 .58505 L
.27448 .55992 L
.27951 .52944 L
.28979 .45828 L
.30063 .38977 L
.30528 .36648 L
.31025 .34657 L
.3157 .3305 L
.32078 .31996 L
.3304 .30495 L
.33463 .29713 L
.33928 .28563 L
.34425 .26885 L
.34951 .2454 L
Mistroke
.3606 .17857 L
.37112 .10449 L
.37612 .07189 L
.38075 .04639 L
.38325 .03527 L
.38551 .02708 L
.38803 .02031 L
.38869 .01895 L
.38941 .01769 L
.39071 .01596 L
.39185 .01505 L
.3931 .01472 L
.39441 .01511 L
.39564 .01617 L
.39678 .01778 L
.39783 .01975 L
.40017 .02594 L
.40284 .03589 L
.40531 .04772 L
.4109 .08285 L
.42105 .16769 L
.43081 .25594 L
.43593 .29798 L
.44132 .336 L
.44606 .36304 L
.44857 .37477 L
.45125 .38532 L
.45372 .39334 L
.45596 .39928 L
.46034 .40771 L
.49811 .47949 L
.50823 .5228 L
.51376 .54482 L
.51646 .55384 L
.51893 .56061 L
.52001 .56306 L
.52119 .56534 L
.52232 .56711 L
.52333 .56834 L
.52452 .56933 L
.52579 .56981 L
.52696 .56972 L
.52807 .56915 L
.52929 .56795 L
.53058 .56603 L
.53169 .56385 L
.5329 .56088 L
.5356 .55204 L
.53811 .54109 L
.54312 .51181 L
Mistroke
.54785 .47581 L
.5567 .39253 L
.56617 .29556 L
.57155 .24437 L
.5766 .20257 L
.58166 .16847 L
.58629 .145 L
.58883 .13537 L
.59124 .1283 L
.59247 .12545 L
.59378 .12295 L
.59522 .12082 L
.59588 .12007 L
.59656 .1194 L
.59781 .11853 L
.59895 .11808 L
.60019 .11795 L
.60089 .11803 L
.60154 .11819 L
.60272 .1187 L
.6038 .11937 L
.60623 .1215 L
.61073 .12693 L
.615 .13257 L
.61959 .13783 L
.62084 .13898 L
.6222 .14004 L
.62347 .14086 L
.62465 .14146 L
.62589 .1419 L
.62655 .14206 L
.62725 .14217 L
.62799 .14222 L
.62869 .1422 L
.63 .14199 L
.63122 .14161 L
.63235 .14108 L
.63359 .14033 L
.6349 .13934 L
.63943 .13453 L
.64367 .12861 L
.64868 .12115 L
.65111 .11789 L
.65332 .11545 L
.65456 .11437 L
.65573 .11359 L
.65696 .11305 L
.65766 .11289 L
.65829 .11283 L
.65955 .113 L
Mistroke
.66027 .11328 L
.66093 .11366 L
.66164 .11419 L
.66228 .11481 L
.66372 .11663 L
.66498 .11876 L
.66615 .12119 L
.66879 .12841 L
.67112 .1368 L
.6736 .14785 L
.67814 .17389 L
.68241 .20468 L
.69213 .29322 L
.70108 .38536 L
.71064 .47514 L
.71613 .51555 L
.72112 .54257 L
.72352 .55202 L
.72479 .55603 L
.72613 .55955 L
.7273 .56203 L
.72859 .5641 L
.72976 .56541 L
.73086 .56614 L
.7321 .56642 L
.7328 .56632 L
.73345 .56606 L
.73465 .5652 L
.73592 .56376 L
.73705 .56202 L
.73811 .56004 L
.74054 .55432 L
.74553 .53832 L
.75093 .51709 L
.761 .47632 L
.76633 .45814 L
.76927 .44977 L
.77204 .44303 L
.77464 .43775 L
.77746 .43312 L
.77981 .43007 L
.7824 .42741 L
.78483 .42543 L
.78741 .42366 L
.79204 .42048 L
.79478 .41799 L
.79729 .4149 L
.80018 .41003 L
.80293 .40376 L
.80546 .39637 L
Mistroke
.80781 .38796 L
.81313 .36309 L
.81821 .3318 L
.8237 .29044 L
.83366 .20203 L
.84329 .1156 L
.84799 .0794 L
.85227 .05238 L
.85447 .04116 L
.85686 .03119 L
.85817 .02675 L
.85939 .02332 L
.86054 .02066 L
.86178 .01846 L
.86302 .01694 L
.86418 .01614 L
.86486 .01596 L
.86557 .01598 L
.86684 .01658 L
.86748 .01715 L
.86817 .01796 L
.86941 .01992 L
.87088 .02307 L
.87221 .0267 L
.87492 .0361 L
.87741 .04703 L
.88305 .07821 L
.89328 .14721 L
.90233 .20666 L
.90732 .23443 L
.91205 .25629 L
.91762 .27607 L
.92271 .28899 L
.92749 .29778 L
.93269 .30567 L
.93802 .31452 L
.94304 .32582 L
.94754 .33977 L
.95241 .35959 L
.96335 .42182 L
.9734 .49268 L
.97619 .5125 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[51]:=",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/A000`40O003h00Oogoo8Goo003oOolQ
Ool00?moob5oo`00ogoo8Goo003oOolQOol00?moob5oo`00ogoo8Goo003oOolQOol0025oo`03001o
ogoo0?aoo`008Goo00<007ooOol0o7oo000QOol2003mOol0025oo`03001oogoo0?aoo`008Goo00<0
07ooOol0G7oo0`00N7oo0P008goo000QOol00`00Oomoo`1KOol200000goo001oo`1fOol200000goo
001oo`0QOol0025oo`8005aoo`05001oogooOol0001fOol01000Oomoo`008Woo000QOol00`00Oomo
o`1KOol01@00Oomoogoo0000MWoo00@007ooOol0029oo`008Goo00<007ooOol0FWoo00<007ooOol0
0Woo00<007ooOol0Lgoo00D007ooOomoo`00029oo`008Goo00<007ooOol0FWoo00<007ooOol00Woo
00<007ooOol0Lgoo00<007ooOol00Woo00<007ooOol07goo000QOol00`00Oomoo`1JOol00`00Oomo
o`02Ool00`00Oomoo`1cOol00`00Oomoo`02Ool00`00Oomoo`0OOol0025oo`8005]oo`03001oogoo
00=oo`03001oogoo079oo`03001oogoo00=oo`03001oogoo01ioo`008Goo00<007ooOol0FWoo00<0
07ooOol00goo00<007ooOol0LGoo00<007ooOol017oo00<007ooOol07Woo0008Ool20004Ool20003
Ool50002Ool20005Ool00`00Oomoo`0GOol20011Ool00`00Oomoo`03Ool00`00Oomoo`1aOol00`00
Oomoo`04Ool00`00Oomoo`0NOol000Moo`04001oogoo000:Ool01@00Oomoogoo00000Woo00<007oo
Ool00Woo00<007ooOol05Woo00@007ooOol003moo`03001oogoo00Aoo`03001oogoo075oo`03001o
ogoo00Aoo`03001oogoo01ioo`001goo00@007ooOol000Yoo`03001oogoo00Aoo`03001oogoo009o
o`8001Moo`03001oogoo008003ioo`03001oogoo00Aoo`03001oogoo075oo`03001oogoo00Aoo`03
001oogoo01ioo`0000=oo`0000000P000Woo00@007ooOol000Yoo`03001oogoo009oo`8000Eoo`03
001oogoo01Eoo`03001oogoo009oo`03001oogoo03aoo`03001oogoo00Aoo`03001oogoo075oo`03
001oogoo00Aoo`03001oogoo01ioo`001goo00@007ooOol000Yoo`03001oogoo009oo`03001oogoo
00Aoo`03001oogoo01Eoo`03001oogoo009oo`03001oogoo03]oo`03001oogoo00Ioo`03001oogoo
06moo`03001oogoo00Ioo`03001oogoo01eoo`0027oo0P002Woo0P0017oo0`0017oo00<007ooOol0
5Goo00<007ooOol00goo00<007ooOol0>Woo00<007ooOol01Woo00<007ooOol0Kgoo00<007ooOol0
1Woo00<007ooOol07Goo000QOol2000FOol00`00Oomoo`03Ool00`00Oomoo`0jOol00`00Oomoo`06
Ool00`00Oomoo`1_Ool00`00Oomoo`06Ool00`00Oomoo`0MOol0025oo`03001oogoo01Eoo`03001o
ogoo00=oo`03001oogoo03Yoo`03001oogoo00Ioo`03001oogoo06moo`03001oogoo00Ioo`03001o
ogoo01eoo`008Goo00<007ooOol057oo00<007ooOol017oo00<007ooOol0>Woo00<007ooOol01Woo
00<007ooOol0Kgoo00<007ooOol01Woo00<007ooOol07Goo000QOol00`00Oomoo`0DOol00`00Oomo
o`04Ool00`00Oomoo`0jOol00`00Oomoo`06Ool00`00Oomoo`1_Ool00`00Oomoo`06Ool00`00Oomo
o`0MOol0025oo`03001oogoo01Aoo`03001oogoo00Eoo`03001oogoo03Qoo`03001oogoo00Moo`03
001oogoo06moo`03001oogoo00Ioo`03001oogoo01eoo`008Goo0P005Goo00<007ooOol01Goo00<0
07ooOol0>7oo00<007ooOol01goo00<007ooOol0Kgoo00<007ooOol01goo00<007ooOol077oo000Q
Ool00`00Oomoo`0DOol00`00Oomoo`05Ool00`00Oomoo`0hOol00`00Oomoo`08Ool00`00Oomoo`1]
Ool00`00Oomoo`08Ool00`00Oomoo`0LOol0025oo`03001oogoo01=oo`03001oogoo00Ioo`03001o
ogoo03Qoo`03001oogoo00Qoo`03001oogoo06eoo`03001oogoo00Qoo`03001oogoo01aoo`008Goo
00<007ooOol04goo00<007ooOol01goo00<007ooOol0=goo00<007ooOol027oo00<007ooOol0KGoo
00<007ooOol027oo00<007ooOol077oo000QOol2000DOol00`00Oomoo`07Ool00`00Oomoo`0gOol0
0`00Oomoo`08Ool00`00Oomoo`1]Ool00`00Oomoo`08Ool00`00Oomoo`0LOol0025oo`03001oogoo
01=oo`03001oogoo00Moo`03001oogoo03Moo`03001oogoo00Qoo`03001oogoo03aoo`<002ioo`03
001oogoo00Qoo`03001oogoo01aoo`008Goo00<007ooOol04goo00<007ooOol01goo00<007ooOol0
=goo00<007ooOol027oo00<007ooOol0;Goo0P0037oo00@007ooOol002ioo`03001oogoo00Uoo`03
001oogoo01]oo`008Goo00<007ooOol04goo00<007ooOol027oo00<007ooOol0=Goo00<007ooOol0
2Goo00<007ooOol0;7oo10002Woo00@007ooOomoo`8002eoo`03001oogoo00Uoo`03001oogoo01]o
o`008Goo0P0057oo00<007ooOol027oo00<007ooOol0=Goo00<007ooOol02Goo00<007ooOol0;7oo
00@007ooOol000Uoo`03001oogoo00=oo`03001oogoo02]oo`03001oogoo00Uoo`03001oogoo01]o
o`008Goo00<007ooOol04goo00<007ooOol027oo00<007ooOol0=Goo00<007ooOol02Woo00<007oo
Ool0:Woo00<007ooOol00Woo00<007ooOol01Woo00<007ooOol00goo00<007ooOol0:goo00<007oo
Ool02Goo00<007ooOol06goo000QOol00`00Oomoo`0BOol00`00Oomoo`0:Ool00`00Oomoo`0dOol0
0`00Oomoo`0:Ool00`00Oomoo`0ZOol00`00Oomoo`03Ool00`00Oomoo`04Ool00`00Oomoo`04Ool0
0`00Oomoo`0ZOol00`00Oomoo`0:Ool00`00Oomoo`0KOol000ioo`8000Aoo`8000=oo`D000=oo`03
001oogoo019oo`03001oogoo00Yoo`03001oogoo03Aoo`03001oogoo00Yoo`03001oogoo02Yoo`03
001oogoo00Aoo`03001oogoo00=oo`03001oogoo00Eoo`03001oogoo02Uoo`03001oogoo00Yoo`03
001oogoo01]oo`003Goo00@007ooOol000Yoo`03001oogoo00=oo`03001oogoo019oo`03001oogoo
00Yoo`03001oogoo03Aoo`03001oogoo00Yoo`03001oogoo02Uoo`03001oogoo00Ioo`800003Ool0
000000Qoo`03001oogoo02Uoo`03001oogoo00]oo`03001oogoo01Yoo`003Goo00@007ooOol000Yo
o`03001oogoo00=oo`8001=oo`03001oogoo00Yoo`03001oogoo03=oo`03001oogoo00]oo`03001o
ogoo02Uoo`03001oogoo00Moo`<000Yoo`03001oogoo02Qoo`03001oogoo00]oo`03001oogoo01Yo
o`001goo10000Woo00@007ooOol000Yoo`03001oogoo00=oo`03001oogoo019oo`03001oogoo00]o
o`03001oogoo039oo`03001oogoo00]oo`03001oogoo02Uoo`03001oogoo01Aoo`03001oogoo02Qo
o`03001oogoo00]oo`03001oogoo01Yoo`003Goo00@007ooOol000Yoo`03001oogoo00=oo`03001o
ogoo019oo`03001oogoo00]oo`03001oogoo039oo`03001oogoo00]oo`03001oogoo02Uoo`03001o
ogoo01Aoo`03001oogoo02Qoo`03001oogoo00]oo`03001oogoo01Yoo`003Woo0P002Woo0P001Goo
00<007ooOol04Woo00<007ooOol02goo00<007ooOol0<Woo00<007ooOol037oo00<007ooOol09goo
00<007ooOol05Goo00<007ooOol0:7oo00<007ooOol02goo00<007ooOol06Woo000QOol2000COol0
0`00Oomoo`0;Ool00`00Oomoo`0bOol00`00Oomoo`0<Ool00`00Oomoo`0WOol00`00Oomoo`0EOol0
0`00Oomoo`0WOol00`00Oomoo`0<Ool00`00Oomoo`0JOol0025oo`03001oogoo019oo`03001oogoo
00]oo`03001oogoo039oo`03001oogoo00aoo`03001oogoo02Moo`03001oogoo01Ioo`03001oogoo
02Ioo`03001oogoo00eoo`03001oogoo01Uoo`008Goo00<007ooOol04Goo00<007ooOol03Goo00<0
07ooOol0<7oo00<007ooOol03Goo00<007ooOol09goo00<007ooOol05Woo00<007ooOol09Woo00<0
07ooOol03Goo00<007ooOol06Goo000QOol00`00Oomoo`0AOol00`00Oomoo`0=Ool00`00Oomoo`0`
Ool00`00Oomoo`0=Ool00`00Oomoo`0WOol00`00Oomoo`0FOol00`00Oomoo`0VOol00`00Oomoo`0=
Ool00`00Oomoo`0IOol0025oo`80019oo`03001oogoo00eoo`03001oogoo031oo`03001oogoo00eo
o`03001oogoo02Moo`03001oogoo01Ioo`03001oogoo02Ioo`03001oogoo00eoo`03001oogoo01Uo
o`008Goo00<007ooOol04Goo00<007ooOol03Woo00<007ooOol0;goo00<007ooOol03Goo00<007oo
Ool09Woo00<007ooOol05goo00<007ooOol09Woo00<007ooOol03Goo00<007ooOol06Goo000QOol0
0`00Oomoo`0AOol00`00Oomoo`0>Ool00`00Oomoo`0_Ool00`00Oomoo`0>Ool00`00Oomoo`0UOol0
0`00Oomoo`0GOol00`00Oomoo`0VOol00`00Oomoo`0=Ool00`00Oomoo`0IOol0025oo`03001oogoo
015oo`03001oogoo00moo`03001oogoo02ioo`03001oogoo00ioo`03001oogoo02Eoo`03001oogoo
01Moo`03001oogoo02Ioo`03001oogoo00eoo`03001oogoo01Uoo`008Goo00<007ooOol04Goo00<0
07ooOol047oo00<007ooOol0;7oo00<007ooOol03goo00<007ooOol09Goo00<007ooOol05goo00<0
07ooOol09Goo00<007ooOol03Woo00<007ooOol06Goo000QOol2000BOol00`00Oomoo`0AOol00`00
Oomoo`0[Ool00`00Oomoo`0?Ool00`00Oomoo`0TOol00`00Oomoo`0IOol00`00Oomoo`0TOol00`00
Oomoo`0?Ool00`00Oomoo`0HOol0025oo`03001oogoo011oo`03001oogoo01=oo`03001oogoo02Yo
o`03001oogoo00moo`03001oogoo02Aoo`03001oogoo01Uoo`03001oogoo02Aoo`03001oogoo00mo
o`03001oogoo01Qoo`008Goo00<007ooOol047oo00<007ooOol057oo00<007ooOol0:Goo00<007oo
Ool03goo00<007ooOol097oo00<007ooOol06Goo00<007ooOol097oo00<007ooOol03goo00<007oo
Ool067oo000QOol00`00Oomoo`0@Ool00`00Oomoo`0EOol00`00Oomoo`0XOol00`00Oomoo`0?Ool0
0`00Oomoo`0TOol00`00Oomoo`0IOol00`00Oomoo`0TOol00`00Oomoo`0?Ool00`00Oomoo`0HOol0
025oo`80015oo`03001oogoo01Ioo`03001oogoo02Moo`03001oogoo00moo`03001oogoo02Aoo`03
001oogoo01Uoo`03001oogoo02Aoo`03001oogoo00moo`03001oogoo01Qoo`008Goo00<007ooOol0
47oo00<007ooOol05Woo00<007ooOol09Woo00<007ooOol04Goo00<007ooOol08goo00<007ooOol0
6Goo00<007ooOol097oo00<007ooOol03goo00<007ooOol067oo0008Ool20004Ool20004Ool20004
Ool20005Ool00`00Oomoo`0@Ool00`00Oomoo`0GOol00`00Oomoo`0UOol00`00Oomoo`0AOol00`00
Oomoo`0SOol00`00Oomoo`0IOol00`00Oomoo`0TOol00`00Oomoo`0?Ool00`00Oomoo`0HOol000Mo
o`04001oogoo0008Ool01000Oomoo`000Woo00@007ooOol000Aoo`03001oogoo011oo`03001oogoo
01Moo`03001oogoo02Eoo`03001oogoo015oo`03001oogoo02=oo`03001oogoo01Uoo`03001oogoo
02Aoo`03001oogoo00moo`03001oogoo01Qoo`001goo00@007ooOol000Qoo`04001oogoo0005Ool0
0`00Oomoo`02Ool2000AOol00`00Oomoo`0GOol00`00Oomoo`0UOol00`00Oomoo`0AOol00`00Oomo
o`0ROol00`00Oomoo`0JOol00`00Oomoo`0SOol00`00Oomoo`0AOol00`00Oomoo`0GOol00003Ool0
00000080009oo`04001oogoo0008Ool01000Oomoo`000goo0P001Goo00<007ooOol047oo00<007oo
Ool067oo00<007ooOol097oo00<007ooOol04Goo00<007ooOol08Woo00<007ooOol06goo00<007oo
Ool08Woo00<007ooOol04Goo00<007ooOol05goo0007Ool01000Oomoo`0027oo00@007ooOol000=o
o`03001oogoo00Aoo`03001oogoo00moo`03001oogoo01Uoo`03001oogoo02Aoo`03001oogoo015o
o`03001oogoo029oo`03001oogoo01]oo`03001oogoo029oo`03001oogoo015oo`03001oogoo01Mo
o`0027oo0P002Woo0P0017oo0`0017oo00<007ooOol03goo00<007ooOol06Goo00<007ooOol08goo
00<007ooOol04Woo00<007ooOol08Woo00<007ooOol06goo00<007ooOol08Woo00<007ooOol04Goo
00<007ooOol05goo000QOol00`00Oomoo`0?Ool00`00Oomoo`0IOol00`00Oomoo`0SOol00`00Oomo
o`0BOol00`00Oomoo`0ROol00`00Oomoo`0KOol00`00Oomoo`0ROol00`00Oomoo`0AOol00`00Oomo
o`0GOol0025oo`80011oo`03001oogoo01Yoo`03001oogoo029oo`03001oogoo01=oo`03001oogoo
025oo`03001oogoo01]oo`03001oogoo029oo`03001oogoo019oo`03001oogoo01Ioo`008Goo00<0
07ooOol03goo00<007ooOol06Woo00<007ooOol08Woo00<007ooOol04goo00<007ooOol08Goo00<0
07ooOol06goo00<007ooOol08Woo00<007ooOol04Woo00<007ooOol05Woo000QOol00`00Oomoo`0?
Ool00`00Oomoo`0JOol00`00Oomoo`0ROol00`00Oomoo`0COol00`00Oomoo`0QOol00`00Oomoo`0K
Ool00`00Oomoo`0ROol00`00Oomoo`0COol00`00Oomoo`0EOol0025oo`03001oogoo00moo`03001o
ogoo01Yoo`03001oogoo029oo`03001oogoo01=oo`03001oogoo025oo`03001oogoo01]oo`03001o
ogoo029oo`03001oogoo01=oo`03001oogoo01Eoo`008Goo0P0047oo00<007ooOol06Woo00<007oo
Ool08Goo00<007ooOol057oo00<007ooOol08Goo00<007ooOol06goo00<007ooOol08Woo00<007oo
Ool04goo00<007ooOol05Goo000QOol00`00Oomoo`0?Ool00`00Oomoo`0JOol00`00Oomoo`0QOol0
0`00Oomoo`0DOol00`00Oomoo`0QOol00`00Oomoo`0KOol00`00Oomoo`0ROol00`00Oomoo`0COol0
0`00Oomoo`0EOol0025oo`03001oogoo00ioo`03001oogoo01aoo`03001oogoo00Ioo`D0009oo`80
015oo`03001oogoo01Aoo`03001oogoo00Yoo`@000=oo`8000ioo`03001oogoo01]oo`03001oogoo
00Ioo`8000Aoo`8001Aoo`03001oogoo01Aoo`03001oogoo00Uoo`<0009oo`8000Aoo`008Goo00<0
07ooOol03Woo00<007ooOol077oo00<007ooOol027oo00D007ooOomoo`00009oo`03001oogoo00io
o`03001oogoo01Aoo`03001oogoo00Yoo`03001oogoo00=oo`04001oogoo000<Ool00`00Oomoo`0L
Ool00`00Oomoo`05Ool01000Oomoo`000Woo00@007ooOol0019oo`03001oogoo01Eoo`03001oogoo
00Yoo`04001oogoo0002Ool00`00Oomoo`01Ool0025oo`8000moo`03001oogoo01aoo`03001oogoo
00Qoo`05001oogooOol00002Ool00`00Oomoo`0>Ool00`00Oomoo`0DOol00`00Oomoo`0;Ool00`00
Oomoo`02Ool01000Oomoo`0037oo00<007ooOol07Goo00<007ooOol01Woo00D007ooOomoo`00009o
o`03001oogoo011oo`03001oogoo01Eoo`03001oogoo00Moo`D00005Ool007ooOol00003Ool0025o
o`03001oogoo00ioo`03001oogoo01aoo`03001oogoo00Qoo`05001oogooOol00002Ool00`00Oomo
o`0=Ool00`00Oomoo`0FOol00`00Oomoo`0;Ool01@00Oomoogoo00000Woo00<007ooOol02Woo00<0
07ooOol07Goo00<007ooOol01goo00@007ooOol0009oo`03001oogoo011oo`03001oogoo01Eoo`03
001oogoo00Moo`04001oogoo0002Ool01000Oomoo`000goo000QOol00`00Oomoo`0>Ool00`00Oomo
o`0LOol00`00Oomoo`08Ool01@00Oomoogoo00000Woo00<007ooOol03Goo00<007ooOol05Woo00<0
07ooOol02Goo00@007ooOol0009oo`04001oogoo000<Ool00`00Oomoo`0MOol00`00Oomoo`04Ool0
1000Oomoo`000Woo00@007ooOol0019oo`03001oogoo01Eoo`03001oogoo00Qoo`03001oo`00009o
o`04001oogoo0003Ool0025oo`03001oogoo00ioo`03001oogoo01aoo`03001oogoo00Moo`8000Ao
o`80011oo`03001oogoo01Ioo`03001oogoo00Yoo`8000Aoo`8000eoo`03001oogoo01eoo`03001o
ogoo00Eoo`8000Aoo`8001=oo`03001oogoo01Ioo`03001oogoo00Qoo`8000=oo`8000Aoo`008Goo
00<007ooOol03Woo00<007ooOol077oo00<007ooOol07goo00<007ooOol05Woo00<007ooOol07goo
00<007ooOol07Goo00<007ooOol087oo00<007ooOol05Woo00<007ooOol04goo000QOol2000?Ool0
0`00Oomoo`0MOol00`00Oomoo`0MOol00`00Oomoo`0GOol00`00Oomoo`0OOol00`00Oomoo`0MOol0
0`00Oomoo`0POol00`00Oomoo`0GOol00`00Oomoo`0BOol0025oo`03001oogoo00ioo`03001oogoo
01eoo`03001oogoo01eoo`03001oogoo01Moo`03001oogoo01moo`03001oogoo01eoo`03001oogoo
021oo`03001oogoo01Qoo`03001oogoo015oo`008Goo00<007ooOol03Woo00<007ooOol07Goo00<0
07ooOol077oo00<007ooOol067oo00<007ooOol07goo00<007ooOol07Goo00<007ooOol07goo00<0
07ooOol06Goo00<007ooOol04Goo000QOol00`00Oomoo`0=Ool00`00Oomoo`0NOol00`00Oomoo`0K
Ool00`00Oomoo`0IOol00`00Oomoo`0OOol00`00Oomoo`0MOol00`00Oomoo`0OOol00`00Oomoo`0J
Ool00`00Oomoo`0@Ool001]oool000H000008Goo00<007ooOol02Goo00D007ooOomoo`0000Qoo`03
001oogoo00Uoo`03001oogoo00Uoo`03001oogoo00Yoo`03001oogoo00Uoo`05001oogooOol00007
Ool00`00Oomoo`0:Ool00`00Oomoo`05Ool01@00Oomoogoo000037oo00<007ooOol02Goo00<007oo
Ool00Woo00<007ooOol017oo00<007ooOol02Woo00<007ooOol02Goo0P002Woo00<007ooOol02Woo
00<007ooOol01Woo00@007ooOol000aoo`03001oogoo00Uoo`03001oo`0000Uoo`03001oogoo00Eo
o`008Goo00<007ooOol03Goo00<007ooOol07goo00<007ooOol02Goo00<007ooOol037oo00<007oo
Ool06goo00<007ooOol03Woo00<007ooOol03Goo00<007ooOol07goo00<007ooOol027oo00<007oo
Ool04goo00<007ooOol077oo00<007ooOol01Woo00<007ooOol01Goo000QOol00`00Oomoo`0=Ool0
0`00Oomoo`0OOol00`00Oomoo`0HOol00`00Oomoo`0LOol00`00Oomoo`0MOol00`00Oomoo`0OOol0
0`00Oomoo`0NOol00`00Oomoo`0LOol00`00Oomoo`0>Ool0025oo`8000ioo`03001oogoo01moo`03
001oogoo01Qoo`03001oogoo01aoo`03001oogoo01eoo`03001oogoo01moo`03001oogoo01eoo`03
001oogoo01ioo`03001oogoo00eoo`008Goo00<007ooOol03Goo00<007ooOol07goo00<007ooOol0
5goo00<007ooOol07Goo00<007ooOol07Goo00<007ooOol07goo00<007ooOol07Goo00<007ooOol0
7Woo00<007ooOol03Goo000QOol00`00Oomoo`0=Ool00`00Oomoo`0OOol00`00Oomoo`0GOol00`00
Oomoo`0MOol00`00Oomoo`0MOol00`00Oomoo`0OOol00`00Oomoo`0MOol00`00Oomoo`0NOol00`00
Oomoo`0=Ool0025oo`03001oogoo00eoo`03001oogoo01moo`03001oogoo01Ioo`03001oogoo01io
o`03001oogoo01eoo`03001oogoo01moo`03001oogoo01eoo`03001oogoo01moo`03001oogoo00ao
o`008Goo00<007ooOol03Goo00<007ooOol087oo00<007ooOol05Goo00<007ooOol07Woo00<007oo
Ool07Goo00<007ooOol07goo00<007ooOol07Goo00<007ooOol07goo00<007ooOol037oo000QOol2
000=Ool00`00Oomoo`0QOol00`00Oomoo`0DOol00`00Oomoo`0POol00`00Oomoo`0LOol00`00Oomo
o`0POol00`00Oomoo`0LOol00`00Oomoo`0OOol00`00Oomoo`0<Ool0025oo`03001oogoo00aoo`03
001oogoo025oo`03001oogoo01Aoo`03001oogoo021oo`03001oogoo01]oo`03001oogoo025oo`03
001oogoo01]oo`03001oogoo021oo`03001oogoo00aoo`008Goo00<007ooOol037oo00<007ooOol0
8Goo00<007ooOol057oo00<007ooOol087oo00<007ooOol06goo00<007ooOol08Goo00<007ooOol0
6goo00<007ooOol08Goo00<007ooOol02goo000QOol00`00Oomoo`0<Ool00`00Oomoo`0QOol00`00
Oomoo`0DOol00`00Oomoo`0POol00`00Oomoo`0KOol00`00Oomoo`0QOol00`00Oomoo`0KOol00`00
Oomoo`0QOol00`00Oomoo`0;Ool0025oo`8000eoo`03001oogoo025oo`03001oogoo01=oo`03001o
ogoo029oo`03001oogoo01Yoo`03001oogoo025oo`03001oogoo01]oo`03001oogoo025oo`03001o
ogoo00]oo`008Goo00<007ooOol037oo00<007ooOol08Goo00<007ooOol04goo00<007ooOol08Woo
00<007ooOol06Woo00<007ooOol08Goo00<007ooOol06goo00<007ooOol08Goo00<007ooOol02goo
000QOol00`00Oomoo`0<Ool00`00Oomoo`0ROol00`00Oomoo`0BOol00`00Oomoo`0ROol00`00Oomo
o`0JOol00`00Oomoo`0QOol00`00Oomoo`0KOol00`00Oomoo`0QOol00`00Oomoo`0;Ool0025oo`03
001oogoo00aoo`03001oogoo029oo`03001oogoo019oo`03001oogoo029oo`03001oogoo01Yoo`03
001oogoo025oo`03001oogoo01]oo`03001oogoo029oo`03001oogoo00Yoo`008Goo0P003Goo00<0
07ooOol08Woo00<007ooOol04Woo00<007ooOol08Woo00<007ooOol06Woo00<007ooOol08Woo00<0
07ooOol06Woo00<007ooOol08Woo00<007ooOol02Woo000QOol00`00Oomoo`0<Ool00`00Oomoo`0R
Ool00`00Oomoo`0AOol00`00Oomoo`0SOol00`00Oomoo`0IOol00`00Oomoo`0SOol00`00Oomoo`0I
Ool00`00Oomoo`0SOol00`00Oomoo`0:Ool0025oo`03001oogoo00aoo`03001oogoo029oo`03001o
ogoo015oo`03001oogoo02Aoo`03001oogoo01Qoo`03001oogoo02=oo`03001oogoo01Uoo`03001o
ogoo02=oo`03001oogoo00Yoo`0027oo0P0017oo0P0017oo0P0017oo0P001Goo00<007ooOol02goo
00<007ooOol08goo00<007ooOol04Goo00<007ooOol097oo00<007ooOol067oo00<007ooOol08goo
00<007ooOol06Goo00<007ooOol08goo00<007ooOol02Woo0007Ool01000Oomoo`0027oo00@007oo
Ool0009oo`04001oogoo0004Ool00`00Oomoo`0;Ool00`00Oomoo`0SOol00`00Oomoo`0AOol00`00
Oomoo`0TOol00`00Oomoo`0HOol00`00Oomoo`0SOol00`00Oomoo`0IOol00`00Oomoo`0SOol00`00
Oomoo`0:Ool000Moo`04001oogoo0008Ool01000Oomoo`001Goo00<007ooOol00Woo0P0037oo00<0
07ooOol08goo00<007ooOol04Goo00<007ooOol09Goo00<007ooOol05goo00<007ooOol08goo00<0
07ooOol067oo00<007ooOol09Goo00<007ooOol02Goo0007Ool01000Oomoo`0027oo00@007ooOol0
00=oo`8000Eoo`03001oogoo00]oo`03001oogoo02Aoo`03001oogoo011oo`03001oogoo02Eoo`03
001oogoo01Moo`03001oogoo02=oo`03001oogoo01Qoo`03001oogoo02Eoo`03001oogoo00Uoo`00
1goo00@007ooOol000Qoo`04001oogoo0003Ool00`00Oomoo`04Ool00`00Oomoo`0;Ool00`00Oomo
o`0TOol00`00Oomoo`0?Ool00`00Oomoo`0VOol00`00Oomoo`0GOol00`00Oomoo`0SOol00`00Oomo
o`0GOol00`00Oomoo`0VOol00`00Oomoo`09Ool000Qoo`8000Yoo`8000Aoo`<000Aoo`03001oogoo
00]oo`03001oogoo02Aoo`03001oogoo00moo`03001oogoo02Moo`03001oogoo01Ioo`03001oogoo
02=oo`03001oogoo01Moo`03001oogoo02Ioo`03001oogoo00Uoo`008Goo0P0037oo00<007ooOol0
97oo00<007ooOol03goo00<007ooOol09goo00<007ooOol05Woo00<007ooOol097oo00<007ooOol0
5Woo00<007ooOol09Woo00<007ooOol02Goo000QOol00`00Oomoo`0;Ool00`00Oomoo`0TOol00`00
Oomoo`0?Ool00`00Oomoo`0XOol00`00Oomoo`0EOol00`00Oomoo`0TOol00`00Oomoo`0EOol00`00
Oomoo`0XOol00`00Oomoo`08Ool0025oo`03001oogoo00Yoo`03001oogoo02Eoo`03001oogoo00mo
o`03001oogoo02Qoo`03001oogoo01Aoo`03001oogoo02Eoo`03001oogoo01Aoo`8002Yoo`03001o
ogoo00Qoo`008Goo00<007ooOol02Woo00<007ooOol09Goo00<007ooOol03goo00<007ooOol0:Goo
00<007ooOol04goo00<007ooOol09Goo00<007ooOol04goo00<007ooOol0:Woo00<007ooOol027oo
000QOol00`00Oomoo`0:Ool00`00Oomoo`0UOol00`00Oomoo`0>Ool00`00Oomoo`0ZOol00`00Oomo
o`0COol00`00Oomoo`0UOol00`00Oomoo`0BOol00`00Oomoo`0[Ool00`00Oomoo`08Ool0025oo`80
00]oo`03001oogoo02Eoo`03001oogoo00ioo`03001oogoo02]oo`03001oogoo019oo`03001oogoo
02Eoo`03001oogoo011oo`8002ioo`03001oogoo00Qoo`008Goo00<007ooOol02Woo00<007ooOol0
9Woo00<007ooOol03Goo00<007ooOol0:goo00<007ooOol04Woo00<007ooOol09Goo00<007ooOol0
47oo00<007ooOol0;Goo00<007ooOol027oo000QOol00`00Oomoo`0:Ool00`00Oomoo`0VOol00`00
Oomoo`0=Ool00`00Oomoo`0/Ool00`00Oomoo`0AOol00`00Oomoo`0UOol00`00Oomoo`0?Ool00`00
Oomoo`0_Ool00`00Oomoo`07Ool0025oo`03001oogoo00Uoo`03001oogoo02Moo`03001oogoo00eo
o`03001oogoo02aoo`03001oogoo015oo`03001oogoo02Eoo`03001oogoo00ioo`03001oogoo031o
o`03001oogoo00Moo`008Goo0P002Woo00<007ooOol09goo00<007ooOol03Goo00<007ooOol0;Goo
00<007ooOol047oo00<007ooOol09Goo00<007ooOol03Woo00<007ooOol0<7oo00<007ooOol01goo
000QOol00`00Oomoo`09Ool00`00Oomoo`0WOol00`00Oomoo`0<Ool00`00Oomoo`0_Ool00`00Oomo
o`0?Ool00`00Oomoo`0UOol00`00Oomoo`0>Ool00`00Oomoo`0`Ool00`00Oomoo`07Ool0025oo`03
001oogoo00Uoo`03001oogoo02Moo`03001oogoo00aoo`03001oogoo02moo`03001oogoo00moo`03
001oogoo02Eoo`03001oogoo00ioo`03001oogoo031oo`03001oogoo00Moo`008Goo00<007ooOol0
2Goo00<007ooOol09goo00<007ooOol037oo00<007ooOol0<7oo00<007ooOol03Goo00<007ooOol0
9goo00<007ooOol037oo00<007ooOol0<Goo00<007ooOol01goo000QOol20009Ool00`00Oomoo`0X
Ool00`00Oomoo`0<Ool00`00Oomoo`0`Ool00`00Oomoo`0=Ool00`00Oomoo`0WOol00`00Oomoo`0<
Ool00`00Oomoo`0bOol00`00Oomoo`06Ool0025oo`03001oogoo00Qoo`03001oogoo02Qoo`03001o
ogoo00aoo`03001oogoo035oo`03001oogoo00aoo`03001oogoo02Moo`03001oogoo00aoo`03001o
ogoo039oo`03001oogoo00Ioo`008Goo00<007ooOol027oo00<007ooOol0:Goo00<007ooOol02goo
00<007ooOol0<Goo00<007ooOol037oo00<007ooOol09goo00<007ooOol02goo00<007ooOol0<goo
00<007ooOol01Woo000>Ool20004Ool20003Ool50003Ool00`00Oomoo`08Ool00`00Oomoo`0YOol0
0`00Oomoo`0:Ool00`00Oomoo`0cOol00`00Oomoo`0;Ool00`00Oomoo`0WOol00`00Oomoo`0;Ool0
0`00Oomoo`0cOol00`00Oomoo`06Ool000eoo`04001oogoo000:Ool00`00Oomoo`03Ool00`00Oomo
o`08Ool00`00Oomoo`0YOol00`00Oomoo`0:Ool00`00Oomoo`0cOol00`00Oomoo`0;Ool00`00Oomo
o`0WOol00`00Oomoo`0:Ool00`00Oomoo`0dOol00`00Oomoo`06Ool000eoo`04001oogoo000:Ool0
0`00Oomoo`03Ool20008Ool00`00Oomoo`0ZOol00`00Oomoo`0:Ool00`00Oomoo`0dOol00`00Oomo
o`0:Ool00`00Oomoo`0WOol00`00Oomoo`0:Ool00`00Oomoo`0dOol00`00Oomoo`06Ool000eoo`04
001oogoo000:Ool00`00Oomoo`03Ool00`00Ool00007Ool00`00Oomoo`0ZOol00`00Oomoo`0:Ool0
0`00Oomoo`0dOol00`00Oomoo`0:Ool00`00Oomoo`0WOol00`00Oomoo`0:Ool00`00Oomoo`0eOol0
0`00Oomoo`05Ool000eoo`04001oogoo000:Ool00`00Oomoo`03Ool01000Oomoo`001Woo00<007oo
Ool0:Woo00<007ooOol02Woo00<007ooOol0=7oo00<007ooOol02Woo00<007ooOol0:7oo00<007oo
Ool02Goo00<007ooOol0=Goo00<007ooOol01Goo000>Ool2000:Ool20005Ool01@00Oomoogoo0000
1Goo00<007ooOol0:Woo00<007ooOol02Woo00<007ooOol0=Goo00<007ooOol02Goo00<007ooOol0
:7oo00<007ooOol027oo00<007ooOol0=Woo00<007ooOol01Goo000QOol20003Ool01@00Oomoogoo
0000;Goo00<007ooOol02Goo00<007ooOol0=Woo00<007ooOol027oo00<007ooOol0:Goo00<007oo
Ool027oo00<007ooOol0=Woo00<007ooOol01Goo000QOol00`00Oomoo`02Ool01@00Oomoogoo0000
;Woo00<007ooOol027oo00<007ooOol0=Woo00<007ooOol027oo00<007ooOol0:Goo00<007ooOol0
27oo00<007ooOol0=Woo00<007ooOol01Goo000QOol00`00Oomoo`03Ool4000^Ool00`00Oomoo`08
Ool00`00Oomoo`0fOol00`00Oomoo`08Ool00`00Oomoo`0YOol00`00Oomoo`08Ool00`00Oomoo`0f
Ool00`00Oomoo`05Ool0025oo`03001oogoo00Aoo`03001oogoo02ioo`03001oogoo00Qoo`03001o
ogoo03Moo`03001oogoo00Moo`03001oogoo02Uoo`03001oogoo00Qoo`03001oogoo03Ioo`03001o
ogoo00Eoo`008Goo0P00=Woo00<007ooOol027oo00<007ooOol0=goo00<007ooOol01goo00<007oo
Ool0:Woo00<007ooOol01Woo00<007ooOol0=goo00<007ooOol01Goo000QOol00`00Oomoo`0eOol0
0`00Oomoo`08Ool00`00Oomoo`0gOol00`00Oomoo`07Ool00`00Oomoo`0ZOol00`00Oomoo`06Ool0
0`00Oomoo`0gOol00`00Oomoo`05Ool0025oo`03001oogoo03Eoo`03001oogoo00Moo`03001oogoo
03Qoo`03001oogoo00Moo`03001oogoo02Yoo`03001oogoo00Ioo`03001oogoo03moo`008Goo00<0
07ooOol0=Goo00<007ooOol01goo00<007ooOol0>Goo00<007ooOol01Goo00<007ooOol0:goo00<0
07ooOol01Woo00<007ooOol0?goo000QOol00`00Oomoo`0eOol00`00Oomoo`07Ool00`00Oomoo`0i
Ool00`00Oomoo`05Ool00`00Oomoo`0[Ool00`00Oomoo`05Ool00`00Oomoo`10Ool0025oo`8003Io
o`03001oogoo00Moo`03001oogoo03Uoo`03001oogoo00Eoo`03001oogoo02]oo`03001oogoo00Eo
o`03001oogoo041oo`008Goo00<007ooOol0=Goo00<007ooOol01goo00<007ooOol0>Goo00<007oo
Ool01Goo00<007ooOol0;7oo00<007ooOol017oo00<007ooOol0@7oo000QOol00`00Oomoo`0fOol0
0`00Oomoo`06Ool00`00Oomoo`0jOol00`00Oomoo`03Ool00`00Oomoo`0]Ool00`00Oomoo`03Ool0
0`00Oomoo`11Ool0025oo`03001oogoo03Ioo`03001oogoo00Ioo`03001oogoo03Yoo`03001oogoo
00=oo`03001oogoo02eoo`03001oogoo00=oo`03001oogoo045oo`008Goo0P00=goo00<007ooOol0
1Goo00<007ooOol0>goo00<007ooOol00goo00<007ooOol0;Goo00<007ooOol00goo00<007ooOol0
@Goo000QOol00`00Oomoo`0fOol00`00Oomoo`05Ool00`00Oomoo`0kOol00`00Oomoo`03Ool00`00
Oomoo`0]Ool00`00Oomoo`02Ool00`00Oomoo`12Ool000Qoo`8000Aoo`8000=oo`D0009oo`8000Eo
o`03001oogoo03Ioo`03001oogoo00Eoo`03001oogoo03aoo`03001oogoo009oo`03001oogoo02io
o`05001oogooOol00014Ool000Moo`04001oogoo000:Ool01@00Oomoogoo00000Woo00<007ooOol0
0Woo00<007ooOol0=Woo00<007ooOol01Goo00<007ooOol0?7oo00D007ooOomoo`00035oo`05001o
ogooOol00014Ool000Moo`04001oogoo000:Ool00`00Oomoo`04Ool00`00Oomoo`02Ool2000gOol0
0`00Oomoo`05Ool00`00Oomoo`0lOol01@00Oomoogoo0000<Goo00@007ooOol004Eoo`001goo00@0
07ooOol000Yoo`03001oogoo009oo`8000Eoo`03001oogoo03Moo`03001oogoo00Aoo`03001oogoo
03eoo`04001oogoo000bOol00`00Ool00015Ool000Moo`04001oogoo000:Ool00`00Oomoo`02Ool0
0`00Oomoo`04Ool00`00Oomoo`0gOol00`00Oomoo`04Ool00`00Oomoo`0mOol3000cOol30015Ool0
00Qoo`8000Yoo`8000Aoo`<000Aoo`03001oogoo03Moo`03001oogoo00Aoo`03001oogoo03ioo`80
07]oo`008Goo00<007ooOol0=goo00<007ooOol00goo00<007ooOol0_7oo000QOol2000hOol00`00
Oomoo`03Ool00`00Oomoo`2lOol0025oo`03001oogoo03Moo`03001oogoo00=oo`03001oogoo0;ao
o`008Goo00<007ooOol0=goo00<007ooOol00goo00<007ooOol0_7oo000QOol00`00Oomoo`0hOol0
1@00Oomoogoo0000_goo000QOol2000iOol01@00Oomoogoo0000_goo000QOol00`00Oomoo`0iOol0
0`00Ool00030Ool0025oo`03001oogoo03Uoo`<00<1oo`008Goo00<007ooOol0>Woo00<007ooOol0
_goo000QOol2003mOol0025oo`03001oogoo0?aoo`008Goo00<007ooOol0o7oo003oOolQOol00?mo
ob5oo`00ogoo8Goo003oOolQOol00?moob5oo`00ogoo8Goo003oOolQOol00?moob5oo`00\
\>"],
  ImageRangeCache->{{{81.0625, 368.063}, {643.063, 466.125}} -> {-18.4704, \
0.868849, 0.161578, 0.00230113}},
  CellTags->"b:14.1"],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  Graphics  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output",
  CellLabel->"Out[51]=",
  CellTags->"b:14.1"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Plot[Evaluate[a2x[t]/.sols],{t,0,40}]", "Input",
  CellLabel->"In[52]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Simulation", "i:2", "b:14.1"}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0238095 0.308908 1.47205 [
[.2619 .29641 -6 -9 ]
[.2619 .29641 6 0 ]
[.5 .29641 -6 -9 ]
[.5 .29641 6 0 ]
[.7381 .29641 -6 -9 ]
[.7381 .29641 6 0 ]
[.97619 .29641 -6 -9 ]
[.97619 .29641 6 0 ]
[.01131 .0145 -24 -4.5 ]
[.01131 .0145 0 4.5 ]
[.01131 .1617 -24 -4.5 ]
[.01131 .1617 0 4.5 ]
[.01131 .45611 -18 -4.5 ]
[.01131 .45611 0 4.5 ]
[.01131 .60332 -18 -4.5 ]
[.01131 .60332 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.2619 .30891 m
.2619 .31516 L
s
[(10)] .2619 .29641 0 1 Mshowa
.5 .30891 m
.5 .31516 L
s
[(20)] .5 .29641 0 1 Mshowa
.7381 .30891 m
.7381 .31516 L
s
[(30)] .7381 .29641 0 1 Mshowa
.97619 .30891 m
.97619 .31516 L
s
[(40)] .97619 .29641 0 1 Mshowa
.125 Mabswid
.07143 .30891 m
.07143 .31266 L
s
.11905 .30891 m
.11905 .31266 L
s
.16667 .30891 m
.16667 .31266 L
s
.21429 .30891 m
.21429 .31266 L
s
.30952 .30891 m
.30952 .31266 L
s
.35714 .30891 m
.35714 .31266 L
s
.40476 .30891 m
.40476 .31266 L
s
.45238 .30891 m
.45238 .31266 L
s
.54762 .30891 m
.54762 .31266 L
s
.59524 .30891 m
.59524 .31266 L
s
.64286 .30891 m
.64286 .31266 L
s
.69048 .30891 m
.69048 .31266 L
s
.78571 .30891 m
.78571 .31266 L
s
.83333 .30891 m
.83333 .31266 L
s
.88095 .30891 m
.88095 .31266 L
s
.92857 .30891 m
.92857 .31266 L
s
.25 Mabswid
0 .30891 m
1 .30891 L
s
.02381 .0145 m
.03006 .0145 L
s
[(-0.2)] .01131 .0145 1 0 Mshowa
.02381 .1617 m
.03006 .1617 L
s
[(-0.1)] .01131 .1617 1 0 Mshowa
.02381 .45611 m
.03006 .45611 L
s
[(0.1)] .01131 .45611 1 0 Mshowa
.02381 .60332 m
.03006 .60332 L
s
[(0.2)] .01131 .60332 1 0 Mshowa
.125 Mabswid
.02381 .04394 m
.02756 .04394 L
s
.02381 .07338 m
.02756 .07338 L
s
.02381 .10282 m
.02756 .10282 L
s
.02381 .13226 m
.02756 .13226 L
s
.02381 .19114 m
.02756 .19114 L
s
.02381 .22058 m
.02756 .22058 L
s
.02381 .25003 m
.02756 .25003 L
s
.02381 .27947 m
.02756 .27947 L
s
.02381 .33835 m
.02756 .33835 L
s
.02381 .36779 m
.02756 .36779 L
s
.02381 .39723 m
.02756 .39723 L
s
.02381 .42667 m
.02756 .42667 L
s
.02381 .48555 m
.02756 .48555 L
s
.02381 .515 m
.02756 .515 L
s
.02381 .54444 m
.02756 .54444 L
s
.02381 .57388 m
.02756 .57388 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .60332 m
.02499 .60256 L
.02605 .60055 L
.02729 .59668 L
.02846 .59151 L
.03053 .57879 L
.03279 .56006 L
.0379 .5006 L
.04262 .42935 L
.05285 .25533 L
.05786 .17994 L
.06244 .1266 L
.06485 .10627 L
.06595 .09888 L
.06713 .09241 L
.06838 .08718 L
.06955 .08384 L
.07026 .08252 L
.07093 .08179 L
.07218 .08175 L
.07348 .08353 L
.07414 .08512 L
.07486 .0874 L
.07614 .09279 L
.07733 .09934 L
.08002 .11924 L
.0829 .1478 L
.09328 .29239 L
.09876 .37559 L
.10176 .41705 L
.10458 .45114 L
.10732 .47849 L
.10982 .49774 L
.11118 .50562 L
.11261 .51188 L
.11335 .5143 L
.11404 .51603 L
.1148 .51733 L
.11561 .51804 L
.11693 .51768 L
.11768 .51664 L
.11836 .51516 L
.11962 .51111 L
.12026 .50842 L
.12096 .50498 L
.12347 .48848 L
.12582 .46748 L
.13089 .4062 L
.13552 .33599 L
.14594 .16538 L
Mistroke
.15072 .09926 L
.15334 .07003 L
.15581 .04809 L
.15812 .03295 L
.15942 .02686 L
.16062 .02285 L
.16162 .02068 L
.16273 .01958 L
.16389 .01988 L
.16496 .02149 L
.16626 .02511 L
.16698 .02797 L
.16766 .03112 L
.17021 .04742 L
.17252 .06801 L
.17498 .09542 L
.1857 .26352 L
.1951 .42513 L
.19767 .46266 L
.20038 .49664 L
.20272 .52072 L
.204 .53166 L
.20521 .54041 L
.20635 .54723 L
.20759 .55301 L
.20871 .55681 L
.20974 .55906 L
.21072 .56009 L
.21179 .56 L
.21291 .55854 L
.21395 .55594 L
.21519 .55133 L
.21636 .54546 L
.21856 .53068 L
.22093 .50961 L
.2235 .48124 L
.23282 .34692 L
.23771 .27046 L
.24287 .19833 L
.24561 .16663 L
.24862 .13901 L
.24994 .12949 L
.25135 .12122 L
.25203 .11794 L
.25268 .11528 L
.25389 .11149 L
.25515 .1092 L
.25578 .10869 L
.25648 .10862 L
.25771 .10978 L
Mistroke
.25886 .11234 L
.25956 .11457 L
.26022 .11716 L
.26146 .12326 L
.26425 .14264 L
.26673 .16593 L
.26897 .19148 L
.27399 .26101 L
.28302 .40546 L
.28808 .48014 L
.29075 .51345 L
.29356 .54254 L
.29603 .5621 L
.29743 .57051 L
.29873 .57651 L
.29993 .58042 L
.30103 .58259 L
.30229 .58345 L
.30349 .58262 L
.30463 .58033 L
.30586 .57624 L
.30702 .57084 L
.30808 .5646 L
.31055 .54546 L
.31315 .5187 L
.31805 .45222 L
.32341 .36238 L
.33289 .19448 L
.33755 .12486 L
.34184 .07603 L
.34424 .05654 L
.34545 .04901 L
.34677 .04258 L
.34747 .03998 L
.34811 .03809 L
.34884 .03646 L
.34954 .03548 L
.35078 .03507 L
.35147 .0356 L
.35212 .03657 L
.35335 .03968 L
.3545 .04409 L
.35553 .04925 L
.35665 .05607 L
.35916 .07593 L
.36146 .09927 L
.37166 .24432 L
.37729 .33508 L
.38242 .4097 L
.38476 .43863 L
Mistroke
.38731 .46516 L
.38971 .48489 L
.39193 .49809 L
.39315 .50317 L
.39428 .50653 L
.39553 .50862 L
.39623 .50906 L
.39687 .509 L
.39816 .50755 L
.39888 .50595 L
.39956 .50397 L
.40079 .49909 L
.40211 .49215 L
.40449 .47527 L
.40707 .45111 L
.41173 .39412 L
.42046 .26123 L
.42505 .1913 L
.42996 .12695 L
.43278 .0979 L
.43536 .07764 L
.43659 .0702 L
.43725 .06692 L
.43794 .06393 L
.43921 .05978 L
.44039 .05752 L
.44169 .0568 L
.44288 .0578 L
.44418 .0607 L
.44484 .06288 L
.44555 .06581 L
.44677 .07205 L
.44807 .08048 L
.4504 .10009 L
.45338 .13277 L
.45619 .1706 L
.46145 .25504 L
.47117 .42573 L
.47648 .50596 L
.47908 .53767 L
.48146 .56131 L
.48363 .57802 L
.48463 .58406 L
.48569 .58931 L
.48682 .59354 L
.48787 .5962 L
.48911 .59779 L
.49024 .59775 L
.49149 .59607 L
.49264 .59302 L
Mistroke
.49389 .58803 L
.49522 .58096 L
.49746 .56498 L
.49991 .54196 L
.51079 .38806 L
.51614 .29892 L
.52108 .22214 L
.52557 .16454 L
.52783 .14154 L
.52913 .13043 L
.53033 .1217 L
.53152 .11441 L
.53282 .10813 L
.53413 .10362 L
.53485 .10194 L
.53551 .10087 L
.53674 .10014 L
.53786 .10091 L
.53915 .10346 L
.54037 .1075 L
.54154 .1128 L
.54263 .11901 L
.5451 .13742 L
.54752 .16067 L
.55011 .19073 L
.55907 .32072 L
.56416 .39782 L
.56958 .46827 L
.57181 .49135 L
.57421 .51143 L
.57558 .52046 L
.57684 .52708 L
.57805 .53186 L
.57932 .53522 L
.58053 .5368 L
.58183 .53671 L
.58305 .53493 L
.58417 .53188 L
.58547 .52657 L
.58689 .51874 L
.58947 .49917 L
.59226 .47073 L
.59522 .43331 L
.60053 .35186 L
.61022 .18669 L
.61556 .10736 L
.61819 .07553 L
.62064 .05138 L
.623 .03352 L
.62432 .02602 L
Mistroke
.62552 .02081 L
.62683 .01687 L
.62758 .01549 L
.62826 .01474 L
.62952 .01472 L
.63015 .01536 L
.63084 .01657 L
.63206 .01997 L
.63317 .02446 L
.63446 .03128 L
.63567 .03932 L
.64018 .08171 L
.64473 .14189 L
.64894 .20866 L
.65848 .37073 L
.66311 .43951 L
.66555 .4697 L
.66811 .49597 L
.67036 .51386 L
.67164 .5217 L
.67283 .52744 L
.67392 .53139 L
.67492 .53384 L
.67607 .53532 L
.67715 .53539 L
.67834 .53398 L
.67965 .53066 L
.68031 .5283 L
.68102 .52525 L
.68229 .51843 L
.68477 .50061 L
.68704 .47939 L
.69253 .41181 L
.69775 .33422 L
.70737 .19342 L
.71014 .16092 L
.71274 .1358 L
.71508 .11845 L
.7164 .11102 L
.7176 .10576 L
.71867 .10238 L
.71982 .10011 L
.72089 .09925 L
.72188 .09957 L
.72303 .10127 L
.72428 .10475 L
.72547 .10957 L
.72656 .11533 L
.72874 .13047 L
.73114 .15248 L
Mistroke
.73601 .21189 L
.74678 .3844 L
.75209 .46844 L
.75703 .533 L
.75932 .55651 L
.76148 .57424 L
.76267 .58207 L
.76397 .58889 L
.76514 .59355 L
.76622 .59651 L
.76743 .59831 L
.76855 .59854 L
.76978 .59719 L
.77048 .59569 L
.77111 .59386 L
.77238 .58887 L
.77376 .58145 L
.77627 .56283 L
.77866 .53927 L
.7809 .51241 L
.78602 .43697 L
.79639 .25583 L
.80075 .18488 L
.80535 .12336 L
.80776 .0984 L
.80903 .08749 L
.81039 .07767 L
.81165 .07037 L
.81279 .06522 L
.81393 .0615 L
.81502 .05932 L
.81617 .05844 L
.8174 .05915 L
.81857 .06137 L
.81964 .0647 L
.82089 .07014 L
.82225 .07793 L
.82473 .09684 L
.82731 .12251 L
.83006 .15565 L
.83501 .22635 L
.84462 .37483 L
.84884 .43068 L
.85107 .45535 L
.85346 .47716 L
.85481 .4872 L
.85605 .49478 L
.85729 .50084 L
.85847 .50511 L
.85978 .50816 L
Mistroke
.861 .50933 L
.86164 .5093 L
.86234 .50877 L
.86307 .50766 L
.86376 .50609 L
.86441 .50413 L
.86512 .50149 L
.86655 .49455 L
.86912 .47698 L
.87211 .44866 L
.87489 .41567 L
.88458 .2689 L
.88967 .18763 L
.89507 .11299 L
.89768 .08404 L
.89907 .07106 L
.90055 .05913 L
.90192 .0501 L
.90316 .04353 L
.90442 .03859 L
.9056 .03548 L
.9068 .03392 L
.90807 .03401 L
.90928 .03577 L
.91036 .03875 L
.91165 .04396 L
.91303 .05163 L
.91555 .07075 L
.9178 .09332 L
.92026 .12342 L
.92524 .19827 L
.93414 .35624 L
.93941 .44528 L
.942 .48342 L
.94439 .51405 L
.94651 .53702 L
.94886 .55717 L
.94999 .56486 L
.95121 .57159 L
.9525 .57692 L
.95371 .58023 L
.95438 .58137 L
.9551 .58202 L
.95639 .58174 L
.95718 .58066 L
.95792 .579 L
.95859 .577 L
.95931 .57429 L
.96066 .56774 L
.96209 .55876 L
Mistroke
.96464 .53768 L
.96941 .48321 L
.97464 .40673 L
.97619 .38215 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  CellLabel->"From In[52]:=",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/A000`40O003h00Oogoo8Goo003oOolQ
Ool00?moob5oo`00ogoo8Goo003oOolQOol00?moob5oo`006goo00<007ooOol0ogoo0goo000KOol0
0`00Oomoo`3oOol3Ool000Qoo`8000Aoo`8000=oo`@000Aoo`03001oogoo0?moo`=oo`001goo00@0
07ooOol000Qoo`03001oogoo00Eoo`03001oogoo0?moo`=oo`001goo00@007ooOol000Uoo`03001o
ogoo00Aoo`8009ioo`8006=oo`0000=oo`0000000P000Woo00@007ooOol000Yoo`03001oogoo00=o
o`03001oogoo025oo`8007Yoo`03001oo`00069oo`001goo00@007ooOol000Qoo`04001oogoo0004
Ool00`00Oomoo`0QOol2001jOol00`00Ool0001ROol000Qoo`8000Yoo`8000Eoo`03001oogoo025o
o`03001oo`0007Uoo`03001oo`00069oo`006goo00<007ooOol08Goo00<007oo0000N7oo00@007oo
Ool0069oo`006goo00<007ooOol08Goo00<007oo0000;goo0P00Agoo00@007ooOol004Moo`8001Uo
o`006goo00<007ooOol08Goo00<007oo0000;goo0P00Agoo00D007ooOomoo`0004Ioo`03001oo`00
01Qoo`006goo00<007ooOol087oo00D007ooOomoo`0002ioo`8004Moo`05001oogooOol00016Ool0
0`00Ool0000HOol001]oo`80025oo`05001oogooOol0000]Ool01000Oomoo`00AWoo00D007ooOomo
o`0004Eoo`04001oogoo000HOol001]oo`03001oogoo021oo`05001oogooOol0000]Ool01000Oomo
o`00AWoo00D007ooOomoo`0004Eoo`04001oogoo000HOol001]oo`03001oogoo021oo`05001oogoo
Ool0000]Ool01000Oomoo`00AWoo00D007ooOomoo`0004Eoo`04001oogoo000HOol001]oo`03001o
ogoo021oo`05001oogooOol0000]Ool01000Oomoo`005Woo00<007ooOol0;Goo00D007ooOomoo`00
02moo`03001oogoo01=oo`04001oogoo000HOol001]oo`03001oogoo01moo`03001oogoo009oo`03
001oogoo02]oo`04001oogoo000EOol00`00Ool0000^Ool01@00Oomoogoo0000;Woo0P005Goo00@0
07ooOol001Qoo`006goo00<007ooOol07goo00<007ooOol00Woo00<007ooOol0:goo00@007ooOol0
01Eoo`03001oo`0002eoo`03001oogoo00=oo`03001oogoo02]oo`03001oo`0001Aoo`05001oogoo
Ool0000GOol001]oo`03001oogoo01moo`03001oogoo009oo`03001oogoo02Yoo`03001oogoo009o
o`03001oogoo019oo`03001oo`0002eoo`03001oogoo00=oo`03001oogoo02Yoo`04001oogoo000D
Ool01@00Oomoogoo00005goo000KOol2000POol00`00Oomoo`02Ool00`00Oomoo`0ZOol00`00Oomo
o`02Ool00`00Oomoo`0BOol00`00Ool0000]Ool00`00Oomoo`03Ool00`00Oomoo`0ZOol01000Oomo
o`0057oo00D007ooOomoo`0001Moo`006goo00<007ooOol07goo00<007ooOol00Woo00<007ooOol0
:Woo00<007ooOol00Woo00<007ooOol04Goo00@007ooOol002eoo`03001oogoo00=oo`03001oogoo
02Yoo`04001oogoo000DOol01@00Oomoogoo00005goo000KOol00`00Oomoo`0OOol00`00Oomoo`02
Ool00`00Oomoo`0ZOol00`00Oomoo`02Ool00`00Oomoo`0AOol01@00Oomoogoo0000;7oo00<007oo
Ool00goo00<007ooOol0:Woo00@007ooOol001=oo`03001oogoo009oo`03001oogoo01Eoo`006goo
00<007ooOol02Goo0P0057oo00<007ooOol00goo00<007ooOol0:Goo00<007ooOol00Woo00<007oo
Ool04Goo00D007ooOomoo`0002aoo`03001oogoo00=oo`03001oogoo02Yoo`04001oogoo000COol0
0`00Oomoo`02Ool00`00Oomoo`0EOol001]oo`03001oogoo00Uoo`8001Aoo`03001oogoo00=oo`03
001oogoo02Uoo`03001oogoo009oo`03001oogoo015oo`05001oogooOol0000/Ool00`00Oomoo`03
Ool00`00Oomoo`0ZOol01000Oomoo`004goo00<007ooOol00goo00<007ooOol057oo000KOol00`00
Oomoo`08Ool00`00Ool0000DOol00`00Oomoo`03Ool00`00Oomoo`0YOol00`00Oomoo`03Ool00`00
Oomoo`0@Ool01@00Oomoogoo0000:goo00<007ooOol017oo00<007ooOol0:Woo00D007ooOomoo`00
019oo`03001oogoo00=oo`03001oogoo01Aoo`006goo00<007ooOol027oo00@007ooOol001=oo`03
001oogoo00=oo`03001oogoo02Uoo`03001oogoo00=oo`03001oogoo011oo`05001oogooOol0000[
Ool00`00Oomoo`04Ool00`00Oomoo`0YOol00`00Oomoo`02Ool00`00Oomoo`0@Ool00`00Oomoo`03
Ool00`00Oomoo`0DOol001]oo`03001oogoo00Qoo`04001oogoo000COol00`00Oomoo`03Ool00`00
Oomoo`0YOol00`00Oomoo`03Ool00`00Oomoo`0@Ool01@00Oomoogoo00005Woo00<007ooOol04Woo
00<007ooOol017oo00<007ooOol04Woo0P005Goo00<007ooOol00Woo00<007ooOol047oo00<007oo
Ool00goo00<007ooOol057oo000KOol20009Ool01000Oomoo`004goo00<007ooOol00goo00<007oo
Ool0:7oo00<007ooOol017oo00<007ooOol047oo00D007ooOomoo`0001Ioo`8001=oo`03001oogoo
00Aoo`03001oogoo019oo`8001Eoo`03001oogoo009oo`03001oogoo011oo`03001oogoo00=oo`03
001oogoo01Aoo`006goo00<007ooOol027oo00@007ooOol001=oo`03001oogoo00=oo`03001oogoo
02Qoo`03001oogoo00Aoo`03001oogoo011oo`05001oogooOol0000EOol00`00Ool0000COol00`00
Oomoo`04Ool00`00Oomoo`0AOol00`00Ool0000EOol00`00Oomoo`02Ool00`00Oomoo`0@Ool00`00
Oomoo`03Ool00`00Oomoo`0DOol001]oo`03001oogoo00Qoo`04001oogoo000COol00`00Oomoo`03
Ool00`00Oomoo`0BOol2000DOol00`00Oomoo`04Ool00`00Oomoo`0@Ool01@00Oomoogoo00005Goo
00<007oo00004goo00<007ooOol017oo00<007ooOol04Goo00<007oo00005Goo00<007ooOol00Woo
00<007ooOol047oo00<007ooOol00goo00<007ooOol057oo000KOol00`00Oomoo`08Ool01@00Oomo
ogoo00004Woo00<007ooOol00goo00<007ooOol04Woo0`004goo00<007ooOol017oo00<007ooOol0
3goo00<007ooOol00Woo00<007ooOol04goo00@007ooOol0019oo`03001oogoo00Aoo`03001oogoo
015oo`03001oo`0001Eoo`03001oogoo00=oo`03001oogoo00moo`03001oogoo00=oo`03001oogoo
01Aoo`006goo00<007ooOol01goo00<007ooOol00Woo00<007ooOol03goo00<007ooOol017oo00<0
07ooOol04Woo00<007oo00004goo00<007ooOol017oo00<007ooOol03goo00<007ooOol00goo00<0
07ooOol04Woo00@007ooOol0019oo`03001oogoo00Eoo`03001oogoo011oo`04001oogoo000DOol0
0`00Oomoo`03Ool00`00Oomoo`0?Ool00`00Oomoo`03Ool00`00Oomoo`0DOol001]oo`03001oogoo
00Moo`03001oogoo009oo`03001oogoo00moo`03001oogoo00Aoo`03001oogoo015oo`04001oogoo
000COol00`00Oomoo`04Ool00`00Oomoo`0?Ool00`00Oomoo`03Ool00`00Oomoo`0BOol01000Oomo
o`004Woo00<007ooOol01Goo00<007ooOol047oo00@007ooOol001Aoo`03001oogoo00=oo`03001o
ogoo00moo`03001oogoo00=oo`03001oogoo01Aoo`006goo00<007ooOol01goo00<007ooOol00Woo
00<007ooOol03goo00<007ooOol01Goo00<007ooOol047oo00@007ooOol001=oo`03001oogoo00Ao
o`03001oogoo00moo`03001oogoo00=oo`03001oogoo019oo`04001oogoo000BOol00`00Oomoo`05
Ool00`00Oomoo`0@Ool01000Oomoo`0057oo00<007ooOol00goo00<007ooOol03goo00<007ooOol0
0goo00<007ooOol057oo000KOol00`00Oomoo`07Ool00`00Oomoo`02Ool00`00Oomoo`0?Ool00`00
Oomoo`05Ool00`00Oomoo`0@Ool01000Oomoo`004goo00<007ooOol017oo00<007ooOol03goo00<0
07ooOol00goo00<007ooOol04Goo00D007ooOomoo`00019oo`03001oogoo00Eoo`03001oogoo00mo
o`05001oogooOol0000DOol00`00Oomoo`03Ool00`00Oomoo`0?Ool00`00Oomoo`03Ool00`00Oomo
o`0DOol001]oo`8000Qoo`03001oogoo009oo`03001oogoo00moo`03001oogoo00Eoo`03001oogoo
011oo`04001oogoo000COol00`00Oomoo`04Ool00`00Oomoo`0?Ool00`00Oomoo`03Ool00`00Oomo
o`0AOol01@00Oomoogoo00004Woo00<007ooOol01Goo00<007ooOol03goo00D007ooOomoo`0001Ao
o`03001oogoo00=oo`03001oogoo00ioo`03001oogoo00Aoo`03001oogoo01Aoo`006goo00<007oo
Ool01goo00<007ooOol00goo00<007ooOol03Woo00<007ooOol01Goo00<007ooOol047oo00D007oo
Oomoo`00019oo`03001oogoo00Eoo`03001oogoo00ioo`03001oogoo00=oo`03001oogoo015oo`05
001oogooOol0000BOol00`00Oomoo`05Ool00`00Oomoo`0?Ool01@00Oomoogoo000057oo00<007oo
Ool00goo00<007ooOol03Woo00<007ooOol017oo00<007ooOol057oo000KOol00`00Oomoo`07Ool0
0`00Oomoo`03Ool00`00Oomoo`0>Ool00`00Oomoo`05Ool00`00Oomoo`0@Ool01@00Oomoogoo0000
4Woo00<007ooOol01Goo00<007ooOol03Woo00<007ooOol00goo00<007ooOol04Goo00D007ooOomo
o`00019oo`03001oogoo00Eoo`03001oogoo00moo`05001oogooOol0000COol00`00Oomoo`04Ool0
0`00Oomoo`0>Ool00`00Oomoo`04Ool00`00Oomoo`0DOol001]oo`03001oogoo00Moo`03001oogoo
00=oo`03001oogoo00ioo`03001oogoo00Eoo`03001oogoo011oo`05001oogooOol0000BOol00`00
Oomoo`05Ool00`00Oomoo`0>Ool00`00Oomoo`03Ool00`00Oomoo`0AOol01@00Oomoogoo00004Woo
00<007ooOol01Goo00<007ooOol03goo00<007ooOol00Woo00<007ooOol047oo00<007ooOol017oo
00<007ooOol03Woo00<007ooOol017oo00<007ooOol057oo000KOol00`00Oomoo`07Ool00`00Oomo
o`03Ool00`00Oomoo`0>Ool00`00Oomoo`05Ool00`00Oomoo`0@Ool01@00Oomoogoo00004Woo00<0
07ooOol01Goo00<007ooOol03Woo00<007ooOol00goo00<007ooOol04Goo00D007ooOomoo`00015o
o`03001oogoo00Ioo`03001oogoo00moo`03001oogoo009oo`03001oogoo011oo`03001oogoo00Ao
o`03001oogoo00ioo`03001oogoo00Eoo`03001oogoo01=oo`006goo00<007ooOol01goo00<007oo
Ool00goo00<007ooOol03Goo00<007ooOol01Woo00<007ooOol047oo00D007ooOomoo`00019oo`03
001oogoo00Eoo`03001oogoo00ioo`03001oogoo00=oo`03001oogoo015oo`03001oogoo009oo`03
001oogoo00ioo`03001oogoo00Ioo`03001oogoo00ioo`03001oogoo00=oo`03001oogoo011oo`03
001oogoo00Aoo`03001oogoo00ioo`03001oogoo00Eoo`03001oogoo01=oo`0027oo0P0017oo0P00
0goo1@000goo00<007ooOol01Woo00<007ooOol017oo00<007ooOol03Goo00<007ooOol01Woo00<0
07ooOol03goo00<007ooOol00Woo00<007ooOol047oo00<007ooOol01Goo00<007ooOol03Woo00<0
07ooOol017oo00<007ooOol03goo00<007ooOol00goo00<007ooOol03Woo00<007ooOol01Woo00<0
07ooOol03Woo00<007ooOol00goo00<007ooOol047oo00<007ooOol017oo00<007ooOol03Woo00<0
07ooOol01Goo00<007ooOol04goo0007Ool01000Oomoo`002Woo00<007ooOol00goo00<007ooOol0
1Woo00<007ooOol017oo00<007ooOol03Goo00<007ooOol01Woo00<007ooOol03goo00<007ooOol0
0Woo00<007ooOol047oo00<007ooOol01Goo00<007ooOol03Woo00<007ooOol017oo00<007ooOol0
3goo00<007ooOol00goo00<007ooOol03Woo00<007ooOol01Woo00<007ooOol03Woo00<007ooOol0
0goo00<007ooOol047oo00<007ooOol017oo00<007ooOol03Woo00<007ooOol01Goo00<007ooOol0
4goo0007Ool01000Oomoo`002Woo00<007ooOol00goo0P001goo00<007ooOol017oo00<007ooOol0
3Goo00<007ooOol01Woo00<007ooOol03goo00<007ooOol00Woo00<007ooOol03goo00<007ooOol0
1Woo00<007ooOol03Goo00<007ooOol01Goo00<007ooOol03goo00<007ooOol00goo00<007ooOol0
3Woo00<007ooOol01Woo00<007ooOol03Woo00<007ooOol00goo00<007ooOol047oo00<007ooOol0
17oo00<007ooOol03Woo00<007ooOol01Goo00<007ooOol04goo00000goo000000020002Ool01000
Oomoo`002Woo00<007ooOol00goo00<007ooOol01Woo00<007ooOol017oo00<007ooOol03Goo00<0
07ooOol01Woo00<007ooOol03goo00<007ooOol00Woo00<007ooOol03goo00<007ooOol01Woo00<0
07ooOol03Goo00<007ooOol01Goo00<007ooOol03goo00<007ooOol00goo00<007ooOol03Woo00<0
07ooOol01Woo00<007ooOol03Woo00<007ooOol00goo00<007ooOol047oo00<007ooOol017oo00<0
07ooOol03Woo00<007ooOol01Goo00<007ooOol04goo0007Ool01000Oomoo`002Woo00<007ooOol0
0goo00<007ooOol01Woo00<007ooOol017oo00<007ooOol03Goo00<007ooOol01Woo00<007ooOol0
3goo00<007ooOol00Woo00<007ooOol03goo00<007ooOol01Woo00<007ooOol03Goo00<007ooOol0
1Goo00<007ooOol03goo00<007ooOol00goo00<007ooOol03Woo00<007ooOol01Woo00<007ooOol0
3Woo00<007ooOol00goo00<007ooOol03goo00<007ooOol01Goo00<007ooOol03Goo00<007ooOol0
1Woo00<007ooOol04goo0008Ool2000:Ool20005Ool00`00Oomoo`06Ool00`00Oomoo`04Ool00`00
Oomoo`0=Ool00`00Oomoo`06Ool00`00Oomoo`0?Ool00`00Oomoo`02Ool00`00Oomoo`0?Ool00`00
Oomoo`06Ool00`00Oomoo`0=Ool00`00Oomoo`05Ool00`00Oomoo`0?Ool00`00Oomoo`03Ool00`00
Oomoo`0>Ool00`00Oomoo`06Ool00`00Oomoo`0>Ool00`00Oomoo`03Ool00`00Oomoo`0?Ool00`00
Oomoo`06Ool00`00Oomoo`0<Ool00`00Oomoo`06Ool00`00Oomoo`0COol001]oo`03001oogoo00Io
o`03001oogoo00Aoo`03001oogoo00eoo`03001oogoo00Ioo`03001oogoo00moo`03001oogoo009o
o`03001oogoo00moo`03001oogoo00Ioo`03001oogoo00eoo`03001oogoo00Eoo`03001oogoo00mo
o`03001oogoo00=oo`03001oogoo00ioo`03001oogoo00Moo`03001oogoo00eoo`03001oogoo00=o
o`03001oogoo00moo`03001oogoo00Ioo`03001oogoo00aoo`03001oogoo00Ioo`03001oogoo01=o
o`006goo00<007ooOol01Woo00<007ooOol017oo00<007ooOol03Goo00<007ooOol01goo00<007oo
Ool03Woo00<007ooOol00goo00<007ooOol03Woo00<007ooOol01Woo00<007ooOol03Goo00<007oo
Ool01Goo00<007ooOol03goo00<007ooOol017oo00<007ooOol03Goo00<007ooOol01goo00<007oo
Ool03Goo00<007ooOol00goo00<007ooOol03goo00<007ooOol01Woo00<007ooOol037oo00<007oo
Ool01Woo00<007ooOol04goo000KOol00`00Oomoo`06Ool00`00Oomoo`05Ool00`00Oomoo`0<Ool0
0`00Oomoo`07Ool00`00Oomoo`0=Ool00`00Oomoo`04Ool00`00Oomoo`0>Ool00`00Oomoo`06Ool0
0`00Oomoo`0=Ool00`00Oomoo`05Ool00`00Oomoo`0?Ool00`00Oomoo`04Ool00`00Oomoo`0=Ool0
0`00Oomoo`07Ool00`00Oomoo`0=Ool00`00Oomoo`04Ool00`00Oomoo`0>Ool00`00Oomoo`06Ool0
0`00Oomoo`0<Ool00`00Oomoo`07Ool00`00Oomoo`0BOol001]oo`03001oogoo00Ioo`03001oogoo
00Eoo`03001oogoo00aoo`03001oogoo00Moo`03001oogoo00eoo`03001oogoo00Aoo`03001oogoo
00ioo`03001oogoo00Ioo`03001oogoo00eoo`03001oogoo00Eoo`03001oogoo00moo`03001oogoo
00Aoo`03001oogoo00eoo`03001oogoo00Moo`03001oogoo00eoo`03001oogoo00Aoo`03001oogoo
00ioo`03001oogoo00Ioo`03001oogoo00aoo`03001oogoo00Moo`03001oogoo019oo`006goo0P00
1goo00<007ooOol01Goo00<007ooOol037oo00<007ooOol01goo00<007ooOol03Goo00<007ooOol0
17oo00<007ooOol03Woo00<007ooOol01Woo00<007ooOol03Goo00<007ooOol01Goo00<007ooOol0
3goo00<007ooOol017oo00<007ooOol03Goo00<007ooOol01goo00<007ooOol03Goo00<007ooOol0
17oo00<007ooOol03Woo00<007ooOol01Woo00<007ooOol037oo00<007ooOol01goo00<007ooOol0
4Woo000KOol00`00Oomoo`06Ool00`00Oomoo`05Ool00`00Oomoo`0<Ool00`00Oomoo`07Ool00`00
Oomoo`0=Ool00`00Oomoo`04Ool00`00Oomoo`0>Ool00`00Oomoo`06Ool00`00Oomoo`0=Ool00`00
Oomoo`05Ool00`00Oomoo`0>Ool00`00Oomoo`05Ool00`00Oomoo`0=Ool00`00Oomoo`07Ool00`00
Oomoo`0=Ool00`00Oomoo`04Ool00`00Oomoo`0>Ool00`00Oomoo`06Ool00`00Oomoo`0<Ool00`00
Oomoo`07Ool00`00Oomoo`0BOol001]oo`03001oogoo00Ioo`03001oogoo00Eoo`03001oogoo00ao
o`03001oogoo00Moo`03001oogoo00eoo`03001oogoo00Aoo`03001oogoo00ioo`03001oogoo00Io
o`03001oogoo00eoo`03001oogoo00Eoo`03001oogoo00ioo`03001oogoo00Eoo`03001oogoo00eo
o`03001oogoo00Moo`03001oogoo00eoo`03001oogoo00Aoo`03001oogoo00ioo`03001oogoo00Io
o`03001oogoo00aoo`03001oogoo00Moo`03001oogoo019oo`006goo00<007ooOol01Woo00<007oo
Ool01Goo00<007ooOol037oo00<007ooOol01goo00<007ooOol03Goo00<007ooOol017oo00<007oo
Ool03Woo00<007ooOol01Woo00<007ooOol03Goo00<007ooOol01Goo00<007ooOol03Woo00<007oo
Ool01Goo00<007ooOol03Goo00<007ooOol01goo00<007ooOol03Goo00<007ooOol017oo00<007oo
Ool03Woo00<007ooOol01Woo00<007ooOol037oo00<007ooOol01goo00<007ooOol04Woo000KOol0
0`00Oomoo`06Ool00`00Oomoo`05Ool00`00Oomoo`0<Ool00`00Oomoo`07Ool00`00Oomoo`0=Ool0
0`00Oomoo`04Ool00`00Oomoo`0>Ool00`00Oomoo`06Ool00`00Oomoo`0=Ool00`00Oomoo`05Ool0
0`00Oomoo`0>Ool00`00Oomoo`05Ool00`00Oomoo`0=Ool00`00Oomoo`07Ool00`00Oomoo`0=Ool0
0`00Oomoo`04Ool00`00Oomoo`0>Ool00`00Oomoo`06Ool00`00Oomoo`0<Ool00`00Oomoo`07Ool0
0`00Oomoo`0BOol001]oo`03001oogoo00Ioo`03001oogoo00Eoo`03001oogoo00]oo`03001oogoo
00Qoo`03001oogoo00eoo`03001oogoo00Aoo`03001oogoo00ioo`03001oogoo00Moo`03001oogoo
00]oo`03001oogoo00Ioo`03001oogoo00ioo`03001oogoo00Eoo`03001oogoo00eoo`03001oogoo
00Moo`03001oogoo00eoo`03001oogoo00Aoo`03001oogoo00ioo`03001oogoo00Moo`03001oogoo
00]oo`03001oogoo00Moo`03001oogoo019oo`006goo00<007ooOol01Woo00<007ooOol01Goo00<0
07ooOol02goo00<007ooOol027oo00<007ooOol03Goo00<007ooOol017oo00<007ooOol03Woo00<0
07ooOol01goo00<007ooOol02goo00<007ooOol01goo00<007ooOol03Goo00<007ooOol01Goo00<0
07ooOol03Goo00<007ooOol01goo00<007ooOol03Goo00<007ooOol017oo00<007ooOol03Woo00<0
07ooOol01goo00<007ooOol02goo00<007ooOol01goo00<007ooOol04Woo000KOol00`00Oomoo`05
Ool00`00Oomoo`06Ool00`00Oomoo`0;Ool00`00Oomoo`08Ool00`00Oomoo`0=Ool00`00Oomoo`04
Ool00`00Oomoo`0>Ool00`00Oomoo`07Ool00`00Oomoo`0;Ool00`00Oomoo`07Ool00`00Oomoo`0=
Ool00`00Oomoo`05Ool00`00Oomoo`0<Ool00`00Oomoo`08Ool00`00Oomoo`0<Ool00`00Oomoo`05
Ool00`00Oomoo`0>Ool00`00Oomoo`07Ool00`00Oomoo`0;Ool00`00Oomoo`07Ool00`00Oomoo`0B
Ool001]oo`8000Ioo`03001oogoo00Ioo`03001oogoo00]oo`03001oogoo00Qoo`03001oogoo00eo
o`03001oogoo00Aoo`03001oogoo00ioo`03001oogoo00Moo`03001oogoo00]oo`03001oogoo00Mo
o`03001oogoo00eoo`03001oogoo00Eoo`03001oogoo00aoo`03001oogoo00Qoo`03001oogoo00ao
o`03001oogoo00Eoo`03001oogoo00eoo`03001oogoo00Qoo`03001oogoo00]oo`03001oogoo00Mo
o`03001oogoo019oo`006goo00<007ooOol01Goo00<007ooOol01Woo00<007ooOol02goo00<007oo
Ool027oo00<007ooOol03Goo00<007ooOol017oo00<007ooOol03Woo00<007ooOol01goo00<007oo
Ool02goo00<007ooOol01goo00<007ooOol03Goo00<007ooOol01Woo00<007ooOol02goo00<007oo
Ool027oo00<007ooOol037oo00<007ooOol01Goo00<007ooOol03Goo00<007ooOol027oo00<007oo
Ool02goo00<007ooOol01goo00<007ooOol04Woo000KOol00`00Oomoo`05Ool00`00Oomoo`06Ool0
0`00Oomoo`0;Ool00`00Oomoo`08Ool00`00Oomoo`0=Ool00`00Oomoo`05Ool00`00Oomoo`0=Ool0
0`00Oomoo`07Ool00`00Oomoo`0;Ool00`00Oomoo`07Ool00`00Oomoo`0=Ool00`00Oomoo`06Ool0
0`00Oomoo`0;Ool00`00Oomoo`08Ool00`00Oomoo`0<Ool00`00Oomoo`05Ool00`00Oomoo`0=Ool0
0`00Oomoo`08Ool00`00Oomoo`0:Ool00`00Oomoo`08Ool00`00Oomoo`0BOol001]oo`03001oogoo
00Eoo`03001oogoo00Ioo`03001oogoo00]oo`03001oogoo00Qoo`03001oogoo00eoo`03001oogoo
00Eoo`03001oogoo00eoo`03001oogoo00Moo`03001oogoo00]oo`03001oogoo00Moo`03001oogoo
00eoo`03001oogoo00Ioo`03001oogoo00]oo`03001oogoo00Qoo`03001oogoo00aoo`03001oogoo
00Eoo`03001oogoo00eoo`03001oogoo00Qoo`03001oogoo00Yoo`03001oogoo00Qoo`03001oogoo
019oo`006goo00<007ooOol01Goo00<007ooOol01Woo00<007ooOol02goo00<007ooOol02Goo00<0
07ooOol02goo00<007ooOol01Woo00<007ooOol037oo00<007ooOol027oo00<007ooOol02goo00<0
07ooOol01goo00<007ooOol03Goo00<007ooOol01Woo00<007ooOol02goo00<007ooOol02Goo00<0
07ooOol02goo00<007ooOol01Goo00<007ooOol03Goo00<007ooOol027oo00<007ooOol02Woo00<0
07ooOol027oo00<007ooOol04Woo000KOol00`00Oomoo`05Ool00`00Oomoo`06Ool00`00Oomoo`0;
Ool00`00Oomoo`09Ool00`00Oomoo`0;Ool00`00Oomoo`06Ool00`00Oomoo`0<Ool00`00Oomoo`08
Ool00`00Oomoo`0;Ool00`00Oomoo`07Ool00`00Oomoo`0=Ool00`00Oomoo`06Ool00`00Oomoo`0;
Ool00`00Oomoo`09Ool00`00Oomoo`0;Ool00`00Oomoo`06Ool00`00Oomoo`0<Ool00`00Oomoo`08
Ool00`00Oomoo`0:Ool00`00Oomoo`09Ool00`00Oomoo`0AOol001]oo`03001oogoo00Eoo`03001o
ogoo00Ioo`03001oogoo00]oo`03001oogoo00Uoo`03001oogoo00]oo`03001oogoo00Ioo`03001o
ogoo00aoo`03001oogoo00Qoo`03001oogoo00Yoo`03001oogoo00Qoo`03001oogoo00eoo`03001o
ogoo00Ioo`03001oogoo00]oo`03001oogoo00Uoo`03001oogoo00]oo`03001oogoo00Ioo`03001o
ogoo00aoo`03001oogoo00Qoo`03001oogoo00Yoo`03001oogoo00Uoo`03001oogoo015oo`006goo
0P001Woo00<007ooOol01Woo00<007ooOol02goo00<007ooOol02Goo00<007ooOol02goo00<007oo
Ool01Woo00<007ooOol037oo00<007ooOol027oo00<007ooOol02Woo00<007ooOol027oo00<007oo
Ool03Goo00<007ooOol01Woo00<007ooOol02goo00<007ooOol02Goo00<007ooOol02goo00<007oo
Ool01Woo00<007ooOol037oo00<007ooOol027oo00<007ooOol02Woo00<007ooOol02Goo00<007oo
Ool04Goo000KOol00`00Oomoo`05Ool00`00Oomoo`06Ool00`00Oomoo`0;Ool00`00Oomoo`09Ool0
0`00Oomoo`0;Ool00`00Oomoo`06Ool00`00Oomoo`0<Ool00`00Oomoo`08Ool00`00Oomoo`0:Ool0
0`00Oomoo`08Ool00`00Oomoo`0=Ool00`00Oomoo`06Ool00`00Oomoo`0;Ool00`00Oomoo`09Ool0
0`00Oomoo`0;Ool00`00Oomoo`06Ool00`00Oomoo`0<Ool00`00Oomoo`08Ool00`00Oomoo`0:Ool0
0`00Oomoo`09Ool00`00Oomoo`0AOol001]oo`03001oogoo00Eoo`03001oogoo00Ioo`03001oogoo
00]oo`03001oogoo00Uoo`03001oogoo00]oo`03001oogoo00Ioo`03001oogoo00aoo`03001oogoo
00Qoo`03001oogoo00Yoo`03001oogoo00Qoo`03001oogoo00eoo`03001oogoo00Ioo`03001oogoo
00]oo`03001oogoo00Uoo`03001oogoo00]oo`03001oogoo00Ioo`03001oogoo00aoo`03001oogoo
00Qoo`03001oogoo00Yoo`03001oogoo00Uoo`03001oogoo015oo`006goo00<007ooOol01Goo00<0
07ooOol01goo00<007ooOol02Woo00<007ooOol02Goo00<007ooOol02goo00<007ooOol01Woo00<0
07ooOol037oo00<007ooOol027oo00<007ooOol02Woo00<007ooOol027oo00<007ooOol037oo00<0
07ooOol01goo00<007ooOol02goo00<007ooOol02Goo00<007ooOol02goo00<007ooOol01Woo00<0
07ooOol037oo00<007ooOol027oo00<007ooOol02Woo00<007ooOol02Goo00<007ooOol04Goo000K
Ool00`00Oomoo`05Ool00`00Oomoo`07Ool00`00Oomoo`0:Ool00`00Oomoo`09Ool00`00Oomoo`0;
Ool00`00Oomoo`06Ool00`00Oomoo`0<Ool00`00Oomoo`08Ool00`00Oomoo`0:Ool00`00Oomoo`08
Ool00`00Oomoo`0<Ool00`00Oomoo`07Ool00`00Oomoo`0;Ool00`00Oomoo`09Ool00`00Oomoo`0:
Ool00`00Oomoo`07Ool00`00Oomoo`0<Ool00`00Oomoo`08Ool00`00Oomoo`0:Ool00`00Oomoo`09
Ool00`00Oomoo`0AOol001]oo`03001oogoo00Eoo`03001oogoo00Moo`03001oogoo00Yoo`03001o
ogoo00Uoo`03001oogoo00]oo`H0009oo`8000ioo`03001oogoo00Qoo`03001oogoo00Yoo`03001o
ogoo00Qoo`03001oogoo009oo`@000=oo`800003Ool007oo00Qoo`03001oogoo00]oo`03001oogoo
00Uoo`03001oogoo00Yoo`03001oogoo00=oo`80009oo`04001oo`00000;Ool00`00Oomoo`09Ool0
0`00Oomoo`09Ool00`00Oomoo`09Ool00`00Oomoo`06Ool30002Ool20004Ool001]oo`03001oogoo
00Eoo`03001oogoo00Moo`03001oogoo00Yoo`03001oogoo00Uoo`03001oogoo00]oo`04001oogoo
0003Ool01000Ool000003Goo00<007ooOol02Goo00<007ooOol02Goo00<007ooOol027oo00<007oo
Ool00Woo00<007ooOol00goo00<007ooOol00P002Goo00<007ooOol02Woo00<007ooOol02Woo00<0
07ooOol02Woo00<007ooOol00Woo00D007ooOol007oo0080009oo`03001oogoo00Qoo`03001oogoo
00Uoo`03001oogoo00Uoo`03001oogoo00Uoo`03001oogoo00Moo`04001oogoo0002Ool00`00Oomo
o`01Ool001]oo`03001oogoo00Eoo`03001oogoo00Moo`03001oogoo00Yoo`03001oogoo00Uoo`03
001oogoo00]oo`04001oogoo0003Ool01000Ool000003Goo00<007ooOol02Goo00<007ooOol02Goo
00<007ooOol027oo00<007ooOol00goo00<007ooOol00Woo00<007ooOol00P002Goo00<007ooOol0
2Woo00<007ooOol02Woo00<007ooOol02Woo00<007ooOol017oo00<007ooOol00P000Woo00<007oo
Ool027oo00<007ooOol02Goo00<007ooOol02Goo00<007ooOol02Goo00<007ooOol017oo1@0000Eo
o`00Oomoo`0000=oo`006goo0P001Woo00<007ooOol01goo00<007ooOol02Woo00<007ooOol02Goo
00<007ooOol02goo00@007ooOol000=oo`04001oo`00000=Ool00`00Oomoo`09Ool00`00Oomoo`09
Ool00`00Oomoo`08Ool00`00Oomoo`04Ool01@00Oomoogoo00000Woo0P002Goo00<007ooOol02Woo
00<007ooOol02Woo00<007ooOol02Woo00<007ooOol01Goo00@007oo0000009oo`03001oogoo00Qo
o`03001oogoo00Uoo`03001oogoo00Uoo`03001oogoo00Uoo`03001oogoo00Aoo`04001oogoo0002
Ool01000Oomoo`000goo000KOol00`00Oomoo`05Ool00`00Oomoo`07Ool00`00Oomoo`0:Ool00`00
Oomoo`09Ool00`00Oomoo`0;Ool01000Oomoo`000goo00@007oo000000eoo`03001oogoo00Uoo`03
001oogoo00Uoo`03001oogoo00Qoo`03001oogoo009oo`04001oogoo0002Ool00`00Oomoo`020009
Ool00`00Oomoo`0:Ool00`00Oomoo`0:Ool00`00Oomoo`0:Ool00`00Oomoo`02Ool01@00Oomoo`00
Ool00P000Woo00<007ooOol027oo00<007ooOol02Goo00<007ooOol02Goo00<007ooOol02Goo00<0
07ooOol01Goo00<007oo00000Woo00@007ooOol000=oo`006goo00<007ooOol017oo00<007ooOol0
27oo00<007ooOol02Woo00<007ooOol02Goo00<007ooOol02goo00@007oo000000Aoo`<000eoo`03
001oogoo00Uoo`03001oogoo00Uoo`03001oogoo00Uoo`03001oogoo009oo`8000Aoo`800003Ool0
07oo00Qoo`03001oogoo00Yoo`03001oogoo00Yoo`03001oogoo00Yoo`03001oogoo00=oo`80009o
o`04001oo`00000;Ool00`00Oomoo`09Ool00`00Oomoo`09Ool00`00Oomoo`09Ool00`00Oomoo`06
Ool20003Ool20004Ool001]oo`03001oogoo00Aoo`03001oogoo00Qoo`03001oogoo00Yoo`03001o
ogoo00Uoo`03001oogoo00Yoo`03001oogoo00Qoo`03001oogoo00]oo`03001oogoo00Uoo`03001o
ogoo00Uoo`03001oogoo00Uoo`03001oogoo00]oo`03001oogoo00Qoo`03001oogoo00Uoo`03001o
ogoo00]oo`03001oogoo00Uoo`03001oogoo00Moo`03001oogoo00aoo`03001oogoo00Uoo`03001o
ogoo00Uoo`03001oogoo00Uoo`03001oogoo015oo`006goo00<007ooOol017oo00<007ooOol027oo
00<007ooOol02Goo00<007ooOol02Woo00<007ooOol02Woo00<007ooOol027oo00<007ooOol02goo
00<007ooOol02Goo00<007ooOol027oo00<007ooOol02Woo00<007ooOol02goo00<007ooOol027oo
00<007ooOol02Goo00<007ooOol02goo00<007ooOol02Goo00<007ooOol01goo00<007ooOol037oo
00<007ooOol02Goo00<007ooOol027oo00<007ooOol02Woo00<007ooOol04Goo000KOol00`00Oomo
o`04Ool00`00Oomoo`08Ool00`00Oomoo`09Ool00`00Oomoo`0:Ool00`00Oomoo`0:Ool00`00Oomo
o`08Ool00`00Oomoo`0;Ool00`00Oomoo`09Ool00`00Oomoo`08Ool00`00Oomoo`0:Ool00`00Oomo
o`0;Ool00`00Oomoo`08Ool00`00Oomoo`09Ool00`00Oomoo`0;Ool00`00Oomoo`09Ool00`00Oomo
o`07Ool00`00Oomoo`0<Ool00`00Oomoo`09Ool00`00Oomoo`08Ool00`00Oomoo`0:Ool00`00Oomo
o`0AOol001]oo`03001oogoo00Aoo`03001oogoo00Qoo`03001oogoo00Uoo`03001oogoo00Yoo`03
001oogoo00Yoo`03001oogoo00Qoo`03001oogoo00]oo`03001oogoo00Uoo`03001oogoo00Qoo`03
001oogoo00Yoo`03001oogoo00]oo`03001oogoo00Qoo`03001oogoo00Uoo`03001oogoo00]oo`03
001oogoo00Uoo`03001oogoo00Qoo`03001oogoo00Yoo`03001oogoo00Yoo`03001oogoo00Qoo`03
001oogoo00Yoo`03001oogoo015oo`006goo00<007ooOol017oo00<007ooOol027oo00<007ooOol0
2Goo00<007ooOol02goo00<007ooOol02Goo00<007ooOol027oo00<007ooOol02goo00<007ooOol0
2Goo00<007ooOol027oo00<007ooOol02Woo00<007ooOol02goo00<007ooOol027oo00<007ooOol0
2Goo00<007ooOol02goo00<007ooOol02Goo00<007ooOol027oo00<007ooOol02Woo00<007ooOol0
2Woo00<007ooOol027oo00<007ooOol02Woo00<007ooOol04Goo000DOooo000=000001]oo`03001o
ogoo00Aoo`03001oogoo009oo`03001oogoo00=oo`03001oogoo00Aoo`03001oogoo009oo`03001o
ogoo00Eoo`03001oogoo00=oo`03001oogoo00=oo`03001oogoo00=oo`03001oogoo00Aoo`05001o
ogooOol00008Ool00`00Oomoo`02Ool00`00Oomoo`04Ool00`00Oomoo`02Ool00`00Oomoo`05Ool0
1000Oomoo`002Goo00@007ooOol000Qoo`03001oogoo009oo`03001oogoo00Eoo`04001oogoo0009
Ool00`00Ool00009Ool01@00Oomoogoo000027oo00<007oo00002Woo0P002Woo00<007oo00002Woo
00<007oo00002Woo00<007ooOol02Goo0P002goo00<007ooOol01Goo000KOol00`00Oomoo`04Ool0
0`00Oomoo`09Ool00`00Oomoo`08Ool00`00Oomoo`0;Ool00`00Oomoo`09Ool00`00Oomoo`04Ool0
1@00Oomoogoo00003Goo00<007ooOol02Woo00<007ooOol01goo00<007ooOol02Woo00<007ooOol0
1Woo00<007ooOol00Woo00<007ooOol027oo00<007ooOol02Goo00<007ooOol02goo00<007ooOol0
27oo00<007ooOol027oo0P0037oo00<007ooOol02Woo00<007ooOol027oo00<007ooOol02Woo00<0
07ooOol02Goo00<007ooOol01Goo000KOol00`00Oomoo`04Ool00`00Oomoo`09Ool00`00Oomoo`08
Ool00`00Oomoo`0;Ool00`00Oomoo`09Ool00`00Oomoo`08Ool00`00Oomoo`0:Ool00`00Oomoo`0;
Ool00`00Oomoo`07Ool00`00Oomoo`0:Ool00`00Oomoo`0:Ool00`00Oomoo`09Ool00`00Oomoo`09
Ool00`00Oomoo`0;Ool00`00Oomoo`08Ool00`00Oomoo`09Ool00`00Oomoo`0:Ool00`00Oomoo`0:
Ool00`00Oomoo`08Ool00`00Oomoo`0;Ool00`00Oomoo`0@Ool001]oo`03001oogoo00Aoo`03001o
ogoo00Uoo`03001oogoo00Qoo`03001oogoo00]oo`03001oogoo00Uoo`03001oogoo00Qoo`03001o
ogoo00Yoo`03001oogoo00]oo`03001oogoo00Moo`03001oogoo00Yoo`03001oogoo00Yoo`03001o
ogoo00Uoo`03001oogoo00Qoo`03001oogoo00aoo`03001oogoo00Qoo`03001oogoo00Uoo`03001o
ogoo00Yoo`03001oogoo00Yoo`03001oogoo00Qoo`03001oogoo00]oo`03001oogoo011oo`006goo
00<007ooOol017oo00<007ooOol02Goo00<007ooOol027oo00<007ooOol02goo00<007ooOol027oo
00<007ooOol02Goo00<007ooOol02Woo00<007ooOol02goo00<007ooOol01goo00<007ooOol02Woo
00<007ooOol02Woo00<007ooOol02Goo00<007ooOol027oo00<007ooOol037oo00<007ooOol027oo
00<007ooOol02Goo00<007ooOol02Woo00<007ooOol02Woo00<007ooOol027oo00<007ooOol02goo
00<007ooOol047oo000KOol00`00Oomoo`04Ool00`00Oomoo`09Ool00`00Oomoo`08Ool00`00Oomo
o`0;Ool00`00Oomoo`08Ool00`00Oomoo`09Ool00`00Oomoo`0:Ool00`00Oomoo`0;Ool00`00Oomo
o`07Ool00`00Oomoo`0:Ool00`00Oomoo`0:Ool00`00Oomoo`09Ool00`00Oomoo`08Ool00`00Oomo
o`0<Ool00`00Oomoo`08Ool00`00Oomoo`09Ool00`00Oomoo`0:Ool00`00Oomoo`0:Ool00`00Oomo
o`08Ool00`00Oomoo`0;Ool00`00Oomoo`0@Ool001]oo`03001oogoo00Aoo`03001oogoo00Uoo`03
001oogoo00Qoo`03001oogoo00]oo`03001oogoo00Qoo`03001oogoo00Yoo`03001oogoo00Uoo`03
001oogoo00]oo`03001oogoo00Moo`03001oogoo00Yoo`03001oogoo00Yoo`03001oogoo00Uoo`03
001oogoo00Qoo`03001oogoo00aoo`03001oogoo00Qoo`03001oogoo00Uoo`03001oogoo00Yoo`03
001oogoo00Yoo`03001oogoo00Qoo`03001oogoo00]oo`03001oogoo011oo`006goo0P001Goo00<0
07ooOol02Goo00<007ooOol027oo00<007ooOol02goo00<007ooOol027oo00<007ooOol02Woo00<0
07ooOol02Goo00<007ooOol02goo00<007ooOol01goo00<007ooOol02Woo00<007ooOol02Woo00<0
07ooOol02Goo00<007ooOol027oo00<007ooOol037oo00<007ooOol027oo00<007ooOol02Goo00<0
07ooOol02Woo00<007ooOol02Woo00<007ooOol027oo00<007ooOol02goo00<007ooOol047oo000K
Ool00`00Oomoo`03Ool00`00Oomoo`0:Ool00`00Oomoo`08Ool00`00Oomoo`0;Ool00`00Oomoo`08
Ool00`00Oomoo`0:Ool00`00Oomoo`09Ool00`00Oomoo`0;Ool00`00Oomoo`07Ool00`00Oomoo`0;
Ool00`00Oomoo`09Ool00`00Oomoo`0:Ool00`00Oomoo`07Ool00`00Oomoo`0=Ool00`00Oomoo`07
Ool00`00Oomoo`09Ool00`00Oomoo`0:Ool00`00Oomoo`0;Ool00`00Oomoo`06Ool00`00Oomoo`0<
Ool00`00Oomoo`0@Ool001]oo`03001oogoo00=oo`03001oogoo00Yoo`03001oogoo00Qoo`03001o
ogoo00]oo`03001oogoo00Qoo`03001oogoo00Yoo`03001oogoo00Uoo`03001oogoo00]oo`03001o
ogoo00Moo`03001oogoo00]oo`03001oogoo00Uoo`03001oogoo00Yoo`03001oogoo00Moo`03001o
ogoo00eoo`03001oogoo00Moo`03001oogoo00Uoo`03001oogoo00Yoo`03001oogoo00]oo`03001o
ogoo00Ioo`03001oogoo00aoo`03001oogoo011oo`006goo00<007ooOol00goo00<007ooOol02Woo
00<007ooOol027oo00<007ooOol02goo00<007ooOol027oo00<007ooOol02Woo00<007ooOol02Goo
00<007ooOol02goo00<007ooOol01goo00<007ooOol02goo00<007ooOol02Goo00<007ooOol02Woo
00<007ooOol01goo00<007ooOol03Goo00<007ooOol01goo00<007ooOol02Goo00<007ooOol02Woo
00<007ooOol02goo00<007ooOol01Woo00<007ooOol037oo00<007ooOol047oo000KOol00`00Oomo
o`03Ool00`00Oomoo`0:Ool00`00Oomoo`07Ool00`00Oomoo`0<Ool00`00Oomoo`08Ool00`00Oomo
o`0:Ool00`00Oomoo`09Ool00`00Oomoo`0;Ool00`00Oomoo`07Ool00`00Oomoo`0;Ool00`00Oomo
o`09Ool00`00Oomoo`0:Ool00`00Oomoo`07Ool00`00Oomoo`0=Ool00`00Oomoo`07Ool00`00Oomo
o`09Ool00`00Oomoo`0:Ool00`00Oomoo`0;Ool00`00Oomoo`06Ool00`00Oomoo`0<Ool00`00Oomo
o`0@Ool001]oo`03001oogoo00=oo`03001oogoo00]oo`03001oogoo00Ioo`03001oogoo00aoo`03
001oogoo00Qoo`03001oogoo00Yoo`03001oogoo00Uoo`03001oogoo00]oo`03001oogoo00Moo`03
001oogoo00]oo`03001oogoo00Uoo`03001oogoo00Yoo`03001oogoo00Moo`03001oogoo00eoo`03
001oogoo00Moo`03001oogoo00Yoo`03001oogoo00Uoo`03001oogoo00]oo`03001oogoo00Ioo`03
001oogoo00aoo`03001oogoo011oo`006goo00<007ooOol00goo00<007ooOol02goo00<007ooOol0
1Woo00<007ooOol037oo00<007ooOol027oo00<007ooOol02Woo00<007ooOol02Goo00<007ooOol0
2goo00<007ooOol01Woo00<007ooOol037oo00<007ooOol02Goo00<007ooOol02Woo00<007ooOol0
1goo00<007ooOol03Goo00<007ooOol01goo00<007ooOol02Woo00<007ooOol02Goo00<007ooOol0
2goo00<007ooOol01Woo00<007ooOol037oo00<007ooOol047oo000KOol00`00Oomoo`03Ool00`00
Oomoo`0;Ool00`00Oomoo`06Ool00`00Oomoo`0<Ool00`00Oomoo`08Ool00`00Oomoo`0:Ool00`00
Oomoo`09Ool00`00Oomoo`0;Ool00`00Oomoo`06Ool00`00Oomoo`0<Ool00`00Oomoo`08Ool00`00
Oomoo`0;Ool00`00Oomoo`07Ool00`00Oomoo`0=Ool00`00Oomoo`07Ool00`00Oomoo`0:Ool00`00
Oomoo`09Ool00`00Oomoo`0;Ool00`00Oomoo`06Ool00`00Oomoo`0<Ool00`00Oomoo`0@Ool001]o
o`8000Aoo`03001oogoo00]oo`03001oogoo00Ioo`03001oogoo00aoo`03001oogoo00Qoo`03001o
ogoo00Yoo`03001oogoo00Uoo`03001oogoo00]oo`03001oogoo00Ioo`03001oogoo00aoo`03001o
ogoo00Qoo`03001oogoo00]oo`03001oogoo00Moo`03001oogoo00eoo`03001oogoo00Moo`03001o
ogoo00Yoo`03001oogoo00Uoo`03001oogoo00]oo`03001oogoo00Ioo`03001oogoo00aoo`03001o
ogoo011oo`006goo00<007ooOol00goo00<007ooOol02goo00<007ooOol01Woo00<007ooOol037oo
00<007ooOol027oo00<007ooOol02Woo00<007ooOol02Goo00<007ooOol02goo00<007ooOol01Woo
00<007ooOol037oo00<007ooOol027oo00<007ooOol02goo00<007ooOol01goo00<007ooOol03Goo
00<007ooOol01Woo00<007ooOol02goo00<007ooOol02Goo00<007ooOol02goo00<007ooOol01Woo
00<007ooOol037oo00<007ooOol047oo000KOol00`00Oomoo`03Ool00`00Oomoo`0;Ool00`00Oomo
o`06Ool00`00Oomoo`0<Ool00`00Oomoo`08Ool00`00Oomoo`0:Ool00`00Oomoo`09Ool00`00Oomo
o`0<Ool00`00Oomoo`05Ool00`00Oomoo`0<Ool00`00Oomoo`08Ool00`00Oomoo`0;Ool00`00Oomo
o`07Ool00`00Oomoo`0=Ool00`00Oomoo`06Ool00`00Oomoo`0;Ool00`00Oomoo`09Ool00`00Oomo
o`0;Ool00`00Oomoo`06Ool00`00Oomoo`0<Ool00`00Oomoo`0@Ool001]oo`03001oogoo00=oo`03
001oogoo00]oo`03001oogoo00Ioo`03001oogoo00aoo`03001oogoo00Moo`03001oogoo00]oo`03
001oogoo00Uoo`03001oogoo00aoo`03001oogoo00Eoo`03001oogoo00aoo`03001oogoo00Qoo`03
001oogoo00aoo`03001oogoo00Ioo`03001oogoo00eoo`03001oogoo00Ioo`03001oogoo00]oo`03
001oogoo00Uoo`03001oogoo00]oo`03001oogoo00Ioo`03001oogoo00aoo`03001oogoo011oo`00
6goo00<007ooOol00goo00<007ooOol02goo00<007ooOol01Woo00<007ooOol037oo00<007ooOol0
1goo00<007ooOol037oo00<007ooOol01goo00<007ooOol03Goo00<007ooOol01Goo00<007ooOol0
37oo00<007ooOol027oo00<007ooOol037oo00<007ooOol01Woo00<007ooOol03Goo00<007ooOol0
1Woo00<007ooOol02goo00<007ooOol02Goo00<007ooOol02goo00<007ooOol01Woo00<007ooOol0
37oo00<007ooOol027oo00<007ooOol01Goo000KOol00`00Oomoo`03Ool00`00Oomoo`0;Ool00`00
Oomoo`06Ool00`00Oomoo`0=Ool00`00Oomoo`06Ool00`00Oomoo`0<Ool00`00Oomoo`07Ool00`00
Oomoo`0=Ool00`00Oomoo`05Ool00`00Oomoo`0<Ool00`00Oomoo`08Ool00`00Oomoo`0<Ool00`00
Oomoo`06Ool00`00Oomoo`0=Ool00`00Oomoo`06Ool00`00Oomoo`0;Ool00`00Oomoo`09Ool00`00
Oomoo`0;Ool00`00Oomoo`06Ool00`00Oomoo`0<Ool00`00Oomoo`08Ool00`00Oomoo`05Ool001]o
o`03001oogoo00=oo`03001oogoo00]oo`03001oogoo00Eoo`03001oogoo00ioo`03001oogoo00Io
o`03001oogoo00aoo`03001oogoo00Moo`03001oogoo00eoo`03001oogoo00Eoo`03001oogoo00ao
o`03001oogoo00Qoo`03001oogoo00aoo`03001oogoo00Eoo`03001oogoo00ioo`03001oogoo00Io
o`03001oogoo00]oo`03001oogoo00Qoo`03001oogoo00aoo`03001oogoo00Eoo`03001oogoo00eo
o`03001oogoo00Qoo`03001oogoo00Eoo`006goo00<007ooOol00goo00<007ooOol02goo00<007oo
Ool01Goo00<007ooOol03Woo00<007ooOol01Woo00<007ooOol037oo00<007ooOol01goo00<007oo
Ool03Goo00<007ooOol01Goo00<007ooOol037oo00<007ooOol027oo00<007ooOol037oo00<007oo
Ool01Goo00<007ooOol03Woo00<007ooOol01Woo00<007ooOol02goo00<007ooOol027oo00<007oo
Ool037oo00<007ooOol01Goo00<007ooOol03Goo00<007ooOol027oo00<007ooOol01Goo000KOol2
0004Ool00`00Oomoo`0<Ool00`00Oomoo`04Ool00`00Oomoo`0>Ool00`00Oomoo`06Ool00`00Oomo
o`0<Ool00`00Oomoo`07Ool00`00Oomoo`0=Ool00`00Oomoo`05Ool00`00Oomoo`0=Ool00`00Oomo
o`07Ool00`00Oomoo`0<Ool00`00Oomoo`05Ool00`00Oomoo`0>Ool00`00Oomoo`06Ool00`00Oomo
o`0;Ool00`00Oomoo`08Ool00`00Oomoo`0<Ool00`00Oomoo`05Ool00`00Oomoo`0=Ool00`00Oomo
o`08Ool00`00Oomoo`05Ool001]oo`03001oogoo009oo`03001oogoo00eoo`03001oogoo00Aoo`03
001oogoo00ioo`03001oogoo00Ioo`03001oogoo00aoo`03001oogoo00Moo`03001oogoo00eoo`03
001oogoo00Eoo`03001oogoo00eoo`03001oogoo00Moo`03001oogoo00aoo`03001oogoo00Eoo`03
001oogoo00ioo`03001oogoo00Ioo`03001oogoo00]oo`03001oogoo00Qoo`03001oogoo00aoo`03
001oogoo00Eoo`03001oogoo00ioo`03001oogoo00Moo`03001oogoo00Eoo`006goo00<007ooOol0
0Woo00<007ooOol03Goo00<007ooOol017oo00<007ooOol03Woo00<007ooOol01Woo00<007ooOol0
37oo00<007ooOol01goo00<007ooOol03Goo00<007ooOol01Goo00<007ooOol03Goo00<007ooOol0
1goo00<007ooOol037oo00<007ooOol01Goo00<007ooOol03goo00<007ooOol01Goo00<007ooOol0
37oo00<007ooOol01goo00<007ooOol03Goo00<007ooOol017oo00<007ooOol03Woo00<007ooOol0
1goo00<007ooOol01Goo000KOol00`00Oomoo`02Ool00`00Oomoo`0=Ool00`00Oomoo`04Ool00`00
Oomoo`0>Ool00`00Oomoo`06Ool00`00Oomoo`0<Ool00`00Oomoo`07Ool00`00Oomoo`0=Ool00`00
Oomoo`05Ool00`00Oomoo`0=Ool00`00Oomoo`07Ool00`00Oomoo`0<Ool00`00Oomoo`05Ool00`00
Oomoo`0?Ool00`00Oomoo`05Ool00`00Oomoo`0<Ool00`00Oomoo`07Ool00`00Oomoo`0=Ool00`00
Oomoo`04Ool00`00Oomoo`0>Ool00`00Oomoo`07Ool00`00Oomoo`05Ool001]oo`03001oogoo009o
o`03001oogoo00eoo`03001oogoo00Aoo`03001oogoo00ioo`03001oogoo00Ioo`03001oogoo00ao
o`03001oogoo00Moo`03001oogoo00eoo`03001oogoo00Eoo`03001oogoo00eoo`03001oogoo00Io
o`03001oogoo00eoo`03001oogoo00Eoo`03001oogoo00moo`03001oogoo00Eoo`03001oogoo00ao
o`03001oogoo00Moo`03001oogoo00eoo`03001oogoo00Aoo`03001oogoo00ioo`03001oogoo00Mo
o`03001oogoo00Eoo`006goo00<007ooOol00Woo00<007ooOol03Goo00<007ooOol017oo00<007oo
Ool03Woo00<007ooOol01Woo00<007ooOol037oo00<007ooOol01goo00<007ooOol03Goo00<007oo
Ool01Goo00<007ooOol03Goo00<007ooOol01Woo00<007ooOol03Goo00<007ooOol01Goo00<007oo
Ool03goo00<007ooOol01Goo00<007ooOol037oo00<007ooOol01goo00<007ooOol03Goo00<007oo
Ool017oo00<007ooOol03Woo00<007ooOol01goo00<007ooOol01Goo000KOol00`00Oomoo`02Ool0
0`00Oomoo`0=Ool00`00Oomoo`04Ool00`00Oomoo`0>Ool00`00Oomoo`06Ool00`00Oomoo`0<Ool0
0`00Oomoo`07Ool00`00Oomoo`0=Ool00`00Oomoo`05Ool00`00Oomoo`0=Ool00`00Oomoo`06Ool0
0`00Oomoo`0=Ool00`00Oomoo`05Ool00`00Oomoo`0?Ool00`00Oomoo`05Ool00`00Oomoo`0<Ool0
0`00Oomoo`07Ool00`00Oomoo`0=Ool00`00Oomoo`04Ool00`00Oomoo`0>Ool00`00Oomoo`07Ool0
0`00Oomoo`05Ool001]oo`8000=oo`03001oogoo00eoo`03001oogoo00Aoo`03001oogoo00ioo`03
001oogoo00Ioo`03001oogoo00aoo`03001oogoo00Moo`03001oogoo00eoo`03001oogoo00Aoo`03
001oogoo00ioo`03001oogoo00Ioo`03001oogoo00eoo`03001oogoo00Eoo`03001oogoo00moo`03
001oogoo00Eoo`03001oogoo00aoo`03001oogoo00Moo`03001oogoo00eoo`03001oogoo00Aoo`03
001oogoo00ioo`03001oogoo00Ioo`03001oogoo00Ioo`006goo00<007ooOol00Woo00<007ooOol0
3Goo00<007ooOol017oo00<007ooOol03Woo00<007ooOol01Woo00<007ooOol037oo00<007ooOol0
1Woo00<007ooOol03goo00<007ooOol00goo00<007ooOol03Woo00<007ooOol01Woo00<007ooOol0
3Goo00<007ooOol01Goo00<007ooOol03goo00<007ooOol017oo00<007ooOol03Goo00<007ooOol0
1goo00<007ooOol03Goo00<007ooOol017oo00<007ooOol03Woo00<007ooOol01Woo00<007ooOol0
1Woo000KOol00`00Oomoo`02Ool00`00Oomoo`0=Ool00`00Oomoo`04Ool00`00Oomoo`0>Ool00`00
Oomoo`06Ool00`00Oomoo`0<Ool00`00Oomoo`06Ool00`00Oomoo`0?Ool00`00Oomoo`03Ool00`00
Oomoo`0>Ool00`00Oomoo`06Ool00`00Oomoo`0=Ool00`00Oomoo`05Ool00`00Oomoo`0?Ool00`00
Oomoo`04Ool00`00Oomoo`0=Ool00`00Oomoo`07Ool00`00Oomoo`0=Ool00`00Oomoo`03Ool00`00
Oomoo`0?Ool00`00Oomoo`06Ool00`00Oomoo`06Ool001]oo`03001oogoo009oo`03001oogoo00eo
o`03001oogoo00=oo`03001oogoo00moo`03001oogoo00Ioo`03001oogoo00aoo`03001oogoo00Io
o`03001oogoo00moo`03001oogoo00=oo`03001oogoo00ioo`03001oogoo00Ioo`03001oogoo00io
o`03001oogoo00Aoo`03001oogoo00moo`03001oogoo00Aoo`03001oogoo00eoo`03001oogoo00Mo
o`03001oogoo00eoo`03001oogoo00=oo`03001oogoo00moo`03001oogoo00Ioo`03001oogoo00Io
o`006goo00<007ooOol00Woo00<007ooOol03Goo00<007ooOol00goo00<007ooOol03goo00<007oo
Ool01Woo00<007ooOol037oo00<007ooOol01Woo00<007ooOol03goo00<007ooOol00goo00<007oo
Ool03Woo00<007ooOol01Woo00<007ooOol03Woo00<007ooOol017oo00<007ooOol03goo00<007oo
Ool017oo00<007ooOol03Goo00<007ooOol01goo00<007ooOol03Goo00<007ooOol00goo00<007oo
Ool03goo00<007ooOol01Woo00<007ooOol01Woo000KOol00`00Oomoo`02Ool00`00Oomoo`0=Ool0
0`00Oomoo`03Ool00`00Oomoo`0?Ool00`00Oomoo`06Ool00`00Oomoo`0<Ool00`00Oomoo`06Ool0
0`00Oomoo`0?Ool00`00Oomoo`03Ool00`00Oomoo`0>Ool00`00Oomoo`06Ool00`00Oomoo`0>Ool0
0`00Oomoo`04Ool00`00Oomoo`0?Ool00`00Oomoo`04Ool00`00Oomoo`0=Ool00`00Oomoo`07Ool0
0`00Oomoo`0>Ool00`00Oomoo`02Ool00`00Oomoo`0?Ool00`00Oomoo`06Ool00`00Oomoo`06Ool0
00Qoo`8000Aoo`8000=oo`D000=oo`03001oogoo009oo`03001oogoo00eoo`03001oogoo00=oo`03
001oogoo011oo`03001oogoo00Aoo`03001oogoo00ioo`03001oogoo00Eoo`03001oogoo00moo`03
001oogoo00=oo`03001oogoo00ioo`03001oogoo00Ioo`03001oogoo00ioo`03001oogoo00Aoo`03
001oogoo00moo`03001oogoo00Aoo`03001oogoo00ioo`03001oogoo00Ioo`03001oogoo00ioo`03
001oogoo009oo`03001oogoo00moo`03001oogoo00Ioo`03001oogoo00Ioo`001goo00@007ooOol0
00Yoo`03001oogoo00=oo`03001oogoo009oo`03001oogoo00eoo`03001oogoo00=oo`03001oogoo
011oo`03001oogoo00Aoo`03001oogoo00ioo`03001oogoo00Eoo`03001oogoo00moo`03001oogoo
00=oo`03001oogoo00ioo`03001oogoo00Ioo`03001oogoo00ioo`03001oogoo00=oo`03001oogoo
011oo`03001oogoo00Aoo`03001oogoo00ioo`03001oogoo00Ioo`03001oogoo00ioo`03001oogoo
009oo`03001oogoo00moo`03001oogoo00Ioo`03001oogoo00Ioo`001goo00@007ooOol000Yoo`03
001oogoo00=oo`8000=oo`03001oogoo00eoo`03001oogoo00=oo`03001oogoo011oo`03001oogoo
00Aoo`03001oogoo00ioo`03001oogoo00Eoo`03001oogoo011oo`03001oogoo009oo`03001oogoo
00ioo`03001oogoo00Ioo`03001oogoo00ioo`03001oogoo00=oo`03001oogoo011oo`03001oogoo
00Aoo`03001oogoo00ioo`03001oogoo00Eoo`03001oogoo00moo`03001oogoo009oo`03001oogoo
00moo`03001oogoo00Ioo`03001oogoo00Ioo`001goo00@007ooOol000Yoo`03001oogoo00=oo`03
001oogoo009oo`03001oogoo00eoo`03001oogoo00=oo`03001oogoo011oo`03001oogoo00Aoo`03
001oogoo00ioo`03001oogoo00Eoo`03001oogoo011oo`03001oogoo009oo`03001oogoo00ioo`03
001oogoo00Ioo`03001oogoo00ioo`03001oogoo00=oo`03001oogoo015oo`03001oogoo00=oo`03
001oogoo00ioo`03001oogoo00Eoo`03001oogoo00moo`03001oogoo009oo`03001oogoo00moo`03
001oogoo00Ioo`03001oogoo00Ioo`001goo00@007ooOol000Yoo`03001oogoo00=oo`03001oogoo
009oo`03001oogoo00eoo`03001oogoo00=oo`03001oogoo011oo`03001oogoo00Aoo`03001oogoo
00ioo`03001oogoo00Eoo`03001oogoo011oo`05001oogooOol0000AOol00`00Oomoo`06Ool00`00
Oomoo`0>Ool00`00Oomoo`03Ool00`00Oomoo`0AOol00`00Oomoo`02Ool00`00Oomoo`0?Ool00`00
Oomoo`05Ool00`00Oomoo`0?Ool00`00Oomoo`02Ool00`00Oomoo`0?Ool00`00Oomoo`05Ool00`00
Oomoo`07Ool000Qoo`8000Yoo`8000Eoo`05001oogooOol0000AOol00`00Oomoo`02Ool00`00Oomo
o`0@Ool00`00Oomoo`04Ool00`00Oomoo`0>Ool00`00Oomoo`05Ool00`00Oomoo`0@Ool01@00Oomo
ogoo00004Woo00<007ooOol017oo00<007ooOol03goo00<007ooOol00goo00<007ooOol04Goo00<0
07ooOol00Woo00<007ooOol03goo00<007ooOol01Goo00<007ooOol03goo00<007ooOol00Woo00<0
07ooOol047oo00<007ooOol017oo00<007ooOol01goo000KOol01@00Oomoogoo00004Goo00<007oo
Ool00Woo00<007ooOol047oo00<007ooOol017oo00<007ooOol03Woo00<007ooOol01Goo00<007oo
Ool047oo00D007ooOomoo`00019oo`03001oogoo00Aoo`03001oogoo00moo`03001oogoo00=oo`03
001oogoo015oo`03001oogoo009oo`03001oogoo00moo`03001oogoo00Eoo`03001oogoo011oo`05
001oogooOol0000BOol00`00Oomoo`04Ool00`00Oomoo`07Ool001]oo`05001oogooOol0000AOol0
0`00Oomoo`02Ool00`00Oomoo`0@Ool00`00Oomoo`04Ool00`00Oomoo`0>Ool00`00Oomoo`05Ool0
0`00Oomoo`0@Ool01@00Oomoogoo00004Woo00<007ooOol017oo00<007ooOol03goo00<007ooOol0
0goo00<007ooOol04Goo00<007ooOol00Woo00<007ooOol03goo00<007ooOol01Goo00<007ooOol0
47oo00D007ooOomoo`00019oo`03001oogoo00Aoo`03001oogoo00Moo`006goo00D007ooOomoo`00
015oo`05001oogooOol0000COol00`00Oomoo`04Ool00`00Oomoo`0>Ool00`00Oomoo`05Ool00`00
Oomoo`0@Ool01@00Oomoogoo00004Woo00<007ooOol017oo00<007ooOol03goo00<007ooOol00goo
00<007ooOol04Goo00<007ooOol00Woo00<007ooOol03goo00<007ooOol01Goo00<007ooOol047oo
00D007ooOomoo`00019oo`03001oogoo00Aoo`03001oogoo00Moo`006goo00D007ooOomoo`00015o
o`05001oogooOol0000DOol00`00Oomoo`03Ool00`00Oomoo`0>Ool00`00Oomoo`05Ool00`00Oomo
o`0@Ool01@00Oomoogoo00004Woo00<007ooOol017oo00<007ooOol047oo00<007ooOol00Woo00<0
07ooOol04Goo00<007ooOol00Woo00<007ooOol03goo00<007ooOol01Goo00<007ooOol047oo00D0
07ooOomoo`00019oo`03001oogoo00Aoo`03001oogoo00Moo`006goo0P000Woo00<007ooOol03goo
00D007ooOomoo`0001Aoo`03001oogoo00=oo`03001oogoo00ioo`03001oogoo00Aoo`03001oogoo
015oo`04001oogoo000COol00`00Oomoo`04Ool00`00Oomoo`0@Ool01@00Oomoogoo000057oo00<0
07ooOol00Woo00<007ooOol03goo00<007ooOol01Goo00<007ooOol047oo00@007ooOol001=oo`03
001oogoo00Aoo`03001oogoo00Moo`006goo00D007ooOomoo`00019oo`04001oogoo000DOol00`00
Oomoo`03Ool00`00Oomoo`0>Ool00`00Oomoo`04Ool00`00Oomoo`0AOol01000Oomoo`004goo00<0
07ooOol017oo00<007ooOol047oo00D007ooOomoo`0001Aoo`03001oogoo009oo`03001oogoo00mo
o`03001oogoo00Eoo`03001oogoo011oo`04001oogoo000COol00`00Oomoo`04Ool00`00Oomoo`07
Ool001]oo`05001oogooOol0000BOol01000Oomoo`0057oo00<007ooOol00goo00<007ooOol03Woo
00<007ooOol017oo00<007ooOol04Woo00<007oo00004goo00<007ooOol017oo00<007ooOol047oo
00D007ooOomoo`0001Aoo`03001oogoo009oo`03001oogoo00moo`03001oogoo00Aoo`03001oogoo
015oo`04001oogoo000COol00`00Oomoo`04Ool00`00Oomoo`07Ool001]oo`05001oogooOol0000B
Ool01000Oomoo`0057oo00<007ooOol00Woo00<007ooOol03goo00<007ooOol017oo00<007ooOol0
4Woo00<007oo00004goo00<007ooOol017oo00<007ooOol047oo00D007ooOomoo`0001Aoo`03001o
ogoo009oo`03001oogoo00moo`03001oogoo00Aoo`03001oogoo015oo`04001oogoo000COol00`00
Oomoo`04Ool00`00Oomoo`07Ool001]oo`05001oogooOol0000BOol01000Oomoo`0057oo00<007oo
Ool00Woo00<007ooOol047oo00<007ooOol00goo00<007ooOol04Woo00<007oo00004goo00<007oo
Ool017oo00<007ooOol047oo00D007ooOomoo`0001Aoo`03001oogoo009oo`03001oogoo00moo`03
001oogoo00Aoo`03001oogoo019oo`03001oo`0001Aoo`03001oogoo00=oo`03001oogoo00Moo`00
6goo00D007ooOomoo`00019oo`04001oogoo000DOol00`00Oomoo`02Ool00`00Oomoo`0@Ool00`00
Oomoo`03Ool00`00Oomoo`0BOol00`00Ool0000COol00`00Oomoo`04Ool00`00Oomoo`0@Ool01@00
Oomoogoo000057oo00<007ooOol00Woo00<007ooOol047oo00<007ooOol00goo00<007ooOol04Woo
0P005Goo00<007ooOol00goo00<007ooOol01goo000KOol01@00Oomoogoo00004Woo00<007oo0000
5Goo00<007ooOol00Woo00<007ooOol047oo00<007ooOol00goo00<007ooOol04Woo0P0057oo00<0
07ooOol017oo00<007ooOol047oo00D007ooOomoo`0001Eoo`05001oogooOol0000BOol00`00Oomo
o`03Ool00`00Oomoo`0BOol2000EOol00`00Oomoo`03Ool00`00Oomoo`07Ool001]oo`05001oogoo
Ool0000BOol00`00Ool0000EOol00`00Oomoo`02Ool00`00Oomoo`0@Ool00`00Oomoo`03Ool00`00
Oomoo`0XOol00`00Oomoo`04Ool00`00Oomoo`0@Ool01@00Oomoogoo00005Goo00@007ooOol001=o
o`03001oogoo00=oo`03001oogoo02Uoo`03001oogoo009oo`03001oogoo00Qoo`006goo0P0000=o
o`00Ool04goo0P005Goo00<007ooOol00Woo00<007ooOol047oo00<007ooOol00goo00<007ooOol0
:7oo00<007ooOol017oo00<007ooOol047oo00D007ooOomoo`0001Eoo`04001oogoo000COol00`00
Oomoo`03Ool00`00Oomoo`0YOol00`00Oomoo`02Ool00`00Oomoo`08Ool001]oo`04001oogoo000D
Ool2000EOol00`00Oomoo`02Ool00`00Oomoo`0@Ool00`00Oomoo`03Ool00`00Oomoo`0XOol00`00
Oomoo`03Ool00`00Oomoo`0BOol01000Oomoo`005Goo00@007ooOol001=oo`03001oogoo00=oo`03
001oogoo02Uoo`03001oogoo009oo`03001oogoo00Qoo`006goo00@007ooOol002]oo`03001oogoo
009oo`03001oogoo011oo`03001oogoo00=oo`03001oogoo02Uoo`03001oogoo009oo`03001oogoo
019oo`03001oo`0001Ioo`04001oogoo000COol00`00Oomoo`03Ool00`00Oomoo`0YOol00`00Oomo
o`02Ool00`00Oomoo`08Ool001]oo`04001oogoo000/Ool01@00Oomoogoo00004Woo00<007ooOol0
0goo00<007ooOol0:Goo00<007ooOol00Woo00<007ooOol04Woo00<007oo00005Woo00@007oo0000
01=oo`03001oogoo00=oo`03001oogoo02Uoo`03001oogoo009oo`03001oogoo00Qoo`006goo00@0
07ooOol002aoo`05001oogooOol0000COol00`00Oomoo`02Ool00`00Oomoo`0YOol00`00Oomoo`02
Ool00`00Oomoo`0BOol00`00Ool0000GOol2000DOol00`00Oomoo`03Ool00`00Oomoo`0YOol00`00
Oomoo`02Ool00`00Oomoo`08Ool001]oo`04001oogoo000/Ool01@00Oomoogoo00004goo00D007oo
Oomoo`0002aoo`03001oogoo009oo`03001oogoo01=oo`8001Moo`8001Aoo`03001oogoo00=oo`03
001oogoo02Uoo`03001oogoo009oo`03001oogoo00Qoo`006goo00@007ooOol002aoo`04001oogoo
000DOol01@00Oomoogoo0000;7oo00<007ooOol00Woo00<007ooOol04goo00<007ooOol0;7oo00<0
07ooOol00goo00<007ooOol0:Goo00<007ooOol00Woo00<007ooOol027oo000KOol01000Oomoo`00
;7oo00@007ooOol001Aoo`05001oogooOol0000/Ool00`00Oomoo`02Ool00`00Oomoo`12Ool00`00
Oomoo`03Ool00`00Oomoo`0YOol00`00Oomoo`02Ool00`00Oomoo`08Ool001]oo`<002eoo`04001o
ogoo000DOol01@00Oomoogoo0000;7oo00<007ooOol00Woo00<007ooOol0@Woo00<007ooOol00goo
00<007ooOol0:Goo00<007ooOol00Woo00<007ooOol027oo000KOol00`00Ool0000]Ool01000Oomo
o`0057oo00D007ooOomoo`0002aoo`03001oogoo009oo`03001oogoo049oo`03001oogoo00=oo`03
001oogoo02Yoo`04001oogoo000;Ool001]oo`03001oo`0002ioo`03001oo`0001Aoo`05001oogoo
Ool0000]Ool01@00Oomoogoo0000A7oo00<007ooOol00Woo00<007ooOol0:goo00@007ooOol000]o
o`006goo00<007oo0000;Woo0P005Goo00@007ooOol002ioo`05001oogooOol00014Ool00`00Oomo
o`02Ool00`00Oomoo`0[Ool01000Oomoo`002goo000KOol00`00Ool0000^Ool2000EOol01000Oomo
o`00;Woo00D007ooOomoo`0004Aoo`03001oogoo009oo`03001oogoo02]oo`04001oogoo000;Ool0
01]oo`03001oo`0004Eoo`04001oogoo000^Ool01@00Oomoogoo0000A7oo00<007ooOol00Woo00<0
07ooOol0:goo00@007ooOol000]oo`006goo00<007oo0000AWoo00<007oo0000;Woo00D007ooOomo
o`0004Eoo`05001oogooOol0000]Ool01000Oomoo`002goo000KOol00`00Ool00016Ool00`00Ool0
000^Ool01000Oomoo`00AWoo00D007ooOomoo`0002ioo`03001oo`0000]oo`006goo0`00AWoo00<0
07oo0000;Woo00@007ooOol004Ioo`04001oogoo000_Ool3000;Ool001]oo`03001oo`0004Ioo`03
001oo`0002ioo`04001oogoo0016Ool01000Oomoo`00;goo0P0037oo000KOol00`00Ool00017Ool0
0`00Oomoo`0]Ool01000Oomoo`00AWoo00@007ooOol002moo`8000aoo`006goo0P00N7oo00@007oo
Ool004Ioo`04001oogoo000mOol001]oo`8007Uoo`03001oo`0004Moo`03001oo`0003eoo`0027oo
0P0017oo0P000goo100017oo0P00NGoo0`00Agoo0`00?Goo0007Ool01000Oomoo`0027oo00<007oo
Ool01Goo0P00NWoo00<007ooOol0AWoo0P00?Woo0007Ool01000Oomoo`002Goo00<007ooOol017oo
0P00ogoo17oo0007Ool01000Oomoo`002Woo00<007ooOol00goo00<007ooOol0ogoo0goo0007Ool0
1000Oomoo`0027oo00@007ooOol000Aoo`03001oogoo0?moo`=oo`0027oo0P002Woo0P001Goo00<0
07ooOol0ogoo0goo000KOol00`00Oomoo`3oOol3Ool00?moob5oo`00ogoo8Goo003oOolQOol00?mo
ob5oo`00ogoo8Goo003oOolQOol00001\
\>"],
  ImageRangeCache->{{{81.0625, 368.063}, {262.75, 85.8125}} -> {-17.1137, \
-0.00678216, 0.157891, 0.0025538}},
  CellTags->"b:14.1"],

Cell[BoxData[
    TagBox[\(\[SkeletonIndicator]  Graphics  \[SkeletonIndicator]\),
      False,
      Editable->False]], "Output",
  CellLabel->"Out[52]=",
  CellTags->"b:14.1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Creating MEX-Files", "Subsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],

Cell["\<\
It may be desireable to use the system model in an external program. TSi \
Dynamics provides functions to create C-Code which compiles as a MEX-File for \
use as an S-function in SIMULINK. An S-function may have inputs, outputs so \
that it can be interconnected with other subsystems within SIMULINK and it \
can have parameters which can be defined from within SIMULINK. In the \
following example, we define the joint torques as inputs and the (y,z) \
position of mass 2 as the outputs. Parameters include the system two masses, \
the two lengths and the gravitational constant.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],

Cell["\<\
We can use the TSi Dynamics function EndEffector to define the functions.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],

Cell[CellGroupData[{

Cell["\<\
ChainLst={{1,1},{2,2}};
TerminalNode=3;
Xout=EndEffector[ChainLst,TerminalNode,BodyLst,X];
yout={Xout[[2,4]],Xout[[3,4]]}\
\>", "Input",
  CellLabel->"In[53]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],

Cell[BoxData[
    \({l1\ Sin[a1x] - 
        l2\ \((\(-Cos[a2x]\)\ Sin[a1x] - Cos[a1x]\ Sin[a2x])\), \(-l1\)\ Cos[
            a1x] - l2\ \((Cos[a1x]\ Cos[a2x] - 
              Sin[a1x]\ Sin[a2x])\)}\)], "Output",
  CellLabel->"Out[56]=",
  CellTags->"b:14.2"]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
  "Inputs = Q;\nOutputs = yout;\nMEXFilename = \"",
  "C:\\Projects\\ProPac\\ProPac_3\\Notebooks",
  "\\dbl_pend.c\";\nparamvec={m1,l1,m2,l2,gc};\nPassedParams = \
{\"X0\",\"m1\",\"l1\",\"m2\",\"l2\",\"gc\"};\nPassedParamsDimensions = \
{{2*Length[p],1},{1,1},{1,1},{1,1},{1,1},{1,1}};\n",
  StyleBox["CreateModelMEX",
    PageWidth->Infinity,
    AspectRatioFixed->True,
    FontFamily->"Courier New"],
  "[p,q,Inputs,Outputs,PassedParams,PassedParamsDimensions,\n               \
V,Cp,Fp,M,MEXFilename,MinimumBreakVariables\[Rule]20];"
}], "Input",
  CellLabel->"In[57]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],

Cell[BoxData[
    \("  *** MEXTools successfully loaded ***"\)], "Print",
  CellLabel->"From In[57]:=",
  CellTags->"b:14.2"],

Cell[BoxData[
    \("...Generating Header Code"\)], "Print",
  CellLabel->"From In[57]:=",
  CellTags->"b:14.2"],

Cell[BoxData[
    \("...Generating Initial Condition Function Code"\)], "Print",
  CellLabel->"From In[57]:=",
  CellTags->"b:14.2"],

Cell[BoxData[
    \("...Generating State Derivative Function Code"\)], "Print",
  CellLabel->"From In[57]:=",
  CellTags->"b:14.2"],

Cell[BoxData[
    \("...Collecting all function terms"\)], "Print",
  CellLabel->"From In[57]:=",
  CellTags->"b:14.2"],

Cell[BoxData[
    \("...Generating temp variable declarations"\)], "Print",
  CellLabel->"From In[57]:=",
  CellTags->"b:14.2"],

Cell[BoxData[
    \("...Converting Expressions to C form"\)], "Print",
  CellLabel->"From In[57]:=",
  CellTags->"b:14.2"],

Cell[BoxData[
    \("...Generating Output Function Code"\)], "Print",
  CellLabel->"From In[57]:=",
  CellTags->"b:14.2"],

Cell[BoxData[
    \("MEX File created with name: \
C:\\Projects\\ProPac\\ProPac_3\\Notebooks\\dbl_pend.c"\)], "Print",
  CellLabel->"From In[57]:=",
  CellTags->"b:14.2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Compiling & Using MEX Files", "Subsection",
  CellTags->{"Creating MEX-Files", "b:14.3"}],

Cell["\<\
Detailed information on compiling and using MEX files with MATLAB/SIMULINK \
can be found in the appropriate MATLAB references. CreateModelMEX (and also \
CreateControllerMEX) assemble C source code that needs to be compiled. During \
the compilation process the compiled code will be linked with additional code \
segments and libraries provided with ProPac or MATLAB or your compiler. It is \
necessary that this code be available at the time of compilation. The easiest \
way to proceed is to perform the compilation from within MATLAB using scripts \
provided with MATLAB (either cmex with MATLAB 4 or mex with MATLAB 5). This \
should automatically define the locations of all required MATLAB or compiler \
code segments. In addition to the file created by ProPac (for example: \
dbl_pend.c) you will need the files linsolv.c, f2c.h and trigfun.h - all \
provided with ProPac. 
\tOne way to compile dbl_pend.c  is to place the files dbl_pend.c, linsolv.c, \
f2c.h and trigfun.h in a common directory, for example C:\\DoublePend. Then \
in the MATLAB command window set the current directory to C:\\DoublePend to \
insure that the four required files are in the search path:
\tcd C:\\DoublePend
then use the command
\tmex dbl_pend.c linsolv.c
in MATLAB 5, or
\t!cmex dbl_pend.c linsolv.c
in MATLAB 4.2.
\tIf you have followed the MATLAB installation instructions as appropriate \
for your compiler this should work fine. Otherwise, you can reinstall MATLAB \
5 and follow the instructions (which setsup the appropriate paths in mex), or \
in MATLAB 4.2 you can edit cmex as described therein.\
\>", "Text",
  TextJustification->1,
  CellTags->{"Creating MEX-Files", "b:14.3"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Building More Complex Models", "Section",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell["Introductory Remarks", "Subsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->"b:15.0"],

Cell["\<\
TSi Dynamics provides several routes to developing models. The most direct is \
to use the functions CreateModel or CreateModelSpecial. These two functions \
work essentially the same way but produce models in slightly different form. \
The former is almost always preferable for creating simulations. One example \
of its use has already been given above for creating a model of the double \
pendulum. The following more elaborate examples illustrate some other \
functions in TSi Dynamics.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->"b:15.0"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Partial Suspension System", "Subsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell["\<\
We consider a simple split axle suspension in planar motion. The system \
consists of a chassis, two axles, two tires, two springs - each connecting \
nodes on one axle and the chassis, and two torsional dampers. The modeling \
process begins in the standard way with joint and body definitions. The \
coordinate system is chosen with z down and motion takes place in the y-z \
plane (y points to the right).
Joint Data\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell[CellGroupData[{

Cell["\<\
(* Joint 1 - planar motion *)
r1={3};
q1={theta1,y,z};p1={w1,vy,vz};
H1={{1,0,0},{0,0,0},{0,0,0},{0,0,0},{0,1,0},{0,0,1}};
(* Joint 2 - suspension *)
r2={1};
q2={theta2};p2={w2};
H2={{1},{0},{0},{0},{0},{0}};
(* Joint 3 - suspension *)
r3={1};
H3={{1},{0},{0},{0},{0},{0}};
q3={theta3};p3={w3};

JointLst={{r1,H1,q1,p1},{r2,H2,q2,p2},{r3,H3,q3,p3}};
{V,X,H}=Joints[JointLst];\
\>", "Input",
  CellLabel->"In[64]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell[BoxData[
    \("Computing joint 1 kinematics"\)], "Print",
  CellLabel->"From In[64]:=",
  CellTags->"b:15.1"],

Cell[BoxData[
    \("Computing joint 2 kinematics"\)], "Print",
  CellLabel->"From In[64]:=",
  CellTags->"b:15.1"],

Cell[BoxData[
    \("Computing joint 3 kinematics"\)], "Print",
  CellLabel->"From In[64]:=",
  CellTags->"b:15.1"]
}, Open  ]],

Cell["Body Data", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell["\<\
(* Body 1 - Chassis *)
cm1={0,0,0};
out1={{2,{0,b/2,a/2}},{3,{0,-b/2,a/2}},{6,{0,b/2+a/2,-a/2}},{7,{0,-b/2-a/2,-a/\
2}}};
I1=DiagonalMatrix[{Ixx,Iyy,Izz}];

(* Body 2 - Wheel & Axle *)
cm2={0,d,0};
out2={{4,{0,d,R}},{8,{0,Sqrt[2]*a,0}}};
I2=DiagonalMatrix[{Wxx,Wyy,Wzz}];

(* Body 3 - Wheel & Axle *)
cm3={0,-d,0};
out3={{5,{0,-d,R}},{9,{0,-Sqrt[2]*a,0}}};
I3=DiagonalMatrix[{Wxx,Wyy,Wzz}];

BodyLst={{cm1,out1,m1,I1},{cm2,out2,mw,I2},{cm3,out3,mw,I3}};
\
\>", "Input",
  CellLabel->"In[78]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell["\<\
System Structure
Notice that this system has a tree structure with two chains.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell["TreeLst={{{1,1},{2,2}},{{1,1},{3,3}}};", "Input",
  CellLabel->"In[91]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell["\<\
Potential Energy Constructions
The potential energy comes from two sources in addition to the gravitational \
potential: (i) the tires, and (ii) the springs. The tires are leaves of the \
tree and they react with the space frame. Computing the potential energy of \
the tires illustrates the use of the function LeafPotential.
It is assumed that each tire potential energy function is defined in terms of \
the space coordinates of the tire contact point (the tire outboard node). In \
this example, we suppose that only the vertical displacements are \
significant.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell["\<\
(* Define Tire PE in terms of Space Coordinates *)

yy={{x2,y2,z2},{x3,y3,z3}};
Pot=(k/2)*(z2-z20)^2+(k/2)*(z3-z30)^2;\
\>", "Input",
  CellLabel->"In[92]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell["\<\
Note that z20 and z30 may be specified as externally defined functions of \
time to simulate riding over rough terrain. LeafPotential replaces the \
absolute coordinates by the generalized coordinates.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell["\<\
(* Obtain Tire PE in terms of generalized Coordinates *)

TirePE=LeafPotential[BodyLst,TreeLst,X,H,Pot,yy];\
\>", "Input",
  CellLabel->"In[95]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell[CellGroupData[{

Cell["TirePE", "Input",
  CellLabel->"In[97]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell[BoxData[
    \(1\/8\ k\ \((\((2\ z - 2\ z20 + a\ Cos[theta1] + 2\ R\ Cos[theta1 + \
theta2] + b\ Sin[theta1] + 2\ d\ Sin[theta1 + theta2])\)\^2 + \((2\ z - 2\ \
z30 + a\ Cos[theta1] + 2\ R\ Cos[theta1 + theta3] - b\ Sin[theta1] - 2\ d\ \
Sin[theta1 + theta3])\)\^2)\)\)], "Output",
  CellLabel->"Out[97]=",
  CellTags->"b:15.1"]
}, Open  ]],

Cell["\<\
The spring potential energy function SpringPotential is a specialized \
function for springs connecting two nodes in the system. It is assumed that \
the potential energy is a function of the length of the spring. Because only \
one variable is involved it is convenient to specify the potential energy as \
a function rather than as an expression in which the variables must be \
separately identified. SpringPotential constructs the potential energy \
function for the spring in terms of the generalized system coordinates.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell["\<\
SpringPot[x_]:=(kdel/2)*(x-ls)^2
q=Join[q1,q2,q3];
SpringPE1=SpringPotential[6,8,SpringPot,TreeLst,BodyLst,X,q];
SpringPE2=SpringPotential[7,9,SpringPot,TreeLst,BodyLst,X,q];

PE=TirePE+SpringPE1+SpringPE2;
\
\>", "Input",
  CellLabel->"In[98]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell[CellGroupData[{

Cell["SpringPE1", "Input",
  CellLabel->"In[103]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell[BoxData[
    \(1\/2\ kdel\ \((\(-ls\) + 1\/2\ \@\(13\ a\^2 - 4\ \@2\ a\^2\ Cos[theta2] \
+ 8\ \@2\ a\^2\ Sin[theta2]\))\)\^2\)], "Output",
  CellLabel->"Out[103]=",
  CellTags->"b:15.1"]
}, Open  ]],

Cell[CellGroupData[{

Cell["SpringPE2", "Input",
  CellLabel->"In[104]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell[BoxData[
    \(1\/2\ kdel\ \((\(-ls\) + 1\/2\ \@\(13\ a\^2 - 4\ \@2\ a\^2\ Cos[theta3] \
- 8\ \@2\ a\^2\ Sin[theta3]\))\)\^2\)], "Output",
  CellLabel->"Out[104]=",
  CellTags->"b:15.1"]
}, Open  ]],

Cell["\<\
Generalized Force
The only generalized forces acting are due to the dampers. We assume that the \
forces due to the torsional dampers are proportional to the relative angular \
velocities of the axles with respect to the chassis. As an alternative one \
could place a linear damper between two nodes and use the function \
DamperForce to compute the generalized force.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell["Q={0,0,0,-c*w2,-c*w3};", "Input",
  CellLabel->"In[105]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell["\<\
Gravity
Since, we have chosen the z-axis to point down, we change the sign of the \
gravitational acceleration\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell["g=-gc;", "Input",
  CellLabel->"In[106]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell["\<\
Obtain Equations of Motion
Notice that since the joint kinematic data V, X, and H has already been \
computed, we include this data in the argument list of CreateModel to avoid \
recomputing it.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell[CellGroupData[{

Cell["\<\
{V,X,H,M,Cm,F,pp,qq}=CreateModel[JointLst,BodyLst,TreeLst,g,PE,Q,V,X,H];\
\>", "Input",
  CellLabel->"In[107]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell[BoxData[
    \("Computing Potential Functions"\)], "Print",
  CellLabel->"From In[107]:=",
  CellTags->"b:15.1"],

Cell[BoxData[
    \("Computing Inertia Matrix"\)], "Print",
  CellLabel->"From In[107]:=",
  CellTags->"b:15.1"],

Cell[BoxData[
    \("Computing Poincare Function"\)], "Print",
  CellLabel->"From In[107]:=",
  CellTags->"b:15.1"]
}, Open  ]],

Cell[CellGroupData[{

Cell["M", "Input",
  CellLabel->"In[108]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell[BoxData[
    \({{Ixx + 2\ d\^2\ mw + 2\ Wxx + 
          d\ mw\ \((1\/2\ b\ Cos[theta2] + 1\/2\ a\ Sin[theta2])\) + 
          mw\ \((\(-\(1\/2\)\)\ a\ Cos[theta2] + 1\/2\ b\ Sin[theta2])\)\^2 + \
\((1\/2\ b\ Cos[theta2] + 1\/2\ a\ Sin[theta2])\)\ \((d\ mw + 
                mw\ \((1\/2\ b\ Cos[theta2] + 1\/2\ a\ Sin[theta2])\))\) - 
          d\ mw\ \((\(-\(1\/2\)\)\ b\ Cos[theta3] + 1\/2\ a\ Sin[theta3])\) + 
          mw\ \((\(-\(1\/2\)\)\ a\ Cos[theta3] - 1\/2\ b\ Sin[theta3])\)\^2 + \
\((\(-\(1\/2\)\)\ b\ Cos[theta3] + 1\/2\ a\ Sin[theta3])\)\ \((\(-d\)\ mw + 
                mw\ \((\(-\(1\/2\)\)\ b\ Cos[theta3] + 
                      1\/2\ a\ Sin[theta3])\))\), 
        1\/2\ \((\(-d\)\ mw\ Sin[theta2] - 
                mw\ Sin[
                    theta2]\ \((1\/2\ b\ Cos[theta2] + 
                      1\/2\ a\ Sin[theta2])\) + 
                2\ mw\ Cos[
                    theta2]\ \((\(-\(1\/2\)\)\ a\ Cos[theta2] + 
                      1\/2\ b\ Sin[theta2])\) - 
                Sin[theta2]\ \((d\ mw + 
                      mw\ \((1\/2\ b\ Cos[theta2] + 
                            1\/2\ a\ Sin[theta2])\))\))\) + 
          1\/2\ \((d\ mw\ Sin[theta3] - 
                mw\ Sin[
                    theta3]\ \((\(-\(1\/2\)\)\ b\ Cos[theta3] + 
                      1\/2\ a\ Sin[theta3])\) + 
                2\ mw\ Cos[
                    theta3]\ \((\(-\(1\/2\)\)\ a\ Cos[theta3] - 
                      1\/2\ b\ Sin[theta3])\) - 
                Sin[theta3]\ \((\(-d\)\ mw + 
                      mw\ \((\(-\(1\/2\)\)\ b\ Cos[theta3] + 
                            1\/2\ a\ Sin[theta3])\))\))\), 
        1\/2\ \((d\ mw\ Cos[theta2] + 
                mw\ Cos[
                    theta2]\ \((1\/2\ b\ Cos[theta2] + 
                      1\/2\ a\ Sin[theta2])\) + 
                2\ mw\ Sin[
                    theta2]\ \((\(-\(1\/2\)\)\ a\ Cos[theta2] + 
                      1\/2\ b\ Sin[theta2])\) + 
                Cos[theta2]\ \((d\ mw + 
                      mw\ \((1\/2\ b\ Cos[theta2] + 
                            1\/2\ a\ Sin[theta2])\))\))\) + 
          1\/2\ \((\(-d\)\ mw\ Cos[theta3] + 
                mw\ Cos[
                    theta3]\ \((\(-\(1\/2\)\)\ b\ Cos[theta3] + 
                      1\/2\ a\ Sin[theta3])\) + 
                2\ mw\ Sin[
                    theta3]\ \((\(-\(1\/2\)\)\ a\ Cos[theta3] - 
                      1\/2\ b\ Sin[theta3])\) + 
                Cos[theta3]\ \((\(-d\)\ mw + 
                      mw\ \((\(-\(1\/2\)\)\ b\ Cos[theta3] + 
                            1\/2\ a\ Sin[theta3])\))\))\), 
        d\^2\ mw + Wxx + 
          d\ mw\ \((1\/2\ b\ Cos[theta2] + 1\/2\ a\ Sin[theta2])\), 
        d\^2\ mw + Wxx - 
          d\ mw\ \((\(-\(1\/2\)\)\ b\ Cos[theta3] + 
                1\/2\ a\ Sin[theta3])\)}, {1\/2\ \((\(-d\)\ mw\ Sin[theta2] - 
                mw\ Sin[
                    theta2]\ \((1\/2\ b\ Cos[theta2] + 
                      1\/2\ a\ Sin[theta2])\) + 
                2\ mw\ Cos[
                    theta2]\ \((\(-\(1\/2\)\)\ a\ Cos[theta2] + 
                      1\/2\ b\ Sin[theta2])\) - 
                Sin[theta2]\ \((d\ mw + 
                      mw\ \((1\/2\ b\ Cos[theta2] + 
                            1\/2\ a\ Sin[theta2])\))\))\) + 
          1\/2\ \((d\ mw\ Sin[theta3] - 
                mw\ Sin[
                    theta3]\ \((\(-\(1\/2\)\)\ b\ Cos[theta3] + 
                      1\/2\ a\ Sin[theta3])\) + 
                2\ mw\ Cos[
                    theta3]\ \((\(-\(1\/2\)\)\ a\ Cos[theta3] - 
                      1\/2\ b\ Sin[theta3])\) - 
                Sin[theta3]\ \((\(-d\)\ mw + 
                      mw\ \((\(-\(1\/2\)\)\ b\ Cos[theta3] + 
                            1\/2\ a\ Sin[theta3])\))\))\), 
        m1 + mw\ Cos[theta2]\^2 + mw\ Cos[theta3]\^2 + mw\ Sin[theta2]\^2 + 
          mw\ Sin[theta3]\^2, 0, \(-d\)\ mw\ Sin[theta2], 
        d\ mw\ Sin[
            theta3]}, {1\/2\ \((d\ mw\ Cos[theta2] + 
                mw\ Cos[
                    theta2]\ \((1\/2\ b\ Cos[theta2] + 
                      1\/2\ a\ Sin[theta2])\) + 
                2\ mw\ Sin[
                    theta2]\ \((\(-\(1\/2\)\)\ a\ Cos[theta2] + 
                      1\/2\ b\ Sin[theta2])\) + 
                Cos[theta2]\ \((d\ mw + 
                      mw\ \((1\/2\ b\ Cos[theta2] + 
                            1\/2\ a\ Sin[theta2])\))\))\) + 
          1\/2\ \((\(-d\)\ mw\ Cos[theta3] + 
                mw\ Cos[
                    theta3]\ \((\(-\(1\/2\)\)\ b\ Cos[theta3] + 
                      1\/2\ a\ Sin[theta3])\) + 
                2\ mw\ Sin[
                    theta3]\ \((\(-\(1\/2\)\)\ a\ Cos[theta3] - 
                      1\/2\ b\ Sin[theta3])\) + 
                Cos[theta3]\ \((\(-d\)\ mw + 
                      mw\ \((\(-\(1\/2\)\)\ b\ Cos[theta3] + 
                            1\/2\ a\ Sin[theta3])\))\))\), 0, 
        m1 + mw\ Cos[theta2]\^2 + mw\ Cos[theta3]\^2 + mw\ Sin[theta2]\^2 + 
          mw\ Sin[theta3]\^2, 
        d\ mw\ Cos[theta2], \(-d\)\ mw\ Cos[theta3]}, {d\^2\ mw + Wxx + 
          d\ mw\ \((1\/2\ b\ Cos[theta2] + 
                1\/2\ a\ Sin[theta2])\), \(-d\)\ mw\ Sin[theta2], 
        d\ mw\ Cos[theta2], d\^2\ mw + Wxx, 
        0}, {d\^2\ mw + Wxx - 
          d\ mw\ \((\(-\(1\/2\)\)\ b\ Cos[theta3] + 1\/2\ a\ Sin[theta3])\), 
        d\ mw\ Sin[theta3], \(-d\)\ mw\ Cos[theta3], 0, 
        d\^2\ mw + Wxx}}\)], "Output",
  CellLabel->"Out[108]=",
  CellTags->"b:15.1"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Simplify[Cm.pp]", "Input",
  CellLabel->"In[109]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell[BoxData[
    \({1\/2\ mw\ \((2\ a\ vz\ w1 + 
            d\ \((2\ vy\ w1 + a\ w2\ \((2\ w1 + w2)\))\)\ Cos[theta2] - 
            d\ \((2\ vy\ w1 + a\ w3\ \((2\ w1 + w3)\))\)\ Cos[theta3] + 
            2\ d\ vz\ w1\ Sin[theta2] - 2\ b\ d\ w1\ w2\ Sin[theta2] - 
            b\ d\ w2\^2\ Sin[theta2] - 2\ d\ vz\ w1\ Sin[theta3] - 
            2\ b\ d\ w1\ w3\ Sin[theta3] - 
            b\ d\ w3\^2\ Sin[theta3])\), \(-\((m1 + 2\ mw)\)\)\ vz\ w1 - 
        d\ mw\ \((w1 + w2)\)\^2\ Cos[theta2] + 
        d\ mw\ \((w1 + w3)\)\^2\ Cos[theta3], 
      w1\ \((m1\ vy + 2\ mw\ vy - a\ mw\ w1)\) - 
        d\ mw\ \((w1 + w2)\)\^2\ Sin[theta2] + 
        d\ mw\ \((w1 + w3)\)\^2\ Sin[theta3], 
      1\/2\ d\ mw\ w1\ \((\((2\ vy - a\ w1)\)\ Cos[
                theta2] + \((2\ vz + b\ w1)\)\ Sin[theta2])\), 
      1\/2\ d\ mw\ w1\ \((\((\(-2\)\ vy + a\ w1)\)\ Cos[
                theta3] + \((\(-2\)\ vz + b\ w1)\)\ Sin[
                theta3])\)}\)], "Output",
  CellLabel->"Out[109]=",
  CellTags->"b:15.1"]
}, Open  ]],

Cell[CellGroupData[{

Cell["F", "Input",
  CellLabel->"In[110]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell[BoxData[
    \({\(-gc\)\ \((d\ mw\ Cos[theta1 + theta2] - 
              d\ mw\ Cos[theta1 + theta3] - a\ mw\ Sin[theta1])\) + 
        1\/8\ k\ \((2\ \((2\ z - 2\ z20 + a\ Cos[theta1] + 
                    2\ R\ Cos[theta1 + theta2] + b\ Sin[theta1] + 
                    2\ d\ Sin[theta1 + theta2])\)\ \((b\ Cos[theta1] + 
                    2\ d\ Cos[theta1 + theta2] - a\ Sin[theta1] - 
                    2\ R\ Sin[theta1 + theta2])\) + 
              2\ \((2\ z - 2\ z30 + a\ Cos[theta1] + 
                    2\ R\ Cos[theta1 + theta3] - b\ Sin[theta1] - 
                    2\ d\ Sin[theta1 + theta3])\)\ \((\(-b\)\ Cos[theta1] - 
                    2\ d\ Cos[theta1 + theta3] - a\ Sin[theta1] - 
                    2\ R\ Sin[theta1 + theta3])\))\), 
      Sin[theta1]\ \((\(-gc\)\ \((m1 + 2\ mw)\) + 
            1\/8\ k\ \((4\ \((2\ z - 2\ z20 + a\ Cos[theta1] + 
                        2\ R\ Cos[theta1 + theta2] + b\ Sin[theta1] + 
                        2\ d\ Sin[theta1 + theta2])\) + 
                  4\ \((2\ z - 2\ z30 + a\ Cos[theta1] + 
                        2\ R\ Cos[theta1 + theta3] - b\ Sin[theta1] - 
                        2\ d\ Sin[theta1 + theta3])\))\))\), 
      Cos[theta1]\ \((\(-gc\)\ \((m1 + 2\ mw)\) + 
            1\/8\ k\ \((4\ \((2\ z - 2\ z20 + a\ Cos[theta1] + 
                        2\ R\ Cos[theta1 + theta2] + b\ Sin[theta1] + 
                        2\ d\ Sin[theta1 + theta2])\) + 
                  4\ \((2\ z - 2\ z30 + a\ Cos[theta1] + 
                        2\ R\ Cos[theta1 + theta3] - b\ Sin[theta1] - 
                        2\ d\ Sin[theta1 + theta3])\))\))\), 
      c\ w2 - d\ gc\ mw\ Cos[
            theta1 + 
              theta2] + \((kdel\ \((8\ \@2\ a\^2\ Cos[theta2] + 
                  4\ \@2\ a\^2\ Sin[theta2])\)\ \((\(-ls\) + 
                  1\/2\ \@\(13\ a\^2 - 4\ \@2\ a\^2\ Cos[theta2] + 8\ \@2\ \
a\^2\ Sin[theta2]\))\))\)/\((4\ \@\(13\ a\^2 - 4\ \@2\ a\^2\ Cos[theta2] + 8\ \
\@2\ a\^2\ Sin[theta2]\))\) + 
        1\/4\ k\ \((2\ z - 2\ z20 + a\ Cos[theta1] + 
              2\ R\ Cos[theta1 + theta2] + b\ Sin[theta1] + 
              2\ d\ Sin[theta1 + theta2])\)\ \((2\ d\ Cos[theta1 + theta2] - 
              2\ R\ Sin[theta1 + theta2])\), 
      c\ w3 + d\ gc\ mw\ Cos[
            theta1 + 
              theta3] + \((kdel\ \((\(-8\)\ \@2\ a\^2\ Cos[theta3] + 
                  4\ \@2\ a\^2\ Sin[theta3])\)\ \((\(-ls\) + 
                  1\/2\ \@\(13\ a\^2 - 4\ \@2\ a\^2\ Cos[theta3] - 8\ \@2\ \
a\^2\ Sin[theta3]\))\))\)/\((4\ \@\(13\ a\^2 - 4\ \@2\ a\^2\ Cos[theta3] - 8\ \
\@2\ a\^2\ Sin[theta3]\))\) + 
        1\/4\ k\ \((2\ z - 2\ z30 + a\ Cos[theta1] + 
              2\ R\ Cos[theta1 + theta3] - b\ Sin[theta1] - 
              2\ d\ Sin[theta1 + theta3])\)\ \((\(-2\)\ d\ Cos[
                  theta1 + theta3] - 
              2\ R\ Sin[theta1 + theta3])\)}\)], "Output",
  CellLabel->"Out[110]=",
  CellTags->"b:15.1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["A Flexible Two Link Manipulator", "Subsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell[CellGroupData[{

Cell["Modeling a Flexible Link", "Subsubsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell["\<\
Consider a two link, planar mechanism in which the lower end of link 1 is \
attached to the space frame by a revolute joint and the upper end of link 1 \
and the lower end of link 2 are joined by a second revolute joint. Assume \
that the links have the same mass per unit length, rho, modulus of \
elasticity, El, cross sectional area moments (constant along the length) \
{Ixx,Iyy,Izz}, and the lengths L1 and L2, respectively.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell["\<\
There are many approaches to modeling individual flexible elements. One view \
fixes a 'primary body reference frame' in an infinitesimal element at the \
inboard joint. The location of any point in the deformed body are specified \
in this frame. As the body undergoes motion, it is necessary to track the \
location and orientation of the primary body frame as well as the location in \
that frame of any interesting body points. Thus, the number of degrees of \
freedom associated with a free deformable body will be the six degrees of \
freedom associated with the primary frame plus some number of degrees of \
freedom associated with the body deformations. Usually, it is desired to \
characterize body deformations by a finite number of degrees of freedom and \
this is often accomplished by using finite element or (truncated) modal \
expansion methods.      
As an example take one of the links in the two link manipulator. We fix a \
reference frame in the link at its inboard joint. Suppose that the beam lies \
along the z-axis (points upward) with the inboard joint at the bottom. \
Moroever, assume that the link bends only in the y-z plane. Within this frame \
the link deforms as a centelevered beam and we can characterize the beam \
deformations in terms of its modes.  A cantelevered (Euler) beam of length L \
has the characteristic equation
\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\t\tCos[b*L] * Cosh[b*L] = -1
which has a countable set of solutions br, r = 1,2,3...; the first three \
roots are approximately: \
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell["bL={0.600*Pi,1.49*Pi,2.50*Pi};", "Input",
  CellLabel->"In[112]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell["They define the natural frequencies by the relation", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell["w[i_]:=(bL[[i]])^2*(El*Ixx/(rho*L^4))^(1/2);", "Input",
  CellLabel->"In[113]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell["Correspondingly, the mode shapes are:", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell["\<\
W[i_,z_]:=((Cosh[bL[[i]]*z/L]-Cos[bL[[i]]*z/L])/(Cosh[bL[[i]]]+Cos[bL[[i]]])

              - \
(Sinh[bL[[i]]*z/L]-Sin[bL[[i]]*z/L])/(Sinh[bL[[i]]]+Sin[bL[[i]]]));  \
\>", "Input",
  CellLabel->"In[114]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell["\<\
From the above information we can construct, for example, a one-mode model \
for a link of length L. First, the inertia, damping and stiffness matrices \
are:\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell["\<\
M={{Ia,0,0,0,-m,0,0},
   {0,Ib,0,m,0,0,a},
   {0,0,Ic,0,0,0,0},
   {0,m,0, m,0,0,b},
   {-m,0,0,0,m,0,0},
   {0,0,0,0,0,m,0},
   {0,a,0,b,0,0,1}};\
\>", "Input",
  CellLabel->"In[115]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell["\<\
B=0*IdentityMatrix[7]; 
K=DiagonalMatrix[{0,0,0,0,0,0,(0.600*Pi)^2}]*(El*Ixx/(m*L^3));\
\>", "Input",
  CellLabel->"In[116]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell["\<\
where Ia, Ib, Ic are respectively Ixx m L^2/A, Iyy m L^2/A, Izz m L^2/A. Let \
y1 denote the modal coordinate, then the y position of the right end and \
anglular rotation about the x-axis (phiend) under deformation are given by:\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell["yend=W[1,L]*y1; phiend=D[W[1,z],z]/.z->L;", "Input",
  CellLabel->"In[118]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell["\<\
and the center of mass (which corresponds to the midpoint) is\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell["ycom=W[1,L/2]*y1;", "Input",
  CellLabel->"In[119]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell["Consequently, we have the matrices Cout and Ccom:", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell["\<\
Cout={{0,D[W[1,z],z]/.z->L},
      {0,0},{0,0},
      {0,0},{0,W[1,L]},{L,0}};
Ccom={{0,0},{0,W[1,L/2]},{L,0}};\
\>", "Input",
  CellLabel->"In[120]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell["The link body data is then", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell[CellGroupData[{

Cell["Link1={N[Ccom],{{2,N[Cout]}},m,{M,B,K},{y11},{v11}}/.L->L1", "Input",
  CellLabel->"In[122]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell[BoxData[
    \({{{0.`, 0.`}, {0.`, 0.223912970343545`}, {L1, 
          0.`}}, {{2, {{0.`, 0.9043710046157647`\/L1}, {0.`, 0.`}, {0.`, 
              0.`}, {0.`, 0.`}, {0.`, 0.6583189111222529`}, {L1, 0.`}}}}, 
      m, {{{Ia, 0, 0, 0, \(-m\), 0, 0}, {0, Ib, 0, m, 0, 0, a}, {0, 0, Ic, 0, 
            0, 0, 0}, {0, m, 0, m, 0, 0, b}, {\(-m\), 0, 0, 0, m, 0, 0}, {0, 
            0, 0, 0, 0, m, 0}, {0, a, 0, b, 0, 0, 1}}, {{0, 0, 0, 0, 0, 0, 
            0}, {0, 0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 
            0, 0}, {0, 0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 
            0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 0, 0}, {0, 0, 
            0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 0, 0}, {0, 
            0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 
            0, \(3.553057584392169`\ El\ Ixx\)\/\(L1\^3\ m\)}}}, {y11}, \
{v11}}\)], "Output",
  CellLabel->"Out[122]=",
  CellTags->"b:15.2"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Link2={N[Ccom],{{3,N[Cout]}},m,{M,B,K},{y21},{v21}}/.L->L2", "Input",
  CellLabel->"In[123]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell[BoxData[
    \({{{0.`, 0.`}, {0.`, 0.223912970343545`}, {L2, 
          0.`}}, {{3, {{0.`, 0.9043710046157647`\/L2}, {0.`, 0.`}, {0.`, 
              0.`}, {0.`, 0.`}, {0.`, 0.6583189111222529`}, {L2, 0.`}}}}, 
      m, {{{Ia, 0, 0, 0, \(-m\), 0, 0}, {0, Ib, 0, m, 0, 0, a}, {0, 0, Ic, 0, 
            0, 0, 0}, {0, m, 0, m, 0, 0, b}, {\(-m\), 0, 0, 0, m, 0, 0}, {0, 
            0, 0, 0, 0, m, 0}, {0, a, 0, b, 0, 0, 1}}, {{0, 0, 0, 0, 0, 0, 
            0}, {0, 0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 
            0, 0}, {0, 0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 
            0, 0, 0}}, {{0, 0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 0, 0}, {0, 0, 
            0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 0, 0}, {0, 
            0, 0, 0, 0, 0, 0}, {0, 0, 0, 0, 0, 
            0, \(3.553057584392169`\ El\ Ixx\)\/\(L2\^3\ m\)}}}, {y21}, \
{v21}}\)], "Output",
  CellLabel->"Out[123]=",
  CellTags->"b:15.2"]
}, Open  ]],

Cell["BodyList={Link1,Link2};", "Input",
  CellLabel->"In[124]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell["\<\
r1={1};
H1=Transpose[{{1,0,0,0,0,0}}];
q1={theta1};p1={w1};

r2={1};
H2=Transpose[{{1,0,0,0,0,0}}];
q2={theta2};p2={w2};

JointList={{r1,H1,q1,p1},{r2,H2,q2,p2}};\
\>", "Input",
  CellLabel->"In[125]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell["\<\
TreeList={{{1,1},{2,2}}};
PE=0; Q={0,0,0,0};\
\>", "Input",
  CellLabel->"In[132]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell[CellGroupData[{

Cell["\<\
{V,X,H,M,Cp,Fp,p,q}=CreateModel[JointList,BodyList,TreeList,g,PE,Q];\
\>", "Input",
  CellLabel->"In[134]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell[BoxData[
    \("Computing Joint Kinematics"\)], "Print",
  CellLabel->"From In[134]:=",
  CellTags->"b:15.2"],

Cell[BoxData[
    \("Computing joint 1 kinematics"\)], "Print",
  CellLabel->"From In[134]:=",
  CellTags->"b:15.2"],

Cell[BoxData[
    \("Computing joint 2 kinematics"\)], "Print",
  CellLabel->"From In[134]:=",
  CellTags->"b:15.2"],

Cell[BoxData[
    \("Computing Potential Functions"\)], "Print",
  CellLabel->"From In[134]:=",
  CellTags->"b:15.2"],

Cell[BoxData[
    \("Computing Inertia Matrix"\)], "Print",
  CellLabel->"From In[134]:=",
  CellTags->"b:15.2"],

Cell[BoxData[
    \("Computing Poincare Function"\)], "Print",
  CellLabel->"From In[134]:=",
  CellTags->"b:15.2"]
}, Open  ]],

Cell["We can examine some of the results:", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell[CellGroupData[{

Cell["M", "Input",
  CellLabel->"In[135]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell[BoxData[
    \({{2\ Ia + 
          m\ \((\((\(-\(\(0.9043710046157647`\ y11\ Cos[theta2]\)\/L1\)\) - \
Sin[theta2])\)\ \((\(-L1\)\ Cos[theta2] - 0.6583189111222529`\ y11\ \
Sin[theta2])\) + \((0.6583189111222529`\ y11\ Cos[theta2] - L1\ \
Sin[theta2])\)\ \((Cos[theta2] - \(0.9043710046157647`\ y11\ \
Sin[theta2]\)\/L1)\))\)\^2 - 
          m\ \((\((\(0.9043710046157647`\ y11\ Cos[theta2]\)\/L1 + 
                      Sin[theta2])\)\ \((0.6583189111222529`\ y11\ Cos[
                          theta2] - 
                      L1\ Sin[theta2])\) + \((\(-L1\)\ Cos[theta2] - 
                      0.6583189111222529`\ y11\ Sin[theta2])\)\ \((Cos[
                        theta2] - \(0.9043710046157647`\ y11\ \
Sin[theta2]\)\/L1)\))\) + \((\((\(0.9043710046157647`\ y11\ Cos[theta2]\)\/L1 \
+ Sin[theta2])\)\ \((0.6583189111222529`\ y11\ Cos[theta2] - 
                      L1\ Sin[theta2])\) + \((\(-L1\)\ Cos[theta2] - 
                      0.6583189111222529`\ y11\ Sin[theta2])\)\ \((Cos[
                        theta2] - \(0.9043710046157647`\ y11\ \
Sin[theta2]\)\/L1)\))\)\ \((\(-m\) + 
                m\ \((\((\(0.9043710046157647`\ y11\ Cos[theta2]\)\/L1 + 
                            Sin[theta2])\)\ \((0.6583189111222529`\ y11\ Cos[
                                theta2] - 
                            L1\ Sin[theta2])\) + \((\(-L1\)\ Cos[theta2] - 
                            0.6583189111222529`\ y11\ Sin[theta2])\)\ \((Cos[
                              theta2] - \(0.9043710046157647`\ y11\ \
Sin[theta2]\)\/L1)\))\))\), 
        Ia - m\ \((\((\(0.9043710046157647`\ y11\ Cos[theta2]\)\/L1 + 
                      Sin[theta2])\)\ \((0.6583189111222529`\ y11\ Cos[
                          theta2] - 
                      L1\ Sin[theta2])\) + \((\(-L1\)\ Cos[theta2] - 
                      0.6583189111222529`\ y11\ Sin[theta2])\)\ \((Cos[
                        theta2] - \(0.9043710046157647`\ y11\ \
Sin[theta2]\)\/L1)\))\), 
        1\/2\ \((\(0.9043710046157647`\ Ia\)\/L1 - 
              0.6583189111222529`\ m\ \((Cos[
                      theta2] - \(0.9043710046157647`\ y11\ \
Sin[theta2]\)\/L1)\) + 
              1.3166378222445059`\ m\ \((\(-\(\(0.9043710046157647`\ y11\ Cos[
                              theta2]\)\/L1\)\) - 
                    Sin[theta2])\)\ \((\((\(-\(\(0.9043710046157647`\ y11\ \
Cos[theta2]\)\/L1\)\) - Sin[theta2])\)\ \((\(-L1\)\ Cos[theta2] - 
                          0.6583189111222529`\ y11\ Sin[
                              theta2])\) + \((0.6583189111222529`\ y11\ Cos[
                              theta2] - 
                          L1\ Sin[theta2])\)\ \((Cos[
                            
                            theta2] - \(0.9043710046157647`\ y11\ Sin[theta2]\
\)\/L1)\))\) + \((\(-\(\(0.9043710046157647`\ m\)\/L1\)\) + 
                    0.6583189111222529`\ m\ \((Cos[
                            theta2] - \(0.9043710046157647`\ y11\ Sin[theta2]\
\)\/L1)\))\)\ \((\((\(0.9043710046157647`\ y11\ Cos[theta2]\)\/L1 + 
                          Sin[theta2])\)\ \((0.6583189111222529`\ y11\ Cos[
                              theta2] - 
                          L1\ Sin[theta2])\) + \((\(-L1\)\ Cos[theta2] - 
                          0.6583189111222529`\ y11\ Sin[theta2])\)\ \((Cos[
                            theta2] - \(0.9043710046157647`\ y11\ Sin[theta2]\
\)\/L1)\))\) + \(1\/L1\) \((0.9043710046157647`\ \((Ia - 
                        m\ \((\((\(0.9043710046157647`\ y11\ \
Cos[theta2]\)\/L1 + Sin[theta2])\)\ \((0.6583189111222529`\ y11\ Cos[
                                        theta2] - 
                                    L1\ Sin[theta2])\) + \((\(-L1\)\ Cos[
                                        theta2] - 
                                    0.6583189111222529`\ y11\ Sin[
                                        theta2])\)\ \((Cos[
                                      theta2] - \(0.9043710046157647`\ y11\ \
Sin[theta2]\)\/L1)\))\))\))\) + 
              0.6583189111222529`\ \((Cos[
                      theta2] - \(0.9043710046157647`\ y11\ \
Sin[theta2]\)\/L1)\)\ \((\(-m\) + 
                    m\ \((\((\(0.9043710046157647`\ y11\ Cos[theta2]\)\/L1 + 
                                Sin[theta2])\)\ \((0.6583189111222529`\ y11\ \
Cos[theta2] - L1\ Sin[theta2])\) + \((\(-L1\)\ Cos[theta2] - 
                                0.6583189111222529`\ y11\ Sin[
                                    theta2])\)\ \((Cos[
                                  theta2] - \(0.9043710046157647`\ y11\ \
Sin[theta2]\)\/L1)\))\))\))\), 
        0}, {Ia - 
          m\ \((\((\(0.9043710046157647`\ y11\ Cos[theta2]\)\/L1 + 
                      Sin[theta2])\)\ \((0.6583189111222529`\ y11\ Cos[
                          theta2] - 
                      L1\ Sin[theta2])\) + \((\(-L1\)\ Cos[theta2] - 
                      0.6583189111222529`\ y11\ Sin[theta2])\)\ \((Cos[
                        theta2] - \(0.9043710046157647`\ y11\ \
Sin[theta2]\)\/L1)\))\), 
        Ia, \(0.9043710046157647`\ Ia\)\/L1 - 
          0.6583189111222529`\ m\ \((Cos[
                  theta2] - \(0.9043710046157647`\ y11\ Sin[theta2]\)\/L1)\), 
        0}, {1\/2\ \((\(0.9043710046157647`\ Ia\)\/L1 - 
              0.6583189111222529`\ m\ \((Cos[
                      theta2] - \(0.9043710046157647`\ y11\ \
Sin[theta2]\)\/L1)\) + 
              1.3166378222445059`\ m\ \((\(-\(\(0.9043710046157647`\ y11\ Cos[
                              theta2]\)\/L1\)\) - 
                    Sin[theta2])\)\ \((\((\(-\(\(0.9043710046157647`\ y11\ \
Cos[theta2]\)\/L1\)\) - Sin[theta2])\)\ \((\(-L1\)\ Cos[theta2] - 
                          0.6583189111222529`\ y11\ Sin[
                              theta2])\) + \((0.6583189111222529`\ y11\ Cos[
                              theta2] - 
                          L1\ Sin[theta2])\)\ \((Cos[
                            theta2] - \(0.9043710046157647`\ y11\ Sin[theta2]\
\)\/L1)\))\) + \((\(-\(\(0.9043710046157647`\ m\)\/L1\)\) + 
                    0.6583189111222529`\ m\ \((Cos[
                            theta2] - \(0.9043710046157647`\ y11\ Sin[theta2]\
\)\/L1)\))\)\ \((\((\(0.9043710046157647`\ y11\ Cos[theta2]\)\/L1 + 
                          Sin[theta2])\)\ \((0.6583189111222529`\ y11\ Cos[
                              theta2] - 
                          L1\ Sin[theta2])\) + \((\(-L1\)\ Cos[theta2] - 
                          0.6583189111222529`\ y11\ Sin[theta2])\)\ \((Cos[
                            theta2] - \(0.9043710046157647`\ y11\ Sin[theta2]\
\)\/L1)\))\) + \(1\/L1\) \((0.9043710046157647`\ \((Ia - 
                        m\ \((\((\(0.9043710046157647`\ y11\ \
Cos[theta2]\)\/L1 + Sin[theta2])\)\ \((0.6583189111222529`\ y11\ Cos[
                                        theta2] - 
                                    L1\ Sin[theta2])\) + \((\(-L1\)\ Cos[
                                        theta2] - 
                                    0.6583189111222529`\ y11\ Sin[
                                        theta2])\)\ \((Cos[
                                      theta2] - \(0.9043710046157647`\ y11\ \
Sin[theta2]\)\/L1)\))\))\))\) + 
              0.6583189111222529`\ \((Cos[
                      theta2] - \(0.9043710046157647`\ y11\ \
Sin[theta2]\)\/L1)\)\ \((\(-m\) + 
                    m\ \((\((\(0.9043710046157647`\ y11\ Cos[theta2]\)\/L1 + 
                                Sin[theta2])\)\ \((0.6583189111222529`\ y11\ \
Cos[theta2] - L1\ Sin[theta2])\) + \((\(-L1\)\ Cos[theta2] - 
                                0.6583189111222529`\ y11\ Sin[
                                    theta2])\)\ \((Cos[
                                  theta2] - \(0.9043710046157647`\ y11\ \
Sin[theta2]\)\/L1)\))\))\))\), \(0.9043710046157647`\ Ia\)\/L1 - 
          0.6583189111222529`\ m\ \((Cos[
                  theta2] - \(0.9043710046157647`\ y11\ Sin[theta2]\)\/L1)\), 
        1 + 0.43338378874118877`\ m\ \((\(-\(\(0.9043710046157647`\ y11\ \
Cos[theta2]\)\/L1\)\) - Sin[theta2])\)\^2 + \(0.9043710046157647`\ \
\((\(0.9043710046157647`\ Ia\)\/L1 - 0.6583189111222529`\ m\ \((Cos[theta2] - \
\(0.9043710046157647`\ y11\ Sin[theta2]\)\/L1)\))\)\)\/L1 + 
          0.6583189111222529`\ \((Cos[
                  theta2] - \(0.9043710046157647`\ y11\ Sin[theta2]\)\/L1)\)\ \
\((\(-\(\(0.9043710046157647`\ m\)\/L1\)\) + 
                0.6583189111222529`\ m\ \((Cos[
                        theta2] - \(0.9043710046157647`\ y11\ \
Sin[theta2]\)\/L1)\))\), 0}, {0, 0, 0, 1}}\)], "Output",
  CellLabel->"Out[135]=",
  CellTags->"b:15.2"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Simplify[Cp.p]", "Input",
  CellLabel->"In[136]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell[BoxData[
    \({\(1\/L1\^2\) \((\((\(\(0.2166918943705944`\)\(\[InvisibleSpace]\)\) + 
                0.`\ \[ImaginaryI])\)\ m\ \((v11\ y11\ \
\((\((\(-4.96954834607716`\) + 
                            0.`\ \[ImaginaryI])\)\ v11 - \
\((\(\(10.990065627299819`\)\(\[InvisibleSpace]\)\) + 
                            0.`\ \[ImaginaryI])\)\ L1\ w1 + \
\((\(\(11.548846405772313`\)\(\[InvisibleSpace]\)\) + 
                            0.`\ \[ImaginaryI])\)\ L1\^2\ w1 - \
\((\(\(5.495032813649909`\)\(\[InvisibleSpace]\)\) + 
                            0.`\ \[ImaginaryI])\)\ L1\ w2 + \
\((\(\(6.543095311917819`\)\(\[InvisibleSpace]\)\) + 
                            0.`\ \[ImaginaryI])\)\ w1\ y11\^2)\) + \
\((\((\(\(0.`\)\(\[InvisibleSpace]\)\) + 
                            0.`\ \[ImaginaryI])\)\ L1\^3\ v11\ w2 + 
                      L1\^2\ \((\((\(\(0.`\)\(\[InvisibleSpace]\)\) + 
                                  0.`\ \[ImaginaryI])\)\ v11\^2 + 
                            v11\ w2\ \((\((\(\(0.`\)\(\[InvisibleSpace]\)\) + 
                                      0.`\ \[ImaginaryI])\) - \((\(\(2.`\)\(\
\[InvisibleSpace]\)\) + 
                                        0.`\ \[ImaginaryI])\)\ y11)\))\) + \
\((\(\(0.`\)\(\[InvisibleSpace]\)\) + 
                            0.`\ \[ImaginaryI])\)\ v11\^2\ y11\^2 - \
\((\(\(1.6357738279794545`\)\(\[InvisibleSpace]\)\) + 
                            0.`\ \[ImaginaryI])\)\ v11\ w2\ y11\^3 + 
                      L1\ \((v11\^2\ \((\((\(\(0.`\)\(\[InvisibleSpace]\)\) + 
                                      0.`\ \[ImaginaryI])\) - \
\((\(\(4.96954834607716`\)\(\[InvisibleSpace]\)\) + 
                                        0.`\ \[ImaginaryI])\)\ y11)\) + 
                            v11\ w2\ \((\((\(\(2.7475164068249547`\)\(\
\[InvisibleSpace]\)\) + 
                                        0.`\ \[ImaginaryI])\)\ y11 + \
\((\(\(0.`\)\(\[InvisibleSpace]\)\) + 
                                        0.`\ \[ImaginaryI])\)\ \
y11\^2)\))\))\)\ Cos[
                    theta2] + \((\((\(\(3.0380412383878643`\)\(\
\[InvisibleSpace]\)\) + 0.`\ \[ImaginaryI])\)\ L1\^3\ v11\ w2 + 
                      L1\^2\ \((\((\(-2.`\) + 0.`\ \[ImaginaryI])\)\ v11\^2 + 
                            v11\ w2\ \((\((\(\(3.0380412383878643`\)\(\
\[InvisibleSpace]\)\) + 
                                      0.`\ \[ImaginaryI])\) + \((\(\(0.`\)\(\
\[InvisibleSpace]\)\) + 
                                        0.`\ \[ImaginaryI])\)\ y11)\))\) - \
\((\(\(4.9073214839383645`\)\(\[InvisibleSpace]\)\) + 
                            0.`\ \[ImaginaryI])\)\ v11\^2\ y11\^2 + \
\((\(\(0.`\)\(\[InvisibleSpace]\)\) + 
                            0.`\ \[ImaginaryI])\)\ v11\ w2\ y11\^3 + 
                      L1\ \((v11\^2\ \((\((\(\(2.7475164068249547`\)\(\
\[InvisibleSpace]\)\) + 
                                      0.`\ \[ImaginaryI])\) + \((\(\(0.`\)\(\
\[InvisibleSpace]\)\) + 0.`\ \[ImaginaryI])\)\ y11)\) + 
                            v11\ w2\ \((\((\(\(0.`\)\(\[InvisibleSpace]\)\) + 
                                        0.`\ \[ImaginaryI])\)\ y11 + \
\((\(\(2.4847741730385806`\)\(\[InvisibleSpace]\)\) + 
                                        0.`\ \[ImaginaryI])\)\ \
y11\^2)\))\))\)\ Sin[
                    theta2])\))\), \(1\/\(4\ L1\^2\)\) \((m\ v11\ \
\((\((1.7335351549647553`\ L1\^2\ w1\ y11 + 
                      1.4178357181868273`\ w1\ y11\^3)\)\ Cos[theta2]\^3 + 
                L1\ Cos[theta2]\^2\ \((\(-1.1102230246251565`*^-16\)\ v11\ \
y11 - 4.762916280073505`\ w1\ y11 + \((\(-2.6332756444890117`\)\ L1\^2\ w1 - 
                            2.1537216905554284`\ w1\ y11\^2)\)\ Sin[
                          theta2])\) + 
                L1\ Sin[
                    theta2]\ \((\(-2.6332756444890117`\)\ L1\ w1 + \
\((1.1102230246251565`*^-16\ v11\ y11 - 4.762916280073505`\ w1\ y11)\)\ Sin[
                          theta2] + \((\(-2.6332756444890117`\)\ L1\^2\ w1 - 
                            2.1537216905554284`\ w1\ y11\^2)\)\ \
Sin[theta2]\^2)\) + 
                y11\ Cos[
                    theta2]\ \((\(-2.1537216905554284`\)\ v11 - 
                      2.3814581400367527`\ L1\ w1 + 
                      1.1102230246251565`*^-16\ v11\ y11\ Sin[
                          theta2] + \((1.7335351549647553`\ L1\^2\ w1 + 
                            1.4178357181868273`\ w1\ y11\^2)\)\ \
Sin[theta2]\^2)\))\))\), \(1\/\(4\ L1\^2\)\) \((m\ \((w1\ w2\ y11\ \
\((\(-1.7335351549647553`\)\ L1\^2 - 
                      1.4178357181868273`\ y11\^2)\)\ Cos[theta2]\^3 + 
                w1\^2\ y11\ \((\(-5.005082810923666`\)\ L1\^2 - 
                      2.835671436373655`\ y11\^2)\)\ Cos[theta2]\^4 + 
                Cos[theta2]\^2\ \((\((1.4178357181868275`\ v11\^2 + 
                            2.220446049250313`*^-16\ L1\ v11\ w2 + 
                            L1\ w1\ \((4.762916280073505`\ w1 + 
                                  4.762916280073505`\ w2)\))\)\ y11 + 
                      L1\ w1\ w2\ \((2.6332756444890117`\ L1\^2 + 
                            2.1537216905554284`\ y11\^2)\)\ Sin[theta2] + 
                      w1\^2\ y11\ \((\(-10.010165621847332`\)\ L1\^2 - 
                            5.67134287274731`\ y11\^2)\)\ Sin[theta2]\^2)\) + 
                Cos[theta2]\ \((w2\ \((4.307443381110857`\ v11 + 
                            2.3814581400367527`\ L1\ w1 + 
                            2.3814581400367527`\ L1\ w2)\)\ y11 + 
                      v11\ \((2.220446049250313`*^-16\ L1\ v11 - 
                            2.220446049250313`*^-16\ w2\ y11\^2)\)\ Sin[
                          theta2] + 
                      w1\ w2\ y11\ \((\(-1.7335351549647553`\)\ L1\^2 - 
                            1.4178357181868273`\ y11\^2)\)\ Sin[theta2]\^2)\) \
+ Sin[theta2]\ \((2.1537216905554284`\ v11\^2 + 
                      4.762916280073505`\ L1\ v11\ w2 + 
                      L1\^2\ w2\ \((2.6332756444890117`\ w1 + 
                            2.6332756444890117`\ w2)\) + \
\((1.4178357181868275`\ v11\^2 - 2.220446049250313`*^-16\ L1\ v11\ w2 + 
                            L1\ w1\ \((4.762916280073505`\ w1 + 
                                  4.762916280073505`\ w2)\))\)\ y11\ Sin[
                          theta2] + 
                      L1\ w1\ w2\ \((2.6332756444890117`\ L1\^2 + 
                            2.1537216905554284`\ y11\^2)\)\ Sin[theta2]\^2 + 
                      w1\^2\ y11\ \((\(-5.005082810923666`\)\ L1\^2 - 
                            2.835671436373655`\ y11\^2)\)\ Sin[theta2]\^3)\))\
\))\), 0}\)], "Output",
  CellLabel->"Out[136]=",
  CellTags->"b:15.2"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Fp", "Input",
  CellLabel->"In[137]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell[BoxData[
    \({\(-\(\(1\/L1\) \((gc\ m\ \((\(-Sin[
                        theta1]\)\ \((2\ L1\^2 + \((L1\ L2 + 
                              0.2025003979360917`\ y11\ y21)\)\ Cos[theta2] - 
                        0.9043710046157647`\ L2\ y11\ Sin[theta2] + 
                        0.223912970343545`\ L1\ y21\ Sin[theta2])\) + 
                  Cos[theta1]\ \((0.8822318814657979`\ L1\ y11 + \
\((\(-0.9043710046157647`\)\ L2\ y11 + 0.223912970343545`\ L1\ y21)\)\ Cos[
                            theta2] - L1\ L2\ Sin[theta2] - 
                        0.2025003979360917`\ y11\ y21\ Sin[
                            theta2])\))\))\)\)\), \(-\(\(1\/L1\) \((gc\ m\ \
\((Sin[theta1]\ \((\(-L1\)\ L2\ Cos[theta2] - 
                        0.2025003979360917`\ y11\ y21\ Cos[
                            theta2] - \((\(-0.9043710046157647`\)\ L2\ y11 + 
                              0.223912970343545`\ L1\ y21)\)\ Sin[
                            theta2])\) + 
                  Cos[theta1]\ \((\(-0.9043710046157647`\)\ L2\ y11\ Cos[
                            theta2] + 
                        0.223912970343545`\ L1\ y21\ Cos[
                            theta2] - \((L1\ L2 + 
                              0.2025003979360917`\ y11\ y21)\)\ Sin[
                            theta2])\))\))\)\)\), \(3.553057584392169`\ El\ \
Ixx\ y11\)\/\(L1\^3\ m\) - \(1\/L1\) \((gc\ m\ \((Cos[
                      theta1]\ \((0.2025003979360917`\ y21\ Cos[theta2] - 
                        0.9043710046157647`\ L2\ Sin[theta2])\) + 
                  Sin[theta1]\ \((0.8822318814657979`\ L1 - 
                        0.9043710046157647`\ L2\ Cos[theta2] - 
                        0.2025003979360917`\ y21\ Sin[
                            theta2])\))\))\), \(3.553057584392169`\ El\ Ixx\ \
y21\)\/\(L2\^3\ m\) - \(1\/L1\) \((gc\ m\ \((Cos[
                      theta1]\ \((0.2025003979360917`\ y11\ Cos[theta2] + 
                        0.223912970343545`\ L1\ Sin[theta2])\) + 
                  Sin[theta1]\ \((0.223912970343545`\ L1\ Cos[theta2] - 
                        0.2025003979360917`\ y11\ Sin[
                            theta2])\))\))\)}\)], "Output",
  CellLabel->"Out[137]=",
  CellTags->"b:15.2"]
}, Open  ]],

Cell["\<\
The absolute position and orientation of a frame at the end of the second \
link can be obtained with the function EndEffector.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell["EndPos=EndEffector[3,TreeList,BodyList,X];", "Input",
  CellLabel->"In[138]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell["Let's take a look at the position vector:", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell[CellGroupData[{

Cell["EndPos[[{1,2,3},4]]", "Input",
  CellLabel->"In[139]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell[BoxData[
    \({0, 
      0.6583189111222529`\ y11\ Cos[theta1] - L1\ Sin[theta1] + 
        0.6583189111222529`\ y21\ \((Cos[
                  theta2]\ \((Cos[
                      theta1] - \(0.9043710046157647`\ y11\ \
Sin[theta1]\)\/L1)\) + \((\(-\(\(0.9043710046157647`\ y11\ Cos[
                              theta1]\)\/L1\)\) - Sin[theta1])\)\ Sin[
                  theta2])\) + 
        L2\ \((Cos[
                  theta2]\ \((\(-\(\(0.9043710046157647`\ y11\ Cos[
                              theta1]\)\/L1\)\) - 
                    Sin[theta1])\) - \((Cos[
                      theta1] - \(0.9043710046157647`\ y11\ \
Sin[theta1]\)\/L1)\)\ Sin[theta2])\), 
      L1\ Cos[theta1] + 0.6583189111222529`\ y11\ Sin[theta1] + 
        L2\ \((Cos[
                  theta2]\ \((Cos[
                      theta1] - \(0.9043710046157647`\ y11\ \
Sin[theta1]\)\/L1)\) - \((\(0.9043710046157647`\ y11\ Cos[theta1]\)\/L1 + 
                    Sin[theta1])\)\ Sin[theta2])\) + 
        0.6583189111222529`\ y21\ \((Cos[
                  theta2]\ \((\(0.9043710046157647`\ y11\ Cos[theta1]\)\/L1 + 
                    Sin[theta1])\) + \((Cos[
                      theta1] - \(0.9043710046157647`\ y11\ \
Sin[theta1]\)\/L1)\)\ Sin[theta2])\)}\)], "Output",
  CellLabel->"Out[139]=",
  CellTags->"b:15.2"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Tractor Semi-Trailer", "Subsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[CellGroupData[{

Cell[" System Description", "Subsubsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell["\<\
The goal is to show how deformable tires, tire parameters, and front end \
alignment parameters affect turning radius of a tractor \
semi\[CapitalEth]trailer. This example uses a simple model for a deformable \
pneumatic tire, namely, a single deformation parameter (proportional to \
sideslip angle), so that the cornering force is given by
                             F=-kappa*arctan[vy/vx]\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[GraphicsData["Metafile", "\<\
CF5dJ6E]HGAYHf4PEfU^I6mgLb15CDHPAVmbKF5d0@000FWT0@0006`00001000000000>01000A0`00
00000000002P>P00b5l00215CDH00040i6T10</700080000000000000000000000@0000300100@00
l0000000000000000000003R1020Z@<0AP0002`0000P0000ADe6:`500@0L00004000004@`=/10000
H000060000160000h0000=@00015CDH[<4020100000400000020?am0100<0000000001i00@0<0000
000001e0000D000020000?goool@0`00:T0002@0000H0000G8nR@P0000000000G8nR@SJTMD=HkVA3
24001S00000T00000A30fj^Z:Sh000000000000000060000D`1I04d0@P1?04`0=T00P400000d0000
0000o`40000100000@00077`1hFe?h8ON3h0080o000000000000080o000000000000024000080000
HP0000`0000100004P0000`000010000600000`0000000005P0000`0000H0000DP00070100010000
l_ooo`000000000000000901000000021`0405<0N@1]0680K`1/0000001=06l0KP1_07@0N@1`06D0
>P1C07T0K@1R06l0K00P03X0MP1U0780L`1Y06l0KP000>UgR?4B04PcjGL;0000:P0008Ca4P2P0nig
000002X0001L04h0C01C05`0CP1/07<0D`1U06<0M01Y06l0KP130500=00b0000`>dB022c1`4@@P1P
0000004000000000V>hB0<C^4P1L:A@0f>dB02Ll060<8A@0a>hB0000000100004>hB0<<l060<8A@0
a>hB0000000<:1@081/0H000000<:1@000000;hJ062lkQ80V>hB000000000000>>hB0>4o062lkQ80
000000400009lQ80V>hB00000020<3/c0000073^4P050000L3WhM`00J35h1fPaI7H02000000U0000
30000040001D0000E00006T1003[0000K`400?/000010000003j@@00nT5Y0@00n@000040001<0000
0000000000000000oooooooooom@0000LO1U0000000U0000300000d008160000304000010015CDH[
:d0000`00000000024010n00003D00000A30faH000000000KQoT@jG25D>_EYY3oBFS@[3:Ud?^L9i2
j/FD@cWRXT:VV9=3=a:]@ZJHTd<h4Ze2YYRC@cPB[D:VV9=3>1:]@Z8a7d>?K[i36]LL@hkj`4?03am3
Doo3@klW94>G;<E3_bLT@iP/aD>Ng9]3JdCW@ieXWT>^LNQ3HVfQ@e]Eid>VVZ93G<WT@jNJXT=LbNA3
O9[V@kLn8d>oanM3^2HN@fb[iT<]7AQ3KAoT@jG25D<00@<30`<30`430`<10@<30`410`>3000D@060
400000@0003om=_o900002@00000080m000000000000080m00000000000200009`0001P000020000
00000=_do`0000009@0000`0000200004`0000`000010000>`0000P0000K0000400008@L001M2@00
=P000100001;4`006PD005P0000d000000000000003oooooooooo`H0003j4_@4VA8H1G@BJ@Ed4VT5
M19Y1G@BJ@Df000040000?@9003>5`00F00002P00000000000000?oooooooooo0`000<h981Sa2H0H
@`ZV65T0000T000000000000003oooooooooo`8000132ZHHO1?Y75P0000X000000000000003ooooo
ooooo`<0003>4`lM;QC[75@DVQaI00009000000000000000ooooooooool20000E1BJ7=@L=0YH0000
:000000000000000ooooooooool30000nAcS2MHLPPV475d9?@0000P0000l0000200003h0000H0000
W00004l0003@0@00d@4001<0000<00000@0002D0000<00000000P2@0000T00000020@@0000000000
0020@@00000000000P0004H0001L0000D00004E=AR/8@082?000030000014<3K000004h000000000
a[N:?P80000200000P00008000014<3K000000000?lE@040400000@000020000900002@00000080m
000000000000080m0000000000020000G`0003P000030000>0000000000h0000000000000@040000
000000000000000000000000000U0000300000<0000U0000300000D0080k0000200001/0000@0000
Q1`005d9000f0000400004/C000J1@00F00003@00000000000000?oooooooooo1P000?XBm0BI4QP5
M19Y1G@BJ@Ed4VT5M19Y1CH0000@0000m0T00<hG001H0000:000000000000000ooooooooool30000
cPTP6?49P1Q32ZHHF@0002@00000000000000?oooooooooo0P0004<:YQQl4nTLF00002P000000000
00000?oooooooooo0`000<hC3ad^5>/LE1BJ75T0000T000000000000003oooooooooo`80001D59XL
e1`d2UP0000X000000000000003oooooooooo`<0003i7><9eQb22H@LG@Tm0000200003`000080000
@00001P0002K0000CP000=41003C0@009@0000`0000700209@0000`000000020900002@000000811
00000000000008110000000000020000:00000`000030000AP0008P0001l0000ADe6:`Q00P8l0000
<000004@`=/00000CP000000002Y4e0o0P000080000200000P00004@`=/000000000o`Q00@</0000
8000004@`=/2000000000>:nTd<gmaQ3CSK=@g/BCT<00@005D010100000400000P0002@0000T0000
0020?@00000000000020?@00000000000P0005l0000h00000`0003P000000000>000000000000040
3@0000000000000000000000000000009@0000`0000300009@0000`000050020E`0002@0000V0@00
U`0009`1003@00000P0007PBT0VW6N8<9@0000`0000700209@0000`000000020900002@000000811
00000000000008110000000000020000:00000`000030000AP000=`0003@0000ADe6:`Q00@>l0000
/000004@`=/B0000000001W/Sd>VIdQ36Nb?@j`@mT8Ik8m3<Bkb@[a8TT<12>m2V3:E@`48kd9d79Q3
0@S_@QQiVT<`;_9267VJ@j/@mT8HNIY3Za3f@QQiVT>/4?I267VJ@j`@mT8HNIY3YVM8@aQiVT?SF4Y3
MAbH@o_[Bd>I<YE3nn];@ke8TT?kjd]36Nb?@n=HBT<Ik8m3YVM8@`010`<30`<30`<30@<30`<3P`00
5401P10000040000`<30obP0000<00000P0002@0000T00000020?@00000000000020?@0000000000
0P0002L0000H00000P0000000030`<00000002D0000<00000P0001<0000<00000@0003/000080000
6`000100003n4@00Q``003H0000@0000oQ400;47001H0000@000000000000000ooooooooool90000
oQ6B1dXBN@NW4WT711=i1e0CTPM@4k47D1>a1e0C/@M@4k47=P000100001@4`00Q``005P0000d0000
00000000003oooooooooo`H0001@4jH<11>o3:LB_`a:4[l<oQ6V3?hAQ``m0000200003`000080000
?P0001P0000O0@00M`0003D1003<00004`0000`0000100009@0000`000000020900002@000000811
00000000000008110000000000020000AP0005`0001@0000ADe6:`Q00P8l0000<000004@`=/00000
CP0000000036]hXn0P000080000200000P00004@`=/000000000oaE00@0@000010000080000T0000
90000000P3d0000000000000P3d0000000000080001O0000>00000<0000h0000000003P000000000
000100@0000000000000000000000000000002D0000<00000`0002D0000<00001@00P3/000080000
6`000100003n4@00Q``003H0000@0000oQ400;47001H0000@000000000000000ooooooooool90000
oQ6B1dXBN@NW4WT711=i1e0CTPM@4k47D1>a1e0C/@M@4k47=P000100001@4`00Q``005P0000d0000
00000000003oooooooooo`H0001@4jH<11>o3:LB_`a:4[l<oQ6V3?hAQ``m0000200003`000080000
@00001P0000N0@00MP0003L1003>00009@0000`0000700209@0000`000000020900002@000000811
00000000000008110000000000020000:00000`000030000AP000=`0003@0000ADe6:`Q00@>l0000
/000004@`=/B0000000001F]a4>RVg935Jg4@e<l9D<E[LA35D/S@kP9ad?m]b53U??9@ofg8D=`gLa3
oKLQ@a@jcd<EBb=353[?@e8l9D<D>/m3DS`U@a@jcd=C?2E353[?@e<l9D<D>/m3XY]b@a@jcd?OS7A3
LMg<@oLOMT>EllU3mamf@kT9ad?g7gI35Jg4@mn<M4<E[LA3XY]b@`010`<30`<30`<30@<30`<3P`00
5401P10000040000`<30ob@0000T00000020?@00000000000020?@00000000000P0002D0000<0000
0P0001<0000<00000@0003/0000800006`000100002F6000:Pl003H0000@0000UQP005@:001H0000
@000000000000000ooooooooool90000UQPe2^8H70Xo6A`:W1TL2^PI=@[X6E@:j1UD2^PIE0[X6E@:
=P000100003X6@00:Pl005P0000d000000000000003oooooooooo`H0003X6DT?W1UR3clIHPoR668?
UQQ93iHH:Plm0000200003`000080000?P0001P000290@00X@0009l1003g00004`0000`000010000
9@0000`000000020900002@00000081100000000000008110000000000020000AP0005`0001@0000
ADe6:`Q00P8l0000<000004@`=/00000CP0000000036]hXn0P000080000200000P00004@`=/00000
0000oaE00@0@000010000080000T000090000000P3d0000000000000P3d0000000000080001O0000
>00000<0000h0000000003P000000000000100@0000000000000000000000000000002D0000<0000
0`0002D0000<00001@00P3/0000800006`000100002F6000:Pl003H0000@0000UQP005@:001H0000
@000000000000000ooooooooool90000UQPe2^8H70Xo6A`:W1TL2^PI=@[X6E@:j1UD2^PIE0[X6E@:
=P000100003X6@00:Pl005P0000d000000000000003oooooooooo`H0003X6DT?W1UR3clIHPoR668?
UQQ93iHH:Plm0000200003`000080000@00001P000280@00X0000:01003h00009@0000`000070020
9@0000`000000020900002@00000081100000000000008110000000000020000:00000`000030000
AP000;P0002/0000ADe6:bY0000T0000600005b?XT800000000005b?XT:mc0K05C36@`Q000I40000
>000004@`=^[ZRXn00000080000000003`0005@0B@1=04D0D`0P04h0A@1G0200DP1?04d0@@1>0000
=T00P400000d00000000o`40000100000@0005T0B6=J?X8ON3h0080o000000000000080o00000000
00000180000<00000@0001P0000<0000000001H0000<0000600002P0000<00000@000580001`0@00
0@000?;oool0000000000000002@0@000@0000L0101D06T0K@1U07<0801>06D0M`0P0580K`1]0640
KP000<L5C@1_06h0K`1d07T0L01U03X0E01Y06d0I@1c0200003ooooo00000000JbD0H5Be1`4P/`L1
G2TD03HM060P/`L1G2TD00PN061L:A@000000000000F@`1PG2TD022c1`4<8A@0`61>H53]4P0001<0
10000000000100000000007^4P1DkQ80e>`B0:2I;3>LkA80UB_hMaPfn7L00?oo[>dB089cj7L001<0
00000>QR7c>V1@X;j68O<l1GmgL0000032PD0000002n6P1PN>dB0:2I;3><ma80003ZMk1ej7Oooooo
0@0004mmm7OXHQlc20000>0f1C?XkA80j68O<l@7mGO0EoMgV0DQG`40000=??Ag`5OgMfAf00P00000
9@0000`000010000E00005@0000?0000T`4001P0002R0@000@000000nT400?Y13`000:0100010000
C00000000000000000000?ooooooooooD00005T0000000009@0000`0000=0020AP0009@000280000
ADe6:b]0000<0000000000Q00P8l0000<000004@`=/00000CP000000002Y4e0o0P00008000020000
0P00004@`=/000000000o`Q00@</00008000004@`=/2000000000>J^WD07C0e4iZjM@0a/d4<00@00
5D010100000400000P0002@0000T00000020?@00000000000020?@00000000000P0005l0000h0000
0`0003P000000000>0000000000000403@0000000000000000000000000000009@0000`000030000
9@0000`000050020E`0002@000030000W`4000L0000g0P000P0004l0Db=?00hJ9@0000`000070020
9@0000`000000020900002@00000081100000000000008110000000000020000:00000`000030000
AP0005`0001@0000ADe6:`Q00@<l0000<000004@`=/40000000005b?XSlGhM13iZjM@;lhcD?kG0U1
5n7@@eb?XSlGhM130041PAA00H0@0000100000000?lX000030000080000T000090000000P3d00000
00000000P3d0000000000080000W0000600000800000000000000000000U000030000080000C0000
30000040000U0000300000P0081F0000;0000040002J0@002@000:81000400005@0M6Tl0Z1V:01dJ
5@0M6RD0000<00001`00P1<0000<00000@0002D0000<00000000P2@0000T00000020@@0000000000
0020@@00000000000P0004H000280000O00004E=AR/8@082?000030000014<3K000004h000000000
ZA=@?`80000200000P00008000014<3K000000000?l8@043;000020000014<3K0P000000003V[Ye0
1d`=A:I5nT87C0e4004001E00@0@000010000080000T000090000000P3d0000000000000P3d00000
00000080001O0000>00000<0000h0000000003P000000000000100d0000000000000000000000000
000002D0000<00000`0002D0000<00001@00P5L0000T00000`0003<2001o0000=`800080001?05<S
d`MC8bD0000<00001`00P2D0000<00000000P2@0000T00000020@@00000000000020@@0000000000
0P0002P0000<00000`0004H0001L0000D00004E=AR/8@043?000030000014<3K10000000001iLOQ2
lF4<A6j90d<7C0e4NG7h@Qdf3TAiLOQ2lF4<A0010H4D@060400000@00000003o900002@00000080m
000000000000080m00000000000200009@0000`0000200004`0000`0000100009@0000`000080020
EP0002`0001l0000<@8008@0000i0P0010000<@76B<i25<Sa0N>8l@76B<U0000300000L0080C0000
30000040000U000030000000080T000090000000P440000000000000P44000000000008000160000
^0000:`00015CDH[:T0002@0000H0000G8nR@P0000000000G8nR@S?`jT87C0e424001T@0000h0000
0A30fj^Z:Sh000000P000000000?0000E01904d0A@1C0200CP1505L0801B04l0C@1104h0000f@020
@00003@00000003o0@00004000010000F02N^1DnPQmh?P00P3l0000000000000P3l0000000000000
4P0000`000010000600000`0000000005P0000`0000H00009@0000`000010000E00005@000200000
?08008/0001;0P000@000000nT400?Y1P@0004T200010000C00000000000000000000?oooooooooo
D00005P0000000009@0000`0000=0020AP0002P1000L0@00ADe6:b]0000<0000000002Y0000T0000
600005b?XT800000000005b?XT;V[Ye01C<=A0Q000I40000>000004@`=_T>0hn0000000000000000
3`0005@0B@1=04D0D`0P04h0A@1G0200DP1?04d0@@1>0000=T00P:@0002H00000000o`4000010000
2`0005<0L01Q06<0I@0P04H0LP1Q06d0I@1U>_4nnII_?V]S33oiUVlnR2XN?oVFKcj8lRdonII_?XRj
?CoiUVlnR89=?oVFKchFIUHonII_?Tl/JSoiUVlnJ`=f?oVFKcjfiH8onII_?[JiT3oiUVln0020?`00
000000000020?`0000000000000B000030000040000H000030000000000F0000300001P0000X0000
30000040001B0000L0400040003eoooo0000000000000000T0400000000700@0E01Y06d0I@1c0200
CP1U07L0801B06l0K@1Q06h0000004d0K`1^06l0M01i0700I@0j05@0J@1]06D0L`0P0000ooooo`00
000006/U061D]@L18;<70E`Y500f7@1P8;<70E`Y50087P1PG2TD0000000000005T<0H5`Y500P/`L1
324D0<1PCV1@kA80000C00@0000000000@0000000001kQ80E>hB0=C/4P2PVB`cW>dB09D[n7LH=_Qg
003oojc]4P22LnQg000C0000003XHQlcYPD:3>QR7c?0EoMg000000`X50000000_QX0H7S]4P2PVB`c
S?LB0000jWN`MNQgooooo`40001?OOAgj68O<`P0003P=PDcj>dB0>QR7c?41oEg`5OgMiP58El10000
3CcdMl1GmgMTMP08000002D0000<00000@0005@0002@0000:`0003d2001T0000BP80004000000?Y1
003j@B/000180P002`0004`00000000000000000003oooooooooof@0001C0700H@1S06D080160780
H@1]06D0000600001P0000D0000500001@0000<000060000100000D000090000000002D0000<0000
3@00P4H000340000^00004E=AR/[@00030000000000Z@000900001P0001LSj9200000000001LSj92
0eRi@`Ye0T<8@006A00003P000014<3KZjXZ?P0000020000000000l0001D04T0C@1505<0801>04D0
E`0P0580C`1=0440CP0003I008100000=00000000?l100000@000040001e07g;5cj27gPn0020?`00
000000000020?`0000000000000B000030000040000H000030000000000F0000300001P0000X0000
30000040001B0000L0400040003boooo0000000000000000T0400040000700@0E01Y06d0I@1c0200
CP1U07L0801B06l0K@1Q06h000371Dd0K`1^06l0M01i0700I@0j05@0J@1]06D0L`0P0000ooooo`00
000006/U061D]@L18;<70E`Y500f7@1P8;<70E`Y50087P1PG2TD0000000000005T<0H5`Y500P/`L1
324D0<1PCV30kA80000C00@0000000000@0000000001kQ80a>hB04C]4P2PVB`c3>hB09D[n7LH=_Qg
003ooac^4P22LnQg000C0000003XHQlcYPD:3NQR7c?0EoMg000000`X50000000_QX0H>S]4P2PVB`c
o?LB0000jWN`MNQgooooo`40001?OOAgj68O<`P0003P=PDcF>hB0>QR7c?41oEg`5OgMiP58El10000
3CcdMl1GmgMTMP08000002D0000<00000@0005@0001D0000O`4008T000250@00V000004000000?Y1
003j@Gl1002F00000@0004`00000000000000000003ooooooooooe00001e0000000002D0000<0000
3@00P4H0002T0000V00004E=AR/[@000300000000008@082?000030000014<3K000004h000000000
ZA=@?`80000200000P00008000014<3K000000000?l8@043?000030000014<3K100000000025Q:U3
>91S@eDEYT=j9ee3bI>Q@nalF4<Sfia3Vg=F@`030`<E@040400000@000020000900002@00000080m
000000000000080m0000000000020000G`0003P000030000>0000000000h0000000000000@0=0000
000000000000000000000000000U0000300000<0000U0000300000D0081E0000;00003P1003E0000
E@400>D000040000<ADi3/<Dd`dc58P=W1=X3BD0000<00001`00P2D0000<00000000P2@0000T0000
0020@@00000000000020@@00000000000P0002P0000<00000`0004H0001L0000D00004E=AR/8@043
?000030000014<3K10000000003b29e3JCAJ@`FaVD?ROUE3bIBM@kWhDT?b29e3JCAJ@`010H4D@060
400000@00000003o900002@00000080m000000000000080m00000000000200009@0000`000020000
4`0000`0000100009@0000`000080020EP0002`0000c0@00d`0003`1003K000010000:8CY0dg4eP=
/a<`3J8CY0dU0000300000L0080C000030000040000U000030000000080T000090000000P4400000
00000000P44000000000008000160000^0000:`00015CDH[:T0002@0000H0000G8nR@P0000000000
G8nR@VH;KD=^liA324001T@0000h00000A30fj^Z:Sh000000P000000000?0000E01904d0A@1C0200
CP1505L0801B04l0C@1104h0000f@020@00003@00000003o0@00004000010000MP2a<gDnPQmh?P00
P3l0000000000000P3l00000000000004P0000`000010000600000`0000000005P0000`0000H0000
9@0000`000010000E00005@000010@00<@4000H100100@000@000000nT400?Y10@4003h100010000
C00000000000000000000?ooooooooooD00007H0000000009@0000`0000=0020AP0007`0001`0000
ADe6:b]0000<0000000000Q00@=@0000A000004@`=/6000000000=WW^4=Ei[<ofW:e@fPLNd7IikQ3
ENJc?omZkd<SXeQ20OK[@hGUR4;oJ^m38j=H@P01P@01P@005401P10000040000ooooobP0000<0000
0P0002@0000T00000020?@00000000000020?@00000000000P0002L0000H00000P000000003oool0
000002D0000<00000P0001<0000<00000@0002D0000<00002000P5/000100000JP400040003O0@00
A@000080000600000`0000<0000M5aL0[aKl01dG5`3^7F<3Oae81>hMH`<U0000300000L0080C0000
30000040000U000030000000080T000090000000P440000000000000P44000000000008000160000
d0000<@00015CDH[24020S`0000`00000A30f`00001>000000000<JgRSh200000P00008000020000
0A30f`000000003o24010g@0001X00000A30f`X000000000fNNh@eGV/coJL[E3J1ak@MWW^4=Ei[<o
of[_@b>SF481m^]3QNF8@_mZkd<SXeQ2k37B@kJU3d8;b[U3Q[54@N`adT>fY@m2c9WZ@`/OKT800@40
0@400@01000E@040400000@000020000900002@00000080m000000000000080m0000000000020000
G`0003P000030000>0000000000h0000000000000@040000000000000000000000000000000U0000
300000<0000U0000300000D0081J0000F00006T100000000h04004H0000400002P0000<000030000
0P000080000M5aL0[aKl01dG5`3^7F<3Oae81>hMH`=76Sl2>QO504LJ?`9D7KT39@0000`000070020
9@0000`000000020900002@00000081100000000000008110000000000020000:00000`000030000
AP0005`0001@0000ADe6:`Q00@<l0000<000004@`=/4000000000?f;^D>dgfU1FO2f@oR>6476S;Y3
XXTW@Of;^D>dgfU10041PAA00H0@0000100000000?lX000030000080000T000090000000P3d00000
00000000P3d0000000000080000W0000600000800000000000000000000U000030000080000C0000
30000040000U0000300000P0081F0000;00006d100090000MP4000l000040000<QOZ0=lFV@1B5jP0
<QOZ02D0000<00001`00P1<0000<00000@0002D0000<00000000P2@0000T00000020@@0000000000
0020@@00000000000P0004H0001L0000D00004E=AR/8@043?000030000014<3K10000000003KenY3
Om=T@WmckD>^9gU24]OY@`AYMD;KenY3Om=T@P010H4D@060400000@00000003o900002@00000080m
000000000000080m00000000000200009@0000`0000200004`0000`0000100009@0000`000080020
EP0002`0003C0@00>@000=/1000o0000100005/MU0>_7ND3>agF0e/MU0<U0000300000L0080C0000
30000040000U000030000000080T000090000000P440000000000000P44000000000008000160000
d0000<@00015CDH[24020S`0000`00000A30f`00001>000000000<JgRSh200000P00008000020000
0A30f`000000003o24010g@0001X00000A30f`X0000000001D?[@n6[Hd=D<>53DJ=J@`E3jd?QZf=3
R5G6@kP`a4?G@[a3L:bo@hQEaT>h<<A3XBWD@cg_V4=@cNE36c5S@j4Ye4<mkiQ3lXG2@ne5`4<00@40
0@400@01000E@040400000@000020000900002@00000080m000000000000080m0000000000020000
G`0003P000030000>0000000000h0000000000000@040000000000000000000000000000000U0000
300000<0000U0000300000D0081J0000F00007L1003I0000f04008X1000400002P0000<000030000
0P000080001Y7C/>9ab[3FTM>`k;68LHRAOf5l/HQaR66QhC^Q`d3XHJ7Q=A60TH9@0000`000070020
9@0000`000000020900002@00000081100000000000008110000000000020000:00000`000030000
AP0005`0001@0000ADe6:`Q00@<l0000<000004@`=/400000000032Ri4=PaV93H:3V@k>3Gd<R_>I3
cZQT@c2Ri4=PaV930041PAA00H0@0000100000000?lT000090000000P3d0000000000000P3d00000
00000080000U000030000080000C000030000040000U0000300000P0081F0000;0000<T1003O0000
cP400>D000040000UA`]3]DLn@gH74/>UA`]3RD0000<00001`00P1<0000<00000@0002D0000<0000
0000P2@0000T00000020@@00000000000020@@00000000000P0004H0001L0000D00004E=AR/8@043
?000030000014<3K10000000000B/L=3Bg_0@n:b`D>Q7<9389O1@aB:_d<B/L=3Bg_0@`010H4D@060
400000@00000003o900002@00000080m000000000000080m00000000000200009@0000`000020000
4`0000`0000100009@0000`000080020EP0002`000230@00O`4008P100250@00100007LH41Pg64@H
<aSb5gLH41PU0000300000L0080C000030000040000U000030000000080T000090000000P4400000
00000000P44000000000008000160000R00007`00015CDH[24020S`0000`00000A30f`00001>0000
00000:TCD3l200000P000080000200000A30f`000000003o24010b`0000P00000A30f`8000000000
CI;A@dH<Cd>Q<>Y3B0>m@P01000E@040400000@000020000900002@00000080m000000000000080m
0000000000020000G`0003P000030000>0000000000h0000000000000@0=00000000000000000000
00000000000U0000300000<0000U0000300000D0081G000090000:41001M0000eP400=4000020000
<a[a34LMj@DU0000300000L0080U000030000000080T000090000000P440000000000000P4400000
00000080000X0000300000<000160000V00008`00015CDH[24020S`0000`00000A30f`00001>0000
00000:TCD3l200000P000080000200000A30f`000000003o24010c`0000`00000A30f`@000000000
]Kc=@`eRbd94cmI3Zfg>@_OFgd<<3n=2mBOS@d9Ao4800`<35D010100000400000P0002@0000T0000
0020?@00000000000020?@00000000000P0005l0000h00000`0003P000000000>000000000000040
3@0000000000000000000000000000009@0000`0000300009@0000`000050020E@0002`0002J0@00
I0000<P10020000010000;PIG0KJ6W@6na/I1fDLh`LU0000300000L0080U000030000000080T0000
90000000P440000000000000P440000000000080000X0000300000<000160000G00005000015CDH[
24010c`0000`00000A30f`@000000000Z2K>@dRmdT:SRLY3W1C;@/TlcT?b7<A2Z2K>@dRmdT800@61
5401P100000400000000ob@0000T00000020?@00000000000020?@00000000000P0002D0000<0000
0P0001<0000<00000@0002D0000<00002000P5H0000/0000U@400680002M0@00JP0000@000356IH6
DQUI1/PI8@K56IH69@0000`0000700204`0000`0000100009@0000`000000020900002@000000811
00000000000008110000000000020000AP000:@0002H0000ADe6:bY0000T0000600005b?XT800000
000005b?XT:i7J53S_j:@PQ000H`00009000004@`=^[ZRXn00000000000000001P0005<0F@1=0480
C`1<03I008100000=00000000?l100000@000040001Tl5c>]Cn27gPn0020?`00000000000020?`00
00000000000B000030000040000H000030000000000F0000300001P0000X000030000040001B0000
L0400040003boooo0000000000000000T0400000008700@0D`1i06d0HP1_06`0000004d0K`1^06l0
M01i0700I@0j05<0N@1]0680K`1/0200>P1f06D0LP1c06T0K`1^0000jGM@lA80B3?YM`/0000Z0000
C?4B0:03kWL00000:P0005`0CP1<05<0G01>06`0L`1C06D0H`1d06T0K`1^04<0D00d03800028kA80
000C00@0000000000@0000000001kQ80S>hB00c]4P2PVB`ce>dB09D[n7LH=_Qg003oonC]4P22LnQg
000C0000003XHQlcYPD:3^QR7c?0EoMg000000`X50000000_QX0H;3]4P2PVB`ca?LB0000jWN`MNQg
ooooo`40001?OOAgj68O<`P0003P=PDc8>hB0>QR7c?41oEg`5OgMiP58El100003CcdMl1GmgMTMP08
000002D0000<00000@0005@0001D0000]P4004/0002l0@00F`00004000000?Y1003j@KH1001I0000
0@0004`00000000000000000003ooooooooooe00001Tl000000002D0000<00003@00P4H0002D0000
R00004E=AR/[@000300000000008@082?000030000014<3K000004h000000000ZA=@?`8000020000
0P00008000014<3K000000000?l8@043;000020000014<3K0P000000002ElLU3mO<i@iGabD=80ke2
004001E00@0@000010000080000T000090000000P3d0000000000000P3d0000000000080001O0000
>00000<0000h0000000003P000000000000100d0000000000000000000000000000002D0000<0000
0`0002D0000<00001@00P5L0000T0000TP4005d0002F0@00_0000080000o6J0;?aWY1BD0000<0000
1`00P2D0000<00000000P2@0000T00000020@@00000000000020@@00000000000P0002P0000<0000
0`0004H0000H200030P004E=AR/Q@0L030000000000U@000400000@000000000;40002@0000H0000
9XTI?@0008000020fKK<?5W2Yd?O?7Y324031I`7002@1`000A30f`8000030000P0L00040001/0000
:00000L0003P0@00:@4000000000000004000:XZ000PADe600010807000V00000`00000000000000
0000000400000`00@0400?0000000000000000000000hP@0P:T301T0000<0000oooo0180000<0000
0P0002D0000<00000000P2D0000<00003@00P2D0000<00001`00P1P0000<0000000001H0000<0000
00000240000800007P0001P000000000000000`2001N0@004@0000`0000800003000010000000000
000000/0000@0000308005h1000H000030000000000B000030000040000I000030000?ooo`0F0000
300001P0000:000040000000000000002@00010000000`0000800580001`0@000@00083nool00000
00000000002@0@000@0000@2011D06T0K@1U07<0801>06D0M`0P0580K`1]0640KP00000000000000
00000000000000000000000000000000000Y<j3Z4P203P00G>`B09D[n7Nh>OQgooooofc/4P0W/_ag
60LC0000002P0`4f000004H0E01G0440DP1505`0@`1/0640L`1c06D0L`1L04<0C01C04T0A01L07/0
<00`0300<P1304D0<00c02d0<00`0300<00]0300<0000300;@130300<00`02d0<00`0300<00`0300
<00`0300<00d03H0O@1L04d0J@1c06<0D`1d0640M01e07<0001m0000YGL00:4f00000>3[4P2mijEg
0P03P:RT;S=/3JIg`USdM`T0003P=PDcj>/B0:ZZ9c?01_Eg0@0006Af00P000009@0000`000010000
E00005@000190@001`000>01000Y0@000@000:^BZT40iZY1l040060100010000C000000000000000
00000?ooooooooooD00004X0002Z0000E00005@0000X00001`000>@0000Y0@000@000:^BZT40iZY1
>P00060100010000C00000000000000000000?ooooooooooD00006d0000E0@00DP00070100020000
P?koo`00000000000000090100000000108045@0J@1]06D0L`0P04h0I@1g0200DP1_06d0H@1^0000
0000000000000000000000000000000000000000000007<0001m0000YGL00:4f00000>3[4P2mijEg
0P03P:RT;S=/3JIg`USdM`T0003P=PDcj>/B0:ZZ9c?01_Eg0@0006Af00P00000I@1c05`0@`1<05<0
B@1405`0N`0`0300<00b04<0A@0`03<0;@0`0300<00`02d0<00`0000<00]04<0<00`0300;@0`0300
<00`0300<00`0300<00`03@0=P1m05`0C@1Y07<0H`1C07@0H@1d07D0L`0007d0002UM`00XCH00000
h>/B0;gWYGL200>0Z:@^<bD0003P=PDcb>/B0?<5002G3?Eg0@00000000010000I7H02000000U0000
30000080000X000030000040001D0000E0000>@0000700009@4002T100010000Zi:Z@@3VZT5>0@00
H0400040001<00000000000000000000oooooooooom@0000;0000640001B0000L0400040000@0000
1`00000000000000_080000000010P8RD`1i07<0M01U06d000000000000000000000000000000000
0000000000000000000000000000000000000000:C?`]2Tc002UM`00XCH00000h>/B0;gWYGL200>0
Z:@^<bD0003P=PDcb>/B0?<5002G3?Eg0@00000000010000I7H02000001U07<0G01304`0D`0001<0
10000000<00`0380@`150040<`0]0300Z>XB0:P51C=`ja80003hMaPfn7OoooooP>/B089cj7L001<0
00000?2d:C?c1@YYl;@Y<l1GmgM=06T0L`1S05<0M01Q07@0C>/B00001C<lkA80EQSZMk1ej7Oooooo
0@0004mmm7O`]2Tc20000>0f1C>lja80l;@Y<l@7mGO0EoMg00000040001TMP08000002D0000<0000
0@0002P0000<00000P0002D0000<00001`00P2D0000<00000000P300000<00003`00P2D0000<0000
3@00P4/000100000<00000D0000P00000@000040000@000000000000000<0P00GP40000000000000
308005h1000X000030000040000R000030000?ooool>000050000000000@0000500001Y00d0/0000
80000?ooool20000000000000000@0=4082_@`00003X0nP38@0000P0001R000030000040000Q0000
2000024000080000L`0000`000000000700000P0000U000030000000080U0000300000L0080U0000
300000h0080`0000300000l0080I000030000?ooo`0H000030000000000B000030000080000C0000
30000040000D0000300000d0000E000030000040000F000030000000000=00004000000000000000
>P0000`0000:00006`00010000000000000002<0000P00000:2<?@0000000000<Z:<?@00Z4<007]3
900002@00000X8`m00000000000bXX`m002X@`00Nd<40000L`0000`0000000003@000100001@0@00
n`000580001`0@000`0001000007000000000000002l0P00000000L20R9C07T0L`1d06D0K@000000
0000000000000000000000000000000000000000000000000000000000000000000002X0003HjA80
=3?YMmSY4P18<nUg2`0002X0003DjA80X0?^McL=jGLXjQ8010000=SU4P000000000002X0000H0000
000001SZ4P10000000000000000/jQ80=3?YMbcZ4P18<nUg2`0003SV4P050000L3WhM`00C00004`0
1@00092oUcD@iQ80P3`00<cW4P2E:oQg^3WhMooooooLia809k;lMgP7C0000000oooooal@07P10000
000003cW4P1Xia80G2TD07cV4P0W?01P324D06SW4P0000000@000;CV4P33?01P324D06SW4P000000
32PD06Af00P000009@0000`000030000AP0002P0000L0000AdA9@`8000000000000000`2001N0@00
000001T0000<0000oooo0180000<00000P0002D0000<00000000P580001`0@001000010000070000
00000000002l0P00000000L20R9C07T0L`1d06D0K@00000000000000000000000000000000000000
00000000000000000000000000000000000B000000000000:P0001P0000000006>XB040000000000
000002cZ4P0d<nUg;>XB04PcjGL;0000>>HB00D0001`>OQg001<0000C0050000T;nG=A3V4P20?000
c>LB09D[n7Nh>OQgooooomcW4P0W/_agN0M<000000000?oo7a00N04000000000?>LB06SW4P1L:A@0
O>HB02Ll060<8A@0J>LB0000000U0000h3H5<f3V4P0@0000003eM`40000I0;01402@0El9mGNH1B5O
0000P2D0000@0901ool001000028iQ80nQgdMiP58El@09019@0006Af00P000009@0000`000040000
9@0000`000070020600000`0000000005P0000`0000000008@0000P0000N00006000000000000000
308005h1000S00008000002PS3d00000000003:RS3d00:Q3001k@b@0000T00000:2<?@0000000000
<Z:<?@00Z4<007]3100002<0000P00000:2<?@0000000000<Z:<?@00Z4<007]3900002@00000X8`m
00000000000bXX`m002X@`00Nd<400008`0002000030j`m2000000000030AL11002X@`00Nd<T0000
90000<3[3d80000000000<15`4400:Q3001k@`@0000H000030000000000B000030000040000I0000
30000?ooo`0F0000300001P0000S000080000<3[3d80000000000<15`4400:Q3001k@b@0000T0000
`>/?@P0000000000`4G0@@00Z4<007]3100002<0000P00000>Do?@0000000000`4E0?@00Z4<007]3
900002@00000iClm000000000030AD0m002X@`00Nd<40000DP00070100050000P?koo`0000000000
00000901000100001`8045@0J@1]06D0L`0P04h0I@1g0200DP1_06d0H@1^00000000000000000000
000000000000000000000000000002TcX>XB080>001Lk180UB_hMkPin7OoooooK>`B02Nbo7LH1a<0
00000:030CH00000AP1D05L0@@1B04D0G01306`0H@1c07<0I@1c05`0@`1<05<0B@1405`0N`0`0300
<00b04<0A@0`03<0;@0`0300<00`02d0<00`0000<00]04<0<00`0300;@0`0300<00`0300<00`0300
<00`03@0=P1m05`0C@1Y07<0H`1C07@0H@1d07D0L`0007d0002UM`00XCH00000h>/B0;gWYGL200>0
Z:@^<f`=YWO2F?Ag2@000>0f1C?Xja80ZZXW<l06mGL10000I7H02000000U0000300000D0001D0000
E00006L1003l0000K`4000l100010000Zi:Z@@3VZT7`0@00H0400040001<00000000000000000000
oooooooooom@0000BP3oojX0001D0000E00005<1003l0000G`4000l100010000Zi:Z@@3VZT4j0000
H0400040001<00000000000000000000oooooooooom@0000K@0001D1001B0000L04000H00020o_oo
0000000000000000T040000000070P0@E01Y06d0I@1c0200CP1U07L0801B06l0K@1Q06h000000000
0000000000000000000000000000000000000000L`0007d0002UM`00XCH00000h>/B0;gWYGL200>0
Z:@^<f`=YWO2F?Ag2@000>0f1C?Xja80ZZXW<l06mGL10000I7H02000001U07<0G01304`0D`1904@0
G01k0300<00`0380@`150300<`0]0300<00`0300;@0`0300000`02d0@`0`0300<00]0300<00`0300
<00`0300<00`0300=00f07d0G01=06T0L`1S05<0M01Q07@0M@1c0000O@000:Eg002Q=P00003Pja80
_NNUM`800h2XY2hc9@000>0f1C?8ja80l`D009L<mGL1000000000040001TMP08000002D0000<0000
1P0002P0000<00001@0005@0001D0000H0400?`0001T0@003`400040002[TZY10>JZ@Dh1001P0@00
0@0004`00000000000000000003ooooooooooe00000/0000H@000580001`0@001@00010000070000
00000000002l0P00000000L20R9C07T0L`1d06D0K@00000000000000000000000000000000000000
00000000000000000000000000000000000Y<o2d:C<00:Eg002Q=P00003Pja80_NNUM`800h2XY2hc
9@000>0f1C?8ja80l`D009L<mGL1000000000040001TMP08000006D0L`1L04<0C01C00004`040000
000`0300<P1304D00@0c02d0<02XjQ80Z0D5<g3[4P000?Qg63KhMooooon0ja80PW?XM`004`000000
l;@Y<o<52VW`]2Tc`5OgMdd0J@1c06<0D`1d0640M01<ja800005<cc]4P1F6>Yg/7GXMoooool10000
CggdMo2d:C<80000h3H5<kc[4P3`]2Tca0OeMl1GmgL000000@0006Af00P000009@0000`000050000
:00000`0000600009@0000`0000700209@0000`0000000209@0000`000040000B`000400000`0000
1@000200000100000@000100001@0@00n`0007@1000C0@00D0400?/0001d0@004`4002P0000<0000
1@000280000<0000ooooodH0000D0000200004M4BD<300008P0000`0003ooooo8P0000`0003ooooo
9@0000`0000=0020:00000`000030000:00000`0000400008P0000`0003oooooAP0009P0002<0000
ADe6:bI0000@0000100000000008@082?000030000014<3K000004h000000000ZA=@?`8000020000
0P00008000014<3K000000000?l8@043;000020000014<3K0P000000000g4713C@B:@hQo/D?^TJA3
004001E00@0@000010000080000T000090000000P3d0000000000000P3d0000000000080001O0000
>00000@0000h0000000003P000000000000100d0000000000000000000000000000002D0000<0000
100002D0000<00001@00P5L0000T0000kP000181001U0@00B`40008000013d4A<1JC52D0000<0000
1`00P2D0000<00000000P2@0000T00000020@@00000000000020@@00000000000P0002P0000<0000
100004H0003T0000f00004E=AR/8@043a0000;P000014<3K4`0000000021LeU3caVJ@kWWND?nmVe3
BKQj@i/cK4<NZgm3k<1/@j9jPT>P<Vm3]An5@eBTLD<<l8I3N`ee@l>7QT?Nd7I3`hN6@mk@MT?3QhI3
g]1f@l>7QT?Nd7I3a8N6@mk@MT=@Vfa3`8JN@kg:Jd=aJ9m3j=MV@lPQWd?2SF53k^RM@ia3G4<D/9a3
kj9H@h3kVT>0LeU3caVJ@`010`<30`<30`<30@430`<30h<05401P10000040000`<30obP0000<0000
0P0002@0000T00000020?@00000000000020?@00000000000P0002L0000H00000P0000000030`<00
000002D0000<00000P0001<0000<00000@0003/0000800006`000100002H3@00A1<003H0000@0000
W`l00>0>001H0000@000000000000000ooooooooool90000[0o43_/?c@i@4?@>Y10K3mh@D@oA46h?
dA1^3m4@KPoA46h?F@0002@00000000000000?oooooooooo0P000=4@KPo:3]4CF00003@000000000
00000?oooooooooo1P000;d>kQ=^3^DC6@jn4lD=UQ>;3F0CV0e44cd000080000?00000P0000n0000
60000=P0003/00003P4003l1000C000030000040000U000030000000080T000090000000P4400000
00000000P44000000000008000160000G00005000015CDH[24020S`0000`00000A30f`00001>0000
00000<JgRSh200000P000080000200000A30f`000000003o5D010100000400000P0002@0000T0000
0020?@00000000000020?@00000000000P0005l0000h0000100003P000000000>000000000000040
100000000000000000000000000000009@0000`0000400009@0000`000050020>`0000P0000K0000
400009P=00144`00=P000100002O3`00h0h005P00010000000000000003oooooooooo`T0002/3l@>
n`o=3U0@m0jT41/?gQ1A3m4@KPoA46h?dA1^3m4@KPmI00009000000000000000ooooooooool20000
dA1^3lX>dA=H0000=000000000000000ooooooooool60000_@k^4fh>iA<I3[hCa@fF4h/=H1>H3D@C
?@0000P0000l000020000400000H0000e`000>/0000?0@00@04002D0000<00001`00P2D0000<0000
0000P2@0000T00000020@@00000000000020@@00000000000P0002P0000<0000100004H0003T0000
f00004E=AR/8@043a0000;P000014<3K4`000000003Poj93c2fc@o`i/d=k3i13A::c@lT]Sd>_6kI3
LWB?@l;0^4=<[I13eFFk@bKVTD</=[e3^IZC@n?=_4=[O9A3hlfl@f]lU4?ScKa3JgbD@n?=_4=[O9A3
hlfl@falU4?8Tja3_9Zg@gh[[4=]O;Q35;:Y@l@e^4<13JM3j_bf@niWY4<@a;E3UiNR@g`?]4?Poj93
bbfc@`010`<30`<30`<30@430`<30h<05401P10000040000`<30ob@0000T00000020?@0000000000
0020?@00000000000P0002D0000<00000P0001<0000<00000@0003/0000800006`000100001P5000
IQH003H0000@0000J1H0008B001H0000@000000000000000ooooooooool90000MAKV4L@Fka4I5aHB
KALm4ZLGM1:J5i0BVQN@4YXGT1:J5i0BF@0002@00000000000000?oooooooooo0P0009XGT1:C5O@F
F00003@00000000000000?oooooooooo1P0008HE41Lg5@LGhQCP5XdD^AIC588FH1AV5Sd000080000
?00000P0000n0000600004D1000N0@00N`400741000C000030000040000U000030000000080T0000
90000000P440000000000000P44000000000008000160000G00005000015CDH[24020S`0000`0000
0A30f`00001>000000000<JgRSh200000P000080000200000A30f`000000003o5D01010000040000
0P0002@0000T00000020?@00000000000020?@00000000000P0005l0000h0000100003P000000000
>000000000000040100000000000000000000000000000009@0000`0000400009@0000`000050020
>`0000P0000K00004000060D001V5P00=P000100001X5P000Q8005P00010000000000000003ooooo
ooooo`T0001e5^HAa1K_4ATG5Q9]5cdBYaMd4YXGT1:J5i0BVQN@4YXGT19I00009000000000000000
ooooooooool20000VQN@4Y<Em1IH0000=000000000000000ooooooooool60000QQD@5cLE1aOR5>0F
SABi5U<DPQIP56HF?@0000P0000l000020000400000H0000A04001d1001l0@00L`4002D0000<0000
1`00P2D0000<00000000P2@0000T00000020@@00000000000020@@00000000000P0002P0000<0000
100004H000280000O00004E=AR/8@082?000030000014<3K000004h000000000ZA=@?`8000020000
0P00008000014<3K000000000?l8@043;000020000014<3K0P000000002dOd]3V^2a@dLgWd<lK/a3
004001E00@0@000010000080000T000090000000P3d0000000000000P3d0000000000080001O0000
>00000@0000h0000000003P000000000000100d0000000000000000000000000000002D0000<0000
100002D0000<00001@00P5L0000T0000bP00068100100@00V`400080002h33dFia>>6BD0000<0000
1`00P2D0000<00000000P2@0000T00000020@@00000000000020@@00000000000P0002P0000<0000
100004H0003T0000f00004E=AR/8@043a0000;P000014<3K4`000000003nhSA37OK1@cIGED?<eii3
aRMF@a[fWD>K6U]3`cbN@l5TH4>MMIm3ijiU@gN^X4>ECfU32V>R@`AoJ4>lA:=317mX@ka4Xd<4OfQ3
_4BS@`AoJ4>lA:=31GmX@ka4Xd?=2TQ33F?6@c/jAd>nA<M3ITM2@aGnaT=0oCa3>lG5@aZc=d=QS<A3
KA8d@lgG`T?nhSA37?K1@`010`<30`<30`<30@430`<30h<05401P10000040000`<30ob@0000T0000
0020?@00000000000020?@00000000000P0002D0000<00000P0001<0000<00000@0003/000080000
6`000100001?2`00?aP003H0000@0000EPd00=/C001H0000@000000000000000ooooooooool90000
H`fo4k8=b1<73^lCF`hF59D>CAB83VTDR0iY58P>JAB83VTDF@0002@00000000000000?oooooooooo
0P0008P>JAB13<dHF00003@00000000000000?oooooooooo1P0007@<jAPU3>0Hd0^i67`;TQQ22e/H
C`/o63d000080000?00000P0000n000060000;@0000k0@00jP0008l1000C000030000040000U0000
30000000080T000090000000P440000000000000P44000000000008000160000G00005000015CDH[
24020S`0000`00000A30f`00001>000000000<JgRSh200000P000080000200000A30f`000000003o
5D010100000400000P0002@0000T00000020?@00000000000020?@00000000000P0005l0000h0000
100003P000000000>000000000000040100000000000000000000000000000009@0000`000040000
9@0000`000050020>`0000P0000K0000400004l;000o6000=P000100001F3@00fa<005P000100000
00000000003oooooooooo`T0001S3KlC/Pg84`L>ka=K3QHDU@i=58P>JAB83VTDR0iY58P>JAAI0000
9000000000000000ooooooooool20000R0iY584<cAQH0000=000000000000000ooooooooool60000
M0cY62D<h1S@2kTHO0^B648;FaQ?2clH?@0000P0000l000020000400000H0000/`0003X1003[0000
T04002D0000<00001`00P2D0000<00000000P2@0000T00000020@@00000000000020@@0000000000
0P0002P0000<0000100004H0003T0000f00004E=AR/8@043a0000;P000014<3K4`000000002O]i13
6@[K@k_aX4?8jkM30UZQ@aL:]d=]dj=3`52g@h1hYT>JRKQ3TafY@gC2^D?ZkJY31gNk@j65ZT>iF;a3
XHFZ@kUH_4>QQJY3^ERl@j65ZT>iF;a3XXFZ@kUH_4>6BiY32GOO@ccSVD>jF>13dVVG@a4Bh4>oa9A3
=mWN@j`OTT=MX=e3EDn@@lW[fd>N]i1360[K@`010`<30`<30`<30@430`<30h<05401P10000040000
`<30ob@0000T00000020?@00000000000020?@00000000000P0002D0000<00000P0001<0000<0000
0@0003/0000800006`000100000G4P00HQ/003H0000@00007a@00?hF001H0000@000000000000000
ooooooooool90000;1CR5W/DjaK@518G91Di5ehEKaMA5H`GDAF<5e4ES1MA5H`GF@0002@000000000
00000?oooooooooo0P00054ES1M:4nlKF00003@00000000000000?oooooooooo1P0003dC31c^4P<L
VA;L6d@B]A/:4WhK5a9R6cd000080000?00000P0000n000060000201001^0@00EP400<41000C0000
30000040000U000030000000080T000090000000P440000000000000P44000000000008000160000
G00005000015CDH[24020S`0000`00000A30f`00001>000000000<JgRSh200000P00008000020000
0A30f`000000003o5D010100000400000P0002@0000T00000020?@00000000000020?@0000000000
0P0005l0000h0000100003P000000000>00000000000004010000000000000000000000000000000
9@0000`0000400009@0000`000050020>`0000P0000K0000400001LB001R6`00=P000100000O5000
oQH005P00010000000000000003oooooooooo`T0000/5>8FNaC[5]0D4QLT5CTGGQE_5e4ES1MA5H`G
DAF<5e4ES1MI00009000000000000000ooooooooool20000DAF<5dXCka]H0000=000000000000000
ooooooooool60000?A<<7>hB0abI4]`KA1:e6`XBOQ/G4V8K?@0000P0000l000020000400000H0000
7`4006d1001G0@00`P4002D0000<00001`00P2D0000<00000000P2@0000T00000020@@0000000000
0020@@00000000000P0002P0000<0000100004H000280000O00004E=AR/8@082?000030000014<3K
000004h000000000]fF]?`80000200000P00008000014<3K000000000?l8@043;000020000014<3K
0P000000002TjCQ3]939@l2d9T=Xc]i3004001E00@0@000010000080000T000090000000P3d00000
00000000P3d0000000000080001O0000>00000@0000h0000000003P000000000000101H000000000
0000000000000000000002D0000<0000100002D0000<00001@00P5L0000T0000Y@000941002k0000
`0400080002?2c<IK0[J6bD0000<00001`00P2D0000<00000000P2@0000T00000020@@0000000000
0020@@00000000000P0002P0000<0000100004H0001L0000D00004E=AR/8@043?000030000014<3K
10000000002QXcA3NSS9@ko>>d<E<<I3j5/l@oOObT>QXcA3NSS9@`010H4D@060400000@00000003o
:00000`000020000900002@00000080m000000000000080m00000000000200009`0001P000020000
00000000000000009@0000`0000200004`0000`0000100009@0000`000080020EP0002`0002d0000
S0400;d0002F0@00100004/;:1Vm2lHHaP]L6D/;:1TU0000300000L0080C000030000040000U0000
30000000080T000090000000P440000000000000P44000000000008000160000^0L00:`70015CDH[
9D00010000040000000002a0000T0000600005PmYS`000200000P2QliSa5o2a3NLOG@`Q010E81`00
?0L0004@`=/200000`0002`700010000K00003P0000V0000=0400>4100000000000002T^0000@000
84E=AP000@0/1`009@0000<000000000000000000000100000<00401003`00000000000000000000
0>84082Y0`0I000030000?ooo`0B000030000080000U000030000000080U0000300000d0080U0000
300000L0080H000030000000000F000030000000000Q0000200001h0000H000000000000001j0@00
30800140000<0000200000`0000@000000000000000;0000400007X1000<0P00600000`000000000
4P0000`0000100006@0000`0003oool05P0000`0000H00002P00010000000000000000T0000@0000
X0400402001B0000L0400040000Poooo0000000000000000T040004000040P0@E01Y06d0I@1c0200
CP1U07L0801B06l0K@1Q06h00000000000000000000000000000000000000000000000007c<`jQ80
P0h00>c[4P2E:oQg^3WhMoooooolja809k;lMaP74`000000d07j=@00001605@0E`110580A@1L04<0
K01Q07<0L`1U07<0G01304`0D`1904@0G01k0300<00`0380@`150300<`0]0300<00`0300;@0`0300
000`02d0@`0`0300<00]0300<00`0300<00`0300<00`0300=00f07d0G01=06T0L`1S05<0M01Q07@0
M@1c0000O@000:Eg002R=P00001`ja80_NNUM`800h3h=2lcK0fVMl9Hm7L90000h3H5<gS[4P22kAlc
`0KeM`40001TMP08000002D0000<00000@0005@0001D0000n0000?h0000d0@00h@400040003I:N=1
>FkS@A41003P0@000@0004`00000000000000000003ooooooooooe00001/0000?`000580001`0@00
0P00083nool0000000000000002@0@000@0000@2011D06T0K@1U07<0801>06D0M`0P0580K`1]0640
KP0000000000000000000000000000000000000000000000001c0000O@000:Eg002R=P00001`ja80
_NNUM`800h3h=2lcK0fVMl9Hm7L90000h3H5<gS[4P22kAlc`0KeM`40001TMP08000006D0L`1L04<0
C01C04T0A01L07/0<00`0300<P1304D0<00c02d0<00`0300<00]0300<0000300;@130300<00`02d0
<00`0300<00`0300<00`0300<00d03H0O@1L04d0J@1c06<0D`1d0640M01e07<0001m0000YGL00:8f
0000073[4P2mijEg0P03P?Pd;c<U0000h3H5<eS[4P3@1@00U`ceM`40000000000@0006Af00P00000
9@0000`000020000:00000`000010000E00005@0000h00009P0002P1002X0@000@000=TYhd4iK^=1
AP00080100010000C00000000000000000000?ooooooooooD00004H0003Z0000DP00070100010000
400000L00000000000000;`2000000000@828U<0N@1c07@0I@1]0000000000000000000000000000
0000000000000000000000000000000000000000000002Tc4;DY<`00YGL00:8f0000073[4P2mijEg
0P03P?Pd;c<U0000h3H5<eS[4P3@1@00U`ceM`40000000000@0006Af00P00000I@1c05`0@`1<05<0
000C00@000000300<00b04<0A@0103<0;@0`03SZ4P1X2ah00>/B0000n7LH=_Qgoooooa3[4P22LnQg
000C0000000@]BTcd0D:g12e:C?0EoMgC@1Y07<0H`1C07@0H@1d0=cZ4P0001h0c>`B05HHjWN`MNQg
ooooo`40001?OOAg4;DY<`P0003P=PDcC>/B012e:C?41oEg`5OgM`0000010000I7H02000000U0000
30000040000X000030000080000U0000300000L0080U000030000000080`0000300000l0080U0000
300000d0081;0000@00003000005000080000040000100004000000000000000NP4000`200000000
000007X1000<0P00:00000`0000100008P0000`0003ooooo3P0001@000000000400001@0000J@0A0
;0000200003ooooo0P00000000000000082m@`100d@00000j0?X0b4000080000HP0000`000010000
8@0000P0000Q0000200007<0000<0000000001`0000800009@0000`0000000209@0000`000070020
9@0000`0000>0020<00000`0000?00206@0000`0003oool0600000`0000000004P0000`000020000
4`0000`000010000500000`0000=00005@0000`0000100005P0000`0000000003@00010000000000
000003X0000<00002P0001/0000@000000000000000S0000800001FSF3d000000000003`DSd002e3
003H@b@0000T00005J=H?@00000000000?1B?@00;D<00=Q3100007<0000<0000000000d0000@0000
[@000;01001B0000L04000@0000@00001`00000000000000_080000000070P8RD`1i07<0M01U06d0
0000000000000000000000000000000000000000000000000000000000000000000000004P000000
iCShM`004`08J34c0000033U4P281Q<0i>DB09D[n7LP>OQgooooooCU4P10]oagB0LC011X<C?c0@[8
46Pa<a1X<C>lja80JbD0H5Be1`4P/`L1G2TD03HM060P/`L1G2TD00PN061L:A@000000000000F@`1P
000D022c1`4<8A@0`61>H=SU4P0P/`L14480H00000010000JbD0H03V4P050000L3WhM`00C01h1d`0
1@000000UcGHiA80P3`009CW4P2E:oQg^3WhMooooonTia809k;lMgP7C0000000oooooal@07P10000
000000CW4P1TMP08000002D0000<0000100004H0000X0000700004M4BD<2000000000000001j0@00
30800000000I000030000?ooo`0B000030000080000U000030000000081B0000L04000<0000@0000
1`00000000000000_080000000070P8RD`1i07<0M01U06d000000000000000000000000000000000
0000000000000000000000000000000000000000o7M81a<046Pa<o<12/P@J34c46Pa<kc[4P1[9@1P
E;D70B2c1`5L:A@0=Qd0H22c1`5L:A@021h0H5`Y50000000000001I3060001@08;<70@`Q5030H4iP
f>DB022c1`4@@P1P00000040001[9@1P0>HB00D0001`>OQg001<07P7C0050000002G=MSU4P20?000
U>LB09D[n7Nh>OQgooooojCW4P0W/_ag9@000>0f1C?`iA8040000000mGL100006@2`0A00T05O2OEg
V0DQG`00080U0000402@0Ooo000@00006>HB0?XMm7NH1B5O402@0BD0001TMP08000002D0000<0000
0`0002D0000<00001`00P1P0000<0000000001H0000<000000000240000800007P0001P000000000
000007X1000<0P008`000200000EXePm000000000000l58m000]@`00f4<T0000900001FSF3d00000
0000003`DSd002e3003H@`@0000S0000800001FSF3d000000000003`DSd002e3003H@b@0000T0000
5J=H?@00000000000?1B?@00;D<00=Q3100002<0000P0000Io2O@@0000000000X>7G@@00;D<00=Q3
900002@0001Wl9m100000000002PhMM1000]@`00f4<40000600000`0000000004P0000`000010000
6@0000`0003oool05P0000`0000H00008`000200001Wl9m100000000002PhMM1000]@`00f4<T0000
900006O`Wd40000000000:3Qed4002e3003H@`@0000S0000800001cIA3d000000000003U?cd002e3
003H@b@0000T00007=U4?@00000000000>Do?@00;D<00=Q310000580001`0@001@00023oool00000
00000000002@0@000@0000L2011D06T0K@1U07<0801>06D0M`0P0580K`1]0640KP00000000000000
00000000000000000000000000000000000O<c3Z4P203P00k>/B09D[n7Nh>OQgooooooc[4P0W/_ag
60LC0000003@0OXe000004H0E01G0440DP1505`0@`1/0640L`1c06D0L`1L04<0C01C04T0A01L07/0
<00`0300<P1304D0<00c02d0<00`0300<00]0300<0000300;@130300<00`02d0<00`0300<00`0300
<00`0300<00d03H0O@1L04d0J@1c06<0D`1d0640M01e07<0001m0000YGL00:8f0000073[4P2mijEg
0P03P?Pd;c=/3JIg`USdM`T0003P=PDcN>/B08;]7c?01_Eg0@0006Af00P000009@0000`000050000
E00005@0002j0000_@400;`000380@000@000=TYhd4iK^=14@400>0100010000C000000000000000
00000?ooooooooooD00006`0000o0000DP00070100060000P?koo`00000000000000090100010000
1`8045@0J@1]06D0L`0P04h0I@1g0200DP1_06d0H@1^000000000000000000000000000000000000
00000000000007<0001m0000YGL00:8f0000073[4P2mijEg0P03P?Pd;c=/3JIg`USdM`T0003P=PDc
N>/B08;]7c?01_Eg0@0006Af00P00000I@1c05`0@`1<05<0B@1405`0N`0`0300<00b04<0A@0`03<0
;@0`0300<00`02d0<00`0000<00]04<0<00`0300;@0`0300<00`0300<00`0300<00`03@0=P1m05`0
C@1Y07<0H`1C07@0H@1d07D0L`0007d0002UM`00XSH00000L>/B0;gWYGL200>0n3@_<bD0003P=PDc
F>/B0=05002G3?Eg0@00000000010000I7H02000000U0000300000H0000X0000300000D0001D0000
E0000;00002b0@00^`000<D100010000fBWS@CU^hd560000P0400040001<00000000000000000000
oooooooooom@0000AP000>X0001B0000L04000D0000@00001`00000000000000_080000000070P8R
D`1i07<0M01U06d00000000000000000000000000000000000000000000000000000000000000000
00000000:C<@]BTc002UM`00XSH00000L>/B0;gWYGL200>0n3@_<bD0003P=PDcF>/B0=05002G3?Eg
0@00000000010000I7H02000001U07<0G01304`0D`0001<010000000<00`0380@`150040<`0]0300
>>XB06P;7P00ja80003hMaPfn7Oooooo4>/B089cj7L001<00000012e:C?@1@[L4;DY<l1GmgM=06T0
L`1S05<0M01Q07@0g>XB00007P3<k180EQSZMk1ej7Oooooo0@0004mmm7L@]BTc20000>0f1C=<ja80
4;DY<l@7mGO0EoMg00000040001TMP08000002D0000<00001@0002P0000<00001P0002D0000<0000
1`00P2D0000<00000000P2D0000<00000`0004/000100000<00000D0000P00000@000040000@0000
[@000;0100310000b`400:d0002`0@00`@000</1000X0000300000D0000R000030000?oooom60000
500000P00017A4U30`000280000<0000ooooob80000<0000ooooobD0000<00003@00P2P0000<0000
100002P0000<00000`000280000<0000ooooodH000381`00_0L004E=AR/V@000400000@000000000
9D00010000040000000002a0000T0000600008Wa/S`000200000P19[fCccLY=3SDO]@`Q01@E81`00
?0L0004@`=/200000`0002`700010000K00003/0000X0000HP400?`10000000000000:`d0000@000
84E=AP000@0/1`009@0000<000000000000000000000100000<00401003`00000000000000000000
0>84082Y0`0I000030000?ooo`0B000030000080000U000030000000080U0000300000d0080U0000
300000L0080H000030000000000F000030000000000Q0000200001h0000H000000000000002`0@00
30800140000<0000200000`0000@000000000000000;000040000;01000<0P00600000`000000000
4P0000`0000100006@0000`0003oool05P0000`0000H00002P00010000000000000000T0000@0000
`0400202001B0000L0400040000Poooo0000000000000000T040004000040P0@E01Y06d0I@1c0200
CP1U07L0801B06l0K@1Q06h00000000000000000000000000000000000000000000000007c<`jQ80
P0h00>c[4P2E:oQg^3WhMoooooolja809k;lMaP74`000000/0cF=@00001605@0E`110580A@1L04<0
K01Q07<0L`1U07<0G01304`0D`1904@0G01k0300<00`0380@`150300<`0]0300<00`0300;@0`0300
000`02d0@`0`0300<00]0300<00`0300<00`0300<00`0300=00f07d0G01=06T0L`1S05<0M01Q07@0
M@1c0000O@000:Eg002S=P00001`ja80_NNUM`800h0XABlcK0fVMl9Hm7L90000h3H5<gS[4P1bBB/c
`0KeM`40001TMP08000002D0000<00000@0005@0001D00003P4000d1001R0@00o040004000194_51
3lo`@AP1003P0@000@0004`00000000000000000003ooooooooooe00001b0000E`000580001`0@00
0P00083nool0000000000000002@0@000@0000@2011D06T0K@1U07<0801>06D0M`0P0580K`1]0640
KP0000000000000000000000000000000000000000000000001c0000O@000:Eg002S=P00001`ja80
_NNUM`800h0XABlcK0fVMl9Hm7L90000h3H5<gS[4P1bBB/c`0KeM`40001TMP08000006D0L`1L04<0
C01C04T0A01L07/0<00`0300<P1304D0<00c02d0<00`0300<00]0300<0000300;@130300<00`02d0
<00`0300<00`0300<00`0300<00d03H0O@1L04d0J@1c06<0D`1d0640M01e07<0001m0000YGL00:<f
0000073[4P2mijEg0P03P2Q5;c<U0000h3H5<eS[4P3L1@00U`ceM`40000000000@0006Af00P00000
9@0000`000020000:00000`000010000E00005@0000k0000:00003/100310@000@0004TBlD4?co11
AP00080100010000C00000000000000000000?ooooooooooD00004H0003Z0000DP00070100010000
400000L00000000000000;`2000000000@828U<0N@1c07@0I@1]0000000000000000000000000000
0000000000000000000000000000000000000000000001P0X34H0000YGL00:<f0000073[4P2mijEg
0P03P2Q5;c<U0000h3H5<eS[4P3L1@00U`ceM`40000000000@0006Af00P00000I@1c05`0@`1<05<0
000C00@000000300<00b04<0A@0103<0;@0`03SZ4P0000000>/B0000n7LH=_Qgoooooa3[4P22LnQg
000C0000002P<AP0g0D:Bj0a6030EoMgC@1Y07<0H`1C07@0H@1d0=cZ4P000000c>`B05HHjWN`MNQg
ooooo`40001?OOAgX34H00P0003P=PDcC>/B0:0a60341oEg`5OgM`0000010000I7H02000000U0000
30000040000X000030000080000U0000300000L0080U000030000000080`0000300000l0080U0000
300000d0081;0000@00003000005000080000040000100004000000000000000/04000`200000000
00000;01000<0P00:00000`0000100008P0000`0003ooooo3P0001@000000000400001@0000J@0E0
;0000200003ooooo0P00000000000000003H@`100d@00000j0?X0b4000080000HP0000`000010000
8@0000P0000Q0000200007<0000<0000000001`0000800009@0000`0000000209@0000`000070020
9@0000`0000>0020<00000`0000?00206@0000`0003oool0600000`0000000004P0000`000020000
4`0000`000010000500000`0000=00005@0000`0000100005P0000`0000000003@00010000000000
000003X0000<00002P0001/0000@000000000000000S000080000>eHAcd000000000000PBcd0P9=3
083]@b@0000T0000kEQ7?@0000000000021;?@20Td<0P>e3100007<0000<0000000000d0000@0000
9`400=/1001B0000L04000<0000@00001`00000000000000_080000000070P8RD`1i07<0M01U06d0
0000000000000000000000000000000000000000000000000000000000000000000000008El00000
000005b48P0@j180E8@R04:6mWNH1B5OV9h]<`<0001LQ280000C00@00000ia809k;lMl9Hm7L10000
8@2:0@002P0hcBlcF_3dM`d0000I0;018@2:0B400022LnQg000C0000000@J34cl`4:bQ1X<C?0EoMg
000QGa00T04U0000/CWdMiSU4P2X1@Dc1>LB05HHjWN`MNQgooooo`400017A4U3200002SV4P0liQ80
Fi_fM`000030QR80000Q0;R68P0PGOMg00000>0:eSD0000000000000000liQ800@0009P58ElliQ80
PD7dMn0:eSETMP08000002D0000<00000`0004H0000X0000700004M4BD<2000000000000002_0@00
30800000000I000030000?ooo`0B000030000080000U000030000000081B0000L04000@0000@0000
1`00000000000000_080000000070P8RD`1i07<0M01U06d000000000000000000000000000000000
00000000000000000000000000000000000000004P0W/_ag`USdM`40000Q08X1000:03S=;c=Jl?Ag
3@0001T0/04Q08X18@00089cj7L001<00000011X<C?c0@[:46Pa<l1GmgL0025O402@0BD0002a>OAg
V>DB0:P51C<4ia80EQSZMk1ej7Oooooo0@0004M4BD<80000000B03cV4P1KVoIg00000<268P000240
^8HR021MmgL00000h0[F=@00000000009@000>0f1C?`iA8040000000mGL100006@2`0A00T05O2OEg
V0DQG`00080U0000402@0Ooo000@00006>HB0?XMm7NH1B5O402@0BD0001TMP08000002D0000<0000
100002D0000<00001`00P1P0000<0000000001H0000<000000000240000800007P0001P000000000
00000;01000<0P008`000200003]F4Lm00000000000084/m082C@`20kD<T000090000>eHAcd00000
0000000PBcd0P9=3083]@`@0000S000080000>eHAcd000000000000PBcd0P9=3083]@b@0000T0000
kEQ7?@0000000000021;?@20Td<0P>e3100002<0000P000023>X@@0000000000`>;?@@20Td<0P>e3
900002@00008<jQ1000000000030h/m1082C@`20kD<40000600000`0000000004P0000`000010000
6@0000`0003oool05P0000`0000H00008`0002000008<jQ1000000000030h/m1082C@`20kD<T0000
900000PcZ440000000000<3Rcd40P9=3083]@`@0000S00008000058j@3d00000000003fX@cd0P9=3
083]@b@0000T0000DSY0?@0000000000?JQ3?@20Td<0P>e310000580001`0@001@00023oool00000
00000000002@0@000@0000L2011D06T0K@1U07<0801>06D0M`0P0580K`1]0640KP00000000000000
00000000000000000000000000000000000O<c3Z4P203P00k>/B09D[n7Nh>OQgooooooc[4P0W/_ag
60LC0000002`3=He000004H0E01G0440DP1505`0@`1/0640L`1c06D0L`1L04<0C01C04T0A01L07/0
<00`0300<P1304D0<00c02d0<00`0300<00]0300<0000300;@130300<00`02d0<00`0300<00`0300
<00`0300<00d03H0O@1L04d0J@1c06<0D`1d0640M01e07<0001m0000YGL00:<f0000073[4P2mijEg
0P03P2Q5;c=/3JIg`USdM`T0003P=PDcN>/B0799:c?01_Eg0@0006Af00P000009@0000`000050000
E00005@0000d0@00i`4003P1003d0@000@0004TBlD4?co1160400>0100010000C000000000000000
00000?ooooooooooD0000780oomG0000DP00070100060000P?koo`00000000000000090100010000
1`8045@0J@1]06D0L`0P04h0I@1g0200DP1_06d0H@1^000000000000000000000000000000000000
00000000000007<0001m0000YGL00:<f0000073[4P2mijEg0P03P2Q5;c=/3JIg`USdM`T0003P=PDc
N>/B0799:c?01_Eg0@0006Af00P00000I@1c05`0@`1<05<0B@1405`0N`0`0300<00b04<0A@0`03<0
;@0`0300<00`02d0<00`0000<00]04<0<00`0300;@0`0300<00`0300<00`0300<00`03@0=P1m05`0
C@1Y07<0H`1C07@0H@1d07D0L`0007d0002UM`00XcH00000L>/B0;gWYGL200>0:4D_<bD0003P=PDc
F>/B0=`5002G3?Eg0@00000000010000I7H02000000U0000300000H0000X0000300000D0001D0000
E00002X1003M0@00=@400?0100010000BA;a@@o?l4560000P0400040001<00000000000000000000
oooooooooom@0000AP000>X0001B0000L04000D0000@00001`00000000000000_080000000070P8R
D`1i07<0M01U06d00000000000000000000000000000000000000000000000000000000000000000
00000000602P<AP0002UM`00XcH00000L>/B0;gWYGL200>0:4D_<bD0003P=PDcF>/B0=`5002G3?Eg
0@00000000010000I7H02000001U07<0G01304`0D`0001<010000000<00`0380@`150040<`0]0300
>>XB00000000ja80003hMaPfn7Oooooo4>/B089cj7L001<000000:0a603L1@Y;X34H0<1GmgM=06T0
L`1S05<0M01Q07@0g>XB0000003<k180EQSZMk1ej7Oooooo0@0004mmm7NP<AP020000>0f1C=<ja80
X34H0<@7mGO0EoMg00000040001TMP08000002D0000<00001@0002P0000<00001P0002D0000<0000
1`00P2D0000<00000000P2D0000<0000100004/000100000<00000D0000P00000@000040000@0000
9`400=/1000l0@00m@4002L1003K0@00?0400?D1000X0000300000D0000R000030000?oooom60000
500000P00017A4U30`000280000<0000ooooob80000<0000ooooobD0000<00003@00P2P0000<0000
0`0002P0000<000010000280000<0000ooooodH0002H0000S00004E=AR/V@000400000@000000000
24020S`0000`00000A30f`00001>000000000;MU[Cl200000P000080000200000A30f`000000003o
24010b`0000P00000A30f`80000000008kZD@`L_hT>aWh]3_6cg@`01000E@040400000@000020000
900002@00000080m000000000000080m0000000000020000G`0003P000040000>0000000000h0000
000000000@0F0000000000000000000000000000000U0000300000@0000U0000300000D0081G0000
900001D100320@00;0400?4100020000V19677@AkQhU0000300000L0080U000030000000080T0000
90000000P440000000000000P440000000000080000X0000300000@000160000G00005000015CDH[
24010c`0000`00000A30f`@0000000008YNB@lgFhD>`;9I3J<kN@dEcUT=:O^=38YNB@lgFhD<00@61
5401P100000400000000ob@0000T00000020?@00000000000020?@00000000000P0002D0000<0000
0P0001<0000<00000@0002D0000<00002000P5H0000/00009@400;d1000]0@00a`4000@0001C4S/L
aQ;J6llBL1aC4S/L9@0000`0000700204`0000`0000100009@0000`000000020900002@000000811
00000000000008110000000000020000AP000:@0002H0000ADe6:bY0000T0000600005b?XT800000
000005b?XT8VfhE3EL^S@`Q000H`00009000004@`=^[ZZXn00000000000000001P0005<0F@1=0480
C`1<03I008100000=00000000?l100000@000040002gl2MfaCeKZIDn0020?`00000000000020?`00
00000000000B000030000040000H000030000000000F0000300001P0000X000030000040001B0000
L0400040003Uoooo0000000000000000T0400000008700@0D`1i06d0HP1_06`0000004d0K`1^06l0
M01i0700I@0j05<0N@1]0680K`1/0200>P1f06D0LP1c06T0K`1^0000jGM@lA80B3?YM`/0000Z0000
C?4B0:03kWL00000:P0005`0CP1<05<0G01>06`0L`1C06D0H`1d06T0K`1^04<0D00d03800028kA80
000C00@0000000000@0000000001kQ80S>hB00c]4P2PVB`ce>dB09D[n7LH=_Qg003oonC]4P22LnQg
000C0000003XHQlcYPD:3nQR7c?0EoMg000000`X50000000_QX0H;3]4P2PVB`ca?LB0000jWN`MNQg
ooooo`40001?OOAgj68O<`P0003P=PDc8>hB0>QR7c?41oEg`5OgMiP58El100003CcdMl1GmgMTMP08
000002D0000<00000@0005@0001D0000504004@1000N0@00I@40004000000?Y1003j@A@1001O0@00
0@0004`00000000000000000003ooooooooooe00002gl000000002D0000<00003@00P4H000340000
^00004E=AR/[@00030000000000Z@000900001P0001LSj9200000000001LSj92e@S2@gn;FT88@006
A00003P000014<3KZjXZ?P0000020000000000l0001D04T0C@1505<0801>04D0E`0P0580C`1=0440
CP0003I008100000=00000000?l100000@000040001h01A;gSj27gPn0020?`00000000000020?`00
00000000000B000030000040000H000030000000000F0000300001P0000X000030000040001B0000
L0400040003boooo0000000000000000T0400040000700@0E01Y06d0I@1c0200CP1U07L0801B06l0
K@1Q06h000371Dd0K`1^06l0M01i0700I@0j05@0J@1]06D0L`0P0000ooooo`00000006/U061D]@L1
8;<70E`Y500f7@1P8;<70E`Y50087P1PG2TD0000000000005T<0H5`Y500P/`L1324D0<1PCV1@kA80
000C00@0000000000@0000000001kQ80E>hB0=C/4P2PVB`cW>dB09D[n7LH=_Qg003oojc]4P22LnQg
000C0000003XHQlcYPD:4>QR7c?0EoMg000000`X50000000_QX0H7S]4P2PVB`cS?LB0000jWN`MNQg
ooooo`40001?OOAgj68O<`P0003P=PDcj>dB0>QR7c?41oEg`5OgMiP58El100003CcdMl1GmgMTMP08
000002D0000<00000@0005@0001D0000Y`4003d0002/0@00C000004000000?Y1003j@JL1001:0000
0@0004`00000000000000000003ooooooooooe00001h0000000002D0000<00003@00P4H0002D0000
R00004E=AR/[@000300000000008@082?000030000014<3K000004h000000000ZA=@?`8000020000
0P00008000014<3K000000000?l8@043;000020000014<3K0P000000003SZhe348>[@cCaaD<@Pj]3
004001E00@0@000010000080000T000090000000P3d0000000000000P3d0000000000080001O0000
>00000@0000h0000000003P000000000000100d0000000000000000000000000000002D0000<0000
100002D0000<00001@00P5L0000T00006P4005D1002>0@00F@400080002f4G4E_aQa5BD0000<0000
1`00P2D0000<00000000P2@0000T00000020@@00000000000020@@00000000000P0002P0000<0000
100004H0002H0000S00004E=AR/8@082?000030000014<3K000004h000000000ZA=@?`8000020000
0P00008000014<3K000000000?l8@043?000030000014<3K10000000000nMIi3V12A@``iYT=AUII3
^oV]@jflXD<dek133H>[@`030`<E@040400000@000020000900002@00000080m000000000000080m
0000000000020000G`0003P000040000>0000000000h0000000000000@0=00000000000000000000
00000000000U0000300000@0000U0000300000D0081E0000;00003/1000P0@00I04005T100040000
ca<S4/PDda;05CPD6aIa5BD0000<00001`00P2D0000<00000000P2@0000T00000020@@0000000000
0020@@00000000000P0002P0000<0000100004H0001L0000D00004E=AR/8@043?000030000014<3K
10000000002ghIe34mbB@cNmVd=WH8m3YlnO@i_0Sd>ghIe34mbB@`010H4D@060400000@00000003o
900002@00000080m000000000000080m00000000000200009@0000`0000200004`0000`000010000
9@0000`000080020EP0002`0000g0@007P400401000V0@0010000;dCG19h4ndAnQ?i4KdCG18U0000
300000L0080C000030000040000U000030000000080T000090000000P440000000000000P4400000
0000008000160000R00007`00015CDH[24020S`0000`00000A30f`00001>000000000:TCD3l20000
0P000080000200000A30f`000000003o24010b`0000P00000A30f`8000000000d>><@a23Zd<39a93
k?j<@`01000E@040400000@000020000900002@00000080m000000000000080m0000000000020000
G`0003P000040000>0000000000h0000000000000@0=0000000000000000000000000000000U0000
300000@0000U0000300000D0081G000090000900000H0@00704005T100020000WA5a5B<9X14U0000
300000L0080U000030000000080T000090000000P440000000000000P440000000000080000X0000
300000@000160000N00006`00015CDH[24010eP0001<00000A30f`L000000000EAX=@a;LRd=CAQE3
H;6;@j=74d>XPHa3jbXB@`>nSD?EAA93@0N?@mI54T=01hm3EAX=@a;LRd<00@<30`6101A00H0@0000
100000000?lT000090000000P3d0000000000000P3d0000000000080000U000030000080000C0000
30000040000k0000200001/0000@0000dPP007`A000f0000400005D9001g4@00F00002P000000000
00000?oooooooooo0`0003D9TA4S2KPA9@WQ4ET0000T000000000000003oooooooooo`80000U2N4A
dPQl4Cd000080000?00000P0000n0000600008d0000G0@00UP0001l1000C000030000040000U0000
30000000080T000090000000P440000000000000P44000000000008000160000R00007`00015CDH[
24020S`0000`00000A30f`00001>000000000:TCD3l200000P000080000200000A30f`000000003o
24010b`0000P00000A30f`8000000000iLB=@a23Zd=KSli3ojm^@P01000E@040400000@000020000
900002@00000080m000000000000080m0000000000020000G`0003P000040000>0000000000h0000
000000000@0=0000000000000000000000000000000U0000300000@0000U0000300000D0081G0000
900001X1000j0000W`4005T100020000^A5a5M8I^`<U0000300000L0080U000030000000080T0000
90000000P440000000000000P440000000000080000X0000300000@000160000N00006`00015CDH[
24010eP0001<00000A30f`L0000000008;K?@gn;FT9if/m3H3ek@[h;cd>B>7=2@M3=@eZfKT;aQ/a3
2A9_@_:6c4<94Vm28;K?@gn;FT800@<30`6101A00H0@0000100000000?lT000090000000P3d00000
00000000P3d0000000000080000U000030000080000C000030000040000k0000200001/0000@0000
maT006/3000f000040000?`I003]0`00F00002P00000000000000?oooooooooo0`000>8Ic@>k6K/3
TAVm0eT0000T000000000000003oooooooooo`80002A6Kd3maU[0cd000080000?00000P0000n0000
600009T1000f0000X04003l0000C000030000040000U000030000000080T000090000000P4400000
00000000P44000000000008000160000R00007`00015CDH[24020S`0000`00000A30f`00001>0000
00000;MU[Cl200000P000080000200000A30f`000000003o24010b`0000P00000A30f`8000000000
d>><@a23Zd?MGZ53F2[@@P01000E@040400000@000020000900002@00000080m000000000000080m
0000000000020000G`0003P000040000>0000000000h0000000000000@0F00000000000000000000
00000000000U0000300000@0000U0000300000D0081G0000900001P1001V0000A@4005T100020000
WA5a5B`DPPHU0000300000L0080U000030000000080T000090000000P440000000000000P4400000
00000080000X0000300000@000160000G00005000015CDH[24010c`0000`00000A30f`@000000000
FcJO@m_@d4;>oJ53^Jg1@_IIXd=aYm=2FcJO@m_@d4800@615401P100000400000000ob@0000T0000
0020?@00000000000020?@00000000000P0002D0000<00000P0001<0000<00000@0002D0000<0000
2000P5H0000/0000?P40060000170@00JP0000@0003W4hL6@1@>1V`DWPKW4hL69@0000`000070020
4`0000`0000100009@0000`000000020900002@00000081100000000000008110000000000020000
AP0009@000280000ADe6:`Q00P980000?000004@`=/00000CP400000002Y4e0o0P00008000020000
0P000080000008100000@04@`=/000000000o`Q00@</00008000004@`=/2000000000<kmXD>i[L52
XN:o@kE^mT800@005D010100000400000P0002@0000T00000020?@00000000000020?@0000000000
0P0005l0000h0000100003P000000000>000000000004P4040000000000000000000000000000000
9@0000`0000400009@0000`0000200004`0000`000020000>`0000P0000K0000400004<D00071P00
=P000100001_50006PH005P0000d000000000000003oooooooooo`H0001c51`6MA@P1W<D9@Ib52T6
KA@Z1VTD:@Hf0000400003dD000E1P00F00003@00000000000000?oooooooooo1P0003TD4`Hg50l6
>A@;1S/D1`Ho50D6@a@71Sd0000800006`000100002K5000;@H003H0000@0000aa@00446001H0000
=000000000000000ooooooooool60000baA21/dDA`K;54/6bAA?1/DDD@K154l6=P000100002E5000
?0H005P0000d000000000000003oooooooooo`H0002A53X6Sa@e1Y4D<@JB52d6Ua@/1Y/D;@Hm0000
200001/0000@0000la@005@6000f0000400001lE001W1P00F00003@00000000000000?oooooooooo
1P0002<EJ@HU5Fh68aEb1R4EMPHM5GP66AEf1SH0000@0000kA@006<6001H0000=000000000000000
ooooooooool60000jAAQ1^LDG0KY55P6jQAD1^lDDPKc55@6?@0000P0000K0000400004/E001k1P00
=P000100001g5@00SPH005P0000d000000000000003oooooooooo`H0001k5I06OAFE1W/EV@Ii5Id6
M1FO1W0EW@Hf0000400004@E00291P00F00003@00000000000000?oooooooooo1P00040ER0Ho5H<6
@1Eo1T8EN`I75GT6BaEk1Sd0000800006`000100002S5@00XPH003H0000@0000caD00;D6001H0000
=000000000000000ooooooooool60000daFg1]DE^`KC5Kl6dAG41/`Ea@K85L@6=P000100002L5@00
/0H005P0000d000000000000003oooooooooo`H0002H5Jh6UQFZ1YPEYPJJ5J86WaFP1Z<EXPHm0000
200001/0000@0000naD00<P6000f0000400002HF003L1P00F00003@00000000000000?oooooooooo
1P0002/Fg@H/5^86:aKV1RTFjPHT5^`681KZ1SH0000@0000m1D00=L6001H0000=000000000000000
ooooooooool60000l1GE1^hEd0K`5L`6lQG81_LEa`Kk5LP6?@0000P0000K00004000058F003_1P00
=P000100001n5P000PL005P0000d000000000000003oooooooooo`H000225P@7Q1H91h8F3@N15Q47
O1HC1gPF4@Lf0000400004`F003n1P00F00003@00000000000000?oooooooooo1P0004PFo0I65_L6
B1Kc1TXFk`I>5^d6DQK_1Sd0000800006`000100002Z5P005PL003H0000@0000eQH002T7001H0000
=000000000000000ooooooooool60000fQH[1m`F<0OJ5S@7f1Hh1m@F>PO@5SP7=P000100002T5P00
90L005P0000d000000000000003oooooooooo`H0002P5R<7WQHN1j0F6PNQ5QH7YQHD1jXF5PLm0000
200001/0000@00000QL003d7000f0000400002hG001@1`00F00003@00000000000000?oooooooooo
1P00038GDPLd5eH7<QMJ1c0GG`L/5f07:1MO1cH0000@0000o1H004/7001H0000=000000000000000
ooooooooool60000n1I91oHFA@Oh5T47nAHm1ohF>`L25cd7?@0000P0000K0000400005XG001S1`00
=P00010000265`00M`L005P0000d000000000000003oooooooooo`H0002:5gP7S1Mm1hXGP@N85hD7
PaN71glGQ@Lf0000400005<G001b1`00F00003@00000000000000?oooooooooo1P0004lGL0M>5f/7
CaMW1e4GH`MF5f87FQMS1cd0000800006`000100002b5`00RPL003H0000@0000gQL009d7001H0000
=000000000000000ooooooooool60000hQNO1n@GY0OR5jP7h1N/1m/G[POG5j`7=P000100002[5`00
V@L005P0000d000000000000003oooooooooo`H0002W5iL7YANB1jLGSPNY5hX7[QN81k8GRPLm0000
200003`000080000?`0001P000110@00GP000801001m00004`0000`0000100009@0000`000000020
9@0000`000070020900002@00000081100000000000008110000000000020000:00000`000040000
AP0009@000280000ADe6:`Q00P980000?000004@`=/00000CP400000002Y4e0o0P00008000020000
0P000080000008100000@04@`=/000000000o`Q00@</00008000004@`=/2000000000<kmXD>i[L52
nHYX@a3fX4<00@005D010100000400000P0002@0000T00000020?@00000000000020?@0000000000
0P0005l0000h0000100003P000000000>000000000004P4040000000000000000000000000000000
9@0000`0000400009@0000`0000200004`0000`000020000>`0000P0000K0000400004PD000A1P00
=P000100000f5000?@H005P0000d000000000000003oooooooooo`H0000d5446;aA31R/D@PHW5406
9A@k1RLD=`Hf0000400003TD000;1P00F00003@00000000000000?oooooooooo1P0003XD1`Ho50D6
@a@61TLD20I950d6B1@A1Sd0000800006`000100000S5000JPH003H0000@00004A@009H6001H0000
=000000000000000ooooooooool6000041BJ1P/DW0H759/60aBI1P4DU0H35906=P000100000E5000
I0H005P0000d000000000000003oooooooooo`H0000F56066aAN1QlDG`HS56469AAV1R<DJPHm0000
200001/0000@0000oa<00<<6000f000040000>dC003_1P00F00003@00000000000000?oooooooooo
1P000>`Cl`KW4oD6ha?d1]lClPKM4nd6gQ?Y1SH0000@0000lA<00;d6001H0000=000000000000000
ooooooooool60000lQ>i1_LC]`Kk4kP6oa>j1P4D_`Ko4l<6?@0000P0000K000040000=/C000L1`00
=P00010000394`00B0L005P0000d000000000000003oooooooooo`H000374d`7`a=>1klCC@Nk4d/7
^A=61kXC@PLf000040000<`C000F1`00F00003@00000000000000?oooooooooo1P000<hC4POC4a07
ea<A1m/C4`OM4aP7fa<L1cd0000800006`000100002g4`00M@L003H0000@0000YA<00:47001H0000
=000000000000000ooooooooool60000Xa>U1ilCY`NK4jH7UQ>T1iDCW`NF4i/7=P000100002X4`00
K`L005P0000d000000000000003oooooooooo`H0002Z4f/7[a=Y1k<CJPNg4f`7^A=a1kLCM@Lm0000
200001/0000@0000Ta<00<h7000f00004000084C003j1`00F00003@00000000000000?oooooooooo
1P0007lCoPMk4`08MQ?n1g8Co@M`4oP7LQ?d1cH0000@0000Q1<00<P7001H0000=000000000000000
ooooooooool60000QQ?31hXC`PN?4l<7Ta?51iDCbPNC4lh7?@0000P0000K0000400006lC000W2000
=P000100001M4`00D`P005P0000d000000000000003oooooooooo`H0001K4eL8EQ=I258CE`Q>4eH8
C1=A24hCC@Pf00004000060C000Q2000F00003@00000000000000?oooooooooo1P00068C70QV4aX8
JQ<L26lC7PQ`4b88Ka<W23d0000800006`000100001;4`00O`P003H0000@0000>A<00:`8001H0000
=000000000000000ooooooooool60000=a>`238C/PP^4k08:Q>_22PCZPPZ4jH8=P000100000l4`00
N@P005P0000d000000000000003oooooooooo`H0000n4gD8@Q=c24HCM@Q:4gL8C1=k24/CO`Pm0000
200001/0000@00009a<00=P8000f0000400001@C00052@00F00003@00000000000000?oooooooooo
1P0001<C2@T>4`/92Q<92@HC20T44`<91Q?o23H0000@000061<00=88001H0000=000000000000000
ooooooooool600006A?>21hCc0PR4lh89Q?@22PCe0PW4mP8?@0000P0000K00004000008C000a2@00
=P000100003`4P00GPT005P0000d000000000000003oooooooooo`H0003_4V89jQ9T2NHBHPWR4V49
h19L2N8BF0Tf000040000?@B000[2@00F00003@00000000000000?oooooooooo1P000?DB9`Wj4RD9
oQ8W2@8C:@T44bd90Q<a2Cd0000800006`000100003N4P00RPT003H0000@0000c1800;L9001H0000
=000000000000000ooooooooool60000ba:k2LHB_@W24[/9_Q:j2K`B]@Vm4[49=P000100003?4P00
Q0T005P0000d000000000000003oooooooooo`H0003A4X09eQ9n2MXBP0WN4X89h1:62MhBRPTm0000
200001/0000@0000^Q800><9000f000040000:PB000@2P00F00003@00000000000000?oooooooooo
1P000:HB50ZR4QH:WQ8D2YXB4PZH4Ph:VA8:2SH0000@0000Za800=d9001H0000=000000000000000
ooooooooool60000[A;I2K8Be`Vf4]T9^Q;J2K`Bg`Vj4^<9?@0000P0000K0000400009HB000l2P00
=P00010000244P00J@X005P0000d000000000000003oooooooooo`H000224Vd:OQ9_2WXBK@Ye4V/:
La9W2WDBH`Xf0000400008LB000f2P00F00003@00000000000000?oooooooooo1P0008TB<PZ>4S0:
TQ8b2YHB<`ZH4SP:UQ8l2Sd0000800006`000100001b4P00U@X003H0000@0000H1800<8:001H0000
=000000000000000ooooooooool60000GQ;62UTBb0YE4/H:DA;42TlB`0YA4[`:=P000100001S4P00
S`X005P0000d000000000000003oooooooooo`H0001U4X/:JA:92VdBR`Yb4X`:M1:A2W8BU@Xm0000
200001/0000@0000CQ800>h:000f0000400003`B000K2`00F00003@00000000000000?oooooooooo
1P0003XB7`/e4R4;<A8O2bdB7@/[4QT;;A8E2cH0000@0000?a800>P:001H0000=000000000000000
ooooooooool60000@A;T2TDBhPY94^@:CA;U2TlBjPY>4^h:?@0000P0000K0000400002XB00172`00
=P000100000H4P00L`/005P0000d000000000000003oooooooooo`H0000F4WP;4A9j2`dBN0/94WH;
1a9b2`TBK@/f0000400001/B00112`00F00003@00000000000000?oooooooooo1P0001`B?@/Q4S/;
9A8m2bTB?P/[4T<;:Q972cd0000800006`00010000054P00X0/003H0000@0000la400<`;001H0000
=000000000000000ooooooooool60000lQ7A2ndAdP_Y4M4;iA7?2n<Ab`_U4LH;=P000100003g4@00
VP/005P0000d000000000000003oooooooooo`H0003h4IH;oA6D2`4BUP/54YL;1a:L2`DBX0/m0000
200001/0000@0000hA400?T;000f000040000<lA000U3000F00003@00000000000000?oooooooooo
1P000<hA:@c94B/<aA4Z3<4A:0bo4B<<`14O33H0000@0000da400?<;001H0000=000000000000000
ooooooooool60000e17_2mTAk@_M4Nh;hA7`2n<Am@_Q4OT;?@0000P0000K000040000;dA001B3000
=P000100002[4@00OP`005P0000d000000000000003oooooooooo`H0002Y4H8<YA643:4AP`bM4H4<
Va5l39`AN0`f000040000:hA001<3000F00003@00000000000000?oooooooooo1P000;0AB0be4DH<
^A573;dAB@bo4Dh<_A5B33d0000800006`000100002I4@00Z``003H0000@0000Qa400=L<001H0000
=000000000000000ooooooooool60000QA7K384Ag@am4M`<NA7J37LAe@ah4M4<=P000100002:4@00
Y@`005P0000d000000000000003oooooooooo`H0002<4J4<TA6O39DAX0bI4J8<Va6W39TAZ``m0000
200001/0000@0000MA4000@=000f0000400006<A000`3@00F00003@00000000000000?oooooooooo
1P00064A=0eM4CH=F14e3E@A<`eB4Bh=E14Z3CH0000@0000IQ400?h<001H0000=000000000000000
ooooooooool60000J17j36`An0aa4OT<MA7k37LA00ee4@@=?@0000P0000K00004000054A001M3@00
=P000100000o4@00R@d005P0000d000000000000003oooooooooo`H0000m4Hd=>16?3C@ASPd`4H`=
;Q673C0AP`df00004000048A001G3@00F00003@00000000000000?oooooooooo1P0004@AD`e84E4=
C15B3E4AE0eB4ET=DA5M3Cd0000800006`000100000]4@00]Pd003H0000@00006a400>8=001H0000
=000000000000000ooooooooool600006A7V3A@Aj0d@4NH=317U3@XAh0d<4M`==P000100000N4@00
/0d005P0000d000000000000003oooooooooo`H0000P4J/=916Z3BPAZ`d/4Jd=;Q6b3BdA]Pdm0000
200001/0000@00002A4000l>000f000040000?H@000k3P00F00003@00000000000000?oooooooooo
1P000?D@?`k`444>k10o3^P@?PkV43T>j10e3SH0000@0000nQ0000T>001H0000=000000000000000
ooooooooool60000na043P0A0Ph44@@>21463PXA2Ph94@l>?@0000P0000K000040000>@@001W3P00
=P000100003B4000U0h005P0000d000000000000003oooooooooo`H0003A49P>c12J3/P@V0k449L>
`Q2B3/@@SPhf000040000=H@001Q3P00F00003@00000000000000?oooooooooo1P000=L@G@kL45/>
h11M3^@@G`kV46<>i11W3Sd0000800006`00010000304000`0h003H0000@0000[Q000>d>001H0000
=000000000000000ooooooooool60000[A3a3ZP@l`jT4?4>X13`3Yh@j`jO4>L>=P000100002a4000
^Ph005P0000d000000000000003oooooooooo`H0002c4;H>^12d3[`@]Pk04;P>`Q2l3/0@`0hm0000
200001/0000@0000W10001T?000f0000400008X@00163`00F00003@00000000000000?oooooooooo
1P0008P@BPn444`?P11:3g`@B@mj44@?Na103cH0000@0000SA0001<?001H0000=000000000000000
ooooooooool60000Sa0?3i@@3@nH40l?W10A3ih@5@nL41T??@0000P0000K0000400007P@001b3`00
=P000100001V4000W`l005P0000d000000000000003oooooooooo`H0001T4:<?H12U3e`@X`mG4:8?
EA2M3eL@V@lf0000400006T@001/3`00F00003@00000000000000?oooooooooo1P0006/@J0m`46H?
M11X3gP@JPmj46h?N11b3cd0000800006`000100001D4000b`l003H0000@0000@Q000?P?001H0000
=000000000000000ooooooooool60000@13l3c`@oPlg4?`?<a3j3c4@mPlc4?8?=P00010000154000
a@l005P0000d000000000000003oooooooooo`H000174<4?Ba2o3dl@`@mD4<8?EQ373e@@b`lm0000
200001/0000@0000<10002@@000f0000400001h@001A4000F00003@00000000000000?oooooooooo
1P0001`@EA0G45L@4a1E40l@Da0=44l@3a1;43H0000@00008A0001h@001H0000=000000000000000
ooooooooool600008a0J42L@610[41X@;a0K434@810`42@@?@0000P0000K0000400000`@001m4000
=P000100003j3`00ZQ0005P0000d000000000000003oooooooooo`H0003h3jh@l`n`4>l?[Q3[3j`@
j@nX4>/?Y10f000040000?d?001g4000F00003@00000000000000?oooooooooo1P000?h?La03474@
1a1c40/@M10=47T@311m43d0000800006`000100003X3`00eQ0003H0000@0000e@l000<A001H0000
=000000000000000ooooooooool60000e0l74Ll?2A7;3`LAa`l54LD?0A773od@=P000100003I3`00
d10005P0000d000000000000003oooooooooo`H0003J3l`@g`o:4><?c13W3ld@j@oB4>P?eQ0m0000
200001/0000@0000``l002lA000f000040000;4?001K4@00F00003@00000000000000?oooooooooo
1P000;0?H16[3f8AY`mP4J<?GQ6Q3eXAXPmE4CH0000@0000]@l002TA001H0000=000000000000000
ooooooooool60000]PlU4K/?8a6o3bDA``lV4LD?:a733blA?@0000P0000K0000400009l?00284@00
=P000100002=3`00]14005P0000d000000000000003oooooooooo`H0002<3kTAQ`nj4H<?^A5o3kLA
O@nb4Gh?[Q4f00004000090?00224@00F00003@00000000000000?oooooooooo1P00098?OQ6G3g`A
V`mn4Il?Oa6Q3h@AW`n84Cd0000800006`000100001k3`00hA4003H0000@0000J@l000dB001H0000
=000000000000000ooooooooool60000I`lA4V<?4a9O3a8BF`l@4UT?2a9J3`LB=P000100001/3`00
fa4005P0000d000000000000003oooooooooo`H0001^3mLAL`oE4GL?eQ5k3mPAO@oM4G/?hA4m0000
200001/0000@0000E`l003XB000f0000400004D?001V4P00F00003@00000000000000?oooooooooo
1P0004<?JQ8o3f`B>Pm[4SH?JA8d3f@B=PmP4SH0000@0000B0l003@B001H0000=000000000000000
ooooooooool60000BPl`4Th?;Q9C3blBE`la4UT?=Q9G3cXB?@0000P0000K0000400003<?002C4P00
=P000100000Q3`00_a8005P0000d000000000000003oooooooooo`H0000O3l<B6Po54QH?a18B3l8B
40nm4Q8?^A8f0000400002@?002=4P00F00003@00000000000000?oooooooooo1P0002H?RA8Z3hLB
;Pn84S<?RQ8e3hlB<`nC4Sd0000800006`000100000?3`00k18003H0000@0000o@h001PC001H0000
=000000000000000ooooooooool60000n`hL4oH>7Q?b3QdCkPhK4n`>5Q?^3Q8C=P00010000003`00
iQ8005P0000d000000000000003oooooooooo`H000023n8B1PoP4PX?hA8>3n<B40oX4Pl?k18m0000
200001/0000@0000j`h004DC000f000040000=T>001a4`00F00003@00000000000000?oooooooooo
1P000=L>MA?B3WLCcPif4lX>M1?83VlCbPi[4cH0000@0000g0h003lC001H0000=000000000000000
ooooooooool60000g@hk4n8>>A?V3SXCjPhl4n`>@A?[3TDC?@0000P0000K000040000<H>002N4`00
=P000100002d3P00bQ<005P0000d000000000000003oooooooooo`H0002c3/hC[Pk@4jX>cQ>V3/dC
Y0k84jH>a1<f000040000;P>002H4`00F00003@00000000000000?oooooooooo1P000;T>Ta>n3Y8C
`PjC4lH>UA?83YXCaPjN4cd0000800006`000100002R3P00ma<003H0000@0000T@h0028D001H0000
=000000000000000ooooooooool60000S`hV58X>:1B63RLDPPhU580>81B23Q`D=P000100002C3P00
lA<005P0000d000000000000003oooooooooo`H0002E3^`CVPkZ4ih>k1>R3^hCY0kb4j8>ma<m0000
200003`000080000?`0001P0003V0000GP0004L100140@004`0000`0000100009@0000`000000020
9@0000`000070020900002@00000081100000000000008110000000000020000:00000`000040000
AP0008P0001l0000ADe6:`Q00P8l0000<000004@`=/00000CP000000002gIJdo0P00008000020000
0P00004@`=/000000000o`Q00@</00008000004@`=/2000000000=3SS4<@Pj]3F@Aa@aadXT<00@00
5D010100000400000P0002@0000T00000020?@00000000000020?@00000000000P0005l0000h0000
100003P000000000>0000000000000405P0000000000000000000000000000009@0000`000040000
9@0000`000050020E`0002@0003_0000@`4001`1001I0@000P0009dALADA3dlD9@0000`000070020
9@0000`000000020900002@00000081100000000000008110000000000020000:00000`000040000
AP0005`0001@0000ADe6:`Q00@<l0000<000004@`=/40000000005=5L4?TUJA38deZ@a3fX4>C[W=3
O;nP@e=5L4?TUJA30041PAA00H0@0000100000000?lT000090000000P3d0000000000000P3d00000
00000080000U000030000080000C000030000040000U0000300000P0081F0000;0000>X000110@00
m00004X1000400001@nC5:D>7a@k3aPD1@nC52D0000<00001`00P1<0000<00000@0002D0000<0000
0000P2@0000T00000020@@00000000000020@@00000000000P0004H000280000O00004E=AR/8@082
?000030000014<3K000004h000000000]fF]?`80000200000P00008000014<3K000000000?l8@043
;000020000014<3K0P000000003jYHi348>[@`UR_T=:k053004001E00@0@000010000080000T0000
90000000P3d0000000000000P3d0000000000080001O0000>00000@0000h0000000003P000000000
000101H0000000000000000000000000000002D0000<0000100002D0000<00001@00P5L0000T0000
6`400800001o0@00F@400080003E4G4EcALO22D0000<00001`00P2D0000<00000000P2@0000T0000
0020@@00000000000020@@00000000000P0002P0000<0000100004H0001L0000D00004E=AR/8@043
?000030000014<3K10000000002;@;a322X1@j7R_d>eK_I2YQG0@aZI14>;@;a322X1@`010H4D@060
400000@00000003o900002@00000080m000000000000080m00000000000200009@0000`000020000
4`0000`0000100009@0000`000080020EP0002`0001h0@00N`00084100250000100008TG4`Sm5k@7
0aQ:28TG4`PU0000300000L0080C000030000040000U000030000000080T000090000000P4400000
00000000P44000000000008000160000^0000:`00015CDH[:T0002@0000H0000G8nR@P0000000000
G8nR@XZc14<`MGQ324061T@0000h00000A30fj^Z:Sh000000P000000000?0000E01904d0A@1C0200
CP1505L0801B04l0C@1104h0000f@0J0@00003@00000003o0@00004000010000N@2a<gDnPQmh?P00
P3l0000000000000P3l00000000000004P0000`000010000600000`0000000005P0000`0000H0000
9@0000`000010000E00005@0002F0000o`0009`0000>0@000@000000nT400?Y1V00000`100010000
C00000000000000000000?ooooooooooD00007T0@`H000009@0000`0000=0020AP000<@0002h0000
ADe6:b]0000<0000000002Y0000T0000600005b?XT800000000005b?XT;DXYM3J6R9@PQ01PI40000
>000004@`=^[ZRXn000000<0000000003`0005@0B@1=04D0D`0P04h0A@1G0200DP1?04d0@@1>0000
=T06P400000d00000000o`40000100000@0005H0Y05<?X8ON3h0080o000000000000080o00000000
00000180000<00000@0001P0000<0000000001H0000<0000600002P0000<00000@000580001`0@00
0@000?;oool0000000000000002l0P000@0000L0101D06T0K@1U07<0801>06D0M`0P0580K`1]0640
KP000000C@1_06h0K`1d07T0L01U03X0E01Y06d0I@1c0200003ooooo00000000JbD0H5Be1`4P/`L1
G2TD03HM060P/`L1G2TD00PN061L:A@000000000000F@`1PG2TD022c1`4<8A@0`61>H1S]4P0001<0
10000000000100000000007]4P0LkQ80W>`B0:2I;3=TkA80UB_hMaPfn7L00?ooM>dB089cj7L001<0
00000>QR7c>V1@XAj68O<l1GmgL0000032PD0000002n6P1P@>dB0:2I;3=Dma80003ZMk1ej7Oooooo
0@0004mmm7OXHQlc20000>0f1C>`kA80j68O<l@7mGO0EoMgV0DQG`40000=??Ag`5OgMfAf00P00000
9@0000`000010000E00005@0000o0@00B`0004T1001K00000@000000nT400?Y1?`4005P000010000
C00000000000000000000?ooooooooooD00005H0L@L000009@0000`0000=0020AP000?0;003T2`00
ADe6:b]0000<0000000002E0000@000010000000000/@000900001P0001R<7<m0000P000080>JmTl
XgU/@aFIS488@0L5M0/006P;00014<3K0P0000<0001H2`000@0006`00007000000000>L1003U0000
000000000000@000Va`00215CDH00040F0/003@000040000000000000000000000@0000300100@00
l0000000000000000000003R1020Z@<06@0000`0003oool04P0000`0000200009@0000`000000020
9@0000`0000=00209@0000`000070020600000`0000000005P0000`0000000008@0000P0000N0000
600000000000000030800>X0000A0000300000P0000<000040000000000000002`000100000<0P00
jP0001P0000<000000000180000<00000@0001T0000<0000oooo01H0000<0000600000X0000@0000
000000000009000040000<04000P0P009P0001`0000100000000010000000000000002D0000<0000
0@0001/0000@00003`<00400000f0000400000l3002h0@006`000100001O1000@00003H0000@0000
G`@00;P1001B0000L04000800020o_oo0000000000000000T040004000000P0@E01Y06d0I@1c0200
CP1U07L0801B06l0K@1Q06h00000000000000000000000000000000000000000000000007@3hjA80
P0h00;C[4P2E:oQg^3WhMoooooo4ja809k;lMaP74`00000040T1=P00001605@0E`110580A@1L04<0
K01Q07<0L`1U07<0G01304`0D`1904@0G01k0300<00`0380@`150300<`0]0300<00`0300;@0`0300
000`02d0@`0`0300<00]0300<00`0300<00`0300<00`0300=00f07d0G01=06T0L`1S05<0M01Q07@0
M@1c0000O@000:Eg002T=P00000hja80_NNUM`800h1HARlc9@000>0f1C<Pja80l@D009L<mGL10000
6@2`0A/0001TMP08000002D0000<00000P0005@0001D00001`000180001m0000bP000040003UMEM1
da9G@A0000200@000@0004`00000000000000000003ooooooooooe00001F0000jP000580001`0@00
0`00023oool0000000000000002@0@000@000002011D06T0K@1U07<0801>06D0M`0P0580K`1]0640
KP0000000000000000000000000000000000000000000000001c0000O@000:Eg002T=P00000hja80
_NNUM`800h1HARlc9@000>0f1C<Pja80I>XB0000003U>?Qg000C03Se9c<00000@>XB08P64`3djQ80
UB_hMb0in7Oooooo1>/B042go7OX1Q<0@?DW<d3e9c<jT1d0@?DW<c00<0000300;@130300<00`02d0
<00`0300<00`0300000C0300<00d03H0O@1L04d0J@0006<00P000640M01e07<0001m0000YGL00:@f
000C00800000mBLc400005Q6;c<100000@0007CZ4P1D0000?>/B09D[n7LH=_QgooooofAf00P00000
9@0000`000030000:00000`000020000E00005@0001Y0000N@0008h0003U00000@000>EeEd7C4UM1
m0000>0100010000C00000000000000000000?ooooooooooD00007<0001G0000DP00070100020000
P?koo`000000000000000901000000020080D5<0N@1]0680K`1/0000000000000000000000000000
0000000000000000000000000000000000000000000001<0<?8C0000YGL00:@f000C00800000mBLc
400005Q6;c<100000@0007CZ4P1TjQ8000000>Dhn7L001<0>?DW<`000010jQ80R0HC0?CZ4P2E:oQg
83WhMoooool4ja80@;OlMnP64`10mBLc@?DW<hJ@7@10mBLcb>XB0000n7LH=_QgooooomSZ4P22LnQg
000C0000000001<0j0D:VC3b4`30EoMgC@1Y0000H`020000H@1d0:CZ4P000000U>`B05HHjWL001<0
0P00003e9c<@0000<?8C004000010000M>XB05@0000lja80UB_hMaPfn7OoooooI7H02000000U0000
30000080000X0000300000<0001D0000E0000<L0000000008@400<P000010000iGEG@M<BEd7=0@00
P0400040001<00000000000000000000oooooooooom@0000?O000=<0001B0000L04000<00020o_oo
0000000000000000_080004000000P0@E01Y06d0I@1c0200CP1U07L0801B06l0K@1Q06h000000000
00000000000000000000000000000000000000004`1@lQ<0U>`B05HHjWL001<00P00003e9c<@0000
<?8C004000010000M>XB06CZ4P000000iCShM`004`0hmBLc0000043Z4P281Q<0m>XB09D[n7LP>OQg
ooooo`C[4P10]oagj0HC043e9c=0mBLcb90M043e9c?8jQ80003hMaPfn7Oooooof>XB089cj7L001<0
000000004`1g1@Z5D?8C0<1GmgM=06T0001S0080001Q07@0Y>XB0000002Dk180EQSZM`004`020000
0?DW<a00001@lQ<00@000040001djQ80E00003c[4P2E:oQg63KhMooooomTMP08000002D0000<0000
0`0002P0000<00000P0005@0001D0000D@400140003@0@00b0000040003UMEM1da9G@@d300200@00
0@0004`00000000000000000003ooooooooooe00001F0000o`000580001`0@000P00010000070000
00000000002l0P00000000420R9C07T0L`1d06D0K@00000000000000000000000000000000000000
00000000000000000000000000000000000C033b4`2Dk180EQSZM`004`0200000?DW<a00001@lQ<0
0@000040001djQ80I>XB0000003U>?Qg000C03Se9c<00000@>XB08P64`3djQ80UB_hMb0in7Oooooo
1>/B042go7OX1Q<0@?DW<d3e9c<DTAd0@?DW<lSZ4P000?Qg63KhMooooooHjQ80PW?XM`004`000000
000C0>P52YX`lQ<0`5OgMdd0J@0006<00P000640M02TjQ80000009C/4P1F6>Yg000C00800000mBLc
4000033b4`0100000@0007CZ4P1D0000?>/B09D[n7LH=_QgooooofAf00P000009@0000`000020000
:00000`0000300009@0000`0000700209@0000`000000020<00000`0000?00209@0000`0000=0020
B`000400000`00001@000200000100000@00010000000000000000`2003Z000000000000000<0P00
jP0002P0000<00000@0002P0000<00000P000280000<0000ooooo`h0000D000000000100000D0000
6T07@2`0000P0000ooooo`8000000000000000100d@006]300000>P3j0<Q000020000680000<0000
0@000240000800008@0000P0001c000030000000000L0000200002D0000<00000000P2D0000<0000
1`00P2D0000<00003P00P300000<00003`00P1T0000<0000oooo01P0000<000000000180000<0000
0P0001<0000<00000@0001@0000<00003@0001D0000<00000@0001H0000<0000000000d0000@0000
00000000000j0000300000X0000K000040000000000000008`0002000000T>8m00000000001LOMXm
001]@`00ST8T00009000002@hSd00000000005amfSd006e3002>@P@0001c000030000000000=0000
40000>d000170000DP00070100040000400000L00000000000000;`2000000001`828U<0N@1c07@0
I@1]0000000000000000000000000000000000000000000000000000000000000000000000000180
00000000000Z0000Y>DB0?lcjGL00000000001000@0<0000700d0080000d1P00X0800;IJ00000000
000001000@0;<_Qg900006/U061D]@L18;<70E`Y500f7@1P8;<70E`Y50087P1PG2TD000000000000
5T<0H000500P/`L1324D0<1PCV2PiA808;<70A12060000000@0006/U0638iA801@00070in7L004`0
N0M<00D0000009LeX>DB080l001Lia80UB_hMkPin7OoooooK>LB02Nbo7Mh1d`000000?oooolO401h
0@000000003<iQ80I7H02000000U0000300000@000160000:00001`00017A4U30P00000000000000
30800>X0000000006@0000`0003oool04P0000`0000200009@0000`000000020DP00070100030000
400000L00000000000000;`2000000001`828U<0N@1c07@0I@1]0000000000000000000000000000
000000000000000000000000000000000000000000000000]UX0000000000000400100/bn7LT0000
JbD0H5Be1`4P/`L1G2TD03HM060P/`L1G2TD00PN061L:A@000000000000F@`1P000D022c1`4<8A@0
`61>H:3U4P0P/`L14480H00000010000JbD0H<SU4P050000L3WhM`00C01h1d`01@000000UcFPiA80
P3`005cW4P2E:oQg^3WhMooooom/ia809k;lMbD0003P=PDc^>DB010000000?Eg0@0001T0/04@0901
G`WeMiP58El000209@000100T07oo`0040000>3U4P3j7OAgV0DQGa00T04U0000I7H02000000U0000
300000<0000U0000300000L0080H000030000000000F000030000000000Q0000200001h0000H0000
00000000000<0P00jP0002<0000P0000093R?@0000000000G7gJ?@00KD<008i2900002@00000T>8m
00000000001LOMXm001]@`00ST8400008`0002000000T>8m00000000001LOMXm001]@`00ST8T0000
9000002@hSd00000000005amfSd006e3002>@P@0000S00008000063OId800000000009Jfad4006e3
002>@R@0000T0000H=mW@P0000000000U[K7@@00KD<008i2100001P0000<000000000180000<0000
0@0001T0000<0000oooo01H0000<0000600002<0000P0000H=mW@P0000000000U[K7@@00KD<008i2
900002@0001PgfM200000000002F]/M1001]@`00ST8400008`000200003G@T<m00000000000Tmc/m
001]@`00ST8T000090000=M2@cd00000000002Cg>cd006e3002>@P@0000V0000700000D000000000
40000000000000009@0000`0000500006`000100000?0`00@00003H0000@00003`<00;P1000K0000
400005l400100000=P000100001O1000^0400580001`0@001P00083nool0000000000000002@0@00
0@0000L2011D06T0K@1U07<0801>06D0M`0P0580K`1]0640KP000000000000000000000000000000
0000000000000000000M0?SY4P203P00]>/B09D[n7Nh>OQgooooolC[4P0W/_ag60LC0000000@2@4f
000004H0E01G0440DP1505`0@`1/0640L`1c06D0L`1L04<0C01C04T0A01L07/0<00`0300<P1304D0
<00c02d0<00`0300<00]0300<0000300;@130300<00`02d0<00`0300<00`0300<00`0300<00d03H0
O@1L04d0J@1c06<0D`1d0640M01e07<0001m0000YGL00:@f000003S[4P2mijEg0P03P5Q6;c<U0000
h3H5<b3[4P3a1@00U`ceM`40000I0;016`0006Af00P000009@0000`000060000E00005@0003^0000
B@000?X0001L00000@000>EeEd7C4UM14000080100010000C00000000000000000000?oooooooooo
D00005H0003Z0000DP000701000700008?ooo`000000000000000901000100001`8045@0J@1]06D0
L`0P04h0I@1g0200DP1_06d0H@1^00000000000000000000000000000000000000000000000007<0
001m0000YGL00:@f000003S[4P2mijEg0P03P5Q6;c<U0000h3H5<b3[4P1TjQ8000000>Dhn7L001<0
>?DW<`000010jQ80R0HC0?CZ4P2E:oQg83WhMoooool4ja80@;OlMnP64`10mBLc@?DW<cZ@7@10mBLc
<00`0000<00]04<0<00`0300;@0`0300<00`0300<00001<0<00`03@0=P1m05`0C@1Y0000H`020000
H@1d07D0L`0007d0002UM`00Y3H001<00P00003e9c<@0000F4H_<`4000010000M>XB05@0000lja80
UB_hMaPfn7OoooooI7H02000000U0000300000L0000X0000300000H0001D0000E0000?T0001C0000
o@0005h000010000iGEG@M<BEd7d0000h0400040001<00000000000000000000oooooooooom@0000
L`0005L0001B0000L04000H00020o_oo0000000000000000T040000000870P1@D`1i06d0HP1_06`0
0000000000000000000000000000000000000000000000000000000000000000000000004`0`lQ<0
002UM`00Y3H001<00P00003e9c<@0000F4H_<`4000010000M>XB06CZ4P000000iCShM`004`0hmBLc
0000043Z4P281Q<0m>XB09D[n7LP>OQgooooo`C[4P10]oagj0HC043e9c=0mBLcQY0M043e9c?8jQ80
003hMaPfn7Oooooof>XB089cj7L001<0000000004`3X1@ZI<?8C0<1GmgM=06T0001S0080001Q07@0
Y>XB0000002Dk180EQSZM`004`0200000?DW<a00000`lQ<00@000040001djQ80E00003c[4P2E:oQg
63KhMooooomTMP08000002D0000<00001P0002P0000<00001`0005@0001D00000`4004L0000=0@00
G0000040003UMEM1da9G@Ld100200@000@0004`00000000000000000003ooooooooooe00000ml000
d`000580001`0@001`00083nool0000000000000002l0P000@0000L2011D06T0K@1U07<0801>06D0
M`0P0580K`1]0640KP0000000000000000000000000000000000000000000000000C053b4`2Dk180
EQSZM`004`0200000?DW<a00000`lQ<00@000040001djQ80I>XB0000003U>?Qg000C03Se9c<00000
@>XB08P64`3djQ80UB_hMb0in7Oooooo1>/B042go7OX1Q<0@?DW<d3e9c?8T1d0@?DW<lSZ4P000?Qg
63KhMooooooHjQ80PW?XM`004`000000000C07L52XE@lQ<0`5OgMdd0J@0006<00P000640M02TjQ80
000009C/4P1F6>Yg000C00800000mBLc4000053b4`0100000@0007CZ4P1D0000?>/B09D[n7LH=_Qg
ooooofAf00P000009@0000`000070000:00000`000060000E00005@0000B0@00B@0001l1001M0000
0@000>EeEd7C4UM13@<0080100010000C00000000000000000000?ooooooooooD00005H0003o0000
DP00070100060000400000L00000000000000;`2000000001`828U<0N@1c07@0I@1]000000000000
00000000000000000000000000000000000000000000000000000000000001<0<?8C09C/4P1F6>Yg
000C00800000mBLc4000053b4`0100000@0007CZ4P1TjQ8000000>Dhn7L001<0>?DW<`000010jQ80
R0HC0?CZ4P2E:oQg83WhMoooool4ja80@;OlMnP64`10mBLc@?DW<aBA7@10mBLcb>XB0000n7LH=_Qg
ooooomSZ4P22LnQg000C0000000001<0j0D:VS3b4`30EoMgC@1Y0000H`020000H@1d0:CZ4P000000
U>`B05HHjWL001<00P00003e9c<@0000<?8C004000010000M>XB05@0000lja80UB_hMaPfn7Oooooo
I7H02000000U0000300000H0000X0000300000L0000U0000300000L0080U000030000000080U0000
300000<0001;0000@000030000050000800000400001000040000>d0001700009`400600003]0000
A`0002L1001P0000:00000`000050000:00000`0000600008P0000`0003oooooAP0001@000080000
AdA9@`<0000R000030000?oooolR000030000?oooolU0000300000d0080X0000300000@0000X0000
300000<0000R000030000?oooom60000804001@10015CDH[9T00010000040000000002Y0000T0000
600005b?XT800000000005b?XT9Ikn52dg9O@`Q01PI40000>000004@`=_T>0hn0000000000000000
3`0005@0B@1=04D0D`0P04h0A@1G0200DP1?04d0@@1>0000=T06P9P0002<00000000o`4000010000
2P000480K`1T07T080160780H@1]06D0d[2a?_VFKch;7n4nnII_?Z9F0SoiUVln_QdD?oVFKckKi2Do
nII_?VW8;SoiUVlnXXi2?oVFKcjnIDhonII_?[h]GSoiUVln_]Ei?oVFKch0080o000000000000080o
00000000000B000030000040000H000030000000000F0000300001P0000X000030000040001B0000
L0400040003eoooo0000000000000000T0400000000700@0E01Y06d0I@1c0200CP1U07L0801B06l0
K@1Q06h0000004d0K`1^06l0M01i0700I@0j05@0J@1]06D0L`0P0000ooooo`00000006/U061D]@L1
8;<70E`Y500f7@1P8;<70E`Y50087P1PG2TD0000000000005T<0H5`Y500P/`L1324D0<1PCV0HkA80
000C00@0000000000@0000000001kA807>hB09c/4P2PVB`cI>dB09D[n7LH=_Qg003oogC]4P22LnQg
000C0000003XHQlcYPD:4^QR7c?0EoMg000000`X50000000_QX0H43]4P2PVB`cE?LB0000jWN`MNQg
ooooo`40001?OOAgj68O<`P0003P=PDc/>dB0>QR7c?41oEg`5OgMiP58El100003CcdMl1GmgMTMP08
000002D0000<00000@0005@000280000S@000>L000350000m000004000000?Y1003j@Hd0003b0000
2P0004`00000000000000000003oooooooooof00001206l0I01i0200AP1b0640K@1U00P000060000
1P0000H0000300001P0000@0000500002@000000000U0000300000d00816000000H00?@50015CDH[
:d0000`0000000009D00010000040000000002a0000T0000600003/OIS`000200000P8Wa/ScSNYi3
:f@h@`Q020F41@00N0D0004@`=/200000`0006P500010000K00001X000000000BP400=P100000000
000002HY0000@00084E=AP000@1X1@008`0000<000000000000000000000100000<00401003`0000
00000000000000000>84082Y0`0I000030000?ooo`0B000030000080000U000030000000080U0000
300000d0080U0000300000L0080H000030000000000F000030000000000Q0000200001h0000H0000
00000000001A0@0030800140000<0000200000`0000@000000000000000;000040000541000<0P00
600000`0000000004P0000`0000100006@0000`0003oool0600000`0000000005P0000`0000H0000
2P00010000000000000000T0000@000080400<01000H000030000000001B0000L04000400020o_oo
0000000000000000T040004000800P1@D`1i06d0HP1_06`000000000000000000000000000000000
00000000000000000000000000000000000000007@0`jQ80P0h00>c[4P2E:oQg^3WhMoooooolja80
9k;lMaP74`000000@0L1=P00001605@0E`110580A@1L04<0K01Q07<0L`1U07<0G01304`0D`1904@0
G01k0300<00`0380@`150300<`0]0300<00`0300;@0`0300000`02d0@`0`0300<00]0300<00`0300
<00`0300<00`0300=00f07d0G01=06T0L`1S05<0M01Q07@0M@1c0000O@000:Eg002U=P00001`ja80
_NNUM`800h0XXBlcK0fVMl9Hm7L90000h3H5<gS[4P3lg_Pb`0KeM`40001TMP08000002D0000<0000
0@0005@0001D00006P000000001:0@00f0400040001aA192TS@B@Q/000100@000@0004`000000000
00000000003ooooooooooe00001Rl000d`000580001`0@000P0001000007000000000000002l0P00
000000420R9C07T0L`1d06D0K@000000000000000000000000000000000000000000000000000000
0000000000000000001c0000O@000:Eg002U=P00001`ja80_NNUM`800h0XXBlcK0fVMl9Hm7L90000
W>XB0000003U>?Qg000C09Q6:S<00000N>XB08P64`0/ja80UB_hMb0in7Oooooo?>/B042go7OX1Q<0
X4HZ<j16:S=jgoPbX4HZ<c00<0000300;@130300<00`02d0<00`0300<00`0300000C0300<00d03H0
O@1L04d0J@0006<00P000640M01e07<0001m0000YGL00:Df000C00800000ARXc400002RQ;c<10000
0@000:cZ4P1D0000M>/B09D[n7LH=_QgooooofAf00P000009@0000`000020000:00000`000010000
9@0000`0000700209@0000`000000020<00000`0000?00209@0000`0000=0020B`000400000`0000
1@000200000100000@0001000000000000000541000<0P0000000000001A0@00308002P0000<0000
0P000280000<0000ooooo`h0000D000000000100000D00006T08@2`0000P0000ooooo`8000000000
00000000ZD<0@0=400000>P3j0<Q000020000680000<00000@000240000800008@0000P0001c0000
30000000000L0000200002D0000<00000000P2D0000<00001`00P2D0000<00003P00P300000<0000
3`00P1T0000<0000oooo01P0000<000000000180000<00000P0001<0000<00000@0001@0000<0000
3@0001D0000<00000@0001H0000<0000000000d0000@000000000000000j0000300000X0000K0000
40000000000000008`000200003emQdm00000000000042@m082N@`00>D<T000090000?Gf7Cd00000
0000000@93d0P9i3000i@`@0001c000030000000000=0000400003d1002i0000DP00070100030000
400000L00000000000000;`2000000001`828U<0N@1c07@0I@1]0000000000000000000000000000
000000000000000000000000000000000000000000000?ag`USdM`40000Q08X1000:03S=;c=[9@1P
E;D70B2c1`5L:A@0=Qd0H22c1`5L:A@021h0H5`Y50000000000001I3061L:A@08;<70@`Q5030H4iP
X>DB022c1`4@@P1P00000040001[9@1Pb>DB00D0001`>OQg001<0000C0050000T;nG=J3U4P20?000
G>LB09D[n7Nh>OQgooooofcW4P0W/_agN0M<0000003ooooo7a00N04000000000c>HB0?SV4P1L:A@0
3>HB02Ll060<8A@0n>HB000000010000A>HB0<<l060<8A@0n>HB0000000<:1@0I7H02000000U0000
300000<000160000:00001`00017A4U30P00000000000000D@4000`2000000006@0000`0003oool0
4P0000`0000200009@0000`000000020DP00070100040000400000L00000000000000;`200000000
1`828U<0N@1c07@0I@1]000000000000000000000000000000000000000000000000000000000000
000000000000001PG2TD0000000000005T<0H5`Y500P/`L1324D0<1PCV2PiA808;<70A1206000000
0@0006/U0638iA801@00070in7L004`0001<00D0002@_iLeX>DB080l001Lia80UB_hMkPin7Oooooo
K>LB02Nbo7Mh1d`000000000oolO401h0@000000003<iQ80n>HB05`Y500<iQ809c`0H0`Q503hiQ80
000002D0003P=PDcl>DB010000000?Eg0@0001T0/04@0901G`WeMiP58El000209@000100T07oo`00
400001SV4P3j7OAgV0DQGa00T04U0000I7H02000000U0000300000@0000U0000300000L0080H0000
30000000000F000030000000000Q0000200001h0000H000000000000001A0@00308002<0000P0000
mOHM?@0000000000010T?@20WT<003U3900002@0003emQdm00000000000042@m082N@`00>D<40000
8`000200003emQdm00000000000042@m082N@`00>D<T000090000?Gf7Cd000000000000@93d0P9i3
000i@`@0000S0000800001WbCd4000000000063XYd40P9i3000i@b@0000T00006O9?@@0000000000
H>RW@@20WT<003U3100001P0000<000000000180000<00000@0001T0000<0000oooo01P0000<0000
000001H0000<0000600002<0000P00006O9?@@0000000000H>RW@@20WT<003U3900002@0000IlTm1
00000000001Pj:M1082N@`00>D<400008`000200000becPm00000000003oi3lm082N@`00>D<T0000
900003;G>3d0000000000?oT?cd0P9i3000i@`@0000H000030000000001B0000L04000H00020o_oo
0000000000000000T040004000870P1@D`1i06d0HP1_06`000000000000000000000000000000000
00000000000000000000000000000000000000007@0`jQ80P0h00>c[4P2E:oQg^3WhMoooooolja80
9k;lMaP74`000000@0L1=P00001605@0E`110580A@1L04<0K01Q07<0L`1U07<0G01304`0D`1904@0
G01k0300<00`0380@`150300<`0]0300<00`0300;@0`0300000`02d0@`0`0300<00]0300<00`0300
<00`0300<00`0300=00f07d0G01=06T0L`1S05<0M01Q07@0M@1c0000O@000:Eg002U=P00001`ja80
_NNUM`800h0XXBlcK0fVMl9Hm7L90000h3H5<gS[4P3lg_Pb`0KeM`40001TMP08000002D0000<0000
1P0005@0001D0000?P400;T000190@00b`000040001aA192TS@B@Q/000100@000@0004`000000000
00000000003ooooooooooe00001Rl000d`000580001`0@001@0001000007000000000000002l0P00
000000L20R9C07T0L`1d06D0K@000000000000000000000000000000000000000000000000000000
0000000000000000001c0000O@000:Eg002U=P00001`ja80_NNUM`800h0XXBlcK0fVMl9Hm7L90000
W>XB0000003U>?Qg000C09Q6:S<00000N>XB08P64`0/ja80UB_hMb0in7Oooooo?>/B042go7OX1Q<0
X4HZ<j16:S=jgoPbX4HZ<c00<0000300;@130300<00`02d0<00`0300<00`0300000C0300<00d03H0
O@1L04d0J@0006<00P000640M01e07<0001m0000YGL00:Df000C00800000ARXc400002RQ;c<10000
0@000:cZ4P1D0000M>/B09D[n7LH=_QgooooofAf00P000009@0000`000050000:00000`000060000
9@0000`0000700209@0000`0000000209@0000`000040000B`000400000`00001@00020000010000
0@000100000m0@00^@0004X1003>0000?@400;T0001:0@00cP0002P0000<00001@000280000<0000
ooooodH0000D0000200004M4BD<300008P0000`0003ooooo8P0000`0003ooooo9@0000`0000=0020
:00000`000030000:00000`0000400008P0000`0003oooooAP000>`5003P1@00ADe6:bI0000@0000
10000000000U@000400000@000000000;40002@0000H0000[XMo?00008000020RO6b?5W2Yd=^OHm3
24091F`5001P1@000A30f`8000030000D0D00040001/00007P00000000130@00oP40000000000000
/Rd00010000PADe600010505000Q00000`000000000000000000000400000`00@0400?0000000000
000000000000hP@0P:T301T0000<0000oooo0180000<00000P0002D0000<00000000P2D0000<0000
3@00P2D0000<00001`00P1P0000<0000000001H0000<000000000240000800007P0001P000000000
000007H1000<0P004@0000`0000800003000010000000000000000/0000@0000MP4000`2000H0000
30000000000B000030000040000I000030000?ooo`0F0000300001P0000:00004000000000000000
2@00010000100@00`0400580001`0@000@00083nool0000000000000002@0@000@000P02051C07T0
K@1R06l0K00000000000000000000000000000000000000000000000000000000000000000000000
000M0:3Z4P203P00G>`B09D[n7Nh>OQgooooofc/4P0W/_ag60LC0000003P2P4f000004H0E01G0440
DP1505`0@`1/0640L`1c06D0L`1L04<0C01C04T0A01L07/0<00`0300<P1304D0<00c02d0<00`0300
<00]0300<0000300;@130300<00`02d0<00`0300<00`0300<00`0300<00d03H0O@1L04d0J@1c06<0
D`1d0640M01e07<0001m0000YGL00:Hf00000>3[4P2mijEg0P03P32U;c=/3JIg`USdM`T0003P=PDc
j>/B0<8U7C?01_Eg0@0006Af00P000009@0000`000010000E00005@0000N0000000004<1003n0@00
0@00000H4T:B=1926P00060100010000C00000000000000000000?ooooooooooD000077`00370000
DP00070100020000400000L00000000000000;`2000000000@828U<0N@1c07@0I@1]000000000000
00000000000000000000000000000000000000000000000000000000000007<0001m0000YGL00:Hf
00000>3[4P2mijEg0P03P32U;c=/3JIg`USdM`T0000<ja8000000>Dhn7L001<0n0LM<`00003XjQ80
R0HC09c[4P2E:oQg83WhMooooon/ja80@;OlMnP64`0021dc00PM<chV7C<021dc<00`0000<00]04<0
<00`0300;@0`0300<00`0300<00001<0<00`03@0=P1m05`0C@1Y0000H`020000H@1d07D0L`0007d0
002UM`00YSH001<00P0000087C<@0000<:D_<`40000100007>/B05@0003Tja80UB_hMaPfn7Oooooo
I7H02000000U000030000080000X000030000040000U0000300000L0080U000030000000080`0000
300000l0080U0000300000d0081;0000@00003000005000080000040000100004000000000000000
MP4000`200000000000007H1000<0P00:00000`0000200008P0000`0003ooooo3P0001@000000000
400001@0000J@0U0;0000200003ooooo0P00000000000000082k@`100d@00000j0?X0b4000080000
HP0000`0000100008@0000P0000Q0000200007<0000<0000000001`0000800009@0000`000000020
9@0000`0000700209@0000`0000>0020<00000`0000?00206@0000`0003oool0600000`000000000
4P0000`0000200004`0000`000010000500000`0000=00005@0000`0000100005P0000`000000000
3@00010000000000000003X0000<00002P0001/0000@000000000000000S00008000070Q93d00000
0000000@93d00:Q3082?@b@0000T0000L24T?@0000000000010T?@00Z4<0P8m3100007<0000<0000
000000d0000@0000D04001l1001B0000L04000@0000@00001`00000000000000_080000000070P8R
D`1i07<0M01U06d00000000000000000000000000000000000000000000000000000000000000000
00000000<C?:1@[446Pa<l1GmgNH1B5O402@0BD0002a>OAgV>DB060a6004ia80EQSZMk1ej7Oooooo
0@0004M4BD<80000:>HB03cV4P1KVoIg0@0008BIn3800240O9Wh<R1MmgL00000h0[F=@0000000000
000003cV4P010000000QGccV4P21@OAgh0[F=N0f1C=4iQ80dYgfMiP58El80000X2Ok<TM4BD<30000
c>HB0>G;mWNH1B5O4>PB0000mWN8iQ8000000>Dhn7L001<026lO<`00001TiQ80R0HC01SW4P2E:oQg
83WhMooooolXia80@;OlMdP74`1TMP08000002D0000<0000100004H0000X0000700004M4BD<20000
00000000001f0@0030800000000I000030000?ooo`0B000030000080000U000030000000081B0000
L04000<0000@00001`00000000000000_080000000070P8RD`1i07<0M01U06d00000000000000000
00000000000000000000000000000000000000000000000000000000ool10000AdA9@`P0000XiQ80
?>HB05^KmWL10000Q9Wh<P008@1lVOPb85ggM`00003P2]He0000000000000000?>HB00400000025O
?>HB0851m7OP2]Heh3H5<dCV4P3BWOIgV0DQG`P0002P9o/bAdA9@`<0003<iQ80003fMiP58El@j180
003fMhSV4P000000iCShM`004`08Kalc000006CV4P281Q<09@000>0f1C=PiQ8040000000mGL10000
6@2`0A00T05O2OEgV0DQG`00080U0000402@0Ooo000@0000R>HB0?XMm7NH1B5O402@0BD0001TMP08
000002D0000<00000`0002D0000<00001`00P1P0000<0000000001H0000<00000000024000080000
7P0001P000000000000007H1000<0P008`000200001`8B@m00000000000042@m002X@`20Sd<T0000
9000070Q93d000000000000@93d00:Q3082?@`@0000S00008000070Q93d000000000000@93d00:Q3
082?@b@0000T0000L24T?@0000000000010T?@00Z4<0P8m3100002<0000P0000f/Q_@@0000000000
H>RW@@00Z4<0P8m3900002@0003Jb6m100000000001Pj:M1002X@`20Sd<40000600000`000000000
4P0000`0000100006@0000`0003oool05P0000`0000H00008`000200003Jb6m100000000001Pj:M1
002X@`20Sd<T000090000=[8Kd4000000000063XYd400:Q3082?@`@0000S000080000>;C?cd00000
00000?oT?cd00:Q3082?@b@0000T0000h]<o?@0000000000on@o?@00Z4<0P8m310000580001`0@00
1@00083nool0000000000000002@0@000@000PL2051C07T0K@1R06l0K00000000000000000000000
000000000000000000000000000000000000000000000000000M0:3Z4P203P00G>`B09D[n7Nh>OQg
ooooofc/4P0W/_ag60LC0000003P2P4f000004H0E01G0440DP1505`0@`1/0640L`1c06D0L`1L04<0
C01C04T0A01L07/0<00`0300<P1304D0<00c02d0<00`0300<00]0300<0000300;@130300<00`02d0
<00`0300<00`0300<00`0300<00d03H0O@1L04d0J@1c06<0D`1d0640M01e07<0001m0000YGL00:Hf
00000>3[4P2mijEg0P03P32U;c=/3JIg`USdM`T0003P=PDcj>/B0<8U7C?01_Eg0@0006Af00P00000
9@0000`000050000E00005@0001A0@007`4005`1000c0@000@00000H4T:B=1926P00060100010000
C00000000000000000000?ooooooooooD000077`U0?70000DP00070100060000400000L000000000
00000;`2000000001`828U<0N@1c07@0I@1]00000000000000000000000000000000000000000000
000000000000000000000000000007<0001m0000YGL00:Hf00000>3[4P2mijEg0P03P32U;c=/3JIg
`USdM`T0000<ja8000000>Dhn7L001<0n0LM<`00003XjQ80R0HC09c[4P2E:oQg83WhMooooon/ja80
@;OlMnP64`0021dc00PM<chV7C<021dc<00`0000<00]04<0<00`0300;@0`0300<00`0300<00001<0
<00`03@0=P1m05`0C@1Y0000H`020000H@1d07D0L`0007d0002UM`00YSH001<00P0000087C<@0000
<:D_<`40000100007>/B05@0003Tja80UB_hMaPfn7OoooooI7H02000000U0000300000H0000X0000
300000D0000U0000300000L0080U000030000000080U0000300000<0001;0000@000030000050000
800000400001000040000501000O0@00G`4003@1001@0@007`4005l1000d0@00:00000`000060000
8P0000`0003oooooAP0001@000080000AdA9@`<0000R000030000?oooolR000030000?oooolU0000
300000d0080X0000300000@0000X0000300000<0000R000030000?oooom60000C04004010015CDH[
9T00010000040000000000Q00P8l0000<000004@`=/00000CP000000002Y4e0o0P00008000020000
0P00004@`=/000000000o`Q00@?P0000e000004@`=/F00000000063V=d>4aSa4R4>d@hC6?4C34KM3
PlHlA09H^D=TXc]40URi@dHl>T@2F;U3AS`jA09H^D=5?3Y40URi@dDl>T@3F;U3AS`jA0=H^D>BWJ]3
0URi@eG?Z4?44KM35hVV@hQ3]4<GRJI3H>Hg@aN9YT?WBC935hVV@fZm;D=EcjQ3J[d]@i6MZd=Z_Be3
TYf[@fZm;D=6?3Y4JKd]@fBS>dCWBC93PlHlA5oV=d>3aSa400430`<30`<10@<30`430`<10@<30`00
5D010100000400000P0002@0000T00000020?@00000000000020?@00000000000P0005l0000h0000
0`0003P000000000>0000000000000403@0000000000000000000000000000009@0000`000030000
9@0000`000050020>`0000P0000K0000400007l;000b;`00=P00010000295P00<Rl005P0000d0000
00000000003oooooooooo`H0003S5S8_:aOY;R/GT2h[5i0^:aN@;R/GT2iI00009000000000000000
ooooooooool20000:aN@;R/GM1EH0000:000000000000000ooooooooool30000:aLJ5N<FdQB95]8D
=P000100001o2`00dQ@005P0000X000000000000003oooooooooo`<0000U2m8Dg0XJ5M`:M1EI0000
9000000000000000ooooooooool20000g0Yd5M`:T2iH0000:000000000000000ooooooooool30000
g0[Y;RD;<Rmo2c8_?00000P00010000060000:`0001;0@00M@400?D2000U0000300000L0080U0000
30000000080T000090000000P440000000000000P440000000000080000X0000300000<000160000
R00007`00015CDH[24020S`0000`00000A30f`00001>000000000<JgRSh200000P00008000020000
0A30f`000000003o24010b`0000P00000A30f`8000000000G1^8@l=gV4=L6hQ3;/m3A001000E@040
400000@000020000900002@00000080m000000000000080m0000000000020000G`0003P000030000
>0000000000h0000000000000@040000000000000000000000000000000U0000300000<0000U0000
300000D0081G0000900000l1000_0@004P40014300020000114?4`@Am30U0000300000L0080U0000
30000000080T000090000000P440000000000000P440000000000080000X0000300000<000160000
b0000;`00015CDH[24010jP0002L00000A30fa00000000001YN4@l4N^4<6UhA3Pbff@ahZQT=[V[A3
Fa^8@f^J]4>I38Y3JiZd@k2ORd>3;KI3/In;@l0N^4>aWh]3`1jh@k6ORd?17[Q3/In;@l4N^4>aWh]3
oPnj@iT<RT<FXk]3G1^8@aJS^d<N:XI35Z>k@`NGQ4?n3kY31YN4@l4N^4<00`<30`<30`<30`<30`>3
5401P100000400000000ob@0000T00000020?@00000000000020?@00000000000P0002D0000<0000
0P0001<0000<00000@0003/000080000E@0005`00000000000000?oooooooooo400009<@11NC4<HF
aQ2D5P@AU1I24I@FM1765W@A11Md4@@GM1445g@A11Md4D8G@Q5e5`@AMAO647DGTa125i<@11Ll0000
200003h0000H00002@4006T1000H0@00N04001<0000<00000@0002D0000<00000000P2@0000T0000
0020@@00000000000020@@00000000000P0004H0001L0000D00004E=AR/8@082?000030000014<3K
000004h000000000ZA=@?`80000200000P00008000014<3K000000000?lE@040400000@000020000
900002@00000080m000000000000080m0000000000020000G`0003P000030000>0000000000h0000
000000000@0=0000000000000000000000000000000U0000300000<0000U0000300000D0081E0000
G00000L1001W0@006@4007T1000@0000Ta045i<@aQK649@F116D5T8AU1Id4LHFM1445g@A11Md4@@G
M1445g@A@QM24GDG115e5lH@MANC448GTa045bD0000<00001`00P2D0000<00000000P2@0000T0000
0020@@00000000000020@@00000000000P0002P0000<00000`0004H0001`0000I00004E=AR/8@043
D00004@000014<3K1P000000001];lm3>1XTA2<</T<h6RA4KBo?@cPJ94A];lm3`Ajh@ji^V4?17[Q3
KBo?@l4N^4<00H400H4001A00H0@000010000?oooolX000030000080000T000090000000P3d00000
00000000P3d0000000000080000W00006000008000000000oooo0000000U000030000080000C0000
30000040000U0000300000P0081K0000@0000301001`0@00W`400942000200001P0000<000030000
iQT7:D8F1bWV6@LYiQT45`hC11OV6@@G9@0000`0000700204`0000`0000100009@0000`000000020
900002@00000081100000000000008110000000000020000AP000=0000340000ADe6:`Q00P8l0000
<000004@`=/00000CP0000000036]hXn0P000080000200000P00004@`=/000000000o`Q00@=d0000
J000004@`=/:0000000006d_cd<h6RA48`bb@cPJ94A];lm3>1XTA6d_cd?17[Q3[VjH@l4N^4=];lm3
`Ajh@g:[bT?<5014LZ_:@nN48TAbZlY3c1@0A7:[bT=RBK]30041004100400@005D01010000040000
0P0002@0000T00000020?@00000000000020?@00000000000P0005l0000h00000`0003P000000000
>000000000000040100000000000000000000000000000009@0000`0000300009@0000`000050020
FP0005P0000_0@00K`400:01002B0P00100000X0000300000`00008000020000iQT7:D8F1bWV6@LY
iQT45`hC11OV6@@GEQT685HIXRQF6@HPEQUZ5bD0000<00001`00P2D0000<00000000P2@0000T0000
0020@@00000000000020@@00000000000P0002P0000<00000`0004H0001L0000D00004E=AR/8@043
?000030000014<3K10000000000hd/]33f0RA7:[bT<h6RA4ZhC9@`mP8T@hd/]33f0RA0010H4D@060
400000@00000003o:00000`000020000900002@00000080m000000000000080m0000000000020000
9`0001P00002000000000000000000009@0000`0000200004`0000`0000100009@0000`000080020
EP0002`0002C0@00R@8009P1002A0P00100007/IV2QF6@LY<AVH:7/IV2PU0000300000L0080C0000
30000040000U000030000000080T000090000000P440000000000000P44000000000008000160000
G00005000015CDH[24010c`0000`00000A30f`@000000000ZhC9@aBC^d=bZlY3`Ajh@cSBbd<DTk]3
ZhC9@aBC^d<00@615401P100000400000000ob@0000T00000020?@00000000000020?@0000000000
0P0002D0000<00000P0001<0000<00000@0002D0000<00002000P5H0000/0000T`400701002H0@00
N04000@0000a6G<GEQT45g/ILaLa6G<G9@0000`0000700204`0000`0000100009@0000`000000020
900002@00000081100000000000008110000000000020000AP0008P0001l0000ADe6:`Q00P8l0000
<000004@`=/00000CP000000002Y4e0o0P000080000200000P00004@`=/000000000o`Q00@</0000
8000004@`=/2000000000;aeFd?lP^m3dH:5@eJ3_4<00@005D010100000400000P0002@0000T0000
0020?@00000000000020?@00000000000P0005l0000h00000`0003P000000000>000000000000040
3@0000000000000000000000000000009@0000`0000300009@0000`000050020E`0002@0003J0000
M`4000d1003Q0@000P000;P=lAfa494G9@0000`0000700209@0000`000000020900002@000000811
00000000000008110000000000020000:00000`000030000AP0009P0002<0000ADe6:`Q00P8l0000
<000004@`=/00000CP000000002Y4e0o0P000080000200000P00004@`=/000000000o`Q00@<l0000
<000004@`=/40000000008o=KT<m3=e38Mac@ed5hD<f`7a33BOT@i@_Pd=YFNE300<30aE00@0@0000
10000080000T000090000000P3d0000000000000P3d0000000000080001O0000>00000<0000h0000
000003P000000000000100d0000000000000000000000000000002D0000<00000`0002D0000<0000
1@00P5D0000/0000k@000;P100080@00c@4000@0003]3Z8K?PlQ7<`?QAaV4:`L9@0000`000070020
9@0000`000000020900002@00000081100000000000008110000000000020000:00000`000030000
AP0005`0001@0000ADe6:`Q00@<l0000<000004@`=/4000000000;/3Pd?DN>=3<UV6@a;GiD<OM893
6AKW@k/3Pd?DN>=30041PAA00H0@0000100000000?lT000090000000P3d0000000000000P3d00000
00000080000U000030000080000C000030000040000U0000300000P0081F0000;00000@100370@00
3@400<l100040000HA1`7<`@^aa?4><LHA1`72D0000<00001`00P1<0000<00000@0002D0000<0000
0000P2@0000T00000020@@00000000000020@@00000000000P0004H0003@1`00a0L004E=AR/U@000
400000@000000000;40002@0000H0000CcfV?00008000020RO6b?6D;KD=Xo>Q3240:1F07001D1`00
0A30f`8000030000A0L00040001/000010000000002a0@002`80000000000000M3/00010000PADe6
000104@7000W00000`000000000000000000000400000`00@0400?0000000000000000000000hP@0
P:T301T0000<0000oooo0180000<00000P0002D0000<00000000P2D0000<00003@00P2D0000<0000
1`00P1P0000<0000000001H0000<000000000240000800007P0001P00000000000000>L1000<0P00
4@0000`0000800003000010000000000000000/0000@0000i`4000`2000H000030000000000B0000
30000040000I000030000?ooo`0H000030000000000F0000300001P0000:00004000000000000000
2@000100002P0@00`04001P0000<000000000580001`0@000@00083nool0000000000000002@0@00
0@000P02051C07T0K@1R06l0K0000000000000000000000000000000000000000000000000000000
0000000000000000000M0:3Z4P203P00G>`B09D[n7Nh>OQgooooofc/4P0W/_ag60LC0000002`304f
000004H0E01G0440DP1505`0@`1/0640L`1c06D0L`1L04<0C01C04T0A01L07/0<00`0300<P1304D0
<00c02d0<00`0300<00]0300<0000300;@130300<00`02d0<00`0300<00`0300<00`0300<00d03H0
O@1L04d0J@1c06<0D`1d0640M01e07<0001m0000YGL00:Pf00000>3[4P2mijEg0P03P3RY;c=/3JIg
`USdM`T0003P=PDcj>/B0?BT9C?01_Eg0@0006Af00P000009@0000`000010000E00005@000040000
00000281003H0@000@0008YE4T:B=1920`00040100010000C00000000000000000000?oooooooooo
D00006C`002n0000DP000701000200008?ooo`00000000000000090100000000008045@0J@1]06D0
L`0P04h0I@1g0200DP1_06d0H@1^00000000000000000000000000000000000000000000000007<0
001m0000YGL00:Pf00000>3[4P2mijEg0P03P3RY;c=/3JIg`USdM`T0000<ja8000000>Dhn7L001<0
Z84E0000003XjQ80R0HC09c[4P2E:oQg83WhMooooon/ja80@;OlMnP64`2`PAD0/84E07:U9C>`PAD0
<00`0000<00]04<0<00`0300;@0`0300<00`0300<00001<0<00`03@0=P1m05`0C@1Y0000H`020000
H@1d07D0L`0007d0002UM`00Z3H001<00P0000215@0@0000>:T_<`40000100007>/B05@0003Tja80
UB_hMaPfn7OoooooI7H02000000U000030000080000X000030000040001D0000E00002h1003n0000
/@4000/200010000RUDB@Y8d4T820@00X0400040001<00000000000000000000oooooooooom@0000
=0000700001B0000L0400040000@00001`00000000000000_080000000010P8RD`1i07<0M01U06d0
000000000000000000000000000000000000000000000000000000000000000000000000;3<XYR`c
002UM`00Z3H001<00P0000215@0@0000>:T_<`40000100007>/B00c[4P000000iCShM`004`2XPAD0
00000>SZ4P281Q<0W>/B09D[n7LP>OQgooooojc[4P10]oagj0HC0;215@2`PAD0_ZDU<k215@1`ja80
003hMaPfn7OoooooP>/B089cj7L001<0000000004`1l1@ZK::H/<l1GmgM=06T0001S0080001Q07@0
C>/B0000:3<lkA80EQSZM`004`020000084E0100000XYR`c0@000040000Lja80E0000>C[4P2E:oQg
63KhMooooomTMP08000002D0000<00000@0002P0000<00000P0002D0000<00001`00P2D0000<0000
0000P300000<00003`00P2D0000<00003@00P4/000100000<00000D0000P00000@000040000@0000
00000000003W0@003080000000000000i`4000`2000X000030000040000R000030000?ooool>0000
50000000000@0000500001Y02T0/000080000?ooool200000000000000000?A30403A000003X0nP3
8@0000P0001R000030000040000Q00002000024000080000L`0000`000000000700000P0000U0000
30000000080U0000300000L0080U0000300000h0080`0000300000l0080I000030000?ooo`0H0000
30000000000B000030000080000C000030000040000D0000300000d0000E000030000040000F0000
30000000000=00004000000000000000>P0000`0000:00006`00010000000000000002<0000P0000
;LXO?@0000000000010T?@00KT<00>U3900002@0000]bQlm00000000000042@m001^@`00jD<40000
L`0000`0000000003@000100003^0000dP400580001`0@000`0001000007000000000000002l0P00
000000L20R9C07T0L`1d06D0K@000000000000000000000000000000000000000000000000000000
0000000000000000000QG`000000000003_i<X3X4P3h>_TbJbD0H5Be1`4P/`L1K3XD03HM060P/`L1
K3XD00PN061/>Q@000000000000F@`1PK3XD022c1`6l>a@0`61>H13V4P0P/`L14480H00000010000
JbD0H3SV4P050000L3WhM`00C00004`01@00003;VcD@iQ80P3`00<cW4P2E:oQg^3WhMooooooLia80
9k;lMgP7C0000000oooooal@07P10000000003cW4P1Xia80K3XD07cV4P0W?01P_3/D06SW4P000000
0@000;CV4P33?01P_3/D06SW4P000000324D06Af00P000009@0000`000030000AP0002P0000L0000
AdA9@`800000000000000>L1000<0P00000001T0000<0000oooo0180000<00000P0002D0000<0000
0000P580001`0@00100001000007000000000000002l0P00000000L20R9C07T0L`1d06D0K@000000
00000000000000000000000000000000000000000000000000000000000000000000H6`j50000000
000001I3061/>Q@08;<70K`k5030H4iP4>HB022c1`4@@P1P00000040001[9@1P>>HB00D0001`>OQg
001<0000C00500000<^K=A3V4P20?000c>LB09D[n7Nh>OQgooooomcW4P0W/_agN0M<000000000?oo
7a00N04000000000?>LB06SW4P1/>Q@0O>HB02Ll062l>a@0J>LB0000000U0000h3H5<f3V4P0@0000
003eM`40000I0;01402@0El9mGNH1B5O0000P2D0000@0901ool001000028iQ80nQgdMiP58El@0901
9@0006Af00P000009@0000`0000400009@0000`000070020600000`0000000005P0000`000000000
8@0000P0000N00006000000000000000i`4000`2000S0000800002g:7cd000000000000@93d006i3
003Y@b@0000T0000;LXO?@0000000000010T?@00KT<00>U3100002<0000P0000;LXO?@0000000000
010T?@00KT<00>U3900002@0000]bQlm00000000000042@m001^@`00jD<400008`000200003>o9M1
00000000001Pj:M1001^@`00jD<T000090000<klUd4000000000063XYd4006i3003Y@`@0000H0000
30000000000B000030000040000I000030000?ooo`0H000030000000000F0000300001P0000S0000
80000<klUd4000000000063XYd4006i3003Y@b@0000T0000c_bG@@0000000000H>RW@@00KT<00>U3
100002<0000P0000``lk?@0000000000on@o?@00KT<00>U3900002@000333c/m00000000003oi3lm
001^@`00jD<40000600000`000000000DP00070100060000P?koo`00000000000000090100010002
1`80D5<0N@1]0680K`1/000000000000000000000000000000000000000000000000000000000000
00000000000001d0X>XB080>001Lk180UB_hMkPin7OoooooK>`B02Nbo7LH1a<000000;0<0CH00000
AP1D05L0@@1B04D0G01306`0H@1c07<0I@1c05`0@`1<05<0B@1405`0N`0`0300<00b04<0A@0`03<0
;@0`0300<00`02d0<00`0000<00]04<0<00`0300;@0`0300<00`0300<00`0300<00`03@0=P1m05`0
C@1Y07<0H`1C07@0H@1d07D0L`0007d0002UM`00Z3H00000h>/B0;gWYGL200>0>:T_<f`=YWO2F?Ag
2@000>0f1C?Xja80m:@U<l06mGL10000I7H02000000U0000300000H0001D0000E0000>h0003B0@00
n@000>@100010000RUDB@Y8d4T830000@0400040001<00000000000000000000oooooooooom@0000
I?000;h0001B0000L04000D0000Poooo0000000000000000T040000000070P0@E01Y06d0I@1c0200
CP1U07L0801B06l0K@1Q06h0000000000000000000000000000000000000000000000000L`0007d0
002UM`00Z3H00000h>/B0;gWYGL200>0>:T_<f`=YWO2F?Ag2@0000c[4P000000iCShM`004`2XPAD0
00000>SZ4P281Q<0W>/B09D[n7LP>OQgooooojc[4P10]oagj0HC0;215@2`PAD0LZDU<k215@0`0300
000`02d0@`0`0300<00]0300<00`0300<00`00004`0`0300=00f07d0G01=06T0001S0080001Q07@0
M@1c0000O@000:Eg002X=P004`020000084E0100000hZBlc0@000040000Lja80E0000>C[4P2E:oQg
63KhMooooomTMP08000002D0000<00001@0002P0000<00001P0005@0001D0000nP000=`1003o0000
iP400040002:EA92TS@B@P81002P0@000@0004`00000000000000000003ooooooooooe00000d0040
L0000580001`0@001P0001000007000000000000002l0P00000000L20R9C07T0L`1d06D0K@000000
0000000000000000000000000000000000000000000000000000000000000000000/<bRV;3<00:Eg
002X=P004`020000084E0100000hZBlc0@000040000Lja803>/B0000003U>?Qg000C0:R15@000000
j>XB08P64`2Lja80UB_hMb0in7Oooooo[>/B042go7OX1Q<0/84E0;215@2nYBDc/84E073[4P000?Qg
63KhMooooon0ja80PW?XM`004`000000000C07`52Y/XYR`c`5OgMdd0J@0006<00P000640M01<ja80
000X<cc]4P1F6>Yg000C00800000PAD0400002RV;3<100000@0001c[4P1D0000i>/B09D[n7LH=_Qg
ooooofAf00P000009@0000`000060000:00000`0000500009@0000`0000700209@0000`000000020
9@0000`000040000B`000400000`00001@000200000100000@000100003^0000dP400041003W0@00
kP000=8100010@00i`4002P0000<00001P000280000<0000ooooodH0000D0000200004M4BD<30000
8P0000`0003ooooo8P0000`0003ooooo9@0000`0000=0020:00000`000030000:00000`000040000
8P0000`0003oooooAP0009P0002<0000ADe6:bI0000@0000100000000008@082?000030000014<3K
000004h000000000ZA=@?`80000200000P00008000014<3K000000000?l8@043;000020000014<3K
0P000000001H_U13nKHSA0j]Yd>`VR=4004001E00@0@000010000080000T000090000000P3d00000
00000000P3d0000000000080001O0000>00000@0000h0000000003P000000000000100d000000000
0000000000000000000002D0000<0000100002D0000<00001@00P5L0000T0000c`0008d2001A0@00
T@800080000<3NhXmQCW:2D0000<00001`00P2D0000<00000000P2@0000T00000020@@0000000000
0020@@00000000000P0002P0000<0000100004H0003T0000f00004E=AR/8@043a0000;P000014<3K
4`0000000012f4U3LV@]A1jCBD>U31Y4HI59@eJ@6D@nBDi3PRXIA?<LE4<e:AU4ZO1I@nLW6DAFZei3
Wh/IA1>]GT?^1aY44jeN@nh76T@D[Ei3kPLJA1>]GT?^1aY45:eN@nl76T@hlUi3^el]A?CcGT<9g2e4
5caJ@me1;TAQJ5A3:T<^A:^DCT=hA2i4o]U9@l3P;DA2f4U3LF@]A0010`<30`<30`<30@430`<30h<0
5401P10000040000`<30obP0000<00000P0002@0000T00000020?@00000000000020?@0000000000
0P0002L0000H00000P0000000030`<00000002D0000<00000P0001<0000<00000@0003/000080000
6`000100002N3000FR/003H0000@0000VP`008@V001H0000@000000000000000ooooooooool90000
VPaU9^D<BbI23D/VX0e:9^/=HbK[3H8Vj`f29^/=PRK[3H8VF@0002@00000000000000?oooooooooo
0P000>/=PRK`3EP[F00003@00000000000000?oooooooooo1P000?0=Mb^T3I4[A`fA:nX<TR^N37T[
WPaJ:cd000080000?00000P0000n000060000<T0001T0P00g`000;X2000C000030000040000U0000
30000000080T000090000000P440000000000000P44000000000008000160000G00005000015CDH[
24020S`0000`00000A30f`00001>000000000<JgRSh200000P000080000200000A30f`000000003o
5D010100000400000P0002@0000T00000020?@00000000000020?@00000000000P0005l0000h0000
100003P000000000>000000000000040100000000000000000000000000000009@0000`000040000
9@0000`000050020>`0000P0000K0000400009h<001J:`00=P000100002J3000Q2H005P000100000
00000000003oooooooooo`T0002J36DVi@a;9T8=BbJP3DXVj`eS9^/=PRK[3H8Vj`f29^/=PRII0000
9000000000000000ooooooooool20000j`f29_0=F2]H0000=000000000000000ooooooooool60000
l0eg:j@=TB]73I4[jPbB:ih<NB^N35X[?@0000P0000l000020000400000H0000b00006<2003Q0000
^`8002D0000<00001`00P2D0000<00000000P2@0000T00000020@@00000000000020@@0000000000
0P0002P0000<0000100004H0003T0000f00004E=AR/8@043a0000;P000014<3K4`000000002_]J13
^dT]A1fCX4?_lAU4?i:P@ime6D@]kZ93b`lIA0SHYD=n3QU4hl6X@c0=6D@i7j]3j70IA1PPZd<gkAU4
622[@cO]6D@H8:]3=ndIA1PPZd<gkAU4622[@cS]6DBZ@Z]314D]A8Q3Zd=B`Be4V^NX@bHW;TBooJE3
LbP^A>@CXd?1:Bi4SKJP@`W6;DB_]J13^TT]A0010`<30`<30`<30@430`<30h<05401P10000040000
`<30ob@0000T00000020?@00000000000020?@00000000000P0002D0000<00000P0001<0000<0000
0@0003/0000800006`000100000G5000Db/003H0000@00004a@007dV001H0000@000000000000000
ooooooooool900004aAN9UhDA2Jk54@V6AE49V@EGBIT5G`VI1El9V@EO2IT5G`VF@0002@000000000
00000?oooooooooo0P0006@EO2IY5E8[F00003@00000000000000?oooooooooo1P0006TELB/M5HX[
`1B;:f<DRb/G578[5aAC:cd000080000?00000P0000n000060000441001T0P00E`400;T2000C0000
30000040000U000030000000080T000090000000P440000000000000P44000000000008000160000
G00005000015CDH[24020S`0000`00000A30f`00001>000000000<JgRSh200000P00008000020000
0A30f`000000003o5D010100000400000P0002@0000T00000020?@00000000000020?@0000000000
0P0005l0000h0000100003P000000000>00000000000004010000000000000000000000000000000
9@0000`0000400009@0000`000050020>`0000P0000K0000400001LD001C:`00=P000100000C5000
OBH005P00010000000000000003oooooooooo`T0000C55hVGQA49[/DA2HI5D@VI1EM9V@EO2IT5G`V
I1El9V@EO2II00009000000000000000ooooooooool20000I1El9VTEDR]H0000=000000000000000
ooooooooool60000JAEa:adERR_058/[HaB;:aLDLR/G55<[?@0000P0000l000020000400000H0000
@04006<2001H0@00^P8002D0000<00001`00P2D0000<00000000P2@0000T00000020@@0000000000
0020@@00000000000P0002P0000<0000100004H0003@1`00a0L004E=AR/U@000400000@000000000
;40002@0000H0000RO6b?00008000020W`;0?:X7h4?3jXe3240;1F07001D1`000A30f`8000030000
A0L00040001/00007P000000002a0@00nP40000000000000YC/00010000PADe6000104@7000W0000
0`000000000000000000000400000`00@0400?0000000000000000000000hP@0P:T301T0000<0000
oooo0180000<00000P0002D0000<00000000P2D0000<00003@00P2D0000<00001`00P1P0000<0000
000001H0000<000000000240000800007P0001P00000000000000>T1000<0P004@0000`000080000
3000010000000000000000/0000@0000j@4000`2000H000030000000000B000030000040000I0000
30000?ooo`0H000030000000000F0000300001P0000:000040000000000000002@00010000300@00
h04001P0000<000000000580001`0@000@00083nool0000000000000002@0@000@000002011D06T0
K@1U07<0801>06D0M`0P0580K`1]0640KP0000000000000000000000000000000000000000000000
000/<mSZ4P203P00U>`B09D[n7Nh>OQgooooojC/4P0W/_ag60LC000000000:Lf000004H0E01G0440
DP1505`0@`1/0640L`1c06D0L`1L04<0C01C04T0A01L07/0<00`0300<P1304D0<00c02d0<00`0300
<00]0300<0000300;@130300<00`02d0<00`0300<00`0300<00`0300<00d03H0O@1L04d0J@1c06<0
D`1d0640M01e07<0001m0000YGL00:Tf000001S/4P2mijEg0P03P42];c=/3JIg`USdM`T0003P=PDc
8>`B09cn8P301_Eg0@0006Af00P000009@0000`000010000E00005@0000N0000000003@1002g0@00
0@00099`249FM@Q26`00040100010000C00000000000000000000?ooooooooooD00007L0003o0000
DP000701000200008?ooo`00000000000000090100010000008045@0J@1]06D0L`0P04h0I@1g0200
DP1_06d0H@1^00000000000000000000000000000000000000000000000007<0001m0000YGL00:Tf
000001S/4P2mijEg0P03P42];c=/3JIg`USdM`T00014ja8000000>Dhn7L001<0b:XU<`00000Pja80
R0HC0=C[4P2E:oQg83WhMooooooTja80@;OlMnP64`3@ZRDcd:XU<bCo8P3@ZRDc<00`0000<00]04<0
<00`0300;@0`0300<00`0300<00001<0<00`03@0=P1m05`0C@1Y0000H`020000H@1d07D0L`0007d0
002UM`00ZCH001<00P00002Z9C<@0000@:d_<`4000010000E>/B05@0000Lk180UB_hMaPfn7Oooooo
I7H02000000U000030000080000X000030000040001D0000E00003L1003/0000/@400?X100010000
TW08@UIe248M0@00X0400040001<00000000000000000000oooooooooom@0000HP000700001B0000
L0400040000@00001`00000000000000_080000000010P8RD`1i07<0M01U06d00000000000000000
00000000000000000000000000000000000000000000000000000000;3<hX2`c002UM`00ZCH001<0
0P00002Z9C<@0000@:d_<`4000010000E>/B04C[4P000000iCShM`004`38ZRDc0000023[4P281Q<0
e>/B09D[n7LP>OQgooooonC[4P10]oagj0HC0=2Z9C?@ZRDcL?lR0=2Z9C>Xja80003hMaPfn7Oooooo
^>/B089cj7L001<0000000004`3R1@XH>:0/<l1GmgM=06T0001S0080001Q07@0Q>/B0000001dkA80
EQSZM`004`0200000:XU<a00000hX2`c0@000040001Dja80E00001c/4P2E:oQg63KhMooooomTMP08
000002D0000<00000@0002P0000<00000P0002D0000<00001`00P2D0000<00000000P300000<0000
3`00P2D0000<00003@00P4/000100000<00000D0000P00000@000040000@000000000000003Y0@00
3080000000000000j@4000`2000X000030000040000R000030000?ooool>000050000000000@0000
500001Y02d0/000080000?ooool200000000000000000?E30403A000003X0nP38@0000P0001R0000
30000040000Q00002000024000080000L`0000`000000000700000P0000U000030000000080U0000
300000L0080U0000300000h0080`0000300000l0080I000030000?ooo`0H000030000000000B0000
30000080000C000030000040000D0000300000d0000E000030000040000F000030000000000=0000
4000000000000000>P0000`0000:00006`00010000000000000002<0000P0000ljPW?@0000000000
0;0c?@20h4<008i3900002@0003cZ2Lm000000000000/3<m083P@`00ST<40000L`0000`000000000
3@00010000310@0070400580001`0@00100001000007000000000000002l0P00000000L20R9C07T0
L`1d06D0K@0000000000000000000000000000000000000000000000000000000000000000000000
00000<9Hm7L100008@2:0@002P0hcBlcF_3dM`d0000I0;018@2:0B400022LnQg000C0000001PJ34c
_@D:ff1X<C?0EoMgV0DQGa00T04U0000/CWdM`SV4P3X3b@cM>LB05HHjWN`MNQgooooo`400017A4U3
200009SV4P000180Fi_fM`400010ioPb000Q03SWn38PGOMg00000>0:eSD0000000000000002/iQ80
0@0009P58En/iQ80003dMn0:eSGP=PDc]>HB0=:MmWNH1B5O20000>QMnc97A4U30`0003cW4P3UboIg
V0DQGh3X4P12Q_Ign>HB06Af00P000009@0000`000040000AP0002P0000L0000AdA9@`8000000000
00000>T1000<0P00000001T0000<0000oooo0180000<00000P0002D0000<00000000P580001`0@00
0`0001000007000000000000002l0P00000000L20R9C07T0L`1d06D0K@0000000000000000000000
0000000000000000000000000000000000000000000000000000061X<C>m1@[KH6Pa<l1GmgNH1B5O
402@0BD0002a>OAg2>HB0>P?93=dia80EQSZMk1ej7Oooooo0@0004M4BD<80000V>HB00004P1KVoIg
0@00043Wn3800240>>Oh<R1MmgL00000h0[F=@000000000000000:cV4P000000V0DQGjcV4P000?Ag
h0[F=N0f1C>diQ80dYgfMiP58El80000j5gk<TM4BD<U0000h3H5<iSV4P0@0000003eM`40000I0;01
402@0El9mGNH1B5O0000P2D0000@0901ool001000030iQ80nQgdMiP58El@09019@0006Af00P00000
9@0000`0000300009@0000`000070020600000`0000000005P0000`0000000008@0000P0000N0000
6000000000000000j@4000`2000S000080000?>X9cd000000000002`<cd0P>13002>@b@0000T0000
ljPW?@00000000000;0c?@20h4<008i3100002<0000P0000ljPW?@00000000000;0c?@20h4<008i3
900002@0003cZ2Lm000000000000/3<m083P@`00ST<400008`000200003L8:1100000000000Pi[M1
083P@`00ST<T000090000=`PX44000000000023V]d40P>13002>@`@0000H000030000000000B0000
30000040000I000030000?ooo`0H000030000000000F0000300001P0000S000080000=`PX4400000
0000023V]d40P>13002>@b@0000T0000g22P@@00000000008>Jg@@20h4<008i3100002<0000P0000
n`0g?@0000000000ZbQ4?@20h4<008i3900002@0003k03Lm00000000002[:4@m083P@`00ST<40000
600000`000000000DP00070100060000P?koo`000000000000000901000100001`8045@0J@1]06D0
L`0P04h0I@1g0200DP1_06d0H@1^00000000000000000000000000000000000000000000000002`c
f>XB080>002Dk180UB_hMkPin7OoooooY>`B02Nbo7LH1a<000000000YcH00000AP1D05L0@@1B04D0
G01306`0H@1c07<0I@1c05`0@`1<05<0B@1405`0N`0`0300<00b04<0A@0`03<0;@0`0300<00`02d0
<00`0000<00]04<0<00`0300;@0`0300<00`0300<00`0300<00`03@0=P1m05`0C@1Y07<0H`1C07@0
H@1d07D0L`0007d0002UM`00ZCH000006>`B0;gWYGL200>0@:d_<f`=YWO2F?Ag2@000>0f1C<Pk180
W?hR0<06mGL10000I7H02000000U0000300000H0001D0000E0000<81000L0@00c@4002h100010000
TW08@UIe248K0000@0400040001<00000000000000000000oooooooooom@0000M`3oool0001B0000
L04000D0000Poooo0000000000000000T040004000070P0@E01Y06d0I@1c0200CP1U07L0801B06l0
K@1Q06h0000000000000000000000000000000000000000000000000L`0007d0002UM`00ZCH00000
6>`B0;gWYGL200>0@:d_<f`=YWO2F?Ag2@0004C[4P000000iCShM`004`38ZRDc0000023[4P281Q<0
e>/B09D[n7LP>OQgooooonC[4P10]oagj0HC0=2Z9C?@ZRDc9?lR0=2Z9C<`0300000`02d0@`0`0300
<00]0300<00`0300<00`00004`0`0300=00f07d0G01=06T0001S0080001Q07@0M@1c0000O@000:Eg
002Y=P004`0200000:XU<a000010[Blc0@000040001Dja80E00001c/4P2E:oQg63KhMooooomTMP08
000002D0000<00001@0002P0000<00001P0005@0001D0000cP4002D1003B0@00<P400040002BL0Q2
EWD8@Qd1002P0@000@0004`00000000000000000003ooooooooooe00001R0008L0000580001`0@00
1P0001000007000000000000002l0P00000000L20R9C07T0L`1d06D0K@0000000000000000000000
000000000000000000000000000000000000000000000000000/<cRP;3<00:Eg002Y=P004`020000
0:XU<a000010[Blc0@000040001Dja80A>/B0000003U>?Qg000C0<RZ9C<000008>/B08P64`3Dja80
UB_hMb0in7Ooooooi>/B042go7OX1Q<0d:XU<m2Z9C=`ob80d:XU<jS[4P000?Qg63KhMooooonhja80
PW?XM`004`000000000C0>852QPhX2`c`5OgMdd0J@0006<00P000640M024ja80000007C]4P1F6>Yg
000C00800000ZRDc400003RP;3<100000@0005C[4P1D00007>`B09D[n7LH=_QgooooofAf00P00000
9@0000`000060000:00000`0000500009@0000`0000700209@0000`0000000209@0000`000030000
B`000400000`00001@000200000100000@00010000310@0070400=D1000c0@00`@4001`1003E0@00
<`4002P0000<00001P000280000<0000ooooodH0000D0000200004M4BD<300008P0000`0003ooooo
8P0000`0003ooooo9@0000`0000=0020:00000`000040000:00000`0000300008P0000`0003ooooo
AP000>07003D1`00ADe6:bI0000@000010000000000U@000400000@000000000;40002@0000H0000
>amV?00008000020/0;0?03of4>kK?M3240<1F07001D1`000A30f`8000030000A0L00040001/0000
50000000000:0@00nP40000000000000FRH00010000PADe6000104@7000W00000`00000000000000
0000000400000`00@0400?0000000000000000000000hP@0P:T301T0000<0000oooo0180000<0000
0P0002D0000<00000000P2D0000<00003@00P2D0000<00001`00P1P0000<0000000001H0000<0000
00000240000800007P0001P000000000000003X1000<0P004@0000`0000800003000010000000000
000000/0000@0000>P4000`2000H000030000000000B000030000040000I000030000?ooo`0H0000
30000000000F0000300001P0000:000040000000000000002@000100000P0@00h04001P0000<0000
00000580001`0@000@00083nool0000000000000002@0@000@000002011D06T0K@1U07<0801>06D0
M`0P0580K`1]0640KP0000000000000000000000000000000000000000000000000/<mSZ4P203P00
U>`B09D[n7Nh>OQgooooojC/4P0W/_ag60LC0000003@0JLf000004H0E01G0440DP1505`0@`1/0640
L`1c06D0L`1L04<0C01C04T0A01L07/0<00`0300<P1304D0<00c02d0<00`0300<00]0300<0000300
;@130300<00`02d0<00`0300<00`0300<00`0300<00d03H0O@1L04d0J@1c06<0D`1d0640M01e07<0
001m0000YGL00:Xf000001S/4P2mijEg0P03P=R_;c=/3JIg`USdM`T0003P=PDc8>`B05a@9C?01_Eg
0@0006Af00P000009@0000`000010000E00005@0000D0000000008l0002g0@000@000>A8249FM@Q2
4P00040100010000C00000000000000000000?ooooooooooD00007@0001[0000DP00070100020000
8?ooo`00000000000000090100010000008045@0J@1]06D0L`0P04h0I@1g0200DP1_06d0H@1^0000
0000000000000000000000000000000000000000000007<0001m0000YGL00:Xf000001S/4P2mijEg
0P03P=R_;c=/3JIg`USdM`T00014ja8000000>Dhn7L001<0F7<X<`00000Pja80R0HC0=C[4P2E:oQg
83WhMooooooTja80@;OlMnP64`1PLbPcH7<X<nA@9C=PLbPc<00`0000<00]04<0<00`0300;@0`0300
<00`0300<00001<0<00`03@0=P1m05`0C@1Y0000H`020000H@1d07D0L`0007d0002UM`00ZSH001<0
0P00001c:3<@0000f:l_<`4000010000E>/B05@0000Lk180UB_hMaPfn7OoooooI7H02000000U0000
30000080000X000030000040001D0000E0000900003/00002P400?X100010000i4P8@UIe24:40000
X0400040001<00000000000000000000oooooooooom@0000HP000700001B0000L0400040000@0000
1`00000000000000_080000000010P8RD`1i07<0M01U06d000000000000000000000000000000000
00000000000000000000000000000000000000000S=hXP8c002UM`00ZSH001<00P00001c:3<@0000
f:l_<`4000010000E>/B04C[4P000000iCShM`004`1HLbPc0000023[4P281Q<0e>/B09D[n7LP>OQg
ooooonC[4P10]oagj0HC061c:3=PLbPc<54U<f1c:3>Xja80003hMaPfn7Oooooo^>/B089cj7L001<0
000000004`0>1PXMN:82<l1GmgM=06T0001S0080001Q07@0Q>/B0000001dkA80EQSZM`004`020000
07<X<a00001hXP8c0@000040001Dja80E00001c/4P2E:oQg63KhMooooomTMP08000002D0000<0000
0@0002P0000<00000P0002D0000<00001`00P2D0000<00000000P300000<00003`00P2D0000<0000
3@00P4/000100000<00000D0000P00000@000040000@000000000000000j0@003080000000000000
>P4000`2000X000030000040000R000030000?ooool>000050000000000@0000500001Y0340/0000
80000?ooool20000000000000000P9e30403A000003X0nP38@0000P0001R000030000040000Q0000
2000024000080000L`0000`000000000700000P0000U000030000000080U0000300000L0080U0000
300000h0080`0000300000l0080I000030000?ooo`0H000030000000000B000030000080000C0000
30000040000D0000300000d0000E000030000040000F000030000000000=00004000000000000000
>P0000`0000:00006`00010000000000000002<0000P0000VhDf?@00000000000;0c?@00fD<0P?M3
900002@0002KQCHm000000000000/3<m003I@`20md<40000L`0000`0000000003@000100002b0@00
k`400580001`0@000`0001000007000000000000002l0P00000000L20R9C07T0L`1d06D0K@000000
0000000000000000000000000000000000000000000000000000000000000000000QG`0000000000
?8Si<[SX4P0dR?Tb@XKfMiP58El08`Hc0`0003b8nC8001<01000001Mnc97A4U3`USdM`40000Q08X1
000:03S=;c=Jl?Ag3@0001T0/04Q08X18@00089cj7L001<00000061X<C>m1@[MH6Pa<l1GmgL0025O
402@0BD0002a>OAg@>HB0>P?93>/ia80EQSZMk1ej7Oooooo0@0004M4BD<80000d>HB0>CV4P1KVoIg
00000:2:nC800240V8[i<R1MmgL00000h0[F=@000000000000000>CV4P010000V0DQGnCV4P21@OAg
h0[F=FAf00P000009@0000`000030000AP0002P0000L0000AdA9@`8000000000000003X1000<0P00
000001T0000<0000oooo0180000<00000P0002D0000<00000000P580001`0@001000010000070000
00000000002l0P00000000L20R9C07T0L`1d06D0K@00000000000000000000000000000000000000
00000000000000000000000000000000003k<TM4BD?2F?Ag0@000240RP4000X0><d_<e[`m7L=0000
6@2`0B40RP4Q0000PW?XM`004`000000H6Pa<kd52]ePJ34c`5OgM`008El@09019@000;4im7M0iQ80
j0lT<jcW4P1F6>Yg/7GXMoooool10000AdA9@`P000000180i>HB05^KmWL00000X8[i<P008@2HR_Tb
85ggM`00003P2]He00000000000U0000h3H5<iSV4P0@0000003eM`40000I0;01402@0El9mGNH1B5O
0000P2D0000@0901ool001000030iQ80nQgdMiP58El@09019@0006Af00P000009@0000`000040000
9@0000`000070020600000`0000000005P0000`0000000008@0000P0000N00006000000000000000
>P4000`2000S0000800009^5=Sd000000000002`<cd00=U3083g@b@0000T0000VhDf?@0000000000
0;0c?@00fD<0P?M3100002<0000P0000VhDf?@00000000000;0c?@00fD<0P?M3900002@0002KQCHm
000000000000/3<m003I@`20md<400008`000200003Pgem100000000000Pi[M1003I@`20md<T0000
90000>3OGd4000000000023V]d400=U3083g@`@0000H000030000000000B000030000040000I0000
30000?ooo`0H000030000000000F0000300001P0000S000080000>3OGd4000000000023V]d400=U3
083g@b@0000T0000h=mO@@00000000008>Jg@@00fD<0P?M3100002<0000P0000i?m6?@0000000000
ZbQ4?@00fD<0P?M3900002@0003TodHm00000000002[:4@m003I@`20md<40000600000`000000000
DP00070100060000P?koo`000000000000000901000100001`8045@0J@1]06D0L`0P04h0I@1g0200
DP1_06d0H@1^00000000000000000000000000000000000000000000000002`cf>XB080>002Dk180
UB_hMkPin7OoooooY>`B02Nbo7LH1a<000000=01YcH00000AP1D05L0@@1B04D0G01306`0H@1c07<0
I@1c05`0@`1<05<0B@1405`0N`0`0300<00b04<0A@0`03<0;@0`0300<00`02d0<00`0000<00]04<0
<00`0300;@0`0300<00`0300<00`0300<00`03@0=P1m05`0C@1Y07<0H`1C07@0H@1d07D0L`0007d0
002UM`00ZSH000006>`B0;gWYGL200>0f:l_<f`=YWO2F?Ag2@000>0f1C<Pk180G50U<l06mGL10000
I7H02000000U0000300000H0001D0000E0000;<1003_0@00^040004200010000i4P8@UIe248B0000
@0400040001<00000000000000000000oooooooooom@0000M00006/0001B0000L04000D0000Poooo
0000000000000000T040004000070P0@E01Y06d0I@1c0200CP1U07L0801B06l0K@1Q06h000000000
0000000000000000000000000000000000000000L`0007d0002UM`00ZSH000006>`B0;gWYGL200>0
f:l_<f`=YWO2F?Ag2@0004C[4P000000iCShM`004`1HLbPc0000023[4P281Q<0e>/B09D[n7LP>OQg
ooooonC[4P10]oagj0HC061c:3=PLbPci50U<f1c:3<`0300000`02d0@`0`0300<00]0300<00`0300
<00`00004`0`0300=00f07d0G01=06T0001S0080001Q07@0M@1c0000O@000:Eg002Z=P004`020000
07<X<a00003H[blc0@000040001Dja80E00001c/4P2E:oQg63KhMooooomTMP08000002D0000<0000
1@0002P0000<00001P0005@0001D0000^0400?P1002m0@001@800040003TB0Q2EWD8@X@0002P0@00
0@0004`00000000000000000003ooooooooooe00001R0000L0000580001`0@001P00010000070000
00000000002l0P00000000L20R9C07T0L`1d06D0K@00000000000000000000000000000000000000
000000000000000000000000000000000002<gRR0S<00:Eg002Z=P004`02000007<X<a00003H[blc
0@000040001Dja80A>/B0000003U>?Qg000C05Qc:3<000008>/B08P64`3Dja80UB_hMb0in7Oooooo
i>/B042go7OX1Q<0H7<X<f1c:3<`DBDcH7<X<jS[4P000?Qg63KhMooooonhja80PW?XM`004`000000
000C00h62QehXP8c`5OgMdd0J@0006<00P000640M024ja80000007C]4P1F6>Yg000C00800000LbPc
400007RR0S<100000@0005C[4P1D00007>`B09D[n7LH=_QgooooofAf00P000009@0000`000060000
:00000`0000500009@0000`0000700209@0000`0000000209@0000`000040000B`000400000`0000
1@000200000100000@000100002b0@00k`400<0100060P00/P400>l100300@001P8002P0000<0000
1P000280000<0000ooooodH0000D0000200004M4BD<300008P0000`0003ooooo8P0000`0003ooooo
9@0000`0000=0020:00000`000030000:00000`0000400008P0000`0003oooooAP000;@0002X0000
ADe6:bI0000@000010000000000Z@000900001P0001LSj9200000000001LSj92HLg7@hVl7T08@0H6
<00002@000014<3KZjXZ?P0000020000000000H0001C05T0C@1204l0C00f@0J0@00003@00000003o
0@00004000010000K?3O<X4nPQmh?P00P3l0000000000000P3l00000000000004P0000`000010000
600000`0000000005P0000`0000H0000:00000`000010000DP00070100010000l_ooo`0000000000
00000901000100021`0405<0N@1]0680K`1/0000001=06l0KP1_07@0N@1`06D0>P1C07T0K@1R06l0
K00P03X0MP1U0780L`1Y06l0KP000>Ugn?4B04PcjGL;0000:P000?Ca4P2P0nig000002X0001L04h0
C01C05`0CP1/07<0D`1U06<0M01Y06l0KP130500=00b0000<>hB00004`0400000000004000000000
0NlB03C_4P2dkA80X9T/<gc^4P2E:oQg63KhM`00oon<kQ80PW?XM`004`000000j68O<jH52Q?XHQlc
`5OgM`00000<:1@000000;hJ061HkQ80X9T/<fch4P000>Yg/7GXMoooool10000CggdMnQR7c<80000
h3H5<lS^4P3XHQlca0OeMl1GmgNH1B5O0@0000dlm7O0EoMgI7H02000000U000030000040001D0000
E0000:@100080000Z`4001P000010000003j@@00nT6T0@005P000040001<00000000000000000000
oooooooooom@0000K?000000000U0000300000d008160000700001000015CDH[:d0000`000000000
C00006@00002000000000>01000@0`000P000000003O0@004@<002T0ZP00000000000000P3l00000
00000000P3l000000000000000000000000000000000000000000280000<0000ooooodH0000L0000
400004E=AR/2@00030000000000>000050000000000@00005000
\>"], "Graphics",
  ImageSize->{318, 521.063},
  ImageMargins->{{0, 0}, {0, 12.1875}},
  ImageRegion->{{0, 1}, {0, 1}},
  CellTags->{"b:15.3", "Tractor-Trailer"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[" Joint Definitions", "Subsubsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell["Main body (tractor) joint", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell["\<\
(* Joint 1  *)
r1={3}; q1={theta,x,y}; p1={wth,vx,vy};
H1={{0,0,0},{0,0,0},{1,0,0},{0,1,0},{0,0,1},{0,0,0}};
\
\>", "Input",
  CellLabel->"In[140]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell["Tractor front end joints", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  FontFamily->"Times New Roman",
  FontSize->12,
  FontWeight->"Bold",
  FontColor->GrayLevel[0],
  Background->GrayLevel[1],
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell["\<\
Define castor sr (right) and sl (left) and camber tr (right) and tl (left). \
Note that positive castor and camber implies spindle points down and forward \
and inward.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell["\<\
(* Castor/Camber *)
(* tr=0.02; tl=0.02; *)

(* Joint 2 right front spindle *)
r2={1}; q2={delta2}; p2={wdel2}; 
H2=Transpose[{{-sr,-tr,1,0,0,0}}];

(* Joint 3 left front spindle *)
r3={1}; q3={delta3}; p3={wdel3}; 
H3=Transpose[{{-sl,tl,1,0,0,0}}];\
\>", "Input",
  CellLabel->"In[143]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell["Tractor-Semi joint", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  FontFamily->"Times New Roman",
  FontSize->12,
  FontWeight->"Bold",
  FontColor->GrayLevel[0],
  Background->GrayLevel[1],
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell["\<\
(* Joint 4 5th wheel *)
r4={1}; q4={delta4}; p4={wdel4};
H4=Transpose[{{0,0,1,0,0,0}}];

JointLst={{r1,H1,q1,p1},{r2,H2,q2,p2},{r3,H3,q3,p3},{r4,H4,q4,p4}};\
\>", "Input",
  CellLabel->"In[151]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Joint constructions", "Subsubsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[CellGroupData[{

Cell["{T,{V,X,H}}=Timing[Joints[JointLst]];", "Input",
  CellLabel->"In[155]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[BoxData[
    \("Computing joint 1 kinematics"\)], "Print",
  CellLabel->"From In[155]:=",
  CellTags->"b:15.3"],

Cell[BoxData[
    \("Computing joint 2 kinematics"\)], "Print",
  CellLabel->"From In[155]:=",
  CellTags->"b:15.3"],

Cell[BoxData[
    \("Computing joint 3 kinematics"\)], "Print",
  CellLabel->"From In[155]:=",
  CellTags->"b:15.3"],

Cell[BoxData[
    \("Computing joint 4 kinematics"\)], "Print",
  CellLabel->"From In[155]:=",
  CellTags->"b:15.3"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Print[T];", "Input",
  CellLabel->"In[156]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[BoxData[
    \(12.788`\ Second\)], "Print",
  CellLabel->"From In[156]:=",
  CellTags->"b:15.3"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Body data", "Subsubsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell["\<\
Tractor body frame is located midway between rear axles on body centerline \
and at axle hight. It is right handed with x-forward, y-left, and z-up. 

\t           wb = wheelbase (186\")
\t           R=tire radius (12\")
\t           2b=track (72-11/16\")
\t           pin= pin location from rear bogie (8\")
\t           d=tractor rear axle location from rear bogie (28\")
\t           tb=trailor base (388\")
\t           kappa=tire coefficient (approx 14,000 N?)\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell["\<\
wb=186*.0254; R=12*.0254; b=(72+11/16)*.0254;
pin=8*.0254; dd=28*.0254; tb=388*.0254;
Jtracxx=1.414*10^3; Jtracyy=5.684*10^4; Jtraczz=5.698*10^4;Jtracxz=1.96*10^3;
Jtrailxx=2.202*10^5; Jtrailyy=4.852*10^5; Jtrailzz=5.065*10^5;
mtrac=1.15*10^4; mtrail=4.376*10^4; mtire=40;
Ixx=mtire*(R^2)/2; Iyy=mtire*(3*R^2+(11*.0254)^2)/12; Izz=Iyy;

(* Body 1 tractor *)
cm1={2.4466,0,.9052};
out1={{2,{wb,-b,0}},{3,{wb,b,0}},{4,{pin,0,0}},{5,{dd,-b,-R}},{6,{dd,b,-R}},{\
7,{-dd,-b,-R}},{8,{-dd,b,-R}}};
I1={{Jtracxx,0,Jtracxz},{0,Jtracyy,0},{Jtracxz,0,Jtraczz}};

(* Body 2 right front wheel *)
cm2={0,0,0}; out2={{9,{sr*R,tr*R,-R}}};
I2=DiagonalMatrix[{Ixx,Iyy,Izz}];

(* Body 3 left front wheel *)
cm3={0,0,0}; out3={{10,{sl*R,-tl*R,-R}}};
I3=DiagonalMatrix[{Ixx,Iyy,Izz}];

(* Body 4 trailor *)
cm4={-5.77,0,1.0592}; 
out4={{11,{-tb,-b,-R}},{12,{-tb,b,-R}}};
I4=DiagonalMatrix[{Jtrailxx,Jtrailyy,Jtrailzz}];

BodyLst={{cm1,out1,mtrac,I1},{cm2,out2,mtire,I2},{cm3,out3,mtire,I3},{cm4,\
out4,mtrail,I4}};
TreeLst={{{1,1},{2,2}},{{1,1},{3,3}},{{1,1},{4,4}}};

q={theta,x,y,delta2,delta3,delta4};
p={wth,vx,vy,wdel2,wdel3,wdel4};
\
\>", "Input",
  CellLabel->"In[157]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Tire Forces", "Subsubsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell["\<\
We will compute the tire generalized forces. This requires eight similar, and \
somewhat complicated, calculations. To reduce computation time, we use the \
function KinematicReplacements. The right front tire force is computed both \
with and without this enhancement. We find a factor of 10 improvement in \
computational speed.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[CellGroupData[{

Cell["\<\
(* Front right tire *)

ChnLst={{1,1},{2,2}};
Force={0,0,0,0,-kappa*ArcTan[v9y/v9x],0};
VelNames={w9x,w9y,w9z,v9x,v9y,v9z};
TerminalNode=9;
{T,Q1}=Timing[GeneralizedForce[ChnLst,TerminalNode,BodyLst,X,H,q,p,Force,\
VelNames]]
\
\>", "Input",
  CellLabel->"In[181]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[BoxData[
    \({42.491`\ Second, {\(-kappa\)\ ArcTan[\((\(vy\ \((tr\^2 + \((1 + sr\^2)\
\)\ Cos[delta2\ \@\(1 + sr\^2 + tr\^2\)])\)\)\/\(1 + sr\^2 + tr\^2\) - \
\(1\/\((1 + sr\^2 + tr\^2)\)\^\(3/2\)\) \((2\ vx\ Sin[
                          1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)]\ \((\((1 + 
                                  sr\^2 + tr\^2)\)\ Cos[
                                1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)] - 
                            sr\ tr\ \@\(1 + sr\^2 + tr\^2\)\ Sin[
                                1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)])\))\) \
+ wth\ \((\(\((\(\(4.7244`\)\(\[InvisibleSpace]\)\) + 4.7244`\ sr\^2 + \
0.30479999999999996`\ sr\^3 + sr\ \((\(\(0.30479999999999996`\)\(\
\[InvisibleSpace]\)\) + 0.30479999999999996`\ tr\^2)\))\)\ Cos[delta2\ \@\(1 \
+ sr\^2 + tr\^2\)]\)\/\(1 + sr\^2 + tr\^2\) + \(1\/\((1 + sr\^2 + \
tr\^2)\)\^\(3/2\)\) \((\((\(-3.692525`\) - 3.692525`\ sr\^2 - 
                                  3.692525`\ tr\^2)\)\ Cos[
                                1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)]\ Sin[
                                1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)])\) + \
\(1\/\(\(\(1.`\)\(\[InvisibleSpace]\)\) + 1.`\ sr\^2 + 
                                1.`\ tr\^2\)\) \((tr\ \((4.7244`\ tr + 
                                  3.692525`\ sr\ Sin[1\/2\ delta2\ \@\(1 + sr\
\^2 + tr\^2\)]\^2 + 0.30479999999999996`\ \@\(1 + sr\^2 + tr\^2\)\ Sin[
                                      delta2\ \@\(1 + sr\^2 + tr\^2\)])\))\))\
\))\)/\((\(vx\ \((sr\^2 + \((1 + tr\^2)\)\ Cos[delta2\ \@\(1 + sr\^2 + \
tr\^2\)])\)\)\/\(1 + sr\^2 + tr\^2\) + \(1\/\((1 + sr\^2 + \
tr\^2)\)\^\(3/2\)\) \((2\ vy\ Sin[
                          1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)]\ \((\((1 + 
                                  sr\^2 + tr\^2)\)\ Cos[
                                1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)] + 
                            sr\ tr\ \@\(1 + sr\^2 + tr\^2\)\ Sin[
                                1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)])\))\) \
+ wth\ \((\(1.8462625`\ \((sr\^2 + \((1 + tr\^2)\)\ Cos[delta2\ \@\(1 + sr\^2 \
+ tr\^2\)])\)\)\/\(1 + sr\^2 + tr\^2\) - \(0.30479999999999996`\ tr\ \((1 + \
\((sr\^2 + tr\^2)\)\ Cos[delta2\ \@\(1 + sr\^2 + tr\^2\)])\)\)\/\(1 + sr\^2 + \
tr\^2\) + \(1\/\((1 + sr\^2 + tr\^2)\)\^\(3/2\)\) \((9.4488`\ Sin[
                                1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)]\ \
\((\((1 + sr\^2 + tr\^2)\)\ Cos[1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)] + 
                                  sr\ tr\ \@\(1 + sr\^2 + tr\^2\)\ Sin[
                                      1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)])\
\))\) + \(1\/\(\(\(1.`\)\(\[InvisibleSpace]\)\) + 1.`\ sr\^2 + 
                                1.`\ tr\^2\)\) \((0.30479999999999996`\ tr - 
                              0.30479999999999996`\ tr\ Cos[
                                  delta2\ \@\(1 + sr\^2 + tr\^2\)] + 
                              0.30479999999999996`\ sr\ \@\(1 + sr\^2 + tr\^2\
\)\ Sin[delta2\ \@\(1 + sr\^2 + tr\^2\)])\))\))\)]\ \((\(4.7244`\ \
tr\^2\)\/\(1 + sr\^2 + tr\^2\) + \(\((\(\(4.7244`\)\(\[InvisibleSpace]\)\) + \
4.7244`\ sr\^2 + 0.30479999999999996`\ sr\^3 + sr\ \
\((\(\(0.30479999999999996`\)\(\[InvisibleSpace]\)\) + 0.30479999999999996`\ \
tr\^2)\))\)\ Cos[delta2\ \@\(1 + sr\^2 + tr\^2\)]\)\/\(1 + sr\^2 + tr\^2\) + \
\(\((\(-3.692525`\) - 3.692525`\ sr\^2 - 3.692525`\ tr\^2)\)\ Cos[1\/2\ \
delta2\ \@\(1 + sr\^2 + tr\^2\)]\ Sin[1\/2\ delta2\ \@\(1 + sr\^2 + \
tr\^2\)]\)\/\((1 + sr\^2 + tr\^2)\)\^\(3/2\) + \(3.692525`\ sr\ tr\ Sin[1\/2\ \
delta2\ \@\(1 + sr\^2 + tr\^2\)]\^2\)\/\(1 + sr\^2 + tr\^2\) + \
\(0.30479999999999996`\ tr\ Sin[delta2\ \@\(1 + sr\^2 + tr\^2\)]\)\/\@\(1 + \
sr\^2 + tr\^2\))\), \(1\/\((1 + sr\^2 + tr\^2)\)\^\(3/2\)\) \((2\ kappa\ \
ArcTan[\((\(vy\ \((tr\^2 + \((1 + sr\^2)\)\ Cos[delta2\ \@\(1 + sr\^2 + tr\^2\
\)])\)\)\/\(1 + sr\^2 + tr\^2\) - \(1\/\((1 + sr\^2 + tr\^2)\)\^\(3/2\)\) \
\((2\ vx\ Sin[1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)]\ \((\((1 + sr\^2 + 
                                      tr\^2)\)\ Cos[
                                    1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)] - 
                                sr\ tr\ \@\(1 + sr\^2 + tr\^2\)\ Sin[
                                    1\/2\ delta2\ \@\(1 + sr\^2 + \
tr\^2\)])\))\) + 
                      wth\ \((\(\((\(\(4.7244`\)\(\[InvisibleSpace]\)\) + \
4.7244`\ sr\^2 + 0.30479999999999996`\ sr\^3 + sr\ \
\((\(\(0.30479999999999996`\)\(\[InvisibleSpace]\)\) + 0.30479999999999996`\ \
tr\^2)\))\)\ Cos[delta2\ \@\(1 + sr\^2 + tr\^2\)]\)\/\(1 + sr\^2 + tr\^2\) + \
\(1\/\((1 + sr\^2 + tr\^2)\)\^\(3/2\)\) \((\((\(-3.692525`\) - 
                                      3.692525`\ sr\^2 - 
                                      3.692525`\ tr\^2)\)\ Cos[
                                    1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)]\ \
Sin[1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)])\) + \(1\/\(\(\(1.`\)\(\
\[InvisibleSpace]\)\) + 1.`\ sr\^2 + 
                                    1.`\ tr\^2\)\) \((tr\ \((4.7244`\ tr + 
                                      3.692525`\ sr\ Sin[1\/2\ delta2\ \@\(1 \
+ sr\^2 + tr\^2\)]\^2 + 0.30479999999999996`\ \@\(1 + sr\^2 + tr\^2\)\ Sin[
                                        delta2\ \@\(1 + sr\^2 + \
tr\^2\)])\))\))\))\)/\((\(vx\ \((sr\^2 + \((1 + tr\^2)\)\ Cos[delta2\ \@\(1 + \
sr\^2 + tr\^2\)])\)\)\/\(1 + sr\^2 + tr\^2\) + \(1\/\((1 + sr\^2 + tr\^2)\)\^\
\(3/2\)\) \((2\ vy\ Sin[1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)]\ \((\((1 + 
                                      sr\^2 + tr\^2)\)\ Cos[
                                    1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)] + 
                                sr\ tr\ \@\(1 + sr\^2 + tr\^2\)\ Sin[
                                    1\/2\ delta2\ \@\(1 + sr\^2 + \
tr\^2\)])\))\) + 
                      wth\ \((\(1.8462625`\ \((sr\^2 + \((1 + tr\^2)\)\ \
Cos[delta2\ \@\(1 + sr\^2 + tr\^2\)])\)\)\/\(1 + sr\^2 + tr\^2\) - \
\(0.30479999999999996`\ tr\ \((1 + \((sr\^2 + tr\^2)\)\ Cos[delta2\ \@\(1 + \
sr\^2 + tr\^2\)])\)\)\/\(1 + sr\^2 + tr\^2\) + \(1\/\((1 + sr\^2 + tr\^2)\)\^\
\(3/2\)\) \((9.4488`\ Sin[1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)]\ \((\((1 + 
                                        sr\^2 + tr\^2)\)\ Cos[
                                        1\/2\ delta2\ \@\(1 + sr\^2 + \
tr\^2\)] + sr\ tr\ \@\(1 + sr\^2 + tr\^2\)\ Sin[
                                        1\/2\ delta2\ \@\(1 + sr\^2 + \
tr\^2\)])\))\) + \(1\/\(\(\(1.`\)\(\[InvisibleSpace]\)\) + 1.`\ sr\^2 + 
                                    1.`\ tr\^2\)\) \((0.30479999999999996`\ \
tr - 0.30479999999999996`\ tr\ Cos[delta2\ \@\(1 + sr\^2 + tr\^2\)] + 
                                  0.30479999999999996`\ sr\ \@\(1 + sr\^2 + \
tr\^2\)\ Sin[delta2\ \@\(1 + sr\^2 + tr\^2\)])\))\))\)]\ Sin[
                1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)]\ \((\((1 + sr\^2 + 
                        tr\^2)\)\ Cos[
                      1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)] - 
                  sr\ tr\ \@\(1 + sr\^2 + tr\^2\)\ Sin[
                      1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)])\))\), \
\(-\(\(1\/\(1 + sr\^2 + 
                  tr\^2\)\) \((kappa\ ArcTan[\((\(vy\ \((tr\^2 + \((1 + \
sr\^2)\)\ Cos[delta2\ \@\(1 + sr\^2 + tr\^2\)])\)\)\/\(1 + sr\^2 + tr\^2\) - \
\(1\/\((1 + sr\^2 + tr\^2)\)\^\(3/2\)\) \((2\ vx\ Sin[
                                1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)]\ \
\((\((1 + sr\^2 + tr\^2)\)\ Cos[1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)] - 
                                  sr\ tr\ \@\(1 + sr\^2 + tr\^2\)\ Sin[
                                      1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)])\
\))\) + wth\ \((\(\((\(\(4.7244`\)\(\[InvisibleSpace]\)\) + 4.7244`\ sr\^2 + \
0.30479999999999996`\ sr\^3 + sr\ \((\(\(0.30479999999999996`\)\(\
\[InvisibleSpace]\)\) + 0.30479999999999996`\ tr\^2)\))\)\ Cos[delta2\ \@\(1 \
+ sr\^2 + tr\^2\)]\)\/\(1 + sr\^2 + tr\^2\) + \(1\/\((1 + sr\^2 + \
tr\^2)\)\^\(3/2\)\) \((\((\(-3.692525`\) - 3.692525`\ sr\^2 - 
                                        3.692525`\ tr\^2)\)\ Cos[
                                      1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)]\ \
Sin[1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)])\) + \(1\/\(\(\(1.`\)\(\
\[InvisibleSpace]\)\) + 1.`\ sr\^2 + 
                                      1.`\ tr\^2\)\) \((tr\ \((4.7244`\ tr + 
                                        3.692525`\ sr\ Sin[1\/2\ delta2\ \
\@\(1 + sr\^2 + tr\^2\)]\^2 + 
                                        0.30479999999999996`\ \@\(1 + sr\^2 + \
tr\^2\)\ Sin[
                                        delta2\ \@\(1 + sr\^2 + \
tr\^2\)])\))\))\))\)/\((\(vx\ \((sr\^2 + \((1 + tr\^2)\)\ Cos[delta2\ \@\(1 + \
sr\^2 + tr\^2\)])\)\)\/\(1 + sr\^2 + tr\^2\) + \(1\/\((1 + sr\^2 + tr\^2)\)\^\
\(3/2\)\) \((2\ vy\ Sin[1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)]\ \((\((1 + 
                                        sr\^2 + tr\^2)\)\ Cos[
                                      1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)] \
+ sr\ tr\ \@\(1 + sr\^2 + tr\^2\)\ Sin[
                                      1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)])\
\))\) + wth\ \((\(1.8462625`\ \((sr\^2 + \((1 + tr\^2)\)\ Cos[delta2\ \@\(1 + \
sr\^2 + tr\^2\)])\)\)\/\(1 + sr\^2 + tr\^2\) - \(0.30479999999999996`\ tr\ \
\((1 + \((sr\^2 + tr\^2)\)\ Cos[delta2\ \@\(1 + sr\^2 + tr\^2\)])\)\)\/\(1 + \
sr\^2 + tr\^2\) + \(1\/\((1 + sr\^2 + tr\^2)\)\^\(3/2\)\) \((9.4488`\ Sin[
                                      1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)]\ \
\((\((1 + sr\^2 + tr\^2)\)\ Cos[1\/2\ delta2\ \@\(1 + sr\^2 + tr\^2\)] + 
                                        sr\ tr\ \@\(1 + sr\^2 + tr\^2\)\ Sin[
                                        1\/2\ delta2\ \@\(1 + sr\^2 + \
tr\^2\)])\))\) + \(1\/\(\(\(1.`\)\(\[InvisibleSpace]\)\) + 1.`\ sr\^2 + 
                                      1.`\ tr\^2\)\) \((0.30479999999999996`\ \
tr - 0.30479999999999996`\ tr\ Cos[delta2\ \@\(1 + sr\^2 + tr\^2\)] + 
                                    0.30479999999999996`\ sr\ \@\(1 + sr\^2 + \
tr\^2\)\ Sin[delta2\ \@\(1 + sr\^2 + tr\^2\)])\))\))\)]\ \((tr\^2 + \((1 + 
                          sr\^2)\)\ Cos[
                        delta2\ \@\(1 + sr\^2 + tr\^2\)])\))\)\)\), 0, 0, 
        0}}\)], "Output",
  CellLabel->"Out[186]=",
  CellTags->"b:15.3"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Print[T]", "Input",
  CellLabel->"In[187]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[BoxData[
    \(42.491`\ Second\)], "Print",
  CellLabel->"From In[187]:=",
  CellTags->"b:15.3"]
}, Open  ]],

Cell["{V,X,H,rules1,rules2}=KinematicReplacements[V,X,H,q];", "Input",
  CellLabel->"In[188]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[CellGroupData[{

Cell["rules1", "Input",
  CellLabel->"In[189]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[BoxData[
    \({t1 \[Rule] Sin[\(delta3\ t17\)\/2], t2 \[Rule] Sin[delta3\ t17], 
      t3 \[Rule] Sin[\(delta2\ t22\)\/2], t4 \[Rule] Sin[delta2\ t22], 
      t5 \[Rule] Cos[delta4], t6 \[Rule] Cos[theta], 
      t7 \[Rule] Cos[\(delta3\ t17\)\/2], t8 \[Rule] Cos[delta3\ t17], 
      t9 \[Rule] Cos[\(delta2\ t22\)\/2], 
      t10 \[Rule] Cos[delta2\ t22]}\)], "Output",
  CellLabel->"Out[189]=",
  CellTags->"b:15.3"]
}, Open  ]],

Cell["\<\
(* Front right tire *)
ChnLst={{1,1},{2,2}};
Force={0,0,0,0,-kappa*ArcTan[v9y/v9x],0};
VelNames={w9x,w9y,w9z,v9x,v9y,v9z};
TerminalNode=9;
Q1=GeneralizedForce[ChnLst,TerminalNode,BodyLst,X,H,q,p,Force,VelNames];\
\>", "Input",
  CellLabel->"In[190]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[CellGroupData[{

Cell["\<\
{T,Q1}=Timing[GeneralizedForce[ChnLst,TerminalNode,BodyLst,X,H,q,p,Force,\
VelNames]]\
\>", "Input",
  CellLabel->"In[196]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[BoxData[
    \({0.01999999999999602`\ Second, {\(-kappa\)\ \((4.7244`\ \((t10 + 
                    t10\ t12 + t18)\)\ t20 + 
              0.30479999999999996`\ sr\ \((1 + t10\ \((t12 + t18)\))\)\ t20 - 
              3.692525`\ t19\ t3\ \((\((1 + t12 + t18)\)\ t9 - 
                    sr\ t22\ t3\ tr)\) + 
              0.30479999999999996`\ \((sr\ \((\(-1\) + t10)\)\ t20 + 
                    t21\ t4\ tr)\))\)\ ArcTan[\((\(-2\)\ t19\ t3\ \((\((1 + 
                              t12 + t18)\)\ t9 - 
                        sr\ t22\ t3\ tr)\)\ vx + \((t10 + t10\ t12 + 
                        t18)\)\ t20\ vy + \((4.7244`\ \((t10 + t10\ t12 + 
                              t18)\)\ t20 + 
                        0.30479999999999996`\ sr\ \((1 + 
                              t10\ \((t12 + t18)\))\)\ t20 - 
                        3.692525`\ t19\ t3\ \((\((1 + t12 + t18)\)\ t9 - 
                              sr\ t22\ t3\ tr)\) + 
                        0.30479999999999996`\ \((sr\ \((\(-1\) + 
                                    t10)\)\ t20 + 
                              t21\ t4\ tr)\))\)\ wth)\)/\((\((t10 + t12 + 
                        t10\ t18)\)\ t20\ vx + 
                  2\ t19\ t3\ \((\((1 + t12 + t18)\)\ t9 + 
                        sr\ t22\ t3\ tr)\)\ vy + \((1.8462625`\ \((t10 + 
                              t12 + t10\ t18)\)\ t20 - 
                        0.30479999999999996`\ \((1 + 
                              t10\ \((t12 + t18)\))\)\ t20\ tr - 
                        0.30479999999999996`\ \((\(-sr\)\ t21\ t4 + \((\(-1\) \
+ t10)\)\ t20\ tr)\) + 
                        9.4488`\ t19\ t3\ \((\((1 + t12 + t18)\)\ t9 + 
                              sr\ t22\ t3\ tr)\))\)\ wth)\)], 
        2\ kappa\ t19\ t3\ \((\((1 + t12 + t18)\)\ t9 - 
              sr\ t22\ t3\ tr)\)\ ArcTan[\((\(-2\)\ t19\ t3\ \((\((1 + t12 + 
                              t18)\)\ t9 - sr\ t22\ t3\ tr)\)\ vx + \((t10 + 
                        t10\ t12 + 
                        t18)\)\ t20\ vy + \((4.7244`\ \((t10 + t10\ t12 + 
                              t18)\)\ t20 + 
                        0.30479999999999996`\ sr\ \((1 + 
                              t10\ \((t12 + t18)\))\)\ t20 - 
                        3.692525`\ t19\ t3\ \((\((1 + t12 + t18)\)\ t9 - 
                              sr\ t22\ t3\ tr)\) + 
                        0.30479999999999996`\ \((sr\ \((\(-1\) + 
                                    t10)\)\ t20 + 
                              t21\ t4\ tr)\))\)\ wth)\)/\((\((t10 + t12 + 
                        t10\ t18)\)\ t20\ vx + 
                  2\ t19\ t3\ \((\((1 + t12 + t18)\)\ t9 + 
                        sr\ t22\ t3\ tr)\)\ vy + \((1.8462625`\ \((t10 + 
                              t12 + t10\ t18)\)\ t20 - 
                        0.30479999999999996`\ \((1 + 
                              t10\ \((t12 + t18)\))\)\ t20\ tr - 
                        0.30479999999999996`\ \((\(-sr\)\ t21\ t4 + \((\(-1\) \
+ t10)\)\ t20\ tr)\) + 
                        9.4488`\ t19\ t3\ \((\((1 + t12 + t18)\)\ t9 + 
                              sr\ t22\ t3\ tr)\))\)\ wth)\)], \(-kappa\)\ \
\((t10 + t10\ t12 + 
              t18)\)\ t20\ ArcTan[\((\(-2\)\ t19\ t3\ \((\((1 + t12 + 
                              t18)\)\ t9 - sr\ t22\ t3\ tr)\)\ vx + \((t10 + 
                        t10\ t12 + 
                        t18)\)\ t20\ vy + \((4.7244`\ \((t10 + t10\ t12 + 
                              t18)\)\ t20 + 
                        0.30479999999999996`\ sr\ \((1 + 
                              t10\ \((t12 + t18)\))\)\ t20 - 
                        3.692525`\ t19\ t3\ \((\((1 + t12 + t18)\)\ t9 - 
                              sr\ t22\ t3\ tr)\) + 
                        0.30479999999999996`\ \((sr\ \((\(-1\) + 
                                    t10)\)\ t20 + 
                              t21\ t4\ tr)\))\)\ wth)\)/\((\((t10 + t12 + 
                        t10\ t18)\)\ t20\ vx + 
                  2\ t19\ t3\ \((\((1 + t12 + t18)\)\ t9 + 
                        sr\ t22\ t3\ tr)\)\ vy + \((1.8462625`\ \((t10 + 
                              t12 + t10\ t18)\)\ t20 - 
                        0.30479999999999996`\ \((1 + 
                              t10\ \((t12 + t18)\))\)\ t20\ tr - 
                        0.30479999999999996`\ \((\(-sr\)\ t21\ t4 + \((\(-1\) \
+ t10)\)\ t20\ tr)\) + 
                        9.4488`\ t19\ t3\ \((\((1 + t12 + t18)\)\ t9 + 
                              sr\ t22\ t3\ tr)\))\)\ wth)\)], 0, 0, 
        0}}\)], "Output",
  CellLabel->"Out[196]=",
  CellTags->"b:15.3"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Print[T];", "Input",
  CellLabel->"In[197]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[BoxData[
    \(0.01999999999999602`\ Second\)], "Print",
  CellLabel->"From In[197]:=",
  CellTags->"b:15.3"]
}, Open  ]],

Cell["\<\
(* Front left tire *)
ChnLst={{1,1},{3,3}};
Force={0,0,0,0,-kappa*ArcTan[v10y/v10x],0};
VelNames={w10x,w10y,w10z,v10x,v10y,v10z};
TerminalNode=10;
Q2=GeneralizedForce[ChnLst,TerminalNode,BodyLst,X,H,q,p,Force,VelNames];\
\>", "Input",
  CellLabel->"In[198]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell["\<\

(* Tractor rear right\[CapitalEth]forward *)
ChnLst={{1,1}};
Force={0,0,0,FDrive,-kappa*ArcTan[v5y/v5x],0};
VelNames={w5x,w5y,w5z,v5x,v5y,v5z};
TerminalNode=5;
Q3=GeneralizedForce[ChnLst,TerminalNode,BodyLst,X,H,q,p,Force,VelNames];\
\>", "Input",
  CellLabel->"In[204]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell["\<\

(* Tracter rear right\[CapitalEth]backward *)
ChnLst={{1,1}};
Force={0,0,0,FDrive,-kappa*ArcTan[v7y/v7x],0};
VelNames={w7x,w7y,w7z,v7x,v7y,v7z};
TerminalNode=7;
Q4=GeneralizedForce[ChnLst,TerminalNode,BodyLst,X,H,q,p,Force,VelNames];\
\>", "Input",
  CellLabel->"In[210]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell["\<\

(* Tractor rear left\[CapitalEth]forward *)
ChnLst={{1,1}};
Force={0,0,0,FDrive,-kappa*ArcTan[v6y/v6x],0};
VelNames={w6x,w6y,w6z,v6x,v6y,v6z};
TerminalNode=6;
Q5=GeneralizedForce[ChnLst,TerminalNode,BodyLst,X,H,q,p,Force,VelNames];\
\>", "Input",
  CellLabel->"In[216]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell["\<\

(* Tractor rear left backward *)
ChnLst={{1,1}};
Force={0,0,0,FDrive,-kappa*ArcTan[v8y/v8x],0};
VelNames={w8x,w8y,w8z,v8x,v8y,v8z};
TerminalNode=8;
Q6=GeneralizedForce[ChnLst,TerminalNode,BodyLst,X,H,q,p,Force,VelNames];\
\>", "Input",
  CellLabel->"In[222]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell["\<\

(* Trailor right *)
ChnLst={{1,1},{4,4}};
Force={0,0,0,0,-kappa*ArcTan[v11y/v11x],0};
VelNames={w11x,w11y,w11z,v11x,v11y,v11z};
TerminalNode=11;
Q7=GeneralizedForce[ChnLst,TerminalNode,BodyLst,X,H,q,p,Force,VelNames];\
\>", "Input",
  CellLabel->"In[228]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell["\<\

(* Trailor left *)
ChnLst={{1,1},{4,4}};
Force={0,0,0,0,-kappa*ArcTan[v12y/v12x],0};
VelNames={w12x,w12y,w12z,v12x,v12y,v12z};
TerminalNode=12;
Q8=GeneralizedForce[ChnLst,TerminalNode,BodyLst,X,H,q,p,Force,VelNames];\
\>", "Input",
  CellLabel->"In[234]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell["Q=Q1+Q2+Q3+Q4+Q5+Q6+Q7+Q8;", "Input",
  CellLabel->"In[240]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Create model", "Subsubsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[CellGroupData[{

Cell["\<\
{V,X,H,Q}=Chop[{V,X,H,Q}/.rules1,.001];
{V,X,H,M,Csys,Fsys,psys,qsys}=CreateModel[JointLst,BodyLst,TreeLst,g,0,Q,V,X,\
H];\
\>", "Input",
  CellLabel->"In[241]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[BoxData[
    \("Computing Potential Functions"\)], "Print",
  CellLabel->"From In[241]:=",
  CellTags->"b:15.3"],

Cell[BoxData[
    \("Computing Inertia Matrix"\)], "Print",
  CellLabel->"From In[241]:=",
  CellTags->"b:15.3"],

Cell[BoxData[
    \("Computing Poincare Function"\)], "Print",
  CellLabel->"From In[241]:=",
  CellTags->"b:15.3"]
}, Open  ]],

Cell["\<\
{V,X,H,M,Csys,Fsys,psys,qsys}={V,X,H,M,Csys,Fsys,psys,qsys}/.rules2;\
\>", "Input",
  CellLabel->"In[243]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell["\<\

yout={X[[1]][[1,4]]/.0254,X[[1]][[2,4]]/.0254};

Inputs = {FDrive};
Outputs = yout;
MEXFilename = \"truck.c\";
paramvec={sr,sl,kappa,g};
PassedParams = {\"X0\",\"sr\",\"sl\",\"kappa\",\"g\"};
PassedParamsDimensions = {{2*Length[p],1},{1,1},{1,1},{1,1},{1,1}};\
\>", "Input",
  CellLabel->"In[244]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["System Reduction & Creation of Mex File", "Subsubsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell["\<\
For analysis purposes we create a simulation model with the following two \
features. First, we assume that delta2 and delta3 are inputs rather than \
coordinates determined by the dynamics. Thus, we eliminate two degrees of \
freedom. Second, we choose to ignore any steering imperfections and assume \
delta2 and delta3 are equal and call them both delta. \
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[CellGroupData[{

Cell["\<\
qred=Part[qsys,{1,2,3,6}];
pred=Part[psys,{1,2,3,6}];
Mred=Part[M,{1,2,3,6},{1,2,3,6}]/.Inner[Rule,{delta2,delta3},{0,0},List];
Cred=Part[Csys,{1,2,3,6},{1,2,3,6}]/.Inner[Rule,{delta2,delta3},{0,0},List];
Fred=Part[Fsys,{1,2,3,6}]/.Inner[Rule,{delta2,delta3},{delta,delta},List];
Vred=V[[{1,4}]];

Inputs = {FDrive,delta};

MEXFilename = \"truck.c\";
paramvec={sr,sl,kappa,g};
PassedParams = {\"X0\",\"sr\",\"sl\",\"kappa\",\"g\"};
PassedParamsDimensions = {{2*Length[pred],1},{1,1},{1,1},{1,1},{1,1}};

FpPrec = SetPrecision[Fred,5];
MPrec = SetPrecision[Mred,5];
CpPrec=SetPrecision[Cred,5];


CreateModelMEX[pred,qred,Inputs,Outputs,PassedParams,PassedParamsDimensions,
                              Vred,CpPrec,FpPrec,MPrec, \
MEXFilename,AssignMaxSize\[Rule]300];
  \
\>", "Input",
  CellLabel->"In[251]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[BoxData[
    \("...Generating Header Code"\)], "Print",
  CellLabel->"From In[251]:=",
  CellTags->"b:15.3"],

Cell[BoxData[
    \("...Generating Initial Condition Function Code"\)], "Print",
  CellLabel->"From In[251]:=",
  CellTags->"b:15.3"],

Cell[BoxData[
    \("...Generating State Derivative Function Code"\)], "Print",
  CellLabel->"From In[251]:=",
  CellTags->"b:15.3"],

Cell[BoxData[
    \("...Collecting all function terms"\)], "Print",
  CellLabel->"From In[251]:=",
  CellTags->"b:15.3"],

Cell[BoxData[
    \("...Generating temp variable declarations"\)], "Print",
  CellLabel->"From In[251]:=",
  CellTags->"b:15.3"],

Cell[BoxData[
    \("...Converting Expressions to C form"\)], "Print",
  CellLabel->"From In[251]:=",
  CellTags->"b:15.3"],

Cell[BoxData[
    \("...Generating Output Function Code"\)], "Print",
  CellLabel->"From In[251]:=",
  CellTags->"b:15.3"],

Cell[BoxData[
    \("MEX File created with name: truck.c"\)], "Print",
  CellLabel->"From In[251]:=",
  CellTags->"b:15.3"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Inputs", "Input",
  CellLabel->"In[267]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[BoxData[
    \({FDrive, delta}\)], "Output",
  CellLabel->"Out[267]=",
  CellTags->"b:15.3"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Outputs", "Input",
  CellLabel->"In[268]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[BoxData[
    \({39.37007874015748`\ x, 39.37007874015748`\ y}\)], "Output",
  CellLabel->"Out[268]=",
  CellTags->"b:15.3"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Constraints", "Section",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell["Introductory Remarks", "Subsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Constraints Intro", "b:16.0"}],

Cell["\<\
Many systems of interest do not have a simple tree structure but can be \
viewed in terms of a tree with additional algebraic and/or differential \
constraints. A key advantage of the Lagrange formulation is that the explicit \
formulation of the relations that constrain the configuration of a system is \
usually avoided by appropriate choice of the generalized coordinates. \
However, with complex systems, a correct choice of generalized coordinates \
may be far from obvious. 

One systematic method for dealing with mechanical configurations containing \
closed loops is to break (free) a minimal number of joints so as to obtain a \
tree structure which, of course, has a natural set of configuration \
coordinates. The loop-breaking joint constraints are then added as \
suplemental algebraic relations. The number of degrees of freedom of the \
system is typically reduced by the number of independent relations. If these \
relations are (globally) solvable for a like number of coordinates then these \
coordinates can be explicitly eliminated. When there are k such constraints, \
the number of coordinates is reduced by k and the number of quasi-velocities \
is also reduced by k. It may not be possible to eliminate all k excess \
coordinates, but it is always possible to eliminate k quasi-velocities.

Some chain and loop structures give rise to differential constraints - such \
as when rolling of one body on another occurs. In general, when k independent \
differential constraints are present it is possible to eliminate k \
quasi-velocities. If the constraints are 'partially integrable' then, in \
principle, it is also possible to eliminate some coordinates as well. When \
the constraints are 'completely integrable' the number of configuration \
degrees of freedom are reduced by k and the system is said to be 'holonomic.' \
Otherwise it is 'nonholonomic.'

Even when it is not possible or practical to eliminate all k configuration \
coordinates it is fairly straightforward to derived the reduced set of \
equations in which the k quasi-velocities are removed. These equations form a \
valid, closed system of differential equations that completely describe the \
system behavior. Presently, TSi Dynamics provides functions to generate this \
latter set of equations unless the user specifies a set of configuration \
coordinates that are to be removed.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Constraints Intro", "b:16.0"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["A Simple Closed Chain", "Subsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell["\<\
Consider a planar two bar linkage in which the lower end of bar 1 is \
constrained by a revolute joint on the y-axis at y=0 and the upper end of bar \
two slides on the z-axis. The upper end of bar 1 and the lower end of bar 2 \
are connected by a revolute joint.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell["\<\
r1={1};
H1=Transpose[{{1,0,0,0,0,0}}];
q1={theta1};p1={\[Omega]1};

r2={1};
H2=Transpose[{{1,0,0,0,0,0}}];
q2={theta2};p2={\[Omega]2};

JointList={{r1,H1,q1,p1},{r2,H2,q2,p2}};\
\>", "Input",
  CellLabel->"In[270]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell["\<\
Mass1=m1; CenterOfMass1={0,0,L1/2}; OutboardNode1={2,{0,0,L1}};
InertiaMatrix1=DiagonalMatrix[{m1*L1^2/12,m1*L1^2/12,0}];

Mass2=m2; CenterOfMass2={0,0,L2/2}; OutboardNode2={3,{0,0,L2}};
InertiaMatrix2=DiagonalMatrix[{m2*L2^2/12,m2*L2^2/12,0}];

BodyList={{CenterOfMass1,{OutboardNode1},Mass1,InertiaMatrix1},
          {CenterOfMass2,{OutboardNode2},Mass2,InertiaMatrix2}};\
\>", "Input",
  CellLabel->"In[277]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell["\<\
TreeList={{{1,1},{2,2}}};
PE=0; Q={0,0};\
\>", "Input",
  CellLabel->"In[282]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell[CellGroupData[{

Cell["\<\
{V,X,H,M,Cp,Fp,p,q}=CreateModel[JointList,BodyList,TreeList,g,PE,Q];\
\>", "Input",
  CellLabel->"In[284]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell[BoxData[
    \("Computing Joint Kinematics"\)], "Print",
  CellLabel->"From In[284]:=",
  CellTags->"b:16.1"],

Cell[BoxData[
    \("Computing joint 1 kinematics"\)], "Print",
  CellLabel->"From In[284]:=",
  CellTags->"b:16.1"],

Cell[BoxData[
    \("Computing joint 2 kinematics"\)], "Print",
  CellLabel->"From In[284]:=",
  CellTags->"b:16.1"],

Cell[BoxData[
    \("Computing Potential Functions"\)], "Print",
  CellLabel->"From In[284]:=",
  CellTags->"b:16.1"],

Cell[BoxData[
    \("Computing Inertia Matrix"\)], "Print",
  CellLabel->"From In[284]:=",
  CellTags->"b:16.1"],

Cell[BoxData[
    \("Computing Poincare Function"\)], "Print",
  CellLabel->"From In[284]:=",
  CellTags->"b:16.1"]
}, Open  ]],

Cell["\<\
ChnBodyList={{CenterOfMass1,{0,0,L1},Mass1,InertiaMatrix1},
             {CenterOfMass2,{0,0,L2},Mass2,InertiaMatrix2}};\
\>", "Input",
  CellLabel->"In[285]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell["EndPos=EndEffector[ChnBodyList,X];", "Input",
  CellLabel->"In[286]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell[CellGroupData[{

Cell["G=EndPos[[{2},4]]", "Input",
  CellLabel->"In[287]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell[BoxData[
    \({\(-L1\)\ Sin[theta1] + 
        L2\ \((\(-Cos[theta2]\)\ Sin[theta1] - 
              Cos[theta1]\ Sin[theta2])\)}\)], "Output",
  CellLabel->"Out[287]=",
  CellTags->"b:16.1"]
}, Open  ]],

Cell["\<\
The relation G=0 is not solvable for either theta1 or theta2 for all values \
of L1 and L2 - although, as we will describe below, it can be solved for \
particular values of these parameters. Thus, we will not try to eliminate any \
configuration coordinate. The constrained dynamics are obtained with:\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell[CellGroupData[{

Cell["\<\
{Mm,Cm,Fm,Vm,phat,qhat}=AlgebraicConstraints[M,Cp,Fp,V,G,p,q]\
\>", "Input",
  CellLabel->"In[288]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell[BoxData[
    \("Vm"\)], "Print",
  CellLabel->"From In[288]:=",
  CellTags->"b:16.1"],

Cell[BoxData[
    \("Mm"\)], "Print",
  CellLabel->"From In[288]:=",
  CellTags->"b:16.1"],

Cell[BoxData[
    \("Cm"\)], "Print",
  CellLabel->"From In[288]:=",
  CellTags->"b:16.1"],

Cell[BoxData[
    \("Fm"\)], "Print",
  CellLabel->"From In[288]:=",
  CellTags->"b:16.1"],

Cell[BoxData[
    \({{{1\/12\ L1\^2\ L2\^2\ \((2\ m1 + 5\ m2 - m2\ Cos[2\ theta1] - 
                3\ m2\ Cos[2\ theta2] + 
                2\ m1\ Cos[2\ \((theta1 + theta2)\)] + 
                3\ m2\ Cos[
                    2\ \((theta1 + 
                          theta2)\)])\)}}, {{\(-\(1\/24\)\)\ "w1"\ L1\^2\ \
L2\^2\ \((3\ L1\ m2\ Sin[theta1 - 2\ theta2] + 
                4\ L2\ m2\ Sin[theta1 - theta2] + 
                L2\ m2\ Sin[3\ theta1 + theta2] + 
                2\ L1\ m1\ Sin[theta1 + 2\ theta2] + 
                2\ L1\ m1\ Sin[3\ theta1 + 2\ theta2] + 
                3\ L1\ m2\ Sin[3\ theta1 + 2\ theta2] - 
                3\ L2\ m2\ Sin[
                    theta1 + 3\ theta2])\)}}, {1\/4\ gc\ L1\ L2\ \((\((m1 + 
                    3\ m2)\)\ Sin[theta2] - \((m1 + m2)\)\ Sin[
                  2\ theta1 + theta2])\)}, {{\(-L2\)\ Cos[
              theta1 + theta2]}, {L1\ Cos[theta1] + 
            L2\ Cos[theta1 + theta2]}}, {"w1"}, {theta1, theta2}}\)], "Output",\

  CellLabel->"Out[288]=",
  CellTags->"b:16.1"]
}, Open  ]],

Cell["\<\
These parameters define the constrained system dynamics. Now, let's consider \
the special case: L1=L2=L. This is the only situation where G=0 can be solved \
for either theta1 or theta2, so that either angle can be used to parameterize \
the system configuration. For simplicity, also set m1=m2=m. In this case the \
system matrices are:\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell[CellGroupData[{

Cell["\<\
{Mm,Cm,Fm,Vm}=Simplify[{Mm,Cm,Fm,Vm}/.{L1->L,L2->L,m1->m,m2->m}]\
\>", "Input",
  CellLabel->"In[289]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell[BoxData[
    \({{{\(-\(1\/12\)\)\ L\^4\ m\ \((\(-7\) + Cos[2\ theta1] + 
                3\ Cos[2\ theta2] - 
                5\ Cos[2\ \((theta1 + 
                          theta2)\)])\)}}, {{\(-\(1\/24\)\)\ "w1"\ L\^5\ m\ \
\((3\ Sin[theta1 - 2\ theta2] + 4\ Sin[theta1 - theta2] + 
                Sin[3\ theta1 + theta2] + 2\ Sin[theta1 + 2\ theta2] + 
                5\ Sin[3\ theta1 + 2\ theta2] - 
                3\ Sin[theta1 + 
                      3\ theta2])\)}}, {\(-\(1\/2\)\)\ gc\ L\^2\ m\ \((\(-2\)\
\ Sin[theta2] + Sin[2\ theta1 + theta2])\)}, {{\(-L\)\ Cos[
              theta1 + theta2]}, {L\ \((Cos[theta1] + 
                Cos[theta1 + theta2])\)}}}\)], "Output",
  CellLabel->"Out[289]=",
  CellTags->"b:16.1"]
}, Open  ]],

Cell["\<\
It is convenient to assemble the corresponding differential equations.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell[CellGroupData[{

Cell["Eqns=MakeODEs[phat,qhat,Vm,Mm,Cm,Fm,t]", "Input",
  CellLabel->"In[290]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          RowBox[{
            SuperscriptBox["theta1", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], 
          "==", \(\(-L\)\ Cos[theta1[t] + theta2[t]]\ w1[t]\)}], ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["theta2", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], 
          "==", \(L\ \((Cos[theta1[t]] + Cos[theta1[t] + theta2[t]])\)\ w1[
              t]\)}], ",", 
        RowBox[{
          
          RowBox[{\(\(-\(1\/2\)\)\ gc\ L\^2\ m\ \((\(-2\)\ Sin[theta2[t]] + 
                  Sin[2\ theta1[t] + theta2[t]])\)\), 
            "-", \(1\/24\ L\^5\ m\ \((3\ Sin[theta1[t] - 2\ theta2[t]] + 
                  4\ Sin[theta1[t] - theta2[t]] + 
                  Sin[3\ theta1[t] + theta2[t]] + 
                  2\ Sin[theta1[t] + 2\ theta2[t]] + 
                  5\ Sin[3\ theta1[t] + 2\ theta2[t]] - 
                  3\ Sin[theta1[t] + 3\ theta2[t]])\)\ w1[t]\^2\), "-", 
            
            RowBox[{\(1\/12\), " ", \(L\^4\), " ", "m", 
              " ", \((\(-7\) + Cos[2\ theta1[t]] + 3\ Cos[2\ theta2[t]] - 
                  5\ Cos[2\ \((theta1[t] + theta2[t])\)])\), " ", 
              RowBox[{
                SuperscriptBox["w1", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}]}], "==", 
          "0"}]}], "}"}]], "Output",
  CellLabel->"Out[290]=",
  CellTags->"b:16.1"]
}, Open  ]],

Cell["\<\
Notice that there are two kinematic equations and one dynamic equation. Since \
L1=L2=L it is easy to show that theta2=-2 theta1. Hence we can make this \
replacement to obtain:\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell[CellGroupData[{

Cell["SpEqns=Simplify[Eqns/.theta2[t]->-2*theta1[t]]", "Input",
  CellLabel->"In[291]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          RowBox[{
            SuperscriptBox["theta1", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], 
          "==", \(\(-L\)\ Cos[theta1[t]]\ w1[t]\)}], ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["theta2", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], 
          "==", \(2\ L\ Cos[theta1[t]]\ w1[t]\)}], ",", 
        RowBox[{
          
          RowBox[{\(1\/6\), " ", \(L\^2\), " ", "m", " ", \(Cos[theta1[t]]\), 
            " ", 
            RowBox[{"(", 
              
              RowBox[{\(12\ gc\ Sin[theta1[t]]\), 
                "+", \(L\^3\ \((\(-2\)\ Sin[2\ theta1[t]] + 
                      3\ Sin[4\ theta1[t]])\)\ w1[t]\^2\), "+", 
                
                RowBox[{\(L\^2\), 
                  " ", \((\(-7\)\ Cos[theta1[t]] + 3\ Cos[3\ theta1[t]])\), 
                  " ", 
                  RowBox[{
                    SuperscriptBox["w1", "\[Prime]",
                      MultilineFunction->None], "[", "t", "]"}]}]}], ")"}]}], 
          "==", "0"}]}], "}"}]], "Output",
  CellLabel->"Out[291]=",
  CellTags->"b:16.1"]
}, Open  ]],

Cell["\<\
Simple trigonometric identities can be used to verify that these equations \
are equivalent to those given by Ginsberg for this example.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(Rule1 = 
      D[w1[t], t] \[Rule] 
        D[D[theta1[t], t]/\((\(-L\)\ Cos[theta1[t]])\), t]\)], "Input",
  CellLabel->"In[292]:=",
  CellTags->"b:16.1"],

Cell[BoxData[
    RowBox[{
      RowBox[{
        SuperscriptBox["w1", "\[Prime]",
          MultilineFunction->None], "[", "t", "]"}], "\[Rule]", 
      RowBox[{
        RowBox[{"-", 
          FractionBox[
            RowBox[{\(Sec[theta1[t]]\), " ", \(Tan[theta1[t]]\), " ", 
              SuperscriptBox[
                RowBox[{
                  SuperscriptBox["theta1", "\[Prime]",
                    MultilineFunction->None], "[", "t", "]"}], "2"]}], "L"]}],
         "-", 
        FractionBox[
          RowBox[{\(Sec[theta1[t]]\), " ", 
            RowBox[{
              SuperscriptBox["theta1", "\[Prime]\[Prime]",
                MultilineFunction->None], "[", "t", "]"}]}], 
          "L"]}]}]], "Output",
  CellLabel->"Out[292]=",
  CellTags->"b:16.1"]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[\((SpEqns[\([3]\)] /. 
            w1[t] \[Rule] D[theta1[t], t]/\((\(-L\)\ Cos[theta1[t]])\))\) /. 
        Rule1]\)], "Input",
  CellLabel->"In[293]:=",
  CellTags->"b:16.1"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(1\/3\), " ", \(L\^2\), " ", "m", " ", \(Cos[theta1[t]]\), " ",
         
        RowBox[{"(", 
          RowBox[{\(6\ gc\ Sin[theta1[t]]\), "+", 
            RowBox[{"3", " ", "L", " ", \(Sin[2\ theta1[t]]\), " ", 
              SuperscriptBox[
                RowBox[{
                  SuperscriptBox["theta1", "\[Prime]",
                    MultilineFunction->None], "[", "t", "]"}], "2"]}], "-", 
            RowBox[{"L", " ", \((\(-5\) + 3\ Cos[2\ theta1[t]])\), " ", 
              RowBox[{
                SuperscriptBox["theta1", "\[Prime]\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}]}], ")"}]}], "==",
       "0"}]], "Output",
  CellLabel->"Out[293]=",
  CellTags->"b:16.1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Another Closed Chain", "Subsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],

Cell["\<\
Consider a two bar linkage in the x-y plane. The left end of bar 1 is \
constrained to slide on the x-axis and the right end of bar two slides on the \
y-axis. The right end of bar 1 and the left end of bar 2 are connected by a \
revolute joint.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],

Cell["\<\
r1={1,1};
H1=Transpose[{{0,0,1,0,0,0},{0,0,0,1,0,0}}];
q1={theta1,x};p1={w1,vx};

r2={1};
H2=Transpose[{{0,0,1,0,0,0}}];
q2={theta2};p2={w2};

JointList={{r1,H1,q1,p1},{r2,H2,q2,p2}};\
\>", "Input",
  CellLabel->"In[294]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],

Cell["\<\
Mass1=m1; CenterOfMass1={L1/2,0,0}; OutboardNode1={2,{L1,0,0}};
InertiaMatrix1=DiagonalMatrix[{0,m1*L1^2/2,m1*L1^2/2}];

Mass2=m2; CenterOfMass2={L2/2,0,0}; OutboardNode2={3,{L2,0,0}};
InertiaMatrix2=DiagonalMatrix[{0,m2*L2^2/2,m2*L2^2/2}];

BodyList={{CenterOfMass1,{OutboardNode1},Mass1,InertiaMatrix1},
          {CenterOfMass2,{OutboardNode2},Mass2,InertiaMatrix2}};\
\>", "Input",
  CellLabel->"In[301]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],

Cell["\<\
TreeList={{{1,1},{2,2}}};
PE=0; Q={0,0,0};\
\>", "Input",
  CellLabel->"In[306]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],

Cell[CellGroupData[{

Cell["\<\
{V,X,H,M,Cp,Fp,p,q}=CreateModel[JointList,BodyList,TreeList,g,PE,Q];\
\>", "Input",
  CellLabel->"In[308]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],

Cell[BoxData[
    \("Computing Joint Kinematics"\)], "Print",
  CellLabel->"From In[308]:=",
  CellTags->"b:16.2"],

Cell[BoxData[
    \("Computing joint 1 kinematics"\)], "Print",
  CellLabel->"From In[308]:=",
  CellTags->"b:16.2"],

Cell[BoxData[
    \("Computing joint 2 kinematics"\)], "Print",
  CellLabel->"From In[308]:=",
  CellTags->"b:16.2"],

Cell[BoxData[
    \("Computing Potential Functions"\)], "Print",
  CellLabel->"From In[308]:=",
  CellTags->"b:16.2"],

Cell[BoxData[
    \("Computing Inertia Matrix"\)], "Print",
  CellLabel->"From In[308]:=",
  CellTags->"b:16.2"],

Cell[BoxData[
    \("Computing Poincare Function"\)], "Print",
  CellLabel->"From In[308]:=",
  CellTags->"b:16.2"]
}, Open  ]],

Cell["\<\
ChnBodyList={{CenterOfMass1,{L1,0,0},Mass1,InertiaMatrix1},
             {CenterOfMass2,{L2,0,0},Mass2,InertiaMatrix2}};\
\>", "Input",
  CellLabel->"In[309]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],

Cell["EndPos=EndEffector[ChnBodyList,X];", "Input",
  CellLabel->"In[310]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],

Cell[CellGroupData[{

Cell["G=EndPos[[{1},4]]", "Input",
  CellLabel->"In[311]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],

Cell[BoxData[
    \({x + L1\ Cos[theta1] + 
        L2\ \((Cos[theta1]\ Cos[theta2] - 
              Sin[theta1]\ Sin[theta2])\)}\)], "Output",
  CellLabel->"Out[311]=",
  CellTags->"b:16.2"]
}, Open  ]],

Cell["\<\
Since G=0 always has a solution for x, we will eliminate x from the \
constrained equations.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],

Cell[CellGroupData[{

Cell["\<\
{Mm,Cm,Fm,Vm,phat,qhat}=AlgebraicConstraints[M,Cp,Fp,V,G,p,q,{2}]\
\>", "Input",
  CellLabel->"In[312]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],

Cell[BoxData[
    \("Vm"\)], "Print",
  CellLabel->"From In[312]:=",
  CellTags->"b:16.2"],

Cell[BoxData[
    \("Mm"\)], "Print",
  CellLabel->"From In[312]:=",
  CellTags->"b:16.2"],

Cell[BoxData[
    \("Cm"\)], "Print",
  CellLabel->"From In[312]:=",
  CellTags->"b:16.2"],

Cell[BoxData[
    \("Fm"\)], "Print",
  CellLabel->"From In[312]:=",
  CellTags->"b:16.2"],

Cell[BoxData[
    \("Eliminating specified coordinates."\)], "Print",
  CellLabel->"From In[312]:=",
  CellTags->"b:16.2"],

Cell[BoxData[
    \({{{1\/4\ \((3\ L1\^2\ m1 + 2\ L2\^2\ m1 + 2\ L1\^2\ m2 + 3\ L2\^2\ m2 + 
                2\ L1\^2\ m2\ Cos[2\ theta1] + 
                2\ L1\ L2\ \((m1 + m2)\)\ Cos[theta2] - 
                2\ L2\^2\ m1\ Cos[2\ \((theta1 + theta2)\)] - 
                2\ L1\ L2\ m1\ Cos[2\ theta1 + theta2] + 
                2\ L1\ L2\ m2\ Cos[
                    2\ theta1 + 
                      theta2])\), \(-\(1\/4\)\)\ L2\ \((\(-2\)\ L2\ m1 - 
                3\ L2\ m2 - L1\ \((m1 + m2)\)\ Cos[theta2] + 
                2\ L2\ m1\ Cos[2\ \((theta1 + theta2)\)] + 
                L1\ m1\ Cos[2\ theta1 + theta2] - 
                L1\ m2\ Cos[
                    2\ theta1 + 
                      theta2])\)}, {\(-\(1\/4\)\)\ L2\ \((\(-2\)\ L2\ m1 - 
                3\ L2\ m2 - L1\ \((m1 + m2)\)\ Cos[theta2] + 
                2\ L2\ m1\ Cos[2\ \((theta1 + theta2)\)] + 
                L1\ m1\ Cos[2\ theta1 + theta2] - 
                L1\ m2\ Cos[
                    2\ theta1 + 
                      theta2])\), \(-\(1\/4\)\)\ L2\^2\ \((\(-2\)\ m1 - 
                3\ m2 + 
                2\ m1\ Cos[
                    2\ \((theta1 + 
                          theta2)\)])\)}}, {{1\/8\ \((\(-4\)\ L1\^2\ m2\ w1\ \
Sin[2\ theta1] + 
                L2\ \((\(-L1\)\ \((3\ m1 + m2)\)\ w2\ Sin[theta2] + 
                      4\ L2\ m1\ \((w1 + w2)\)\ Sin[
                          2\ \((theta1 + theta2)\)] + 
                      L1\ \((4\ m1\ w1 - 4\ m2\ w1 + m1\ w2 - 
                            3\ m2\ w2)\)\ Sin[2\ theta1 + theta2])\))\), 
          1\/8\ L2\ \((\(-L1\)\ \((m1\ w1 + 3\ m2\ w1 + 2\ m1\ w2 + 
                      2\ m2\ w2)\)\ Sin[theta2] + 
                4\ L2\ m1\ \((w1 + w2)\)\ Sin[2\ \((theta1 + theta2)\)] + 
                L1\ \((3\ m1\ w1 - m2\ w1 + 2\ m1\ w2 - 2\ m2\ w2)\)\ Sin[
                    2\ theta1 + theta2])\)}, {1\/8\ L2\ \((L1\ \((2\ m1\ w1 + 
                      m2\ \((2\ w1 + w2)\))\)\ Sin[theta2] + 
                4\ L2\ m1\ \((w1 + w2)\)\ Sin[2\ \((theta1 + theta2)\)] + 
                L1\ \((2\ m1\ w1 - 2\ m2\ w1 + m2\ w2)\)\ Sin[
                    2\ theta1 + theta2])\), 
          1\/4\ L2\ \((\(-L1\)\ m2\ w1\ Cos[theta1] + 
                4\ L2\ m1\ \((w1 + w2)\)\ Cos[theta1 + theta2])\)\ Sin[
              theta1 + theta2]}}, {0, 0}, {{1, 0}, {0, 1}}, {w1, 
        w2}, {theta1, theta2}}\)], "Output",
  CellLabel->"Out[312]=",
  CellTags->"b:16.2"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Simplify[Cm.phat]", "Input",
  CellLabel->"In[313]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],

Cell[BoxData[
    \({1\/4\ \((\(-2\)\ L1\^2\ m2\ w1\^2\ Sin[2\ theta1] + 
            L2\ \((\(-L1\)\ \((m1 + m2)\)\ w2\ \((2\ w1 + w2)\)\ Sin[
                      theta2] + 
                  2\ L2\ m1\ \((w1 + w2)\)\^2\ Sin[
                      2\ \((theta1 + theta2)\)] + 
                  L1\ \((m1 - m2)\)\ \((2\ w1\^2 + 2\ w1\ w2 + w2\^2)\)\ Sin[
                      2\ theta1 + theta2])\))\), 
      1\/4\ L2\ \((L1\ \((m1 + m2)\)\ w1\^2\ Sin[theta2] + 
            2\ L2\ m1\ \((w1 + w2)\)\^2\ Sin[2\ \((theta1 + theta2)\)] + 
            L1\ \((m1 - m2)\)\ w1\^2\ Sin[2\ theta1 + theta2])\)}\)], "Output",\

  CellLabel->"Out[313]=",
  CellTags->"b:16.2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
The Rolling Disk: a Classic Example of Differential Constraints\
\>", "Subsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell["\<\
As an illustration we take an example from Neimark and Fufaev. Consider a \
disk that rolls without slipping on the x-y plane (z is up). Assume that the \
disk is of mass m and radius R. One approach is to ignore the the rolling \
constraints and formulate the equations of motion for the free disk in space, \
and then add the required constraints. A body frame is established with \
origin at the center of the disk. The six degree of freedom, simple joint is \
defined by: \
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[GraphicsData["Metafile", "\<\
CF5dJ6E]HGAYHf4PEfU^I6mgLb15CDHPAVmbKF5d0@0008A40@0006`0000100000@000?@1003G0000
00000000000E?@00FaX00215CDH00040A8@004l300070000000000000000000000@0000300100@00
l0000000000000000000003R1020Z@<0AP0002`0000P0000ADe6:`500@0L00004000004@`=/10000
H000060000160000Y00009P00015CDH[<4020100000400000020?am0100<0000000001i00@0<0000
000001e0000D000020000000003G000024000dP0000l00000A30f`D00000000058ki@eb?XSmLKk93
Aj5F@eb?XSm7XEI3Aj5F@eb?XSlDS_U3G8nR?`010@6100005400P10000040000i^KVob4000080000
HP0000`000010000900002@00000080m000000000000080m00000000000200009`0001P000010000
00000>KViP0000009@0000`0000100004`0000`0000100009@0000`000080020EP00030000010000
0@000?@1003G00001@00038O5@1>5V/=5@1[3F/=5@0b7aD09@0000`0000700204`0000`000010000
9@0000`000000020900002@00000081100000000000008110000000000020000AP000>`0003P0000
ADe6:`Q00@8l0000<000004@`=/00000CP000000002gIJdo0P000080000200000P00004@`=/00000
omnoo`Q000>@0000Q000004@`=/=0000000009UI1D>5ZaU3L?d]@hF[6D?4HU93Ph_d@VbDMT<ncNm2
U;>4@kmEkD;f5Hi3^i7j@X<AV4=IgP139Z^R@iF/14=mlJe3A0l5@jfJ]D>IFOI2OhFl@nUoi48=O/13
<>O1@TO1aT>5JkU200<30`<30`<30`<30`0001E0000@000010000040000T000090000000P3d00000
00000000P3d0000000000080001O0000>0000080000h0000000003P000000000000101H000000000
_moo000000000000000002D0000<00000P0002D0000<00001@00P5D0001@0000P`0005/0002@0@00
W00000d0001F29/9h0ZK2BL=Y@MZ3gl7Ua1[1l<Ae@L34`h8EQA;2;lED@Rd5[<7TALT1a0H40KI6<`5
9@0000`0000700209@0000`000000020900002@00000081100000000000008110000000000020000
:00000`000020000AP000:@0002H0000ADe6:bY0000T0000600005b?XT800000000005b?XT:GffI2
TD0=@`Q00PH`00009000004@`=^[ZRXn00000000000000001@000440DP190440C00003I00X100000
=00000000?l100000@000040001H01b7HchGK44n0020?`00000000000020?`0000000000000B0000
30000040000H000030000000000F0000300001P0001B0000L0400080003boooo0000000000000000
T0400000000700@0@@1b06T0H@1/0000001=06l0KP1_07@0N@1`06D0>P110780J@1Q06`0801B06D0
I`1e06`0H@1b03X0EP1U0000ooooo`00000006/U060<f_P0<;_i02`e500f7@1P<;_i02`e50087P1P
;3DD0000000000005T<0H2`e500`^oT0o3TD0<1PCV28kA80<;_i0112060000000@000000001PkQ80
S>hB02`e502PkA809c`0H?`i502<kQ8000000040003HkA80`c`0H?`i502<kQ80000001@d500P6`1P
000001@d50000000_QX0H8C^4P1PkQ80000000000000kQ80hCl0H8C^4P0000000@000=7a4P1PkQ80
0000080`>c<00000>>hB00D0001`>OQg001X<GP7J35TMP08000002D0000<00000P0005@0001D0000
C0000900001C0000W`00004000000?Y1003j@D`0002M00000@0004`00000000000000000003ooooo
oooooe00001H0020000002D0000<00003@00P4H0002`0000Y00004E=AR/[@00030000000000Z@000
900001P0001LSj9200000000001LSj92E_A7@dcMlD88@086<00002@000014<3KZjXZ?P0000000000
000000D000110580B@1104`0000f@0:0@00003@00000003o0@00004000010000F@0i3/Lm5fa1?P00
P3l0000000000000P3l00000000000004P0000`000010000600000`0000000005P0000`0000H0000
9@0000`000020000E00005@0003@0000N`000=P0002:00000@000000nT400?Y1d00008P000010000
C00000000000000000000?ooooooooooD00005T0mGL000009@0000`0000=0020AP000;00002T0000
ADe6:b]0000<0000000002Y0000T0000600005b?XT800000000005b?XT9`_Le2D[R/@@Q00PH`0000
9000004@`=^[ZRXn00000000000000001@000440DP190440C00003I00X100000=00000000?l10000
0@000040001J0<MA63hGK44n0020?`00000000000020?`0000000000000B000030000040000H0000
30000000000F0000300001P0000U000030000080001D0000E00007<0000H0000NP0002L000010000
003j@@00nT5c00009@000040001<00000000000000000000oooooooooom@0000FP000000000U0000
300000d008160000e0000<P00015CDH[:d0000`00000000024000jP0002L00000A30fa0000000000
Pf^i@Ul?hT:0JkU2Vb^e@VW?gD:^ai12UET5@jo7T49fba]3Z/N@@Vgm;D>A:kE2Kod]@e/?hT9`oBe3
F`oR@W3m;D=L3n92Kod]@e`?hT9`oBe3SgT7@g[;6d>4ZaU3VET5@hB[6D=`cme2Q:/I@hE[^D:@N@M3
Q6^i@Uh?hT800`<30`<30`<30`<30`>35400P10000040000WOooobP0000<00000@0002@0000T0000
0020?@00000000000020?@00000000000P0002L0000H00000@000000003ooid0000002D0000<0000
0@0001<0000<00000@0003/000080000E@0005`00000000000000?oooooooooo40000<`54@O<1JX5
k`J715H8Q`Bm2HL4h0ZZ1N0:4@OP2Q47h0XA1n0:4@OP2WP8_@VK2EH8V`W_1Y/9c0Eh2<`54@Ll0000
200003h0000H0000G00004P0002^0000VP0001<0000<00000@0002D0000<00000000P2@0000T0000
0020@@00000000000020@@00000000000P0004H0001L0000D00004E=AR/8@042?000030000014<3K
000004h000000000a[N:?P80000200000P00008000014<3K000000000?lE@000400000@000010000
900002@00000080m000000000000080m0000000000020000G`0003P000030000>0000000000h0000
000000000@040000000000000000000000000000000U0000300000<0000U0000300000D0081E0000
G00005/000170000/00009/0000@0000c0DA1l`5ZPG_1XL4EPR71;d9Q`CP2ZX5h0XA1n0:4@OP2Q47
h0XA1n0:N0Rm2I/9EPRK2Nl6V`W<1GP8c0DA1bD0000<00001`00P2D0000<00000000P2@0000T0000
0020@@00000000000020@@00000000000P0002P0000<00000`0004H000280000O00004E=AR/8@042
?000030000014<3K000004h000000000ZA=@?`80000200000P00008000014<3K000000000?l8@003
;000020000014<3K0P000000002IF@E3G0oR@YUI1D=[VQQ2004001E0000@000010000040000T0000
90000000P3d0000000000000P3d0000000000080001O0000>00000<0000h0000000003P000000000
000100d0000000000000000000000000000002D0000<00000`0002D0000<00001@00P5L0000T0000
Q00002@000270000L`000080001F2147EPQS0RD0000<00001`00P2D0000<00000000P2@0000T0000
0020@@00000000000020@@00000000000P0002P0000<00000`0004H0001L0000D00004E=AR/8@003
?000030000014<3K100000000011/@53a48L@YUI1D<00?i1lP49@lA27491/@53a48L@P010H4D@020
400000@00000003o:00000`000010000900002@00000080m000000000000080m0000000000020000
9`0001P00001000000000000000000009@0000`0000100004`0000`0000100009@0000`000080020
EP0002`0002100007`0008X0000X0000100001`8LP9F2?`1T@Qb0Q`8LP8U0000300000L0080C0000
30000040000U000030000000080T000090000000P440000000000000P44000000000008000160000
R00007`00015CDH[24010S`0000`00000A30f`00001>000000000:TCD3l200000P00008000020000
0A30f`000000003o24000b`0000P00000A30f`8000000000VET5@e`?hT:]>U13G0oR@P01000E@000
400000@000010000900002@00000080m000000000000080m0000000000020000G`0003P000030000
>0000000000h0000000000000@0=0000000000000000000000000000000U0000300000<0000U0000
300000D0081G0000900008@0001_0000dP0007<000020000EPPA1`@=4@LU0000300000L0080U0000
30000000080T000090000000P440000000000000P440000000000080000X0000300000<000160000
G00005000015CDH[24000c`0000`00000A30f`@000000000Ue1?@j^nfT97XEI3G0oR@YM@Cd<=H>U2
Ue1?@j^nfT800@615400P100000400000000ob@0000T00000020?@00000000000020?@0000000000
0P0002D0000<00000@0001<0000<00000@0002D0000<00002000P5H0000/0000c`0006d0003G0000
M@0000@0003f3=H6J`dA1oH<B`Of3=H69@0000`0000700204`0000`0000100009@0000`000000020
900002@00000081100000000000008110000000000020000AP0008P0001l0000ADe6:`Q00@8l0000
<000004@`=/00000CP000000002Y4e0o0P000080000200000P00004@`=/000000000o`Q000</0000
8000004@`=/20000000009UI1D=L3n92fgS2@]XT5D<00@005D000100000400000@0002@0000T0000
0020?@00000000000020?@00000000000P0005l0000h00000`0003P000000000>000000000000040
3@0000000000000000000000000000009@0000`0000300009@0000`000050020E`0002@0001O0000
K`0008L0002G00000P0005H84@LD1U<99@0000`0000700209@0000`000000020900002@000000811
00000000000008110000000000020000:00000`000030000AP0005`0001@0000ADe6:`Q000<l0000
<000004@`=/40000000001W`b49/5AM3QF^i@XF[6D>fUki2>nTA@aW`b49/5AM30041PAA0080@0000
100000000?lT000090000000P3d0000000000000P3d0000000000080000U000030000040000C0000
30000040000U0000300000P0081F0000;00005`0002A0000I@0009X000040000B0Ib2L`5V`We1Al9
B0Ib2BD0000<00001`00P1<0000<00000@0002D0000<00000000P2@0000T00000020@@0000000000
0020@@00000000000P0004H000380000_00004E=AR/8@003Z00009`000014<3K40000000002_kj13
dEk5@SICVd>YOYi2DK>K@cTgId9GaZ53hBQ;@UgIYd>06Rm2Le:a@jQSEd;]k[I30I:B@^g^]T<1TY92
kNjf@`6BTT;]k[I30Y:B@VF;_4<YL[U2BR^l@cAEi4936;I3HEcb@Sd5/4?7<@139hbV@oLnk4:_kj13
d5k5@P030`<30`<30`<30`<30h<D@020400000@0002Moooo:00000`000010000900002@00000080m
000000000000080m00000000000200009`0001P00001000000000?ooW@0000009@0000`000010000
4`0000`000010000>`0000P0001E0000G000000000000000ooooooooool@00007Q@[1V/Cm0Ag4id3
>A@]0o`D_@8[5Uh3gQJE1=hFU@CN5YD4gQJE198Gc0F65b<7a1JC1`4F10SB56877Q@[1S`000080000
?P0001P0000f0@00:`0007X1002100004`0000`0000100009@0000`000000020900002@000000811
00000000000008110000000000020000AP0005`0001@0000ADe6:`Q00@8l0000<000004@`=/00000
CP000000002Y4e0o0P000080000200000P00004@`=/000000000oaE0000@000010000040000T0000
90000000P3d0000000000000P3d0000000000080001O0000>00000<0000h0000000003P000000000
000100d0000000000000000000000000000002D0000<00000`0002D0000<00001@00P5D0001L0000
=@4002X0001k0@00PP000100000N52/6Ja?d17LCW@<i52d3o1Bm0R/FGP?N5YD4gQJE1=hFU@CN5YD4
TQO<1HHG8`O45Y<70AH42=8DHPLN52/69@0000`0000700209@0000`000000020900002@000000811
00000000000008110000000000020000:00000`000030000AP000;P0002/0000ADe6:bY0000T0000
60000?G7S49LSb;2G8lR@_G7S4;8P<Q3lNY9@PQ00PI40000>000004@`=^[ZRXn000000<000000000
3`0005@0B@1=04D0D`0P04h0A@1G0200DP1?04d0@@1>0000=T02P400000d00000000o`4000010000
0@0007T0i7SR?AM/@Ch0080o000000000000080o0000000000000180000<00000@0001P0000<0000
000001H0000<0000600002P0000<00000P000580001`0@000P000?;oool00000;04002`1002l0P00
0@0000L0101D06T0K@1U07<0801>06D0M`0P0580K`1]0640KP000000C@1_06h0K`1d07T0L01U03X0
E01Y06d0I@1c0200003ooooo00000000JbD0H0cJn00`^oT0;3DD03HM060`^oT0;3DD00PN060/=A@0
00000000000F@`1P;3DD032kn@3l>A@0`61>H53]4P0001<010000000000100000000007^4P1DkQ80
e>`B0000002LkA80UB_hMaPfn7L00?oo[>dB089cj7L001<0000003RP;3?V1@X`>:0/<l1GmgL00000
53@D0000002n6P1PN>dB0000002<ma80003ZMk1ej7Oooooo0@0004mmm7LhX2`c40000>0f1C?XkA80
>:0/<l@7mGO0EoMgM@DQgP80003V1@X`iPD:<6Af00P000009@0000`000020000E00005@0002D0@00
:0000:X1001200000@000000nT400?Y1X04003/000010000C00000000000000000000?oooooooooo
D00007T0000000009@0000`0000=0020AP0009@000280000ADe6:b]0000<0000000000Q00@8l0000
<000004@`=/00000CP000000002Y4e0o0P000080000200000P00004@`=/000000000o`Q000</0000
8000004@`=/2000000000;i;Zd?dJ:Q2NI>H@i;LVT400@005D000100000400000@0002@0000T0000
0020?@00000000000020?@00000000000P0005l0000h00000`0003P000000000>000000000000040
3@0000000000000000000000000000009@0000`0000300009@0000`000050020E`0002@0000_0@00
4P0005P1001F00000P0006XEA0DC4cH19@0000`0000700209@0000`000000020900002@000000811
00000000000008110000000000020000:00000`000030000AP0005`0001@0000ADe6:`Q000<l0000
<000004@`=/40000000008`hUd?0djm1doVF@oP7GD5bHiY3oI2B@H`hUd?0djm10041PAA0080@0000
100000000?lX000030000040000T000090000000P3d0000000000000P3d0000000000080000W0000
600000400000000000000000000U000030000040000C000030000040000U0000300000P0081F0000
;00002h1000=0000=@4001H000040000j19P0N0Bg@1=4bH1j19P0BD0000<00001`00P1<0000<0000
0@0002D0000<00000000P2@0000T00000020@@00000000000020@@00000000000P0004H000280000
O00004E=AR/8@042?000030000014<3K000004h000000000ZA=@?`80000200000P00008000014<3K
000000000?l8@003;000020000014<3K0P000000002nBj]3m6RX@S:hbd?23c]2004001E0000@0000
10000040000T000090000000P3d0000000000000P3d0000000000080001O0000>00000<0000h0000
000003P000000000000100d0000000000000000000000000000002D0000<00000`0002D0000<0000
1@00P5L0000T0000E@4002d0002I0@00EP000080001Z5D@5MaW]0RD0000<00001`00P2D0000<0000
0000P2@0000T00000020@@00000000000020@@00000000000P0002P0000<00000`0004H0001L0000
D00004E=AR/8@003?000030000014<3K10000000002oJ<Y3ESP`@[amcT>=@Ri2k3c<@hN?BD:oJ<Y3
ESP`@P010H4D@020400000@00000003o900002@00000080m000000000000080m0000000000020000
9@0000`0000100004`0000`0000100009@0000`000080020EP0002`0002D0@00:`0009d1000c0000
100004hI`@;@6KX2R1TW0dhI`@8U0000300000L0080C000030000040000U000030000000080T0000
90000000P440000000000000P44000000000008000160000R00007`00015CDH[24010S`0000`0000
0A30f`00001>000000000:TCD3l200000P000080000200000A30f`000000003o24000b`0000P0000
0A30f`8000000000_T^[@oAXZ4;2[jA3WXT5@`01000E@000400000@000010000900002@00000080m
000000000000080m0000000000020000G`0003P000030000>0000000000h0000000000000@0=0000
000000000000000000000000000U0000300000<0000U0000300000D0081G0000900004P1001B0000
F04008L000020000JQE41IHDF@PU0000300000L0080U000030000000080T000090000000P4400000
00000000P440000000000080000X0000300000<000160000G00005000015CDH[24000c`0000`0000
0A30f`@000000000AY:V@mZI1D>efj=3HkP;@mD9Xd<Y]@=3AY:V@mZI1D<00@615400P10000040000
0000ob@0000T00000020?@00000000000020?@00000000000P0002D0000<00000@0001<0000<0000
0@0002D0000<00002000P5H0000/0000AP4008<0001>0@00S00000@0003C55X8O1Bl268D?0SC55X8
9@0000`0000700204`0000`0000100009@0000`000000020900002@0000008110000000000000811
0000000000020000AP000;P0002/0000ADe6:bY0000T000060000?G7S49LSb;2G8lR@_G7S488I9A3
n1d6@`Q00PI40000>000004@`=^[ZRXn000000<0000000003`0005@0B@1=04D0D`0P04h0A@1G0200
DP1?04d0@@1>0000=T02P400000d00000000o`40000100000@0007P0ag4/?QM/@Ch0080o00000000
0000080o0000000000000180000<00000@0001P0000<0000000001H0000<0000600002D0000<0000
0P0005@0001D0000<@4007X000170@00T`00004000000?Y1003j@C`1002=00000@0004`000000000
00000000003ooooooooooe00001h0000000002D0000<00003@00P4H000340000^00004E=AR/[@000
30000000000Z@000900001P0003eaha2G8lR`Ub?8T;eaha21MB:@gT:F448@086A00003P000014<3K
ZjXZ?P0000030000000000l0001D04T0C@1505<0801>04D0E`0P0580C`1=0440CP0003I00X100000
=00000000?l100000@000040001j0<Naa3fJVATn0020?`00000000000020?`0000000000000B0000
30000040000H000030000000000F0000300001P0000U000030000080001D0000E00001P100020000
;04001X000010000003j@@00nT4S0@0050000040001<00000000000000000000oooooooooom@0000
NP000000000U0000300000d008160000X00009@00015CDH[:d0000`00000000024010TP0000l0000
0A30f`00001>0@0000000:TCD3l200000P000080000200000P000000044008100A30f`000000003o
24000b`0000P00000A30f`8000000000<g>[@mNSYd:h?[I3cDca@P01000E@000400000@000010000
900002@00000080m000000000000080m0000000000020000G`0003P000030000>0000000000h0000
0000000B0@0@0000000000000000000000000000000U0000300000<0000U000030000040000C0000
30000080000k0000200001/0000@0000MQD003X5000f000040000:hE002J1@00F00003@000000000
00000?oooooooooo1P000;4EWPF_5J<5[1FU1JPEY`FS5JH5XAFR1CH0000@0000J1D00485001H0000
=000000000000000ooooooooool60000IQDn1FLE>@E[5CL5KaDd1G@E=PEf5CX5?@0000P0000K0000
40000=LE003O1@00=P000100000@5P00@0H005P0000d000000000000003oooooooooo`H0000B5T@6
41I91PdFB`H95Td611I<1P8FB0Hf000040000<TE003W1@00F00003@00000000000000?oooooooooo
1P000<LEh`G85Ml5c1GL1M0EfPGE5M/5eaGO1Cd0000800006`000100000h5P00Q@H003H0000@0000
LAH00>D6001H0000=000000000000000ooooooooool60000LaKY1W8FkPI^5_06JQKc1VDFl@IS5^h6
=P000100000Z5P00S@H005P0000d000000000000003oooooooooo`H0000X5XT6:AJ41RdFPPHa5X06
=QJ11SPFQ@Hm0000200001/0000@0000VAH002X7000f000040000<lF00271`00F00003@000000000
00000?oooooooooo1P000=4FR`O@5Y07c1JB1lTFU0O45Y<7`AJ?1cH0000@0000RaH003<7001H0000
=000000000000000ooooooooool60000RAH_1hXF:PN>5RP7TQHU1iLF9`NI5RX7?@0000P0000l0000
200003l0000H0000E0400540001_0@00N`0001<0000<00000@0002D0000<00000000P2D0000<0000
1`00P2@0000T00000020@@00000000000020@@00000000000P0002P0000<00000`0004H000280000
O00004E=AR/8@042?000030000014<3K000004h000000000ZA=@?`80000200000P00008000014<3K
0000003o0?l8@003;000020000014<3K0P000000000m:[5385;o@QnUcd>?@Zm2004001E0000@0000
10000040000T000090000000P3d0000000000000P3d0000000000080001O0000>00000<0000h0000
000003P000000000000100d0000000000?l0000000000000000002D0000<00000`0002D0000<0000
1@00P5L0000T0000H@4005H0002Q0@00PP000080000V5_/7mAUk1BD0000<00001`00P2D0000<0000
0000P2@0000T00000020@@00000000000020@@00000000000P0002P0000<00000`0004H000280000
O00004E=AR/8@042?000030000014<3K000004h000000000ZA=@?`80000200000P00008000014<3K
0000003o0?l8@003;000020000014<3K0P000000002h?[I3cDca@SgZd4?=C?52004001E0000@0000
10000040000T000090000000P3d0000000000000P3d0000000000080001O0000>00000<0000h0000
000003P000000000000100d0000000000?l0000000000000000002D0000<00000`0002D0000<0000
1@00P5L0000T0000J`4007L0002T0@00N`00008000385X/77QZ;1bD0000<00001`00P2D0000<0000
0000P2@0000T00000020@@00000000000020@@00000000000P0002P0000<00000`0004H0002D0000
R00004E=AR/8@042B00003`000014<3K000004h100000000ZA=@?`80000200000P00008000020000
0020?`00P4014<3K00000?l00?l8@003;000020000014<3K0P00000000325Ja352jZ@ZB`^T<0P;i2
004001E0000@000010000040000X000030000040000T000090000000P3d0000000000000P3d00000
00000080001O0000>0000040000h0000000003P0000000000181010000000000003o000000000000
000002D0000<00000@0002L0000H00000`00000000000?l0000002D0000<00000`0001<0000<0000
0P0003/0000800006`00010000265@00BPD003H0000@0000QQD004X5001H0000=000000000000000
ooooooooool60000RQE<1H`ED0F;5ED5RAEI1HDEF`F15ET5=P00010000205@00F@D005P0000d0000
00000000003oooooooooo`H0001l5EP5NQEC1G`EC`Em5D/5PQE91HHEBPDm0000200001/0000@0000
dAD006D5000f000040000=4E001U1@00F00003@00000000000000?oooooooooo1P000=HEIPGH5F/5
eQE_1MDEL`G@5GD5c1Ed1CH0000@0000c1D007@5001H0000=000000000000000ooooooooool60000
b1Eb1LHEKPG75FT5bAEU1LdEH`GA5FD5?@0000P0000K0000400001dF001o1@00=P000100000M5P00
O`D005P0000d000000000000003oooooooooo`H0000Q5X058aJ51B8FR@DP5Xd571J@1APFSPDf0000
400001PF002>1@00F00003@00000000000000?oooooooooo1P0001<FS@DA5XP54aJ41A@FP0DI5Wd5
7AIo1Cd0000800006`000100001Y5P00V@D003H0000@0000JAH009T5001H0000=000000000000000
ooooooooool60000KAJK1FlFW`E]5Z<5K1JX1FLFZPES5ZP5=P000100001S5P00Z0D005P0000d0000
00000000003oooooooooo`H0001O5ZL5GAJR1EhFWPEP5YX5I1JH1FTFV@Dm0000200001/0000@0000
]1H00;@5000f000040000;@F002d1@00F00003@00000000000000?oooooooooo1P000;PF]@Fk5[X5
^AJn1KPF`PFc5/@5[aK31CH0000@0000[aH00<<5001H0000=000000000000000ooooooooool60000
ZaK11JPF_@FZ5[P5ZaJd1K0F/PFd5[@5?@0000P0000K00004000000G003>1@00=P00010000005`00
cPD005P0000d000000000000003oooooooooo`H000045ll51QOD1@DGf0D35m`5oaKN1OXFg@Df0000
40000?XF003M1@00F00003@00000000000000?oooooooooo1P000?HFf`Gd5]L5mQKC1OLFcPGl5/`5
01O>1Cd0000800006`000100001;5`00j0D003H0000@0000BaL00>P5001H0000=000000000000000
ooooooooool60000D1OZ1E8GkPE@5o85CaOf1DXGn@E65oL5=P00010000165`00m`D005P0000d0000
00000000003oooooooooo`H000125oH5@1Oa1D4Gk@E35nT5AaOW1D/Gj0Dm0000200003`000080000
?`0001P0001F0@00D`0007L1001R00004`0000`0000100009@0000`0000000209@0000`000070020
900002@00000081100000000000008110000000000020000:00000`000010000AP0009P0002<0000
ADe6:`Q00@8l0000<000004@`=/00000CP0000000036]hXn0P000080000200000P00004@`=/00000
0000o`Q000<l0000<000004@`=/40000000005]_/T=Wi]M2BR2c@d:UdD9`Xk=38bo8@RS5/d?b?Km2
00<30aE0000@000010000040000T000090000000P3d0000000000000P3d0000000000080001O0000
>0000040000h0000000003P000000000000100@0000000000000000000000000000002D0000<0000
0@0002D0000<00001@00P5D0000/0000H`4005h0001Y0@00KP0000@0001>5/06IAJ>1WDF@PIi5_X5
9@0000`0000700209@0000`000000020900002@00000081100000000000008110000000000020000
:00000`000010000AP0005`0001@0000ADe6:`Q000<l0000<000004@`=/4000000000=Ph/T?5l/12
NkBc@`YG]48nE[E3PZK0@]Ph/T?5l/120041PAA0080@0000100000000?lX0000300000<0000T0000
90000000P3d0000000000000P3d0000000000080000W0000600000<00000000000000000000U0000
300000<0000C000030000040000U0000300000P0081F0000;00006@1001J0000J`40064000040000
B1H81WLFX`F[5PH6B1H81RD0000<00001`00P1<0000<00000@0002D0000<00000000P2@0000T0000
0020@@00000000000020@@00000000000P0004H000280000O00004E=AR/8@042?000030000014<3K
000004h000000000ZA=@?`80000200000P00008000014<3K0000003o0?l8@003;000020000014<3K
0P000000002h?[I3cDca@[Pn]T<YG=E1004001E0000@000010000040000T000090000000P3d00000
00000000P3d0000000000080001O0000>0000040000h0000000003P000000000000100d000000000
0?l0000000000000000002D0000<00000@0002D0000<00001@00P5L0000T0000J`4001T0001^0@00
N`00008000385X/7b1J[0BD0000<00001`00P2D0000<00000000P2@0000T00000020@@0000000000
0020@@00000000000P0002P0000<00000@0004H0002D0000R00004E=AR/8@042B00003`000014<3K
000004h100000000ZA=@?`80000200000P000080000200000020?`00P4014<3K00000?l00?l8@003
;000020000014<3K0P000000002IF@E3G0oR@YUI1D>5ZaU3004001E0000@000010000040000X0000
300000<0000T000090000000P3d0000000000000P3d0000000000080001O0000>00000<0000h0000
000003P0000000000181010000000000003o000000000000000002D0000<00000`0002L0000H0000
0@00000000000?l0000002D0000<00000@0001<0000<00000P0003/0000800006`000100001N2000
4@L003H0000@0000GPP00147001H0000=000000000000000ooooooooool60000GPPE1eX86@MF21T7
DPPI1dh85@M>2147=P000100001>20004@L005P0000d000000000000003oooooooooo`H0001>20d7
DPP91eH82@MJ20T7GPP=1eh84@Lm0000200001/0000@0000GPP00647000f0000400005h8001Q1`00
F00003@00000000000000?oooooooooo1P0005h8I@MJ26T7EPQY1e88J@M>26D7CPQQ1cH0000@0000
CPP00647001H0000=000000000000000ooooooooool60000CPQM1e88F@MF25T7FPQI1eh8G@MN2647
?@0000P0000K0000400005h8002a1`00=P000100001N2000/@L005P0000d000000000000003ooooo
ooooo`H0001N2;D7FPRi1eH8^@MB2;T7CPRe1dh8/@Lf0000400004h8002a1`00F00003@000000000
00000?oooooooooo1P0004h8[@MB2:T7EPRY1eX8Z@MN2:d7GPRa1cd0000800006`000100001N2000
0@P003H0000@0000GPP00048001H0000=000000000000000ooooooooool60000GPP525X82@QF20T8
DPP924h81@Q>2048=P000100001>20000@P005P0000d000000000000003oooooooooo`H0001>2?d7
DPSi1eH8n@MJ2?T7GPSm1eh80@Pm0000200001/0000@0000GPP00548000f0000400005h8001A2000
F00003@00000000000000?oooooooooo1P0005h8E@QJ25T8EPQI2588F@Q>25D8CPQA23H0000@0000
CPP00548001H0000=000000000000000ooooooooool60000CPQ=2588B@QF24T8FPQ925h8C@QN2548
?@0000P0000K0000400005h8002Q2000=P000100001N2000X@P005P0000d000000000000003ooooo
ooooo`H0001N2:D8FPRY25H8Z@QB2:T8CPRU24h8X@Pf0000400004h8002Q2000F00003@000000000
00000?oooooooooo1P0004h8W@QB29T8EPRI25X8V@QN29d8GPRQ23d0000800006`000100001N2000
l@P003H0000@0000GPP00?48001H0000=000000000000000ooooooooool60000GPSe25X8n@QF2?T8
DPSi24h8m@Q>2?48=P000100001>2000l@P005P0000d000000000000003oooooooooo`H0001>2>d8
DPSY25H8j@QJ2>T8GPS]25h8l@Pm0000200001/0000@0000GPP00449000f0000400005h800112@00
F00003@00000000000000?oooooooooo1P0005h8A@UJ24T9EPQ92E88B@U>24D9CPQ12CH0000@0000
CPP00449001H0000=000000000000000ooooooooool60000CPPm2E88>@UF23T9FPPi2Eh8?@UN2449
?@0000P0000K0000400005h8002A2@00=P000100001N2000T@T005P0000d000000000000003ooooo
ooooo`H0001N29H9FPRI2EH8V@UB29T9CPRF2Dh8T@Tf0000400004h8002A2@00F00003@000000000
00000?oooooooooo1P0004h8S@UB28T9EPR92EX8R@UN28d9GPRA2Cd000080000?00000P0000o0000
600008<0001_0000R00009`0000C000030000040000U000030000000080U0000300000L0080T0000
90000000P440000000000000P440000000000080000X0000300000<00016000000H00?@50015CDH[
8D0700`0000000009D00010000040000000002a0000T0000600008^<>3`000200000P:`lPCbSd:Y3
3BdA@PQ00`F41@00N0D0004@`=/200000`0006P500010000K00001h000000000@`400?h100000000
00000;8]0000@00084E=AP000@1X1@008`0000<000000000000000000000100000<00401003`0000
00000000000000000>84082Y0`0I000030000?ooo`0B000030000080000U000030000000080U0000
300000d0080U0000300000L0080H000030000000000F000030000000000Q0000200001h0000H0000
00000000001f0@0030800140000<0000200000`0000@000000000000000;0000400007H1000<0P00
600000`0000000004P0000`0000100006@0000`0003oool0600000`0000000005P0000`0000H0000
2P00010000000000000000T0000@0000@0400<01000H000030000000001B0000L04000400020o_oo
0000000000000000T040004000800P0@D`1i06d0HP1_06`000000000000000000000000000000000
00000000000000000000000000000000000000005`2PjQ8001<005c/4P2E:oQg^3WhMooooom/k180
9k;lMaP74`000000@0Oj=@00001605@0E`110580A@1L04<0K01Q07<0L`1U07<0G01304`0D`1904@0
G01k0300<00`0380@`150300<`0]0300<00`0300;@0`0300000`02d0@`0`0300<00]0300<00`0300
<00`0300<00`0300=00f07d0G01=06T0L`1S05<0M01Q07@0M@1c0000O@000:Eg000n>000003Pja80
_NNUM`800h1@mBLcK0fVMl9Hm7L90000h3H5<nS[4P1/T2<c`0KeM`40001TMP08000002D0000<0000
0@0005@0001D00007P00000000130@00oP40004000006192TS@B@QX0001P0@000@0004`000000000
00000000003ooooooooooe00001al000b@000580001`0@000P0001000007000000000000002l0P00
000000420R9C07T0L`1d06D0K@000000000000000000000000000000000000000000000000000000
0000000000000000001c0000O@000:Eg000n>000003Pja80_NNUM`800h1@mBLcK0fVMl9Hm7L90000
3>/B0000003U>?Qg000C0=PX<C<00000j>XB08P64`2Lja80UB_hMb0in7Oooooo[>/B042go7OX1Q<0
h2Pa<n0X<C?ZT2<ch2Pa<c00<0000300;@130300<00`02d0<00`0300<00`0300000C0300<00d03H0
O@1L04d0J@0006<00P000640M01e07<0001m0000YGL003hh000C00800000:34c4000053e9c<10000
0@0001c[4P1D0000i>/B09D[n7LH=_QgooooofAf00P000009@0000`000020000:00000`000010000
9@0000`0000700209@0000`000000020<00000`0000?00209@0000`0000=0020B`000400000`0000
1@000200000100000@00010000000000000007H1000<0P0000000000001f0@00308002P0000<0000
0P000280000<0000ooooo`h0000D000000000100000D00006T03@2`0000P0000ooooo`8000000000
00000020^d<0@0=400000>P3j0<Q000020000680000<00000@000240000800008@0000P0001c0000
30000000000L0000200002D0000<00000000P2D0000<00001`00P2D0000<00003P00P300000<0000
3`00P1T0000<0000oooo01P0000<000000000180000<00000P0001<0000<00000@0001@0000<0000
3@0001D0000<00000@0001H0000<0000000000d0000@000000000000000j0000300000X0000K0000
40000000000000008`0002000010emXl000000000000H>Xl002[@`00548T00009000043GfS`00000
0000001PjS`00:]3000D@P@0001c000030000000000=0000400005H1000U0000DP00070100030000
400000L00000000000000;`2000000001`828U<0N@1c07@0I@1]0000000000000000000000000000
0000000000000000000000000000000000000000000001801@00070in7L004`0N0M<00D00038O9`e
X>DB001=001Lia80UB_hMkPin7OoooooK>LB02Nbo7Mh1d`000000?oooolO401h0@000000003<iQ80
n>HB02`e500<iQ809c`0H?`i503hiQ80000000400014iQ80`c`0H000503hiQ80000001@d500P6`1P
000001@d50000000_QX0H?3V4P3<iQ80000009d0001/iQ80hCl0H?3V4P0000000@0003gZ4P3<iQ80
0000000000000000gch0H0l0000mjQ80`>HB0:CW4P1AC01P0M[h049B0634iQ80I7H02000000U0000
300000<000160000:00001`00017A4U30P00000000000000MP4000`2000000006@0000`0003oool0
4P0000`0000200009@0000`000000020DP00070100040000400000L00000000000000;`200000000
1`828U<0N@1c07@0I@1]000000000000000000000000000000000000000000000000000000000000
00000000000001809k;lMgP7C0000000oooooal@07P1000000000<cV4P3hiQ80;3DD00cV4P0W?01P
o3TD0?SV4P0000000@0004CV4P33?01P000D0?SV4P00000053@D020K0600000053@D0000002n6P1P
l>HB0<cV4P000000W@0000004P3Q?`1Pl>HB000000010000?NXB0<cV4P00000000000000003O?P1P
3`0002D0003P=PDcH>HB010000000?Eg0@0001T0/04@0901G`WeMgD58Mh000209@000100T07oo`00
400008SV4P3j7OAgM@DQgQ00T04U0000I7H02000000U0000300000@0000U0000300000L0080H0000
30000000000F000030000000000Q0000200001h0000H000000000000001f0@00308002<0000P0000
@=OJ?00000000000063Z?000Zd<001A2900002@00010emXl000000000000H>Xl002[@`0054840000
8`0002000010emXl000000000000H>Xl002[@`00548T00009000043GfS`000000000001PjS`00:]3
000D@P@0000S0000800003cK7d4000000000043NKd400:]3000D@R@0000T0000?=/O@@0000000000
@=i_@@00Zd<001A2100001P0000<000000000180000<00000@0001T0000<0000oooo01P0000<0000
000001H0000<0000600002<0000P0000?=/O@@0000000000@=i_@@00Zd<001A2900002@0000lfam1
000000000010gVm1002[@`00548400008`000200000]aOll0000000000194@Tm002[@`00548T0000
900002g5oc`00000000004TA2Cd00:]3000D@P@0000H000030000000001B0000L04000D00020o_oo
0000000000000000T040004000870P0@D`1i06d0HP1_06`000000000000000000000000000000000
00000000000000000000000000000000000000005`2PjQ8001<005c/4P2E:oQg^3WhMooooom/k180
9k;lMaP74`000000@0Oj=@00001605@0E`110580A@1L04<0K01Q07<0L`1U07<0G01304`0D`1904@0
G01k0300<00`0380@`150300<`0]0300<00`0300;@0`0300000`02d0@`0`0300<00]0300<00`0300
<00`0300<00`0300=00f07d0G01=06T0L`1S05<0M01Q07@0M@1c0000O@000:Eg000n>000003Pja80
_NNUM`800h1@mBLcK0fVMl9Hm7L90000h3H5<nS[4P1/T2<c`0KeM`40001TMP08000002D0000<0000
1@0005@0001D0000E`4002D0001N0@00<`00004000006192TS@B@QX0001P0@000@0004`000000000
00000000003ooooooooooe00001al040b@000580001`0@001P0001000007000000000000002l0P00
000000L20R9C07T0L`1d06D0K@000000000000000000000000000000000000000000000000000000
0000000000000000001c0000O@000:Eg000n>000003Pja80_NNUM`800h1@mBLcK0fVMl9Hm7L90000
3>/B0000003U>?Qg000C0=PX<C<00000j>XB08P64`2Lja80UB_hMb0in7Oooooo[>/B042go7OX1Q<0
h2Pa<n0X<C?ZT2<ch2Pa<c00<0000300;@130300<00`02d0<00`0300<00`0300000C0300<00d03H0
O@1L04d0J@0006<00P000640M01e07<0001m0000YGL003hh000C00800000:34c4000053e9c<10000
0@0001c[4P1D0000i>/B09D[n7LH=_QgooooofAf00P000009@0000`000060000:00000`000050000
9@0000`0000700209@0000`0000000209@0000`000040000B`000400000`00001@00020000010000
0@000100001F0@009@000601000d0000EP4002D0001P0@00=00002P0000<00001P000280000<0000
ooooodH0000D0000200004M4BD<300008P0000`0003ooooo8P0000`0003ooooo9@0000`0000=0020
:00000`000030000:00000`0000400008P0000`0003oooooAP0000@6003h1@00ADe6:bI0000@0000
10000000000U@000400000@000000000;40002@0000H0000[8`h?00008000020J=VC?3Bnb4<Kf/E2
24041H@5001h1@000A30f`8000030000J0D00040001/00009P000000000?0@00_@40000000000000
lBL00010000PADe6000106P5000S00000`000000000000000000000400000`00@0400?0000000000
000000000000hP@0P:T301T0000<0000oooo0180000<00000P0002D0000<00000000P2D0000<0000
3@00P2D0000<00001`00P1P0000<0000000001H0000<000000000240000800007P0001P000000000
000004L1000<0P004@0000`0000800003000010000000000000000/0000@0000A`4000`2000H0000
30000000000B000030000040000I000030000?ooo`0H000030000000000F0000300001P0000:0000
40000000000000002@00010000100@00008001P0000<000000000580001`0@000@00083nool00000
00000000002@0@000@000P02011C07T0K@1R06l0K000000000000000000000000000000000000000
00000000000000000000000000000000000G0:3Z4P004`00G>`B09D[n7Nh>OQgooooofc/4P0W/_ag
60LC0000003P2Q0c000004H0E01G0440DP1505`0@`1/0640L`1c06D0L`1L04<0C01C04T0A01L07/0
<00`0300<P1304D0<00c02d0<00`0300<00]0300<0000300;@130300<00`02d0<00`0300<00`0300
<00`0300<00d03H0O@1L04d0J@1c06<0D`1d0640M01e07<0001m0000YGL003lh00000>3[4P2mijEg
0P03P>2[9c=/3JIg`USdM`T0003P=PDcj>/B04C98S?01_Eg0@0006Af00P000009@0000`000010000
E00005@0000V0000000000l1002m0@000@00001hod40g?m19@00060100010000C000000000000000
00000?ooooooooooD00006K`00390000DP00070100020000400000L00000000000000;`200000000
0@828U<0N@1c07@0I@1]000000000000000000000000000000000000000000000000000000000000
00000000000007<0001m0000YGL003lh00000>3[4P2mijEg0P03P>2[9c=/3JIg`USdM`T0000<ja80
00000>Dhn7L001<0>2la<`00003XjQ80R0HC09c[4P2E:oQg83WhMooooon/ja80@;OlMnP64`10;c4c
@2la<l;98S=0;c4c<00`0000<00]04<0<00`0300;@0`0300<00`0300<00001<0<00`03@0=P1m05`0
C@1Y0000H`020000H@1d07D0L`0007d0002UM`00?cP001<00P00000_<C<@0000h:/W<`4000010000
7>/B05@0003Tja80UB_hMaPfn7OoooooI7H02000000U000030000080000X000030000040000U0000
300000L0080U000030000000080`0000300000l0080U0000300000d0081;0000@000030000050000
80000040000100004000000000000000A`4000`200000000000004L1000<0P00:00000`000020000
8P0000`0003ooooo3P0001@000000000400001@0000J@0A0;0000200003ooooo0P00000000000000
002T@`100d@00000j0?X0b4000080000HP0000`0000100008@0000P0000Q0000200007<0000<0000
000001`0000800009@0000`0000000209@0000`0000700209@0000`0000>0020<00000`0000?0020
6@0000`0003oool0600000`0000000004P0000`0000200004`0000`000010000500000`0000=0000
5@0000`0000100005P0000`0000000003@00010000000000000003X0000<00002P0001/0000@0000
00000000000S000080000>=MnS`000000000003@13d00<U30036@R@0000T0000hegj?00000000000
0=04?@00bD<00<I2100007<0000<0000000000d0000@0000TP4006<0001B0000L04000@0000@0000
1`00000000000000_080000000070P8RD`1i07<0M01U06d000000000000000000000000000000000
00000000000000000000000000000000000000008Mh00000000000330S>0j180n<82<d:6mWMe1B7N
07/Z<`<00000``8c000C00@00000?P1P3`000<9Hm7L100008@2:0@002P0hcBlcF_3dM`d0000I0;01
8@2:0B400022LnQg000C0000000XYR`ce0D:=RRV;3?0EoMg000QgQ00T04U0000/CWdM`SV4P1`2R/c
M>LB05HHjWN`MNQgooooo`400017A4U3200009SV4P2/iQ80Fi_fM`00001Ta@8c000Q05c50S<PGOMg
00000=01YcH0000000000000002/iQ800@0007D58Mj/iQ80PD7dMm01YcITMP08000002D0000<0000
100004H0000X0000700004M4BD<200000000000000170@0030800000000I000030000?ooo`0B0000
30000080000U000030000000081B0000L04000<0000@00001`00000000000000_080000000070P8R
D`1i07<0M01U06d00000000000000000000000000000000000000000000000000000000000000000
00000000060?0000`USdM`40000Q08X1000:03S=;c=Jl?Ag3@0001T0/04Q08X18@00089cj7L001<0
000002RV;3?D1@Xf::H/<l1GmgL0027N402@0BD0002a>OAg2>HB070::c=dia80EQSZMk1ej7Oooooo
0@0004M4BD<80000000B0:cV4P1KVoIg000006C50S<00240G<D2<b1MmgL00000d06W=P0000000000
9@000>0f1C=PiQ8040000000mGL100006@2`0A00T05O2OEgM@DQgP00080U0000402@0Ooo000@0000
R>HB0?XMm7Me1B7N402@0BD0001TMP08000002D0000<00000`0002D0000<00001`00P1P0000<0000
000001H0000<000000000240000800007P0001P000000000000004L1000<0P008`000200003SGOXl
000000000000d0@m0039@`00aT8T000090000>=MnS`000000000003@13d00<U30036@P@0000S0000
80000>=MnS`000000000003@13d00<U30036@R@0000T0000hegj?000000000000=04?@00bD<00<I2
100002<0000P0000m^HO@@0000000000h>b7@@00bD<00<I2900002@0003fiQm100000000003Pk8M1
0039@`00aT840000600000`0000000004P0000`0000100006@0000`0003oool0600000`000000000
5P0000`0000H00008`000200003fiQm100000000003Pk8M10039@`00aT8T000090000?KV7d400000
00000>3/Qd400<U30036@P@0000S000080000?7Goc`0000000000>3/1cd00<U30036@R@0000T0000
lMOo?00000000000h>`7?@00bD<00<I2100001P0000<000000000580001`0@001P00083nool00000
00000000002@0@000@000PL2011C07T0K@1R06l0K000000000000000000000000000000000000000
00000000000000000000000000000000000G0:3Z4P004`00G>`B09D[n7Nh>OQgooooofc/4P0W/_ag
60LC0000003P2Q0c000004H0E01G0440DP1505`0@`1/0640L`1c06D0L`1L04<0C01C04T0A01L07/0
<00`0300<P1304D0<00c02d0<00`0300<00]0300<0000300;@130300<00`02d0<00`0300<00`0300
<00`0300<00d03H0O@1L04d0J@1c06<0D`1d0640M01e07<0001m0000YGL003lh00000>3[4P2mijEg
0P03P>2[9c=/3JIg`USdM`T0003P=PDcj>/B04C98S?01_Eg0@0006Af00P000009@0000`000060000
E00005@0002C0@00H`0009T1001a00000@00001hod40g?m19@00060100010000C000000000000000
00000?ooooooooooD00006K`00390000DP00070100050000400000L00000000000000;`200000000
1`828U<0N@1c07@0I@1]000000000000000000000000000000000000000000000000000000000000
00000000000007<0001m0000YGL003lh00000>3[4P2mijEg0P03P>2[9c=/3JIg`USdM`T0000<ja80
00000>Dhn7L001<0>2la<`00003XjQ80R0HC09c[4P2E:oQg83WhMooooon/ja80@;OlMnP64`10;c4c
@2la<l;98S=0;c4c<00`0000<00]04<0<00`0300;@0`0300<00`0300<00001<0<00`03@0=P1m05`0
C@1Y0000H`020000H@1d07D0L`0007d0002UM`00?cP001<00P00000_<C<@0000h:/W<`4000010000
7>/B05@0003Tja80UB_hMaPfn7OoooooI7H02000000U0000300000D0000X0000300000H0000U0000
300000L0080U000030000000080U0000300000<0001;0000@0000300000500008000004000010000
40000981001S0000W04007@0002B0@00H`0009`1001d0000:00000`0000500008P0000`0003ooooo
AP0001@000080000AdA9@`<0000R000030000?oooolR000030000?oooolU0000300000d0080X0000
300000@0000X0000300000<0000R000030000?oooom6000010H00?P50015CDH[9T00010000040000
000002E0000@000010000000000/@000900001P0002n7g0l0000P000082f7g0li;2Y@gZD`d88@0D5
Q0D007P500014<3K0P0000<0001X1@000@0006`0000J000000000:@1002[0@00000000000000@000
04000215CDH00040J0D002<000030000000000000000000000@0000300100@00l000000000000000
0000003R1020Z@<06@0000`0003oool04P0000`0000200009@0000`0000000209@0000`0000=0020
9@0000`000070020600000`0000000005P0000`0000000008@0000P0000N00006000000000000000
308000`2000A0000300000P0000<000040000000000000002`000100000<0P00308001P0000<0000
00000180000<00000@0001T0000<0000oooo01P0000<0000000001H0000<0000600000X0000@0000
000000000009000040000:01002P0@00600000`000000000DP00070100010000P?koo`0000000000
0000090100010002008045<0N@1]0680K`1/00000000000000000000000000000000000000000000
000000000000000000000000000001L0X>XB000C001Lk180UB_hMkPin7OoooooK>`B02Nbo7LH1a<0
00000407OcD00000AP1D05L0@@1B04D0G01306`0H@1c07<0I@1c05`0@`1<05<0B@1405`0N`0`0300
<00b04<0A@0`03<0;@0`0300<00`02d0<00`0000<00]04<0<00`0300;@0`0300<00`0300<00`0300
<00`03@0=P1m05`0C@1Y07<0H`1C07@0H@1d07D0L`0007d0002UM`00@3P00000h>/B0;gWYGL200>0
h:@W<f`=YWO2F?Ag2@000>0f1C?Xja80E0dX<l06mGL10000I7H02000000U000030000040001D0000
E00001X000000000Y0400:/100010000/W<M@[9c7D8E000000400040001<00000000000000000000
oooooooooom@0000NO0000H1001B0000L0400080000@00001`00000000000000_080000000010P8R
D`1i07<0M01U06d00000000000000000000000000000000000000000000000000000000000000000
00000000L`0007d0002UM`00@3P00000h>/B0;gWYGL200>0h:@W<f`=YWO2F?Ag2@0000c[4P000000
iCShM`004`3Pfalc00000>SZ4P281Q<0W>/B09D[n7LP>OQgooooojc[4P10]oagj0HC0>SK7c?Xfalc
dPdX<nSK7c<`0300000`02d0@`0`0300<00]0300<00`0300<00`00004`0`0300=00f07d0G01=06T0
001S0080001Q07@0M@1c0000O@000:Eg0010>0004`0200000=/O<a00003PY2Lc0@000040000Lja80
E0000>C[4P2E:oQg63KhMooooomTMP08000002D0000<00000P0002P0000<00000@0002D0000<0000
1`00P2D0000<00000000P300000<00003`00P2D0000<00003@00P4/000100000<00000D0000P0000
0@000040000@000000000000000<0P003080000000000000308000`2000X000030000080000R0000
30000?ooool>000050000000000@0000500001Y01D0/000080000?ooool20000000000000000@0=4
0403A000003X0nP38@0000P0001R000030000040000Q00002000024000080000L`0000`000000000
700000P0000U000030000000080U0000300000L0080U0000300000h0080`0000300000l0080I0000
30000?ooo`0H000030000000000B000030000080000C000030000040000D0000300000d0000E0000
30000040000F000030000000000=00004000000000000000>P0000`0000:00006`00010000000000
000002<0000P00000<3J?000000000000<3J?000ZT<00<A2900002@00000`=Xl000000000000`=Xl
002Z@`00a4840000L`0000`0000000003@000100001D0@00HP000580001`0@000`00010000070000
00000000002l0P00000000L20R9C07T0L`1d06D0K@00000000000000000000000000000000000000
00000000000000000000000000000000000QgP0000000000J9DS<h3X4P1PUB<c@XKfMgD58MjP8bXc
0`0006RE8c<001<01000000000000000`USdM`40000Q08X1000:03S=;c=Jl?Ag3@0001T0/04Q08X1
8@00089cj7L001<0000000Q5601l1@ZQ24DH0<1GmgL0027N402@0BD0002a>OAg2>HB0000001dia80
EQSZMk1ej7Oooooo0@0004M4BD<80000V>HB0:cV4P1KVoIg00000<bG8c<00240a9LS<b1MmgL00000
d06W=P000000000000000:cV4P010000M@DQgZcV4P21@OAgd06W=VAf00P000009@0000`000030000
AP0002P0000L0000AdA9@`8000000000000000`2000<0P00000001T0000<0000oooo0180000<0000
0P0002D0000<00000000P580001`0@00100001000007000000000000002l0P00000000L20R9C07T0
L`1d06D0K@0000000000000000000000000000000000000000000000000000000000000000000000
000000000032F?Ag0@000240RP4000X0><d_<e[`m7L=00006@2`0B40RP4Q0000PW?XM`004`000000
24DH07`52Z48AAP0`5OgM`008Mh@09019@000;4im7L8iQ80000007CW4P1F6>Yg/7GXMoooool10000
AdA9@`P000000180[>HB05^KmWL00000c9LS<`008@34Ub<c85ggM`00003@0JLf00000000000U0000
h3H5<f3V4P0@0000003eM`40000I0;01402@0El9mGMe1B7N0000P2D0000@0901ool001000028iQ80
nQgdMgD58Mh@09019@0006Af00P000009@0000`0000400009@0000`000070020600000`000000000
5P0000`0000000008@0000P0000N00006000000000000000308000`2000S000080000030fS`00000
00000030fS`00:Y30034@R@0000T00000<3J?000000000000<3J?000ZT<00<A2100002<0000P0000
0<3J?000000000000<3J?000ZT<00<A2900002@00000`=Xl000000000000`=Xl002Z@`00a4840000
8`0002000020h5m1000000000020h5m1002Z@`00a48T00009000083PGd4000000000083PGd400:Y3
0034@P@0000H000030000000000B000030000040000I000030000?ooo`0H000030000000000F0000
300001P0000S00008000083PGd4000000000083PGd400:Y30034@R@0000T0000P>1O@@0000000000
P>1O@@00ZT<00<A2100002<0000P0000>lD9?@0000000000>lD9?@00ZT<00<A2900002@0000ka@Tm
00000000000ka@Tm002Z@`00a4840000600000`000000000DP00070100050000P?koo`0000000000
00000901000100021`8045<0N@1]0680K`1/00000000000000000000000000000000000000000000
000000000000000000000000000001L0X>XB000C001Lk180UB_hMkPin7OoooooK>`B02Nbo7LH1a<0
00000407OcD00000AP1D05L0@@1B04D0G01306`0H@1c07<0I@1c05`0@`1<05<0B@1405`0N`0`0300
<00b04<0A@0`03<0;@0`0300<00`02d0<00`0000<00]04<0<00`0300;@0`0300<00`0300<00`0300
<00`03@0=P1m05`0C@1Y07<0H`1C07@0H@1d07D0L`0007d0002UM`00@3P00000h>/B0;gWYGL200>0
h:@W<f`=YWO2F?Ag2@000>0f1C?Xja80E0dX<l06mGL10000I7H02000000U0000300000D0001D0000
E00005D1001R0000GP4006d000010000/W<M@[9c7D8E000000400040001<00000000000000000000
oooooooooom@0000NO0V1PH1001B0000L04000H0000@00001`00000000000000_080000000070P8R
D`1i07<0M01U06d00000000000000000000000000000000000000000000000000000000000000000
00000000L`0007d0002UM`00@3P00000h>/B0;gWYGL200>0h:@W<f`=YWO2F?Ag2@0000c[4P000000
iCShM`004`3Pfalc00000>SZ4P281Q<0W>/B09D[n7LP>OQgooooojc[4P10]oagj0HC0>SK7c?Xfalc
dPdX<nSK7c<`0300000`02d0@`0`0300<00]0300<00`0300<00`00004`0`0300=00f07d0G01=06T0
001S0080001Q07@0M@1c0000O@000:Eg0010>0004`0200000=/O<a00003PY2Lc0@000040000Lja80
E0000>C[4P2E:oQg63KhMooooomTMP08000002D0000<00001P0002P0000<00001@0002D0000<0000
1`00P2D0000<00000000P2D0000<0000100004/000100000<00000D0000P00000@000040000@0000
E0400680001R0@00L00005@1001R0000HP400700000X0000300000H0000R000030000?oooom60000
500000P00017A4U30`000280000<0000ooooob80000<0000ooooobD0000<00003@00P2P0000<0000
0`0002P0000<000010000280000<0000ooooodH0002X0000W00004E=AR/V@000400000@000000000
24010S`0000`00000A30f`00001>000000000<JgRSh200000P000080000200000A30f`000000003o
24000c`0000`00000A30f`@000000000^Sjf@lK<<D897[53W0/n@Whd[4?H_E52<n2W@baZJd800`<3
5D000100000400000@0002@0000T00000020?@00000000000020?@00000000000P0005l0000h0000
100003P000000000>000000000000040100000000000000000000000000000009@0000`000040000
9@0000`000050020E@0002`0001>0@00:`0006h1000l000010000<PFb08T5_T2QaE70o`D[P<U0000
300000L0080U000030000000080T000090000000P440000000000000P440000000000080000X0000
300000@000160000G00005000015CDH[24000c`0000`00000A30f`@000000000KBFY@ii5Ld8Y_:E3
QN]h@XPgYd<@/em2KBFY@ii5Ld800@615400P100000400000000obP0000<00000@0002@0000T0000
0020?@00000000000020?@00000000000P0002L0000H00000@00000000000000000002D0000<0000
0@0001<0000<00000@0002D0000<00002000P5H0000/0000B`4003L0001C0@00?`0000@0000U5Lh3
^1CT0nLDO`<U5Lh39@0000`0000700204`0000`0000100009@0000`000000020900002@000000811
00000000000008110000000000020000AP0009P0002<0000ADe6:`Q00@8l0000<000004@`=/00000
CP0000000036]hXn0P000080000200000P00004@`=/000000000o`Q000<l0000<000004@`=/40000
000004K1aT?AC?52m5o7@dJ8id;SR/M3[fCM@_@oad=@H]=200<30aE0000@000010000040000T0000
90000000P3d0000000000000P3d0000000000080001O0000>00000@0000h0000000003P000000000
000100@0000000000000000000000000000002D0000<0000100002D0000<00001@00P5D0000/0000
S04006P0002A0@00NP0000@0003I68/7k1Pm1o8Hk0KX69`69@0000`0000700209@0000`000000020
900002@00000081100000000000008110000000000020000:00000`000040000AP0005`0001@0000
ADe6:`Q000<l0000<000004@`=/4000000000=g9aD<63=I2Al76@oJXb48nf/Q3k/SC@]g9aD<63=I2
0041PAA0080@0000100000000?lT000090000000P3d0000000000000P3d0000000000080000U0000
30000040000C000030000040000U0000300000P0081F0000;00008/1001T0000TP4006`000040000
^QRa1]THAPHL6Il6^QRa1RD0000<00001`00P1<0000<00000@0002D0000<00000000P2@0000T0000
0020@@00000000000020@@00000000000P0004H0002D0000R00004E=AR/8@042B00003`000014<3K
000004h100000000ZA=@?`80000200000P000080000200000020?`00P4014<3K00000?l00?l8@003
;000020000014<3K0P000000002IF@E3QJ/I@g^D`d9LCd93004001E0000@000010000040000X0000
30000040000T000090000000P3d0000000000000P3d0000000000080001O0000>0000040000h0000
000003P0000000000181010000000000003o000000000000000002D0000<00000@0002L0000H0000
1000000000000?l0000002D0000<0000100001<0000<00000P0003/0000800006`000100001L2000
X@T003H0000@0000G0P00:49001H0000=000000000000000ooooooooool60000F@RT2E@8Y0UA2:49
C@RN2Dd8V@U@29H9=P000100001@2000UPT005P0000d000000000000003oooooooooo`H0001C29<9
F0RB2E/8U@UO29P9G`RM2E`8X@Tm0000200001/0000@00009`P00=d9000f0000400002L8003M2@00
F00003@00000000000000?oooooooooo1P0002@8h0TO2>0970SM2AT8fPTH2=D96`SB2CH0000@0000
6`P00=89001H0000=000000000000000ooooooooool600007PS?2B<8c`TW2=49:PSD2BX8f@TW2=d9
?@0000P0000K000040000?<7000I2P00=P000100003c1`006@X005P0000d000000000000003ooooo
ooooo`H0003`1a`:j`LM2^L76P[T1aL:i0LB2^L73PXf000040000>L7000>2P00F00003@000000000
00000?oooooooooo1P000>X72`[_1`/:lPL>2_D74@[f1aH:l`LI2Sd0000800006`000100002n1`00
E@X003H0000@0000_PL005D:001H0000=000000000000000ooooooooool60000^`MH2[H7F@Zc1eH:
[`MC2Zl7CPZb1d/:=P000100002b1`00B`X005P0000d000000000000003oooooooooo`H0002e1dL:
^PM72[d7BP[11dd:`@MB2[h7E@Xm0000200001/0000@0000R@L0094:000f0000400008T7002A2P00
F00003@00000000000000?oooooooooo1P0008H7U@Z11iD:OPNB2W/7S`Yj1hX:O@N72SH0000@0000
O@L008L:001H0000=000000000000000ooooooooool60000P0N32XD7P`Z91hH:S0N92X`7SPZ91i4:
?@0000P0000K0000400005D7003>2P00=P000100001E1`00cPX005P0000d000000000000003ooooo
ooooo`H0001B1m4:C@OA2TT7cPY61l/:APO62TT7``Xf0000400004T700332P00F00003@000000000
00000?oooooooooo1P0004`7`0YA1kl:E0O22UL7a@YH1lX:E@O>2Sd0000800006`000100000P1`00
2P/003H0000@000080L000X;001H0000=000000000000000ooooooooool600007@L=2aP73@/E1`X;
4@L82a470`/D1ol:=P000100000D1`00o`X005P0000d000000000000003oooooooooo`H0000G1o`:
70Ol2Ql7o`XS1`4;8`L72b072P/m0000200001/0000@0000j`H004H;000f000040000>/600162`00
F00003@00000000000000?oooooooooo1P000>P6B@_S1TX;h0I72md6A0_L1Sl;g`Hk2cH0000@0000
g`H003/;001H0000=000000000000000ooooooooool60000hPHh2nL6>0_[1S/;kPHn2nh6@`_[1TH;
?@0000P0000K000040000;L600222`00=P000100002g1P00PP/005P0000d000000000000003ooooo
ooooo`H0002d1XH;[`J62j/6P`^X1X0;Z0Ik2j/6N0/f000040000:/6001h2`00F00003@000000000
00000?oooooooooo1P000:d6M0^c1W@;]PIg2kT6NP^i1Wl;]`J22cd0000800006`00010000221P00
_P/003H0000@0000PPH00;h;001H0000=000000000000000ooooooooool60000O`K22gX6`P]g1[l;
L`Jl2g<6]`]f1[@;=P000100001f1P00]0/005P0000d000000000000003oooooooooo`H0001i1[4;
OPJ`2h46/`^41[H;Q@Jk2h86_P/m0000200001/0000@0000C@H00?/;000f0000400004d6003k2`00
F00003@00000000000000?oooooooooo1P0004X6oP]51_h;@PKk2cl6n0/n1_<;@@K`2cH0000@0000
@@H00?0;001H0000=000000000000000ooooooooool60000A0K]2dT6k0]<1^l;D0Kb2e06m`]=1_/;
?@0000P0000l0000200003l0000H0000HP0009L000280000`P0001<0000<00000@0002D0000<0000
0000P2D0000<00001`00P2@0000T00000020@@00000000000020@@00000000000P0002P0000<0000
0@0004H000280000O00004E=AR/8@042?000030000014<3K000004h000000000a[N:?P8000020000
0P00008000014<3K000000000?l8@003;000020000014<3K0P000000001L3n92UhTX@e`?hT:`E`e3
004001E0000@000010000040000T000090000000P3d0000000000000P3d0000000000080001O0000
>0000040000h0000000003P000000000000100@0000000000000000000000000000002D0000<0000
0@0002D0000<00001@00P5L0000T0000K`0008`0001c0000ZP000080000A1hT:4@OF22D0000<0000
1`00P2D0000<00000000P2@0000T00000020@@00000000000020@@00000000000P0002P0000<0000
0@0004H0001L0000D00004E=AR/8@003?000030000014<3K10000000002MB^Q2;l8W@e`?hT9`oBe3
6mCK@Ro29d>MB^Q2;l8W@`010H4D@020400000@00000003o:00000`000040000900002@00000080m
000000000000080m00000000000200009`0001P00004000000000000000000009@0000`000040000
4`0000`0000100009@0000`000080020EP0002`0001]0000Y`0007D0002^0000100004<7O@XA1n0:
g`Im2T<7O@XU0000300000L0080C000030000040000U000030000000080T000090000000P4400000
00000000P44000000000008000160000G00005000015CDH[24000c`0000`00000A30f`@000000000
6mCK@QPO3T=L3n92en<7@ie:j48H7`i36mCK@QPO3T<00@615400P100000400000000ob@0000T0000
0020?@00000000000020?@00000000000P0002D0000<0000100001<0000<00000@0002D0000<0000
2000P5H0000/0000K@0008L0001e0000S`0000@0003O1^884@Mo24<7hPSO1^889@0000`000070020
4`0000`0000100009@0000`000000020900002@00000081100000000000008110000000000020000
AP000;P0002/0000ADe6:bY0000T0000600005b?XT800000000005b?XT:4S;Y2QfTG@`Q00PI40000
>000004@`=^[ZRXn00000080000000003`0005@0B@1=04D0D`0P04h0A@1G0200DP1?04d0@@1>0000
=T02P400000d00000000o`40000100000@000580SZ?@?AM/@Ch0080o000000000000080o00000000
00000180000<00000@0001P0000<0000000001H0000<0000600002P0000<00000P000580001`0@00
0P000?;oool0000000000000002@0@000@0000L0101D06T0K@1U07<0801>06D0M`0P0580K`1]0640
KP000<L5C@1_06h0K`1d07T0L01U03X0E01Y06d0I@1c0200003ooooo00000000JbD0H0cJn00`^oT0
;3DD03HM060`^oT0;3DD00PN060/=A@000000000000F@`1P;3DD032kn@3l>A@0`61>H?S]4P0001<0
10000000000100000000007^4P3lkQ80O>dB00000014kQ80UB_hMaPfn7L00?ooE>hB089cj7L001<0
000003RP;3?V1@Xa>:0/<l1GmgL0000053@D0000002n6P1P8>hB0000000dn180003ZMk1ej7Oooooo
0@0004mmm7LhX2`c40000>0f1C>@kQ80>:0/<l@7mGO0EoMgM@DQgP80003V1@XaiPD:<FAf00P00000
9@0000`000020000E00005@0001V0000VP0006h0002Y00000@000000nT400?Y1IP000:L000010000
C00000000000000000000?ooooooooooD0000580ool000009@0000`0000=0020AP0001`0000@0000
ADe6:b]0000<0000000004`0001T00000P000080003c0@00eP00008000020000lP400=D0000Y0:X0
000000000000080o000000000000080o0000000000000000000000000000000000000000000R0000
30000?oooom60000700001000015CDH[0T0000`0000000003P0001@000000000400001@0
\>"], "Graphics",
  ImageSize->{499, 214},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  CellTags->{"b:16.3", "Rolling Disk"}],

Cell[CellGroupData[{

Cell["\<\
r={6};
H=IdentityMatrix[6]\
\>", "Input",
  CellLabel->"In[314]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[BoxData[
    \({{1, 0, 0, 0, 0, 0}, {0, 1, 0, 0, 0, 0}, {0, 0, 1, 0, 0, 0}, {0, 0, 0, 
        1, 0, 0}, {0, 0, 0, 0, 1, 0}, {0, 0, 0, 0, 0, 1}}\)], "Output",
  CellLabel->"Out[315]=",
  CellTags->"b:16.3"]
}, Open  ]],

Cell["\<\
q={psi,theta,phi,x,y,z};
p={wx,wy,wz,vx,vy,vz};
JointList={{r,H,q,p}};
\
\>", "Input",
  CellLabel->"In[316]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell["\<\
In this setup the x,y,z coordinates locate the center of the disk in the \
space frame and psi, theta, phi are Euler angles in the z-y-x (or 3-2-1) \
convention. 
The body data is:\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell["\<\
(* reassign som variables *)
g=\"g\"; m=\"m\"; J=\"J\"; Iy=\"Iy\"; R=\"R\";\
\>", "Input",
  CellLabel->"In[319]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell["\<\
Mass=m; CenterOfMass={0,0,0}; OutboardNode={2,{0,-R*Sin[psi],-R*Cos[psi]}};
InertiaMatrix=DiagonalMatrix[{J,Iy,Iy}];
BodyList={{CenterOfMass,{OutboardNode},Mass,InertiaMatrix}};\
\>", "Input",
  CellLabel->"In[321]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell["and the remaining data:", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell["\<\
TreeList={{{1,1}}};
PE=0; Q={0,0,0,0,0,0};\
\>", "Input",
  CellLabel->"In[324]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell["The unconstrained disk model is obtained with", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{

Cell["\<\
{V,X,H,M,Cp,Fp,p,q}=CreateModel[JointList,BodyList,TreeList,g,PE,Q];\
\>", "Input",
  CellLabel->"In[326]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[BoxData[
    \("Computing Joint Kinematics"\)], "Print",
  CellLabel->"From In[326]:=",
  CellTags->"b:16.3"],

Cell[BoxData[
    \("Computing joint 1 kinematics"\)], "Print",
  CellLabel->"From In[326]:=",
  CellTags->"b:16.3"],

Cell[BoxData[
    \("Computing Potential Functions"\)], "Print",
  CellLabel->"From In[326]:=",
  CellTags->"b:16.3"],

Cell[BoxData[
    \("Computing Inertia Matrix"\)], "Print",
  CellLabel->"From In[326]:=",
  CellTags->"b:16.3"],

Cell[BoxData[
    \("Computing Poincare Function"\)], "Print",
  CellLabel->"From In[326]:=",
  CellTags->"b:16.3"]
}, Open  ]],

Cell["\<\
Now, we formulate the differential constraints. Rolling without slipping \
implies that the velocity of the disk contact point must be zero. An \
expression for the (angular and translational) velocity at the outboard node \
as a function of the configuration variables is easily obtained using the \
function EndEffectorVelocity or NodeVelocity.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{

Cell["\<\
{ChainList}=TreeList;
VCont=NodeVelocity[ChainList,BodyList,X,H,p]\
\>", "Input",
  CellLabel->"In[327]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[BoxData[
    \({wx, wy, wz, vx - "R"\ wy\ Cos[psi] + "R"\ wz\ Sin[psi], 
      vy + "R"\ wx\ Cos[psi], vz - "R"\ wx\ Sin[psi]}\)], "Output",
  CellLabel->"Out[328]=",
  CellTags->"b:16.3"]
}, Open  ]],

Cell["\<\
The rolling constraint requires that the translation velocity of the contact \
point (the last three of the elements of the six\[CapitalEth]vector VCont) be \
zero. Thus, we obtain the constrained system:\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{

Cell["\<\
{Mm,Cm,Fm,Vm,T,phat}=DifferentialConstraints[M,Cp,Fp,V,VCont[[Range[4,6]]],p,\
q,{4,5,6}];\
\>", "Input",
  CellLabel->"In[329]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[BoxData[
    \("Vm"\)], "Print",
  CellLabel->"From In[329]:=",
  CellTags->"b:16.3"],

Cell[BoxData[
    \("Mm"\)], "Print",
  CellLabel->"From In[329]:=",
  CellTags->"b:16.3"],

Cell[BoxData[
    \("Cm"\)], "Print",
  CellLabel->"From In[329]:=",
  CellTags->"b:16.3"],

Cell[BoxData[
    \("Fm"\)], "Print",
  CellLabel->"From In[329]:=",
  CellTags->"b:16.3"]
}, Open  ]],

Cell["\<\
The dynamics are reduced to three dimensions and the original six quasi\
\[CapitalEth]velocities are reduced to three, in fact, we have:\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{

Cell["phat", "Input",
  CellLabel->"In[330]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[BoxData[
    \({wx, wy, wz}\)], "Output",
  CellLabel->"Out[330]=",
  CellTags->"b:16.3"]
}, Open  ]],

Cell["\<\
The set of configuration coordinates is not reduced by the function \
DiffConstrainedSys. In general, a set of differential constraints may not \
admit any such reduction. Such would be the case if the constraints were \
completely nonholonomic. In the present case, however, the constraints are \
'partially' integrable and from basic geometry one can see that (cg) hight = \
Rcos(theta)-R. Using this relationship, the coordinate z can be eliminated \
from the equations.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{

Cell["{Sols}=Solve[{X[[1]][[3,4]]==R*(Cos[theta]-1)},{z}]", "Input",
  CellLabel->"In[331]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[BoxData[
    \({{z \[Rule] "R"\ \((\(-1\) + Cos[theta])\)}}\)], "Output",
  CellLabel->"Out[331]=",
  CellTags->"b:16.3"]
}, Open  ]],

Cell["\<\
Because the translation parameters x,y,z are, in fact, the space coordinates, \
z is precisely the hight of the center of gravity. Inspection shows that only \
Mm and Vm depend on z so we define:\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell["Mmo=Simplify[Mm/.Sols];", "Input",
  CellLabel->"In[332]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell["Vmo=Simplify[Vm[[ {1,2,3,4,5} ]]/.Sols];", "Input",
  CellLabel->"In[333]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell["Now, we assemble the governing equations.", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{

Cell["Eqns=MakeODEs[phat,q[[{1,2,3,4,5}]],Vmo,Mmo,Cm,Fm,t]", "Input",
  CellLabel->"In[334]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          RowBox[{
            SuperscriptBox["psi", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], 
          "==", \(wx[t] + Sin[psi[t]]\ Tan[theta[t]]\ wy[t] + 
            Cos[psi[t]]\ Tan[theta[t]]\ wz[t]\)}], ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["theta", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], 
          "==", \(Cos[psi[t]]\ wy[t] - Sin[psi[t]]\ wz[t]\)}], ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["phi", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], 
          "==", \(Sec[theta[t]]\ Sin[psi[t]]\ wy[t] + 
            Cos[psi[t]]\ Sec[theta[t]]\ wz[t]\)}], ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["x", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], 
          "==", \("R"\ Sin[phi[t]]\ wx[t] + 
            "R"\ Cos[phi[t]]\ Cos[psi[t]]\ Cos[theta[t]]\ wy[t] - 
            "R"\ Cos[phi[t]]\ Cos[theta[t]]\ Sin[psi[t]]\ wz[t]\)}], ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["y", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], 
          "==", \(\(-"R"\)\ Cos[phi[t]]\ wx[t] + 
            "R"\ Cos[psi[t]]\ Cos[theta[t]]\ Sin[phi[t]]\ wy[t] - 
            "R"\ Cos[theta[t]]\ Sin[phi[t]]\ Sin[psi[t]]\ wz[t]\)}], ",", 
        RowBox[{
          
          RowBox[{\(wy[
                t]\ \((\(-"Iy"\)\ wz[t] + 
                  "R"\ Cos[
                      psi[t]]\ \((\(-"m"\)\ "R"\ Sin[psi[t]]\ wy[t] - 
                        "m"\ "R"\ Cos[psi[t]]\ wz[t])\))\)\), 
            "+", \(wz[
                t]\ \(("Iy"\ wy[t] - 
                  "R"\ Sin[
                      psi[t]]\ \((\(-"m"\)\ "R"\ Sin[psi[t]]\ wy[t] - 
                        "m"\ "R"\ Cos[psi[t]]\ wz[t])\))\)\), "+", 
            RowBox[{\(("J" + "m"\ "R"\^2)\), " ", 
              RowBox[{
                SuperscriptBox["wx", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}]}], "==", "0"}], 
        ",", 
        RowBox[{
          
          RowBox[{\(\(-"g"\)\ "m"\ "R"\ Cos[psi[t]]\ Sin[theta[t]]\), 
            "+", \(wz[
                t]\ \((\(-"Iy"\)\ wx[t] - 
                  "m"\ "R"\^2\ Sin[psi[t]]\^2\ wx[t] + 
                  "m"\ "R"\ Cos[psi[t]]\^2\ Tan[
                      theta[t]]\ \((\(-"R"\)\ Sin[psi[t]]\ wy[t] - 
                        "R"\ Cos[psi[t]]\ wz[t])\))\)\), 
            "+", \(wy[
                t]\ \(("m"\ "R"\^2\ Cos[psi[t]]\ Sin[psi[t]]\ wx[t] + 
                  "m"\ "R"\ Cos[psi[t]]\ Sin[psi[t]]\ Tan[
                      theta[t]]\ \((\(-"R"\)\ Sin[psi[t]]\ wy[t] - 
                        "R"\ Cos[psi[t]]\ wz[t])\))\)\), 
            "+", \(wx[
                t]\ \(("J"\ wz[t] + "m"\ "R"\^2\ Cos[psi[t]]\^2\ wz[t] + 
                  "m"\ "R"\ Cos[
                      psi[t]]\ \((\(-"R"\)\ Sin[psi[t]]\ wy[t] - 
                        "R"\ Cos[psi[t]]\ wz[t])\) + 
                  "R"\ Sin[
                      psi[t]]\ \(("m"\ "R"\ Cos[psi[t]]\ wy[t] - 
                        "m"\ \(("R"\ Cos[psi[t]]\ wy[t] - 
                              "R"\ Sin[psi[t]]\ wz[t])\))\))\)\), "+", 
            RowBox[{\(("Iy" + "m"\ "R"\^2\ Cos[psi[t]]\^2)\), " ", 
              RowBox[{
                SuperscriptBox["wy", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}], "-", 
            
            RowBox[{"\<\"m\"\>", " ", \("R"\^2\), " ", \(Cos[psi[t]]\), 
              " ", \(Sin[psi[t]]\), " ", 
              RowBox[{
                SuperscriptBox["wz", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}]}], "==", "0"}], 
        ",", 
        RowBox[{
          
          RowBox[{\("g"\ "m"\ "R"\ Sin[psi[t]]\ Sin[theta[t]]\), 
            "+", \(wz[
                t]\ \((\(-"m"\)\ "R"\^2\ Cos[psi[t]]\ Sin[psi[t]]\ wx[t] - 
                  "m"\ "R"\ Cos[psi[t]]\ Sin[psi[t]]\ Tan[
                      theta[t]]\ \((\(-"R"\)\ Sin[psi[t]]\ wy[t] - 
                        "R"\ Cos[psi[t]]\ wz[t])\))\)\), 
            "+", \(wy[
                t]\ \(("Iy"\ wx[t] + "m"\ "R"\^2\ Cos[psi[t]]\^2\ wx[t] - 
                  "m"\ "R"\ Sin[psi[t]]\^2\ Tan[
                      theta[t]]\ \((\(-"R"\)\ Sin[psi[t]]\ wy[t] - 
                        "R"\ Cos[psi[t]]\ wz[t])\))\)\), 
            "+", \(wx[
                t]\ \((\(-"J"\)\ wy[t] - "m"\ "R"\^2\ Sin[psi[t]]\^2\ wy[t] - 
                  "m"\ "R"\ Sin[
                      psi[t]]\ \((\(-"R"\)\ Sin[psi[t]]\ wy[t] - 
                        "R"\ Cos[psi[t]]\ wz[t])\) - 
                  "R"\ Cos[
                      psi[t]]\ \(("m"\ "R"\ Sin[psi[t]]\ wz[t] + 
                        "m"\ \(("R"\ Cos[psi[t]]\ wy[t] - 
                              "R"\ Sin[psi[t]]\ wz[t])\))\))\)\), "-", 
            RowBox[{"\<\"m\"\>", " ", \("R"\^2\), " ", \(Cos[psi[t]]\), 
              " ", \(Sin[psi[t]]\), " ", 
              RowBox[{
                SuperscriptBox["wy", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}], "+", 
            RowBox[{\(("Iy" + "m"\ "R"\^2\ Sin[psi[t]]\^2)\), " ", 
              RowBox[{
                SuperscriptBox["wz", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}]}], "==", 
          "0"}]}], "}"}]], "Output",
  CellLabel->"Out[334]=",
  CellTags->"b:16.3"]
}, Open  ]],

Cell["\<\
Since x,y,z locate the center of gravity it is of interest to compute the \
contact point.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell["\<\
ChnBodyList={{CenterOfMass,{0,-R*Sin[psi],-R*Cos[psi]},Mass,InertiaMatrix}};
\
\>", "Input",
  CellLabel->"In[335]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{

Cell["\<\
{ContactPoint}=Simplify[Transpose[EndEffector[ChnBodyList,X][[{1,2,3},{4}]]]/.\
Sols]\
\>", "Input",
  CellLabel->"In[336]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[BoxData[
    \({{x - "R"\ Cos[phi]\ Sin[theta], 
        y - "R"\ Sin[phi]\ Sin[theta], \(-"R"\)}}\)], "Output",
  CellLabel->"Out[336]=",
  CellTags->"b:16.3"]
}, Open  ]],

Cell["\<\
Consider the equations of motion as given by the list Eqns. Careful \
inspection of the equations suggests that representation of the angular \
velocity in a fame that does not rotate about the body-x axis may simplify \
them. Thus, we transform the angular velocity coordinates, wy, wz, via the \
relations:\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{

Cell["\<\
a1=wyy[t]==Cos[psi[t]]*wy[t]-Sin[psi[t]]*wz[t]
a2=wzz[t]==Sin[psi[t]]*wy[t]+Cos[psi[t]]*wz[t]
\
\>", "Input",
  CellLabel->"In[337]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[BoxData[
    \(wyy[t] == Cos[psi[t]]\ wy[t] - Sin[psi[t]]\ wz[t]\)], "Output",
  CellLabel->"Out[337]=",
  CellTags->"b:16.3"],

Cell[BoxData[
    \(wzz[t] == Sin[psi[t]]\ wy[t] + Cos[psi[t]]\ wz[t]\)], "Output",
  CellLabel->"Out[338]=",
  CellTags->"b:16.3"]
}, Open  ]],

Cell["\<\
The inverse transformation rules are obtained using Mathematica's Solve \
function\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{

Cell["Trans=Simplify[Flatten[Solve[{a1,a2},{wy[t],wz[t]}]]]", "Input",
  CellLabel->"In[339]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[BoxData[
    \({wy[t] \[Rule] Cos[psi[t]]\ wyy[t] + Sin[psi[t]]\ wzz[t], 
      wz[t] \[Rule] \(-Sin[psi[t]]\)\ wyy[t] + 
          Cos[psi[t]]\ wzz[t]}\)], "Output",
  CellLabel->"Out[339]=",
  CellTags->"b:16.3"]
}, Open  ]],

Cell["\<\
The transformed equations are obtained with the function StateTransformation:\
\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{

Cell["StateVars={psi,theta,phi,x,y,wx,wy,wz}", "Input",
  CellLabel->"In[340]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[BoxData[
    \({psi, theta, phi, x, y, wx, wy, wz}\)], "Output",
  CellLabel->"Out[340]=",
  CellTags->"b:16.3"]
}, Open  ]],

Cell[CellGroupData[{

Cell["TrEqns=StateTransformation[Eqns,StateVars,Trans,{wyy,wzz},t]", "Input",
  CellLabel->"In[341]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          RowBox[{
            SuperscriptBox["theta", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], "==", \(wyy[t]\)}], 
        ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["phi", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], 
          "==", \(Sec[theta[t]]\ wzz[t]\)}], ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["x", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], 
          "==", \("R"\ \((Sin[phi[t]]\ wx[t] + 
                Cos[phi[t]]\ Cos[theta[t]]\ wyy[t])\)\)}], ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["y", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], 
          "==", \(\(-"R"\)\ Cos[phi[t]]\ wx[t] + 
            "R"\ Cos[theta[t]]\ Sin[phi[t]]\ wyy[t]\)}], ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["wx", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], 
          "==", \(\("m"\ "R"\^2\ wyy[t]\ wzz[t]\)\/\("J" + "m"\ "R"\^2\)\)}], 
        ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["wzz", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], 
          "==", \(wyy[
              t]\ \((\("J"\ wx[t]\)\/"Iy" + Tan[theta[t]]\ wzz[t])\)\)}], ",",
         
        RowBox[{
          RowBox[{
            SuperscriptBox["wyy", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], 
          "==", \(\("g"\ "m"\ "R"\ Sin[
                    theta[t]] - \(("J" + "m"\ "R"\^2)\)\ wx[t]\ wzz[t] - 
                "Iy"\ Tan[theta[t]]\ wzz[t]\^2\)\/\("Iy" + "m"\ "R"\^2\)\)}], 
        ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["psi", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], 
          "==", \(wx[t] + Tan[theta[t]]\ wzz[t]\)}]}], "}"}]], "Output",
  CellLabel->"Out[341]=",
  CellTags->"b:16.3"]
}, Open  ]],

Cell["\<\
These equations are obviously equivalent to thedisk equations given by \
Neimark and Fufaev. Another derivation is given by Meirovitch. To compare our \
equations with his, it is necessary to reduce them to second order form by \
eliminating the quasi-velociites (thus, we get Lagrange's equations), and to \
perform a minor transformation of angular coordinates:\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{

Cell["\<\
Rules1=Flatten[Solve[TrEqns[[{1,2,8}]],{wx[t],wyy[t],wzz[t]}]]\
\>", "Input",
  CellLabel->"In[342]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{\(wx[t]\), "\[Rule]", 
          RowBox[{
            RowBox[{\(-Sin[theta[t]]\), " ", 
              RowBox[{
                SuperscriptBox["phi", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}], "+", 
            RowBox[{
              SuperscriptBox["psi", "\[Prime]",
                MultilineFunction->None], "[", "t", "]"}]}]}], ",", 
        RowBox[{\(wyy[t]\), "\[Rule]", 
          RowBox[{
            SuperscriptBox["theta", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}]}], ",", 
        RowBox[{\(wzz[t]\), "\[Rule]", 
          RowBox[{\(Cos[theta[t]]\), " ", 
            RowBox[{
              SuperscriptBox["phi", "\[Prime]",
                MultilineFunction->None], "[", "t", "]"}]}]}]}], 
      "}"}]], "Output",
  CellLabel->"Out[342]=",
  CellTags->"b:16.3"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Rules2={wx'[t]->D[wx[t]/.Rules1,t],
        wyy'[t]->D[wyy[t]/.Rules1,t],
        wzz'[t]->D[wzz[t]/.Rules1,t]}\
\>", "Input",
  CellLabel->"In[343]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          RowBox[{
            SuperscriptBox["wx", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], "\[Rule]", 
          RowBox[{
            RowBox[{\(-Cos[theta[t]]\), " ", 
              RowBox[{
                SuperscriptBox["phi", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}], " ", 
              RowBox[{
                SuperscriptBox["theta", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}], "-", 
            RowBox[{\(Sin[theta[t]]\), " ", 
              RowBox[{
                SuperscriptBox["phi", "\[Prime]\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}], "+", 
            RowBox[{
              SuperscriptBox["psi", "\[Prime]\[Prime]",
                MultilineFunction->None], "[", "t", "]"}]}]}], ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["wyy", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], "\[Rule]", 
          RowBox[{
            SuperscriptBox["theta", "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "t", "]"}]}], ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["wzz", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], "\[Rule]", 
          RowBox[{
            RowBox[{\(-Sin[theta[t]]\), " ", 
              RowBox[{
                SuperscriptBox["phi", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}], " ", 
              RowBox[{
                SuperscriptBox["theta", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}], "+", 
            RowBox[{\(Cos[theta[t]]\), " ", 
              RowBox[{
                SuperscriptBox["phi", "\[Prime]\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}]}]}]}], 
      "}"}]], "Output",
  CellLabel->"Out[343]=",
  CellTags->"b:16.3"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Rules3={theta[t]->th[t]-Pi/2,
        theta'[t]->th'[t],
        theta''[t]->th''[t]}\
\>", "Input",
  CellLabel->"In[344]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{\(theta[t] \[Rule] \(-\(\[Pi]\/2\)\) + th[t]\), ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["theta", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], "\[Rule]", 
          RowBox[{
            SuperscriptBox["th", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}]}], ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["theta", "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], "\[Rule]", 
          RowBox[{
            SuperscriptBox["th", "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "t", "]"}]}]}], "}"}]], "Output",\

  CellLabel->"Out[344]=",
  CellTags->"b:16.3"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Rules4={phi[t]->-pi/2+ph[t],
        phi'[t]->ph'[t],
        phi''[t]->ph''[t]}\
\>", "Input",
  CellLabel->"In[345]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{\(phi[t] \[Rule] \(-\(pi\/2\)\) + ph[t]\), ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["phi", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], "\[Rule]", 
          RowBox[{
            SuperscriptBox["ph", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}]}], ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["phi", "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], "\[Rule]", 
          RowBox[{
            SuperscriptBox["ph", "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "t", "]"}]}]}], "}"}]], "Output",\

  CellLabel->"Out[345]=",
  CellTags->"b:16.3"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
LagEqns=Simplify[TrEqns[[{5,6,7}]]/.Rules1/.Rules2/.Rules3/.Rules4]\
\>", "Input",
  CellLabel->"In[346]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          RowBox[{
            RowBox[{\(Cos[th[t]]\), " ", 
              RowBox[{
                SuperscriptBox["ph", "\[Prime]\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}], "+", 
            RowBox[{
              SuperscriptBox["psi", "\[Prime]\[Prime]",
                MultilineFunction->None], "[", "t", "]"}]}], "==", 
          FractionBox[
            RowBox[{\(("J" + 2\ "m"\ "R"\^2)\), " ", \(Sin[th[t]]\), " ", 
              RowBox[{
                SuperscriptBox["ph", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}], " ", 
              RowBox[{
                SuperscriptBox["th", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}], \("J" + 
              "m"\ "R"\^2\)]}], ",", 
        RowBox[{
          FractionBox[
            RowBox[{
              RowBox[{\((2\ "Iy" - "J")\), " ", \(Cos[th[t]]\), " ", 
                RowBox[{
                  SuperscriptBox["ph", "\[Prime]",
                    MultilineFunction->None], "[", "t", "]"}], " ", 
                RowBox[{
                  SuperscriptBox["th", "\[Prime]",
                    MultilineFunction->None], "[", "t", "]"}]}], "-", 
              RowBox[{"\<\"J\"\>", " ", 
                RowBox[{
                  SuperscriptBox["psi", "\[Prime]",
                    MultilineFunction->None], "[", "t", "]"}], " ", 
                RowBox[{
                  SuperscriptBox["th", "\[Prime]",
                    MultilineFunction->None], "[", "t", "]"}]}], "+", 
              RowBox[{"\<\"Iy\"\>", " ", \(Sin[th[t]]\), " ", 
                RowBox[{
                  SuperscriptBox["ph", "\[Prime]\[Prime]",
                    MultilineFunction->None], "[", "t", 
                  "]"}]}]}], "\<\"Iy\"\>"], "==", "0"}], ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["th", "\[Prime]\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], "==", 
          RowBox[{\(1\/\("Iy" + "m"\ "R"\^2\)\), 
            RowBox[{"(", 
              RowBox[{\(\(-"g"\)\ "m"\ "R"\ Cos[th[t]]\), "+", 
                
                RowBox[{"\<\"Iy\"\>", " ", \(Cos[th[t]]\), 
                  " ", \(Sin[th[t]]\), " ", 
                  SuperscriptBox[
                    RowBox[{
                      SuperscriptBox["ph", "\[Prime]",
                        MultilineFunction->None], "[", "t", "]"}], "2"]}], 
                "-", 
                RowBox[{\(("J" + "m"\ "R"\^2)\), " ", \(Sin[th[t]]\), " ", 
                  RowBox[{
                    SuperscriptBox["ph", "\[Prime]",
                      MultilineFunction->None], "[", "t", "]"}], " ", 
                  RowBox[{"(", 
                    RowBox[{
                      RowBox[{\(Cos[th[t]]\), " ", 
                        RowBox[{
                          SuperscriptBox["ph", "\[Prime]",
                            MultilineFunction->None], "[", "t", "]"}]}], "+", 
                      
                      RowBox[{
                        SuperscriptBox["psi", "\[Prime]",
                          MultilineFunction->None], "[", "t", "]"}]}], 
                    ")"}]}]}], ")"}]}]}]}], "}"}]], "Output",
  CellLabel->"Out[346]=",
  CellTags->"b:16.3"]
}, Open  ]],

Cell["\<\
These equations are easily confirmed to be equivalent to those given by \
Meirovitch.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["A Rigid Drive Train", "Subsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}],

Cell[CellGroupData[{

Cell["Description", "Subsubsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}],

Cell["\<\
The system is a chain consisting of 3 bodies (coupler, forward shaft and aft \
shaft), 3 joints (one revolute andtwo universal) and and one constraints \
(cylindrical). \
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Body Data", "Subsubsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}],

Cell["Body 1: Coupler", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}],

Cell["\<\
com1={0,0,0}; mass1=12925.9; out1={{2,{0,0,0}}};
Inertia1=DiagonalMatrix[{3.6112 10^5,3.8792 10^5,3.6794 10^5}];

Body1={com1,out1,mass1,Inertia1};
\
\>", "Input",
  CellLabel->"In[347]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}],

Cell["Body 2: Forward Shaft", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}],

Cell["\<\
com2={0,175/2,0}; mass2=24232.2; out2={{3,{0,175,0}}};
Inertia2=DiagonalMatrix[{(mass2*(3*25^2+175^2))/12,(mass2*25^2)/2,(mass2*(3*\
25^2+175^2))/12}];

Body2={com2,out2,mass2,Inertia2};\
\>", "Input",
  CellLabel->"In[350]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}],

Cell["Body 3: Aft Shaft", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}],

Cell["\<\
com3={0,10,0}; mass3=19618.8; out3={{6,{0,10,0}},{4,{0,20,0}}};
Inertia3=DiagonalMatrix[{2.2558 10^6,2.8837 10^5,2.2416 10^6}];

Body3={com3,out3,mass3,Inertia3};\
\>", "Input",
  CellLabel->"In[353]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}],

Cell["BodyLst=Chop[{Body1,Body2,Body3},10^-5];", "Input",
  CellLabel->"In[356]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Joint Data", "Subsubsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}],

Cell["Joint 1: Revolute", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  FontFamily->"Times New Roman",
  FontSize->12,
  FontWeight->"Bold",
  FontColor->GrayLevel[0],
  Background->GrayLevel[1],
  CellTags->{"Drive Chain", "b:16.4"}],

Cell["\<\
r1={1};H1={{0},{1},{0},{0},{0},{0}};
q1={a1y};p1={w1y};
Joint1={r1,H1,q1,p1};\
\>", "Input",
  CellLabel->"In[357]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}],

Cell["Joints 2 and 3: Universal", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  FontFamily->"Times New Roman",
  FontSize->12,
  FontWeight->"Bold",
  FontColor->GrayLevel[0],
  Background->GrayLevel[1],
  CellTags->{"Drive Chain", "b:16.4"}],

Cell["\<\
r2={1,1};H2={{1,0},{0,0},{0,1},{0,0},{0,0},{0,0}};
q2={a2x,a2z};p2={w2x,w2z};
Joint2={r2,H2,q2,p2};\
\>", "Input",
  CellLabel->"In[360]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}],

Cell["\<\
Universal joints have singularities which restritc them to 'local' two degree \
of freedom rotations, if the joint operates within the allowable limits the \
relative motion is equivalent to a spherical joint. Treating universal joints \
a spherical generally produces simpler dynamical equations. For variety, we \
do that with the last three joints.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}],

Cell["\<\
r3={1,1};H3={{1,0},{0,0},{0,1},{0,0},{0,0},{0,0}};
q3={a3x,a3z};p3={w3x,w3z};
Joint3={r3,H3,q3,p3};


JointLst={Joint1,Joint2,Joint3};
\
\>", "Input",
  CellLabel->"In[363]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Tree Definition", "Subsubsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}],

Cell["The tree consists of one chain, so we have:", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}],

Cell["TreeLst={{{1,1},{2,2},{3,3}}};", "Input",
  CellLabel->"In[367]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["The Unconstrained Tree Model", "Subsubsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}],

Cell["PE=0;Q={0,0,0,0,0};", "Input",
  CellLabel->"In[368]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}],

Cell[CellGroupData[{

Cell["{V,X,H}=Joints[JointLst];", "Input",
  CellLabel->"In[369]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}],

Cell[BoxData[
    \("Computing joint 1 kinematics"\)], "Print",
  CellLabel->"From In[369]:=",
  CellTags->"b:16.4"],

Cell[BoxData[
    \("Computing joint 2 kinematics"\)], "Print",
  CellLabel->"From In[369]:=",
  CellTags->"b:16.4"],

Cell[BoxData[
    \("Computing joint 3 kinematics"\)], "Print",
  CellLabel->"From In[369]:=",
  CellTags->"b:16.4"]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
{V,X,H,M,Cp,Fp,p,q}=CreateModel[JointLst,BodyLst,TreeLst,0,PE,Q,V,X,H];\
\>", "Input",
  CellLabel->"In[370]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "b:16.4"}],

Cell[BoxData[
    \("Computing Potential Functions"\)], "Print",
  CellLabel->"From In[370]:=",
  CellTags->"b:16.4"],

Cell[BoxData[
    \("Computing Inertia Matrix"\)], "Print",
  CellLabel->"From In[370]:=",
  CellTags->"b:16.4"],

Cell[BoxData[
    \("Computing Poincare Function"\)], "Print",
  CellLabel->"From In[370]:=",
  CellTags->"b:16.4"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Adding the Constraints", "Subsubsection",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "i:10", "b:16.4"}],

Cell["\<\
There are two algebraic constraint sets. One at node 3 of the fore shaft \
constrains the node x-z coordinates. Similarly, the second applies at node 4 \
of the aft shaft and fixes its x and z coordinates. To set up the \
constraints, we compute the relevent node locations (3 and 4) in the space \
frame using the EndEffector function.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "i:10", "b:16.4"}],

Cell["\<\
AftShaftPos=Chop[EndEffector[4,TreeLst,BodyLst,X],10^-5];
Gangles={AftShaftPos[[1,2]],AftShaftPos[[2,1]],AftShaftPos[[2,3]],
    AftShaftPos[[2,2]]-1,AftShaftPos[[3,2]]};
ForeShaftPos=Chop[EndEffector[3,TreeLst,BodyLst,X],10^-5];\
\>", "Input",
  CellLabel->"In[371]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "i:10", "b:16.4"}],

Cell["\<\
G={ForeShaftPos[[1,4]]/175-alpha,ForeShaftPos[[3,4]]/175+alpha,
    AftShaftPos[[1,4]]/175-alpha,AftShaftPos[[3,4]]/175+alpha};
Gs=SeriesExpansion[G,{a2x,a2z,a3x,a3z},1];\
\>", "Input",
  CellLabel->"In[374]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "i:10", "b:16.4"}],

Cell["\<\
Ms=Chop[N[SeriesExpansion[M/10^8,{a2x,a2z,a3x,a3z},1]],10^-4];
Cps=Chop[N[SeriesExpansion[Cp/10^8,{a2x,a2z,a3x,a3z},1]],10^-4];
Fps=Fp/10^8;\
\>", "Input",
  CellLabel->"In[376]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "i:10", "b:16.4"}],

Cell["Vs=V;", "Input",
  CellLabel->"In[379]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "i:10", "b:16.4"}],

Cell[CellGroupData[{

Cell["\<\
{Mm,Cm,Fm,Vm,phat,qhat}=AlgebraicConstraints[Ms,Cps,Fps,Vs,Gs,p,q,{2,3,4,5},{\
2,3,4,5},{2,3,4,5}];\
\>", "Input",
  CellLabel->"In[380]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "i:10", "b:16.4"}],

Cell[BoxData[
    \("Vm"\)], "Print",
  CellLabel->"From In[380]:=",
  CellTags->{"i:10", "b:16.4"}],

Cell[BoxData[
    \("Mm"\)], "Print",
  CellLabel->"From In[380]:=",
  CellTags->{"i:10", "b:16.4"}],

Cell[BoxData[
    \("Cm"\)], "Print",
  CellLabel->"From In[380]:=",
  CellTags->{"i:10", "b:16.4"}],

Cell[BoxData[
    \("Fm"\)], "Print",
  CellLabel->"From In[380]:=",
  CellTags->{"i:10", "b:16.4"}],

Cell[BoxData[
    \("Eliminating specified coordinates."\)], "Print",
  CellLabel->"From In[380]:=",
  CellTags->{"i:10", "b:16.4"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Mm", "Input",
  CellLabel->"In[381]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "i:10", "b:16.4"}],

Cell[BoxData[
    \({{\(\(0.082488525`\)\(\[InvisibleSpace]\)\) - 
          16.888196874999995`\ alpha\^2\ Cos[a1y]\^2 - 
          0.32597209999999777`\ alpha\^2\ Cos[a1y]\ Sin[a1y] - 
          16.888196874999995`\ alpha\^2\ Sin[a1y]\^2}}\)], "Output",
  CellLabel->"Out[381]=",
  CellTags->{"i:10", "b:16.4"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Cm", "Input",
  CellLabel->"In[382]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "i:10", "b:16.4"}],

Cell[BoxData[
    \({{\(-0.0014418499999999668`\)\ alpha\^4\ w1y\ Cos[a1y]\^4 - 
          0.7563091`\ alpha\^4\ w1y\ Cos[a1y]\^3\ Sin[a1y] + 
          alpha\^2\ w1y\ Sin[a1y]\^2\ \((\(\(0.12223953749999783`\)\(\
\[InvisibleSpace]\)\) + 0.0014418499999999668`\ alpha\^2\ Sin[a1y]\^2)\) + 
          Cos[a1y]\^2\ \((\(-0.12223953749999872`\)\ alpha\^2\ w1y + 
                0.`\ alpha\^4\ w1y\ Sin[a1y]\^2)\) + 
          Cos[a1y]\ \((0.`\ alpha\^2\ w1y\ Sin[a1y] + 
                0.7563091`\ alpha\^4\ w1y\ Sin[a1y]\^3)\)}}\)], "Output",
  CellLabel->"Out[382]=",
  CellTags->{"i:10", "b:16.4"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Fm", "Input",
  CellLabel->"In[383]:=",
  PageWidth->Infinity,
  AspectRatioFixed->True,
  CellTags->{"Drive Chain", "i:10", "b:16.4"}],

Cell[BoxData[
    \({0}\)], "Output",
  CellLabel->"Out[383]=",
  CellTags->{"i:10", "b:16.4"}]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Bibliography", "Section",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"References", "b:17"}],

Cell["\<\
To use TSi Dynamics requires some knowledge about dynamics. For those users \
seeking more information on the concepts behind the TSi Dynamics modeling \
formalism, we list a few reference books. Some have been cited above in our \
examples.

V. I. Arnold, Mathematical Methods of Classical Mechanics, Springer-Verlag: \
New York, 1989 (2nd ed.)
N. G. Chetaev, Theoretical Mechanics, Springer-Verlag: New York, 1989.
S. H. Crandall, D. C. Karnopp, E. F. Kurtz, D. C. Pridmore-Brown, Dynamics of \
Mechanical and Electromechanial Systems, McGraw-Hill: New York, 1968.
F. Gantmacher, Lectures in Analytical Mechanics, MIR: Moscow, 1975.
J. H. Ginsberg, Advanced Engineering Dynamics, Harper & Row: New York, 1988.
H. Goldstein, Classical Mechanics, Addison-Wesley: Reading, 1981 (2nd ed.)
D. T. Greenwood, Classical Mechanics, Prentice-Hall: Englewood Cliffs, 1977.
T. R. Kane and D. A. Levinson, Dynamics: Theory and Application, McGraw-Hill: \
New York, 1985.
C. Lanczos, The Variational Principles of Mechanics, University of Toronto \
Press: Toronto, 1970 (4th ed.)
L.Meirovitch, Methods of Analytical Dynamics, McGraw-Hill: New York, 1970.
Ju. I. Neimark and N. A. Fufaev, Dynamics of Nonholonomic Systems, American \
Mathematical Society: Providence, 1972.
R. M. Rosenberg, Analytical Dynamics of Discrete Systems, Plenum: New York, \
1977.
H. G. Kwatny and G. L. Blankenship, Nonlinear Contril and Analystical \
Mechanics: a computational approach, Birkhauser, Boston, 2000.\
\>", "Text",
  Evaluatable->False,
  CellHorizontalScrolling->False,
  AspectRatioFixed->True,
  CellTags->{"References", "b:17"}]
}, Open  ]]
},
FrontEndVersion->"5.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 695}},
ShowPageBreaks->False,
WindowToolbars->{},
WindowSize->{808, 593},
WindowMargins->{{6, Automatic}, {Automatic, -3}},
PrintingCopies->1,
PrintingStartingPageNumber->29,
PrintingPageRange->{Automatic, Automatic},
PrivateNotebookOptions->{"ColorPalette"->{RGBColor, -1}},
TaggingRules:>{"NextIndexingNumber" -> 11, "IndexEntries" -> {{"i:1", \
"CreateModel"}, {"i:10", "SeriesExpansion"}, {"i:2", "MakeODEs"}, {"i:3", \
"CreateModelMEX"}, {"i:4", "CreateModel"}, {"i:4", "Joints"}, {"i:4", \
"LeafPotential"}, {"i:4", "SpringPotential"}, {"i:5", "CreateModel"}, {"i:5", \
"EndEffector"}, {"i:6", "CreateModel"}, {"i:6", "CreateModelMEX"}, {"i:6", \
"GeneralizedForce"}, {"i:6", "Joints"}, {"i:6", "KinematicReplacements"}, \
{"i:7", "AlgebraicConstraints"}, {"i:7", "EndEffector"}, {"i:8", \
"AlgebraicConstraints"}, {"i:9", "DifferentialConstraints"}, {"i:9", \
"NodeVelocity"}}, "PaginationCache" -> {{2003, 8, 1, 21, 15, \
44.1344032`9.397352127438063}, {29, 29, 29, 29, 29, 29, 29, 29, 30, 30, 30, \
30, 30, 30, 30, 30, 30, 31, 31, 32, 32, 33, 33, 33, 33, 33, 34, 34, 35, 35, \
35, 35, 35, 35, 35, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 37, 37, \
37, 37, 37, 37, 38, 38, 39, 39, 40, 40, 40, 40, 40, 40, 41, 41, 41, 41, 41, \
41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 41, 42, 42, 42, 42, 42, 42, 42, 42, \
42, 42, 42, 42, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 43, 44, 44, 44, 44, \
44, 44, 44, 44, 45, 45, 45, 45, 45, 45, 45, 45, 46, 46, 46, 46, 46, 46, 46, \
46, 46, 46, 47, 47, 47, 47, 47, 48, 48, 48, 48, 48, 48, 48, 49, 49, 49, 49, \
49, 49, 49, 50, 50, 50, 50, 50, 50, 50, 50, 50, 50, 51, 51, 51, 51, 51, 51, \
51, 51, 52, 52, 53, 54, 54, 55, 55, 55, 55, 55, 55, 55, 55, 55, 56, 56, 56, \
56, 56, 56, 56, 56, 56, 56, 57, 57, 57, 57, 57, 57, 57, 58, 58, 58, 58, 58, \
58, 58, 58, 58, 60, 60, 61, 62, 62, 62, 62, 62, 62, 62, 63, 63, 63, 64, 64, \
64, 65, 65, 65, 65, 65, 65, 66, 66, 66, 66, 66, 66, 66, 66, 66, 66, 67, 67, \
67, 68, 70, 70, 70, 70, 71, 71, 71, 71, 71, 72, 72, 72, 72, 72, 72, 73, 73, \
73, 73, 73, 73, 73, 73, 73, 73, 74, 74, 74, 75, 75, 75, 75, 75, 75, 75, 75, \
75, 75, 75, 76, 76, 76, 76, 76, 77, 77, 77, 77, 77, 78, 78, 78, 78, 78, 78, \
78, 78, 78, 78, 78, 78, 79, 79, 79, 79, 79, 79, 79, 79, 79, 79, 80, 80, 80, \
80, 80, 80, 80, 80, 80, 80, 81, 81, 81, 81, 81, 82, 82, 82, 82, 82, 82, 82, \
82, 82, 82, 82, 82, 83, 83, 83, 83, 83, 83, 83, 84, 84, 84, 84, 84, 84, 84, \
85, 85, 85, 85, 85, 85, 85, 85, 85, 85, 85, 85, 85, 86, 86, 86, 86, 86, 86, \
86, 86, 86, 86, 86, 86, 86, 87, 87, 87, 87, 87, 87, 87, 87, 88, 88, 88, 88, \
88, 88, 88, 88, 88, 88, 88, 88, 89, 89, 89, 89, 89, 89, 89, 89, 89, 90, 90, \
90, 90, 90, 90, 90, 90, 90, 90, 90, 90, 91, 91, 91, 91, 91, 91, 91, 91, 91, \
91, 91, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 92, 93, 93, \
93, 93, 93, 93, 93, 93, 93, 93, 93, 93, 94, 94, 94, 94, 94, 94, 95, 95}}},
ShowCellLabel->True,
CellLabelAutoDelete->False,
ShowCellTags->True,
RenderingOptions->{"ObjectDithering"->True,
"RasterDithering"->False},
StyleDefinitions -> "HelpBrowser.nb"
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "Introduction"->{
    Cell[1776, 53, 151, 4, 70, "Section",
      Evaluatable->False,
      CellTags->{"Introduction", "b:11"}],
    Cell[2373, 78, 1782, 30, 292, "Text",
      Evaluatable->False,
      CellTags->{"Introduction", "b:11"}]},
  "b:11"->{
    Cell[1776, 53, 151, 4, 70, "Section",
      Evaluatable->False,
      CellTags->{"Introduction", "b:11"}],
    Cell[1930, 59, 145, 4, 63, "Input",
      CellTags->"b:11"],
    Cell[2100, 67, 134, 3, 43, "Input",
      CellTags->"b:11"],
    Cell[2237, 72, 121, 3, 41, "Output",
      CellTags->"b:11"],
    Cell[2373, 78, 1782, 30, 292, "Text",
      Evaluatable->False,
      CellTags->{"Introduction", "b:11"}]},
  "Getting Started"->{
    Cell[4192, 113, 157, 4, 70, "Section",
      Evaluatable->False,
      CellTags->{"Getting Started", "b:12"}],
    Cell[4352, 119, 772, 15, 90, "Text",
      Evaluatable->False,
      CellTags->{"Getting Started", "b:12"}],
    Cell[5127, 136, 256, 9, 39, "Text",
      Evaluatable->False,
      CellTags->{"Getting Started", "b:12"}],
    Cell[5408, 149, 140, 4, 45, "Input",
      CellTags->{"Getting Started", "b:12"}],
    Cell[6275, 169, 172, 6, 39, "Text",
      CellTags->{"Getting Started", "b:12"}]},
  "b:12"->{
    Cell[4192, 113, 157, 4, 70, "Section",
      Evaluatable->False,
      CellTags->{"Getting Started", "b:12"}],
    Cell[4352, 119, 772, 15, 90, "Text",
      Evaluatable->False,
      CellTags->{"Getting Started", "b:12"}],
    Cell[5127, 136, 256, 9, 39, "Text",
      Evaluatable->False,
      CellTags->{"Getting Started", "b:12"}],
    Cell[5408, 149, 140, 4, 45, "Input",
      CellTags->{"Getting Started", "b:12"}],
    Cell[5551, 155, 709, 11, 417, "Print",
      CellTags->"b:12"],
    Cell[6275, 169, 172, 6, 39, "Text",
      CellTags->{"Getting Started", "b:12"}]},
  "What's in the Package"->{
    Cell[6636, 187, 175, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"What's in the Package", "b:13.0"}],
    Cell[6814, 193, 234, 7, 39, "Text",
      Evaluatable->False,
      CellTags->{"What's in the Package", "b:13.0"}],
    Cell[7073, 204, 147, 4, 45, "Input",
      CellTags->{"What's in the Package", "b:13.0"}],
    Cell[7475, 223, 119, 3, 43, "Input",
      CellTags->{"What's in the Package", "b:13.0"}],
    Cell[8813, 252, 127, 3, 43, "Input",
      CellTags->{"What's in the Package", "b:13.0"}],
    Cell[10199, 278, 1892, 31, 341, "Text",
      Evaluatable->False,
      CellTags->{"What's in the Package", "b:13.0"}]},
  "b:13.0"->{
    Cell[6636, 187, 175, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"What's in the Package", "b:13.0"}],
    Cell[6814, 193, 234, 7, 39, "Text",
      Evaluatable->False,
      CellTags->{"What's in the Package", "b:13.0"}],
    Cell[7073, 204, 147, 4, 45, "Input",
      CellTags->{"What's in the Package", "b:13.0"}],
    Cell[7223, 210, 124, 3, 37, "Print",
      CellTags->"b:13.0"],
    Cell[7350, 215, 88, 3, 41, "Output",
      CellTags->"b:13.0"],
    Cell[7475, 223, 119, 3, 43, "Input",
      CellTags->{"What's in the Package", "b:13.0"}],
    Cell[7597, 228, 1179, 19, 588, "Print",
      CellTags->{"Info3248853457-1624643", "b:13.0"}],
    Cell[8813, 252, 127, 3, 43, "Input",
      CellTags->{"What's in the Package", "b:13.0"}],
    Cell[8943, 257, 1241, 18, 569, "Print",
      CellTags->{"Info3248853457-6771253", "b:13.0"}],
    Cell[10199, 278, 1892, 31, 341, "Text",
      Evaluatable->False,
      CellTags->{"What's in the Package", "b:13.0"}]},
  "Info3248853457-1624643"->{
    Cell[7597, 228, 1179, 19, 588, "Print",
      CellTags->{"Info3248853457-1624643", "b:13.0"}]},
  "Info3248853457-6771253"->{
    Cell[8943, 257, 1241, 18, 569, "Print",
      CellTags->{"Info3248853457-6771253", "b:13.0"}]},
  "Data Structures"->{
    Cell[12128, 314, 162, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[12315, 322, 165, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[12483, 328, 363, 8, 56, "Text",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[12849, 338, 34187, 425, 319, 34048, 421, "GraphicsData", 
    "Metafile", "Graphics",
      CellTags->{"b:13.1", "Data Structures"}],
    Cell[47039, 765, 3093, 51, 609, "Text",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[50169, 821, 84, 1, 49, "Subsubsection",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[50256, 824, 1086, 16, 141, "Text",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[51379, 845, 156, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[51538, 851, 1941, 34, 391, "Text",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[53482, 887, 216, 7, 65, "Input",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[53701, 896, 228, 7, 39, "Text",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[53932, 905, 219, 7, 65, "Input",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[54154, 914, 396, 9, 56, "Text",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[54575, 927, 208, 8, 85, "Input",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[55056, 950, 142, 4, 45, "Input",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[55366, 964, 142, 4, 45, "Input",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[56016, 983, 142, 4, 45, "Input",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[56386, 996, 636, 12, 90, "Text",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[57047, 1012, 270, 11, 145, "Input",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[57472, 1033, 142, 4, 45, "Input",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[58289, 1056, 142, 4, 45, "Input",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[58932, 1076, 142, 4, 45, "Input",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[59344, 1092, 162, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[59509, 1098, 1358, 26, 340, "Text",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[60904, 1129, 165, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[61072, 1135, 2927, 49, 625, "Text",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}]},
  "b:13.1"->{
    Cell[12128, 314, 162, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[12315, 322, 165, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[12483, 328, 363, 8, 56, "Text",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[12849, 338, 34187, 425, 319, 34048, 421, "GraphicsData", 
    "Metafile", "Graphics",
      CellTags->{"b:13.1", "Data Structures"}],
    Cell[47039, 765, 3093, 51, 609, "Text",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[50169, 821, 84, 1, 49, "Subsubsection",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[50256, 824, 1086, 16, 141, "Text",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[51379, 845, 156, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[51538, 851, 1941, 34, 391, "Text",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[53482, 887, 216, 7, 65, "Input",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[53701, 896, 228, 7, 39, "Text",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[53932, 905, 219, 7, 65, "Input",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[54154, 914, 396, 9, 56, "Text",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[54575, 927, 208, 8, 85, "Input",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[54786, 937, 115, 3, 37, "Print",
      CellTags->"b:13.1"],
    Cell[54904, 942, 115, 3, 37, "Print",
      CellTags->"b:13.1"],
    Cell[55056, 950, 142, 4, 45, "Input",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[55201, 956, 128, 3, 37, "Print",
      CellTags->"b:13.1"],
    Cell[55366, 964, 142, 4, 45, "Input",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[55511, 970, 468, 8, 94, "Print",
      CellTags->"b:13.1"],
    Cell[56016, 983, 142, 4, 45, "Input",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[56161, 989, 210, 4, 56, "Print",
      CellTags->"b:13.1"],
    Cell[56386, 996, 636, 12, 90, "Text",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[57047, 1012, 270, 11, 145, "Input",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[57320, 1025, 115, 3, 37, "Print",
      CellTags->"b:13.1"],
    Cell[57472, 1033, 142, 4, 45, "Input",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[57617, 1039, 635, 12, 113, "Print",
      CellTags->"b:13.1"],
    Cell[58289, 1056, 142, 4, 45, "Input",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[58434, 1062, 461, 9, 94, "Print",
      CellTags->"b:13.1"],
    Cell[58932, 1076, 142, 4, 45, "Input",
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[59077, 1082, 218, 4, 56, "Print",
      CellTags->"b:13.1"],
    Cell[59344, 1092, 162, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[59509, 1098, 1358, 26, 340, "Text",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[60904, 1129, 165, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}],
    Cell[61072, 1135, 2927, 49, 625, "Text",
      Evaluatable->False,
      CellTags->{"Data Structures", "b:13.1"}]},
  "Building Equations"->{
    Cell[64222, 1198, 182, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[64407, 1204, 58025, 720, 218, 57887, 716, "GraphicsData", 
    "Metafile", "Graphics",
      CellTags->{"b:14.0", "Building Equations"}],
    Cell[122435, 1926, 446, 13, 214, "Text",
      Evaluatable->False,
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[122884, 1941, 305, 12, 165, "Input",
      CellTags->{"Building Equations", "i:1", "b:14.0"}],
    Cell[123192, 1955, 160, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[123355, 1961, 375, 12, 165, "Input",
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[123733, 1975, 176, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[123912, 1981, 161, 4, 45, "Input",
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[124076, 1987, 315, 8, 64, "Text",
      Evaluatable->False,
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[124394, 1997, 159, 4, 45, "Input",
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[124578, 2005, 210, 6, 45, "Input",
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[125505, 2044, 172, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[125702, 2052, 157, 4, 45, "Input",
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[126364, 2075, 384, 9, 56, "Text",
      Evaluatable->False,
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[126773, 2088, 157, 4, 45, "Input",
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[127493, 2111, 158, 4, 45, "Input",
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[128444, 2145, 446, 9, 56, "Text",
      Evaluatable->False,
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[128893, 2156, 305, 9, 85, "Input",
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[129201, 2167, 192, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[129396, 2173, 195, 4, 45, "Input",
      CellTags->{"Building Equations", "b:14.0"}]},
  "b:14.0"->{
    Cell[64222, 1198, 182, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[64407, 1204, 58025, 720, 218, 57887, 716, "GraphicsData", 
    "Metafile", "Graphics",
      CellTags->{"b:14.0", "Building Equations"}],
    Cell[122435, 1926, 446, 13, 214, "Text",
      Evaluatable->False,
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[122884, 1941, 305, 12, 165, "Input",
      CellTags->{"Building Equations", "i:1", "b:14.0"}],
    Cell[123192, 1955, 160, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[123355, 1961, 375, 12, 165, "Input",
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[123733, 1975, 176, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[123912, 1981, 161, 4, 45, "Input",
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[124076, 1987, 315, 8, 64, "Text",
      Evaluatable->False,
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[124394, 1997, 159, 4, 45, "Input",
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[124578, 2005, 210, 6, 45, "Input",
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[124791, 2013, 113, 3, 37, "Print",
      CellTags->"b:14.0"],
    Cell[124907, 2018, 115, 3, 37, "Print",
      CellTags->"b:14.0"],
    Cell[125025, 2023, 115, 3, 37, "Print",
      CellTags->"b:14.0"],
    Cell[125143, 2028, 116, 3, 37, "Print",
      CellTags->"b:14.0"],
    Cell[125262, 2033, 111, 3, 37, "Print",
      CellTags->"b:14.0"],
    Cell[125376, 2038, 114, 3, 37, "Print",
      CellTags->"b:14.0"],
    Cell[125505, 2044, 172, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[125702, 2052, 157, 4, 45, "Input",
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[125862, 2058, 487, 14, 51, "Print",
      CellTags->"b:14.0"],
    Cell[126364, 2075, 384, 9, 56, "Text",
      Evaluatable->False,
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[126773, 2088, 157, 4, 45, "Input",
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[126933, 2094, 523, 12, 55, "Print",
      CellTags->"b:14.0"],
    Cell[127493, 2111, 158, 4, 45, "Input",
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[127654, 2117, 775, 25, 49, "Print",
      CellTags->"b:14.0"],
    Cell[128444, 2145, 446, 9, 56, "Text",
      Evaluatable->False,
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[128893, 2156, 305, 9, 85, "Input",
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[129201, 2167, 192, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Building Equations", "b:14.0"}],
    Cell[129396, 2173, 195, 4, 45, "Input",
      CellTags->{"Building Equations", "b:14.0"}]},
  "i:1"->{
    Cell[122884, 1941, 305, 12, 165, "Input",
      CellTags->{"Building Equations", "i:1", "b:14.0"}]},
  "Simulation"->{
    Cell[129628, 2182, 179, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[129810, 2188, 362, 8, 56, "Text",
      Evaluatable->False,
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[130197, 2200, 171, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[132866, 2265, 223, 6, 39, "Text",
      Evaluatable->False,
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[133114, 2275, 196, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[133536, 2290, 174, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[135817, 2351, 150, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[136108, 2365, 244, 7, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[136525, 2382, 216, 6, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[136922, 2399, 181, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[139293, 2459, 246, 7, 39, "Text",
      Evaluatable->False,
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[139564, 2470, 174, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[140574, 2502, 173, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[172337, 3359, 173, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}]},
  "i:2"->{
    Cell[129628, 2182, 179, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[129810, 2188, 362, 8, 56, "Text",
      Evaluatable->False,
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[130197, 2200, 171, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[132866, 2265, 223, 6, 39, "Text",
      Evaluatable->False,
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[133114, 2275, 196, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[133536, 2290, 174, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[135817, 2351, 150, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[136108, 2365, 244, 7, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[136525, 2382, 216, 6, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[136922, 2399, 181, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[139293, 2459, 246, 7, 39, "Text",
      Evaluatable->False,
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[139564, 2470, 174, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[140574, 2502, 173, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[172337, 3359, 173, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}]},
  "b:14.1"->{
    Cell[129628, 2182, 179, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[129810, 2188, 362, 8, 56, "Text",
      Evaluatable->False,
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[130197, 2200, 171, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[130371, 2206, 2480, 56, 218, "Output",
      CellTags->"b:14.1"],
    Cell[132866, 2265, 223, 6, 39, "Text",
      Evaluatable->False,
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[133114, 2275, 196, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[133313, 2281, 186, 4, 41, "Output",
      CellTags->"b:14.1"],
    Cell[133536, 2290, 174, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[133713, 2296, 2067, 50, 199, "Output",
      CellTags->"b:14.1"],
    Cell[135817, 2351, 150, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[135970, 2357, 101, 3, 41, "Output",
      CellTags->"b:14.1"],
    Cell[136108, 2365, 244, 7, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[136355, 2374, 133, 3, 41, "Output",
      CellTags->"b:14.1"],
    Cell[136525, 2382, 216, 6, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[136744, 2390, 141, 4, 41, "Output",
      CellTags->"b:14.1"],
    Cell[136922, 2399, 181, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[137106, 2405, 2172, 51, 218, "Output",
      CellTags->"b:14.1"],
    Cell[139293, 2459, 246, 7, 39, "Text",
      Evaluatable->False,
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[139564, 2470, 174, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[139741, 2476, 796, 21, 60, "Output",
      CellTags->"b:14.1"],
    Cell[140574, 2502, 173, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[140750, 2508, 31370, 839, 198, 8066, 546, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"b:14.1"],
    Cell[172123, 3349, 177, 5, 41, "Output",
      CellTags->"b:14.1"],
    Cell[172337, 3359, 173, 4, 45, "Input",
      CellTags->{"Simulation", "i:2", "b:14.1"}],
    Cell[172513, 3365, 53974, 1166, 198, 9163, 607, "GraphicsData", 
    "PostScript", "Graphics",
      CellTags->"b:14.1"],
    Cell[226490, 4533, 177, 5, 41, "Output",
      CellTags->"b:14.1"]},
  "Creating MEX-Files"->{
    Cell[226716, 4544, 175, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],
    Cell[226894, 4550, 740, 13, 90, "Text",
      Evaluatable->False,
      CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],
    Cell[227637, 4565, 232, 6, 39, "Text",
      Evaluatable->False,
      CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],
    Cell[227894, 4575, 273, 9, 105, "Input",
      CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],
    Cell[228467, 4597, 693, 16, 185, "Input",
      CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],
    Cell[230395, 4665, 95, 1, 53, "Subsection",
      CellTags->{"Creating MEX-Files", "b:14.3"}],
    Cell[230493, 4668, 1690, 29, 375, "Text",
      CellTags->{"Creating MEX-Files", "b:14.3"}]},
  "i:3"->{
    Cell[226716, 4544, 175, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],
    Cell[226894, 4550, 740, 13, 90, "Text",
      Evaluatable->False,
      CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],
    Cell[227637, 4565, 232, 6, 39, "Text",
      Evaluatable->False,
      CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],
    Cell[227894, 4575, 273, 9, 105, "Input",
      CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],
    Cell[228467, 4597, 693, 16, 185, "Input",
      CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}]},
  "b:14.2"->{
    Cell[226716, 4544, 175, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],
    Cell[226894, 4550, 740, 13, 90, "Text",
      Evaluatable->False,
      CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],
    Cell[227637, 4565, 232, 6, 39, "Text",
      Evaluatable->False,
      CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],
    Cell[227894, 4575, 273, 9, 105, "Input",
      CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],
    Cell[228170, 4586, 260, 6, 60, "Output",
      CellTags->"b:14.2"],
    Cell[228467, 4597, 693, 16, 185, "Input",
      CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],
    Cell[229163, 4615, 125, 3, 37, "Print",
      CellTags->"b:14.2"],
    Cell[229291, 4620, 112, 3, 37, "Print",
      CellTags->"b:14.2"],
    Cell[229406, 4625, 132, 3, 37, "Print",
      CellTags->"b:14.2"],
    Cell[229541, 4630, 131, 3, 37, "Print",
      CellTags->"b:14.2"],
    Cell[229675, 4635, 119, 3, 37, "Print",
      CellTags->"b:14.2"],
    Cell[229797, 4640, 127, 3, 37, "Print",
      CellTags->"b:14.2"],
    Cell[229927, 4645, 122, 3, 37, "Print",
      CellTags->"b:14.2"],
    Cell[230052, 4650, 121, 3, 37, "Print",
      CellTags->"b:14.2"],
    Cell[230176, 4655, 170, 4, 37, "Print",
      CellTags->"b:14.2"]},
  "b:14.3"->{
    Cell[230395, 4665, 95, 1, 53, "Subsection",
      CellTags->{"Creating MEX-Files", "b:14.3"}],
    Cell[230493, 4668, 1690, 29, 375, "Text",
      CellTags->{"Creating MEX-Files", "b:14.3"}]},
  "b:15.0"->{
    Cell[232386, 4710, 146, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->"b:15.0"],
    Cell[232535, 4716, 619, 12, 90, "Text",
      Evaluatable->False,
      CellTags->"b:15.0"]},
  "Suspension"->{
    Cell[233191, 4733, 174, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[233368, 4739, 570, 12, 98, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[233963, 4755, 519, 20, 325, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[234851, 4793, 152, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[235006, 4799, 598, 23, 365, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[235607, 4824, 229, 7, 64, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[235839, 4833, 174, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[236016, 4839, 717, 14, 123, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[236736, 4855, 262, 9, 105, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[237001, 4866, 352, 8, 56, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[237356, 4876, 251, 8, 85, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[237632, 4888, 142, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[238125, 4903, 676, 12, 90, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[238804, 4917, 351, 12, 165, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[239180, 4933, 146, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[239557, 4948, 146, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[239912, 4961, 519, 11, 98, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[240434, 4974, 159, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[240596, 4980, 261, 8, 64, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[240860, 4990, 143, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[241006, 4996, 345, 9, 81, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[241376, 5009, 217, 6, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[241983, 5035, 138, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[247661, 5155, 152, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[248866, 5185, 138, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}]},
  "i:4"->{
    Cell[233191, 4733, 174, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[233368, 4739, 570, 12, 98, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[233963, 4755, 519, 20, 325, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[234851, 4793, 152, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[235006, 4799, 598, 23, 365, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[235607, 4824, 229, 7, 64, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[235839, 4833, 174, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[236016, 4839, 717, 14, 123, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[236736, 4855, 262, 9, 105, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[237001, 4866, 352, 8, 56, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[237356, 4876, 251, 8, 85, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[237632, 4888, 142, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[238125, 4903, 676, 12, 90, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[238804, 4917, 351, 12, 165, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[239180, 4933, 146, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[239557, 4948, 146, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[239912, 4961, 519, 11, 98, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[240434, 4974, 159, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[240596, 4980, 261, 8, 64, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[240860, 4990, 143, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[241006, 4996, 345, 9, 81, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[241376, 5009, 217, 6, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[241983, 5035, 138, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[247661, 5155, 152, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[248866, 5185, 138, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}]},
  "b:15.1"->{
    Cell[233191, 4733, 174, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[233368, 4739, 570, 12, 98, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[233963, 4755, 519, 20, 325, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[234485, 4777, 115, 3, 37, "Print",
      CellTags->"b:15.1"],
    Cell[234603, 4782, 115, 3, 37, "Print",
      CellTags->"b:15.1"],
    Cell[234721, 4787, 115, 3, 37, "Print",
      CellTags->"b:15.1"],
    Cell[234851, 4793, 152, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[235006, 4799, 598, 23, 365, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[235607, 4824, 229, 7, 64, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[235839, 4833, 174, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[236016, 4839, 717, 14, 123, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[236736, 4855, 262, 9, 105, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[237001, 4866, 352, 8, 56, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[237356, 4876, 251, 8, 85, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[237632, 4888, 142, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[237777, 4894, 333, 6, 74, "Output",
      CellTags->"b:15.1"],
    Cell[238125, 4903, 676, 12, 90, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[238804, 4917, 351, 12, 165, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[239180, 4933, 146, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[239329, 4939, 191, 4, 57, "Output",
      CellTags->"b:15.1"],
    Cell[239557, 4948, 146, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[239706, 4954, 191, 4, 57, "Output",
      CellTags->"b:15.1"],
    Cell[239912, 4961, 519, 11, 98, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[240434, 4974, 159, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[240596, 4980, 261, 8, 64, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[240860, 4990, 143, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[241006, 4996, 345, 9, 81, "Text",
      Evaluatable->False,
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[241376, 5009, 217, 6, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[241596, 5017, 117, 3, 37, "Print",
      CellTags->"b:15.1"],
    Cell[241716, 5022, 112, 3, 37, "Print",
      CellTags->"b:15.1"],
    Cell[241831, 5027, 115, 3, 37, "Print",
      CellTags->"b:15.1"],
    Cell[241983, 5035, 138, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[242124, 5041, 5500, 109, 1328, "Output",
      CellTags->"b:15.1"],
    Cell[247661, 5155, 152, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[247816, 5161, 1013, 19, 198, "Output",
      CellTags->"b:15.1"],
    Cell[248866, 5185, 138, 4, 45, "Input",
      CellTags->{"Suspension", "i:4", "b:15.1"}],
    Cell[249007, 5191, 2941, 51, 622, "Output",
      CellTags->"b:15.1"]},
  "Flexible Manipulator"->{
    Cell[251997, 5248, 190, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[252212, 5256, 186, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[252401, 5262, 590, 11, 73, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[252994, 5275, 1689, 27, 267, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[254686, 5304, 177, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[254866, 5310, 204, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[255073, 5316, 191, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[255267, 5322, 190, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[255460, 5328, 320, 9, 85, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[255783, 5339, 319, 8, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[256105, 5349, 301, 12, 165, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[256409, 5363, 241, 7, 65, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[256653, 5372, 390, 8, 56, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[257046, 5382, 188, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[257237, 5388, 222, 6, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[257462, 5396, 164, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[257629, 5402, 202, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[257834, 5408, 266, 9, 105, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[258103, 5419, 179, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[258307, 5427, 205, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[259506, 5453, 205, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[260683, 5477, 170, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[260856, 5483, 317, 14, 205, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[261176, 5499, 199, 7, 65, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[261400, 5510, 223, 6, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[262346, 5549, 188, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[262559, 5557, 148, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[271289, 5710, 161, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[278135, 5831, 149, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[280526, 5876, 288, 7, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[280817, 5885, 189, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[281009, 5891, 194, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[281228, 5899, 166, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}]},
  "i:5"->{
    Cell[251997, 5248, 190, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[252212, 5256, 186, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[252401, 5262, 590, 11, 73, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[252994, 5275, 1689, 27, 267, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[254686, 5304, 177, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[254866, 5310, 204, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[255073, 5316, 191, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[255267, 5322, 190, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[255460, 5328, 320, 9, 85, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[255783, 5339, 319, 8, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[256105, 5349, 301, 12, 165, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[256409, 5363, 241, 7, 65, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[256653, 5372, 390, 8, 56, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[257046, 5382, 188, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[257237, 5388, 222, 6, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[257462, 5396, 164, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[257629, 5402, 202, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[257834, 5408, 266, 9, 105, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[258103, 5419, 179, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[258307, 5427, 205, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[259506, 5453, 205, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[260683, 5477, 170, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[260856, 5483, 317, 14, 205, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[261176, 5499, 199, 7, 65, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[261400, 5510, 223, 6, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[262346, 5549, 188, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[262559, 5557, 148, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[271289, 5710, 161, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[278135, 5831, 149, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[280526, 5876, 288, 7, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[280817, 5885, 189, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[281009, 5891, 194, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[281228, 5899, 166, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}]},
  "b:15.2"->{
    Cell[251997, 5248, 190, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[252212, 5256, 186, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[252401, 5262, 590, 11, 73, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[252994, 5275, 1689, 27, 267, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[254686, 5304, 177, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[254866, 5310, 204, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[255073, 5316, 191, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[255267, 5322, 190, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[255460, 5328, 320, 9, 85, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[255783, 5339, 319, 8, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[256105, 5349, 301, 12, 165, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[256409, 5363, 241, 7, 65, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[256653, 5372, 390, 8, 56, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[257046, 5382, 188, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[257237, 5388, 222, 6, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[257462, 5396, 164, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[257629, 5402, 202, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[257834, 5408, 266, 9, 105, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[258103, 5419, 179, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[258307, 5427, 205, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[258515, 5433, 954, 15, 220, "Output",
      CellTags->"b:15.2"],
    Cell[259506, 5453, 205, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[259714, 5459, 954, 15, 220, "Output",
      CellTags->"b:15.2"],
    Cell[260683, 5477, 170, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[260856, 5483, 317, 14, 205, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[261176, 5499, 199, 7, 65, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[261400, 5510, 223, 6, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[261626, 5518, 114, 3, 37, "Print",
      CellTags->"b:15.2"],
    Cell[261743, 5523, 116, 3, 37, "Print",
      CellTags->"b:15.2"],
    Cell[261862, 5528, 116, 3, 37, "Print",
      CellTags->"b:15.2"],
    Cell[261981, 5533, 117, 3, 37, "Print",
      CellTags->"b:15.2"],
    Cell[262101, 5538, 112, 3, 37, "Print",
      CellTags->"b:15.2"],
    Cell[262216, 5543, 115, 3, 37, "Print",
      CellTags->"b:15.2"],
    Cell[262346, 5549, 188, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[262559, 5557, 148, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[262710, 5563, 8542, 142, 1484, "Output",
      CellTags->"b:15.2"],
    Cell[271289, 5710, 161, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[271453, 5716, 6645, 110, 529, "Output",
      CellTags->"b:15.2"],
    Cell[278135, 5831, 149, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[278287, 5837, 2224, 36, 298, "Output",
      CellTags->"b:15.2"],
    Cell[280526, 5876, 288, 7, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[280817, 5885, 189, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[281009, 5891, 194, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[281228, 5899, 166, 4, 45, "Input",
      CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
    Cell[281397, 5905, 1319, 27, 317, "Output",
      CellTags->"b:15.2"]},
  "Tractor-Trailer"->{
    Cell[282777, 5939, 174, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[282976, 5947, 176, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[283155, 5953, 549, 11, 98, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[283707, 5966, 125314, 1550, 554, 125169, 1546, "GraphicsData", 
    "Metafile", "Graphics",
      CellTags->{"b:15.3", "Tractor-Trailer"}],
    Cell[409058, 7521, 175, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[409236, 7527, 173, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[409412, 7533, 259, 9, 105, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[409674, 7544, 298, 9, 39, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[409975, 7555, 324, 8, 56, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[410302, 7565, 399, 15, 225, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[410704, 7582, 292, 9, 39, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[410999, 7593, 306, 10, 125, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[411342, 7608, 176, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[411543, 7616, 179, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[412235, 7645, 151, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[412539, 7660, 166, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[412708, 7666, 621, 15, 239, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[413332, 7683, 1267, 38, 645, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[414636, 7726, 168, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[414807, 7732, 486, 10, 73, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[415318, 7746, 376, 14, 185, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[425969, 7917, 150, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[426238, 7929, 195, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[426458, 7937, 148, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[427049, 7954, 361, 11, 145, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[427435, 7969, 235, 7, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[432306, 8060, 151, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[432589, 8072, 369, 11, 145, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[432961, 8085, 383, 12, 165, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[433347, 8099, 384, 12, 165, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[433734, 8113, 382, 12, 165, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[434119, 8127, 371, 12, 165, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[434493, 8141, 368, 12, 165, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[434864, 8155, 367, 12, 165, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[435234, 8169, 168, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[435439, 8178, 169, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[435633, 8186, 272, 8, 65, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[436273, 8212, 218, 6, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[436494, 8220, 410, 14, 205, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[436941, 8239, 196, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[437140, 8245, 514, 10, 73, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[437679, 8259, 921, 28, 465, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[439655, 8332, 148, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[439940, 8346, 149, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}]},
  "i:6"->{
    Cell[282777, 5939, 174, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[282976, 5947, 176, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[283155, 5953, 549, 11, 98, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[409058, 7521, 175, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[409236, 7527, 173, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[409412, 7533, 259, 9, 105, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[409674, 7544, 298, 9, 39, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[409975, 7555, 324, 8, 56, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[410302, 7565, 399, 15, 225, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[410704, 7582, 292, 9, 39, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[410999, 7593, 306, 10, 125, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[411342, 7608, 176, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[411543, 7616, 179, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[412235, 7645, 151, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[412539, 7660, 166, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[412708, 7666, 621, 15, 239, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[413332, 7683, 1267, 38, 645, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[414636, 7726, 168, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[414807, 7732, 486, 10, 73, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[415318, 7746, 376, 14, 185, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[425969, 7917, 150, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[426238, 7929, 195, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[426458, 7937, 148, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[427049, 7954, 361, 11, 145, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[427435, 7969, 235, 7, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[432306, 8060, 151, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[432589, 8072, 369, 11, 145, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[432961, 8085, 383, 12, 165, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[433347, 8099, 384, 12, 165, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[433734, 8113, 382, 12, 165, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[434119, 8127, 371, 12, 165, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[434493, 8141, 368, 12, 165, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[434864, 8155, 367, 12, 165, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[435234, 8169, 168, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[435439, 8178, 169, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[435633, 8186, 272, 8, 65, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[436273, 8212, 218, 6, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[436494, 8220, 410, 14, 205, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[436941, 8239, 196, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[437140, 8245, 514, 10, 73, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[437679, 8259, 921, 28, 465, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[439655, 8332, 148, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[439940, 8346, 149, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}]},
  "b:15.3"->{
    Cell[282777, 5939, 174, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[282976, 5947, 176, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[283155, 5953, 549, 11, 98, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[283707, 5966, 125314, 1550, 554, 125169, 1546, "GraphicsData", 
    "Metafile", "Graphics",
      CellTags->{"b:15.3", "Tractor-Trailer"}],
    Cell[409058, 7521, 175, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[409236, 7527, 173, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[409412, 7533, 259, 9, 105, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[409674, 7544, 298, 9, 39, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[409975, 7555, 324, 8, 56, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[410302, 7565, 399, 15, 225, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[410704, 7582, 292, 9, 39, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[410999, 7593, 306, 10, 125, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[411342, 7608, 176, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[411543, 7616, 179, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[411725, 7622, 116, 3, 37, "Print",
      CellTags->"b:15.3"],
    Cell[411844, 7627, 116, 3, 37, "Print",
      CellTags->"b:15.3"],
    Cell[411963, 7632, 116, 3, 37, "Print",
      CellTags->"b:15.3"],
    Cell[412082, 7637, 116, 3, 37, "Print",
      CellTags->"b:15.3"],
    Cell[412235, 7645, 151, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[412389, 7651, 101, 3, 37, "Print",
      CellTags->"b:15.3"],
    Cell[412539, 7660, 166, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[412708, 7666, 621, 15, 239, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[413332, 7683, 1267, 38, 645, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[414636, 7726, 168, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[414807, 7732, 486, 10, 73, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[415318, 7746, 376, 14, 185, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[415697, 7762, 10235, 150, 2267, "Output",
      CellTags->"b:15.3"],
    Cell[425969, 7917, 150, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[426122, 7923, 101, 3, 37, "Print",
      CellTags->"b:15.3"],
    Cell[426238, 7929, 195, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[426458, 7937, 148, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[426609, 7943, 425, 8, 122, "Output",
      CellTags->"b:15.3"],
    Cell[427049, 7954, 361, 11, 145, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[427435, 7969, 235, 7, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[427673, 7978, 4596, 77, 402, "Output",
      CellTags->"b:15.3"],
    Cell[432306, 8060, 151, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[432460, 8066, 114, 3, 37, "Print",
      CellTags->"b:15.3"],
    Cell[432589, 8072, 369, 11, 145, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[432961, 8085, 383, 12, 165, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[433347, 8099, 384, 12, 165, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[433734, 8113, 382, 12, 165, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[434119, 8127, 371, 12, 165, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[434493, 8141, 368, 12, 165, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[434864, 8155, 367, 12, 165, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[435234, 8169, 168, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[435439, 8178, 169, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[435633, 8186, 272, 8, 65, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[435908, 8196, 117, 3, 37, "Print",
      CellTags->"b:15.3"],
    Cell[436028, 8201, 112, 3, 37, "Print",
      CellTags->"b:15.3"],
    Cell[436143, 8206, 115, 3, 37, "Print",
      CellTags->"b:15.3"],
    Cell[436273, 8212, 218, 6, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[436494, 8220, 410, 14, 205, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[436941, 8239, 196, 4, 49, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[437140, 8245, 514, 10, 73, "Text",
      Evaluatable->False,
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[437679, 8259, 921, 28, 465, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[438603, 8289, 113, 3, 37, "Print",
      CellTags->"b:15.3"],
    Cell[438719, 8294, 133, 3, 37, "Print",
      CellTags->"b:15.3"],
    Cell[438855, 8299, 132, 3, 37, "Print",
      CellTags->"b:15.3"],
    Cell[438990, 8304, 120, 3, 37, "Print",
      CellTags->"b:15.3"],
    Cell[439113, 8309, 128, 3, 37, "Print",
      CellTags->"b:15.3"],
    Cell[439244, 8314, 123, 3, 37, "Print",
      CellTags->"b:15.3"],
    Cell[439370, 8319, 122, 3, 37, "Print",
      CellTags->"b:15.3"],
    Cell[439495, 8324, 123, 3, 37, "Print",
      CellTags->"b:15.3"],
    Cell[439655, 8332, 148, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[439806, 8338, 97, 3, 41, "Output",
      CellTags->"b:15.3"],
    Cell[439940, 8346, 149, 4, 45, "Input",
      CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
    Cell[440092, 8352, 128, 3, 41, "Output",
      CellTags->"b:15.3"]},
  "Constraints Intro"->{
    Cell[440430, 8370, 169, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Constraints Intro", "b:16.0"}],
    Cell[440602, 8376, 2531, 41, 427, "Text",
      Evaluatable->False,
      CellTags->{"Constraints Intro", "b:16.0"}]},
  "b:16.0"->{
    Cell[440430, 8370, 169, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Constraints Intro", "b:16.0"}],
    Cell[440602, 8376, 2531, 41, 427, "Text",
      Evaluatable->False,
      CellTags->{"Constraints Intro", "b:16.0"}]},
  "Closed Chain"->{
    Cell[443170, 8422, 172, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[443345, 8428, 416, 9, 56, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[443764, 8439, 323, 14, 205, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[444090, 8455, 521, 13, 185, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[444614, 8470, 187, 7, 65, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[444826, 8481, 215, 6, 45, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[445764, 8520, 267, 7, 65, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[446034, 8529, 173, 4, 45, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[446232, 8537, 156, 4, 45, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[446603, 8551, 455, 9, 56, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[447083, 8564, 208, 6, 45, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[448735, 8616, 491, 10, 73, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[449251, 8630, 211, 6, 45, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[450223, 8655, 223, 6, 39, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[450471, 8665, 177, 4, 45, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[452134, 8707, 330, 8, 56, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[452489, 8719, 185, 4, 45, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[453886, 8759, 289, 7, 39, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[456194, 8833, 171, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[456368, 8839, 398, 9, 56, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[456769, 8850, 330, 14, 205, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[457102, 8866, 517, 13, 185, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[457622, 8881, 189, 7, 65, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[457836, 8892, 215, 6, 45, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[458774, 8931, 267, 7, 65, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[459044, 8940, 173, 4, 45, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[459242, 8948, 156, 4, 45, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[459607, 8962, 245, 7, 39, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[459877, 8973, 212, 6, 45, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[463086, 9056, 156, 4, 45, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}]},
  "i:7"->{
    Cell[443170, 8422, 172, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[443345, 8428, 416, 9, 56, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[443764, 8439, 323, 14, 205, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[444090, 8455, 521, 13, 185, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[444614, 8470, 187, 7, 65, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[444826, 8481, 215, 6, 45, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[445764, 8520, 267, 7, 65, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[446034, 8529, 173, 4, 45, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[446232, 8537, 156, 4, 45, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[446603, 8551, 455, 9, 56, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[447083, 8564, 208, 6, 45, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[448735, 8616, 491, 10, 73, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[449251, 8630, 211, 6, 45, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[450223, 8655, 223, 6, 39, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[450471, 8665, 177, 4, 45, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[452134, 8707, 330, 8, 56, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[452489, 8719, 185, 4, 45, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[453886, 8759, 289, 7, 39, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:7", "b:16.1"}]},
  "b:16.1"->{
    Cell[443170, 8422, 172, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[443345, 8428, 416, 9, 56, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[443764, 8439, 323, 14, 205, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[444090, 8455, 521, 13, 185, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[444614, 8470, 187, 7, 65, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[444826, 8481, 215, 6, 45, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[445044, 8489, 114, 3, 37, "Print",
      CellTags->"b:16.1"],
    Cell[445161, 8494, 116, 3, 37, "Print",
      CellTags->"b:16.1"],
    Cell[445280, 8499, 116, 3, 37, "Print",
      CellTags->"b:16.1"],
    Cell[445399, 8504, 117, 3, 37, "Print",
      CellTags->"b:16.1"],
    Cell[445519, 8509, 112, 3, 37, "Print",
      CellTags->"b:16.1"],
    Cell[445634, 8514, 115, 3, 37, "Print",
      CellTags->"b:16.1"],
    Cell[445764, 8520, 267, 7, 65, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[446034, 8529, 173, 4, 45, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[446232, 8537, 156, 4, 45, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[446391, 8543, 197, 5, 41, "Output",
      CellTags->"b:16.1"],
    Cell[446603, 8551, 455, 9, 56, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[447083, 8564, 208, 6, 45, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[447294, 8572, 90, 3, 37, "Print",
      CellTags->"b:16.1"],
    Cell[447387, 8577, 90, 3, 37, "Print",
      CellTags->"b:16.1"],
    Cell[447480, 8582, 90, 3, 37, "Print",
      CellTags->"b:16.1"],
    Cell[447573, 8587, 90, 3, 37, "Print",
      CellTags->"b:16.1"],
    Cell[447666, 8592, 1054, 21, 233, "Output",
      CellTags->"b:16.1"],
    Cell[448735, 8616, 491, 10, 73, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[449251, 8630, 211, 6, 45, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[449465, 8638, 743, 14, 171, "Output",
      CellTags->"b:16.1"],
    Cell[450223, 8655, 223, 6, 39, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[450471, 8665, 177, 4, 45, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[450651, 8671, 1468, 33, 170, "Output",
      CellTags->"b:16.1"],
    Cell[452134, 8707, 330, 8, 56, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[452489, 8719, 185, 4, 45, "Input",
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[452677, 8725, 1194, 31, 104, "Output",
      CellTags->"b:16.1"],
    Cell[453886, 8759, 289, 7, 39, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:7", "b:16.1"}],
    Cell[454200, 8770, 175, 5, 43, "Input",
      CellTags->"b:16.1"],
    Cell[454378, 8777, 768, 21, 57, "Output",
      CellTags->"b:16.1"],
    Cell[455183, 8803, 205, 5, 43, "Input",
      CellTags->"b:16.1"],
    Cell[455391, 8810, 754, 17, 74, "Output",
      CellTags->"b:16.1"]},
  "i:8"->{
    Cell[456194, 8833, 171, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[456368, 8839, 398, 9, 56, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[456769, 8850, 330, 14, 205, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[457102, 8866, 517, 13, 185, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[457622, 8881, 189, 7, 65, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[457836, 8892, 215, 6, 45, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[458774, 8931, 267, 7, 65, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[459044, 8940, 173, 4, 45, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[459242, 8948, 156, 4, 45, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[459607, 8962, 245, 7, 39, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[459877, 8973, 212, 6, 45, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[463086, 9056, 156, 4, 45, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}]},
  "b:16.2"->{
    Cell[456194, 8833, 171, 4, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[456368, 8839, 398, 9, 56, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[456769, 8850, 330, 14, 205, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[457102, 8866, 517, 13, 185, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[457622, 8881, 189, 7, 65, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[457836, 8892, 215, 6, 45, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[458054, 8900, 114, 3, 37, "Print",
      CellTags->"b:16.2"],
    Cell[458171, 8905, 116, 3, 37, "Print",
      CellTags->"b:16.2"],
    Cell[458290, 8910, 116, 3, 37, "Print",
      CellTags->"b:16.2"],
    Cell[458409, 8915, 117, 3, 37, "Print",
      CellTags->"b:16.2"],
    Cell[458529, 8920, 112, 3, 37, "Print",
      CellTags->"b:16.2"],
    Cell[458644, 8925, 115, 3, 37, "Print",
      CellTags->"b:16.2"],
    Cell[458774, 8931, 267, 7, 65, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[459044, 8940, 173, 4, 45, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[459242, 8948, 156, 4, 45, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[459401, 8954, 191, 5, 41, "Output",
      CellTags->"b:16.2"],
    Cell[459607, 8962, 245, 7, 39, "Text",
      Evaluatable->False,
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[459877, 8973, 212, 6, 45, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[460092, 8981, 90, 3, 37, "Print",
      CellTags->"b:16.2"],
    Cell[460185, 8986, 90, 3, 37, "Print",
      CellTags->"b:16.2"],
    Cell[460278, 8991, 90, 3, 37, "Print",
      CellTags->"b:16.2"],
    Cell[460371, 8996, 90, 3, 37, "Print",
      CellTags->"b:16.2"],
    Cell[460464, 9001, 122, 3, 37, "Print",
      CellTags->"b:16.2"],
    Cell[460589, 9006, 2460, 45, 441, "Output",
      CellTags->"b:16.2"],
    Cell[463086, 9056, 156, 4, 45, "Input",
      CellTags->{"Closed Chain", "i:8", "b:16.2"}],
    Cell[463245, 9062, 671, 13, 151, "Output",
      CellTags->"b:16.2"]},
  "Rolling Disk"->{
    Cell[463965, 9081, 222, 6, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[464190, 9089, 629, 12, 73, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[464822, 9103, 45941, 570, 234, 45809, 566, "GraphicsData", 
    "Metafile", "Graphics",
      CellTags->{"b:16.3", "Rolling Disk"}],
    Cell[510788, 9677, 173, 7, 65, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[511190, 9693, 218, 9, 105, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[511411, 9704, 333, 9, 64, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[511747, 9715, 222, 7, 65, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[511972, 9724, 324, 8, 85, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[512299, 9734, 168, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[512470, 9740, 189, 7, 65, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[512662, 9749, 190, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[512877, 9757, 215, 6, 45, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[513696, 9791, 499, 10, 61, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[514220, 9805, 213, 7, 57, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[514644, 9821, 357, 8, 48, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[515026, 9833, 237, 7, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[515650, 9863, 289, 7, 35, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[515964, 9874, 143, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[516219, 9886, 626, 12, 74, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[516870, 9902, 190, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[517204, 9914, 348, 8, 48, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[517555, 9924, 162, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[517720, 9930, 179, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[517902, 9936, 186, 4, 35, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[518113, 9944, 191, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[523783, 10071, 243, 7, 35, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[524029, 10080, 224, 7, 57, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[524278, 10091, 232, 7, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[524693, 10107, 461, 10, 61, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[525179, 10121, 241, 8, 70, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[525703, 10142, 235, 7, 35, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[525963, 10153, 192, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[526392, 10167, 232, 7, 35, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[526649, 10178, 177, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[526983, 10192, 199, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[529207, 10253, 516, 10, 61, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[529748, 10267, 209, 6, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[530889, 10303, 258, 8, 70, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[533132, 10365, 232, 8, 70, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[534147, 10399, 227, 8, 70, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[535148, 10433, 214, 6, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[538677, 10518, 238, 7, 35, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}]},
  "i:9"->{
    Cell[463965, 9081, 222, 6, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[464190, 9089, 629, 12, 73, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[510788, 9677, 173, 7, 65, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[511190, 9693, 218, 9, 105, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[511411, 9704, 333, 9, 64, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[511747, 9715, 222, 7, 65, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[511972, 9724, 324, 8, 85, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[512299, 9734, 168, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[512470, 9740, 189, 7, 65, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[512662, 9749, 190, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[512877, 9757, 215, 6, 45, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[513696, 9791, 499, 10, 61, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[514220, 9805, 213, 7, 57, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[514644, 9821, 357, 8, 48, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[515026, 9833, 237, 7, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[515650, 9863, 289, 7, 35, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[515964, 9874, 143, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[516219, 9886, 626, 12, 74, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[516870, 9902, 190, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[517204, 9914, 348, 8, 48, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[517555, 9924, 162, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[517720, 9930, 179, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[517902, 9936, 186, 4, 35, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[518113, 9944, 191, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[523783, 10071, 243, 7, 35, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[524029, 10080, 224, 7, 57, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[524278, 10091, 232, 7, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[524693, 10107, 461, 10, 61, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[525179, 10121, 241, 8, 70, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[525703, 10142, 235, 7, 35, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[525963, 10153, 192, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[526392, 10167, 232, 7, 35, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[526649, 10178, 177, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[526983, 10192, 199, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[529207, 10253, 516, 10, 61, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[529748, 10267, 209, 6, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[530889, 10303, 258, 8, 70, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[533132, 10365, 232, 8, 70, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[534147, 10399, 227, 8, 70, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[535148, 10433, 214, 6, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[538677, 10518, 238, 7, 35, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}]},
  "b:16.3"->{
    Cell[463965, 9081, 222, 6, 53, "Subsection",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[464190, 9089, 629, 12, 73, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[464822, 9103, 45941, 570, 234, 45809, 566, "GraphicsData", 
    "Metafile", "Graphics",
      CellTags->{"b:16.3", "Rolling Disk"}],
    Cell[510788, 9677, 173, 7, 65, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[510964, 9686, 211, 4, 60, "Output",
      CellTags->"b:16.3"],
    Cell[511190, 9693, 218, 9, 105, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[511411, 9704, 333, 9, 64, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[511747, 9715, 222, 7, 65, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[511972, 9724, 324, 8, 85, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[512299, 9734, 168, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[512470, 9740, 189, 7, 65, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[512662, 9749, 190, 4, 39, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[512877, 9757, 215, 6, 45, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[513095, 9765, 114, 3, 37, "Print",
      CellTags->"b:16.3"],
    Cell[513212, 9770, 116, 3, 29, "Print",
      CellTags->"b:16.3"],
    Cell[513331, 9775, 117, 3, 29, "Print",
      CellTags->"b:16.3"],
    Cell[513451, 9780, 112, 3, 29, "Print",
      CellTags->"b:16.3"],
    Cell[513566, 9785, 115, 3, 29, "Print",
      CellTags->"b:16.3"],
    Cell[513696, 9791, 499, 10, 61, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[514220, 9805, 213, 7, 57, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[514436, 9814, 193, 4, 43, "Output",
      CellTags->"b:16.3"],
    Cell[514644, 9821, 357, 8, 48, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[515026, 9833, 237, 7, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[515266, 9842, 90, 3, 29, "Print",
      CellTags->"b:16.3"],
    Cell[515359, 9847, 90, 3, 29, "Print",
      CellTags->"b:16.3"],
    Cell[515452, 9852, 90, 3, 29, "Print",
      CellTags->"b:16.3"],
    Cell[515545, 9857, 90, 3, 29, "Print",
      CellTags->"b:16.3"],
    Cell[515650, 9863, 289, 7, 35, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[515964, 9874, 143, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[516110, 9880, 94, 3, 43, "Output",
      CellTags->"b:16.3"],
    Cell[516219, 9886, 626, 12, 74, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[516870, 9902, 190, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[517063, 9908, 126, 3, 43, "Output",
      CellTags->"b:16.3"],
    Cell[517204, 9914, 348, 8, 48, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[517555, 9924, 162, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[517720, 9930, 179, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[517902, 9936, 186, 4, 35, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[518113, 9944, 191, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[518307, 9950, 5461, 118, 345, "Output",
      CellTags->"b:16.3"],
    Cell[523783, 10071, 243, 7, 35, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[524029, 10080, 224, 7, 57, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[524278, 10091, 232, 7, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[524513, 10100, 165, 4, 43, "Output",
      CellTags->"b:16.3"],
    Cell[524693, 10107, 461, 10, 61, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[525179, 10121, 241, 8, 70, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[525423, 10131, 131, 3, 43, "Output",
      CellTags->"b:16.3"],
    Cell[525557, 10136, 131, 3, 43, "Output",
      CellTags->"b:16.3"],
    Cell[525703, 10142, 235, 7, 35, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[525963, 10153, 192, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[526158, 10159, 219, 5, 56, "Output",
      CellTags->"b:16.3"],
    Cell[526392, 10167, 232, 7, 35, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[526649, 10178, 177, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[526829, 10184, 117, 3, 43, "Output",
      CellTags->"b:16.3"],
    Cell[526983, 10192, 199, 4, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[527185, 10198, 2007, 52, 136, "Output",
      CellTags->"b:16.3"],
    Cell[529207, 10253, 516, 10, 61, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[529748, 10267, 209, 6, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[529960, 10275, 892, 23, 56, "Output",
      CellTags->"b:16.3"],
    Cell[530889, 10303, 258, 8, 70, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[531150, 10313, 1945, 47, 56, "Output",
      CellTags->"b:16.3"],
    Cell[533132, 10365, 232, 8, 70, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[533367, 10375, 743, 19, 47, "Output",
      CellTags->"b:16.3"],
    Cell[534147, 10399, 227, 8, 70, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[534377, 10409, 734, 19, 50, "Output",
      CellTags->"b:16.3"],
    Cell[535148, 10433, 214, 6, 44, "Input",
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
    Cell[535365, 10441, 3297, 74, 116, "Output",
      CellTags->"b:16.3"],
    Cell[538677, 10518, 238, 7, 35, "Text",
      Evaluatable->False,
      CellTags->{"Rolling Disk", "i:9", "b:16.3"}]},
  "Drive Chain"->{
    Cell[538952, 10530, 162, 4, 54, "Subsection",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[539139, 10538, 157, 4, 34, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[539299, 10544, 314, 8, 48, "Text",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[539650, 10557, 155, 4, 34, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[539808, 10563, 152, 4, 35, "Text",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[539963, 10569, 287, 10, 96, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[540253, 10581, 158, 4, 35, "Text",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[540414, 10587, 325, 10, 83, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[540742, 10599, 154, 4, 35, "Text",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[540899, 10605, 301, 9, 83, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[541203, 10616, 171, 4, 44, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[541411, 10625, 156, 4, 34, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[541570, 10631, 280, 9, 37, "Text",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[541853, 10642, 216, 8, 70, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[542072, 10652, 288, 9, 37, "Text",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[542363, 10663, 238, 8, 70, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[542604, 10673, 496, 10, 61, "Text",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[543103, 10685, 274, 12, 122, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[543414, 10702, 161, 4, 34, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[543578, 10708, 180, 4, 35, "Text",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[543761, 10714, 161, 4, 44, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[543959, 10723, 174, 4, 34, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[544136, 10729, 150, 4, 44, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[544311, 10737, 156, 4, 44, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[544861, 10761, 210, 6, 44, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[545473, 10788, 176, 4, 34, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[545652, 10794, 489, 10, 61, "Text",
      Evaluatable->False,
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[546144, 10806, 376, 9, 83, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[546523, 10817, 317, 8, 70, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[546843, 10827, 287, 8, 70, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[547133, 10837, 144, 4, 44, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[547302, 10845, 246, 7, 44, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[548132, 10882, 141, 4, 44, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[548626, 10899, 141, 4, 44, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[549407, 10920, 141, 4, 44, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}]},
  "b:16.4"->{
    Cell[538952, 10530, 162, 4, 54, "Subsection",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[539139, 10538, 157, 4, 34, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[539299, 10544, 314, 8, 48, "Text",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[539650, 10557, 155, 4, 34, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[539808, 10563, 152, 4, 35, "Text",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[539963, 10569, 287, 10, 96, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[540253, 10581, 158, 4, 35, "Text",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[540414, 10587, 325, 10, 83, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[540742, 10599, 154, 4, 35, "Text",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[540899, 10605, 301, 9, 83, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[541203, 10616, 171, 4, 44, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[541411, 10625, 156, 4, 34, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[541570, 10631, 280, 9, 37, "Text",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[541853, 10642, 216, 8, 70, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[542072, 10652, 288, 9, 37, "Text",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[542363, 10663, 238, 8, 70, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[542604, 10673, 496, 10, 61, "Text",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[543103, 10685, 274, 12, 122, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[543414, 10702, 161, 4, 34, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[543578, 10708, 180, 4, 35, "Text",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[543761, 10714, 161, 4, 44, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[543959, 10723, 174, 4, 34, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[544136, 10729, 150, 4, 44, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[544311, 10737, 156, 4, 44, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[544470, 10743, 116, 3, 29, "Print",
      CellTags->"b:16.4"],
    Cell[544589, 10748, 116, 3, 29, "Print",
      CellTags->"b:16.4"],
    Cell[544708, 10753, 116, 3, 29, "Print",
      CellTags->"b:16.4"],
    Cell[544861, 10761, 210, 6, 44, "Input",
      CellTags->{"Drive Chain", "b:16.4"}],
    Cell[545074, 10769, 117, 3, 29, "Print",
      CellTags->"b:16.4"],
    Cell[545194, 10774, 112, 3, 29, "Print",
      CellTags->"b:16.4"],
    Cell[545309, 10779, 115, 3, 29, "Print",
      CellTags->"b:16.4"],
    Cell[545473, 10788, 176, 4, 34, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[545652, 10794, 489, 10, 61, "Text",
      Evaluatable->False,
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[546144, 10806, 376, 9, 83, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[546523, 10817, 317, 8, 70, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[546843, 10827, 287, 8, 70, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[547133, 10837, 144, 4, 44, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[547302, 10845, 246, 7, 44, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[547551, 10854, 100, 3, 29, "Print",
      CellTags->{"i:10", "b:16.4"}],
    Cell[547654, 10859, 100, 3, 29, "Print",
      CellTags->{"i:10", "b:16.4"}],
    Cell[547757, 10864, 100, 3, 29, "Print",
      CellTags->{"i:10", "b:16.4"}],
    Cell[547860, 10869, 100, 3, 29, "Print",
      CellTags->{"i:10", "b:16.4"}],
    Cell[547963, 10874, 132, 3, 29, "Print",
      CellTags->{"i:10", "b:16.4"}],
    Cell[548132, 10882, 141, 4, 44, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[548276, 10888, 313, 6, 56, "Output",
      CellTags->{"i:10", "b:16.4"}],
    Cell[548626, 10899, 141, 4, 44, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[548770, 10905, 600, 10, 82, "Output",
      CellTags->{"i:10", "b:16.4"}],
    Cell[549407, 10920, 141, 4, 44, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[549551, 10926, 95, 3, 43, "Output",
      CellTags->{"i:10", "b:16.4"}]},
  "i:10"->{
    Cell[545473, 10788, 176, 4, 34, "Subsubsection",
      Evaluatable->False,
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[545652, 10794, 489, 10, 61, "Text",
      Evaluatable->False,
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[546144, 10806, 376, 9, 83, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[546523, 10817, 317, 8, 70, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[546843, 10827, 287, 8, 70, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[547133, 10837, 144, 4, 44, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[547302, 10845, 246, 7, 44, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[547551, 10854, 100, 3, 29, "Print",
      CellTags->{"i:10", "b:16.4"}],
    Cell[547654, 10859, 100, 3, 29, "Print",
      CellTags->{"i:10", "b:16.4"}],
    Cell[547757, 10864, 100, 3, 29, "Print",
      CellTags->{"i:10", "b:16.4"}],
    Cell[547860, 10869, 100, 3, 29, "Print",
      CellTags->{"i:10", "b:16.4"}],
    Cell[547963, 10874, 132, 3, 29, "Print",
      CellTags->{"i:10", "b:16.4"}],
    Cell[548132, 10882, 141, 4, 44, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[548276, 10888, 313, 6, 56, "Output",
      CellTags->{"i:10", "b:16.4"}],
    Cell[548626, 10899, 141, 4, 44, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[548770, 10905, 600, 10, 82, "Output",
      CellTags->{"i:10", "b:16.4"}],
    Cell[549407, 10920, 141, 4, 44, "Input",
      CellTags->{"Drive Chain", "i:10", "b:16.4"}],
    Cell[549551, 10926, 95, 3, 43, "Output",
      CellTags->{"i:10", "b:16.4"}]},
  "References"->{
    Cell[549719, 10937, 149, 4, 51, "Section",
      Evaluatable->False,
      CellTags->{"References", "b:17"}],
    Cell[549871, 10943, 1621, 30, 367, "Text",
      Evaluatable->False,
      CellTags->{"References", "b:17"}]},
  "b:17"->{
    Cell[549719, 10937, 149, 4, 51, "Section",
      Evaluatable->False,
      CellTags->{"References", "b:17"}],
    Cell[549871, 10943, 1621, 30, 367, "Text",
      Evaluatable->False,
      CellTags->{"References", "b:17"}]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"Introduction", 555079, 11037},
  {"b:11", 555314, 11044},
  {"Getting Started", 555756, 11057},
  {"b:12", 556282, 11071},
  {"What's in the Package", 556893, 11087},
  {"b:13.0", 557563, 11103},
  {"Info3248853457-1624643", 558583, 11127},
  {"Info3248853457-6771253", 558713, 11130},
  {"Data Structures", 558836, 11133},
  {"b:13.1", 561643, 11200},
  {"Building Equations", 565089, 11285},
  {"b:14.0", 567299, 11336},
  {"i:1", 570148, 11405},
  {"Simulation", 570271, 11408},
  {"i:2", 571705, 11441},
  {"b:14.1", 573142, 11474},
  {"Creating MEX-Files", 575572, 11533},
  {"i:3", 576375, 11551},
  {"b:14.2", 576985, 11565},
  {"b:14.3", 578306, 11599},
  {"b:15.0", 578517, 11604},
  {"Suspension", 578735, 11611},
  {"i:4", 581365, 11672},
  {"b:15.1", 583998, 11733},
  {"Flexible Manipulator", 587514, 11818},
  {"i:5", 591332, 11899},
  {"b:15.2", 595153, 11980},
  {"Tractor-Trailer", 599860, 12085},
  {"i:6", 604782, 12191},
  {"b:15.3", 609558, 12294},
  {"Constraints Intro", 616141, 12446},
  {"b:16.0", 616403, 12453},
  {"Closed Chain", 616671, 12460},
  {"i:7", 619834, 12531},
  {"b:16.1", 621763, 12575},
  {"i:8", 625054, 12657},
  {"b:16.2", 626306, 12685},
  {"Rolling Disk", 628562, 12741},
  {"i:9", 633106, 12844},
  {"b:16.3", 637511, 12944},
  {"Drive Chain", 643875, 13097},
  {"b:16.4", 647551, 13184},
  {"i:10", 652316, 13299},
  {"References", 654024, 13338},
  {"b:17", 654265, 13345}
  }
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 151, 4, 70, "Section",
  Evaluatable->False,
  CellTags->{"Introduction", "b:11"}],
Cell[1930, 59, 145, 4, 63, "Input",
  CellTags->"b:11"],

Cell[CellGroupData[{
Cell[2100, 67, 134, 3, 43, "Input",
  CellTags->"b:11"],
Cell[2237, 72, 121, 3, 41, "Output",
  CellTags->"b:11"]
}, Open  ]],
Cell[2373, 78, 1782, 30, 292, "Text",
  Evaluatable->False,
  CellTags->{"Introduction", "b:11"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[4192, 113, 157, 4, 70, "Section",
  Evaluatable->False,
  CellTags->{"Getting Started", "b:12"}],
Cell[4352, 119, 772, 15, 90, "Text",
  Evaluatable->False,
  CellTags->{"Getting Started", "b:12"}],
Cell[5127, 136, 256, 9, 39, "Text",
  Evaluatable->False,
  CellTags->{"Getting Started", "b:12"}],

Cell[CellGroupData[{
Cell[5408, 149, 140, 4, 45, "Input",
  CellTags->{"Getting Started", "b:12"}],
Cell[5551, 155, 709, 11, 417, "Print",
  CellTags->"b:12"]
}, Open  ]],
Cell[6275, 169, 172, 6, 39, "Text",
  CellTags->{"Getting Started", "b:12"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[6484, 180, 127, 3, 56, "Section",
  Evaluatable->False],

Cell[CellGroupData[{
Cell[6636, 187, 175, 4, 53, "Subsection",
  Evaluatable->False,
  CellTags->{"What's in the Package", "b:13.0"}],
Cell[6814, 193, 234, 7, 39, "Text",
  Evaluatable->False,
  CellTags->{"What's in the Package", "b:13.0"}],

Cell[CellGroupData[{
Cell[7073, 204, 147, 4, 45, "Input",
  CellTags->{"What's in the Package", "b:13.0"}],
Cell[7223, 210, 124, 3, 37, "Print",
  CellTags->"b:13.0"],
Cell[7350, 215, 88, 3, 41, "Output",
  CellTags->"b:13.0"]
}, Open  ]],

Cell[CellGroupData[{
Cell[7475, 223, 119, 3, 43, "Input",
  CellTags->{"What's in the Package", "b:13.0"}],
Cell[7597, 228, 1179, 19, 588, "Print",
  CellTags->{"Info3248853457-1624643", "b:13.0"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[8813, 252, 127, 3, 43, "Input",
  CellTags->{"What's in the Package", "b:13.0"}],
Cell[8943, 257, 1241, 18, 569, "Print",
  CellTags->{"Info3248853457-6771253", "b:13.0"}]
}, Open  ]],
Cell[10199, 278, 1892, 31, 341, "Text",
  Evaluatable->False,
  CellTags->{"What's in the Package", "b:13.0"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[12128, 314, 162, 4, 53, "Subsection",
  Evaluatable->False,
  CellTags->{"Data Structures", "b:13.1"}],

Cell[CellGroupData[{
Cell[12315, 322, 165, 4, 49, "Subsubsection",
  Evaluatable->False,
  CellTags->{"Data Structures", "b:13.1"}],
Cell[12483, 328, 363, 8, 56, "Text",
  Evaluatable->False,
  CellTags->{"Data Structures", "b:13.1"}],
Cell[12849, 338, 34187, 425, 319, 34048, 421, "GraphicsData", "Metafile", \
"Graphics",
  CellTags->{"b:13.1", "Data Structures"}],
Cell[47039, 765, 3093, 51, 609, "Text",
  Evaluatable->False,
  CellTags->{"Data Structures", "b:13.1"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[50169, 821, 84, 1, 49, "Subsubsection",
  CellTags->{"Data Structures", "b:13.1"}],
Cell[50256, 824, 1086, 16, 141, "Text",
  CellTags->{"Data Structures", "b:13.1"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[51379, 845, 156, 4, 49, "Subsubsection",
  Evaluatable->False,
  CellTags->{"Data Structures", "b:13.1"}],
Cell[51538, 851, 1941, 34, 391, "Text",
  Evaluatable->False,
  CellTags->{"Data Structures", "b:13.1"}],
Cell[53482, 887, 216, 7, 65, "Input",
  CellTags->{"Data Structures", "b:13.1"}],
Cell[53701, 896, 228, 7, 39, "Text",
  Evaluatable->False,
  CellTags->{"Data Structures", "b:13.1"}],
Cell[53932, 905, 219, 7, 65, "Input",
  CellTags->{"Data Structures", "b:13.1"}],
Cell[54154, 914, 396, 9, 56, "Text",
  Evaluatable->False,
  CellTags->{"Data Structures", "b:13.1"}],

Cell[CellGroupData[{
Cell[54575, 927, 208, 8, 85, "Input",
  CellTags->{"Data Structures", "b:13.1"}],
Cell[54786, 937, 115, 3, 37, "Print",
  CellTags->"b:13.1"],
Cell[54904, 942, 115, 3, 37, "Print",
  CellTags->"b:13.1"]
}, Open  ]],

Cell[CellGroupData[{
Cell[55056, 950, 142, 4, 45, "Input",
  CellTags->{"Data Structures", "b:13.1"}],
Cell[55201, 956, 128, 3, 37, "Print",
  CellTags->"b:13.1"]
}, Open  ]],

Cell[CellGroupData[{
Cell[55366, 964, 142, 4, 45, "Input",
  CellTags->{"Data Structures", "b:13.1"}],
Cell[55511, 970, 468, 8, 94, "Print",
  CellTags->"b:13.1"]
}, Open  ]],

Cell[CellGroupData[{
Cell[56016, 983, 142, 4, 45, "Input",
  CellTags->{"Data Structures", "b:13.1"}],
Cell[56161, 989, 210, 4, 56, "Print",
  CellTags->"b:13.1"]
}, Open  ]],
Cell[56386, 996, 636, 12, 90, "Text",
  Evaluatable->False,
  CellTags->{"Data Structures", "b:13.1"}],

Cell[CellGroupData[{
Cell[57047, 1012, 270, 11, 145, "Input",
  CellTags->{"Data Structures", "b:13.1"}],
Cell[57320, 1025, 115, 3, 37, "Print",
  CellTags->"b:13.1"]
}, Open  ]],

Cell[CellGroupData[{
Cell[57472, 1033, 142, 4, 45, "Input",
  CellTags->{"Data Structures", "b:13.1"}],
Cell[57617, 1039, 635, 12, 113, "Print",
  CellTags->"b:13.1"]
}, Open  ]],

Cell[CellGroupData[{
Cell[58289, 1056, 142, 4, 45, "Input",
  CellTags->{"Data Structures", "b:13.1"}],
Cell[58434, 1062, 461, 9, 94, "Print",
  CellTags->"b:13.1"]
}, Open  ]],

Cell[CellGroupData[{
Cell[58932, 1076, 142, 4, 45, "Input",
  CellTags->{"Data Structures", "b:13.1"}],
Cell[59077, 1082, 218, 4, 56, "Print",
  CellTags->"b:13.1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[59344, 1092, 162, 4, 49, "Subsubsection",
  Evaluatable->False,
  CellTags->{"Data Structures", "b:13.1"}],
Cell[59509, 1098, 1358, 26, 340, "Text",
  Evaluatable->False,
  CellTags->{"Data Structures", "b:13.1"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[60904, 1129, 165, 4, 49, "Subsubsection",
  Evaluatable->False,
  CellTags->{"Data Structures", "b:13.1"}],
Cell[61072, 1135, 2927, 49, 625, "Text",
  Evaluatable->False,
  CellTags->{"Data Structures", "b:13.1"}]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[64060, 1191, 137, 3, 56, "Section",
  Evaluatable->False],

Cell[CellGroupData[{
Cell[64222, 1198, 182, 4, 53, "Subsection",
  Evaluatable->False,
  CellTags->{"Building Equations", "b:14.0"}],
Cell[64407, 1204, 58025, 720, 218, 57887, 716, "GraphicsData", "Metafile", \
"Graphics",
  CellTags->{"b:14.0", "Building Equations"}],
Cell[122435, 1926, 446, 13, 214, "Text",
  Evaluatable->False,
  CellTags->{"Building Equations", "b:14.0"}],
Cell[122884, 1941, 305, 12, 165, "Input",
  CellTags->{"Building Equations", "i:1", "b:14.0"}],
Cell[123192, 1955, 160, 4, 39, "Text",
  Evaluatable->False,
  CellTags->{"Building Equations", "b:14.0"}],
Cell[123355, 1961, 375, 12, 165, "Input",
  CellTags->{"Building Equations", "b:14.0"}],
Cell[123733, 1975, 176, 4, 39, "Text",
  Evaluatable->False,
  CellTags->{"Building Equations", "b:14.0"}],
Cell[123912, 1981, 161, 4, 45, "Input",
  CellTags->{"Building Equations", "b:14.0"}],
Cell[124076, 1987, 315, 8, 64, "Text",
  Evaluatable->False,
  CellTags->{"Building Equations", "b:14.0"}],
Cell[124394, 1997, 159, 4, 45, "Input",
  CellTags->{"Building Equations", "b:14.0"}],

Cell[CellGroupData[{
Cell[124578, 2005, 210, 6, 45, "Input",
  CellTags->{"Building Equations", "b:14.0"}],
Cell[124791, 2013, 113, 3, 37, "Print",
  CellTags->"b:14.0"],
Cell[124907, 2018, 115, 3, 37, "Print",
  CellTags->"b:14.0"],
Cell[125025, 2023, 115, 3, 37, "Print",
  CellTags->"b:14.0"],
Cell[125143, 2028, 116, 3, 37, "Print",
  CellTags->"b:14.0"],
Cell[125262, 2033, 111, 3, 37, "Print",
  CellTags->"b:14.0"],
Cell[125376, 2038, 114, 3, 37, "Print",
  CellTags->"b:14.0"]
}, Open  ]],
Cell[125505, 2044, 172, 4, 39, "Text",
  Evaluatable->False,
  CellTags->{"Building Equations", "b:14.0"}],

Cell[CellGroupData[{
Cell[125702, 2052, 157, 4, 45, "Input",
  CellTags->{"Building Equations", "b:14.0"}],
Cell[125862, 2058, 487, 14, 51, "Print",
  CellTags->"b:14.0"]
}, Open  ]],
Cell[126364, 2075, 384, 9, 56, "Text",
  Evaluatable->False,
  CellTags->{"Building Equations", "b:14.0"}],

Cell[CellGroupData[{
Cell[126773, 2088, 157, 4, 45, "Input",
  CellTags->{"Building Equations", "b:14.0"}],
Cell[126933, 2094, 523, 12, 55, "Print",
  CellTags->"b:14.0"]
}, Open  ]],

Cell[CellGroupData[{
Cell[127493, 2111, 158, 4, 45, "Input",
  CellTags->{"Building Equations", "b:14.0"}],
Cell[127654, 2117, 775, 25, 49, "Print",
  CellTags->"b:14.0"]
}, Open  ]],
Cell[128444, 2145, 446, 9, 56, "Text",
  Evaluatable->False,
  CellTags->{"Building Equations", "b:14.0"}],
Cell[128893, 2156, 305, 9, 85, "Input",
  CellTags->{"Building Equations", "b:14.0"}],
Cell[129201, 2167, 192, 4, 39, "Text",
  Evaluatable->False,
  CellTags->{"Building Equations", "b:14.0"}],
Cell[129396, 2173, 195, 4, 45, "Input",
  CellTags->{"Building Equations", "b:14.0"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[129628, 2182, 179, 4, 53, "Subsection",
  Evaluatable->False,
  CellTags->{"Simulation", "i:2", "b:14.1"}],
Cell[129810, 2188, 362, 8, 56, "Text",
  Evaluatable->False,
  CellTags->{"Simulation", "i:2", "b:14.1"}],

Cell[CellGroupData[{
Cell[130197, 2200, 171, 4, 45, "Input",
  CellTags->{"Simulation", "i:2", "b:14.1"}],
Cell[130371, 2206, 2480, 56, 218, "Output",
  CellTags->"b:14.1"]
}, Open  ]],
Cell[132866, 2265, 223, 6, 39, "Text",
  Evaluatable->False,
  CellTags->{"Simulation", "i:2", "b:14.1"}],

Cell[CellGroupData[{
Cell[133114, 2275, 196, 4, 45, "Input",
  CellTags->{"Simulation", "i:2", "b:14.1"}],
Cell[133313, 2281, 186, 4, 41, "Output",
  CellTags->"b:14.1"]
}, Open  ]],

Cell[CellGroupData[{
Cell[133536, 2290, 174, 4, 45, "Input",
  CellTags->{"Simulation", "i:2", "b:14.1"}],
Cell[133713, 2296, 2067, 50, 199, "Output",
  CellTags->"b:14.1"]
}, Open  ]],

Cell[CellGroupData[{
Cell[135817, 2351, 150, 4, 45, "Input",
  CellTags->{"Simulation", "i:2", "b:14.1"}],
Cell[135970, 2357, 101, 3, 41, "Output",
  CellTags->"b:14.1"]
}, Open  ]],

Cell[CellGroupData[{
Cell[136108, 2365, 244, 7, 45, "Input",
  CellTags->{"Simulation", "i:2", "b:14.1"}],
Cell[136355, 2374, 133, 3, 41, "Output",
  CellTags->"b:14.1"]
}, Open  ]],

Cell[CellGroupData[{
Cell[136525, 2382, 216, 6, 45, "Input",
  CellTags->{"Simulation", "i:2", "b:14.1"}],
Cell[136744, 2390, 141, 4, 41, "Output",
  CellTags->"b:14.1"]
}, Open  ]],

Cell[CellGroupData[{
Cell[136922, 2399, 181, 4, 45, "Input",
  CellTags->{"Simulation", "i:2", "b:14.1"}],
Cell[137106, 2405, 2172, 51, 218, "Output",
  CellTags->"b:14.1"]
}, Open  ]],
Cell[139293, 2459, 246, 7, 39, "Text",
  Evaluatable->False,
  CellTags->{"Simulation", "i:2", "b:14.1"}],

Cell[CellGroupData[{
Cell[139564, 2470, 174, 4, 45, "Input",
  CellTags->{"Simulation", "i:2", "b:14.1"}],
Cell[139741, 2476, 796, 21, 60, "Output",
  CellTags->"b:14.1"]
}, Open  ]],

Cell[CellGroupData[{
Cell[140574, 2502, 173, 4, 45, "Input",
  CellTags->{"Simulation", "i:2", "b:14.1"}],
Cell[140750, 2508, 31370, 839, 198, 8066, 546, "GraphicsData", "PostScript", \
"Graphics",
  CellTags->"b:14.1"],
Cell[172123, 3349, 177, 5, 41, "Output",
  CellTags->"b:14.1"]
}, Open  ]],

Cell[CellGroupData[{
Cell[172337, 3359, 173, 4, 45, "Input",
  CellTags->{"Simulation", "i:2", "b:14.1"}],
Cell[172513, 3365, 53974, 1166, 198, 9163, 607, "GraphicsData", "PostScript", \
"Graphics",
  CellTags->"b:14.1"],
Cell[226490, 4533, 177, 5, 41, "Output",
  CellTags->"b:14.1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[226716, 4544, 175, 4, 53, "Subsection",
  Evaluatable->False,
  CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],
Cell[226894, 4550, 740, 13, 90, "Text",
  Evaluatable->False,
  CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],
Cell[227637, 4565, 232, 6, 39, "Text",
  Evaluatable->False,
  CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],

Cell[CellGroupData[{
Cell[227894, 4575, 273, 9, 105, "Input",
  CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],
Cell[228170, 4586, 260, 6, 60, "Output",
  CellTags->"b:14.2"]
}, Open  ]],

Cell[CellGroupData[{
Cell[228467, 4597, 693, 16, 185, "Input",
  CellTags->{"Creating MEX-Files", "i:3", "b:14.2"}],
Cell[229163, 4615, 125, 3, 37, "Print",
  CellTags->"b:14.2"],
Cell[229291, 4620, 112, 3, 37, "Print",
  CellTags->"b:14.2"],
Cell[229406, 4625, 132, 3, 37, "Print",
  CellTags->"b:14.2"],
Cell[229541, 4630, 131, 3, 37, "Print",
  CellTags->"b:14.2"],
Cell[229675, 4635, 119, 3, 37, "Print",
  CellTags->"b:14.2"],
Cell[229797, 4640, 127, 3, 37, "Print",
  CellTags->"b:14.2"],
Cell[229927, 4645, 122, 3, 37, "Print",
  CellTags->"b:14.2"],
Cell[230052, 4650, 121, 3, 37, "Print",
  CellTags->"b:14.2"],
Cell[230176, 4655, 170, 4, 37, "Print",
  CellTags->"b:14.2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[230395, 4665, 95, 1, 53, "Subsection",
  CellTags->{"Creating MEX-Files", "b:14.3"}],
Cell[230493, 4668, 1690, 29, 375, "Text",
  CellTags->{"Creating MEX-Files", "b:14.3"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[232232, 4703, 129, 3, 56, "Section",
  Evaluatable->False],

Cell[CellGroupData[{
Cell[232386, 4710, 146, 4, 53, "Subsection",
  Evaluatable->False,
  CellTags->"b:15.0"],
Cell[232535, 4716, 619, 12, 90, "Text",
  Evaluatable->False,
  CellTags->"b:15.0"]
}, Open  ]],

Cell[CellGroupData[{
Cell[233191, 4733, 174, 4, 53, "Subsection",
  Evaluatable->False,
  CellTags->{"Suspension", "i:4", "b:15.1"}],
Cell[233368, 4739, 570, 12, 98, "Text",
  Evaluatable->False,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell[CellGroupData[{
Cell[233963, 4755, 519, 20, 325, "Input",
  CellTags->{"Suspension", "i:4", "b:15.1"}],
Cell[234485, 4777, 115, 3, 37, "Print",
  CellTags->"b:15.1"],
Cell[234603, 4782, 115, 3, 37, "Print",
  CellTags->"b:15.1"],
Cell[234721, 4787, 115, 3, 37, "Print",
  CellTags->"b:15.1"]
}, Open  ]],
Cell[234851, 4793, 152, 4, 39, "Text",
  Evaluatable->False,
  CellTags->{"Suspension", "i:4", "b:15.1"}],
Cell[235006, 4799, 598, 23, 365, "Input",
  CellTags->{"Suspension", "i:4", "b:15.1"}],
Cell[235607, 4824, 229, 7, 64, "Text",
  Evaluatable->False,
  CellTags->{"Suspension", "i:4", "b:15.1"}],
Cell[235839, 4833, 174, 4, 45, "Input",
  CellTags->{"Suspension", "i:4", "b:15.1"}],
Cell[236016, 4839, 717, 14, 123, "Text",
  Evaluatable->False,
  CellTags->{"Suspension", "i:4", "b:15.1"}],
Cell[236736, 4855, 262, 9, 105, "Input",
  CellTags->{"Suspension", "i:4", "b:15.1"}],
Cell[237001, 4866, 352, 8, 56, "Text",
  Evaluatable->False,
  CellTags->{"Suspension", "i:4", "b:15.1"}],
Cell[237356, 4876, 251, 8, 85, "Input",
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell[CellGroupData[{
Cell[237632, 4888, 142, 4, 45, "Input",
  CellTags->{"Suspension", "i:4", "b:15.1"}],
Cell[237777, 4894, 333, 6, 74, "Output",
  CellTags->"b:15.1"]
}, Open  ]],
Cell[238125, 4903, 676, 12, 90, "Text",
  Evaluatable->False,
  CellTags->{"Suspension", "i:4", "b:15.1"}],
Cell[238804, 4917, 351, 12, 165, "Input",
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell[CellGroupData[{
Cell[239180, 4933, 146, 4, 45, "Input",
  CellTags->{"Suspension", "i:4", "b:15.1"}],
Cell[239329, 4939, 191, 4, 57, "Output",
  CellTags->"b:15.1"]
}, Open  ]],

Cell[CellGroupData[{
Cell[239557, 4948, 146, 4, 45, "Input",
  CellTags->{"Suspension", "i:4", "b:15.1"}],
Cell[239706, 4954, 191, 4, 57, "Output",
  CellTags->"b:15.1"]
}, Open  ]],
Cell[239912, 4961, 519, 11, 98, "Text",
  Evaluatable->False,
  CellTags->{"Suspension", "i:4", "b:15.1"}],
Cell[240434, 4974, 159, 4, 45, "Input",
  CellTags->{"Suspension", "i:4", "b:15.1"}],
Cell[240596, 4980, 261, 8, 64, "Text",
  Evaluatable->False,
  CellTags->{"Suspension", "i:4", "b:15.1"}],
Cell[240860, 4990, 143, 4, 45, "Input",
  CellTags->{"Suspension", "i:4", "b:15.1"}],
Cell[241006, 4996, 345, 9, 81, "Text",
  Evaluatable->False,
  CellTags->{"Suspension", "i:4", "b:15.1"}],

Cell[CellGroupData[{
Cell[241376, 5009, 217, 6, 45, "Input",
  CellTags->{"Suspension", "i:4", "b:15.1"}],
Cell[241596, 5017, 117, 3, 37, "Print",
  CellTags->"b:15.1"],
Cell[241716, 5022, 112, 3, 37, "Print",
  CellTags->"b:15.1"],
Cell[241831, 5027, 115, 3, 37, "Print",
  CellTags->"b:15.1"]
}, Open  ]],

Cell[CellGroupData[{
Cell[241983, 5035, 138, 4, 45, "Input",
  CellTags->{"Suspension", "i:4", "b:15.1"}],
Cell[242124, 5041, 5500, 109, 1328, "Output",
  CellTags->"b:15.1"]
}, Open  ]],

Cell[CellGroupData[{
Cell[247661, 5155, 152, 4, 45, "Input",
  CellTags->{"Suspension", "i:4", "b:15.1"}],
Cell[247816, 5161, 1013, 19, 198, "Output",
  CellTags->"b:15.1"]
}, Open  ]],

Cell[CellGroupData[{
Cell[248866, 5185, 138, 4, 45, "Input",
  CellTags->{"Suspension", "i:4", "b:15.1"}],
Cell[249007, 5191, 2941, 51, 622, "Output",
  CellTags->"b:15.1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[251997, 5248, 190, 4, 53, "Subsection",
  Evaluatable->False,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell[CellGroupData[{
Cell[252212, 5256, 186, 4, 49, "Subsubsection",
  Evaluatable->False,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[252401, 5262, 590, 11, 73, "Text",
  Evaluatable->False,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[252994, 5275, 1689, 27, 267, "Text",
  Evaluatable->False,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[254686, 5304, 177, 4, 45, "Input",
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[254866, 5310, 204, 4, 39, "Text",
  Evaluatable->False,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[255073, 5316, 191, 4, 45, "Input",
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[255267, 5322, 190, 4, 39, "Text",
  Evaluatable->False,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[255460, 5328, 320, 9, 85, "Input",
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[255783, 5339, 319, 8, 39, "Text",
  Evaluatable->False,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[256105, 5349, 301, 12, 165, "Input",
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[256409, 5363, 241, 7, 65, "Input",
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[256653, 5372, 390, 8, 56, "Text",
  Evaluatable->False,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[257046, 5382, 188, 4, 45, "Input",
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[257237, 5388, 222, 6, 39, "Text",
  Evaluatable->False,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[257462, 5396, 164, 4, 45, "Input",
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[257629, 5402, 202, 4, 39, "Text",
  Evaluatable->False,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[257834, 5408, 266, 9, 105, "Input",
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[258103, 5419, 179, 4, 39, "Text",
  Evaluatable->False,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell[CellGroupData[{
Cell[258307, 5427, 205, 4, 45, "Input",
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[258515, 5433, 954, 15, 220, "Output",
  CellTags->"b:15.2"]
}, Open  ]],

Cell[CellGroupData[{
Cell[259506, 5453, 205, 4, 45, "Input",
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[259714, 5459, 954, 15, 220, "Output",
  CellTags->"b:15.2"]
}, Open  ]],
Cell[260683, 5477, 170, 4, 45, "Input",
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[260856, 5483, 317, 14, 205, "Input",
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[261176, 5499, 199, 7, 65, "Input",
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell[CellGroupData[{
Cell[261400, 5510, 223, 6, 45, "Input",
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[261626, 5518, 114, 3, 37, "Print",
  CellTags->"b:15.2"],
Cell[261743, 5523, 116, 3, 37, "Print",
  CellTags->"b:15.2"],
Cell[261862, 5528, 116, 3, 37, "Print",
  CellTags->"b:15.2"],
Cell[261981, 5533, 117, 3, 37, "Print",
  CellTags->"b:15.2"],
Cell[262101, 5538, 112, 3, 37, "Print",
  CellTags->"b:15.2"],
Cell[262216, 5543, 115, 3, 37, "Print",
  CellTags->"b:15.2"]
}, Open  ]],
Cell[262346, 5549, 188, 4, 39, "Text",
  Evaluatable->False,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell[CellGroupData[{
Cell[262559, 5557, 148, 4, 45, "Input",
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[262710, 5563, 8542, 142, 1484, "Output",
  CellTags->"b:15.2"]
}, Open  ]],

Cell[CellGroupData[{
Cell[271289, 5710, 161, 4, 45, "Input",
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[271453, 5716, 6645, 110, 529, "Output",
  CellTags->"b:15.2"]
}, Open  ]],

Cell[CellGroupData[{
Cell[278135, 5831, 149, 4, 45, "Input",
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[278287, 5837, 2224, 36, 298, "Output",
  CellTags->"b:15.2"]
}, Open  ]],
Cell[280526, 5876, 288, 7, 39, "Text",
  Evaluatable->False,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[280817, 5885, 189, 4, 45, "Input",
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[281009, 5891, 194, 4, 39, "Text",
  Evaluatable->False,
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],

Cell[CellGroupData[{
Cell[281228, 5899, 166, 4, 45, "Input",
  CellTags->{"Flexible Manipulator", "i:5", "b:15.2"}],
Cell[281397, 5905, 1319, 27, 317, "Output",
  CellTags->"b:15.2"]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[282777, 5939, 174, 4, 53, "Subsection",
  Evaluatable->False,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[CellGroupData[{
Cell[282976, 5947, 176, 4, 49, "Subsubsection",
  Evaluatable->False,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[283155, 5953, 549, 11, 98, "Text",
  Evaluatable->False,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[283707, 5966, 125314, 1550, 554, 125169, 1546, "GraphicsData", \
"Metafile", "Graphics",
  CellTags->{"b:15.3", "Tractor-Trailer"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[409058, 7521, 175, 4, 49, "Subsubsection",
  Evaluatable->False,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[409236, 7527, 173, 4, 39, "Text",
  Evaluatable->False,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[409412, 7533, 259, 9, 105, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[409674, 7544, 298, 9, 39, "Text",
  Evaluatable->False,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[409975, 7555, 324, 8, 56, "Text",
  Evaluatable->False,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[410302, 7565, 399, 15, 225, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[410704, 7582, 292, 9, 39, "Text",
  Evaluatable->False,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[410999, 7593, 306, 10, 125, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[411342, 7608, 176, 4, 49, "Subsubsection",
  Evaluatable->False,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[CellGroupData[{
Cell[411543, 7616, 179, 4, 45, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[411725, 7622, 116, 3, 37, "Print",
  CellTags->"b:15.3"],
Cell[411844, 7627, 116, 3, 37, "Print",
  CellTags->"b:15.3"],
Cell[411963, 7632, 116, 3, 37, "Print",
  CellTags->"b:15.3"],
Cell[412082, 7637, 116, 3, 37, "Print",
  CellTags->"b:15.3"]
}, Open  ]],

Cell[CellGroupData[{
Cell[412235, 7645, 151, 4, 45, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[412389, 7651, 101, 3, 37, "Print",
  CellTags->"b:15.3"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[412539, 7660, 166, 4, 49, "Subsubsection",
  Evaluatable->False,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[412708, 7666, 621, 15, 239, "Text",
  Evaluatable->False,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[413332, 7683, 1267, 38, 645, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[414636, 7726, 168, 4, 49, "Subsubsection",
  Evaluatable->False,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[414807, 7732, 486, 10, 73, "Text",
  Evaluatable->False,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[CellGroupData[{
Cell[415318, 7746, 376, 14, 185, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[415697, 7762, 10235, 150, 2267, "Output",
  CellTags->"b:15.3"]
}, Open  ]],

Cell[CellGroupData[{
Cell[425969, 7917, 150, 4, 45, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[426122, 7923, 101, 3, 37, "Print",
  CellTags->"b:15.3"]
}, Open  ]],
Cell[426238, 7929, 195, 4, 45, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[CellGroupData[{
Cell[426458, 7937, 148, 4, 45, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[426609, 7943, 425, 8, 122, "Output",
  CellTags->"b:15.3"]
}, Open  ]],
Cell[427049, 7954, 361, 11, 145, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[CellGroupData[{
Cell[427435, 7969, 235, 7, 45, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[427673, 7978, 4596, 77, 402, "Output",
  CellTags->"b:15.3"]
}, Open  ]],

Cell[CellGroupData[{
Cell[432306, 8060, 151, 4, 45, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[432460, 8066, 114, 3, 37, "Print",
  CellTags->"b:15.3"]
}, Open  ]],
Cell[432589, 8072, 369, 11, 145, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[432961, 8085, 383, 12, 165, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[433347, 8099, 384, 12, 165, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[433734, 8113, 382, 12, 165, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[434119, 8127, 371, 12, 165, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[434493, 8141, 368, 12, 165, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[434864, 8155, 367, 12, 165, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[435234, 8169, 168, 4, 45, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[435439, 8178, 169, 4, 49, "Subsubsection",
  Evaluatable->False,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[CellGroupData[{
Cell[435633, 8186, 272, 8, 65, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[435908, 8196, 117, 3, 37, "Print",
  CellTags->"b:15.3"],
Cell[436028, 8201, 112, 3, 37, "Print",
  CellTags->"b:15.3"],
Cell[436143, 8206, 115, 3, 37, "Print",
  CellTags->"b:15.3"]
}, Open  ]],
Cell[436273, 8212, 218, 6, 45, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[436494, 8220, 410, 14, 205, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[436941, 8239, 196, 4, 49, "Subsubsection",
  Evaluatable->False,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[437140, 8245, 514, 10, 73, "Text",
  Evaluatable->False,
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],

Cell[CellGroupData[{
Cell[437679, 8259, 921, 28, 465, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[438603, 8289, 113, 3, 37, "Print",
  CellTags->"b:15.3"],
Cell[438719, 8294, 133, 3, 37, "Print",
  CellTags->"b:15.3"],
Cell[438855, 8299, 132, 3, 37, "Print",
  CellTags->"b:15.3"],
Cell[438990, 8304, 120, 3, 37, "Print",
  CellTags->"b:15.3"],
Cell[439113, 8309, 128, 3, 37, "Print",
  CellTags->"b:15.3"],
Cell[439244, 8314, 123, 3, 37, "Print",
  CellTags->"b:15.3"],
Cell[439370, 8319, 122, 3, 37, "Print",
  CellTags->"b:15.3"],
Cell[439495, 8324, 123, 3, 37, "Print",
  CellTags->"b:15.3"]
}, Open  ]],

Cell[CellGroupData[{
Cell[439655, 8332, 148, 4, 45, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[439806, 8338, 97, 3, 41, "Output",
  CellTags->"b:15.3"]
}, Open  ]],

Cell[CellGroupData[{
Cell[439940, 8346, 149, 4, 45, "Input",
  CellTags->{"Tractor-Trailer", "i:6", "b:15.3"}],
Cell[440092, 8352, 128, 3, 41, "Output",
  CellTags->"b:15.3"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[440293, 8363, 112, 3, 56, "Section",
  Evaluatable->False],

Cell[CellGroupData[{
Cell[440430, 8370, 169, 4, 53, "Subsection",
  Evaluatable->False,
  CellTags->{"Constraints Intro", "b:16.0"}],
Cell[440602, 8376, 2531, 41, 427, "Text",
  Evaluatable->False,
  CellTags->{"Constraints Intro", "b:16.0"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[443170, 8422, 172, 4, 53, "Subsection",
  Evaluatable->False,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],
Cell[443345, 8428, 416, 9, 56, "Text",
  Evaluatable->False,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],
Cell[443764, 8439, 323, 14, 205, "Input",
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],
Cell[444090, 8455, 521, 13, 185, "Input",
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],
Cell[444614, 8470, 187, 7, 65, "Input",
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell[CellGroupData[{
Cell[444826, 8481, 215, 6, 45, "Input",
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],
Cell[445044, 8489, 114, 3, 37, "Print",
  CellTags->"b:16.1"],
Cell[445161, 8494, 116, 3, 37, "Print",
  CellTags->"b:16.1"],
Cell[445280, 8499, 116, 3, 37, "Print",
  CellTags->"b:16.1"],
Cell[445399, 8504, 117, 3, 37, "Print",
  CellTags->"b:16.1"],
Cell[445519, 8509, 112, 3, 37, "Print",
  CellTags->"b:16.1"],
Cell[445634, 8514, 115, 3, 37, "Print",
  CellTags->"b:16.1"]
}, Open  ]],
Cell[445764, 8520, 267, 7, 65, "Input",
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],
Cell[446034, 8529, 173, 4, 45, "Input",
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell[CellGroupData[{
Cell[446232, 8537, 156, 4, 45, "Input",
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],
Cell[446391, 8543, 197, 5, 41, "Output",
  CellTags->"b:16.1"]
}, Open  ]],
Cell[446603, 8551, 455, 9, 56, "Text",
  Evaluatable->False,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell[CellGroupData[{
Cell[447083, 8564, 208, 6, 45, "Input",
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],
Cell[447294, 8572, 90, 3, 37, "Print",
  CellTags->"b:16.1"],
Cell[447387, 8577, 90, 3, 37, "Print",
  CellTags->"b:16.1"],
Cell[447480, 8582, 90, 3, 37, "Print",
  CellTags->"b:16.1"],
Cell[447573, 8587, 90, 3, 37, "Print",
  CellTags->"b:16.1"],
Cell[447666, 8592, 1054, 21, 233, "Output",
  CellTags->"b:16.1"]
}, Open  ]],
Cell[448735, 8616, 491, 10, 73, "Text",
  Evaluatable->False,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell[CellGroupData[{
Cell[449251, 8630, 211, 6, 45, "Input",
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],
Cell[449465, 8638, 743, 14, 171, "Output",
  CellTags->"b:16.1"]
}, Open  ]],
Cell[450223, 8655, 223, 6, 39, "Text",
  Evaluatable->False,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell[CellGroupData[{
Cell[450471, 8665, 177, 4, 45, "Input",
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],
Cell[450651, 8671, 1468, 33, 170, "Output",
  CellTags->"b:16.1"]
}, Open  ]],
Cell[452134, 8707, 330, 8, 56, "Text",
  Evaluatable->False,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell[CellGroupData[{
Cell[452489, 8719, 185, 4, 45, "Input",
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],
Cell[452677, 8725, 1194, 31, 104, "Output",
  CellTags->"b:16.1"]
}, Open  ]],
Cell[453886, 8759, 289, 7, 39, "Text",
  Evaluatable->False,
  CellTags->{"Closed Chain", "i:7", "b:16.1"}],

Cell[CellGroupData[{
Cell[454200, 8770, 175, 5, 43, "Input",
  CellTags->"b:16.1"],
Cell[454378, 8777, 768, 21, 57, "Output",
  CellTags->"b:16.1"]
}, Open  ]],

Cell[CellGroupData[{
Cell[455183, 8803, 205, 5, 43, "Input",
  CellTags->"b:16.1"],
Cell[455391, 8810, 754, 17, 74, "Output",
  CellTags->"b:16.1"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[456194, 8833, 171, 4, 53, "Subsection",
  Evaluatable->False,
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],
Cell[456368, 8839, 398, 9, 56, "Text",
  Evaluatable->False,
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],
Cell[456769, 8850, 330, 14, 205, "Input",
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],
Cell[457102, 8866, 517, 13, 185, "Input",
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],
Cell[457622, 8881, 189, 7, 65, "Input",
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],

Cell[CellGroupData[{
Cell[457836, 8892, 215, 6, 45, "Input",
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],
Cell[458054, 8900, 114, 3, 37, "Print",
  CellTags->"b:16.2"],
Cell[458171, 8905, 116, 3, 37, "Print",
  CellTags->"b:16.2"],
Cell[458290, 8910, 116, 3, 37, "Print",
  CellTags->"b:16.2"],
Cell[458409, 8915, 117, 3, 37, "Print",
  CellTags->"b:16.2"],
Cell[458529, 8920, 112, 3, 37, "Print",
  CellTags->"b:16.2"],
Cell[458644, 8925, 115, 3, 37, "Print",
  CellTags->"b:16.2"]
}, Open  ]],
Cell[458774, 8931, 267, 7, 65, "Input",
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],
Cell[459044, 8940, 173, 4, 45, "Input",
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],

Cell[CellGroupData[{
Cell[459242, 8948, 156, 4, 45, "Input",
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],
Cell[459401, 8954, 191, 5, 41, "Output",
  CellTags->"b:16.2"]
}, Open  ]],
Cell[459607, 8962, 245, 7, 39, "Text",
  Evaluatable->False,
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],

Cell[CellGroupData[{
Cell[459877, 8973, 212, 6, 45, "Input",
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],
Cell[460092, 8981, 90, 3, 37, "Print",
  CellTags->"b:16.2"],
Cell[460185, 8986, 90, 3, 37, "Print",
  CellTags->"b:16.2"],
Cell[460278, 8991, 90, 3, 37, "Print",
  CellTags->"b:16.2"],
Cell[460371, 8996, 90, 3, 37, "Print",
  CellTags->"b:16.2"],
Cell[460464, 9001, 122, 3, 37, "Print",
  CellTags->"b:16.2"],
Cell[460589, 9006, 2460, 45, 441, "Output",
  CellTags->"b:16.2"]
}, Open  ]],

Cell[CellGroupData[{
Cell[463086, 9056, 156, 4, 45, "Input",
  CellTags->{"Closed Chain", "i:8", "b:16.2"}],
Cell[463245, 9062, 671, 13, 151, "Output",
  CellTags->"b:16.2"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[463965, 9081, 222, 6, 53, "Subsection",
  Evaluatable->False,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[464190, 9089, 629, 12, 73, "Text",
  Evaluatable->False,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[464822, 9103, 45941, 570, 234, 45809, 566, "GraphicsData", "Metafile", \
"Graphics",
  CellTags->{"b:16.3", "Rolling Disk"}],

Cell[CellGroupData[{
Cell[510788, 9677, 173, 7, 65, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[510964, 9686, 211, 4, 60, "Output",
  CellTags->"b:16.3"]
}, Open  ]],
Cell[511190, 9693, 218, 9, 105, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[511411, 9704, 333, 9, 64, "Text",
  Evaluatable->False,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[511747, 9715, 222, 7, 65, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[511972, 9724, 324, 8, 85, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[512299, 9734, 168, 4, 39, "Text",
  Evaluatable->False,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[512470, 9740, 189, 7, 65, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[512662, 9749, 190, 4, 39, "Text",
  Evaluatable->False,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{
Cell[512877, 9757, 215, 6, 45, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[513095, 9765, 114, 3, 37, "Print",
  CellTags->"b:16.3"],
Cell[513212, 9770, 116, 3, 29, "Print",
  CellTags->"b:16.3"],
Cell[513331, 9775, 117, 3, 29, "Print",
  CellTags->"b:16.3"],
Cell[513451, 9780, 112, 3, 29, "Print",
  CellTags->"b:16.3"],
Cell[513566, 9785, 115, 3, 29, "Print",
  CellTags->"b:16.3"]
}, Open  ]],
Cell[513696, 9791, 499, 10, 61, "Text",
  Evaluatable->False,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{
Cell[514220, 9805, 213, 7, 57, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[514436, 9814, 193, 4, 43, "Output",
  CellTags->"b:16.3"]
}, Open  ]],
Cell[514644, 9821, 357, 8, 48, "Text",
  Evaluatable->False,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{
Cell[515026, 9833, 237, 7, 44, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[515266, 9842, 90, 3, 29, "Print",
  CellTags->"b:16.3"],
Cell[515359, 9847, 90, 3, 29, "Print",
  CellTags->"b:16.3"],
Cell[515452, 9852, 90, 3, 29, "Print",
  CellTags->"b:16.3"],
Cell[515545, 9857, 90, 3, 29, "Print",
  CellTags->"b:16.3"]
}, Open  ]],
Cell[515650, 9863, 289, 7, 35, "Text",
  Evaluatable->False,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{
Cell[515964, 9874, 143, 4, 44, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[516110, 9880, 94, 3, 43, "Output",
  CellTags->"b:16.3"]
}, Open  ]],
Cell[516219, 9886, 626, 12, 74, "Text",
  Evaluatable->False,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{
Cell[516870, 9902, 190, 4, 44, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[517063, 9908, 126, 3, 43, "Output",
  CellTags->"b:16.3"]
}, Open  ]],
Cell[517204, 9914, 348, 8, 48, "Text",
  Evaluatable->False,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[517555, 9924, 162, 4, 44, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[517720, 9930, 179, 4, 44, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[517902, 9936, 186, 4, 35, "Text",
  Evaluatable->False,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{
Cell[518113, 9944, 191, 4, 44, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[518307, 9950, 5461, 118, 345, "Output",
  CellTags->"b:16.3"]
}, Open  ]],
Cell[523783, 10071, 243, 7, 35, "Text",
  Evaluatable->False,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[524029, 10080, 224, 7, 57, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{
Cell[524278, 10091, 232, 7, 44, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[524513, 10100, 165, 4, 43, "Output",
  CellTags->"b:16.3"]
}, Open  ]],
Cell[524693, 10107, 461, 10, 61, "Text",
  Evaluatable->False,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{
Cell[525179, 10121, 241, 8, 70, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[525423, 10131, 131, 3, 43, "Output",
  CellTags->"b:16.3"],
Cell[525557, 10136, 131, 3, 43, "Output",
  CellTags->"b:16.3"]
}, Open  ]],
Cell[525703, 10142, 235, 7, 35, "Text",
  Evaluatable->False,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{
Cell[525963, 10153, 192, 4, 44, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[526158, 10159, 219, 5, 56, "Output",
  CellTags->"b:16.3"]
}, Open  ]],
Cell[526392, 10167, 232, 7, 35, "Text",
  Evaluatable->False,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{
Cell[526649, 10178, 177, 4, 44, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[526829, 10184, 117, 3, 43, "Output",
  CellTags->"b:16.3"]
}, Open  ]],

Cell[CellGroupData[{
Cell[526983, 10192, 199, 4, 44, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[527185, 10198, 2007, 52, 136, "Output",
  CellTags->"b:16.3"]
}, Open  ]],
Cell[529207, 10253, 516, 10, 61, "Text",
  Evaluatable->False,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],

Cell[CellGroupData[{
Cell[529748, 10267, 209, 6, 44, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[529960, 10275, 892, 23, 56, "Output",
  CellTags->"b:16.3"]
}, Open  ]],

Cell[CellGroupData[{
Cell[530889, 10303, 258, 8, 70, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[531150, 10313, 1945, 47, 56, "Output",
  CellTags->"b:16.3"]
}, Open  ]],

Cell[CellGroupData[{
Cell[533132, 10365, 232, 8, 70, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[533367, 10375, 743, 19, 47, "Output",
  CellTags->"b:16.3"]
}, Open  ]],

Cell[CellGroupData[{
Cell[534147, 10399, 227, 8, 70, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[534377, 10409, 734, 19, 50, "Output",
  CellTags->"b:16.3"]
}, Open  ]],

Cell[CellGroupData[{
Cell[535148, 10433, 214, 6, 44, "Input",
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}],
Cell[535365, 10441, 3297, 74, 116, "Output",
  CellTags->"b:16.3"]
}, Open  ]],
Cell[538677, 10518, 238, 7, 35, "Text",
  Evaluatable->False,
  CellTags->{"Rolling Disk", "i:9", "b:16.3"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[538952, 10530, 162, 4, 54, "Subsection",
  Evaluatable->False,
  CellTags->{"Drive Chain", "b:16.4"}],

Cell[CellGroupData[{
Cell[539139, 10538, 157, 4, 34, "Subsubsection",
  Evaluatable->False,
  CellTags->{"Drive Chain", "b:16.4"}],
Cell[539299, 10544, 314, 8, 48, "Text",
  Evaluatable->False,
  CellTags->{"Drive Chain", "b:16.4"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[539650, 10557, 155, 4, 34, "Subsubsection",
  Evaluatable->False,
  CellTags->{"Drive Chain", "b:16.4"}],
Cell[539808, 10563, 152, 4, 35, "Text",
  Evaluatable->False,
  CellTags->{"Drive Chain", "b:16.4"}],
Cell[539963, 10569, 287, 10, 96, "Input",
  CellTags->{"Drive Chain", "b:16.4"}],
Cell[540253, 10581, 158, 4, 35, "Text",
  Evaluatable->False,
  CellTags->{"Drive Chain", "b:16.4"}],
Cell[540414, 10587, 325, 10, 83, "Input",
  CellTags->{"Drive Chain", "b:16.4"}],
Cell[540742, 10599, 154, 4, 35, "Text",
  Evaluatable->False,
  CellTags->{"Drive Chain", "b:16.4"}],
Cell[540899, 10605, 301, 9, 83, "Input",
  CellTags->{"Drive Chain", "b:16.4"}],
Cell[541203, 10616, 171, 4, 44, "Input",
  CellTags->{"Drive Chain", "b:16.4"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[541411, 10625, 156, 4, 34, "Subsubsection",
  Evaluatable->False,
  CellTags->{"Drive Chain", "b:16.4"}],
Cell[541570, 10631, 280, 9, 37, "Text",
  Evaluatable->False,
  CellTags->{"Drive Chain", "b:16.4"}],
Cell[541853, 10642, 216, 8, 70, "Input",
  CellTags->{"Drive Chain", "b:16.4"}],
Cell[542072, 10652, 288, 9, 37, "Text",
  Evaluatable->False,
  CellTags->{"Drive Chain", "b:16.4"}],
Cell[542363, 10663, 238, 8, 70, "Input",
  CellTags->{"Drive Chain", "b:16.4"}],
Cell[542604, 10673, 496, 10, 61, "Text",
  Evaluatable->False,
  CellTags->{"Drive Chain", "b:16.4"}],
Cell[543103, 10685, 274, 12, 122, "Input",
  CellTags->{"Drive Chain", "b:16.4"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[543414, 10702, 161, 4, 34, "Subsubsection",
  Evaluatable->False,
  CellTags->{"Drive Chain", "b:16.4"}],
Cell[543578, 10708, 180, 4, 35, "Text",
  Evaluatable->False,
  CellTags->{"Drive Chain", "b:16.4"}],
Cell[543761, 10714, 161, 4, 44, "Input",
  CellTags->{"Drive Chain", "b:16.4"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[543959, 10723, 174, 4, 34, "Subsubsection",
  Evaluatable->False,
  CellTags->{"Drive Chain", "b:16.4"}],
Cell[544136, 10729, 150, 4, 44, "Input",
  CellTags->{"Drive Chain", "b:16.4"}],

Cell[CellGroupData[{
Cell[544311, 10737, 156, 4, 44, "Input",
  CellTags->{"Drive Chain", "b:16.4"}],
Cell[544470, 10743, 116, 3, 29, "Print",
  CellTags->"b:16.4"],
Cell[544589, 10748, 116, 3, 29, "Print",
  CellTags->"b:16.4"],
Cell[544708, 10753, 116, 3, 29, "Print",
  CellTags->"b:16.4"]
}, Open  ]],

Cell[CellGroupData[{
Cell[544861, 10761, 210, 6, 44, "Input",
  CellTags->{"Drive Chain", "b:16.4"}],
Cell[545074, 10769, 117, 3, 29, "Print",
  CellTags->"b:16.4"],
Cell[545194, 10774, 112, 3, 29, "Print",
  CellTags->"b:16.4"],
Cell[545309, 10779, 115, 3, 29, "Print",
  CellTags->"b:16.4"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[545473, 10788, 176, 4, 34, "Subsubsection",
  Evaluatable->False,
  CellTags->{"Drive Chain", "i:10", "b:16.4"}],
Cell[545652, 10794, 489, 10, 61, "Text",
  Evaluatable->False,
  CellTags->{"Drive Chain", "i:10", "b:16.4"}],
Cell[546144, 10806, 376, 9, 83, "Input",
  CellTags->{"Drive Chain", "i:10", "b:16.4"}],
Cell[546523, 10817, 317, 8, 70, "Input",
  CellTags->{"Drive Chain", "i:10", "b:16.4"}],
Cell[546843, 10827, 287, 8, 70, "Input",
  CellTags->{"Drive Chain", "i:10", "b:16.4"}],
Cell[547133, 10837, 144, 4, 44, "Input",
  CellTags->{"Drive Chain", "i:10", "b:16.4"}],

Cell[CellGroupData[{
Cell[547302, 10845, 246, 7, 44, "Input",
  CellTags->{"Drive Chain", "i:10", "b:16.4"}],
Cell[547551, 10854, 100, 3, 29, "Print",
  CellTags->{"i:10", "b:16.4"}],
Cell[547654, 10859, 100, 3, 29, "Print",
  CellTags->{"i:10", "b:16.4"}],
Cell[547757, 10864, 100, 3, 29, "Print",
  CellTags->{"i:10", "b:16.4"}],
Cell[547860, 10869, 100, 3, 29, "Print",
  CellTags->{"i:10", "b:16.4"}],
Cell[547963, 10874, 132, 3, 29, "Print",
  CellTags->{"i:10", "b:16.4"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[548132, 10882, 141, 4, 44, "Input",
  CellTags->{"Drive Chain", "i:10", "b:16.4"}],
Cell[548276, 10888, 313, 6, 56, "Output",
  CellTags->{"i:10", "b:16.4"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[548626, 10899, 141, 4, 44, "Input",
  CellTags->{"Drive Chain", "i:10", "b:16.4"}],
Cell[548770, 10905, 600, 10, 82, "Output",
  CellTags->{"i:10", "b:16.4"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[549407, 10920, 141, 4, 44, "Input",
  CellTags->{"Drive Chain", "i:10", "b:16.4"}],
Cell[549551, 10926, 95, 3, 43, "Output",
  CellTags->{"i:10", "b:16.4"}]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[549719, 10937, 149, 4, 51, "Section",
  Evaluatable->False,
  CellTags->{"References", "b:17"}],
Cell[549871, 10943, 1621, 30, 367, "Text",
  Evaluatable->False,
  CellTags->{"References", "b:17"}]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

