(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.0'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[      8241,        240]*)
(*NotebookOutlinePosition[     16495,        423]*)
(*  CellTagsIndexPosition[     16270,        412]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["Introduction", "Section",
  CellTags->"b:42"],

Cell[BoxData[{
    \(\(Off[General::spell];\)\), "\n", 
    \(\(Off[General::spell1];\)\)}], "Input",
  CellLabel->"In[1]:=",
  CellTags->"b:42"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Exponential Observer", "Section",
  CellTags->{"Exponential Observer", "i:1"}],

Cell[CellGroupData[{

Cell["Introduction", "Subsection",
  CellTags->{"Exponential Observer", "i:1", "b:43.0"}],

Cell["\<\
If a nonlinear system has a detectable Taylor linearization at an equibrium \
point of interest, then an observer with exponential ly stable error dynamics \
can be designed. Of course, the observer is only local.
Load the package:\
\>", "Text",
  CellTags->{"Exponential Observer", "i:1", "b:43.0"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(<< ProPac`\)], "Input",
  CellLabel->"In[3]:=",
  CellTags->{"Exponential Observer", "i:1", "b:43.0"}],

Cell[BoxData[
    \("\n    This is ProPac - Development Version 3.xx\n \n ProPac has been \
successfully loaded. It contains several packages\n for multibody dynamics, \
linear control and nonlinear control.\n The package names are:\n    GeoTools\n\
    Dynamics\n    ControlL\n    ControlN\n    DifferentialForms\n    \
ObserverTools\n    NDTools\n    VSCTools\n    MatlabInterface\n    MEXTools\n \
Appropriate packages will be loaded automatically as required.\n To obtain a \
list of functions in a package and links to function\n declarations, enter:\n \
   ?ProPac`PackageName`*\n Help is also available in the Help Browser under \
Add-ons."\)], "Print",
  CellLabel->"From In[3]:=",
  CellTags->{"Exponential Observer", "i:1", "b:43.0"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(NonlinearControl\)], "Input",
  CellLabel->"In[4]:=",
  CellTags->{"Exponential Observer", "i:1", "b:43.0"}],

Cell[BoxData[
    \("  *** LinearControl successfully loaded ***"\)], "Print",
  CellLabel->"From In[4]:=",
  CellTags->{"Exponential Observer", "i:1", "b:43.0"}],

Cell[BoxData[
    \("  *** NonlinearControl successfully loaded ***"\)], "Print",
  CellLabel->"From In[4]:=",
  CellTags->{"Exponential Observer", "i:1", "b:43.0"}],

Cell[BoxData[
    \(NonlinearControl\)], "Output",
  CellLabel->"Out[4]=",
  CellTags->{"Exponential Observer", "i:1", "b:43.0"}]
}, Open  ]],

Cell["Here is the definition of the function ExponentialObserver", "Text",
  CellTags->{"Exponential Observer", "i:1", "b:43.0"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?ExponentialObserver\)\)], "Input",
  CellLabel->"In[5]:=",
  CellTags->{"Exponential Observer", "i:1", "b:43.0"}],

Cell[BoxData[
    \("ExponentialObserver[f,h,x,u,y,x0,u0,decay] returns\n {rhs,xhat,eigs} \
that defines an exponential observer\n of the form:\n \n    \
xhat_dot=rhs(xhat,u,y)\n \n for a system defined by\n \n     x_dot=f(x,u), \
y=h(x)\n \n x0,u0 is an equilibrium point, f(x0,u0)=0, and decay is\n a \
specified exponential decay rate. The observer has an equilibrium\n point at \
xhat=x0, i.e., rhs(x0,u0,h(x0))=0, and its linearization at\n this point has \
eigenvalues eigs. Furthermore, Re(eigs)<decay.\n \n By default the state \
estimates assume the names of the states with 'hat'\n adjoined. The calling \
syntax:\n \n    {rhs,xhat,eigs}=ExponentialObserver[f,h,x,u,y,x0,u0,decay,z]\n\
 \n returns state estimates with names z1,z2,...,zn\n "\)], "Print",
  CellTags->{"Info3268717555-1940392", "b:43.0"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Example", "Subsection",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}],

Cell[BoxData[{
    \(\(f = {x1^2\  + Cos[x2]*x3 - x2, x2*Cos[x1], 
          Sin[x3] + Cos[x1] + u1};\)\), "\n", 
    \(\(h = {x1 + x1*x2 + x3, x2 + x3^3};\)\)}], "Input",
  CellLabel->"In[6]:=",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}],

Cell[CellGroupData[{

Cell[BoxData[
    \(ExponentialObserver[f, h, {x1, x2, x3}, {u1}, {y1, y2}, {0, 0, 0}, {0}, 
      4]\)], "Input",
  CellLabel->"In[8]:=",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}],

Cell[BoxData[
    \({{x1hat\^2 - x2hat - 27\ \((\(-x1hat\) - x1hat\ x2hat - x3hat + y1)\) - 
          5\ \((\(-x2hat\) - x3hat\^3 + y2)\) + x3hat\ Cos[x2hat], 
        10\ \((\(-x2hat\) - x3hat\^3 + y2)\) + x2hat\ Cos[x1hat], 
        u1 + 45\ \((\(-x1hat\) - x1hat\ x2hat - x3hat + y1)\) + 
          5\ \((\(-x2hat\) - x3hat\^3 + y2)\) + Cos[x1hat] + 
          Sin[x3hat]}, {x1hat, x2hat, 
        x3hat}, {\(-9\), \(-9\), \(-8\)}}\)], "Output",
  CellLabel->"Out[8]=",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ExponentialObserver[f, h, {x1, x2, x3}, {u1}, {y1, y2}, {0, 0, 0}, {0}, 
      4, z]\)], "Input",
  CellLabel->"In[9]:=",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}],

Cell[BoxData[
    \({{z1\^2 - z2 - 27\ \((y1 - z1 - z1\ z2 - z3)\) - 
          5\ \((y2 - z2 - z3\^3)\) + z3\ Cos[z2], 
        10\ \((y2 - z2 - z3\^3)\) + z2\ Cos[z1], 
        u1 + 45\ \((y1 - z1 - z1\ z2 - z3)\) + 5\ \((y2 - z2 - z3\^3)\) + 
          Cos[z1] + Sin[z3]}, {z1, z2, 
        z3}, {\(-9\), \(-9\), \(-8\)}}\)], "Output",
  CellLabel->"Out[9]=",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
    \(ExponentialObserver[f, h, {x1, x2, x3}, {u1}, {y1, y2}, {0, 0, 0}, {0}, 
      4, x]\)], "Input",
  CellLabel->"In[10]:=",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}],

Cell[BoxData[
    \({{x1\^2 - x2 - 27\ \((\(-x1\) - x1\ x2 - x3 + y1)\) - 
          5\ \((\(-x2\) - x3\^3 + y2)\) + x3\ Cos[x2], 
        10\ \((\(-x2\) - x3\^3 + y2)\) + x2\ Cos[x1], 
        u1 + 45\ \((\(-x1\) - x1\ x2 - x3 + y1)\) + 
          5\ \((\(-x2\) - x3\^3 + y2)\) + Cos[x1] + Sin[x3]}, {x1, x2, 
        x3}, {\(-9\), \(-9\), \(-8\)}}\)], "Output",
  CellLabel->"Out[10]=",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}]
}, Open  ]],

Cell["If the (Taylor) linearized system is not observable", "Text",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(h = {x1*x2 + x3, x2 + x3^3};\)\), "\n", 
    \(ExponentialObserver[f, h, {x1, x2, x3}, {u1}, {y1, y2}, {0, 0, 0}, {0}, 
      4]\)}], "Input",
  CellLabel->"In[11]:=",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}],

Cell[BoxData[
    \("Exponential observer may not exist"\)], "Print",
  CellLabel->"From In[11]:=",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}]
}, Open  ]]
}, Open  ]]
}, Open  ]]
},
FrontEndVersion->"5.0 for Microsoft Windows",
ScreenRectangle->{{0, 1024}, {0, 695}},
ShowPageBreaks->False,
WindowSize->{687, 557},
WindowMargins->{{11, Automatic}, {Automatic, 24}},
PrintingStartingPageNumber->189,
TaggingRules:>{"NextIndexingNumber" -> 2, "IndexEntries" -> {{"i:1", \
"ExponentialObserver"}}, "PaginationCache" -> {{2003, 8, 1, 21, 15, \
48.961344`9.442428189021717}, {189, 189, 189, 189, 189, 189, 189, 190, 190, \
190, 190, 190, 190, 190, 190, 190, 191, 191, 191, 191, 191, 191, 191, 191, \
191}}},
ShowCellTags->True,
StyleDefinitions -> "HelpBrowser.nb"
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "b:42"->{
    Cell[1776, 53, 51, 1, 70, "Section",
      CellTags->"b:42"],
    Cell[1830, 56, 145, 4, 63, "Input",
      CellTags->"b:42"]},
  "Exponential Observer"->{
    Cell[2012, 65, 84, 1, 70, "Section",
      CellTags->{"Exponential Observer", "i:1"}],
    Cell[2121, 70, 89, 1, 53, "Subsection",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[2213, 73, 310, 6, 81, "Text",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[2548, 83, 122, 3, 43, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[2673, 88, 744, 11, 417, "Print",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[3454, 104, 128, 3, 33, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[3585, 109, 162, 3, 29, "Print",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[3750, 114, 165, 3, 29, "Print",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[3918, 119, 129, 3, 33, "Output",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[4062, 125, 129, 1, 35, "Text",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[4216, 130, 136, 3, 33, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[5218, 152, 84, 1, 54, "Subsection",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[5305, 155, 250, 5, 46, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[5580, 164, 193, 4, 33, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[5776, 170, 528, 9, 72, "Output",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[6341, 184, 196, 4, 33, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[6540, 190, 417, 8, 59, "Output",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[6994, 203, 197, 4, 33, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[7194, 209, 443, 8, 59, "Output",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[7652, 220, 122, 1, 35, "Text",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[7799, 225, 245, 5, 46, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[8047, 232, 154, 3, 29, "Print",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}]},
  "i:1"->{
    Cell[2012, 65, 84, 1, 70, "Section",
      CellTags->{"Exponential Observer", "i:1"}],
    Cell[2121, 70, 89, 1, 53, "Subsection",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[2213, 73, 310, 6, 81, "Text",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[2548, 83, 122, 3, 43, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[2673, 88, 744, 11, 417, "Print",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[3454, 104, 128, 3, 33, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[3585, 109, 162, 3, 29, "Print",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[3750, 114, 165, 3, 29, "Print",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[3918, 119, 129, 3, 33, "Output",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[4062, 125, 129, 1, 35, "Text",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[4216, 130, 136, 3, 33, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[5218, 152, 84, 1, 54, "Subsection",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[5305, 155, 250, 5, 46, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[5580, 164, 193, 4, 33, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[5776, 170, 528, 9, 72, "Output",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[6341, 184, 196, 4, 33, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[6540, 190, 417, 8, 59, "Output",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[6994, 203, 197, 4, 33, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[7194, 209, 443, 8, 59, "Output",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[7652, 220, 122, 1, 35, "Text",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[7799, 225, 245, 5, 46, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[8047, 232, 154, 3, 29, "Print",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}]},
  "b:43.0"->{
    Cell[2121, 70, 89, 1, 53, "Subsection",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[2213, 73, 310, 6, 81, "Text",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[2548, 83, 122, 3, 43, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[2673, 88, 744, 11, 417, "Print",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[3454, 104, 128, 3, 33, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[3585, 109, 162, 3, 29, "Print",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[3750, 114, 165, 3, 29, "Print",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[3918, 119, 129, 3, 33, "Output",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[4062, 125, 129, 1, 35, "Text",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[4216, 130, 136, 3, 33, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
    Cell[4355, 135, 814, 11, 305, "Print",
      CellTags->{"Info3268717555-1940392", "b:43.0"}]},
  "Info3268717555-1940392"->{
    Cell[4355, 135, 814, 11, 305, "Print",
      CellTags->{"Info3268717555-1940392", "b:43.0"}]},
  "b:43.1"->{
    Cell[5218, 152, 84, 1, 54, "Subsection",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[5305, 155, 250, 5, 46, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[5580, 164, 193, 4, 33, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[5776, 170, 528, 9, 72, "Output",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[6341, 184, 196, 4, 33, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[6540, 190, 417, 8, 59, "Output",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[6994, 203, 197, 4, 33, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[7194, 209, 443, 8, 59, "Output",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[7652, 220, 122, 1, 35, "Text",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[7799, 225, 245, 5, 46, "Input",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
    Cell[8047, 232, 154, 3, 29, "Print",
      CellTags->{"Exponential Observer", "i:1", "b:43.1"}]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"b:42", 9249, 264},
  {"Exponential Observer", 9409, 269},
  {"i:1", 11640, 314},
  {"b:43.0", 13874, 359},
  {"Info3268717555-1940392", 15014, 382},
  {"b:43.1", 15127, 385}
  }
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 51, 1, 70, "Section",
  CellTags->"b:42"],
Cell[1830, 56, 145, 4, 63, "Input",
  CellTags->"b:42"]
}, Open  ]],

Cell[CellGroupData[{
Cell[2012, 65, 84, 1, 70, "Section",
  CellTags->{"Exponential Observer", "i:1"}],

Cell[CellGroupData[{
Cell[2121, 70, 89, 1, 53, "Subsection",
  CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
Cell[2213, 73, 310, 6, 81, "Text",
  CellTags->{"Exponential Observer", "i:1", "b:43.0"}],

Cell[CellGroupData[{
Cell[2548, 83, 122, 3, 43, "Input",
  CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
Cell[2673, 88, 744, 11, 417, "Print",
  CellTags->{"Exponential Observer", "i:1", "b:43.0"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[3454, 104, 128, 3, 33, "Input",
  CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
Cell[3585, 109, 162, 3, 29, "Print",
  CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
Cell[3750, 114, 165, 3, 29, "Print",
  CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
Cell[3918, 119, 129, 3, 33, "Output",
  CellTags->{"Exponential Observer", "i:1", "b:43.0"}]
}, Open  ]],
Cell[4062, 125, 129, 1, 35, "Text",
  CellTags->{"Exponential Observer", "i:1", "b:43.0"}],

Cell[CellGroupData[{
Cell[4216, 130, 136, 3, 33, "Input",
  CellTags->{"Exponential Observer", "i:1", "b:43.0"}],
Cell[4355, 135, 814, 11, 305, "Print",
  CellTags->{"Info3268717555-1940392", "b:43.0"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[5218, 152, 84, 1, 54, "Subsection",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
Cell[5305, 155, 250, 5, 46, "Input",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}],

Cell[CellGroupData[{
Cell[5580, 164, 193, 4, 33, "Input",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
Cell[5776, 170, 528, 9, 72, "Output",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[6341, 184, 196, 4, 33, "Input",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
Cell[6540, 190, 417, 8, 59, "Output",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}]
}, Open  ]],

Cell[CellGroupData[{
Cell[6994, 203, 197, 4, 33, "Input",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
Cell[7194, 209, 443, 8, 59, "Output",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}]
}, Open  ]],
Cell[7652, 220, 122, 1, 35, "Text",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}],

Cell[CellGroupData[{
Cell[7799, 225, 245, 5, 46, "Input",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}],
Cell[8047, 232, 154, 3, 29, "Print",
  CellTags->{"Exponential Observer", "i:1", "b:43.1"}]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

