(* Content-type: application/mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 7.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       145,          7]
NotebookDataLength[    221517,       6165]
NotebookOptionsPosition[    103994,       3613]
NotebookOutlinePosition[    190636,       5356]
CellTagsIndexPosition[    189993,       5332]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["Introduction", "Section",
 CellTags->{"Introduction", "b:36"}],

Cell["\<\
Every observable system has an 'observable form.' Some can be transformed \
into a (time-invariant or time-varying) linear from up to output injection, \
called an 'observer form'. Observer design is facilitated by transforming the \
system to one of these forms. Here we describe and illustrate the main tools \
in ProPac that generate these transformations.
The functions provided in this package complement basic functions in the \
nonlinear control package, e.g., Observability, ObservabilityCodistribution, \
and Exponential Observer.\
\>", "Text",
 CellTags->{"Introduction", "i:1", "b:36"}],

Cell["First load the package.", "Text",
 CellTags->{"Introduction", "i:1", "b:36"}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"<<", "ProPac`"}]], "Input",
 CellTags->{"Introduction", "i:1", "b:36"},
 CellLabel->"In[1]:="],

Cell[BoxData["\<\"\\n    This is ProPac - Development Version 3.xx\\n \\n \
ProPac has been successfully loaded. It contains several packages\\n for \
multibody dynamics, linear control and nonlinear control.\\n The package \
names are:\\n    GeoTools\\n    Dynamics\\n    ControlL\\n    ControlN\\n    \
DifferentialForms\\n    ObserverTools\\n    NDTools\\n    VSCTools\\n    \
MatlabInterface\\n    MEXTools\\n Appropriate packages will be loaded \
automatically as required.\\n To obtain a list of functions in a package and \
links to function\\n declarations, enter:\\n    ?ProPac`PackageName`*\\n Help \
is also available in the Help Browser under Add-ons.\"\>"], "Print",
 CellTags->{"Introduction", "i:1", "b:36"},
 CellLabel->"From In[1]:="]
}, Open  ]],

Cell["The available functions in the ObserverTools package are:", "Text",
 CellTags->{"Observer Tools", "Introduction", "i:1", "b:36"}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"?", "ProPac`ObserverTools`*"}]], "Input",
 CellTags->{"Observer Tools", "Introduction", "i:1", "b:36"},
 CellLabel->"In[2]:="],

Cell[BoxData[GridBox[{
   {
    StyleBox["ProPac`ObserverTools`",
     FontFamily->"Helvetica",
     FontSize->12,
     FontWeight->"Bold"]},
   {GridBox[{
      {
       ButtonBox[
        StyleBox["LinearizeToOutputInjection",
         FontVariations->{"Underline"->True},
         FontColor->RGBColor[0, 0, 1]],
        Appearance->{Automatic, None},
        ButtonData:>{
         "Info3268713131-3644900", 
          "ProPac`ObserverTools`LinearizeToOutputInjection"},
        ButtonFunction:>(Internal`PutInformation[#, LongForm -> False]& ),
        ButtonNote->"ProPac`ObserverTools`",
        Evaluator->Automatic], 
       ButtonBox[
        StyleBox["ObservableForm",
         FontVariations->{"Underline"->True},
         FontColor->RGBColor[0, 0, 1]],
        Appearance->{Automatic, None},
        ButtonData:>{
         "Info3268713131-3644900", "ProPac`ObserverTools`ObservableForm"},
        ButtonFunction:>(Internal`PutInformation[#, LongForm -> False]& ),
        ButtonNote->"ProPac`ObserverTools`",
        Evaluator->Automatic], 
       ButtonBox[
        StyleBox["PDESolve",
         FontVariations->{"Underline"->True},
         FontColor->RGBColor[0, 0, 1]],
        Appearance->{Automatic, None},
        ButtonData:>{
         "Info3268713131-3644900", "ProPac`ObserverTools`PDESolve"},
        ButtonFunction:>(Internal`PutInformation[#, LongForm -> False]& ),
        ButtonNote->"ProPac`ObserverTools`",
        Evaluator->Automatic]},
      {
       ButtonBox[
        StyleBox["ObservabilityIndices",
         FontVariations->{"Underline"->True},
         FontColor->RGBColor[0, 0, 1]],
        Appearance->{Automatic, None},
        ButtonData:>{
         "Info3268713131-3644900", 
          "ProPac`ObserverTools`ObservabilityIndices"},
        ButtonFunction:>(Internal`PutInformation[#, LongForm -> False]& ),
        ButtonNote->"ProPac`ObserverTools`",
        Evaluator->Automatic], 
       ButtonBox[
        StyleBox["ObserverTools",
         FontVariations->{"Underline"->True},
         FontColor->RGBColor[0, 0, 1]],
        Appearance->{Automatic, None},
        ButtonData:>{
         "Info3268713131-3644900", "ProPac`ObserverTools`ObserverTools"},
        ButtonFunction:>(Internal`PutInformation[#, LongForm -> False]& ),
        ButtonNote->"ProPac`ObserverTools`",
        Evaluator->Automatic], ""}
     },
     GridBoxItemSize->{
      "Columns" -> {{All}}, "ColumnsIndexed" -> {}, "Rows" -> {1, {1.05}}, 
       "RowsIndexed" -> {}}]}
  },
  ColumnsEqual->True,
  GridBoxAlignment->{
   "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
    "RowsIndexed" -> {}},
  GridBoxSpacings->{"Columns" -> {
      Offset[0.27999999999999997`], {
       Offset[0.5599999999999999]}, 
      Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
      Offset[0.2], 
      Offset[0.8], {
       Offset[1.2]}, 
      Offset[0.2]}, "RowsIndexed" -> {}}]], "Print",
 CellMargins->{{20, Inherited}, {Inherited, Inherited}},
 ShowCellLabel->False,
 CellFrameMargins->{{Inherited, Inherited}, {14, 14}},
 Background->RGBColor[0.964706, 0.929412, 0.839216],
 ButtonBoxOptions->{Active->True},
 CellTags->{"Observer Tools", "Introduction", "i:1", "b:36"}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Autonomous, Linearly Observable Examples", "Section",
 CellTags->{"Autonomous Examples", "i:2"}],

Cell[CellGroupData[{

Cell["Introduction", "Subsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.0"}],

Cell["\<\
We consider the simplest situation. Each of the following examples involves \
an autonomous system whose linearization at the origin is observable. In \
other words, the observability rank condition is satisfied.\
\>", "Text",
 CellTags->{"Autonomous Examples", "i:2", "b:37.0"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["SISO Example: after Krener & Respondek (Example 7.3)", "Subsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],

Cell[CellGroupData[{

Cell["System", "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],

Cell[BoxData[
 RowBox[{"Clear", "[", 
  RowBox[{"c", ",", "rho"}], "]"}]], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"In[3]:="],

Cell[BoxData[{
 RowBox[{
  RowBox[{"f", "=", 
   RowBox[{"{", 
    RowBox[{"x2", ",", "x3", ",", 
     RowBox[{
      RowBox[{
       RowBox[{"c", "[", "x1", "]"}], " ", 
       RowBox[{
        RowBox[{"x2", "^", "2"}], "/", "2"}]}], "+", 
      RowBox[{
       RowBox[{"rho", "[", "x1", "]"}], " ", "x3"}]}]}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"x", "=", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", "x3"}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"h", "=", 
   RowBox[{"{", "x1", "}"}]}], ";"}]}], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"In[4]:="],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityIndices", "[", 
  RowBox[{"f", ",", "h", ",", "x", ",", 
   RowBox[{"{", "u", "}"}]}], "]"}]], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"In[7]:="],

Cell[BoxData[
 RowBox[{"{", "3", "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"Out[7]="]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["observable form", "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{"Trans", ",", "Ind"}], "}"}], "=", 
  RowBox[{"ObservableForm", "[", 
   RowBox[{"f", ",", "h", ",", "x", ",", 
    RowBox[{"{", "u", "}"}]}], "]"}]}]], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"In[8]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", "x3"}], "}"}], ",", 
   RowBox[{"{", "3", "}"}]}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"Out[8]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
z={z1,z2,z3};
InvTrans=InverseTransformation[x,z,Trans]\
\>", "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"In[9]:="],

Cell[BoxData["\<\"  *** LinearControl successfully loaded ***\"\>"], "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"From In[9]:="],

Cell[BoxData["\<\"  *** NonlinearControl successfully loaded ***\"\>"], \
"Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"From In[9]:="],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Inverse Transformation: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", "x3"}], "}"}], 
   "\[InvisibleSpace]", "\<\" =  \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"z1", ",", "z2", ",", "z3"}], "}"}]}],
  SequenceForm["Inverse Transformation: ", {x1, x2, x3}, " =  ", {z1, z2, z3}],
  Editable->False]], "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"From In[9]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"z1", ",", "z2", ",", "z3"}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"Out[10]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{"newf", ",", "newg", ",", "newh"}], "}"}], "=", 
  RowBox[{"TransformSystem", "[", 
   RowBox[{
    RowBox[{"f", "/.", 
     RowBox[{"u", "\[Rule]", "0"}]}], ",", 
    RowBox[{"Coefficient", "[", 
     RowBox[{"f", ",", "u"}], "]"}], ",", "h", ",", "x", ",", "z", ",", 
    "Trans", ",", "InvTrans"}], "]"}]}]], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"In[11]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"z2", ",", "z3", ",", 
     RowBox[{
      RowBox[{
       FractionBox["1", "2"], " ", 
       SuperscriptBox["z2", "2"], " ", 
       RowBox[{"c", "[", "z1", "]"}]}], "+", 
      RowBox[{"z3", " ", 
       RowBox[{"rho", "[", "z1", "]"}]}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", "z1", "}"}]}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"Out[11]="]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Observer Form", "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"LinearizeToOutputInjection", "[", 
  RowBox[{"f", ",", "h", ",", "x", ",", 
   RowBox[{"{", "u", "}"}]}], "]"}]], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"In[12]:="],

Cell[BoxData["\<\"Integrability conditions are not satisfied.\"\>"], "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"From In[12]:="],

Cell[BoxData[
 RowBox[{"{", "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"Out[12]="]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["A Special Case Linearizable up to Output Injection", \
"Subsubsection"]], "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],

Cell[TextData[{
 "According to the theory in Krener and Respondek, an output transformation \
will not help. They establish that the required condition ",
 Cell[BoxData[
  RowBox[{
   RowBox[{
    RowBox[{"c", "[", "x1", "]"}], "=", " ", 
    RowBox[{
     SuperscriptBox["rho", "\[Prime]",
      MultilineFunction->None], "[", "x1", "]"}]}], ","}]]],
 " but our analysis leads to the necessary condition ",
 Cell[BoxData[
  RowBox[{
   RowBox[{"c", "[", "x1", "]"}], "=", 
   RowBox[{"2", " ", 
    RowBox[{
     SuperscriptBox["rho", "\[Prime]",
      MultilineFunction->None], "[", "x1", "]"}]}]}]]],
 ". "
}], "Text",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"rho", "[", "x1_", "]"}], ":=", 
  RowBox[{"x1", "^", "3"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"c", "[", "x1_", "]"}], ":=", 
  RowBox[{"2", " ", 
   RowBox[{"D", "[", 
    RowBox[{
     RowBox[{"rho", "[", "x1", "]"}], ",", "x1"}], 
    "]"}]}]}], "\[IndentingNewLine]", 
 RowBox[{"Trans", "=", 
  RowBox[{"LinearizeToOutputInjection", "[", 
   RowBox[{"f", ",", "h", ",", "x", ",", 
    RowBox[{"{", "u", "}"}]}], "]"}]}]}], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"In[13]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"x1", ",", 
   RowBox[{
    FractionBox["1", "4"], " ", 
    RowBox[{"(", 
     RowBox[{
      SuperscriptBox["x1", "4"], "-", 
      RowBox[{"4", " ", "x2"}]}], ")"}]}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      SuperscriptBox["x1", "3"]}], " ", "x2"}], "+", "x3"}]}], 
  "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"Out[15]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"z", "=", 
   RowBox[{"{", 
    RowBox[{"z1", ",", "z2", ",", "z3"}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"InvTrans", "=", 
  RowBox[{"InverseTransformation", "[", 
   RowBox[{"x", ",", "z", ",", "Trans"}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"{", 
   RowBox[{"newf", ",", "newg", ",", "newh"}], "}"}], "=", 
  RowBox[{"TransformSystem", "[", 
   RowBox[{
    RowBox[{"f", "/.", 
     RowBox[{"u", "\[Rule]", "0"}]}], ",", 
    RowBox[{"Coefficient", "[", 
     RowBox[{"f", ",", "u"}], "]"}], ",", "h", ",", "x", ",", "z", ",", 
    "Trans", ",", "InvTrans"}], "]"}]}]}], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"In[16]:="],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Inverse Transformation: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", "x3"}], "}"}], 
   "\[InvisibleSpace]", "\<\" =  \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"z1", ",", 
     RowBox[{
      FractionBox["1", "4"], " ", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["z1", "4"], "-", 
        RowBox[{"4", " ", "z2"}]}], ")"}]}], ",", 
     RowBox[{
      FractionBox["1", "4"], " ", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["z1", "7"], "-", 
        RowBox[{"4", " ", 
         SuperscriptBox["z1", "3"], " ", "z2"}], "+", 
        RowBox[{"4", " ", "z3"}]}], ")"}]}]}], "}"}]}],
  SequenceForm[
  "Inverse Transformation: ", {x1, x2, x3}, " =  ", {
   z1, Rational[1, 4] (z1^4 - 4 z2), 
    Rational[1, 4] (z1^7 - 4 z1^3 z2 + 4 z3)}],
  Editable->False]], "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"From In[16]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"z1", ",", 
   RowBox[{
    FractionBox["1", "4"], " ", 
    RowBox[{"(", 
     RowBox[{
      SuperscriptBox["z1", "4"], "-", 
      RowBox[{"4", " ", "z2"}]}], ")"}]}], ",", 
   RowBox[{
    FractionBox[
     SuperscriptBox["z1", "7"], "4"], "-", 
    RowBox[{
     SuperscriptBox["z1", "3"], " ", "z2"}], "+", "z3"}]}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"Out[17]="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      FractionBox["1", "4"], " ", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["z1", "4"], "-", 
        RowBox[{"4", " ", "z2"}]}], ")"}]}], ",", 
     RowBox[{"-", "z3"}], ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", "z1", "}"}]}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"Out[18]="]
}, Open  ]],

Cell["\<\
This example seems to indicate that our condition is correct. For comparison, \
use the Krener-Isidori approach.\
\>", "Text",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Y0", "=", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "1"}], "}"}]}], ";", 
  RowBox[{"Y1", "=", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"-", "1"}], ")"}], 
    RowBox[{"Ad", "[", 
     RowBox[{"f", ",", "Y0", ",", "x"}], "]"}]}]}], ";", 
  RowBox[{"Y2", "=", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"-", "1"}], ")"}], " ", 
    RowBox[{"Ad", "[", 
     RowBox[{"f", ",", "Y1", ",", "x"}], "]"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Transpose", "[", 
   RowBox[{"{", 
    RowBox[{"Y2", ",", "Y1", ",", "Y0"}], "}"}], "]"}], "//", 
  "MatrixForm"}]}], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"In[19]:="],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {"1", "0", "0"},
     {
      SuperscriptBox["x1", "3"], "1", "0"},
     {
      RowBox[{
       SuperscriptBox["x1", "6"], "+", 
       RowBox[{"3", " ", 
        SuperscriptBox["x1", "2"], " ", "x2"}]}], 
      SuperscriptBox["x1", "3"], "1"}
    }], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"Out[20]//MatrixForm="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"W1", "=", 
  RowBox[{"Inverse", "[", 
   RowBox[{"Transpose", "[", 
    RowBox[{"{", 
     RowBox[{"Y2", ",", "Y1", ",", "Y0"}], "}"}], "]"}], "]"}]}]], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"In[21]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", 
      SuperscriptBox["x1", "3"]}], ",", "1", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{"-", "3"}], " ", 
      SuperscriptBox["x1", "2"], " ", "x2"}], ",", 
     RowBox[{"-", 
      SuperscriptBox["x1", "3"]}], ",", "1"}], "}"}]}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"Out[21]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Trans", "=", 
  RowBox[{"Map", "[", 
   RowBox[{
    RowBox[{
     RowBox[{"PDESolve", "[", 
      RowBox[{"#", ",", "x"}], "]"}], "&"}], ",", "W1"}], "]"}]}]], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"In[22]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"x1", ",", 
   RowBox[{
    RowBox[{"-", 
     FractionBox[
      SuperscriptBox["x1", "4"], "4"]}], "+", "x2"}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", 
      SuperscriptBox["x1", "3"]}], " ", "x2"}], "+", "x3"}]}], 
  "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"Out[22]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"z", "=", 
   RowBox[{"{", 
    RowBox[{"z1", ",", "z2", ",", "z3"}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"InvTrans", "=", 
  RowBox[{"InverseTransformation", "[", 
   RowBox[{"x", ",", "z", ",", "Trans"}], "]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"{", 
   RowBox[{"newf", ",", "newg", ",", "newh"}], "}"}], "=", 
  RowBox[{"TransformSystem", "[", 
   RowBox[{
    RowBox[{"f", "/.", 
     RowBox[{"u", "\[Rule]", "0"}]}], ",", 
    RowBox[{"Coefficient", "[", 
     RowBox[{"f", ",", "u"}], "]"}], ",", "h", ",", "x", ",", "z", ",", 
    "Trans", ",", "InvTrans"}], "]"}]}]}], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"In[23]:="],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Inverse Transformation: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", "x3"}], "}"}], 
   "\[InvisibleSpace]", "\<\" =  \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"z1", ",", 
     RowBox[{
      FractionBox["1", "4"], " ", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["z1", "4"], "+", 
        RowBox[{"4", " ", "z2"}]}], ")"}]}], ",", 
     RowBox[{
      FractionBox["1", "4"], " ", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["z1", "7"], "+", 
        RowBox[{"4", " ", 
         SuperscriptBox["z1", "3"], " ", "z2"}], "+", 
        RowBox[{"4", " ", "z3"}]}], ")"}]}]}], "}"}]}],
  SequenceForm[
  "Inverse Transformation: ", {x1, x2, x3}, " =  ", {
   z1, Rational[1, 4] (z1^4 + 4 z2), 
    Rational[1, 4] (z1^7 + 4 z1^3 z2 + 4 z3)}],
  Editable->False]], "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"From In[23]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"z1", ",", 
   RowBox[{
    FractionBox[
     SuperscriptBox["z1", "4"], "4"], "+", "z2"}], ",", 
   RowBox[{
    FractionBox[
     SuperscriptBox["z1", "7"], "4"], "+", 
    RowBox[{
     SuperscriptBox["z1", "3"], " ", "z2"}], "+", "z3"}]}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"Out[24]="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      FractionBox[
       SuperscriptBox["z1", "4"], "4"], "+", "z2"}], ",", "z3", ",", "0"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", "z1", "}"}]}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"},
 CellLabel->"Out[25]="]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["2-Output Example: after Krener & Respondek ", "Subsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],

Cell[CellGroupData[{

Cell["System", "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"a", "[", "x_", "]"}], ":=", 
   RowBox[{"x", " ", 
    RowBox[{"Sin", "[", "x", "]"}]}]}], ";", 
  RowBox[{
   RowBox[{"c", "[", "x_", "]"}], ":=", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"1", "+", "x"}], ")"}], "/", "2"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
    RowBox[{
     RowBox[{"b", "[", "x_", "]"}], ":=", 
     RowBox[{"x", "-", 
      RowBox[{"Cos", "[", "x", "]"}]}]}], ";"}], " ", "*)"}]}]}], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"In[26]:="],

Cell[BoxData[{
 RowBox[{
  RowBox[{"f", "=", 
   RowBox[{"{", 
    RowBox[{"x2", ",", 
     RowBox[{
      RowBox[{"a", "[", "x1", "]"}], "+", 
      RowBox[{
       RowBox[{"a", "[", "x1", "]"}], " ", "x2"}], "+", 
      RowBox[{
       RowBox[{"c", "[", "x1", "]"}], " ", 
       RowBox[{"x2", "^", "2"}]}]}], ",", " ", 
     RowBox[{
      RowBox[{"b", "[", "x1", "]"}], " ", "x2"}]}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"x", "=", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", "x3"}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"h", "=", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x3"}], "}"}]}], ";"}]}], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"In[27]:="],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityIndices", "[", 
  RowBox[{"f", ",", "h", ",", "x", ",", 
   RowBox[{"{", "u", "}"}]}], "]"}]], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"In[30]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"2", ",", "1"}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"Out[30]="]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Observable form", "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{"Trans", ",", "Ind"}], "}"}], "=", 
  RowBox[{"ObservableForm", "[", 
   RowBox[{"f", ",", "h", ",", "x", ",", 
    RowBox[{"{", "u", "}"}]}], "]"}]}]], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"In[31]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", "x3"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2", ",", "1"}], "}"}]}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"Out[31]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
z={z1,z2,z3};
InvTrans=InverseTransformation[x,z,Trans]\
\>", "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"In[32]:="],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Inverse Transformation: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", "x3"}], "}"}], 
   "\[InvisibleSpace]", "\<\" =  \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"z1", ",", "z2", ",", "z3"}], "}"}]}],
  SequenceForm["Inverse Transformation: ", {x1, x2, x3}, " =  ", {z1, z2, z3}],
  Editable->False]], "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"From In[32]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"z1", ",", "z2", ",", "z3"}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"Out[33]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{"newf", ",", "newg", ",", "newh"}], "}"}], "=", 
  RowBox[{"TransformSystem", "[", 
   RowBox[{
    RowBox[{"f", "/.", 
     RowBox[{"u", "\[Rule]", "0"}]}], ",", 
    RowBox[{"Coefficient", "[", 
     RowBox[{"f", ",", "u"}], "]"}], ",", "h", ",", "x", ",", "z", ",", 
    "Trans", ",", "InvTrans"}], "]"}]}]], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"In[34]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"z2", ",", 
     RowBox[{
      RowBox[{
       FractionBox["1", "2"], " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "z1"}], ")"}], " ", 
       SuperscriptBox["z2", "2"]}], "+", 
      RowBox[{"z1", " ", 
       RowBox[{"(", 
        RowBox[{"1", "+", "z2"}], ")"}], " ", 
       RowBox[{"Sin", "[", "z1", "]"}]}]}], ",", 
     RowBox[{"z2", " ", 
      RowBox[{"b", "[", "z1", "]"}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"z1", ",", "z3"}], "}"}]}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"Out[34]="]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Observer form", "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"LinearizeToOutputInjection", "[", 
  RowBox[{"f", ",", "h", ",", "x", ",", 
   RowBox[{"{", "u", "}"}]}], "]"}]], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"In[35]:="],

Cell[BoxData["\<\"dim \[CapitalOmega] = 2, 1 expected.\"\>"], "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"From In[35]:="],

Cell[BoxData["\<\"Integrability conditions are not satisfied.\"\>"], "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"From In[35]:="],

Cell[BoxData[
 RowBox[{"{", "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"Out[35]="]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["A Special Case Linearizable to Output Injection ", "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"a", "[", "x_", "]"}], ":=", 
   RowBox[{"x", " ", "x"}]}], ";", " ", 
  RowBox[{"(*", " ", 
   RowBox[{
    RowBox[{
     RowBox[{"c", "[", "x_", "]"}], ":=", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"1", "+", "x"}], ")"}], "/", "2"}]}], ";"}], " ", "*)"}], 
  RowBox[{
   RowBox[{"c", "[", "x_", "]"}], ":=", "0"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"b", "[", 
    RowBox[{"x1_", ",", "x3_"}], "]"}], ":=", 
   RowBox[{
    RowBox[{"-", "x3"}], "/", 
    RowBox[{"(", 
     RowBox[{"1", "+", "x1"}], ")"}]}]}], ";"}]}], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"In[36]:="],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"\[Psi]2", "[", 
   RowBox[{"x1", ",", "x3"}], "]"}], "=", 
  RowBox[{"x3", " ", "+", 
   RowBox[{"x1", "  ", "x3"}]}]}], "\[IndentingNewLine]", 
 RowBox[{"Eqn", "=", 
  RowBox[{
   RowBox[{
    RowBox[{"D", "[", 
     RowBox[{
      RowBox[{"\[Psi]2", "[", 
       RowBox[{"x1", ",", "x3"}], "]"}], ",", "x1"}], "]"}], "+", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{
       RowBox[{"-", "x3"}], "/", 
       RowBox[{"(", 
        RowBox[{"1", "+", "x1"}], ")"}]}], ")"}], " ", 
     RowBox[{"D", "[", 
      RowBox[{
       RowBox[{"\[Psi]2", "[", 
        RowBox[{"x1", ",", "x3"}], "]"}], ",", "x3"}], "]"}]}]}], "\[Equal]", 
   "0"}]}]}], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"In[38]:="],

Cell[BoxData[
 RowBox[{"x3", "+", 
  RowBox[{"x1", " ", "x3"}]}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"Out[38]="],

Cell[BoxData["True"], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"Out[39]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{"f", "=", 
  RowBox[{"{", 
   RowBox[{"x2", ",", 
    RowBox[{
     RowBox[{"a", "[", "x1", "]"}], "+", 
     RowBox[{
      RowBox[{"a", "[", "x1", "]"}], " ", "x2"}], "+", 
     RowBox[{
      RowBox[{"c", "[", "x1", "]"}], " ", 
      RowBox[{"x2", "^", "2"}]}]}], ",", " ", 
    RowBox[{
     RowBox[{"b", "[", 
      RowBox[{"x1", ",", "x3"}], "]"}], " ", "x2"}]}], 
   "}"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"x", "=", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", "x3"}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"h", "=", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x3"}], "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"h", "=", 
   RowBox[{"{", 
    RowBox[{"x1", ",", 
     RowBox[{"x3", " ", "+", 
      RowBox[{"x1", "  ", "x3"}]}]}], "}"}]}], ";"}]}], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"In[40]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"x2", ",", 
   RowBox[{
    SuperscriptBox["x1", "2"], "+", 
    RowBox[{
     SuperscriptBox["x1", "2"], " ", "x2"}]}], ",", 
   RowBox[{"-", 
    FractionBox[
     RowBox[{"x2", " ", "x3"}], 
     RowBox[{"1", "+", "x1"}]]}]}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"Out[40]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityIndices", "[", 
  RowBox[{"f", ",", "h", ",", "x", ",", 
   RowBox[{"{", "u", "}"}]}], "]"}]], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"In[44]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"2", ",", "1"}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"Out[44]="]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["observable form", "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{"Trans", ",", "Ind"}], "}"}], "=", 
  RowBox[{"ObservableForm", "[", 
   RowBox[{"f", ",", "h", ",", "x", ",", 
    RowBox[{"{", "u", "}"}]}], "]"}]}]], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"In[45]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", 
     RowBox[{"x3", "+", 
      RowBox[{"x1", " ", "x3"}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2", ",", "1"}], "}"}]}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"Out[45]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
z={z1,z2,z3};
InvTrans=InverseTransformation[x,z,Trans]\
\>", "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"In[46]:="],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Inverse Transformation: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", "x3"}], "}"}], 
   "\[InvisibleSpace]", "\<\" =  \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"z1", ",", "z2", ",", 
     FractionBox["z3", 
      RowBox[{"1", "+", "z1"}]]}], "}"}]}],
  SequenceForm[
  "Inverse Transformation: ", {x1, x2, x3}, " =  ", {
   z1, z2, (1 + z1)^(-1) z3}],
  Editable->False]], "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"From In[46]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"z1", ",", "z2", ",", 
   FractionBox["z3", 
    RowBox[{"1", "+", "z1"}]]}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"Out[47]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{"newf", ",", "newg", ",", "newh"}], "}"}], "=", 
  RowBox[{"TransformSystem", "[", 
   RowBox[{
    RowBox[{"f", "/.", 
     RowBox[{"u", "\[Rule]", "0"}]}], ",", 
    RowBox[{"Coefficient", "[", 
     RowBox[{"f", ",", "u"}], "]"}], ",", "h", ",", "x", ",", "z", ",", 
    "Trans", ",", "InvTrans"}], "]"}]}]], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"In[48]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"z2", ",", 
     RowBox[{
      SuperscriptBox["z1", "2"], " ", 
      RowBox[{"(", 
       RowBox[{"1", "+", "z2"}], ")"}]}], ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"z1", ",", "z3"}], "}"}]}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"Out[48]="]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Observer form", "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Trans", "=", 
  RowBox[{"LinearizeToOutputInjection", "[", 
   RowBox[{"f", ",", "h", ",", "x", ",", 
    RowBox[{"{", "u", "}"}]}], "]"}]}]], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"In[49]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"x1", ",", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{"1", "+", "x1"}], ")"}], " ", "x3"}], ",", 
   RowBox[{
    FractionBox["1", "3"], " ", 
    RowBox[{"(", 
     RowBox[{
      SuperscriptBox["x1", "3"], "-", 
      RowBox[{"3", " ", "x2"}]}], ")"}]}]}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"Out[49]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
z={z1,z2,z3};
InvTrans=InverseTransformation[x,z,Trans]\
\>", "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"In[50]:="],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Inverse Transformation: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", "x3"}], "}"}], 
   "\[InvisibleSpace]", "\<\" =  \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"z1", ",", 
     RowBox[{
      FractionBox["1", "3"], " ", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["z1", "3"], "-", 
        RowBox[{"3", " ", "z3"}]}], ")"}]}], ",", 
     FractionBox["z2", 
      RowBox[{"1", "+", "z1"}]]}], "}"}]}],
  SequenceForm[
  "Inverse Transformation: ", {x1, x2, x3}, " =  ", {
   z1, Rational[1, 3] (z1^3 - 3 z3), (1 + z1)^(-1) z2}],
  Editable->False]], "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"From In[50]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"z1", ",", 
   RowBox[{
    FractionBox["1", "3"], " ", 
    RowBox[{"(", 
     RowBox[{
      SuperscriptBox["z1", "3"], "-", 
      RowBox[{"3", " ", "z3"}]}], ")"}]}], ",", 
   FractionBox["z2", 
    RowBox[{"1", "+", "z1"}]]}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"Out[51]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{"newf", ",", "newg", ",", "newh"}], "}"}], "=", 
  RowBox[{"TransformSystem", "[", 
   RowBox[{
    RowBox[{"f", "/.", 
     RowBox[{"u", "\[Rule]", "0"}]}], ",", 
    RowBox[{"Coefficient", "[", 
     RowBox[{"f", ",", "u"}], "]"}], ",", "h", ",", "x", ",", "z", ",", 
    "Trans", ",", "InvTrans"}], "]"}]}]], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"In[52]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      FractionBox["1", "3"], " ", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["z1", "3"], "-", 
        RowBox[{"3", " ", "z3"}]}], ")"}]}], ",", "0", ",", 
     RowBox[{"-", 
      SuperscriptBox["z1", "2"]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"z1", ",", "z2"}], "}"}]}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"},
 CellLabel->"Out[52]="]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["2-Output Example: Hou & Pugh", "Subsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],

Cell[CellGroupData[{

Cell["System", "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"f", "=", 
   RowBox[{"{", 
    RowBox[{"x2", ",", 
     RowBox[{"x3", " ", "x2"}], ",", "x2"}], "}"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"CC", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"0", ",", "1"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "0"}], "}"}]}], "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"h", "=", 
   RowBox[{"CC", ".", 
    RowBox[{"{", 
     RowBox[{"x1", ",", "x3"}], "}"}]}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"x", "=", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", "x3"}], "}"}]}], ";"}], "\n", 
 RowBox[{"ObservabilityIndices", "[", 
  RowBox[{"f", ",", "h", ",", "x", ",", 
   RowBox[{"{", "u", "}"}]}], "]"}]}], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"},
 CellLabel->"In[53]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"2", ",", "1"}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"},
 CellLabel->"Out[57]="]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Observable Form", "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{"Trans", ",", "Ind"}], "}"}], "=", 
  RowBox[{"ObservableForm", "[", 
   RowBox[{"f", ",", "h", ",", "x", ",", 
    RowBox[{"{", "u", "}"}]}], "]"}]}]], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"},
 CellLabel->"In[58]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"x3", ",", "x2", ",", "x1"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2", ",", "1"}], "}"}]}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"},
 CellLabel->"Out[58]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
z={z1,z2,z3};
InvTrans=InverseTransformation[x,z,Trans]\
\>", "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"},
 CellLabel->"In[59]:="],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Inverse Transformation: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", "x3"}], "}"}], 
   "\[InvisibleSpace]", "\<\" =  \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"z3", ",", "z2", ",", "z1"}], "}"}]}],
  SequenceForm["Inverse Transformation: ", {x1, x2, x3}, " =  ", {z3, z2, z1}],
  Editable->False]], "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"},
 CellLabel->"From In[59]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"z3", ",", "z2", ",", "z1"}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"},
 CellLabel->"Out[60]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
{newf,newg,newh}=TransformSystem[f/.u\[Rule]0,Coefficient[f,u],h,x,z,Trans,\
InvTrans]
newf+newg u\
\>", "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"},
 CellLabel->"In[61]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"z2", ",", 
     RowBox[{"z1", " ", "z2"}], ",", "z2"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"z1", ",", "z3"}], "}"}]}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"},
 CellLabel->"Out[61]="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"z2", ",", 
   RowBox[{"z1", " ", "z2"}], ",", "z2"}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"},
 CellLabel->"Out[62]="]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Observer Form", "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Trans", "=", 
  RowBox[{"LinearizeToOutputInjection", "[", 
   RowBox[{"f", ",", "h", ",", "x", ",", 
    RowBox[{"{", "u", "}"}]}], "]"}]}]], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"},
 CellLabel->"In[63]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"x3", ",", "x1", ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", "2"}], " ", "x2"}], "+", 
    SuperscriptBox["x3", "2"]}]}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"},
 CellLabel->"Out[63]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"z", "=", 
   RowBox[{"{", 
    RowBox[{"z1", ",", "z2", ",", "z3"}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{"InvTrans", "=", 
  RowBox[{"InverseTransformation", "[", 
   RowBox[{"x", ",", "z", ",", "Trans"}], "]"}]}]}], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"},
 CellLabel->"In[64]:="],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Inverse Transformation: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", "x3"}], "}"}], 
   "\[InvisibleSpace]", "\<\" =  \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"z2", ",", 
     RowBox[{
      FractionBox["1", "2"], " ", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["z1", "2"], "-", "z3"}], ")"}]}], ",", "z1"}], "}"}]}],
  SequenceForm[
  "Inverse Transformation: ", {x1, x2, x3}, " =  ", {
   z2, Rational[1, 2] (z1^2 - z3), z1}],
  Editable->False]], "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"},
 CellLabel->"From In[64]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"z2", ",", 
   RowBox[{
    FractionBox["1", "2"], " ", 
    RowBox[{"(", 
     RowBox[{
      SuperscriptBox["z1", "2"], "-", "z3"}], ")"}]}], ",", "z1"}], 
  "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"},
 CellLabel->"Out[65]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{"newf", ",", "newg", ",", "newh"}], "}"}], "=", 
  RowBox[{"TransformSystem", "[", 
   RowBox[{
    RowBox[{"f", "/.", 
     RowBox[{"u", "\[Rule]", "0"}]}], ",", 
    RowBox[{"Coefficient", "[", 
     RowBox[{"f", ",", "u"}], "]"}], ",", "h", ",", "x", ",", "z", ",", 
    "Trans", ",", "InvTrans"}], "]"}]}]], "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"},
 CellLabel->"In[66]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      FractionBox["1", "2"], " ", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["z1", "2"], "-", "z3"}], ")"}]}], ",", 
     RowBox[{
      FractionBox["1", "2"], " ", 
      RowBox[{"(", 
       RowBox[{
        SuperscriptBox["z1", "2"], "-", "z3"}], ")"}]}], ",", "0"}], "}"}], 
   ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"z1", ",", "z2"}], "}"}]}], "}"}]], "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"},
 CellLabel->"Out[66]="]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Autonomous: locally observable, not linearly observable", "Section",
 CellTags->{"Failure of Rank Condition", "i:2"}],

Cell[CellGroupData[{

Cell["Xia & Zeitz example 2", "Subsection",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],

Cell[CellGroupData[{

Cell["System", "Subsubsection",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],

Cell["\<\
In the following example, the autonomous system is locally observable but it \
is not linearly observable at the origin, i.e., the observability rank \
condition fails. In contrast to the previous examples, the transformations \
are merely continuous, not smooth.\
\>", "Text",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"f", "=", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2"}], "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"h", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"x1", "^", "3"}], "+", 
     RowBox[{"x2", "^", "5"}]}], "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"x", "=", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2"}], "}"}]}], ";"}]}], "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"},
 CellLabel->"In[67]:="],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityIndices", "[", 
  RowBox[{"f", ",", "h", ",", "x", ",", 
   RowBox[{"{", "u", "}"}]}], "]"}]], "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"},
 CellLabel->"In[70]:="],

Cell[BoxData[
 RowBox[{"{", "2", "}"}]], "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"},
 CellLabel->"Out[70]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityCodistribution", "[", 
  RowBox[{"f", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0"}], "}"}], ",", "h", ",", "x"}], "]"}]], "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"},
 CellLabel->"In[71]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1"}], "}"}]}], "}"}]], "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"},
 CellLabel->"Out[71]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityCodistribution", "[", 
  RowBox[{"f", ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0"}], "}"}], ",", "h", ",", "x", ",", 
   RowBox[{"0", "*", "x"}]}], "]"}]], "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"},
 CellLabel->"In[72]:="],

Cell[BoxData[
 RowBox[{"{", "}"}]], "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"},
 CellLabel->"Out[72]="]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Observable form", "Subsubsection",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{"Trans", ",", "Ind"}], "}"}], "=", 
  RowBox[{"ObservableForm", "[", 
   RowBox[{"f", ",", "h", ",", "x", ",", 
    RowBox[{"{", "u", "}"}]}], "]"}]}]], "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"},
 CellLabel->"In[73]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      SuperscriptBox["x1", "3"], "+", 
      SuperscriptBox["x2", "5"]}], ",", 
     RowBox[{
      RowBox[{"3", " ", 
       SuperscriptBox["x1", "3"]}], "+", 
      RowBox[{"5", " ", 
       SuperscriptBox["x2", "5"]}]}]}], "}"}], ",", 
   RowBox[{"{", "2", "}"}]}], "}"}]], "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"},
 CellLabel->"Out[73]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
z={z1,z2};
InvTrans=InverseTransformation[x,z,Trans]\
\>", "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"},
 CellLabel->"In[74]:="],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Inverse Transformation: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2"}], "}"}], "\[InvisibleSpace]", "\<\" =  \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"-", 
          FractionBox["1", "2"]}], ")"}], 
        RowBox[{"1", "/", "3"}]]}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         RowBox[{"5", " ", "z1"}], "-", "z2"}], ")"}], 
       RowBox[{"1", "/", "3"}]]}], ",", 
     RowBox[{
      RowBox[{"-", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"-", 
          FractionBox["1", "2"]}], ")"}], 
        RowBox[{"1", "/", "5"}]]}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "3"}], " ", "z1"}], "+", "z2"}], ")"}], 
       RowBox[{"1", "/", "5"}]]}]}], "}"}]}],
  SequenceForm[
  "Inverse Transformation: ", {x1, x2}, 
   " =  ", {-Rational[-1, 2]^Rational[1, 3] (5 z1 - z2)^Rational[1, 3], -
    Rational[-1, 2]^Rational[1, 5] ((-3) z1 + z2)^Rational[1, 5]}],
  Editable->False]], "Print",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"},
 CellLabel->"From In[74]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"-", 
        FractionBox["1", "2"]}], ")"}], 
      RowBox[{"1", "/", "3"}]]}], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       RowBox[{"5", " ", "z1"}], "-", "z2"}], ")"}], 
     RowBox[{"1", "/", "3"}]]}], ",", 
   RowBox[{
    RowBox[{"-", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"-", 
        FractionBox["1", "2"]}], ")"}], 
      RowBox[{"1", "/", "5"}]]}], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "3"}], " ", "z1"}], "+", "z2"}], ")"}], 
     RowBox[{"1", "/", "5"}]]}]}], "}"}]], "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"},
 CellLabel->"Out[75]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
{newf,newg,newh}=TransformSystem[f/.u\[Rule]0,Coefficient[f,u],h,x,z,Trans,\
InvTrans]\
\>", "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"},
 CellLabel->"In[76]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"z2", ",", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "15"}], " ", "z1"}], "+", 
      RowBox[{"8", " ", "z2"}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", "z1", "}"}]}], "}"}]], "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"},
 CellLabel->"Out[76]="]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Observer form", "Subsubsection",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Trans", "=", 
  RowBox[{"LinearizeToOutputInjection", "[", 
   RowBox[{"f", ",", "h", ",", "x", ",", 
    RowBox[{"{", "u", "}"}]}], "]"}]}]], "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"},
 CellLabel->"In[77]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    SuperscriptBox["x1", "3"], "+", 
    SuperscriptBox["x2", "5"]}], ",", 
   RowBox[{
    RowBox[{"5", " ", 
     SuperscriptBox["x1", "3"]}], "+", 
    RowBox[{"3", " ", 
     SuperscriptBox["x2", "5"]}]}]}], "}"}]], "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"},
 CellLabel->"Out[77]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
z={z1,z2};
InvTrans=InverseTransformation[x,z,Trans]\
\>", "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"},
 CellLabel->"In[78]:="],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Inverse Transformation: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2"}], "}"}], "\[InvisibleSpace]", "\<\" =  \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"-", 
          FractionBox["1", "2"]}], ")"}], 
        RowBox[{"1", "/", "3"}]]}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", "3"}], " ", "z1"}], "+", "z2"}], ")"}], 
       RowBox[{"1", "/", "3"}]]}], ",", 
     RowBox[{
      RowBox[{"-", 
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{"-", 
          FractionBox["1", "2"]}], ")"}], 
        RowBox[{"1", "/", "5"}]]}], " ", 
      SuperscriptBox[
       RowBox[{"(", 
        RowBox[{
         RowBox[{"5", " ", "z1"}], "-", "z2"}], ")"}], 
       RowBox[{"1", "/", "5"}]]}]}], "}"}]}],
  SequenceForm[
  "Inverse Transformation: ", {x1, x2}, 
   " =  ", {-Rational[-1, 2]^Rational[1, 3] ((-3) z1 + z2)^Rational[1, 3], -
    Rational[-1, 2]^Rational[1, 5] (5 z1 - z2)^Rational[1, 5]}],
  Editable->False]], "Print",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"},
 CellLabel->"From In[78]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"-", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"-", 
        FractionBox["1", "2"]}], ")"}], 
      RowBox[{"1", "/", "3"}]]}], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       RowBox[{
        RowBox[{"-", "3"}], " ", "z1"}], "+", "z2"}], ")"}], 
     RowBox[{"1", "/", "3"}]]}], ",", 
   RowBox[{
    RowBox[{"-", 
     SuperscriptBox[
      RowBox[{"(", 
       RowBox[{"-", 
        FractionBox["1", "2"]}], ")"}], 
      RowBox[{"1", "/", "5"}]]}], " ", 
    SuperscriptBox[
     RowBox[{"(", 
      RowBox[{
       RowBox[{"5", " ", "z1"}], "-", "z2"}], ")"}], 
     RowBox[{"1", "/", "5"}]]}]}], "}"}]], "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"},
 CellLabel->"Out[79]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
{newf,newg,newh}=TransformSystem[f/.u\[Rule]0,Coefficient[f,u],h,x,z,Trans,\
InvTrans]\
\>", "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"},
 CellLabel->"In[80]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{"8", " ", "z1"}], "-", "z2"}], ",", 
     RowBox[{"15", " ", "z1"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", "z1", "}"}]}], "}"}]], "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"},
 CellLabel->"Out[80]="]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
Nonautonomous: zero-input observable, not linearly observable\
\>", "Section",
 CellTags->{"Failure of Rank Condition", "i:2"}],

Cell[CellGroupData[{

Cell["Xia & Zeitz example 3", "Subsection",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],

Cell[CellGroupData[{

Cell["System", "Subsubsection",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],

Cell["\<\
Here we have a nonautonomous system. It is observable without any input, but \
it is not linearly observable.\
\>", "Text",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"p", "=", "3"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"f", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"x2", "^", "p"}], ",", 
     RowBox[{"x2", " ", "u"}]}], "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"h", "=", 
   RowBox[{"{", "x1", "}"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"x", "=", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2"}], "}"}]}], ";"}]}], "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"In[81]:="],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityIndices", "[", 
  RowBox[{"f", ",", "h", ",", "x", ",", 
   RowBox[{"{", "u", "}"}]}], "]"}]], "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"In[85]:="],

Cell[BoxData[
 RowBox[{"{", "2", "}"}]], "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"Out[85]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityCodistribution", "[", 
  RowBox[{
   RowBox[{"f", "/.", 
    RowBox[{"u", "\[Rule]", "0"}]}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "x2"}], "}"}], ",", "h", ",", "x"}], "]"}]], "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"In[86]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1"}], "}"}]}], "}"}]], "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"Out[86]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityCodistribution", "[", 
  RowBox[{
   RowBox[{"f", "/.", 
    RowBox[{"u", "\[Rule]", "0"}]}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "x2"}], "}"}], ",", "h", ",", "x", ",", 
   RowBox[{"0", "*", "x"}]}], "]"}]], "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"In[87]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"{", 
   RowBox[{"1", ",", "0"}], "}"}], "}"}]], "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"Out[87]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityCodistribution", "[", 
  RowBox[{
   RowBox[{"f", "/.", 
    RowBox[{"u", "\[Rule]", "0"}]}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0"}], "}"}], ",", "h", ",", "x"}], "]"}]], "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"In[88]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1"}], "}"}]}], "}"}]], "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"Out[88]="]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Observable form", "Subsubsection",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{"Trans", ",", "Ind"}], "}"}], "=", 
  RowBox[{"ObservableForm", "[", 
   RowBox[{"f", ",", "h", ",", "x", ",", 
    RowBox[{"{", "u", "}"}]}], "]"}]}]], "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"In[89]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"x1", ",", 
     SuperscriptBox["x2", "3"]}], "}"}], ",", 
   RowBox[{"{", "2", "}"}]}], "}"}]], "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"Out[89]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
z={z1,z2};
InvTrans=InverseTransformation[x,z,Trans]\
\>", "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"In[90]:="],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Inverse Transformation: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2"}], "}"}], "\[InvisibleSpace]", "\<\" =  \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"z1", ",", 
     SuperscriptBox["z2", 
      RowBox[{"1", "/", "3"}]]}], "}"}]}],
  SequenceForm[
  "Inverse Transformation: ", {x1, x2}, " =  ", {z1, z2^Rational[1, 3]}],
  Editable->False]], "Print",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"From In[90]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"z1", ",", 
   SuperscriptBox["z2", 
    RowBox[{"1", "/", "3"}]]}], "}"}]], "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"Out[91]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
{newf,newg,newh}=TransformSystem[f/.u\[Rule]0,Coefficient[f,u],h,x,z,Trans,\
InvTrans]\
\>", "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"In[92]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"z2", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", 
     RowBox[{"3", " ", "z2"}]}], "}"}], ",", 
   RowBox[{"{", "z1", "}"}]}], "}"}]], "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"Out[92]="]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Observer form", "Subsubsection",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Trans", "=", 
  RowBox[{"LinearizeToOutputInjection", "[", 
   RowBox[{"f", ",", "h", ",", "x", ",", 
    RowBox[{"{", "u", "}"}]}], "]"}]}]], "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"In[93]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"x1", ",", 
   RowBox[{"-", 
    SuperscriptBox["x2", "3"]}]}], "}"}]], "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"Out[93]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
z={z1,z2};
InvTrans=InverseTransformation[x,z,Trans]\
\>", "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"In[94]:="],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Inverse Transformation: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2"}], "}"}], "\[InvisibleSpace]", "\<\" =  \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"z1", ",", 
     RowBox[{"-", 
      SuperscriptBox["z2", 
       RowBox[{"1", "/", "3"}]]}]}], "}"}]}],
  SequenceForm[
  "Inverse Transformation: ", {x1, x2}, " =  ", {z1, -z2^Rational[1, 3]}],
  Editable->False]], "Print",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"From In[94]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"z1", ",", 
   RowBox[{"-", 
    SuperscriptBox["z2", 
     RowBox[{"1", "/", "3"}]]}]}], "}"}]], "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"Out[95]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
{newf,newg,newh}=TransformSystem[f/.u\[Rule]0,Coefficient[f,u],h,x,z,Trans,\
InvTrans]\
\>", "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"In[96]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "z2"}], ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", 
     RowBox[{"3", " ", "z2"}]}], "}"}], ",", 
   RowBox[{"{", "z1", "}"}]}], "}"}]], "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"},
 CellLabel->"Out[96]="]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Nonautonomous: not observable for all inputs", "Section",
 CellTags->{"Zero-input Unobservable", "i:2"}],

Cell[CellGroupData[{

Cell["Inroduction", "Subsection",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.0"}],

Cell["\<\
Now, we give three examples of nonautonomous systems that are not zero-input \
observable. In each case the observability rank condition is satisfied.\
\>", "Text",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.0"}]
}, Open  ]],

Cell[CellGroupData[{

Cell["SISO Example: Hammouri", "Subsection",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],

Cell["\<\
f = {Exp[x1+x2]-1+u x1^2,-Exp[x1+x2]+1+u \
(Exp[x3-x2]-Exp[-x1-x2]-x1^2),-Exp[x1+x2]+1+x1^3 Exp[-x1-x3]-u x1^2};
h=x1;
x={x1,x2,x3};\
\>", "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"In[97]:="],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityIndices", "[", 
  RowBox[{"f", ",", 
   RowBox[{"{", "h", "}"}], ",", "x", ",", 
   RowBox[{"{", "u", "}"}]}], "]"}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"In[100]:="],

Cell[BoxData[
 RowBox[{"{", "3", "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"Out[100]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityIndices", "[", 
  RowBox[{
   RowBox[{"f", "/.", 
    RowBox[{"u", "\[Rule]", "0"}]}], ",", 
   RowBox[{"{", "h", "}"}], ",", "x", ",", 
   RowBox[{"{", "u", "}"}]}], "]"}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"In[101]:="],

Cell[BoxData["\<\"codimension = 0 at k= 2! System is not observable.\"\>"], \
"Print",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"From In[101]:="],

Cell[BoxData[
 RowBox[{"{", "1", "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"Out[101]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityCodistribution", "[", 
  RowBox[{
   RowBox[{"f", "/.", 
    RowBox[{"u", "\[Rule]", "0"}]}], ",", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"f", ",", "u"}], "]"}], ",", 
   RowBox[{"{", "h", "}"}], ",", "x"}], "]"}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"In[102]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "1"}], "}"}]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"Out[102]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityCodistribution", "[", 
  RowBox[{
   RowBox[{"f", "/.", 
    RowBox[{"u", "\[Rule]", "0"}]}], ",", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"f", ",", "u"}], "]"}], ",", 
   RowBox[{"{", "h", "}"}], ",", "x", ",", 
   RowBox[{"0", "*", "x"}]}], "]"}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"In[103]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "1"}], "}"}]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"Out[103]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityCodistribution", "[", 
  RowBox[{
   RowBox[{"f", "/.", 
    RowBox[{"u", "\[Rule]", "0"}]}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", "h", "}"}], ",", "x"}], "]"}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"In[104]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "0", ",", "0"}], "}"}]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"Out[104]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["observable form", "Subsubsection",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{"Trans", ",", "Ind"}], "}"}], "=", 
  RowBox[{"ObservableForm", "[", 
   RowBox[{"f", ",", 
    RowBox[{"{", "h", "}"}], ",", "x", ",", 
    RowBox[{"{", "u", "}"}]}], "]"}]}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"In[105]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"x1", ",", 
     RowBox[{
      RowBox[{"-", "1"}], "+", 
      SuperscriptBox["\[ExponentialE]", 
       RowBox[{"x1", "+", "x2"}]]}], ",", 
     RowBox[{
      RowBox[{
       SuperscriptBox["\[ExponentialE]", 
        RowBox[{"x1", "+", "x2"}]], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         SuperscriptBox["\[ExponentialE]", 
          RowBox[{
           RowBox[{"-", "x1"}], "-", "x2"}]], "-", 
         SuperscriptBox["\[ExponentialE]", 
          RowBox[{"x1", "+", "x2"}]], "+", 
         SuperscriptBox["\[ExponentialE]", 
          RowBox[{
           RowBox[{"-", "x2"}], "+", "x3"}]], "-", 
         SuperscriptBox["x1", "2"]}], ")"}]}], "+", 
      RowBox[{
       SuperscriptBox["\[ExponentialE]", 
        RowBox[{"x1", "+", "x2"}]], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", 
         SuperscriptBox["\[ExponentialE]", 
          RowBox[{"x1", "+", "x2"}]], "+", 
         SuperscriptBox["x1", "2"]}], ")"}]}]}]}], "}"}], ",", 
   RowBox[{"{", "3", "}"}]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"Out[105]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
z={z1,z2,z3};
InvTrans=InverseTransformation[x,z,Trans]\
\>", "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"In[106]:="],

Cell[BoxData[
 RowBox[{
  RowBox[{"Solve", "::", "\<\"ifun\"\>"}], 
  RowBox[{
  ":", " "}], "\<\"Inverse functions are being used by \\!\\(Solve\\), so \
some solutions may not be found; use Reduce for complete solution \
information. \\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", \
ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"Solve::ifun\\\"]\\)\"\>"}]], "Message",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"From In[106]:="],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Inverse Transformation: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", "x3"}], "}"}], 
   "\[InvisibleSpace]", "\<\" =  \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"z1", ",", 
     RowBox[{
      RowBox[{"-", "z1"}], "+", 
      RowBox[{"Log", "[", 
       RowBox[{"1", "+", "z2"}], "]"}]}], ",", 
     RowBox[{
      RowBox[{"-", "z1"}], "+", 
      RowBox[{"Log", "[", 
       RowBox[{"1", "+", "z3"}], "]"}]}]}], "}"}]}],
  SequenceForm[
  "Inverse Transformation: ", {x1, x2, x3}, " =  ", {
   z1, -z1 + Log[1 + z2], -z1 + Log[1 + z3]}],
  Editable->False]], "Print",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"From In[106]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"z1", ",", 
   RowBox[{
    RowBox[{"-", "z1"}], "+", 
    RowBox[{"Log", "[", 
     RowBox[{"1", "+", "z2"}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"-", "z1"}], "+", 
    RowBox[{"Log", "[", 
     RowBox[{"1", "+", "z3"}], "]"}]}]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"Out[107]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{"newf", ",", "newg", ",", "newh"}], "}"}], "=", 
  RowBox[{"TransformSystem", "[", 
   RowBox[{
    RowBox[{"f", "/.", 
     RowBox[{"u", "\[Rule]", "0"}]}], ",", 
    RowBox[{"Coefficient", "[", 
     RowBox[{"f", ",", "u"}], "]"}], ",", "h", ",", "x", ",", "z", ",", 
    "Trans", ",", "InvTrans"}], "]"}]}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"In[108]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"z2", ",", "0", ",", 
     SuperscriptBox["z1", "3"]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     SuperscriptBox["z1", "2"], ",", "z3", ",", "0"}], "}"}], ",", "z1"}], 
  "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"Out[108]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["newf+newg u", "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"In[109]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"u", " ", 
     SuperscriptBox["z1", "2"]}], "+", "z2"}], ",", 
   RowBox[{"u", " ", "z3"}], ",", 
   SuperscriptBox["z1", "3"]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"Out[109]="]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Observer form", "Subsubsection",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Trans", "=", 
  RowBox[{"LinearizeToOutputInjection", "[", 
   RowBox[{"f", ",", 
    RowBox[{"{", "h", "}"}], ",", "x", ",", 
    RowBox[{"{", "u", "}"}]}], "]"}]}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"In[110]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"x1", ",", 
   RowBox[{"1", "-", 
    SuperscriptBox["\[ExponentialE]", 
     RowBox[{"x1", "+", "x2"}]]}], ",", 
   RowBox[{
    RowBox[{"-", "1"}], "+", 
    SuperscriptBox["\[ExponentialE]", 
     RowBox[{"x1", "+", "x3"}]]}]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"Out[110]="]
}, Open  ]],

Cell[BoxData[
 RowBox[{"(*", " ", 
  RowBox[{"Trans", "=", 
   RowBox[{"Trans", "-", 
    RowBox[{"(", 
     RowBox[{"Trans", "/.", 
      RowBox[{"Inner", "[", 
       RowBox[{"Rule", ",", "x", ",", 
        RowBox[{"0", "*", "x"}], ",", "List"}], "]"}]}], ")"}]}]}], " ", 
  "*)"}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"In[111]:="],

Cell[CellGroupData[{

Cell["\<\
z={z1,z2,z3};
InvTrans=InverseTransformation[x,z,Trans]\
\>", "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"In[112]:="],

Cell[BoxData[
 RowBox[{
  RowBox[{"Solve", "::", "\<\"ifun\"\>"}], 
  RowBox[{
  ":", " "}], "\<\"Inverse functions are being used by \\!\\(Solve\\), so \
some solutions may not be found; use Reduce for complete solution \
information. \\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", \
ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"Solve::ifun\\\"]\\)\"\>"}]], "Message",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"From In[112]:="],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Inverse Transformation: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", "x3"}], "}"}], 
   "\[InvisibleSpace]", "\<\" =  \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"z1", ",", 
     RowBox[{
      RowBox[{"-", "z1"}], "+", 
      RowBox[{"Log", "[", 
       RowBox[{"1", "-", "z2"}], "]"}]}], ",", 
     RowBox[{
      RowBox[{"-", "z1"}], "+", 
      RowBox[{"Log", "[", 
       RowBox[{"1", "+", "z3"}], "]"}]}]}], "}"}]}],
  SequenceForm[
  "Inverse Transformation: ", {x1, x2, x3}, " =  ", {
   z1, -z1 + Log[1 - z2], -z1 + Log[1 + z3]}],
  Editable->False]], "Print",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"From In[112]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"z1", ",", 
   RowBox[{
    RowBox[{"-", "z1"}], "+", 
    RowBox[{"Log", "[", 
     RowBox[{"1", "-", "z2"}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"-", "z1"}], "+", 
    RowBox[{"Log", "[", 
     RowBox[{"1", "+", "z3"}], "]"}]}]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"Out[113]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
{newf,newg,newh}=TransformSystem[f/.u\[Rule]0,Coefficient[f,u],h,x,z,Trans,\
InvTrans]\
\>", "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"In[114]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "z2"}], ",", "0", ",", 
     SuperscriptBox["z1", "3"]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     SuperscriptBox["z1", "2"], ",", 
     RowBox[{"-", "z3"}], ",", "0"}], "}"}], ",", "z1"}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"Out[114]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["newf+newg u", "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"In[115]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{
    RowBox[{"u", " ", 
     SuperscriptBox["z1", "2"]}], "-", "z2"}], ",", 
   RowBox[{
    RowBox[{"-", "u"}], " ", "z3"}], ",", 
   SuperscriptBox["z1", "3"]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"},
 CellLabel->"Out[115]="]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["SISO Example: Besancon & Bornard", "Subsection",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"f", "=", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"x2", " ", 
      RowBox[{"Exp", "[", 
       RowBox[{"-", "x1"}], "]"}]}], ",", 
     RowBox[{
      RowBox[{
       RowBox[{"x2", "^", "2"}], " ", 
       RowBox[{"Exp", "[", 
        RowBox[{"-", "x1"}], "]"}]}], "+", 
      RowBox[{"x4", " ", 
       RowBox[{"Exp", "[", 
        RowBox[{"2", " ", "x1"}], "]"}], " ", "u"}]}], ",", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"x1", "^", "2"}], "-", 
        RowBox[{
         RowBox[{"x4", "^", "2"}], " ", 
         RowBox[{"Exp", "[", 
          RowBox[{"2", " ", "x1"}], "]"}]}], "-", 
        RowBox[{"x2", " ", "x3", " ", 
         RowBox[{"Exp", "[", 
          RowBox[{"-", "x1"}], "]"}]}], "-", 
        RowBox[{
         RowBox[{"x2", "^", "2"}], " ", "x4", " ", 
         RowBox[{"Exp", "[", 
          RowBox[{"-", "x1"}], "]"}]}]}], ")"}], " ", "u"}], ",", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "x2"}], " ", "x4", " ", 
       RowBox[{"Exp", "[", 
        RowBox[{"-", "x1"}], "]"}]}], "+", 
      RowBox[{
       RowBox[{"(", 
        RowBox[{"x3", "+", 
         RowBox[{"x2", " ", "x4"}]}], ")"}], " ", 
       RowBox[{"Exp", "[", 
        RowBox[{"-", "x1"}], "]"}], " ", "u"}]}]}], "}"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"h", "=", "x1"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"x", "=", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", "x3", ",", "x4"}], "}"}]}], ";"}]}], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"In[116]:="],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityIndices", "[", 
  RowBox[{"f", ",", 
   RowBox[{"{", "h", "}"}], ",", "x", ",", 
   RowBox[{"{", "u", "}"}]}], "]"}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"In[119]:="],

Cell[BoxData[
 RowBox[{"{", "4", "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"Out[119]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityIndices", "[", 
  RowBox[{
   RowBox[{"f", "/.", 
    RowBox[{"u", "\[Rule]", "0"}]}], ",", 
   RowBox[{"{", "h", "}"}], ",", "x", ",", 
   RowBox[{"{", "u", "}"}]}], "]"}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"In[120]:="],

Cell[BoxData["\<\"codimension = 0 at k= 2! System is not observable.\"\>"], \
"Print",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"From In[120]:="],

Cell[BoxData[
 RowBox[{"{", "1", "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"Out[120]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityCodistribution", "[", 
  RowBox[{
   RowBox[{"f", "/.", 
    RowBox[{"u", "\[Rule]", "0"}]}], ",", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"f", ",", "u"}], "]"}], ",", "h", ",", "x"}], "]"}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"In[121]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "1", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0", ",", "1"}], "}"}]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"Out[121]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityCodistribution", "[", 
  RowBox[{
   RowBox[{"f", "/.", 
    RowBox[{"u", "\[Rule]", "0"}]}], ",", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"f", ",", "u"}], "]"}], ",", "h", ",", "x", ",", 
   RowBox[{"0", "*", "x"}]}], "]"}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"In[122]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "1", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0", ",", "1"}], "}"}]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"Out[122]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityCodistribution", "[", 
  RowBox[{
   RowBox[{"f", "/.", 
    RowBox[{"u", "\[Rule]", "0"}]}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", "h", ",", "x"}], 
  "]"}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"In[123]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "0", ",", "0", ",", "0"}], "}"}]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"Out[123]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["observable form", "Subsubsection",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{"Trans", ",", "Ind"}], "}"}], "=", 
  RowBox[{"ObservableForm", "[", 
   RowBox[{"f", ",", 
    RowBox[{"{", "h", "}"}], ",", "x", ",", 
    RowBox[{"{", "u", "}"}]}], "]"}]}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"In[124]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"x1", ",", 
     RowBox[{
      SuperscriptBox["\[ExponentialE]", 
       RowBox[{"-", "x1"}]], " ", "x2"}], ",", 
     RowBox[{
      RowBox[{
       RowBox[{"-", 
        SuperscriptBox["\[ExponentialE]", 
         RowBox[{
          RowBox[{"-", "2"}], " ", "x1"}]]}], " ", 
       SuperscriptBox["x2", "2"]}], "+", 
      RowBox[{
       SuperscriptBox["\[ExponentialE]", 
        RowBox[{"-", "x1"}]], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{"-", "x1"}]], " ", 
          SuperscriptBox["x2", "2"]}], "+", 
         RowBox[{
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{"2", " ", "x1"}]], " ", "x4"}]}], ")"}]}]}], ",", 
     RowBox[{
      RowBox[{
       SuperscriptBox["\[ExponentialE]", 
        RowBox[{"-", "x1"}]], " ", "x2", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"2", " ", 
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{
            RowBox[{"-", "2"}], " ", "x1"}]], " ", 
          SuperscriptBox["x2", "2"]}], "-", 
         RowBox[{
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{"-", "x1"}]], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             SuperscriptBox["\[ExponentialE]", 
              RowBox[{"-", "x1"}]], " ", 
             SuperscriptBox["x2", "2"]}], "+", 
            RowBox[{
             SuperscriptBox["\[ExponentialE]", 
              RowBox[{"2", " ", "x1"}]], " ", "x4"}]}], ")"}]}], "+", 
         RowBox[{
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{"-", "x1"}]], " ", 
          RowBox[{"(", 
           RowBox[{
            RowBox[{
             RowBox[{"-", 
              SuperscriptBox["\[ExponentialE]", 
               RowBox[{"-", "x1"}]]}], " ", 
             SuperscriptBox["x2", "2"]}], "+", 
            RowBox[{"2", " ", 
             SuperscriptBox["\[ExponentialE]", 
              RowBox[{"2", " ", "x1"}]], " ", "x4"}]}], ")"}]}]}], ")"}]}], 
      "+", 
      RowBox[{
       SuperscriptBox["\[ExponentialE]", "x1"], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          RowBox[{"-", 
           SuperscriptBox["\[ExponentialE]", 
            RowBox[{"-", "x1"}]]}], " ", "x2", " ", "x4"}], "+", 
         RowBox[{
          SuperscriptBox["\[ExponentialE]", 
           RowBox[{"-", "x1"}]], " ", 
          RowBox[{"(", 
           RowBox[{"x3", "+", 
            RowBox[{"x2", " ", "x4"}]}], ")"}]}]}], ")"}]}]}]}], "}"}], ",", 
   RowBox[{"{", "4", "}"}]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"Out[124]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
z={z1,z2,z3,z4};
InvTrans=InverseTransformation[x,z,Trans]\
\>", "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"In[125]:="],

Cell[BoxData[
 RowBox[{
  RowBox[{"Solve", "::", "\<\"incnst\"\>"}], 
  RowBox[{
  ":", " "}], "\<\"Inconsistent or redundant transcendental equation. After \
reduction, the bad equation is \\!\\(\\(\\(-z1\\)\\) + \\(\\(Log[\\(\\(\
\[ExponentialE]\\^x1\\)\\)]\\)\\)\\) == 0. \\!\\(\\*ButtonBox[\\\"More\
\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"Solve::incnst\\\"]\\)\"\>"}]], "Message",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"From In[125]:="],

Cell[BoxData[
 RowBox[{
  RowBox[{"Solve", "::", "\<\"ifun\"\>"}], 
  RowBox[{
  ":", " "}], "\<\"Inverse functions are being used by \\!\\(Solve\\), so \
some solutions may not be found; use Reduce for complete solution \
information. \\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", \
ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"Solve::ifun\\\"]\\)\"\>"}]], "Message",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"From In[125]:="],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Inverse Transformation: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", "x3", ",", "x4"}], "}"}], 
   "\[InvisibleSpace]", "\<\" =  \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"z1", ",", 
     RowBox[{
      SuperscriptBox["\[ExponentialE]", "z1"], " ", "z2"}], ",", 
     RowBox[{
      RowBox[{
       RowBox[{"-", "z2"}], " ", "z3"}], "+", "z4"}], ",", 
     RowBox[{
      SuperscriptBox["\[ExponentialE]", 
       RowBox[{"-", "z1"}]], " ", "z3"}]}], "}"}]}],
  SequenceForm[
  "Inverse Transformation: ", {x1, x2, x3, x4}, " =  ", {
   z1, E^z1 z2, -z2 z3 + z4, E^(-z1) z3}],
  Editable->False]], "Print",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"From In[125]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"z1", ",", 
   RowBox[{
    SuperscriptBox["\[ExponentialE]", "z1"], " ", "z2"}], ",", 
   RowBox[{
    RowBox[{
     RowBox[{"-", "z2"}], " ", "z3"}], "+", "z4"}], ",", 
   RowBox[{
    SuperscriptBox["\[ExponentialE]", 
     RowBox[{"-", "z1"}]], " ", "z3"}]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"Out[126]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{"newf", ",", "newg", ",", "newh"}], "}"}], "=", 
  RowBox[{"TransformSystem", "[", 
   RowBox[{
    RowBox[{"f", "/.", 
     RowBox[{"u", "\[Rule]", "0"}]}], ",", 
    RowBox[{"Coefficient", "[", 
     RowBox[{"f", ",", "u"}], "]"}], ",", "h", ",", "x", ",", "z", ",", 
    "Trans", ",", "InvTrans"}], "]"}]}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"In[127]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"z2", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "z3", ",", "z4", ",", 
     SuperscriptBox["z1", "2"]}], "}"}], ",", "z1"}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"Out[127]="]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Observer form", "Subsubsection",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Trans", "=", 
  RowBox[{"LinearizeToOutputInjection", "[", 
   RowBox[{"f", ",", 
    RowBox[{"{", "h", "}"}], ",", "x", ",", 
    RowBox[{"{", "u", "}"}]}], "]"}]}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"In[128]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"x1", ",", 
   RowBox[{
    RowBox[{"-", 
     SuperscriptBox["\[ExponentialE]", 
      RowBox[{"-", "x1"}]]}], " ", "x2"}], ",", 
   RowBox[{
    SuperscriptBox["\[ExponentialE]", "x1"], " ", "x4"}], ",", 
   RowBox[{
    RowBox[{"-", "x3"}], "-", 
    RowBox[{"x2", " ", "x4"}]}]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"Out[128]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
z={z1,z2,z3,z4};
InvTrans=InverseTransformation[x,z,Trans]\
\>", "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"In[129]:="],

Cell[BoxData[
 RowBox[{
  RowBox[{"Solve", "::", "\<\"incnst\"\>"}], 
  RowBox[{
  ":", " "}], "\<\"Inconsistent or redundant transcendental equation. After \
reduction, the bad equation is \\!\\(\\(\\(-z1\\)\\) + \\(\\(Log[\\(\\(\
\[ExponentialE]\\^x1\\)\\)]\\)\\)\\) == 0. \\!\\(\\*ButtonBox[\\\"More\
\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"Solve::incnst\\\"]\\)\"\>"}]], "Message",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"From In[129]:="],

Cell[BoxData[
 RowBox[{
  RowBox[{"Solve", "::", "\<\"ifun\"\>"}], 
  RowBox[{
  ":", " "}], "\<\"Inverse functions are being used by \\!\\(Solve\\), so \
some solutions may not be found; use Reduce for complete solution \
information. \\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", \
ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"Solve::ifun\\\"]\\)\"\>"}]], "Message",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"From In[129]:="],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Inverse Transformation: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", "x3", ",", "x4"}], "}"}], 
   "\[InvisibleSpace]", "\<\" =  \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"z1", ",", 
     RowBox[{
      RowBox[{"-", 
       SuperscriptBox["\[ExponentialE]", "z1"]}], " ", "z2"}], ",", 
     RowBox[{
      RowBox[{"z2", " ", "z3"}], "-", "z4"}], ",", 
     RowBox[{
      SuperscriptBox["\[ExponentialE]", 
       RowBox[{"-", "z1"}]], " ", "z3"}]}], "}"}]}],
  SequenceForm[
  "Inverse Transformation: ", {x1, x2, x3, x4}, " =  ", {
   z1, -E^z1 z2, z2 z3 - z4, E^(-z1) z3}],
  Editable->False]], "Print",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"From In[129]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"z1", ",", 
   RowBox[{
    RowBox[{"-", 
     SuperscriptBox["\[ExponentialE]", "z1"]}], " ", "z2"}], ",", 
   RowBox[{
    RowBox[{"z2", " ", "z3"}], "-", "z4"}], ",", 
   RowBox[{
    SuperscriptBox["\[ExponentialE]", 
     RowBox[{"-", "z1"}]], " ", "z3"}]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"Out[130]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
{newf,newg}=TransformSystem[f/.u\[Rule]0,Coefficient[f,u],x,z,Trans,InvTrans]\
\>", "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"In[131]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "z2"}], ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", 
     RowBox[{"-", "z3"}], ",", 
     RowBox[{"-", "z4"}], ",", 
     RowBox[{"-", 
      SuperscriptBox["z1", "2"]}]}], "}"}]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"},
 CellLabel->"Out[131]="]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["2-Output (modified Hammouri)", "Subsection",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],

Cell["\<\
f = {Exp[x1+x2]-1+u x1^2,-Exp[x1+x2]+1+u \
(Exp[x3-x2]-Exp[-x1-x2]-x1^2),-Exp[x1+x2]+1+x1^3 Exp[-x1-x3]-u x1^2,x5,x1};
h={x1,x4};
x={x1,x2,x3,x4,x5};\
\>", "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"In[132]:="],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityIndices", "[", 
  RowBox[{"f", ",", "h", ",", "x", ",", 
   RowBox[{"{", "u", "}"}]}], "]"}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"In[135]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"3", ",", "2"}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"Out[135]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityIndices", "[", 
  RowBox[{
   RowBox[{"f", "/.", 
    RowBox[{"u", "\[Rule]", "0"}]}], ",", "h", ",", "x", ",", 
   RowBox[{"{", "u", "}"}]}], "]"}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"In[136]:="],

Cell[BoxData["\<\"codimension = 0 at k= 2! System is not observable.\"\>"], \
"Print",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"From In[136]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"1", ",", "1"}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"Out[136]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityCodistribution", "[", 
  RowBox[{
   RowBox[{"f", "/.", 
    RowBox[{"u", "\[Rule]", "0"}]}], ",", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"f", ",", "u"}], "]"}], ",", "h", ",", "x"}], "]"}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"In[137]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "1", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0", ",", "1", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "1"}], "}"}]}], 
  "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"Out[137]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityCodistribution", "[", 
  RowBox[{
   RowBox[{"f", "/.", 
    RowBox[{"u", "\[Rule]", "0"}]}], ",", 
   RowBox[{"Coefficient", "[", 
    RowBox[{"f", ",", "u"}], "]"}], ",", "h", ",", "x", ",", 
   RowBox[{"0", "*", "x"}]}], "]"}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"In[138]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", "0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "1", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0", ",", "1", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "1"}], "}"}]}], 
  "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"Out[138]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"ObservabilityCodistribution", "[", 
  RowBox[{
   RowBox[{"f", "/.", 
    RowBox[{"u", "\[Rule]", "0"}]}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "0"}], "}"}], ",", "h", 
   ",", "x"}], "]"}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"In[139]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0", ",", "0", ",", "1"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0", ",", "1", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "1", ",", "0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"1", ",", "0", ",", "0", ",", "0", ",", "0"}], "}"}]}], 
  "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"Out[139]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["Observable form", "Subsubsection",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{"Trans", ",", "Ind"}], "}"}], "=", 
  RowBox[{"ObservableForm", "[", 
   RowBox[{"f", ",", "h", ",", "x", ",", 
    RowBox[{"{", "u", "}"}]}], "]"}]}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"In[140]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"x1", ",", 
     RowBox[{
      RowBox[{"-", "1"}], "+", 
      SuperscriptBox["\[ExponentialE]", 
       RowBox[{"x1", "+", "x2"}]]}], ",", 
     RowBox[{
      RowBox[{
       SuperscriptBox["\[ExponentialE]", 
        RowBox[{"x1", "+", "x2"}]], " ", 
       RowBox[{"(", 
        RowBox[{"1", "-", 
         SuperscriptBox["\[ExponentialE]", 
          RowBox[{
           RowBox[{"-", "x1"}], "-", "x2"}]], "-", 
         SuperscriptBox["\[ExponentialE]", 
          RowBox[{"x1", "+", "x2"}]], "+", 
         SuperscriptBox["\[ExponentialE]", 
          RowBox[{
           RowBox[{"-", "x2"}], "+", "x3"}]], "-", 
         SuperscriptBox["x1", "2"]}], ")"}]}], "+", 
      RowBox[{
       SuperscriptBox["\[ExponentialE]", 
        RowBox[{"x1", "+", "x2"}]], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"-", "1"}], "+", 
         SuperscriptBox["\[ExponentialE]", 
          RowBox[{"x1", "+", "x2"}]], "+", 
         SuperscriptBox["x1", "2"]}], ")"}]}]}], ",", "x4", ",", "x5"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"3", ",", "2"}], "}"}]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"Out[140]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
z={z1,z2,z3,z4,z5};
InvTrans=InverseTransformation[x,z,Trans]\
\>", "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"In[141]:="],

Cell[BoxData[
 RowBox[{
  RowBox[{"Solve", "::", "\<\"ifun\"\>"}], 
  RowBox[{
  ":", " "}], "\<\"Inverse functions are being used by \\!\\(Solve\\), so \
some solutions may not be found; use Reduce for complete solution \
information. \\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", \
ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"Solve::ifun\\\"]\\)\"\>"}]], "Message",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"From In[141]:="],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Inverse Transformation: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", "x3", ",", "x4", ",", "x5"}], "}"}], 
   "\[InvisibleSpace]", "\<\" =  \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"z1", ",", 
     RowBox[{
      RowBox[{"-", "z1"}], "+", 
      RowBox[{"Log", "[", 
       RowBox[{"1", "+", "z2"}], "]"}]}], ",", 
     RowBox[{
      RowBox[{"-", "z1"}], "+", 
      RowBox[{"Log", "[", 
       RowBox[{"1", "+", "z3"}], "]"}]}], ",", "z4", ",", "z5"}], "}"}]}],
  SequenceForm[
  "Inverse Transformation: ", {x1, x2, x3, x4, x5}, " =  ", {
   z1, -z1 + Log[1 + z2], -z1 + Log[1 + z3], z4, z5}],
  Editable->False]], "Print",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"From In[141]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"z1", ",", 
   RowBox[{
    RowBox[{"-", "z1"}], "+", 
    RowBox[{"Log", "[", 
     RowBox[{"1", "+", "z2"}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"-", "z1"}], "+", 
    RowBox[{"Log", "[", 
     RowBox[{"1", "+", "z3"}], "]"}]}], ",", "z4", ",", "z5"}], 
  "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"Out[142]="]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"{", 
   RowBox[{"newf", ",", "newg", ",", "newh"}], "}"}], "=", 
  RowBox[{"TransformSystem", "[", 
   RowBox[{
    RowBox[{"f", "/.", 
     RowBox[{"u", "\[Rule]", "0"}]}], ",", 
    RowBox[{"Coefficient", "[", 
     RowBox[{"f", ",", "u"}], "]"}], ",", "h", ",", "x", ",", "z", ",", 
    "Trans", ",", "InvTrans"}], "]"}]}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"In[143]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"z2", ",", "0", ",", 
     SuperscriptBox["z1", "3"], ",", "z5", ",", "z1"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     SuperscriptBox["z1", "2"], ",", "z3", ",", "0", ",", "0", ",", "0"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"z1", ",", "z4"}], "}"}]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"Out[143]="]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Observer form", "Subsubsection",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Trans", "=", 
  RowBox[{"LinearizeToOutputInjection", "[", 
   RowBox[{"f", ",", "h", ",", "x", ",", 
    RowBox[{"{", "u", "}"}]}], "]"}]}]], "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"In[144]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"x1", ",", "x4", ",", 
   RowBox[{"1", "-", 
    SuperscriptBox["\[ExponentialE]", 
     RowBox[{"x1", "+", "x2"}]]}], ",", 
   RowBox[{"-", "x5"}], ",", 
   RowBox[{
    RowBox[{"-", "1"}], "+", 
    SuperscriptBox["\[ExponentialE]", 
     RowBox[{"x1", "+", "x3"}]]}]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"Out[144]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
z={z1,z2,z3,z4,z5};
InvTrans=InverseTransformation[x,z,Trans]\
\>", "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"In[145]:="],

Cell[BoxData[
 RowBox[{
  RowBox[{"Solve", "::", "\<\"ifun\"\>"}], 
  RowBox[{
  ":", " "}], "\<\"Inverse functions are being used by \\!\\(Solve\\), so \
some solutions may not be found; use Reduce for complete solution \
information. \\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", \
ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"Solve::ifun\\\"]\\)\"\>"}]], "Message",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"From In[145]:="],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Inverse Transformation: \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"x1", ",", "x2", ",", "x3", ",", "x4", ",", "x5"}], "}"}], 
   "\[InvisibleSpace]", "\<\" =  \"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"z1", ",", 
     RowBox[{
      RowBox[{"-", "z1"}], "+", 
      RowBox[{"Log", "[", 
       RowBox[{"1", "-", "z3"}], "]"}]}], ",", 
     RowBox[{
      RowBox[{"-", "z1"}], "+", 
      RowBox[{"Log", "[", 
       RowBox[{"1", "+", "z5"}], "]"}]}], ",", "z2", ",", 
     RowBox[{"-", "z4"}]}], "}"}]}],
  SequenceForm[
  "Inverse Transformation: ", {x1, x2, x3, x4, x5}, " =  ", {
   z1, -z1 + Log[1 - z3], -z1 + Log[1 + z5], z2, -z4}],
  Editable->False]], "Print",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"From In[145]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"z1", ",", 
   RowBox[{
    RowBox[{"-", "z1"}], "+", 
    RowBox[{"Log", "[", 
     RowBox[{"1", "-", "z3"}], "]"}]}], ",", 
   RowBox[{
    RowBox[{"-", "z1"}], "+", 
    RowBox[{"Log", "[", 
     RowBox[{"1", "+", "z5"}], "]"}]}], ",", "z2", ",", 
   RowBox[{"-", "z4"}]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"Out[146]="]
}, Open  ]],

Cell[CellGroupData[{

Cell["\<\
{newf,newg,newh}=TransformSystem[f/.u\[Rule]0,Coefficient[f,u],h,x,z,Trans,\
InvTrans]\
\>", "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"In[147]:="],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "z3"}], ",", 
     RowBox[{"-", "z4"}], ",", "0", ",", 
     RowBox[{"-", "z1"}], ",", 
     SuperscriptBox["z1", "3"]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     SuperscriptBox["z1", "2"], ",", "0", ",", 
     RowBox[{"-", "z5"}], ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"z1", ",", "z2"}], "}"}]}], "}"}]], "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"},
 CellLabel->"Out[147]="]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Bibliography", "Section",
 CellTags->{"Bibliography", "b:41"}],

Cell["\<\
A. J. Krener and A. Isidori, ``Linearization by Output Injection and \
Nonlinear Observers,\" Systems and Control Letters, vol. 3, pp. 47-52, 1983.
A. J. Krener and W. Respondek, ``Nonlinear Observers with Linearizable Error \
Dynamics,\" SIAM Journal of Control and Optimization, vol. 23, pp. 197-216, \
1985.
H. Hammouri and J. P. Gauthier, ``Bilinearization up to Output Injection,\" \
Systems and Control Letters, vol. 11, pp. 139-149, 1988.
H. Hammouri and M. Kinnaert, ``A New Procedure for Time-Varying Linearization \
up to Output Injection,\" Systems and Control Letters, vol. 28, pp. 151-157, \
1996.
 H. Hammouri and J. P. Gauthier, ``Time Varying Linearization up to Output \
Injection,\" presented at 28th Conference on Decision and Control, Tampa, \
Florida, 1999.
 G. Besancon and G. Bornard, ``Symbolic Computation for Nonlinear Systems \
Analysis: the Example of Bilinearization,\" Mathematics and Computers in \
Simulation, vol. 44, pp. 11-27, 1997.
X. Xia and M. Zeitz, ``On Nonlinear Continuous Observers,\" International \
Journal of Control, vol. 66, pp. 943-954, 1997.
 M. Hou and A. C. Pugh, ``Observer with Linear Error Dynamics for Nonlinear \
Multi-Output Systems,\" Systems and Control Letters, vol. 37, pp. 1-9, 1999.
\
\>", "Text",
 CellTags->{"Bibliography", "b:41"}],

Cell[BoxData["\<\"  *** LinearControl successfully loaded ***\"\>"], "Print",
 CellChangeTimes->{3.474161331707037*^9},
 CellLabel->"During evaluation of In[248]:="],

Cell[BoxData["\<\"  *** NonlinearControl successfully loaded ***\"\>"], \
"Print",
 CellChangeTimes->{3.4741613317851624`*^9},
 CellLabel->"During evaluation of In[248]:="]
}, Open  ]]
},
WindowSize->{818, 573},
WindowMargins->{{83, Automatic}, {Automatic, -1}},
PrintingStartingPageNumber->168,
ShowSelection->True,
TaggingRules:>{
 "NextIndexingNumber" -> 3, 
  "IndexEntries" -> {{"i:1", "LinearizeToOutputInjection"}, {
    "i:1", "ObservabilityIndices"}, {"i:1", "ObservableForm"}, {
    "i:1", "PDESolve"}, {"i:2", "InverseTransformation"}, {
    "i:2", "LinearizeToOutputInjection"}, {"i:2", "ObservabilityIndices"}, {
    "i:2", "PDESolve"}, {"i:2", "TransformSystem"}, {
    "i:2", "ObservableForm", "", ""}}, 
  "PaginationCache" -> {{2003, 8, 1, 21, 15, 48.4405952`9.437784329011935}, {
    168, 168, 168, 168, 168, 168, 169, 169, 169, 169, 169, 169, 169, 169, 169,
     169, 169, 169, 169, 169, 170, 170, 170, 170, 170, 170, 170, 170, 170, 
    170, 170, 170, 170, 170, 170, 171, 171, 171, 171, 171, 171, 171, 171, 171,
     171, 171, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 172, 
    172, 173, 173, 173, 173, 173, 173, 173, 173, 173, 173, 173, 173, 173, 173,
     173, 174, 174, 174, 174, 174, 174, 174, 174, 174, 174, 174, 174, 174, 
    174, 174, 175, 175, 175, 175, 175, 175, 175, 175, 175, 175, 175, 175, 175,
     175, 175, 176, 176, 176, 176, 176, 176, 176, 176, 176, 176, 176, 176, 
    176, 176, 176, 177, 177, 177, 177, 177, 177, 177, 177, 177, 177, 177, 177,
     177, 177, 177, 178, 178, 178, 178, 178, 178, 178, 178, 178, 178, 178, 
    178, 178, 178, 178, 179, 179, 179, 179, 179, 179, 179, 179, 179, 179, 179,
     179, 179, 179, 179, 179, 179, 180, 180, 180, 180, 180, 180, 180, 180, 
    180, 180, 180, 180, 181, 181, 181, 181, 181, 181, 181, 181, 181, 181, 181,
     181, 182, 182, 182, 182, 182, 182, 182, 182, 182, 182, 182, 182, 183, 
    183, 183, 183, 183, 183, 183, 183, 183, 183, 183, 183, 184, 184, 184, 184,
     184, 184, 184, 184, 184, 184, 184, 184, 185, 185, 185, 185, 185, 185, 
    185, 185, 185, 185, 185, 185, 185, 185, 185, 186, 186, 186, 186, 186, 186,
     186, 186, 186, 186, 186, 186, 186, 187, 187, 187, 187, 187, 187, 187, 
    187, 187, 187, 187, 187, 187, 187, 188, 188, 188, 188, 188, 188, 188, 
    188}}},
ShowCellTags->True,
FrontEndVersion->"7.0 for Microsoft Windows (32-bit) (November 10, 2008)",
StyleDefinitions->Notebook[{
   Cell[
    StyleData[StyleDefinitions -> "HelpBrowser.nb"]]}, 
  WindowSize -> {616, 505}, 
  WindowMargins -> {{110, Automatic}, {Automatic, 0}}, Visible -> False, 
  FrontEndVersion -> "7.0 for Microsoft Windows (32-bit) (November 10, 2008)",
   StyleDefinitions -> "PrivateStylesheetFormatting.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "Introduction"->{
  Cell[567, 22, 68, 1, 77, "Section",
   CellTags->{"Introduction", "b:36"}],
  Cell[638, 25, 607, 10, 109, "Text",
   CellTags->{"Introduction", "i:1", "b:36"}],
  Cell[1248, 37, 83, 1, 37, "Text",
   CellTags->{"Introduction", "i:1", "b:36"}],
  Cell[1356, 42, 118, 3, 41, "Input",
   CellTags->{"Introduction", "i:1", "b:36"}],
  Cell[1477, 47, 751, 10, 416, "Print",
   CellTags->{"Introduction", "i:1", "b:36"}],
  Cell[2243, 60, 135, 1, 37, "Text",
   CellTags->{"Observer Tools", "Introduction", "i:1", "b:36"}],
  Cell[2403, 65, 150, 3, 41, "Input",
   CellTags->{"Observer Tools", "Introduction", "i:1", "b:36"}],
  Cell[2556, 70, 3225, 83, 102, "Print",
   CellTags->{"Observer Tools", "Introduction", "i:1", "b:36"}]},
 "b:36"->{
  Cell[567, 22, 68, 1, 77, "Section",
   CellTags->{"Introduction", "b:36"}],
  Cell[638, 25, 607, 10, 109, "Text",
   CellTags->{"Introduction", "i:1", "b:36"}],
  Cell[1248, 37, 83, 1, 37, "Text",
   CellTags->{"Introduction", "i:1", "b:36"}],
  Cell[1356, 42, 118, 3, 41, "Input",
   CellTags->{"Introduction", "i:1", "b:36"}],
  Cell[1477, 47, 751, 10, 416, "Print",
   CellTags->{"Introduction", "i:1", "b:36"}],
  Cell[2243, 60, 135, 1, 37, "Text",
   CellTags->{"Observer Tools", "Introduction", "i:1", "b:36"}],
  Cell[2403, 65, 150, 3, 41, "Input",
   CellTags->{"Observer Tools", "Introduction", "i:1", "b:36"}],
  Cell[2556, 70, 3225, 83, 102, "Print",
   CellTags->{"Observer Tools", "Introduction", "i:1", "b:36"}]},
 "i:1"->{
  Cell[638, 25, 607, 10, 109, "Text",
   CellTags->{"Introduction", "i:1", "b:36"}],
  Cell[1248, 37, 83, 1, 37, "Text",
   CellTags->{"Introduction", "i:1", "b:36"}],
  Cell[1356, 42, 118, 3, 41, "Input",
   CellTags->{"Introduction", "i:1", "b:36"}],
  Cell[1477, 47, 751, 10, 416, "Print",
   CellTags->{"Introduction", "i:1", "b:36"}],
  Cell[2243, 60, 135, 1, 37, "Text",
   CellTags->{"Observer Tools", "Introduction", "i:1", "b:36"}],
  Cell[2403, 65, 150, 3, 41, "Input",
   CellTags->{"Observer Tools", "Introduction", "i:1", "b:36"}],
  Cell[2556, 70, 3225, 83, 102, "Print",
   CellTags->{"Observer Tools", "Introduction", "i:1", "b:36"}]},
 "Observer Tools"->{
  Cell[2243, 60, 135, 1, 37, "Text",
   CellTags->{"Observer Tools", "Introduction", "i:1", "b:36"}],
  Cell[2403, 65, 150, 3, 41, "Input",
   CellTags->{"Observer Tools", "Introduction", "i:1", "b:36"}],
  Cell[2556, 70, 3225, 83, 102, "Print",
   CellTags->{"Observer Tools", "Introduction", "i:1", "b:36"}]},
 "Autonomous Examples"->{
  Cell[5830, 159, 102, 1, 77, "Section",
   CellTags->{"Autonomous Examples", "i:2"}],
  Cell[5957, 164, 87, 1, 51, "Subsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.0"}],
  Cell[6047, 167, 289, 5, 53, "Text",
   CellTags->{"Autonomous Examples", "i:2", "b:37.0"}],
  Cell[6373, 177, 127, 1, 51, "Subsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[6525, 182, 84, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[6612, 185, 159, 4, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[6774, 191, 644, 22, 82, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[7443, 217, 216, 5, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[7662, 224, 126, 3, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[7837, 233, 93, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[7955, 238, 288, 8, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[8246, 248, 235, 7, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[8518, 260, 156, 5, 61, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[8677, 267, 158, 2, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[8838, 271, 163, 3, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[9004, 276, 489, 11, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[9496, 289, 163, 4, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[9696, 298, 448, 12, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[10147, 312, 522, 16, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[10718, 334, 91, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[10834, 339, 223, 5, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[11060, 346, 159, 2, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[11222, 350, 122, 3, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[11393, 359, 167, 2, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[11563, 363, 674, 20, 53, "Text",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[12262, 387, 559, 16, 82, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[12824, 405, 420, 15, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[13281, 425, 726, 20, 82, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[14010, 447, 980, 28, 52, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[14993, 477, 456, 15, 59, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[15452, 494, 492, 16, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[15959, 513, 189, 4, 37, "Text",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[16173, 521, 709, 24, 62, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[16885, 547, 493, 16, 85, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[17415, 568, 262, 7, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[17680, 577, 525, 17, 42, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[18242, 599, 270, 8, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[18515, 609, 363, 13, 59, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[18915, 627, 726, 20, 82, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[19644, 649, 980, 28, 52, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[20627, 679, 377, 12, 59, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[21007, 693, 391, 13, 59, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[21459, 713, 118, 1, 51, "Subsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[21602, 718, 84, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[21689, 721, 587, 20, 62, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[22279, 743, 747, 25, 82, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[23051, 772, 217, 5, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[23271, 779, 150, 4, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[23470, 789, 93, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[23588, 794, 289, 8, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[23880, 804, 261, 8, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[24178, 817, 157, 5, 61, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[24338, 824, 490, 11, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[24831, 837, 163, 4, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[25031, 846, 448, 12, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[25482, 860, 699, 22, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[26230, 888, 91, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[26346, 893, 223, 5, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[26572, 900, 152, 2, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[26727, 904, 159, 2, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[26889, 908, 122, 3, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[27060, 917, 126, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[27189, 920, 692, 23, 62, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[27906, 947, 768, 25, 62, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[28677, 974, 153, 4, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[28833, 980, 108, 2, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[28978, 987, 927, 31, 102, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[29908, 1020, 363, 12, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[30308, 1037, 217, 5, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[30528, 1044, 150, 4, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[30727, 1054, 93, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[30845, 1059, 289, 8, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[31137, 1069, 316, 10, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[31490, 1084, 157, 5, 61, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[31650, 1091, 563, 15, 52, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[32216, 1108, 211, 6, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[32464, 1119, 448, 12, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[32915, 1133, 441, 14, 42, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[33405, 1153, 91, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[33521, 1158, 252, 6, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[33776, 1166, 393, 13, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[34206, 1184, 157, 5, 61, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[34366, 1191, 755, 21, 52, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[35124, 1214, 365, 12, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[35526, 1231, 448, 12, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[35977, 1245, 546, 18, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[36584, 1270, 103, 1, 51, "Subsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[36712, 1275, 84, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[36821, 1280, 825, 27, 122, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[37649, 1309, 150, 4, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[37848, 1319, 93, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[37966, 1324, 289, 8, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[38258, 1334, 261, 8, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[38556, 1347, 157, 5, 61, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[38716, 1354, 490, 11, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[39209, 1367, 163, 4, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[39409, 1376, 200, 6, 61, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[39612, 1384, 359, 11, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[39974, 1397, 188, 5, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[40211, 1408, 91, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[40327, 1413, 252, 6, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[40582, 1421, 260, 8, 42, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[40879, 1434, 354, 10, 62, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[41236, 1446, 658, 18, 52, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[41897, 1466, 293, 10, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[42227, 1481, 448, 12, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[42678, 1495, 609, 21, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}]},
 "i:2"->{
  Cell[5830, 159, 102, 1, 77, "Section",
   CellTags->{"Autonomous Examples", "i:2"}],
  Cell[5957, 164, 87, 1, 51, "Subsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.0"}],
  Cell[6047, 167, 289, 5, 53, "Text",
   CellTags->{"Autonomous Examples", "i:2", "b:37.0"}],
  Cell[6373, 177, 127, 1, 51, "Subsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[6525, 182, 84, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[6612, 185, 159, 4, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[6774, 191, 644, 22, 82, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[7443, 217, 216, 5, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[7662, 224, 126, 3, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[7837, 233, 93, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[7955, 238, 288, 8, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[8246, 248, 235, 7, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[8518, 260, 156, 5, 61, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[8677, 267, 158, 2, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[8838, 271, 163, 3, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[9004, 276, 489, 11, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[9496, 289, 163, 4, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[9696, 298, 448, 12, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[10147, 312, 522, 16, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[10718, 334, 91, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[10834, 339, 223, 5, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[11060, 346, 159, 2, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[11222, 350, 122, 3, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[11393, 359, 167, 2, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[11563, 363, 674, 20, 53, "Text",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[12262, 387, 559, 16, 82, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[12824, 405, 420, 15, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[13281, 425, 726, 20, 82, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[14010, 447, 980, 28, 52, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[14993, 477, 456, 15, 59, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[15452, 494, 492, 16, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[15959, 513, 189, 4, 37, "Text",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[16173, 521, 709, 24, 62, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[16885, 547, 493, 16, 85, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[17415, 568, 262, 7, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[17680, 577, 525, 17, 42, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[18242, 599, 270, 8, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[18515, 609, 363, 13, 59, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[18915, 627, 726, 20, 82, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[19644, 649, 980, 28, 52, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[20627, 679, 377, 12, 59, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[21007, 693, 391, 13, 59, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[21459, 713, 118, 1, 51, "Subsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[21602, 718, 84, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[21689, 721, 587, 20, 62, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[22279, 743, 747, 25, 82, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[23051, 772, 217, 5, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[23271, 779, 150, 4, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[23470, 789, 93, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[23588, 794, 289, 8, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[23880, 804, 261, 8, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[24178, 817, 157, 5, 61, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[24338, 824, 490, 11, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[24831, 837, 163, 4, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[25031, 846, 448, 12, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[25482, 860, 699, 22, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[26230, 888, 91, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[26346, 893, 223, 5, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[26572, 900, 152, 2, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[26727, 904, 159, 2, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[26889, 908, 122, 3, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[27060, 917, 126, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[27189, 920, 692, 23, 62, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[27906, 947, 768, 25, 62, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[28677, 974, 153, 4, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[28833, 980, 108, 2, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[28978, 987, 927, 31, 102, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[29908, 1020, 363, 12, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[30308, 1037, 217, 5, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[30528, 1044, 150, 4, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[30727, 1054, 93, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[30845, 1059, 289, 8, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[31137, 1069, 316, 10, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[31490, 1084, 157, 5, 61, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[31650, 1091, 563, 15, 52, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[32216, 1108, 211, 6, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[32464, 1119, 448, 12, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[32915, 1133, 441, 14, 42, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[33405, 1153, 91, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[33521, 1158, 252, 6, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[33776, 1166, 393, 13, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[34206, 1184, 157, 5, 61, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[34366, 1191, 755, 21, 52, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[35124, 1214, 365, 12, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[35526, 1231, 448, 12, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[35977, 1245, 546, 18, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[36584, 1270, 103, 1, 51, "Subsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[36712, 1275, 84, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[36821, 1280, 825, 27, 122, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[37649, 1309, 150, 4, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[37848, 1319, 93, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[37966, 1324, 289, 8, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[38258, 1334, 261, 8, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[38556, 1347, 157, 5, 61, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[38716, 1354, 490, 11, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[39209, 1367, 163, 4, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[39409, 1376, 200, 6, 61, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[39612, 1384, 359, 11, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[39974, 1397, 188, 5, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[40211, 1408, 91, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[40327, 1413, 252, 6, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[40582, 1421, 260, 8, 42, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[40879, 1434, 354, 10, 62, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[41236, 1446, 658, 18, 52, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[41897, 1466, 293, 10, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[42227, 1481, 448, 12, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[42678, 1495, 609, 21, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[43360, 1524, 123, 1, 77, "Section",
   CellTags->{"Failure of Rank Condition", "i:2"}],
  Cell[43508, 1529, 102, 1, 51, "Subsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[43635, 1534, 90, 1, 47, "Subsubsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[43728, 1537, 346, 6, 53, "Text",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[44077, 1545, 493, 16, 82, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[44595, 1565, 223, 5, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[44821, 1572, 133, 3, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[44991, 1580, 255, 6, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[45249, 1588, 254, 8, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[45540, 1601, 289, 7, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[45832, 1610, 128, 3, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[46009, 1619, 99, 1, 47, "Subsubsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[46133, 1624, 295, 8, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[46431, 1634, 453, 15, 42, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[46921, 1654, 160, 5, 61, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[47084, 1661, 1285, 39, 52, "Print",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[48372, 1702, 800, 29, 56, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[49209, 1736, 194, 5, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[49406, 1743, 396, 13, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[49851, 1762, 97, 1, 47, "Subsubsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[49973, 1767, 258, 6, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[50234, 1775, 364, 12, 42, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[50635, 1792, 160, 5, 61, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[50798, 1799, 1285, 39, 52, "Print",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[52086, 1840, 800, 29, 56, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[52923, 1874, 194, 5, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[53120, 1881, 372, 12, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[53565, 1901, 137, 3, 77, "Section",
   CellTags->{"Failure of Rank Condition", "i:2"}],
  Cell[53727, 1908, 102, 1, 51, "Subsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[53854, 1913, 90, 1, 47, "Subsubsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[53947, 1916, 192, 4, 37, "Text",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[54142, 1922, 535, 17, 102, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[54702, 1943, 223, 5, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[54928, 1950, 133, 3, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[55098, 1958, 312, 8, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[55413, 1968, 254, 8, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[55704, 1981, 346, 9, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[56053, 1992, 180, 5, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[56270, 2002, 311, 8, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[56584, 2012, 254, 8, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[56887, 2026, 99, 1, 47, "Subsubsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[57011, 2031, 295, 8, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[57309, 2041, 258, 8, 42, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[57604, 2054, 160, 5, 61, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[57767, 2061, 538, 14, 38, "Print",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[58308, 2077, 208, 6, 42, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[58553, 2088, 194, 5, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[58750, 2095, 317, 10, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[59116, 2111, 97, 1, 47, "Subsubsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[59238, 2116, 258, 6, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[59499, 2124, 203, 6, 42, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[59739, 2135, 160, 5, 61, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[59902, 2142, 562, 15, 38, "Print",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[60467, 2159, 229, 7, 42, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[60733, 2171, 194, 5, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[60930, 2178, 338, 11, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[61341, 2197, 110, 1, 77, "Section",
   CellTags->{"Zero-input Unobservable", "i:2"}],
  Cell[61476, 2202, 90, 1, 51, "Subsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.0"}],
  Cell[61569, 2205, 231, 4, 37, "Text",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.0"}],
  Cell[61837, 2214, 101, 1, 51, "Subsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[61941, 2217, 238, 7, 101, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[62204, 2228, 246, 6, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[62453, 2236, 132, 3, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[62622, 2244, 302, 8, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[62927, 2254, 173, 3, 35, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[63103, 2259, 132, 3, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[63272, 2267, 349, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[63624, 2278, 342, 10, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[64003, 2293, 383, 10, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[64389, 2305, 342, 10, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[64768, 2320, 344, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[65115, 2331, 273, 8, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[65425, 2344, 97, 1, 47, "Subsubsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[65547, 2349, 319, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[65869, 2360, 1196, 35, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[67102, 2400, 162, 5, 61, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[67267, 2407, 484, 10, 48, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[67754, 2419, 754, 21, 35, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[68511, 2442, 374, 12, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[68922, 2459, 453, 12, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[69378, 2473, 345, 11, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[69760, 2489, 110, 2, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[69873, 2493, 294, 9, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[70216, 2508, 95, 1, 47, "Subsubsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[70336, 2513, 282, 7, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[70621, 2522, 370, 11, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[71006, 2536, 376, 11, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[71407, 2551, 162, 5, 61, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[71572, 2558, 484, 10, 48, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[72059, 2570, 754, 21, 35, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[72816, 2593, 374, 12, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[73227, 2610, 193, 5, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[73423, 2617, 384, 12, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[73844, 2634, 110, 2, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[73957, 2638, 314, 10, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[74332, 2655, 111, 1, 51, "Subsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[74446, 2658, 1598, 50, 102, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[76069, 2712, 246, 6, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[76318, 2720, 132, 3, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[76487, 2728, 302, 8, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[76792, 2738, 173, 3, 35, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[76968, 2743, 132, 3, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[77137, 2751, 325, 8, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[77465, 2761, 451, 12, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[77953, 2778, 359, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[78315, 2789, 451, 12, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[78803, 2806, 333, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[79139, 2817, 293, 8, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[79469, 2830, 97, 1, 47, "Subsubsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[79591, 2835, 319, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[79913, 2846, 2720, 80, 86, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[82670, 2931, 165, 5, 61, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[82838, 2938, 525, 10, 51, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[83366, 2950, 484, 10, 48, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[83853, 2962, 793, 21, 38, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[84649, 2985, 402, 12, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[85088, 3002, 453, 12, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[85544, 3016, 335, 9, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[85928, 3031, 95, 1, 47, "Subsubsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[86048, 3036, 282, 7, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[86333, 3045, 423, 13, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[86793, 3063, 165, 5, 61, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[86961, 3070, 525, 10, 51, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[87489, 3082, 484, 10, 48, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[87976, 3094, 793, 21, 38, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[88772, 3117, 402, 12, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[89211, 3134, 184, 4, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[89398, 3140, 409, 13, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[89868, 3160, 107, 1, 51, "Subsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[89978, 3163, 256, 7, 101, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[90259, 3174, 222, 5, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[90484, 3181, 155, 4, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[90676, 3190, 278, 7, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[90957, 3199, 173, 3, 35, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[91133, 3204, 155, 4, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[91325, 3213, 325, 8, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[91653, 3223, 583, 15, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[92273, 3243, 359, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[92635, 3254, 583, 15, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[93255, 3274, 344, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[93602, 3285, 494, 13, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[94133, 3303, 97, 1, 47, "Subsubsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[94255, 3308, 294, 8, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[94552, 3318, 1248, 37, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[95837, 3360, 168, 5, 61, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[96008, 3367, 484, 10, 48, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[96495, 3379, 814, 21, 35, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[97312, 3402, 399, 13, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[97748, 3420, 453, 12, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[98204, 3434, 439, 13, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[98692, 3453, 95, 1, 47, "Subsubsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[98812, 3458, 257, 6, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[99072, 3466, 411, 12, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[99520, 3483, 168, 5, 61, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[99691, 3490, 484, 10, 48, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[100178, 3502, 836, 22, 35, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[101017, 3526, 415, 13, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[101469, 3544, 193, 5, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[101665, 3551, 518, 16, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}]},
 "b:37.0"->{
  Cell[5957, 164, 87, 1, 51, "Subsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.0"}],
  Cell[6047, 167, 289, 5, 53, "Text",
   CellTags->{"Autonomous Examples", "i:2", "b:37.0"}]},
 "b:37.1"->{
  Cell[6373, 177, 127, 1, 51, "Subsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[6525, 182, 84, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[6612, 185, 159, 4, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[6774, 191, 644, 22, 82, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[7443, 217, 216, 5, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[7662, 224, 126, 3, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[7837, 233, 93, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[7955, 238, 288, 8, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[8246, 248, 235, 7, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[8518, 260, 156, 5, 61, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[8677, 267, 158, 2, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[8838, 271, 163, 3, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[9004, 276, 489, 11, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[9496, 289, 163, 4, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[9696, 298, 448, 12, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[10147, 312, 522, 16, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[10718, 334, 91, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[10834, 339, 223, 5, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[11060, 346, 159, 2, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[11222, 350, 122, 3, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[11393, 359, 167, 2, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[11563, 363, 674, 20, 53, "Text",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[12262, 387, 559, 16, 82, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[12824, 405, 420, 15, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[13281, 425, 726, 20, 82, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[14010, 447, 980, 28, 52, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[14993, 477, 456, 15, 59, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[15452, 494, 492, 16, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[15959, 513, 189, 4, 37, "Text",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[16173, 521, 709, 24, 62, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[16885, 547, 493, 16, 85, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[17415, 568, 262, 7, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[17680, 577, 525, 17, 42, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[18242, 599, 270, 8, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[18515, 609, 363, 13, 59, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[18915, 627, 726, 20, 82, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[19644, 649, 980, 28, 52, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[20627, 679, 377, 12, 59, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
  Cell[21007, 693, 391, 13, 59, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.1"}]},
 "b:37.2"->{
  Cell[21459, 713, 118, 1, 51, "Subsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[21602, 718, 84, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[21689, 721, 587, 20, 62, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[22279, 743, 747, 25, 82, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[23051, 772, 217, 5, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[23271, 779, 150, 4, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[23470, 789, 93, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[23588, 794, 289, 8, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[23880, 804, 261, 8, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[24178, 817, 157, 5, 61, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[24338, 824, 490, 11, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[24831, 837, 163, 4, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[25031, 846, 448, 12, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[25482, 860, 699, 22, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[26230, 888, 91, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[26346, 893, 223, 5, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[26572, 900, 152, 2, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[26727, 904, 159, 2, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[26889, 908, 122, 3, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[27060, 917, 126, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[27189, 920, 692, 23, 62, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[27906, 947, 768, 25, 62, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[28677, 974, 153, 4, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[28833, 980, 108, 2, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[28978, 987, 927, 31, 102, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[29908, 1020, 363, 12, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[30308, 1037, 217, 5, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[30528, 1044, 150, 4, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[30727, 1054, 93, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[30845, 1059, 289, 8, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[31137, 1069, 316, 10, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[31490, 1084, 157, 5, 61, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[31650, 1091, 563, 15, 52, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[32216, 1108, 211, 6, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[32464, 1119, 448, 12, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[32915, 1133, 441, 14, 42, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[33405, 1153, 91, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[33521, 1158, 252, 6, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[33776, 1166, 393, 13, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[34206, 1184, 157, 5, 61, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[34366, 1191, 755, 21, 52, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[35124, 1214, 365, 12, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[35526, 1231, 448, 12, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
  Cell[35977, 1245, 546, 18, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.2"}]},
 "b:37.3"->{
  Cell[36584, 1270, 103, 1, 51, "Subsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[36712, 1275, 84, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[36821, 1280, 825, 27, 122, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[37649, 1309, 150, 4, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[37848, 1319, 93, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[37966, 1324, 289, 8, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[38258, 1334, 261, 8, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[38556, 1347, 157, 5, 61, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[38716, 1354, 490, 11, 35, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[39209, 1367, 163, 4, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[39409, 1376, 200, 6, 61, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[39612, 1384, 359, 11, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[39974, 1397, 188, 5, 39, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[40211, 1408, 91, 1, 47, "Subsubsection",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[40327, 1413, 252, 6, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[40582, 1421, 260, 8, 42, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[40879, 1434, 354, 10, 62, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[41236, 1446, 658, 18, 52, "Print",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[41897, 1466, 293, 10, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[42227, 1481, 448, 12, 41, "Input",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
  Cell[42678, 1495, 609, 21, 56, "Output",
   CellTags->{"Autonomous Examples", "i:2", "b:37.3"}]},
 "Failure of Rank Condition"->{
  Cell[43360, 1524, 123, 1, 77, "Section",
   CellTags->{"Failure of Rank Condition", "i:2"}],
  Cell[43508, 1529, 102, 1, 51, "Subsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[43635, 1534, 90, 1, 47, "Subsubsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[43728, 1537, 346, 6, 53, "Text",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[44077, 1545, 493, 16, 82, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[44595, 1565, 223, 5, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[44821, 1572, 133, 3, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[44991, 1580, 255, 6, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[45249, 1588, 254, 8, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[45540, 1601, 289, 7, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[45832, 1610, 128, 3, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[46009, 1619, 99, 1, 47, "Subsubsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[46133, 1624, 295, 8, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[46431, 1634, 453, 15, 42, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[46921, 1654, 160, 5, 61, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[47084, 1661, 1285, 39, 52, "Print",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[48372, 1702, 800, 29, 56, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[49209, 1736, 194, 5, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[49406, 1743, 396, 13, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[49851, 1762, 97, 1, 47, "Subsubsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[49973, 1767, 258, 6, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[50234, 1775, 364, 12, 42, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[50635, 1792, 160, 5, 61, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[50798, 1799, 1285, 39, 52, "Print",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[52086, 1840, 800, 29, 56, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[52923, 1874, 194, 5, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[53120, 1881, 372, 12, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[53565, 1901, 137, 3, 77, "Section",
   CellTags->{"Failure of Rank Condition", "i:2"}],
  Cell[53727, 1908, 102, 1, 51, "Subsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[53854, 1913, 90, 1, 47, "Subsubsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[53947, 1916, 192, 4, 37, "Text",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[54142, 1922, 535, 17, 102, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[54702, 1943, 223, 5, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[54928, 1950, 133, 3, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[55098, 1958, 312, 8, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[55413, 1968, 254, 8, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[55704, 1981, 346, 9, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[56053, 1992, 180, 5, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[56270, 2002, 311, 8, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[56584, 2012, 254, 8, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[56887, 2026, 99, 1, 47, "Subsubsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[57011, 2031, 295, 8, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[57309, 2041, 258, 8, 42, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[57604, 2054, 160, 5, 61, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[57767, 2061, 538, 14, 38, "Print",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[58308, 2077, 208, 6, 42, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[58553, 2088, 194, 5, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[58750, 2095, 317, 10, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[59116, 2111, 97, 1, 47, "Subsubsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[59238, 2116, 258, 6, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[59499, 2124, 203, 6, 42, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[59739, 2135, 160, 5, 61, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[59902, 2142, 562, 15, 38, "Print",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[60467, 2159, 229, 7, 42, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[60733, 2171, 194, 5, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[60930, 2178, 338, 11, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}]},
 "b:38.0"->{
  Cell[43508, 1529, 102, 1, 51, "Subsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[43635, 1534, 90, 1, 47, "Subsubsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[43728, 1537, 346, 6, 53, "Text",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[44077, 1545, 493, 16, 82, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[44595, 1565, 223, 5, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[44821, 1572, 133, 3, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[44991, 1580, 255, 6, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[45249, 1588, 254, 8, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[45540, 1601, 289, 7, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[45832, 1610, 128, 3, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[46009, 1619, 99, 1, 47, "Subsubsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[46133, 1624, 295, 8, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[46431, 1634, 453, 15, 42, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[46921, 1654, 160, 5, 61, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[47084, 1661, 1285, 39, 52, "Print",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[48372, 1702, 800, 29, 56, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[49209, 1736, 194, 5, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[49406, 1743, 396, 13, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[49851, 1762, 97, 1, 47, "Subsubsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[49973, 1767, 258, 6, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[50234, 1775, 364, 12, 42, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[50635, 1792, 160, 5, 61, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[50798, 1799, 1285, 39, 52, "Print",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[52086, 1840, 800, 29, 56, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[52923, 1874, 194, 5, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
  Cell[53120, 1881, 372, 12, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}]},
 "b:39.0"->{
  Cell[53727, 1908, 102, 1, 51, "Subsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[53854, 1913, 90, 1, 47, "Subsubsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[53947, 1916, 192, 4, 37, "Text",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[54142, 1922, 535, 17, 102, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[54702, 1943, 223, 5, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[54928, 1950, 133, 3, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[55098, 1958, 312, 8, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[55413, 1968, 254, 8, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[55704, 1981, 346, 9, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[56053, 1992, 180, 5, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[56270, 2002, 311, 8, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[56584, 2012, 254, 8, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[56887, 2026, 99, 1, 47, "Subsubsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[57011, 2031, 295, 8, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[57309, 2041, 258, 8, 42, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[57604, 2054, 160, 5, 61, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[57767, 2061, 538, 14, 38, "Print",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[58308, 2077, 208, 6, 42, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[58553, 2088, 194, 5, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[58750, 2095, 317, 10, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[59116, 2111, 97, 1, 47, "Subsubsection",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[59238, 2116, 258, 6, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[59499, 2124, 203, 6, 42, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[59739, 2135, 160, 5, 61, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[59902, 2142, 562, 15, 38, "Print",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[60467, 2159, 229, 7, 42, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[60733, 2171, 194, 5, 41, "Input",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
  Cell[60930, 2178, 338, 11, 39, "Output",
   CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}]},
 "Zero-input Unobservable"->{
  Cell[61341, 2197, 110, 1, 77, "Section",
   CellTags->{"Zero-input Unobservable", "i:2"}],
  Cell[61476, 2202, 90, 1, 51, "Subsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.0"}],
  Cell[61569, 2205, 231, 4, 37, "Text",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.0"}],
  Cell[61837, 2214, 101, 1, 51, "Subsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[61941, 2217, 238, 7, 101, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[62204, 2228, 246, 6, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[62453, 2236, 132, 3, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[62622, 2244, 302, 8, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[62927, 2254, 173, 3, 35, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[63103, 2259, 132, 3, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[63272, 2267, 349, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[63624, 2278, 342, 10, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[64003, 2293, 383, 10, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[64389, 2305, 342, 10, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[64768, 2320, 344, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[65115, 2331, 273, 8, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[65425, 2344, 97, 1, 47, "Subsubsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[65547, 2349, 319, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[65869, 2360, 1196, 35, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[67102, 2400, 162, 5, 61, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[67267, 2407, 484, 10, 48, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[67754, 2419, 754, 21, 35, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[68511, 2442, 374, 12, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[68922, 2459, 453, 12, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[69378, 2473, 345, 11, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[69760, 2489, 110, 2, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[69873, 2493, 294, 9, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[70216, 2508, 95, 1, 47, "Subsubsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[70336, 2513, 282, 7, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[70621, 2522, 370, 11, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[71006, 2536, 376, 11, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[71407, 2551, 162, 5, 61, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[71572, 2558, 484, 10, 48, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[72059, 2570, 754, 21, 35, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[72816, 2593, 374, 12, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[73227, 2610, 193, 5, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[73423, 2617, 384, 12, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[73844, 2634, 110, 2, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[73957, 2638, 314, 10, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[74332, 2655, 111, 1, 51, "Subsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[74446, 2658, 1598, 50, 102, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[76069, 2712, 246, 6, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[76318, 2720, 132, 3, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[76487, 2728, 302, 8, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[76792, 2738, 173, 3, 35, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[76968, 2743, 132, 3, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[77137, 2751, 325, 8, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[77465, 2761, 451, 12, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[77953, 2778, 359, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[78315, 2789, 451, 12, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[78803, 2806, 333, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[79139, 2817, 293, 8, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[79469, 2830, 97, 1, 47, "Subsubsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[79591, 2835, 319, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[79913, 2846, 2720, 80, 86, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[82670, 2931, 165, 5, 61, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[82838, 2938, 525, 10, 51, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[83366, 2950, 484, 10, 48, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[83853, 2962, 793, 21, 38, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[84649, 2985, 402, 12, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[85088, 3002, 453, 12, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[85544, 3016, 335, 9, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[85928, 3031, 95, 1, 47, "Subsubsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[86048, 3036, 282, 7, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[86333, 3045, 423, 13, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[86793, 3063, 165, 5, 61, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[86961, 3070, 525, 10, 51, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[87489, 3082, 484, 10, 48, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[87976, 3094, 793, 21, 38, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[88772, 3117, 402, 12, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[89211, 3134, 184, 4, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[89398, 3140, 409, 13, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[89868, 3160, 107, 1, 51, "Subsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[89978, 3163, 256, 7, 101, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[90259, 3174, 222, 5, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[90484, 3181, 155, 4, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[90676, 3190, 278, 7, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[90957, 3199, 173, 3, 35, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[91133, 3204, 155, 4, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[91325, 3213, 325, 8, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[91653, 3223, 583, 15, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[92273, 3243, 359, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[92635, 3254, 583, 15, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[93255, 3274, 344, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[93602, 3285, 494, 13, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[94133, 3303, 97, 1, 47, "Subsubsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[94255, 3308, 294, 8, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[94552, 3318, 1248, 37, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[95837, 3360, 168, 5, 61, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[96008, 3367, 484, 10, 48, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[96495, 3379, 814, 21, 35, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[97312, 3402, 399, 13, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[97748, 3420, 453, 12, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[98204, 3434, 439, 13, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[98692, 3453, 95, 1, 47, "Subsubsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[98812, 3458, 257, 6, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[99072, 3466, 411, 12, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[99520, 3483, 168, 5, 61, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[99691, 3490, 484, 10, 48, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[100178, 3502, 836, 22, 35, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[101017, 3526, 415, 13, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[101469, 3544, 193, 5, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[101665, 3551, 518, 16, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}]},
 "b:40.0"->{
  Cell[61476, 2202, 90, 1, 51, "Subsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.0"}],
  Cell[61569, 2205, 231, 4, 37, "Text",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.0"}]},
 "b:40.1"->{
  Cell[61837, 2214, 101, 1, 51, "Subsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[61941, 2217, 238, 7, 101, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[62204, 2228, 246, 6, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[62453, 2236, 132, 3, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[62622, 2244, 302, 8, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[62927, 2254, 173, 3, 35, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[63103, 2259, 132, 3, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[63272, 2267, 349, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[63624, 2278, 342, 10, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[64003, 2293, 383, 10, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[64389, 2305, 342, 10, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[64768, 2320, 344, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[65115, 2331, 273, 8, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[65425, 2344, 97, 1, 47, "Subsubsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[65547, 2349, 319, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[65869, 2360, 1196, 35, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[67102, 2400, 162, 5, 61, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[67267, 2407, 484, 10, 48, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[67754, 2419, 754, 21, 35, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[68511, 2442, 374, 12, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[68922, 2459, 453, 12, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[69378, 2473, 345, 11, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[69760, 2489, 110, 2, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[69873, 2493, 294, 9, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[70216, 2508, 95, 1, 47, "Subsubsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[70336, 2513, 282, 7, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[70621, 2522, 370, 11, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[71006, 2536, 376, 11, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[71407, 2551, 162, 5, 61, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[71572, 2558, 484, 10, 48, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[72059, 2570, 754, 21, 35, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[72816, 2593, 374, 12, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[73227, 2610, 193, 5, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[73423, 2617, 384, 12, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[73844, 2634, 110, 2, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
  Cell[73957, 2638, 314, 10, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}]},
 "b:40.2"->{
  Cell[74332, 2655, 111, 1, 51, "Subsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[74446, 2658, 1598, 50, 102, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[76069, 2712, 246, 6, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[76318, 2720, 132, 3, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[76487, 2728, 302, 8, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[76792, 2738, 173, 3, 35, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[76968, 2743, 132, 3, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[77137, 2751, 325, 8, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[77465, 2761, 451, 12, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[77953, 2778, 359, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[78315, 2789, 451, 12, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[78803, 2806, 333, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[79139, 2817, 293, 8, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[79469, 2830, 97, 1, 47, "Subsubsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[79591, 2835, 319, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[79913, 2846, 2720, 80, 86, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[82670, 2931, 165, 5, 61, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[82838, 2938, 525, 10, 51, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[83366, 2950, 484, 10, 48, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[83853, 2962, 793, 21, 38, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[84649, 2985, 402, 12, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[85088, 3002, 453, 12, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[85544, 3016, 335, 9, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[85928, 3031, 95, 1, 47, "Subsubsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[86048, 3036, 282, 7, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[86333, 3045, 423, 13, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[86793, 3063, 165, 5, 61, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[86961, 3070, 525, 10, 51, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[87489, 3082, 484, 10, 48, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[87976, 3094, 793, 21, 38, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[88772, 3117, 402, 12, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[89211, 3134, 184, 4, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
  Cell[89398, 3140, 409, 13, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}]},
 "b:40.3"->{
  Cell[89868, 3160, 107, 1, 51, "Subsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[89978, 3163, 256, 7, 101, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[90259, 3174, 222, 5, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[90484, 3181, 155, 4, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[90676, 3190, 278, 7, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[90957, 3199, 173, 3, 35, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[91133, 3204, 155, 4, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[91325, 3213, 325, 8, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[91653, 3223, 583, 15, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[92273, 3243, 359, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[92635, 3254, 583, 15, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[93255, 3274, 344, 9, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[93602, 3285, 494, 13, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[94133, 3303, 97, 1, 47, "Subsubsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[94255, 3308, 294, 8, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[94552, 3318, 1248, 37, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[95837, 3360, 168, 5, 61, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[96008, 3367, 484, 10, 48, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[96495, 3379, 814, 21, 35, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[97312, 3402, 399, 13, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[97748, 3420, 453, 12, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[98204, 3434, 439, 13, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[98692, 3453, 95, 1, 47, "Subsubsection",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[98812, 3458, 257, 6, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[99072, 3466, 411, 12, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[99520, 3483, 168, 5, 61, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[99691, 3490, 484, 10, 48, "Message",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[100178, 3502, 836, 22, 35, "Print",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[101017, 3526, 415, 13, 39, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[101469, 3544, 193, 5, 41, "Input",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
  Cell[101665, 3551, 518, 16, 42, "Output",
   CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}]},
 "Bibliography"->{
  Cell[102256, 3575, 68, 1, 77, "Section",
   CellTags->{"Bibliography", "b:41"}],
  Cell[102327, 3578, 1308, 23, 309, "Text",
   CellTags->{"Bibliography", "b:41"}]},
 "b:41"->{
  Cell[102256, 3575, 68, 1, 77, "Section",
   CellTags->{"Bibliography", "b:41"}],
  Cell[102327, 3578, 1308, 23, 309, "Text",
   CellTags->{"Bibliography", "b:41"}]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"Introduction", 106632, 3661},
 {"b:36", 107373, 3678},
 {"i:1", 108113, 3695},
 {"Observer Tools", 108786, 3710},
 {"Autonomous Examples", 109124, 3717},
 {"i:2", 119613, 3932},
 {"b:37.0", 146505, 4465},
 {"b:37.1", 146712, 4470},
 {"b:37.2", 150519, 4549},
 {"b:37.3", 154858, 4638},
 {"Failure of Rank Condition", 156971, 4681},
 {"b:38.0", 162823, 4794},
 {"b:39.0", 165560, 4847},
 {"Zero-input Unobservable", 168516, 4904},
 {"b:40.0", 179092, 5111},
 {"b:40.1", 179311, 5116},
 {"b:40.2", 183015, 5189},
 {"b:40.3", 186418, 5256},
 {"Bibliography", 189623, 5319},
 {"b:41", 189802, 5324}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[567, 22, 68, 1, 77, "Section",
 CellTags->{"Introduction", "b:36"}],
Cell[638, 25, 607, 10, 109, "Text",
 CellTags->{"Introduction", "i:1", "b:36"}],
Cell[1248, 37, 83, 1, 37, "Text",
 CellTags->{"Introduction", "i:1", "b:36"}],
Cell[CellGroupData[{
Cell[1356, 42, 118, 3, 41, "Input",
 CellTags->{"Introduction", "i:1", "b:36"}],
Cell[1477, 47, 751, 10, 416, "Print",
 CellTags->{"Introduction", "i:1", "b:36"}]
}, Open  ]],
Cell[2243, 60, 135, 1, 37, "Text",
 CellTags->{"Observer Tools", "Introduction", "i:1", "b:36"}],
Cell[CellGroupData[{
Cell[2403, 65, 150, 3, 41, "Input",
 CellTags->{"Observer Tools", "Introduction", "i:1", "b:36"}],
Cell[2556, 70, 3225, 83, 102, "Print",
 CellTags->{"Observer Tools", "Introduction", "i:1", "b:36"}]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5830, 159, 102, 1, 77, "Section",
 CellTags->{"Autonomous Examples", "i:2"}],
Cell[CellGroupData[{
Cell[5957, 164, 87, 1, 51, "Subsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.0"}],
Cell[6047, 167, 289, 5, 53, "Text",
 CellTags->{"Autonomous Examples", "i:2", "b:37.0"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[6373, 177, 127, 1, 51, "Subsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[CellGroupData[{
Cell[6525, 182, 84, 1, 47, "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[6612, 185, 159, 4, 41, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[6774, 191, 644, 22, 82, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[CellGroupData[{
Cell[7443, 217, 216, 5, 41, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[7662, 224, 126, 3, 39, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7837, 233, 93, 1, 47, "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[CellGroupData[{
Cell[7955, 238, 288, 8, 41, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[8246, 248, 235, 7, 39, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[8518, 260, 156, 5, 61, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[8677, 267, 158, 2, 35, "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[8838, 271, 163, 3, 35, "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[9004, 276, 489, 11, 35, "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[9496, 289, 163, 4, 39, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[9696, 298, 448, 12, 41, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[10147, 312, 522, 16, 56, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[10718, 334, 91, 1, 47, "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[CellGroupData[{
Cell[10834, 339, 223, 5, 41, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[11060, 346, 159, 2, 35, "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[11222, 350, 122, 3, 39, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[11393, 359, 167, 2, 47, "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[11563, 363, 674, 20, 53, "Text",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[CellGroupData[{
Cell[12262, 387, 559, 16, 82, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[12824, 405, 420, 15, 56, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[13281, 425, 726, 20, 82, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[14010, 447, 980, 28, 52, "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[14993, 477, 456, 15, 59, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[15452, 494, 492, 16, 56, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}]
}, Open  ]],
Cell[15959, 513, 189, 4, 37, "Text",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[CellGroupData[{
Cell[16173, 521, 709, 24, 62, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[16885, 547, 493, 16, 85, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[17415, 568, 262, 7, 41, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[17680, 577, 525, 17, 42, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[18242, 599, 270, 8, 41, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[18515, 609, 363, 13, 59, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[18915, 627, 726, 20, 82, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[19644, 649, 980, 28, 52, "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[20627, 679, 377, 12, 59, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}],
Cell[21007, 693, 391, 13, 59, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.1"}]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[21459, 713, 118, 1, 51, "Subsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[CellGroupData[{
Cell[21602, 718, 84, 1, 47, "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[21689, 721, 587, 20, 62, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[22279, 743, 747, 25, 82, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[CellGroupData[{
Cell[23051, 772, 217, 5, 41, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[23271, 779, 150, 4, 39, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[23470, 789, 93, 1, 47, "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[CellGroupData[{
Cell[23588, 794, 289, 8, 41, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[23880, 804, 261, 8, 39, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[24178, 817, 157, 5, 61, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[24338, 824, 490, 11, 35, "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[24831, 837, 163, 4, 39, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[25031, 846, 448, 12, 41, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[25482, 860, 699, 22, 56, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[26230, 888, 91, 1, 47, "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[CellGroupData[{
Cell[26346, 893, 223, 5, 41, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[26572, 900, 152, 2, 35, "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[26727, 904, 159, 2, 35, "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[26889, 908, 122, 3, 39, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[27060, 917, 126, 1, 47, "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[27189, 920, 692, 23, 62, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[CellGroupData[{
Cell[27906, 947, 768, 25, 62, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[28677, 974, 153, 4, 39, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[28833, 980, 108, 2, 39, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[28978, 987, 927, 31, 102, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[29908, 1020, 363, 12, 56, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[30308, 1037, 217, 5, 41, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[30528, 1044, 150, 4, 39, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[30727, 1054, 93, 1, 47, "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[CellGroupData[{
Cell[30845, 1059, 289, 8, 41, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[31137, 1069, 316, 10, 39, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[31490, 1084, 157, 5, 61, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[31650, 1091, 563, 15, 52, "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[32216, 1108, 211, 6, 56, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[32464, 1119, 448, 12, 41, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[32915, 1133, 441, 14, 42, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[33405, 1153, 91, 1, 47, "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[CellGroupData[{
Cell[33521, 1158, 252, 6, 41, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[33776, 1166, 393, 13, 56, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[34206, 1184, 157, 5, 61, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[34366, 1191, 755, 21, 52, "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[35124, 1214, 365, 12, 56, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[35526, 1231, 448, 12, 41, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}],
Cell[35977, 1245, 546, 18, 56, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.2"}]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[36584, 1270, 103, 1, 51, "Subsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
Cell[CellGroupData[{
Cell[36712, 1275, 84, 1, 47, "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
Cell[CellGroupData[{
Cell[36821, 1280, 825, 27, 122, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
Cell[37649, 1309, 150, 4, 39, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[37848, 1319, 93, 1, 47, "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
Cell[CellGroupData[{
Cell[37966, 1324, 289, 8, 41, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
Cell[38258, 1334, 261, 8, 39, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[38556, 1347, 157, 5, 61, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
Cell[38716, 1354, 490, 11, 35, "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
Cell[39209, 1367, 163, 4, 39, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[39409, 1376, 200, 6, 61, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
Cell[39612, 1384, 359, 11, 39, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
Cell[39974, 1397, 188, 5, 39, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[40211, 1408, 91, 1, 47, "Subsubsection",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
Cell[CellGroupData[{
Cell[40327, 1413, 252, 6, 41, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
Cell[40582, 1421, 260, 8, 42, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[40879, 1434, 354, 10, 62, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
Cell[41236, 1446, 658, 18, 52, "Print",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
Cell[41897, 1466, 293, 10, 56, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[42227, 1481, 448, 12, 41, "Input",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}],
Cell[42678, 1495, 609, 21, 56, "Output",
 CellTags->{"Autonomous Examples", "i:2", "b:37.3"}]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[43360, 1524, 123, 1, 77, "Section",
 CellTags->{"Failure of Rank Condition", "i:2"}],
Cell[CellGroupData[{
Cell[43508, 1529, 102, 1, 51, "Subsection",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
Cell[CellGroupData[{
Cell[43635, 1534, 90, 1, 47, "Subsubsection",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
Cell[43728, 1537, 346, 6, 53, "Text",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
Cell[44077, 1545, 493, 16, 82, "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
Cell[CellGroupData[{
Cell[44595, 1565, 223, 5, 41, "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
Cell[44821, 1572, 133, 3, 39, "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[44991, 1580, 255, 6, 41, "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
Cell[45249, 1588, 254, 8, 39, "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[45540, 1601, 289, 7, 41, "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
Cell[45832, 1610, 128, 3, 39, "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[46009, 1619, 99, 1, 47, "Subsubsection",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
Cell[CellGroupData[{
Cell[46133, 1624, 295, 8, 41, "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
Cell[46431, 1634, 453, 15, 42, "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[46921, 1654, 160, 5, 61, "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
Cell[47084, 1661, 1285, 39, 52, "Print",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
Cell[48372, 1702, 800, 29, 56, "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[49209, 1736, 194, 5, 41, "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
Cell[49406, 1743, 396, 13, 39, "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[49851, 1762, 97, 1, 47, "Subsubsection",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
Cell[CellGroupData[{
Cell[49973, 1767, 258, 6, 41, "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
Cell[50234, 1775, 364, 12, 42, "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[50635, 1792, 160, 5, 61, "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
Cell[50798, 1799, 1285, 39, 52, "Print",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
Cell[52086, 1840, 800, 29, 56, "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[52923, 1874, 194, 5, 41, "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}],
Cell[53120, 1881, 372, 12, 39, "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:38.0"}]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[53565, 1901, 137, 3, 77, "Section",
 CellTags->{"Failure of Rank Condition", "i:2"}],
Cell[CellGroupData[{
Cell[53727, 1908, 102, 1, 51, "Subsection",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
Cell[CellGroupData[{
Cell[53854, 1913, 90, 1, 47, "Subsubsection",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
Cell[53947, 1916, 192, 4, 37, "Text",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
Cell[54142, 1922, 535, 17, 102, "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
Cell[CellGroupData[{
Cell[54702, 1943, 223, 5, 41, "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
Cell[54928, 1950, 133, 3, 39, "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[55098, 1958, 312, 8, 41, "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
Cell[55413, 1968, 254, 8, 39, "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[55704, 1981, 346, 9, 41, "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
Cell[56053, 1992, 180, 5, 39, "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[56270, 2002, 311, 8, 41, "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
Cell[56584, 2012, 254, 8, 39, "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[56887, 2026, 99, 1, 47, "Subsubsection",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
Cell[CellGroupData[{
Cell[57011, 2031, 295, 8, 41, "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
Cell[57309, 2041, 258, 8, 42, "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[57604, 2054, 160, 5, 61, "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
Cell[57767, 2061, 538, 14, 38, "Print",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
Cell[58308, 2077, 208, 6, 42, "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[58553, 2088, 194, 5, 41, "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
Cell[58750, 2095, 317, 10, 39, "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[59116, 2111, 97, 1, 47, "Subsubsection",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
Cell[CellGroupData[{
Cell[59238, 2116, 258, 6, 41, "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
Cell[59499, 2124, 203, 6, 42, "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[59739, 2135, 160, 5, 61, "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
Cell[59902, 2142, 562, 15, 38, "Print",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
Cell[60467, 2159, 229, 7, 42, "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[60733, 2171, 194, 5, 41, "Input",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}],
Cell[60930, 2178, 338, 11, 39, "Output",
 CellTags->{"Failure of Rank Condition", "i:2", "b:39.0"}]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[61341, 2197, 110, 1, 77, "Section",
 CellTags->{"Zero-input Unobservable", "i:2"}],
Cell[CellGroupData[{
Cell[61476, 2202, 90, 1, 51, "Subsection",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.0"}],
Cell[61569, 2205, 231, 4, 37, "Text",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.0"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[61837, 2214, 101, 1, 51, "Subsection",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[61941, 2217, 238, 7, 101, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[CellGroupData[{
Cell[62204, 2228, 246, 6, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[62453, 2236, 132, 3, 39, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[62622, 2244, 302, 8, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[62927, 2254, 173, 3, 35, "Print",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[63103, 2259, 132, 3, 39, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[63272, 2267, 349, 9, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[63624, 2278, 342, 10, 39, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[64003, 2293, 383, 10, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[64389, 2305, 342, 10, 39, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[64768, 2320, 344, 9, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[65115, 2331, 273, 8, 39, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[65425, 2344, 97, 1, 47, "Subsubsection",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[CellGroupData[{
Cell[65547, 2349, 319, 9, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[65869, 2360, 1196, 35, 42, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[67102, 2400, 162, 5, 61, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[67267, 2407, 484, 10, 48, "Message",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[67754, 2419, 754, 21, 35, "Print",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[68511, 2442, 374, 12, 39, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[68922, 2459, 453, 12, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[69378, 2473, 345, 11, 42, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[69760, 2489, 110, 2, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[69873, 2493, 294, 9, 42, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[70216, 2508, 95, 1, 47, "Subsubsection",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[CellGroupData[{
Cell[70336, 2513, 282, 7, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[70621, 2522, 370, 11, 42, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}]
}, Open  ]],
Cell[71006, 2536, 376, 11, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[CellGroupData[{
Cell[71407, 2551, 162, 5, 61, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[71572, 2558, 484, 10, 48, "Message",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[72059, 2570, 754, 21, 35, "Print",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[72816, 2593, 374, 12, 39, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[73227, 2610, 193, 5, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[73423, 2617, 384, 12, 42, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[73844, 2634, 110, 2, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}],
Cell[73957, 2638, 314, 10, 42, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.1"}]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[74332, 2655, 111, 1, 51, "Subsection",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[74446, 2658, 1598, 50, 102, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[CellGroupData[{
Cell[76069, 2712, 246, 6, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[76318, 2720, 132, 3, 39, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[76487, 2728, 302, 8, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[76792, 2738, 173, 3, 35, "Print",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[76968, 2743, 132, 3, 39, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[77137, 2751, 325, 8, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[77465, 2761, 451, 12, 39, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[77953, 2778, 359, 9, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[78315, 2789, 451, 12, 39, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[78803, 2806, 333, 9, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[79139, 2817, 293, 8, 39, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[79469, 2830, 97, 1, 47, "Subsubsection",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[CellGroupData[{
Cell[79591, 2835, 319, 9, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[79913, 2846, 2720, 80, 86, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[82670, 2931, 165, 5, 61, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[82838, 2938, 525, 10, 51, "Message",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[83366, 2950, 484, 10, 48, "Message",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[83853, 2962, 793, 21, 38, "Print",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[84649, 2985, 402, 12, 42, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[85088, 3002, 453, 12, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[85544, 3016, 335, 9, 42, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[85928, 3031, 95, 1, 47, "Subsubsection",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[CellGroupData[{
Cell[86048, 3036, 282, 7, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[86333, 3045, 423, 13, 42, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[86793, 3063, 165, 5, 61, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[86961, 3070, 525, 10, 51, "Message",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[87489, 3082, 484, 10, 48, "Message",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[87976, 3094, 793, 21, 38, "Print",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[88772, 3117, 402, 12, 42, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[89211, 3134, 184, 4, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}],
Cell[89398, 3140, 409, 13, 42, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.2"}]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[89868, 3160, 107, 1, 51, "Subsection",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
Cell[89978, 3163, 256, 7, 101, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
Cell[CellGroupData[{
Cell[90259, 3174, 222, 5, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
Cell[90484, 3181, 155, 4, 39, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[90676, 3190, 278, 7, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
Cell[90957, 3199, 173, 3, 35, "Print",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
Cell[91133, 3204, 155, 4, 39, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[91325, 3213, 325, 8, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
Cell[91653, 3223, 583, 15, 39, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[92273, 3243, 359, 9, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
Cell[92635, 3254, 583, 15, 39, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[93255, 3274, 344, 9, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
Cell[93602, 3285, 494, 13, 39, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[94133, 3303, 97, 1, 47, "Subsubsection",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
Cell[CellGroupData[{
Cell[94255, 3308, 294, 8, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
Cell[94552, 3318, 1248, 37, 42, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[95837, 3360, 168, 5, 61, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
Cell[96008, 3367, 484, 10, 48, "Message",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
Cell[96495, 3379, 814, 21, 35, "Print",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
Cell[97312, 3402, 399, 13, 39, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[97748, 3420, 453, 12, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
Cell[98204, 3434, 439, 13, 42, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[98692, 3453, 95, 1, 47, "Subsubsection",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
Cell[CellGroupData[{
Cell[98812, 3458, 257, 6, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
Cell[99072, 3466, 411, 12, 42, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[99520, 3483, 168, 5, 61, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
Cell[99691, 3490, 484, 10, 48, "Message",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
Cell[100178, 3502, 836, 22, 35, "Print",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
Cell[101017, 3526, 415, 13, 39, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}]
}, Open  ]],
Cell[CellGroupData[{
Cell[101469, 3544, 193, 5, 41, "Input",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}],
Cell[101665, 3551, 518, 16, 42, "Output",
 CellTags->{"Zero-input Unobservable", "i:2", "b:40.3"}]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[102256, 3575, 68, 1, 77, "Section",
 CellTags->{"Bibliography", "b:41"}],
Cell[102327, 3578, 1308, 23, 309, "Text",
 CellTags->{"Bibliography", "b:41"}],
Cell[103638, 3603, 165, 2, 21, "Print"],
Cell[103806, 3607, 172, 3, 21, "Print"]
}, Open  ]]
}
]
*)

(* End of internal cache information *)

